/*!
 * @file ltp.h
 * @brief This is the external interface for our transport protocol
 * @version $Revision: 1.4 $ $Author: stephen $
 * @version Copyright (C) 2006-2012 Trinity College Dublin
 * @version Copyright (C) 2013-2014 Tolerant Networks Limited
 * My LTP is based upon RFC 5325, 5326 and 5327 of course:-)
 * This API is modelled on the sockets API as a kind of mixture between connection 
 * and connectionless functions, see e.g.:
 * http://www.fifi.org/cgi-bin/man2html/usr/share/man/man7/udp.7.gz
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef LTP_H
#define LTP_H

#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/*!
 * @struct ltpaddr is either a sockaddr or else an octet string
 * @param sock is the sockaddr option
 * @param magic is a fixed value 0xff which is illegal in a sockaddr
 * @param len is the number of octets in the value
 * @param val is the octet string
 */
typedef  union {
	struct sockaddr_in 	sock;
	struct {
		unsigned char		magic;
		unsigned int 		len;
		unsigned char  		*val;
	}					octets;
} ltpaddr; 

/// this is the value in the magic field 
#define LTP_ADDR_OCT_MAGIC 0xff

#define LTP_BUILD_VERSION "$Revision: 1.32 $"
#define LTP_BUILD_DATE __DATE__
#define LTP_BUILD_TIME __TIME__

/// sizing of those
typedef size_t ltpaddr_len; 

/// A "special" LTP address - the LTP layer loopback
extern ltpaddr ltplb;

/// the length to use in calls with the loopback address (doesn't matter much, but worth doing!)
#define LTPLBLEN	sizeof(ltplb)

/// Another "special" address - "any" (equivalent to broadcast IP)
extern ltpaddr ltpany;

/// The default sin_port to use at the LTP layer, which is mapped to the LTP client service ID
/// (and not the UDP port!) Value is irrelevant here for now.
#define UDP_DEF_PORT 1113

/// the length to use in calls with the "any" address (doesn't matter much, but worth doing!)
#define LTPANYLEN	sizeof(ltpany)

/// AF_LTP is our equivalent to AF_INET - 1st parameter to ltp_socket()
#define AF_LTP 26 ///< currently 1-23 are used and MAX is 32 (on cygwin anyway)

/// SOCK_LTP_SESSION - 2nd parameter to ltp_socket()
#define SOCK_LTP_SESSION 16 ///< 1-6 currently used, no idea about max

/// Protocols - 3rd parameter to ltp_socket()
#define LTP_FARRELL	0 ///< the default protocol for sockets over LTP
#define LTP_BUNDLING	1 ///< when the bundling protocol works do that then

/// red/green socket options - otherwise the value is the length of the red part (order & values are important here!)
#define LTP_REDTIDE_INIT	-2				///< this is the default for the red level actually seen 
#define LTP_ALLRED			-1				///< set this (which is the default) to make the entire block red
#define LTP_ALLGREEN		0				///< set this (which is the default) to make the entire block greed
#define LTP_MAXRED			0x1000000		///< if not LTP_ALLRED or LTP_ALLGREEN the value must less than this (2^24)

/// ciphersuite options
#define LTP_CS_NONE		254					///< no crypto (default)
#define LTP_CS_MAC		0					///< OriginAuth ciphersuite
#define LTP_CS_SIG		1					///< Signature ciphersuite
#define LTP_CS_NULL		255					///< NULL ciphersuite

#ifndef MAXPATH
#define MAXPATH 1024
#endif

/// longest KeyID allowed for now
#define LTP_CS_MAX_KID_LEN					MAXPATH

/// layer 2 MTU range for the sockopt() call - currently this is not maintained in the state through a reboot
/// the logic for that being that the MTU is more likely to vary based on the environment (and I can't be
/// bothered re-segmenting yet in any case - the segments are each separately in the state)
#define LTP_MINMTU		68			/// the smallest layer 2 MTU we allow the application to set (RFC 791) - TC of TM/TC is 200 up
#define LTP_MAXMTU		0x100000 	/// the largest layer 2 MTU we allow the application to set - I've no idea why this value

/*!
 * @struct Generic authentication structure for sockopt
 * @param ciphersuite is an LTP_CS_* value specifying which ciphersuite to use
 * @param keyid is the length of KeyID value 
 * @param keyid is the KeyID value as a (often a non-null-termindated string)
 */
typedef struct ltp_auth_so_defn {
	int					ciphersuite;
	int					keyidlen;
	unsigned char		keyid[LTP_CS_MAX_KID_LEN];
} ltp_auth_so;

/// the amount of space we allow for authentication results
#define LTP_AUTHBUFSIZE 1024

/// value to indicate omission of an LTP-T port
#define LTP_T_NO_PORT -1

/// max LTP-T port
#define LTP_T_MAX_PORT 0x10000

/// value to indicate omission of an LTP-T hopcount
#define LTP_T_NO_HOPCOUNT -1

/// max hop count - it is *the* answer after all:-)
#define LTP_T_MAX_HOPCOUNT 42

/// default # relay sockets
#define LTP_T_DEF_RELAY_SOCKETS 5

/// max # relay sockets
#define LTP_T_MAX_RELAY_SOCKETS 50

/// crypto map file for e2e stuff
#define LTP_T_DEF_MAP_FILE "ltpd.crypt"

/*!
 * @struct an LTP-T configuration
 * @param on specifies whether LTP-T is on or off for this socket (default off)
 * @param relay_sockets specifies how many relay_sockets to setup (detault 5)
 * @param include_ebs specifies whether or not to include an estimated block size extension (omitted by default)
 * @param port is the LTP-T port to use (omitted by default)
 * @param hopcount specifies the hopcount to use (omitted by default)
 * @param incude_via specifies whether or not to include a via extension (omitted by default)
 * @param e2e_int specifies the end-to-end authenticatoin settings (see ltp_auth_so definition)
 * @param do_ecn specifies whether or not to inform peers about congestion (default is to be quiet)
 * @param need_e2e_int specifies whether a receiver insists on verifying an e2e ltp auth (true) or not (false)
 * @param map_file contains the map file
 */
typedef struct ltp_t_cfg_so_defn {
	bool				on; 
	int					relay_sockets;
	bool				include_est;
	int					port;
	int					hopcount;
	bool				include_via;
	ltp_auth_so			e2e_int;
	bool				do_ecn;
	bool				need_e2e_int;
	char				map_file[MAXPATH];
} ltp_t_cfg_so;

/// the length of an LTP-T route string
#define LTP_T_ROUTE_SLEN 	MAXPATH

/*!
 * @struct a single LTP-T route
 * @param network is a whitespace-ignored, comma-sep string representing the networks, e.g. "10.18/16,10.19.25/24"
 * @param gateway is a dot-sep IP address string [with optional port], e.g. "10.18.0.1" "10.18.1.0:1113"
 */
typedef struct ltp_t_route_so_defn {
	char 				network[LTP_T_ROUTE_SLEN];
	char				gateway[LTP_T_ROUTE_SLEN];
	// later on we may add an optiona route configuration file name (contents like an LTP-T config)
} ltp_t_route_so;

// --- stuff about state-of-play

///< MAX #blocks/socket we can provide SOP for
#define MAXBINFOS 10

///< MAX #scopes per blocl we can provide info for (we just stop if there're more)
#define MAXSCOPES 100

/*!
 * @struct information about a block
 * @param outbound is whether or not this is an outbound block
 * @param complete is whether or not LTP processing of this block is complete
 * @param radone is whether or not a receiver has the last outstanding RA for a red block
 * @param bidlen is the length of the block identifier
 * @param bid is the block identifier (LTP session no)
 * @param redlen is the length of the red part (seen so far; -1 = all; 0=all green)
 * @param nscopes is the number of scopes so far
 * @param lbs is the set of scope lower bounds
 * @param ubs is the set of scope upper bounds
 *
 * Note all ptrs in this stuct are to statically allocated buffers that may be
 * reallocated (i.e. moved) or freed (e.g. if the socket is closed). 
 * 
 * Take care! TODO: make these reliably static
 */
typedef struct ltp_binfo_defn {
	bool				inbound;
	bool				outbound;
	bool				complete;
	bool 				radone;
	int					bidlen;
	unsigned char		*bid;
	int					redlen;
	int					nscopes;
	int					*lbs;
	int					*ubs;
} ltp_binfo;

/*!
 * @struct LTP state-of-plau (sop) definintion
 * @param nblocks is the number of blocks in play under this socket (other fields are per-block)
 * @param completes say whether the block is complete or not
 *
 * We're exposing LTPlib internal memory with this. Not good. But I guess one could
 * write a bit better glue sometime.
 */
typedef struct ltp_sop_defn {
	int					nblocks;
	ltp_binfo			*infos;
} ltp_sop;

/// LTP sockopts
#define LTP_SO_LINGER		0				///< identifies the sockopt() as the linger option
#define LTP_SO_INACTIVITY	1				///< identifies the sockopt() as the inactivity timer option
#define LTP_SO_COOKIE		2				///< requests/reports cookies are on for this socket
#define LTP_SO_RED			3				///< set/get red/green-ness for the socket
#define LTP_SO_COOKIE_GRACE	4				///< set/get grace period during which we don't insist on seeing cookies
#define LTP_SO_AUTH			5				///< set/get LTP auth parameters to use - parameter is an ltp_auth_so
#define LTP_SO_CLIENTONLY	6				///< turn off "layer 2" listening
#define LTP_SO_L2MTU		7				///< set the l2 MTU from the application, ignoring lower layers
#define LTP_SO_T_CFG		8				///< feed in an LTP-T configuration
#define LTP_SO_T_ROUTE		9				///< feed in an LTP-T route
#define LTP_SO_T_AUTH		10				///< recipient indicates it requires good LTP-T e2e auth
#define LTP_SO_BLOCK		11				///< calls to sendto will block until the LTP session has succeeded
#define LTP_SO_SOP			12				///< get will return state of play for the socket (ltp_sop struct)
#define LTP_SO_FORCE		13				///< force responder address (ltpaddr struct)
#define LTP_SO_ION			14				///< force identifiers to be short, suitable for interop with ION
#define LTP_MAX_SO			14				///< used to ensure sockopt is sensible
#define LTP_MAX_OPTLEN		4096			///< largest optlen input on a setsockopt


/// Number of peers that we're able (by default) to handle transmitting to in 
/// a single "ping"
#define LTP_LISTENERS_PER_GO 10

/// Number of peers that we're able (by default) to handle listening to in 
/// a single "ping"
#define LTP_TALKERS_PER_GO 10

/*!
 * @fn ltp_socket(int domain, int type, int protocol)
 * @brief create (for originator) a new LTP session (analagous to the socket() function)
 * @param domain has to be AF_LTP
 * @param type has to be SOCK_LTP_SESSION
 * @param protocol has to be LTP_BUNDLING (not here though) or LTP_FARRELL (all that works here!)
 * @return is the socket number for use in later calls (e.g. to ltp_sendto())
 */
int ltp_socket(int domain, int type, int protocol);

/*!
 * @fn ltp_bind(int sock, ltpaddr *my_addr, ltpaddr_len addrlen); 
 * @brief bind my address to a socket (for a server)
 * @param sock is the socket
 * @param my_addr is the my address
 * @param addrlen is the length of my address
 * @return is zero for success and -1 for error
 */
int ltp_bind(int sock, ltpaddr *my_addr, ltpaddr_len addrlen); 

/*!
 * @fn ltp_sendto(int sock, const void *msg, size_t len, int flags, ltpaddr *to, ltpaddr_len tolen)
 * @brief sendto equivalent for LTP
 * @param sock is a destination LTP "session" (like a socket)
 * @param msg are the bytes to send
 * @param len is the number of bytes
 * @param flags contains flags for the LTP session
 * @param to is the destination address
 * @param tolen is the length of the destination address
 * @return is the number of bytes sent or -1 for an error
 */
int ltp_sendto(int sock, const void *msg, size_t len, int flags, ltpaddr *to, ltpaddr_len tolen); 

/*!
 * @fn ltp_recvfrom(int sock, unsigned char *msg, size_t len, int flags, ltpaddr *addr, ltpaddr_len *addr_len)
 * @brief recvfrom equivalent for LTP
 * @param sock is a destination LTP "session" (like a socket)
 * @param msg are the bytes rx'd (the buffer has to be allocated by the caller)
 * @param len is the number of bytes
 * @param flags contains flags for the LTP session
 * @param addr contains the address (engine ID) of the LTP peer
 * @param addr_len contains the size of the above
 * @return is the number of bytes sent or a negative number  for an error (if the input buffer was too small
 *     we return the negative of the number that would have been ok and a subsequent call with that sized
 *     buffer should be fine. (Note this differs from UDP's recvfrom return, because LTP blocks can be BIG!)
 */
int ltp_recvfrom(int sock, unsigned char *msg, size_t len, int flags, ltpaddr *addr, ltpaddr_len *addr_len);

/*!
 * @fn ltp_close(int sock)
 * @brief shutdown a socket
 * @param sock is the socket to close
 */
void ltp_close(int sock);

/// Types of ltp_status reports
#define LTP_PEER_REP 		0	///< A type of status report about peers (i.e. indpendent of paricular blocks)
#define LTP_STATISTICS		1	///< A report about how we're doing generally
#define LTP_ERROR			2	///< About a timer expiring or other error affecting nominal data handling
#define LTP_TIMERS			3	///< About currently counting LTP timers
#define LTP_LOWER_LAYERS	4	///< About lower layers

/*!
 * @struct ltp_status
 * contains the state of play of the LTP layer, reporting on any timers that have expired or
 * are pending, as well as about blocks which weren't successfully sent and information about
 * peers (but I've still to figure out exactly what goes in here)
 */
typedef struct ltp_status_defn {
	int type; ///< the type of report - see above
	int vallen;  ///< the length of the actual report
	char *rptstr; ///< the report string (alloc'd by implementation - don't free!)
	/// @todo add some union of report detail structures here as we go
} ltp_status;

/*!
 * @fn ltp_get_status(int *n, ltp_status *stats, bool *more2go)
 * @brief check the status of the LTP layer
 * @param n is the number of status reports 
 * @param stats is an array of status reports
 * @param more2go specifies whether there are more status reports available (which don't fit in the current stats)
 * @return is zero for success, or non-zero for error
 * On input, the parameter n specifies the size of the array stats (which is allocated by the 
 * caller); on output it contains the number of status reports which have been "filled in";
 * if there are more status reports that the (input) n, then more2go will be set to true.
 * Pointers within report structures are allocated by the implementation and must not be
 * free'd by the caller (@todo think about threads).
 */
int ltp_get_status(int *n,ltp_status *stats, bool *more2go);

/*!
 * @fn int ltp_ping(void)
 * @brief give the LTP layer a chance to do some work
 * @return zero or error
 */
int ltp_ping(void);

/*!
 * @fn int ltp_save
 * @brief tells LTP to save its state etc to non-volatile storage
 * @return zero or error
 */
int ltp_save(void);

/*!
 * @fn int ltp_restore
 * @brief tells LTP to restore its state etc from non-volatile storage
 * @return zero or error
 */
int ltp_restore(void);

/*!
 * @fn ltp_init(void)
 * @brief initialisation function - called upon 1st call to a ltp_*
 * @return is zero for success, or else an error
 */
int ltp_init(void);


/*!
 * @brief setsockopt() equivalent
 * @param s is the socket
 * @param level is the level to be used and MUST be SOL_SOCKET here
 * @param optname is the #define'd name of the option, e.g. LTP_SO_LINGER
 * @param optval is the value of the option
 * @param optlen is the length of the value
 * @return zero or error
 */
int ltp_setsockopt(int s, int level, int optname, const void *optval, socklen_t optlen);

/*!
 * @brief getsockopt() equivalent
 * @param s is the socket
 * @param level is the level to be used and MUST be SOL_SOCKET here
 * @param optname is the #define'd name of the option, e.g. LTP_SO_LINGER
 * @param optval is the value of the option
 * @param optlen is the length of the value
 * @return zero or error
 */
int ltp_getsockopt(int s, int level, int optname, void *optval, socklen_t *optlen);

/*!
 * @brief map a string into an LTP socket address 
 * @param str is the string (can be DNS or dot.sep. IP)
 * @param addr is the resuting address
 * @return zero or error
 */
int str2ltpaddr(char *str, ltpaddr *addr);

/*!
 * @brief string-ify an ltpaddr
 * @param addr is the address to string-fy
 * @return is the string-form (in a static buffer)
 */
char *ltpaddr2str(ltpaddr *addr);

/*!
 * @brief compare an address to the ltpaddr wildcard
 * @param guy is the comparee
 * @param len is the length of the guy
 * @return true if the same
 */
bool isany(ltpaddr *guy,int len);

/*!
 * @brief set the response from whoami to a pre-cooked one
 * @param ltpaddr *me
 * @return zero or error
 *
 * This is needed when you want to get responses to the LTP
 * segments you send to a peer LTP engine, and where you 
 * don't have any relevant LTP configuration (e.g. schedules
 * that determine who you are.)
 */
int ltp_set_whoiam(ltpaddr *me);


/*!
 * @brief tells a node which other nodes are listening now (so we can Tx to them if there're packets)
 * @param nlisteners on input is the size of the listeners array, and on output is the number of real entries in the array
 * @param listeners is an array (caller allocated) of the LTP addresses of the currently listening nodes
 * @return zero or error
 *
 * This can be called by an application that stores up LTP blocks until a 
 * contact is open, e.g. a bundle protocol agent like DTN2's dtnd.
 * If there is a substantial light trip time or latency between the
 * peers, this function tells the caller when to transmit, i.e. the
 * "now" is relative to when to transmit, not when LTP segments 
 * will be received.
 */
int ltp_whos_listening_now(int *nlisteners, ltpaddr *listeners);

/*!
 * @brief compare two LTP addresses
 * @param a1 is a comparee
 * @param a2 is a comparee
 * @param a1len is the length of significant bytes of a1
 * @return zero if they're the same, non-zero otherwise
 *
 * This compares a1len bytes of a1 with a2 (allowing NULL=NULL to return zero). The funny
 * length handling is due to idiosyncracies of the underlying sockaddr structure.
 */
int ltpaddr_cmp(ltpaddr *a1, ltpaddr *a2,int a1len );

#endif
