/*!
 * @file ltp_codec.h
 * @brief This API hides details of the LTP codec
 * @version $Revision: 1.16 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This API "knows" abstract definitions for the various
 * segment types in LTP. For each we will have an encode
 * and decode function, e.g. ltpc_encode_ds() and 
 * ltpc_decode_ds(). 
 *
 * The memory model is that the caller allocates all
 * storage.
 *
 * There are only "flat" new types defined for this api, we
 * just deal with raw 'C' types as far as posible.
 *
 * #define'd constants here are all LTPC_*, functions
 * and structures are ltpc_*
 */


/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef LTP_CODEC_H
#define LTP_CODEC_H

#include "ltp.h"

/*! 
 * @defgroup g0 misc. stuff
 */

#define LTPC_DS_OVERHEAD 32 ///< This is the maximum amount of space, over and above the number of data bytes required for a data segment
#define LTPC_RS_OVERHEAD 100 ///< This is the maximum amount of space, over and above the number of claims required for a data segment
#define LTPC_RA_OVERHEAD 200 ///< This is the maximum amount of space required for a report ack segment
#define LTPC_CO_OVERHEAD 200 ///< This is the maximum amount of space required for an originator's cancel segment
#define LTPC_CAO_OVERHEAD 200 ///< This is the maximum amount of space required for an originatpr's cancel ack segment
#define LTPC_CR_OVERHEAD 200 ///< This is the maximum amount of space required for a recipient's cancel segment
#define LTPC_CAR_OVERHEAD 200 ///< This is the maximum amount of space required for a recipient's cancel ack segment

#define LTPC_DS_HEAD 0x00 ///< Byte 0 of a "standard" data segment (other FLAG bits can be combined with this!)
#define LTPC_RS_HEAD 0x08 ///< Byte 0 of a report segment (only poss).
#define LTPC_RA_HEAD 0x09 ///< Byte 0 of a report ack segment (only poss).
#define LTPC_CO_HEAD  0x0C ///< Byte 0 of an originator's cancel segment (only poss).
#define LTPC_CAO_HEAD 0x0D ///< Byte 0 of an originatgor's cancel ack segment (only poss).
#define LTPC_CR_HEAD  0x0E ///< Byte 0 of a recipient's cancel segment (only poss).
#define LTPC_CAR_HEAD 0x0F ///< Byte 0 of a recipient's cancel ack segment (only poss).

#define LTP_VERSION 0 ///< this is the 4-bit value of the LTP version
#define LTP_VERSION_MASK 0x0F ///< a mask to or-in the version to a segment control byte
#define LTP_VERSION_EXTRACTOR(x) (x/16) ///< extract the version bits from the control byte

/// flag bits are still a mess - go read the spec
#define LTPC_FLAG0_BIT 		0x01 ///< rightmost bit of byte[0] 
#define LTPC_FLAG1_BIT  	0x02 ///< rightmost-but-one bit of byte[0] 
#define LTPC_EXC_FLAG_BIT  	0x04 ///< rightmost-but-two bit of byte[0] 
#define LTPC_CTRL_FLAG_BIT 	0x08 ///< rightmost-but-two bit of byte[0] 

#define LTPC_C_APP			0x00 ///< The application caused the cancel segment to be sent
#define LTPC_C_UNREACH		0x01 ///< The client service is unreachable (for some system's reason maybe)
#define LTPC_C_TIMER		0x02 ///< Some timer or re-tx limit was reached

#define LTPC_C_MISCOLOURED	0x03 	///< Received either a red-part data segment
									///< at block offset above any green-part
									///< data segment offset or a green-part
									///< data segment at block offset below any
									///< red-part data segment offset.

#define LTPC_C_OTHER		0xff ///< Something we didnt't think of caused the error!

/*!
 * @defgroup g1 codec packet types
 *  We #define new values here (there are some in ltp_int.h too) so as to 
 *  allow for e.g. "truncated". All these values are >100 so that functions can
 *  return one of these or else an error
 */
/// The type of an LTP block (a specific value)
typedef int ltp_segment_type;

#define LTPC_SEG_DS			100				///< a (complete) data segment
#define LTPC_SEG_RS			101				///< a (complete) report segment
#define LTPC_SEG_RA			102				///< a (complete) report ack segment
#define LTPC_SEG_CO			103				///< a (complete) originator's cancel segment
#define LTPC_SEG_CAO		104				///< a (complete) originator's cancel ack segment
#define LTPC_SEG_CR			105				///< a (complete) recipient's cancel segment
#define LTPC_SEG_CAR		106				///< a (complete) recipient's cancel ack segment
#define LTPC_SEG_ENCODED	107				///< one of the above, but already encoded
// new values should be less than LTPC_SEG_TRUNC
#define LTPC_SEG_TRUNC		149				///< a truncated segment of some sort - waiting for the next buffer might help
#define LTPC_SEG_TRUNCATED	149				///< a truncated segment of some sort - waiting for the next buffer might help
#define LTPC_SEG_CRAP		150				///< no idea - looks like total crap to me - recommend storing in /dev/null 
#define LTPC_SEG_BAD		150				///< no idea - looks like total crap to me - recommend storing in /dev/null 

#define LTPC_ERROR(xXx) (xXx<LTPC_SEG_DS || xXx>LTPC_SEG_ENCODED)	///< a macro to spot errors

/// a macro to spot control segments
#define LTPC_CTRL_SEG(xXxType) (xXxType >= LTPC_SEG_RS && xXxType <=LTPC_SEG_CAR) 

#define LTPC_SEG_MIN_LEN 4 ///< This is the minimum possible length of a segment

/*!
 * @defgroup g6 codec reasoncodes (for C segment)
 */
#define LTPC_RC_CLOSE 0x00	///< Client chose to close socket
#define LTPC_RC_UNREACHABLE 0x01 ///< No sign of that socket here
#define LTPC_RC_TIMER 0x02 ///< Some timer somewhere expired


#define LTPC_MAXFSEG 4000 ///< max. amount of space needed for a file-encoding of a segment

/*!
 * @defgroup g2 abstract types for the various segments
 */

/*!
 * @struct ltpc_octets
 * @brief len/val stuff here.
 */
typedef struct ltpc_octets_defn {
		int len; ///< the length
		unsigned char *val; ///< guess what?
} ltpc_octets;


/*
 * In all the ltpc_xx structures below, we depend on the peer, sno and exp
 * fields being at the start and the same. I could make this a separate part
 * of the ltpc_seg structure but that'd also be a pain sometimes. So don't
 * change their placment or names anywhere - ok?
 */

/*!
 * @struct ltpc_ds
 * @brief an abstract definition of an LTP data segment
 * @param eob indicates end-of-block - this segment is the last of this block
 * @param accelerated indicates that this segment should be delivered in accelerated mode
 * @param red is a boolean indicating whether the data segment is red (true) or green (false)
 * @param eorp is a boolean indicating that this segment is the last red one from the block
 * @param retransmit indicates that this segment is being re-transmitted (and hence references a report serial number)
 * @param bytes_offset the offset of this segment's bytes within the block to which it belongs
 * @param bytes is a buffer containing the bytes to be sent (1st byte to send is at element zero of this buffer)
 * @param csi is the client service ID which is like a port at the LTP layer 
 * @param csn is a check point serial number and is only used if checkpoint is true
 * @param rsn is a report point serial number - only used if retransmit is true (note: retransmit overrides checkpoint so a csn is needed here too)
 */
typedef struct ltpc_ds_defn {
	ltpc_octets		csn;
	ltpc_octets		rsn;
	ltpc_octets		csi;
	bool			eob;
	bool			red;
	bool			eorp;
	bool			checkpoint;
	bool			retransmit;
	int				offset;
	ltpc_octets		bytes;
} ltpc_ds;

/*!
 * @struct ltpc_rs
 * @brief an abstract definition of a report segment
 * @param rsn is a report point serial number - only used if retransmit is true (note: retransmit overrides checkpoint so a csn is needed here too)
 * @param csn is a check point serial number and is only used if checkpoint is true
 * @param ub is the upper bound - how far into the block these claims pertain
 * @param lb is the lower bound - how far into the block these claims pertain
 * @param cc is the claim count - note the "int" type here is a limitation not imposed by the protocol
 * @param offsets is an array (cc elements) of offsets
 * @param lengths is an array (cc elements) of lengths
 */
typedef struct ltpc_rs_defn {
	ltpc_octets		rsn;
	ltpc_octets		csn;
	int				ub;
	int				lb;
	int				cc;
	int				*offsets;
	int				*lengths;
} ltpc_rs;

/*!
 * @struct ltpc_ra
 * @brief an abstract definition of a report ack segment
 * @param rsn is a report point serial number - only used if retransmit is true (note: retransmit overrides checkpoint so a csn is needed here too)
 */
typedef struct ltpc_ra_defn {
	ltpc_octets		rsn;
} ltpc_ra;

/*!
 * @struct ltpc_co
 * @brief an abstract definition of a cancel segment
 * @param reasoncode is the reason for sending the cancel
 */
typedef struct ltpc_co_defn {
	int				reasoncode;
} ltpc_co;

/*!
 * @struct ltpc_cao
 * @brief an abstract definition of a cancel ack segment
 * @param peer is the guy sending
 * @param sno is the session number part of sessionID (and is the ltp_instance .id field of the session)
 * @param exp is the usually empty expansion zone.
 */
typedef struct ltpc_cao_defn {
} ltpc_cao;

/*!
 * @struct ltpc_cr
 * @brief an abstract definition of a cancel segment
 * @param reasoncode is the reason for sending the cancel
 */
typedef struct ltpc_cr_defn {
	int				reasoncode;
} ltpc_cr;

/*!
 * @struct ltpc_car
 * @brief an abstract definition of a cancel ack segment
 * @param peer is the guy sending
 * @param sno is the session number part of sessionID (and is the ltp_instance .id field of the session)
 * @param exp is the usually empty expansion zone.
 */
typedef struct ltpc_car_defn {
} ltpc_car;

/*!
 * @struct ltpc_ext
 * @param header determines if the extension is a header (true) or trailer (false)
 * @param type is the type of the extension
 * @param len is the size of the extension value
 * @param val is the extension value cast to void*
 * @param enclen is the length of the encoded version of the extension
 * @param encval is the encoded version of the extension
 */
typedef struct ltpc_ext_defn {
	bool				header;
	int					type;
	int					len;
	void		 		*val;
	int					enclen;
	unsigned char		*encval;
} ltpc_ext;

/*!
 * @struct ltpc_seg
 * @brief any oul segment
 * @param type is one of LTPC_SEG_* (e.g. LTPC_SEG_DS)
 * @param peer is the guy sending
 * @param sno is the session number part of sessionID (and is the ltp_instance .id field of the session)
 * @param nexts is the number of (header and trailer) extensions present
 * @param exts are the extensions themselves
 * @param val is one of the message specific structures (e.g. ltpc_ds)
 */
typedef struct ltpc_seg_defn {
	int type; 
	int version;
	ltpc_octets		peer;
	ltpc_octets		sno;
	int				nexts;
	ltpc_ext		*exts;
	union {
		ltpc_ds		ds;		///< data 
		ltpc_rs		rs;		///< report 
		ltpc_ra		ra;		///< report ack
		ltpc_co		co;		///< originator's cancel
		ltpc_cao	cao;	///< originator's cancel ack
		ltpc_cr		cr;		///< recipient's cancel
		ltpc_car	car;	///< recipient's cancel ack
	} val;
} ltpc_seg;

/*!
 * @defgroup g3 function prototypes
 */

/*!
 * @brief utilituy free fnc in case we use different memory models or whatever.
 * @param foo is the thing to be freed - should have been alloc'd with ltpc_alloc()
 */
void ltpc_free(unsigned char *foo);

/*!
 * @brief utilituy malloc fnc in case we use different memory models or whatever.
 * @param len is the amount of space to allocate (won't be zero'd!)
 * @return is the pointer to the allocated space
 */
unsigned char* ltpc_alloc(size_t len);

/*!
 * @brief utilituy realloc fnc in case we use different memory models or whatever.
 * @param len is the amount of space to allocate (won't be zero'd!)
 * @return is the pointer to the allocated space
 */
unsigned char* ltpc_realloc(unsigned char *ptr,size_t len);

/*!
 * @brief free up everything inside an ltpc_seg
 * @param seg is the segment to free up.
 */
void ltpc_free_seg(ltpc_seg *seg);

/*!
 * @brief free up all the extensions in an ltpc_seg
 * @param seg is the segment to free up stuff from
 */
void ltpc_free_seg_exts(ltpc_seg *seg);

/*!
 * @brief encode any segment at all
 * @param seg is an abstract data segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode(ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet);

/*!
 * @brief first level decoding of a buffer - see what type of segment it might be
 * @param len is the buffer length on input, and the remaining buffer length on output (zero if no more)
 * @param val is (a ptr to) the buffer on input, and points to the byte after the end of the 1st complete segment on output (NULL if no more)
 * @param seg is a (caller allocated) structure for the decoded stuff
 * @return is the type of segment e.g. LTPC_SEG_DS etc.
 */
int ltpc_decode(int *len, unsigned char **val, ltpc_seg *seg);

/*!
 * @brief map from ltpc_octets to an int if possible
 * @param oct is the (ptr to an) ltpc_octets str
 * @param val is the (ptr to the) int output
 * @return zero or error
 */
int ltpc_oct2int(ltpc_octets *oct, int *val);

/*!
 * @brief make a copy of an ltpc_octet
 * @param to is the destinatin
 * @param from is the source
 * @return zero or error
 */
int ltpc_oct_cpy(ltpc_octets *to, ltpc_octets *from);

/*!
 * @fn int sdnv_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf)
 * @brief Encode up an SDNV-8 from an input buffer
 * @param ilen is the length of the input
 * @param ibuf is the input buffer
 * @param olen is the size of the output buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-8 should be written
 * @return zero or error as usual 
 *
 * Encode up an SDNV. See the LTP specification for details. 
 */
int sdnv_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf);

/*!
 * @brief decode an sdnv-8
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param olen is the size of the output buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-8 should be written
 * @return zero or error
 */
int sdnv_dec(int *blen, unsigned char *buf, int *olen, unsigned char *obuf);

/*!
 * @brief utility fnc that encodes an unsigned int an network byte order using as few bytes as possible
 * @param val is the to-be-encoded value
 * @param blen the buffer length and used buffer on output
 * @param buf is the buffer for the encoding
 * @return zero or error
 */
int ltpc_small_htonl(unsigned int val, int *blen, unsigned char *buf);

/*!
 * @brief pretty print an encoded message into a buffer
 * @param msg is the encoded message
 * @param msglen is the encoded message length
 * @param buffer is the (called allocated) buffer 
 * @param len is the length of the output
 * @return zero or error
 */
int ltpc_pp(unsigned char *msg, int msglen, char *buffer, int *len);

/*!
 * @brief binary to ascii hex converter
 * @param buf if the binary buffer
 * @param len is the buffer length
 * @param str is the called allocated string buffer (which is sufficiently long)
 * @param slen is the length of the string buffer
 * @return zero or error
 */
int bin2AH(unsigned char *buf, int len, char *str,int slen);

/*! 
 * @brief octets 2 int
 * @param octets is an ltpc_octets ptr
 * @return is an integer (-1 is used for dodgy stuff)
 */
int ltpc_oct2int_r(ltpc_octets *in);

/*!
 * @brief encode an int as an SDNV
 * @param val is the integer to be encoded
 * @param olen is the size of the uutput buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-16 should be written
 * @return zero or error as usual 
 */
int sdnv_enc_int(int val, int *olen, unsigned char *obuf);

/*!
 * @brief decode an sndv to an int
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param iout is a ptr to the output int
 * @return zero or error
 */
int sdnv_dec_int(int *blen, unsigned char *buf, int *iout);

#endif
