/*!
 * @file ltp_crypt.h
 * @brief Codec for specific (actually, all-known) LTP extensions
 * @version $Revision: 1.6 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This API handles key management and use of crypto libraries
 * (in this case openssl by default). The intent is that it
 * be replaceable if someone wants to use a different (typically
 * more complex) key management scheme and/or crypto library.
 *
 * The key management scheme used here is described in the
 * C file documentation - ltp_crypto.cc
 *
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef LTP_CRYPT
#define LTP_CRYPT

#include "ltp.h" // for ltp_auth_so definition
#include "ltp_codec.h" // for ltpc_octets definition

/// the default crypto configuration file name
#define LTP_CRYPT_FNAME_DEF "ltp.crypt"

/*!
 * @brief make is easy to specify a ciphersuite
 * @param str is the optarg string from a -A
 * @return is the ciphersuite or -1 for error
 */
int ltp_crypt_mapciphersuite(char *str);

/*!
 * @brief specify and read the crypto configuration file
 * @param fname is the file name 
 * @return zero or error
 */
int ltp_crypt_setup(char *fname);

/*!
 * @brief do the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param peer is the peer for this segment
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param olen is a ptr to the output buffer length (length used on output)
 * @param obuf is the caller allocated output buffer
 * @return zero or error
 */
int ltp_crypt_do(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int *olen, unsigned char **obuf);

/*!
 * @brief check the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param peer is the peer for this segment
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param alen is a ptr to the authenticator buffer length (length used on output)
 * @param abuf is the authenticator buffer
 * @param result is the check result (true = ok, false = bad)
 * @return zero or error
 */
int ltp_crypt_check(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int alen, unsigned char *abuf, bool *result);

/*!
 * @brief check if an address matches an address range
 * @param rlen is the length of the address range specifier
 * @param rval is the address range specifier
 * @param vlen is the value length
 * @param val is the value
 * @return true if the value is specified by the range specifier
 */
bool match_char_ip(int rlen, char *rval, int vlen, void *val);

#endif

