/*!
 * @file ltp_exts.h
 * @brief Codec for specific (actually, all-known) LTP extensions
 * @version $Revision: 1.17 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This API "knows" abstract definitions for the various
 * extension types in LTP/LTP-T. For each we will have an encode
 * and decode function, e.g. ltpc_encode_cookie() and 
 * ltpc_decode_cookie(). We also have a table of extensions
 * with function pointers etc.
 *
 * The memory model is that the caller allocates all
 * storage - all required inputs are passed in 
 * separately.
 *
 * There are no new types defined for this api, we
 * just deal with raw 'C' types as far as posible.
 * See ltp_exts.h for definitions.
 *
 * Both of the above are aimed at avoiding me making
 * unwarranted assumptions about how the codec will
 * evolve over time. Not sure if it'll help that much
 * but worth a try.
 *
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef LTP_EXT_H
#define LTP_EXT_H

#include "ltp.h"
#include "ltp_int.h"
#include "ltp_codec.h"

/*!
 * @defgroup g0 basic extension header type numbers
 */
#define LTPE_AUTH				0x0 				///< Standard authentication header (structured)
#define LTPE_COOKIE				0x1 				///< Standard cookie header	(octets)


/// the ciphersuites
#define LTPE_CIPHERSUITE_ORIGINAUTH		LTP_CS_MAC
#define LTPE_CIPHERSUITE_SIGNATURE		LTP_CS_SIG
#define LTPE_CIPHERSUITE_NULL			LTP_CS_NULL


/*!
 * @defgroup g1 LTP-T extension header type numbers (order is as in TCD CS Tech Rep)
 */
#define LTPTE_FIRST_TAG			0xC0				///< Since my tags are liable to move about let's make that easy
#define LTPTE_SOURCE			LTPTE_FIRST_TAG+0	///< The source address (an SDNV)
#define LTPTE_DEST				LTPTE_FIRST_TAG+1	///< The destination address (octets)
#define LTPTE_EBS				LTPTE_FIRST_TAG+2	///< The estimated block size (octets)
#define LTPTE_PORT				LTPTE_FIRST_TAG+3	///< Like a UDP/TCP port (not sure if wanted really) (octets)
#define LTPTE_HOP_COUNT			LTPTE_FIRST_TAG+4	///< Number of subsequent hops allowed (one octet)
#define LTPTE_VIA	 			LTPTE_FIRST_TAG+5	///< Debug header (structured)
#define LTPTE_AUTH	 			LTPTE_FIRST_TAG+6	///< End-to-end authentication (structured like LTPE_AUTH)
#define LTPTE_CONFID			LTPTE_FIRST_TAG+7	///< End-to-end confidentiality (not yet defined)
#define LTPTE_ECN	 			LTPTE_FIRST_TAG+8	///< In-band congestion notification (structured)

/// Maximum number of hops allowed in an LTPT segment LTPTE_HOP_COUNT extension
/// This has to be 32 bits or less (with anything v. big being a bit crazy)
#define LTPT_MAX_HOPS LTP_T_MAX_HOPCOUNT

/// The max peer for LTPT is bounded by the congestion structure LTPTE_ECN to, for now,
/// 16 bits
#define LTPT_MAX_PEER 0xffff

/// The min and max of time units for ecn LTPTE_ECN strutures, basically its an eight bit signed number
#define LTPT_MIN_UNITS -127
#define LTPT_MAX_UNITS 127

/// The max duration for an LTPTE_ECN - its 24 bits
#define LTPT_MAX_DURATION (1<<24-1)

/*!
 * @defgroup g2 abstract types for the standard extensions
 */

/*!
 * @struct ltpe_auth_header - uses key handles etc. from socket
 * @param ciphersuite is the ciphersuite identifier
 * @param keyID is the key identifier sent over the wire (zero length if none applies)
 */
typedef struct ltpe_auth_header_defn {
	unsigned char	ciphersuite;
	ltpc_octets		keyid;
} ltpe_auth_header;

/*!
 * @struct ltpe_auth_trailer
 * @param authVal is the authentication result
 */
typedef struct ltpe_auth_trailer_defn {
	ltpc_octets		authval;					
} ltpe_auth_trailer;

/*!
 * @struct ltpe_e2eauth
 * @param ciphersuite is the ciphersuite identifier
 * @param keyID is the key identifier sent over the wire (zero length if none applies)
 * @param authVal is the authentication result
 * 
 * This is like the above two combined which works fine in the LTP-T e2e case
 */
typedef struct ltpe_e2eauth_defn {
	unsigned char	ciphersuite;
	ltpc_octets		keyid;
	ltpc_octets		authval;					
} ltpe_e2eauth;

/*!
 * @struct ltpe_cookie
 * @param cval the cookie value
 */
typedef struct ltpe_cookie_defn {
	ltpc_octets		cval;
} ltpe_cookie;

/*!
 * @defgroup g3 abstract types for the LTP-T extensions
 */

/*!
 * @struct ltpe_src
 * @param src is the peer's name/id - 0xc0
 */
typedef struct ltpe_src_defn {
	ltpaddr			src;			///< the peer's address 
} ltpe_src;

/*!
 * @struct ltpe_dest
 * @param dest is the peer's name/id - 0xc1
 */
typedef struct ltpe_dest_defn {
	ltpaddr			dest;			///< the peer's address
} ltpe_dest;

/*!
 * @struct ltpe_ebs
 * @param ebs is the estimated block size  - 0xc2
 */
typedef struct ltpe_ebs_defn {
	unsigned int	ebs;			///< estimated block size
} ltpe_ebs;

/*!
 * @struct ltpe_port
 * @param port is the LTP-T port - 0xc3
 */
typedef struct ltpe_port_defn {
	ltpc_octets			port;
} ltpe_port;

/*!
 * @struct ltpe_hop_count
 * @param hop_count is the LTP-T hop_count - 0xc4
 */
typedef struct ltpe_hop_count_defn {
	unsigned int	hop_count;
} ltpe_hop_count;

/*!
 * @struct ltpe_via
 * @param via is a big long string with hosts/times in UTF8 
 */
typedef struct ltpe_via_defn {
	ltpc_octets		via;		
} ltpe_via;

/// LTP-T's authentication (0xC6) uses the same data structure as LTP's and hence ltpe_auth
/// LTP-T's confid is TBD (0xC7)

/*!
 * @struct ltpe_ecn
 * @param necn is the number of ecn's contained in the extension - 0xC8
 * @param peers are the peers claimed to be suffering congestion
 * @param units are the time units of each ecn - represented as a signed power of 10
 * @param durations are the durations (in time units) for which congestion is expected to persist (from the sending time)
 */
typedef struct ltpe_ecn_defn {
	int				necn;
	ltpc_octets		*peers;
	int				*units;
	unsigned int	*durations;
} ltpe_ecn;

/*!
 * @defgroup g3 generic apis
 */

/*!
 * @brief populate the extension codec tables and other init
 * @return zero or error
 */
int ltpe_init(void);


/*!
 * @brief general encoder (calls type specific fncs.)
 * @param ext is an extension structure
 * @return zero or error
 */
int ltpe_encode(ltpc_ext *ext);

/*!
 * @brief general decoder (calls type specific fncs.)
 * @param ext is an extension structure
 * @return zero or error
 */
int ltpe_decode(ltpc_ext *ext);

/*!
 * @brief call a print fnc if it exists
 * @param ext is the guy to print
 * @param str is the caller-allocated print buffer
 * @param slen is the length of str
 * @return zero or error
 */
int ltpe_print(ltpc_ext *ext, int slen, char *str);

/*!
 * @defgroup g4 encoding and decoding apis
 */

/*! 
 * @brief decode a auth extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_auth(ltpc_ext *ext);

/*!
 * @brief encode a auth extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_auth(ltpc_ext *ext);

/*!
 * @brief print a auth extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_auth(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a cookie extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_cookie(ltpc_ext *ext);

/*!
 * @brief encode a cookie extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_cookie(ltpc_ext *ext);

/*!
 * @brief print a cookie extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_cookie(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a source extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_source(ltpc_ext *ext);

/*!
 * @brief encode a source extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_source(ltpc_ext *ext);

/*!
 * @brief print a source extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_source(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a dest extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_dest(ltpc_ext *ext);

/*!
 * @brief encode a dest extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_dest(ltpc_ext *ext);

/*!
 * @brief print a dest extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_dest(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a ebs extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_ebs(ltpc_ext *ext);

/*!
 * @brief encode a ebs extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_ebs(ltpc_ext *ext);

/*!
 * @brief print a ebs extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_ebs(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a port extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_port(ltpc_ext *ext);

/*!
 * @brief encode a port extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_port(ltpc_ext *ext);

/*!
 * @brief print a port extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_port(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a hop_count extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_hop_count(ltpc_ext *ext);

/*!
 * @brief encode a hop_count extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_hop_count(ltpc_ext *ext);

/*!
 * @brief print a hop_count extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_hop_count(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a via extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_via(ltpc_ext *ext);

/*!
 * @brief encode a via extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_via(ltpc_ext *ext);

/*!
 * @brief print a via extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_via(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a auth extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_t_auth(ltpc_ext *ext);

/*!
 * @brief encode a auth extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_t_auth(ltpc_ext *ext);

/*!
 * @brief print a auth extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_t_auth(ltpc_ext *ext, int strlen, char *strbuf);

/*! 
 * @brief decode a ecn extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_ecn(ltpc_ext *ext);

/*!
 * @brief encode a ecn extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_ecn(ltpc_ext *ext);

/*!
 * @brief print a ecn extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_ecn(ltpc_ext *ext, int strlen, char *strbuf);

/*!
 * @defgroup g4 encoding and decoding apis
 */

/*!
 * @brief free an extension that only contais an ltpc_octets field (cookie,port,via)
 * @param ext is the extension structure
 */
void ltpe_free_octets_ext(ltpc_ext *ext);

/*!
 * @brief free an extension that only contais an ltpaddr field (src,dest)
 * @param ext is the extension structure
 */
void ltpe_free_addr_ext(ltpc_ext *ext);

/*!
 * @brief free an auth extension 
 * @param ext is the extension structure
 */
void ltpe_free_auth_ext(ltpc_ext *ext);

/*!
 * @brief free an ecn extension 
 * @param ext is the extension structure
 */
void ltpe_free_ecn_ext(ltpc_ext *ext);

/*!
 * @brief generic free fnc that calls others via jump table
 * @param ext is the extension to free the internals for
 */
void ltpe_free_ext(ltpc_ext *ext);

/*! 
 * @brief decode a auth extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_e2eauth(ltpc_ext *ext);

/*!
 * @brief encode a auth extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_e2eauth(ltpc_ext *ext);

/*!
 * @brief print a auth extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_e2eauth(ltpc_ext *ext, int strlen, char *strbuf);

/*!
 * @brief free an e2eauth extension 
 * @param ext is the extension structure
 */
void ltpe_free_e2eauth(ltpc_ext *ext);

#endif
