/*!
 * @file ltp_int.h
 * @brief This is the internal header for our transport protocol
 * @version $Revision: 1.36 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef LTP_INT_H
#define LTP_INT_H

#include "ltp_codec.h"

/// This is for the MUDSAT build, TODO:  make this a configure param later
#define MUDSAT


/// a type for ltp structure instances (so I can flatten to file!)
typedef int ltp_instance;

/// These are 'C' programming artefacts. Due to realloc() potentially moving things,
/// I need the fixparent() function (in ltp.cc). That finds good parent
/// pointers given an ltp_instance (a parentid). But we've go to know which
/// type of parent we're fixing hence...
typedef int32_t ltp_parent;

/// a length type for LTP segments, buffers etc
// typedef unsigned long ltp_len;
typedef uint32_t ltp_len;

/*!
 * @struct ltp_octets
 * This structure is the generic octet holder
 * Pointer elements here (filename,fp,encoded.val) can be in essentially any state,
 * so code must be able, e.g. to handle a structure just containing a filename when
 * it wants to manipulat the encoded form, etc. Of course, one of those pointers
 * has to be non-NULL if we want to read the bytes, but which one will vary!
 */
typedef struct ltp_octets_defn {
	ltp_len					len; ///< the number of octets
	unsigned char			*val; ///< a raw buffer containing the octets
} ltp_octets;

/// Some known codecs - only one really for now
#define LTP_CODEC_SDNV			0 ///< use SDNV (unsigned) to represent buffers/integers
#define LTP_DEFAULT_CODEC		LTP_CODEC_SDNV ///< default codec


/*!
 * @struct ltp_mib_data
 * @brief this is the structrure for stuff I want out of a MIB (@todo snmp:-)
 * Notes: (1) that these values don't change with the clock, except when the relevant
 * comms. contact with the relevant peer is "on". For example, if I send a segment
 * to a peer and am configured to expect a report within 10 seconds, then (first
 * adding in our extimated LTT and latency) this value is stored in the time2report
 * fields of the relevant active timer. However, the value will only be decremented
 * when there's an inbound socket active with that peer. (2) The values stored
 * here do change, in contrast to the same named fields in the ltp_mib_data structure
 * which contain the initial values for these guys (that clear?) 
 * (3) The section numbers refer to where each timer is discussed in draft-08 of 
 * the spec (as sent to the DTNRG list, not yet published)
 * And finally, (4) We'll call such timer handing "punctuated time" (since we stop clocks ticking
 * when we're "offline").
 */
typedef struct ltp_mib_data_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	ltp_len				segmentsize; ///< how big is a segment (assumes only 1 layer2 - fine for now)
	ltp_time_t			time2report; ///< the punctuated time in which I expect a report about a data segment (section 9.2)
	ltp_time_t			time2rack; ///< the punctuated time in which a report ack is expected (section 9.3)
	ltp_time_t			time2cack; ///< the punctuated time in which a cancel ack is expected (section 9.15)
	int					txlimit; ///< the number of times to transmit (excl. originally) a data segment
								///<after a timer's expiry (section 9.7)
	int					reptxlimit; ///< the number of times to transmit (excl. originally) a report
								///<after a timer's expiry (section 9.8)
	int					cantxlimit; ///< the number of times to transmit (excl. originally) a cancel
								///<after a timer's expiry (section 9.16)
	ltp_time_t			cacheflushfreq; ///< specifies how often to do LTP garbage collection
	bool				linger; ///< if true then socket state is not immediately flushed after a ltp_close() call
	ltp_time_t			inactivitylimit; ///< this specifies (in punctuated time) for how long a socket will be allow sit doing nothing
	int					cookie; ///< length of cookies for this socket (0 if off)
	ltp_time_t			cookie_grace; ///< the grace period (seconds) before cookies are required in responses
	int					ciphersuite;	///< which (if any) way to secure segments
	bool				nonblocking;	///< whether to return from sendto or not before session is complete
} ltp_mib_data;

/// some MIB defaults and other known values
#define LTP_MIB_DEFAULT_SEGSIZE		1400	///< generally we can get segments of 1400 bytes payload out ok
#define LTP_MIB_DEFAULT_TIME2REP   	10		///< ignoring LTT and "comms-off" periods, I expect a report within this many seconds
#define LTP_MIB_DEFAULT_TIME2RACK	10		///< as above, but for ACKs  (note: assuming default is comms symmetry here!)
#define LTP_MIB_DEFAULT_TIME2CACK	10		///< as above, but for cancel ACKs  (note: assuming default is comms symmetry here!)
#define LTP_MIB_DEFAULT_TXLIMIT		3		///< should be small anyhow
#define LTP_MIB_DEFAULT_REPTXLIMIT	2		///< should be small anyhow
#define LTP_MIB_DEFAULT_CANTXLIMIT	1		///< should be small anyhow
#define LTP_MIB_DEFAULT_GARBAGE		60*60*24 ///< clean up sockets once per day by default
#define LTP_MIB_DEFAULT_LINGER		false   ///< immediately clean up socket state by default
#define LTP_MIB_DEFAULT_INACTIVITY	60*10 	///< allow 2 punctuated minutes of doing nothing
#define LTP_MAX_COOKIE				16		///< max cookie len (Bytes)
#define LTP_MIN_COOKIE				4		///< min cookie len (Bytes)
#define LTP_MAX_COOKIE_GRACE		60*60*24	///< max cookie grace period (1 day)
#define LTP_MIN_COOKIE_GRACE		0		///< min cookie grace period (0 seconds)
#define LTP_MIB_DEFAULT_COOKIE_LEN		0	///< default is no cookies
#define LTP_MIB_DEFAULT_COOKIE_GRACE	LTP_MIB_DEFAULT_TIME2REP	///< default is 5 seconds
#define LTP_MIB_DEFAULT_REDLEN		LTP_ALLRED	///< all red default
#define LTP_MIB_DEFAULT_CIPHER		LTP_CS_NONE	///< no crypto default
#define LTP_MIB_DEFAULT_NONBLOCK	true ///< don't wait until success

/// local mib variable
static ltp_mib_data ltp_mib_var={
		0,
		LTP_MIB_DEFAULT_SEGSIZE,
		LTP_MIB_DEFAULT_TIME2REP,
		LTP_MIB_DEFAULT_TIME2RACK,
		LTP_MIB_DEFAULT_TIME2CACK,
		LTP_MIB_DEFAULT_TXLIMIT,
		LTP_MIB_DEFAULT_REPTXLIMIT,
		LTP_MIB_DEFAULT_CANTXLIMIT,
		LTP_MIB_DEFAULT_GARBAGE,
		LTP_MIB_DEFAULT_LINGER,
		LTP_MIB_DEFAULT_INACTIVITY,
		LTP_MIB_DEFAULT_COOKIE_LEN,
		LTP_MIB_DEFAULT_COOKIE_GRACE,
		LTP_MIB_DEFAULT_CIPHER,
		LTP_MIB_DEFAULT_NONBLOCK,
};

/*!
 * @struct ltp_scopes
 * @brief this is the structure where we record which bits of a block we've gotten
 * I'm not sure when, but at some stage we need to merge consequetive scopes. But
 * at the start there'll be one per rx'd segment.
 */
typedef struct ltp_scopes_defn {
	int 		nscopes; ///< the number of ub/lb pairs
	int			alloced; ///< the number of ubs/lbs allocated
	int			*lbs; ///< the lower bounds of the scopes
	int			*ubs; ///< the upper bounds of the scopes
} ltp_scopes;

/// these are the types of scopes in a block
/// nothing, just so zero isn't used
#define LTP_NO_SOCPES	0
/// scopes that I've rx'd (not session initiator)
#define LTP_RX_SCOPES	1
/// scopes that I've tx'd and seen acked (session initiator)
#define LTP_TX_SCOPES	2
/// scopes that I've fwd'd (regardless of acks), so I know not to send twice
#define LTP_FWD_SCOPES	3
/// this is an artefact for representing the difference of two scopes
#define LTP_SUB_SCOPES	4

/*!
 * @struct ltpt_vals
 * @param flags is a bit mask with a 1 whereever a value is set here - specifics below
 * @param forwarded is a boolean that is set to true after all bytes of the block have been forwarded (once)
 * @param source is the original source
 * @param dest is the final destination
 * @param gw is the gw for this block 
 * @param ebs is the est. (or exact) block size
 * @param port is the LTP-T port
 * @param hop_count is the hop_count to send out
 * @param via is the via to send out
 * @param e2e_a_cs end-to-end ciphersuite
 * @param e2e_a_kid end-to-end key id
 * @param e2e_a_val end-to-end auth val
 * @param necn number of ECNs
 * @param peers - ECN peer array
 * @param units - ECN units array
 * @param durations - ECN durations array
 */
typedef struct ltpt_vals_defn {
	int					flags;
	bool				forwarded;
	ltpaddr				source;
	ltpaddr				dest;
	ltpaddr				gw;
	int					ebs;
	int					port;
	int 				hop_count;
	ltp_octets			via;
	int					e2e_a_cs;
	ltp_octets			e2e_a_kid;
	ltp_octets			e2e_a_val;
	uint				necn;
	ltpaddr 			*peers;
	int					*units;
	unsigned int		*durations;
} ltpt_vals;

#define LTPT_SET_SOURCE					0x0001	/// whether this is set in an ltpt_vals flags field
#define LTPT_SET_DEST 					0x0002	/// whether this is set in an ltpt_vals flags field
#define LTPT_SET_EBS 					0x0004	/// whether this is set in an ltpt_vals flags field
#define LTPT_SET_PORT 					0x0008	/// whether this is set in an ltpt_vals flags field
#define LTPT_SET_HOP_COUNT				0x0010	/// whether this is set in an ltpt_vals flags field
#define LTPT_SET_VIA 					0x0020	/// whether this is set in an ltpt_vals flags field
#define LTPT_SET_E2E_AUTH				0x0040	/// whether this is set in an ltpt_vals flags field
#define LTPT_SET_ECN 					0x0080	/// whether this is set in an ltpt_vals flags field

/*!
 * @struct ltp_block_defn
 * @brief this is the structure for an entire LTP block
 */
typedef struct ltp_block_defn {
	ltp_instance			id; ///< a unique number (helps mapping from data str to flat file etc)
	int						remoteidlen; ///< the (length of) the peer's idea of the session id 
	unsigned char			*remoteid; ///< the the peer's idea of the session id 
	bool					complete; ///< whether the remaining information is about a partial or complete block
	bool					delivered; ///< whether the block data has been read via a ltp_recvfrom call
	bool					eob_seen; ///< whether or not the eob segment has arrived
	bool					eorp_seen; ///< whether or not the eorp segment has arrived
	bool					eorp_sent; ///< whether or not the eorp segment has been fwd'd
	bool					radone; ///< whether or not all RAs have been sent/rx'd
	bool					finishedwithwhendelivered; ///< once delivered, then this is ok to mark finishedwith
	bool					finishedwith; ///< whether this is now ready for garbage collection
	ltp_octets				inboundcookie; ///< if non-zero the other side's cookie
	ltp_octets				outboundcookie; ///< if non-zero this side's cookie
	bool					cookie_grace_expired; ///< whether or not the cookie's grace period has expired
	short 					csi; ///< the client service indicator for this block
	ltp_time_t				created; ///< time of creation of this record
	ltp_octets				decoded; ///< the bare block data 
	struct ltp_socket_defn *parent; ///< the "parent" socket for this block
	ltp_instance			parentid; ///< a unique number (helps mapping from data str to flat file etc)
	ltp_scopes				rx_scopes; ///< holds the ranges of the block we've gotten for rx'd blocks
	ltp_scopes				tx_scopes; ///< holds the ranges of the block we've gotten for tx'd blocks
	ltp_scopes				fwd_scopes; ///< holds the ranges of the block we've forwarded for router blocks
	ltpt_vals				t_vals; ///< holds the LTP-T values for this session/block
	int						current_rsn; ///< holds the most current report serial number
	int						redtide; ///< the highest tide of red we've seen for an inbound block
	bool 					redcomplete; ///< whether red data is complete for an outbound block
	bool					eobsent; ///< whether the eob (green or red has been sent)
#ifdef MUDSAT
    int                     mudsat_re_tx_ctr; ///<  counter for how many MUDSSAT re-tx's to do
#endif
} ltp_block;

/// the state of a segment
typedef int ltp_segment_state;

/// the status values for segment
#define LTP_PS_NEW				0 ///< just created (INITIAL STATE for both)
#define LTP_PS_WAITING_SEND		1 ///< segment is to be sent whenever possible
#define LTP_PS_SENDING_NOW		2 ///< segment is being sent (note: not the same as in transit!)
#define LTP_PS_SENT				3 ///< segment wsa sent but not ack'd yet
#define LTP_PS_ACKED			4 ///< segment's rx has been ack'd (FINAL STATE for Tx)
#define LTP_PS_ARRIVING			5 ///< segment is in the process of arriving now
#define LTP_PS_WAITING_RS		6 ///< segment was rx'd but RS hasn't been sent yet
#define LTP_PS_SENT_RS			7 ///< segment was rx'd and RS has been sent 
#define LTP_PS_RS_ACKED			8 ///< segment's RS has been ACKED (FINAL STATE for Rx)

#define LTP_NOCACHE				-1 ///< this value is used to indicate that an r_entry field is unused
#define LTP_CACHE				-2 ///< this value is used to indicate that an r_entry field is unused
#define LTP_NOTIMER				-1 ///< this value is used to indicate that a t_entry field is unused

/*!
 * @struct ltp_segment
 * @brief The structure which contains an abstract description of an LTP segment; 
 */
typedef struct ltp_segment_defn {
	ltp_instance			id; ///< a unique number (helps mapping from data str to flat file etc)
	ltp_segment_state		state; ///< the state of this segment
	int						r_entry; ///< the place to put a response for replay detection purposes if necessary
									///< LTP_NOCACHE means ignore this
	ltp_instance			r_id; ///< the identifier of the above
	int						t_entry; ///< the index of the timer associated with this segment
									///< so we can reinit_timer on that at the right time if we later re-tx  this seg
	ltp_instance			t_id; ///< the identifier of the above
	bool					finishedwith; ///< whether this is now ready for garbage collection
	ltp_block				*parent; ///< the "parent" block for this segment (if any)
	ltp_instance			parentid; ///< a unique number (helps mapping from data str to flat file etc)
	ltpaddr					peer; ///< the guy on the other end
	ltpaddr_len				peerlen; ///< the length of peer (wish sockaddr was better structured)
	ltpc_seg				seg; ///< the abstract segment for/from the codec API
	ltpc_octets				encoded; ///< for some re-tx cases we've already got the encoding done
	int						encoded_type; ///< if the segment is encoded then here's the encoded's segment type
} ltp_segment;

/*!
 * @struct ltp_sockopts
 * @brief the things you can set via setsocktop() calls
 */
typedef struct ltp_sockopts_defn {
	bool				linger; 			///< whether this socket's state is immediately killed on a ltp_close() call
	ltp_time_t			inactivitylimit; 	///< this specifies (in punctuated time) for how long this socket will be allow sit doing nothing
	int					cookie; 			///< whether cookies are enabled for this socket & if so, length
	ltp_time_t			cookie_grace; ///< the (punctuated?) time after 1st cookie out until we start dropping segments
	int					redlen; ///< how many bytes from each block are to be red - -1 means all
	ltp_auth_so			auth; ///< authentication settings (ciphersuite, kid)
	ltp_t_cfg_so		t_cfg; ///< the LTP-T configuration
	bool				need_e2e_int; ///< Tell LTP-T that I insist on good e2e int on rx'd blocks
	bool				nonblocking; ///< don't wait to see if sendto() finished
	int					mtu; ///< the MTU to use for this guy
	bool				mtu_set; ///< the MTU to use for this guy
} ltp_sockopts;

/// This is used to initialise new sockets with our default sockopts
static ltp_sockopts LTP_DEFAULT_SOCKOPT={
	LTP_MIB_DEFAULT_LINGER, 
	LTP_MIB_DEFAULT_INACTIVITY,
	LTP_MIB_DEFAULT_COOKIE_LEN,
	LTP_MIB_DEFAULT_COOKIE_GRACE,
	LTP_MIB_DEFAULT_REDLEN,
	{ LTP_MIB_DEFAULT_CIPHER, 0, "" },
	{ false, LTP_T_DEF_RELAY_SOCKETS, false, LTP_T_NO_PORT, LTP_T_NO_HOPCOUNT, false, { LTP_MIB_DEFAULT_CIPHER, 0, "" }, false },
	false,
	LTP_MIB_DEFAULT_NONBLOCK,
	0,false,
	};

/*!
 * @struct ltp_socket
 * @brief a container for the sockets currently known to the higher layers
 */
typedef struct ltp_socket_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	bool				inbound; ///< is true if we will rx stuff here
	bool				outbound; ///< is true if we'll send stuff 
	int					sockfd; ///< the socket handle for this
	bool				finishedwith; ///< whether this is now ready for garbage collection
	int					alloced; ///< the number of blocks allocated
	int					nblocks; ///< number of blocks in use
	ltp_block			*blocks; ///< ...currently "in-play" blocks  
	ltpaddr				ibpeer; ///< the other guy in this connection - inbound
	ltpaddr_len			ibpeerlen; ///< the length of inbound peer (wish sockaddr was better structured)
	ltpaddr				obpeer; ///< the other guy in this connection - outbound
	ltpaddr_len			obpeerlen; ///< the length of outbound peer (wish sockaddr was better structured)
	ltp_sockopts		opts; ///< the socket options for this socket
	bool				ltptrelay; ///< true iff this is an LTP-T relay socket
	bool				close_called; ///< true iff ltp_close was called (see SO_LINGER)
} ltp_socket_t;

/*!
 * @struct ltp_loopback_queue
 * @brief this is the structure where we queue blocks being looped back at the LTP layer
 */
typedef struct ltp_loopback_queue_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	int					alloced; ///< the number of loopback allocated
	int					nblocks; ///< number of...
	ltp_block			*blocks; ///< ...currently "in-play" blocks  
} ltp_loopback;

/*!
 * @struct ltp_tx_queue
 * @brief this is the structure for holding segments queued for transmission and info on peers to whom I've recently transmitted
 */
typedef struct ltp_tx_queue_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	int					alloced; ///< the number of segments allocated
	int					nsegments; ///< the number of segments in the tx-queue
	struct ltp_segment_defn	*segments; ///< the segments themselves
} ltp_tx_queue;

/*!
 * @struct ltp_cache_detail_defn
 * @brief the details of a relpay cache entry
 */
typedef struct ltp_cache_detail_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	bool				finishedwith; ///< true=>can ignore this guy - he's just for garbage collection
	ltpc_octets			hash; ///< the hash of the previously rx'd message
	ltp_segment_type	type; ///< the type of the response message here
	ltpc_octets			response; ///< the responses to those messages (mostly empty!)
	ltp_instance		parentid; ///< the id of the matching segment or block or socket, depending
} ltp_cache_detail;

/*!
 * @struct ltp_rx_cache
 * @brief this is the list of finger prints of rx'd packets
 *
 * Our replay detection scheme is to store a hash
 * as an the fingerprint for each rx'd buffer. When we get
 * a hit, we simply drop the incoming packet.
 */
typedef struct ltp_rx_cache_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	int					alloced; ///< the number of sockets allocated
	int					nentries; ///< the number of hash table entries
	ltp_cache_detail	*details; ///< the entries themselves
} ltp_rx_cache;

typedef int ltp_timer_type; ///< These are the timers that we need

#define LTP_TIMER_CP		0	///< checkpoint timer, default is MIB.time2report, a ltp_time_t expirer
								///< the number of times to transmit (excl. originally) a data segment, (default 
								///< is MIB.txlimit) after a timer's expiry (section 9.7) - a countdown timer
#define LTP_TIMER_RS		1	///< report segment timer, default is MIB.time2rack, a ltp_timer_t expirer
								///< the number of times to transmit (excl. originally) a report segment, (default 
								///< is MIB.reptxlimit) after a timer's expiry (section 9.7) - a countdown timer
#define LTP_TIMER_C			2	///< cancel segment timer, default is MIB.time2cack, a ltp_timer_t expirer
								///< the number of times to transmit (excl. originally) a cancel segment, (default 
								///< is MIB.cantxlimit) after a timer's expiry (section 9.7) - a countdown timer
#define LTP_TIMER_INACTIVITY	3	///< the inactivity timer on the socket
#define LTP_TIMER_COOKIE	4	///< the grace period after a 1st cookie was set

/*!
 * @struct ltp_timer_detail
 * @brief this is where we keep details of a given timer
 */
typedef struct ltp_timer_detail_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	ltp_timer_type		type;	///< the type of this timer (one of LTP_TIMER_<<foo>>, e.g. LTP_TIMER_CP) 
								///< which can be a ltp_time_t or a counter (which "expires" at zero
	bool				finishedwith; ///< true=>can ignore this guy - he's just for garbage collection
	bool				active; ///< true if decrementing happening now
	bool				everexpired; ///< true if this guy has ever expired
	ltp_time_t			lastdecremented; ///< last time I decremented this
	ltp_time_t			remaining; ///< the (punctuated!) time remaining before this timer expires
	int					counter; ///<the number of decrements remaining until expiry
	ltp_segment			*segment; ///< a pointer to that segment (note: can move!)
	ltp_instance		segmentid; ///< the id of the matching segment
	ltp_block			*block; ///< a pointer to that block (note: can move!)
	ltp_instance		blockid; ///< the id of the matching block
	ltp_socket_t		*socket; ///< a pointer to that socket (note: can move!)
	ltp_instance		socketid; ///< the id of the matching socket
	ltp_time_t			absstart; ///< the real world time to start punctuated decrements (usually one RTT after init_timer call)
	ltp_time_t			rtt; ///< the est. RTT for the connection
} ltp_timer_detail;

/*!
 * @struct ltp_timers
 * @brief this structure contains information about the currently running timers
 */
typedef struct ltp_timers_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	int					alloced; ///< the number of sockets allocated
	int					nentries; ///< the number of hash table entries
	ltp_timer_detail	*details; ///< the individual timer details
} ltp_timers;

/*!
 * @struct ltp_state
 * @brief This is our global status variable. 
 *
 * Everything we have to remember had better be linked in here, since this is
 * what we sync to disk.
 */
typedef struct ltp_state_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	bool 				inited;	///< whether ltp's internals have been initialised 
	int					nsockets; ///< the number of open sockets
	int					alloced; ///< the number of sockets allocated
	ltp_socket_t		*sockets; ///< them themselves
	ltp_loopback		lbq; ///< the loopback queue
	ltp_tx_queue		txq; ///< the transmit queue
	ltp_rx_cache		replays; ///< the list of received segment fingerprints
	ltp_timers			timers; ///< the structure for timers
	ltp_time_t			lastbincollection; ///< time of last garbage collection
	ltp_t_cfg_so		t_cfg; ///< the LTP-T configuration
} ltp_state;

/*!
 * @struct ltp_stats
 * @brief a container for statistics, gathered on the fly
 */
typedef struct ltp_stats_defn {
	ltp_instance		id; ///< a unique number (helps mapping from data str to flat file etc)
	ltp_time_t		start;	///< time when ltp initialised
	int					blocks_in; ///< number of complete blocks recveived
	int					blocks_out; ///< number of complete blocks transmitted
	int					segments_in; ///< number of complete segments recveived
	int					segments_out; ///< number of complete segments transmitted
	int					segment_errs; ///< number of segment errors
	ltp_len				bytes_in; ///< number of complete bytes recveived
	ltp_len				bytes_out; ///< number of complete blocks transmitted
	int					timeouts; ///< the number of timeouts seen
} ltp_stats;

/// our statistics variable
static ltp_stats ltp_stat_var={0,0,0,0,0,0,0,0,0,0};

#define LTP_PARENT_BLOCK		0	///< an ltp_parent value - when a block is parent of a segment
#define LTP_PARENT_SOCKET 		1	///< an ltp_parent value - when a socket is parent of a block
#define LTP_PARENT_SEGMENT		2	///< an ltp_parent value - when a segment is the parent of a timer

/*!
 * @brief fix the association between a parentid and the associated parent pointer
 * @param parent is the to-be-modified-if-necessary parent pointer
 * @param parentid is the ltp_instance to which we'd like to point
 * @param type is LTP_PARENT_[SOCKET|BLOCK|SEGMENT]
 * @return zero or error
 */
int ltp_fixparent(ltp_state *tree, void **parent,ltp_instance parentid, ltp_parent type);

/// This is the max number of consequetive errors allowed in the ltp_ping() function
/// When exceeded the pinger thread will exit (not sure what'll happen then!)
#define LTP_MAX_PING_ERRORS 100

/*!
 * @brief add a new, or the first extension
 * @param seg to the guy to add to 
 * @param added is the new guy
 * @return is the new total number of segments or a negative number for errors
 */
int add_ext(ltpc_seg *seg);

/*!
 * @brief map ltpaddr to an octet string 
 * @param ad is the ltp address
 * @param sp is the octet string
 * @return zero or error
 *
 * The mapping is to use an IP address in network byte order with an optional
 * port number, if the default isn't in use.
 */
int ltpaddr2octets(ltpaddr *ad, ltpc_octets *sp);

/*!
 * @brief map an octet string to an LTP address
 * @param sp is the octet string
 * @param ad is the LTP address
 * @return zero or error
 *
 * The inverse of ltpaddr2octets
 */
int octets2ltpaddr(ltpc_octets *sp, ltpaddr *ad);

#ifdef LTP_MULTICAST
/*!
 * @fn int ltp_is_multi(addr)
 * @brief utility fnc to determine if target addr is MC or not
 * @return zero MC, error (1) if not
 */
int ltp_is_multi(ltpaddr *addr);
#endif

#endif

