/*!
 * @file ltp_l2.h
 * @brief This is the external interface for our link layer protocol
 * @version $Revision: 1.5 $ $Author: stephen $
 * @version Copyright (C) 2004 6rinity College Dublin
 *
 * The LTP implementation codes down to this, and from here we can
 * go down to a) an OMNET++ simulation, b) LTP/UDP on cygwin and eventually
 * c) 802.11 on the SeNDT node. We'll see what happens.
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef LTP_L2_H
#define LTP_L2_H

#include "ltp.h"
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/*!
 * @struct ltp_l2_buffer
 * Contains the usual (len,val) but also some memory allocation flags.
 *
 * If the heap flag is true, then val can be freed sometime, (its in the heap), otherwise never (
 * its static or on the stack)
 * If the l2alloc flag is true then val was alloc'd inside l2, false by the caller
 * If the rx_complete is true, then this is a complete l2 block, otherwise something went wrong
 * during reception (ignore for 2b-transmitted blocks)
 */
typedef struct ltp_l2_buffer_defn {
	bool	heap; ///< whether this is in the heap (and so can be freed) or is static/on the stack and cannot
	bool	l2alloc; ///< whether alloc'd by the l2 code or the caller
	bool 	dirty; ///< set to true when bytes added from network, then false then decoded to spot missed buffers if decoding is slow
	unsigned int len; ///< buffer length
	unsigned char *val; ///< buffer value
} ltp_l2_buffer;

/*!
 * @struct ltp_l2_id
 * @brief a simple name gatherum all structure
 *
 * Either the pname should contain the loopback string or else
 * the addr should contain something usable with UDP sockets.
 */
typedef struct ltp_l2_id_defn {
	bool	heap; ///< whether this is in the heap (and so can be freed) or is static/on the stack and cannot
	bool	l2alloc; ///< whether alloc'd by the l2 code or the caller
	char				*pname; ///< printable version of name for logging/convience maybe a DNS name, or URI
	unsigned int		addr_len; ///< the length of the address
	ltpaddr 	addr; ///< the L2 address
} ltp_l2_id;


/// This is the Loopback identity - use it as a destination and it'll pop back to your data handler
extern const ltp_l2_id l2lb;

/// This is the wildcard identity - use it as a destination and it'll pop back to your data handler
extern const ltp_l2_id l2any;


/*!
 * @fn ltp_copy_l2_id(ltp_l2_id *from, ltp_l2_id *to) 
 * @brief deep copy a ltp_l2_id
 * @param from,to guess what? (caller alloc'd top level structs)
 * @return zero or error
 *
 * Any combination of "empty" fields are ok in "from", they'll just be replicated.
 */
int ltp_copy_l2_id(ltp_l2_id *from, ltp_l2_id *to);

/*!
 * @fn ltp_free_l2_id(ltp_l2_id *it)
 * @brief yada yada, doesn't free top level thing
 * @return zero or error
 */
int ltp_free_l2_id(ltp_l2_id *it);

/*!
 * @brief a callback function that is called when there is a layer2 link allowing for potential tx/rx
 * @return zero or error
 */
typedef int (*ltp_l2_cb_cts)(void);

/*!
 * @brief a callback function that's called when there's a block of data off the net ready for leyer3
 * @param buf is a buffer (probably alloc'd by l2) which contains the data
 * @return zero or error
 */
typedef int (*ltp_l2_cb_data)(ltp_l2_buffer	*buf, ltp_l2_id *peer);

/// Layer 2 errors
///
#define LTP_L2_OK		0			///< Not an error I guess:-)
#define LTP_L2_ERR		1			///< General, unspecified error
#define LTP_L2_GONE		2			///< Used to have an l2 connection, but its unexpectedly disappered!
#define LTP_L2_TOUT		3			///< Some l2 (bad) timeout happened

/*!
 * @brief a callback function to announce an l2 error
 * @param errno is the error number (according to l2)
 * @param str is an error string
 * @return zero or error
 */
typedef int (*ltp_l2_cb_err)(int errno, char *str);

// Layer 2 states (visible from l3)
#define LTP_L2S_RESET		0		///< state prior to ltp_l2_init() or post a call to ltp_l2_shutdown()
#define LTP_L2S_INITED		1		///< state post ltp_l2_init() or post ltp_l2_off() (but before ltp_l2_shutdown())
#define LTP_L2S_ON_TX		2		///< "turned on" and can connect to one or more peers (post ltp_l2_on(), before ltp_l2_off())
#define LTP_L2S_ON_RX		3		///< "turned on" and can connect to one or more peers (post ltp_l2_on(), before ltp_l2_off())
#define LTP_L2S_ON_BOTH		4		///< "turned on" and can connect to one or more peers (post ltp_l2_on(), before ltp_l2_off())
#define LTP_L2S_SLEEP1		5		///< "turned on" but sleeping in sleep mode 1 (semantics depends on device)
#define LTP_L2S_SLEEP2		6		///< "turned on" but sleeping in sleep mode 2 (semantics depends on device)
#define LTP_L2S_SLEEP3		7		///< "turned on" but sleeping in sleep mode 3 (semantics depends on device)
#define LTP_L2S_SLEEP4		8		///< "turned on" but sleeping in sleep mode 4 (semantics depends on device)
#define LTP_L2S_SLEEP5		9		///< "turned on" but sleeping in sleep mode 5 (semantics depends on device)

/*! 
 * @struct ltp_l2_callbacks
 * @brief this structure contains pointers to the layer2 callback functions provided by layer 3 (LTP)
 */
typedef struct ltp_l2_callbacks_defn {
	ltp_l2_cb_cts		cts; ///< Clear-to-send callback
	ltp_l2_cb_data		data; ///< Data in rx buffer callback
	ltp_l2_cb_err		err; ///< Error handler
} ltp_l2_callbacks;

// The default callback variables

/// The default callback structure
extern ltp_l2_callbacks* defcbs_p;

/// @struct ltp_l2_cb_queue 
/// This is a structure that can contain a number of buffers and their associated meta-data
typedef struct ltp_l2_cb_queue_defn {
	int		size; ///< the number of buffers in the circular array
	int 	position; ///< the current position in the queue
	ltp_l2_buffer	*bufs; ///< the circular array (initialised to 'size' elements)
	ltp_l2_id		*peers; ///< the origin(rx) or destination(tx) of the packet
	ltp_time_t		*times; ///< the time of arrival/transmission as appropriate (initialised to 'size' elements)
} ltp_l2_cb_queue;

#ifdef THREADS
/// this protects queues from badness (during writes)
extern pthread_mutex_t qutex;

#undef QUTEXDEBUG
#ifdef QUTEXDEBUG
#define GRABQUTEX { \
	fprintf(stderr,"Grabbing qutex (in %s @ %d, tid=%p)...",__FUNCTION__,__LINE__,(void*)pthread_self()); \
	pthread_mutex_lock(&qutex); \
	fprintf(stderr,"..got qutex\n"); }
/// Give back the l2 listener mutex
#define DROPQUTEX { \
	fprintf(stderr,"Dropping qutex (in %s @ %d, tid=%p)...",__FUNCTION__,__LINE__,(void*)pthread_self()); \
	pthread_mutex_unlock(&qutex); \
	fprintf(stderr,"...dropped qutex\n"); }
#else
#define GRABQUTEX pthread_mutex_lock(&qutex);
#define DROPQUTEX pthread_mutex_unlock(&qutex);
#endif

#else
#define GRABQUTEX
#define DROPQUTEX
#endif

/// This is the default Rx queue
extern ltp_l2_cb_queue* rxqueue_p;

/// And the default Tx queue
extern ltp_l2_cb_queue* txqueue_p;

/// This is the size of the default tx & rx queues
#define LTP_L2_CB_DEFAULT_QUEUE_SIZE	1000

/*!
 * @fn ltp_l2_init(ltp_l2_callbacks *cbs, int port)
 * @param cbs is a pointer to a callback structure (has ptrs to c-b functions)
 * @param rport is the "port" to listen with (actually in the case of e.g. UDP/IP, virtually otherwise)
 * @brief boot-time initialise the layer 2 stuff 
 * @return zero or error
 *
 * This function initialises a l2 implementation so that it'll be ready to 
 * be turned on using ltp_l2_on()
 */
int ltp_l2_init(ltp_l2_callbacks *cbs, int rport);

/*!
 * @fn ltp_l2_on(unsigned int *mtu, int state2go2)
 * @brief runtime initialise the layer 2 stuff and return the MTU for this link
 * @param mtu is the (caller alloc'd) place where the MTU will be returned
 * @param state2go2 is the state we want to go to, which can be LTP_L2S_ON[_TX,_RX, or _BOTH]
 * @return zero or error
 *
 * This function "turns on" a layer2 implementation, so that its ready to
 * bind/accept-connections/whatever according to the desired state (state2go2)
 *
 * MTU means the max-transmission-unit and is interpreted here as the biggest
 * payload that can be sent or rx'd by this l2.
 */
int ltp_l2_on(unsigned int *mtu, int state2go2);

#define LTP_L2_OFF_NONE 0
#define LTP_L2_OFF_BOTH 1
#define LTP_L2_OFF_TX   2
#define LTP_L2_OFF_RX	3

/*!
 * @fn ltp_l2_off(int what)
 * @param what one of the LTP_L2_OFF_* values
 * @brief turn off the layer 2 stuff (can be turned back on with ltp_l2_on later on)
 * @return zero or error
 */
int ltp_l2_off(int what);

/*!
 * @fn ltp_l2_shutdown(void)
 * @brief shutdown the layer 2 stuff, before we go to sleep or shutdown CPU (CPU reset
 * may be needed prior to next ltp_l2_init()
 * @return zero or error
 */
int ltp_l2_shutdown(void);

/*!
 * @fn ltp_l2_get_state(void)
 * @brief get the state of l2
 * @return is the state of l2 (one of LTP_L2S_* above)
 */
int ltp_l2_get_state(void);

/*!
 * @fn ltp_l2_sleep(int mode)
 * @brief set the sleep mode using one of the defined LTP_L2S_SLEEP* modes
 * @param mode is the ltp sleep state we want to get to (e.g. LTP_L2S_SLEEP1)
 * @return zero or error (e.g. if mode is not a sleep state)
 */
int ltp_l2_sleep(int mode);

/*!
 * @fn ltp_l2_send( ltp_l2_buffer	*buf, ltp_l2_id		*recip)
 * @brief send a block to an l2 peer
 * @param buf is the buffer to sent
 * @param recip is the peer to whom it should be sent
 */
int ltp_l2_send(
		ltp_l2_buffer	*buf,
		ltp_l2_id		*recip);

#endif
