/*!
 * @file ltp_t.h
 * @brief The LTP-T protocol's stuff
 * @version $Revision: 1.14 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef LTP_T_H
#define LTP_T_H

#include "ltp.h"
#include "ltp_int.h"
#include "ltp_ext.h"
#include "ltp_codec.h"

/*!
 * @brief load an LTP-T configuration
 * @param fname is the filename to use
 * @param cfg is a ptr to the caller allocated configuration structure
 * @return zero or error
 */
int ltpt_load_cfg(char *fname, ltp_t_cfg_so *cfg);

/*!
 * @brief load LTP-T routes
 * @param fname is the filename to use
 * @param nroutes is a ptr to an integer on output containing the number of routes
 * @param routes is a ptr to an array of ltp_t_route_so's allocated inside this function
 * @return zero or error
 */
int ltpt_load_routes(char *fname, int *nroutes,ltp_t_route_so **routes);

/*!
 * @brief estimate the length of stuff extension encodings take up
 * @param theblock is the block concerned
 * @param thesock is the socket
 * @return the estimeate
 */
int ltpt_extensions_estimate(ltp_block *theblock, ltp_socket_t *thesock);

/*!
 * @brief initialise a cfg to default
 * @param cfg is a ptr to an LTP-T cfg
 * @return zero or error
 */
int ltpt_def_cfg(ltp_t_cfg_so *cfg);

/*!
 * @brief initialise using a cfg
 * @param cfg is a ptr to an LTP-T cfg
 * @return zero or error
 */
int ltpt_init(ltp_t_cfg_so *cfg);

/*!
 * @brief re-initialise using a cfg
 * @param cfg is a ptr to an LTP-T cfg
 * @return zero or error
 */
int ltpt_reinit(ltp_t_cfg_so *cfg);

/*!
 * @brief add basic LTP-T related extensions
 * @param seg is the ltpc_seg
 * @param block is the block
 * @param sock is the socket
 * @return zero or error
 */
int add_ltpt_exts(ltpc_seg *seg, ltp_block *block, ltp_socket_t *sock);

/*!
 * @brief figure out whether this segment is an LTP-T one or not
 * @param seg is the segment
 * @return true if the segment has any LTP-T extensions
 */
bool is_ltpt_seg(ltpc_seg *seg);

/*!
 * @brief figure out whether this LTP-T segment is from me 
 * @param seg is the segment
 * @return true if the segment is from me
 */
bool is_ltpt_fromme(ltpc_seg *seg);

/*!
 * @brief figure out whether this LTP-T segment is for me (LTP peer= LTP-T dest)
 * @param seg is the segment
 * @return true if the segment is for me
 */
bool is_ltpt_forme(ltpc_seg *seg);

/*!
 * @brief do the LTP-T processing for this just received segment
 * @param segstr is the segment structure
 * @param theblock is the parent block (it can be NULL)
 * @param thesock is the parent socket (can it be NULL - not sure?)
 * @param badcookie is an output - true if this segments cookies are awry in some way  
 * @param inlen is the length of the encoded segment
 */
int ltpt_inbound_processing(ltpc_seg *segstr, ltp_block *theblock, ltp_socket_t *thesock, bool *badauth, int inlen);

/*!
 * @brief figure out the LTP-T gateway to use for an LTP-T address
 * @param dest is the destination
 * @param gw is the gateway to use (called allocated)
 * @param nlisteners is the number of current contacts
 * @param listeners are the current contacts
 * @param matched is true (on return) if some route matches, otherwise false
 * @return zero or error
 */
int ltpt_gateway_listeners(ltpaddr *dest, ltpaddr *gw, int nlisteners, ltpaddr *listeners, bool *matched);

/*!
 * @brief figure out the LTP-T gateway to use for an LTP-T address
 * @param dest is the destination
 * @param gw is the gateway to use (called allocated)
 * @return zero or error
 */
int ltpt_gateway(ltpaddr *dest, ltpaddr *gw);

/*!
 * @brief add a route to the local set
 * @param route is a ptr to an ltp_t_route_so
 * @return zero or error
 */
int ltpt_set_route(ltp_t_route_so *route);

/*!
 * @brief return a printable version of the local routes
 * @return a ptr to a static char * with route info
 */
char *ltpt_print_routes(void);

/*!
 * @brief find the ultimate destination for this LTP-T block
 * @param block is the block we're dealing wiht
 * @param dest is the (caller alloc'd) output dest
 * @return zero or error
 */
int ltpt_dest(ltp_block *block, ltpaddr *dest);

/*!
 * @brief final LTP-T processing before delivery
 * @param theblock is the parent block (it can be NULL)
 * @param thesock is the parent socket (can it be NULL - not sure?)
 * @param status returns whether things look good or not
 * @param status_str explains why not, if not
 * @param sstr_len is the size of the status_str buffer
 * @return zero or error
 */
int ltpt_final_processing(ltp_block *block, ltp_socket_t *sock, bool *status, char *status_str, int sstr_len);

/*!
 * @brief add basic LTP-T related extensions
 * @param seg is the ltpc_seg
 * @param block is the block
 * @param sock is the socket
 * @return zero or error
 */
int ltpt_add_departure_exts(ltpc_seg *seg, ltp_block *block, ltp_socket_t *sock);

#endif
