/*!
 * @file ltp_codec.cc
 * @brief This API hides details of the LTP codec
 * @version $Revision: 1.53 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This API "knows" abstract definitions for the various
 * segment types in LTP. For each we will have an encode
 * and decode function, e.g. ltpc_encode_ds() and 
 * ltpc_decode_ds(). 
 *
 * The memory model is that the caller allocates all
 * storage - all required inputs are passed in 
 * separately.
 *
 * There are no new types defined for this api, we
 * just deal with raw 'C' types as far as posible.
 *
 * Both of the above are aimed at avoiding me making
 * unwarranted assumptions about how the codec will
 * evolve over time. Not sure if it'll help that much
 * but worth a try.
 *
 * There's a little history here too. There used to be
 * two sdnv types sdnv8 and sdnv16, but those were
 * deprecated, but I left the function names for now
 * just to make the code transition easier. 
 * TODO: Clean up that sdnv code sometime.
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "ltp_base.h"
#include "ltp_codec.h"
#include "ltp_ext.h"

/*!
 * @section g1 redefinitions of stuff...
 * potentially redefine this here since we might want to separate out this code
 * from the rest of the build
 */
#ifndef LTPERR
#define LTPERR(xxFoo) { return(xxFoo);}
#endif
#ifndef LTP_OK
#define LTP_OK 0
#endif
#ifndef LTP_BAD_PARAM
#define LTP_BAD_PARAM 1
#endif
#ifndef LTP_MEMORY
#define LTP_MEMORY 2
#endif

/*!
 * @section g2 sdnv codec
 */

/// Since callers allocated memory we have to check sizing often, this is just a bit of sugar for that
#define BUFLIMIT(xX,yY) if (xX>=yY) LTPERR(LTP_MEMORY);

/*!
 * @brief decode an sdnv of some sort
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param oct is the output len/buffer
 * @param slen is the input sdnv type (i.e. 8 or 18) NOW IGNORED
 * @return zero or error
 */
static int sdnv_dec_aux(int *blen, unsigned char *buf, ltpc_octets *oct,int slen)
{
	// check for zero case(s!)
	if (!blen || *blen <0) {
		LTPERR(LTP_BAD_PARAM); // no, or a negative length is bad
	}
	if (*blen==0 || !buf) { // zero's of various sorts are ok though
		oct->len=0;
		oct->val=NULL;
		return(LTP_OK);
	}
	if (*blen==1 & buf[0]==0x00) {
		oct->len=1;
		oct->val=ltpc_alloc(oct->len);
		if (!oct->val) LTPERR(LTP_MEMORY);
		oct->val[0]=0x00;
		*blen=0;
		return(LTP_OK);
	}
	if (buf[0]<=0x7f) { // its a short encoding
		if (*blen<1) LTPERR(LTPC_SEG_CRAP);
		oct->len=1;
		oct->val=ltpc_alloc(oct->len);
		if (!oct->val) LTPERR(LTP_MEMORY);
		oct->val[0]=buf[0];
		*blen -= 1;
	} else {
		// long encoding
		// read until error or a byte without top bit set
		// bit de-stuffing for decoded value required
		// could be more efficient I guess but who cares for now
		int bytes_used=0;
		int len_left=*blen;
		while (len_left-- >0 && buf[bytes_used++] >= 0x80) ;
		if (len_left<=0 && buf[bytes_used-1] >= 0x80) { // ERROR sdnv extends beyond buffer
			LTPERR(LTPC_SEG_CRAP);
		}
		int leadingzeros=0;
		unsigned char *ibp=&buf[0];
		while (*ibp == 0x80) { ibp++; leadingzeros+=8; }
		unsigned char flag=0x40;
		while (flag && !(*ibp&flag)) {
			leadingzeros++;
			flag = flag>>1;
		}
		int outlen=bytes_used*7 - leadingzeros; // in bits
		outlen = outlen/8 + (outlen%8?1:0); // bytes
		oct->len=outlen;
		oct->val=ltpc_alloc(outlen); 
		int carrybits=0;
		unsigned char *outp=&oct->val[outlen-1];
		ibp=&buf[bytes_used-1]; // start at lsb
		for (int j=0;j!=outlen-1;j++) { // all but topmost byte (of output)!
			/*
			     buf[i-1] buf[i]   buf[i+1]
				 ibp-1    ibp      ibp+1
				+--------+--------+--------+
				|1yyyyxxx|1xxxxxaa|1aaaaaab|
				+--------+--------+--------+
				output bytes=xxxx xxxx, aaaa aaaa
				carrybits here is 2, i.e. the number of a's in buf[i]
				we get the most-sig (ignoring bit 8) 7-carrybits x's from buf[i]
				we get least-sig carrybits+1 x's from buf[i-1]
			*/
			unsigned char working;
			// the x's we want from buf[i], shifted to the right
			working = (*ibp-- & 0x7f) >> carrybits  ;
			// the x's we want from buf[i-1], shifted to the left
			working |=  ((*ibp & 0x7f) << (7-carrybits));
			*outp-- = working;
			carrybits=(carrybits+1)%8;
			if (!carrybits) {
				ibp--;
				carrybits=1;
			}
		}
		// take care of last byte of decoding - in the above, those the the y's
		// and we have 7-carrybits of them (carrybits is 3 when we get here in the example)
		// careful here 0x81 0x84 -> 0x42, depends on how many leadingzeros
		if (ibp==buf) {
			*outp= (*ibp & 0x7f) >> carrybits;
		} else {
			*outp= (*ibp-- & 0x7f) >> carrybits;
			*outp|= (*ibp & 0x7f)  << (7-carrybits);
		}
		// update remaining buffer length
		*blen = *blen-bytes_used;
	}
	if (*blen < 0 ) 
		LTPERR(LTPC_SEG_CRAP);
	 return(LTP_OK);
}

/*!
 * @brief decode an sdnv-8
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param out is the output len/buffer
 * @return zero or error
 */
static int sdnv8_dec(int *blen, unsigned char *buf, ltpc_octets *out)
{
	return(sdnv_dec_aux(blen,buf,out,8));
}

/*!
 * @brief decode an sdnv-16
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param out is the output len/buffer
 * @return zero or error
 */
static int sdnv16_dec(int *blen, unsigned char *buf, ltpc_octets *out)
{
	return(sdnv_dec_aux(blen,buf,out,16));
}

/*!
 * @brief decode an sndv to an int
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param iout is a ptr to the output int
 * @return zero or error
 */
static int sdnv16_dec_int(int *blen, unsigned char *buf, int *iout)
{
	ltpc_octets out;
	int rv=sdnv_dec_aux(blen,buf,&out,16);
	if (rv) LTPERR(rv);
	rv=ltpc_oct2int(&out,iout);
	if (rv) LTPERR(rv);
	ltpc_free(out.val);
	return (LTP_OK);
}

/*!
 * @brief decode an sndv to an int
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param iout is a ptr to the output int
 * @return zero or error
 */
int sdnv_dec_int(int *blen, unsigned char *buf, int *iout)
{
	return sdnv16_dec_int(blen,buf,iout);
}

/*!
 * @brief decode an sndv to an int
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param iout is a ptr to the output int
 * @return zero or error
 */
static int sdnv8_dec_int(int *blen, unsigned char *buf, int *iout)
{
	ltpc_octets out;
	int rv=sdnv_dec_aux(blen,buf,&out,8);
	if (rv) LTPERR(rv);
	rv=ltpc_oct2int(&out,iout);
	if (rv) LTPERR(rv);
	ltpc_free(out.val);
	return (LTP_OK);
}


/*!
 * @brief Encode up an SDNV-16 from an input buffer
 * @param ilen is the length of the input
 * @param ibuf is the input buffer
 * @param olen is the size of the output buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-16 should be written
 * @param slen is either 8 or 16 - NOW IGNORED
 * @return zero or error as usual 
 */
static int sdnv_enc_aux(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf, int slen)
{
	if (ilen==1 && ibuf[0]<=0x7f) {
		// short encoding allowed
		*olen=1;
		obuf[0]=ibuf[0];
		return(LTP_OK);
	}
	// long encoding 
	int leadingzeros=0;
	unsigned char *ibp=ibuf;
	int ilencnt=ilen;
	while (ilencnt && (*ibp == 0x00)) { 
		ibp++; 
		ilencnt--;
		leadingzeros=leadingzeros+8; 
	}
	if (ilencnt<=0) {
		LTPERR(LTP_BAD_VALUE);
	}
	// valgrind was complaining about the while below. Let's see what if...
	if (ibp>=(ibuf+ilen)) {
		LTPERR(LTP_BAD_VALUE);
		// fprintf(stderr,"ibp out of whack (ibp=%p, ibuf=%p, ilen=%d, ilencnt=%d, leadingzeros=%d\n",
		// ibp,ibuf,ilen,ilencnt,leadingzeros);
	}
	unsigned char flag=0x80;
	while (flag && !(*ibp&flag)) {
		leadingzeros++;
		flag = flag>>1;
	}
	int realilen=8*ilen-leadingzeros; // in bits
	realilen = realilen/8 + (realilen%8?1:0); // in bytes
	int outlen=8*realilen-leadingzeros; // in bits
	outlen = outlen/7 + ((outlen%7)?1:0); // in bytes
	if (*olen<outlen) LTPERR(LTP_BAD_PARAM); // input too long for output
	/*
         buf[i-1] buf[i]   buf[i+1] ... buf[1]   buf[0]
		+--------+--------+--------+...+--------+--------+
        |aaaaaaaa|bbbbbbbb|cccccccc|   |yyyyyyyy|xxxxxxxx|
		+--------+--------+--------+...+--------+--------+
		encodes to:
         tp-1      tp       tp+1
		+--------+--------+--------+...+--------+--------+--------+
        |1aaaabbb|1bbbbbcc|1ccccccd|   |1zzzzzyy|1yyyyyyx|0xxxxxxx|
		+--------+--------+--------+...+--------+--------+--------+
		at the start of processing buf[i] carrybits is 2
		carry should be 0000 00cc
	*/
	unsigned char *tp=&obuf[outlen-1];
	int carrybits=0;
	unsigned char carry=0x00;
	unsigned char *ib=&ibuf[ilen-1];
	for (int i=0;i!=realilen-1 && tp!=obuf;i++,ib--) {
		*tp-- = 0x80 | carry | ( ( *ib  & (0xff >> carrybits )) << carrybits );
		carry = (*ib >> (7-carrybits)) ;
        // |1aaaabbb|1bbbbbcc|1ccccccd|1ddddddd|1eeeeeee|1effffff|1ffggggg|...|1zzzzzyy|1yyyyyyx|0xxxxxxx|
		// carry  3        2        1        0        7        6        5
		carrybits++;
		if (carrybits == 7) {
			*tp-- = 0x80 | carry;
			carrybits=carry=0x00;
		} 
	}
	// one or two more bytes needed?
	if (carrybits+8-leadingzeros>7) {
		// two bytes
		*tp--=0x80 | (*ib << carrybits) | carry;
		*tp=0x80 | (*ib >> (7-carrybits));
	} else {
		// just one'll do
		*tp=0x80 | (*ib << carrybits) | carry;
	}
	obuf[outlen-1] &= 0x7f;
	*olen=outlen;
	return(LTP_OK);
}

/*!
 * @fn int sdnv8_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf)
 * @brief Encode up an SDNV-8 from an input buffer
 * @param ilen is the length of the input
 * @param ibuf is the input buffer
 * @param olen is the size of the uutput buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-8 should be written
 * @return zero or error as usual 
 *
 * Encode up an SDNV. See the LTP specification for details. 
 */
static int sdnv8_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf)
{
	return(sdnv_enc_aux(ilen,ibuf,olen,obuf,8));
}

/*!
 * @fn int sdnv16_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf)
 * @brief Encode up an SDNV-16 from an input buffer
 * @param ilen is the length of the input
 * @param ibuf is the input buffer
 * @param olen is the size of the uutput buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-16 should be written
 * @return zero or error as usual 
 *
 * Encode up an SDNV-16. See the LTP specification for details. 
 */
static int sdnv16_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf)
{
	return(sdnv_enc_aux(ilen,ibuf,olen,obuf,16));
}

/*!
 * @brief Encode an int (usually, though not always, 32-bit) as an SDNV-16
 * @param val is the value to be encoded
 * @param olen is the size of the uutput buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-16 should be written
 * @return zero or error as usual 
 */
static int sdnv16_enc_int(int val, int *olen, unsigned char *obuf)
{
	if (!val) {
		if (*olen <=1) LTPERR(LTP_BAD_PARAM);
		*olen=1;
		obuf[0]=0x00; // have to take 16 bits to allow unambiguous decoding
		return(LTP_OK);
	}
	int ilen=sizeof(int);
	int nlval=htonl(val);
	unsigned char intbuf[ilen];
	unsigned char *ibuf=intbuf;
	memcpy(intbuf,&nlval,ilen);
	int i=0;
	while (!(intbuf[i++])) { ilen--; ibuf++; }
	return(sdnv_enc_aux(ilen,ibuf,olen,obuf,16));
}

/*!
 * @brief encode an int as an SDNV
 * @param val is the integer to be encoded
 * @param olen is the size of the uutput buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-16 should be written
 * @return zero or error as usual 
 */
int sdnv_enc_int(int val, int *olen, unsigned char *obuf)
{
	return sdnv16_enc_int(val,olen,obuf);
}

/*!
 * @brief Encode an int (usually, though not always, 32-bit) as an SDNV-8
 * @param val is the value to be encoded
 * @param olen is the size of the output buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-16 should be written
 * @return zero or error as usual 
 */
static int sdnv8_enc_int(int val, int *olen, unsigned char *obuf)
{
	if (!val) {
		if (*olen <=1) LTPERR(LTP_BAD_PARAM);
		*olen=1;
		obuf[0]=0x00; 
		return(LTP_OK);
	}
	int ilen=sizeof(int);
	int nlval=htonl(val);
	unsigned char intbuf[ilen];
	unsigned char *ibuf=intbuf;
	memcpy(intbuf,&nlval,ilen);
	int i=0;
	while (!(intbuf[i++])) { ilen--; ibuf++; }
	return(sdnv_enc_aux(ilen,ibuf,olen,obuf,8));
}

/*!
 * @section g3 ltp encoding functions
 */


/*! 
 * @brief encode some extensions
 * @param seg is the segment 
 * @param obufp is the pointer into the encoding buffer
 * @param olenp is a pointer to the current length of the encoding
 * @param availp is a pointer to the currently remaining space in the encoding buffer
 * @param fornet determines if the encoding is for sending on the net or not (then to a file)
 * @param header determines if we should now encode header (true) or trailer (false) extensions 
 */
static int ltpc_encode_ext(	ltpc_seg *seg,
							unsigned char **obufpp,
							int *olenp,
							int *availp,
							bool fornet,
							bool header)
{
	if (!seg || !obufpp || !availp) LTPERR(LTP_BAD_PARAM);
	int olen=*olenp;
	unsigned char *obufp=*obufpp;
	int sdnvlen;
	int avail=*availp;
	int rv;
	// encode up all header/trailer extensions in the order given 
	for (int i=0;i!=seg->nexts;i++) {
		if (seg->exts[i].header==header) {
			// just add in the relevant type, sdnv'd length and value octets
			// but first encode if necessary
			rv=ltpe_encode(&seg->exts[i]);
			if (rv && rv!=LTP_UNKNOWN_EXT) LTPERR(rv);
			*obufp++=seg->exts[i].type; olen++; BUFLIMIT(olen,avail);
			sdnvlen=avail;
			rv=sdnv8_enc_int(seg->exts[i].enclen,&sdnvlen,obufp);
			if (rv) LTPERR(rv);
			obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
			if (seg->exts[i].enclen) { // its ok to have a zero length
				memcpy(obufp,seg->exts[i].encval,seg->exts[i].enclen);
				obufp+=seg->exts[i].enclen; olen+=seg->exts[i].enclen; BUFLIMIT(olen,avail);
			}
		}
	}
	*olenp=olen;
	*availp=avail;
	*obufpp=obufp;
	return (LTP_OK);
}

/*!
 * @brief decode out header or trailer extensions 
 * @param buf is the encoded buffer
 * @param blen is (a ptr) to the current remaining bytes in buf (modified on output)
 * @param exts is a ptr to the (possibly NULL) extensions
 * @param next is the number of extensions that should be present
 * @param header says if these are header or trailer exts
 */
static int ltpc_decode_ext(unsigned char *buf, int *blen, ltpc_ext *exts, int next, bool header)
{
	if (!buf || !blen || !exts) LTPERR(LTP_BAD_PARAM);
	if (!next) return(LTP_OK);
	unsigned char *bv=buf;
	int len=*blen;
	int olen;
	int rv;
	for (int i=0;i!=next;i++) {
		// type
		exts[i].type=*bv;
		len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
		// length
		olen=len;
		rv=sdnv8_dec_int(&len,bv,&exts[i].enclen);
		if (rv) LTPERR(rv);
		bv+=(olen-len);
		// value
		exts[i].len=0;
		exts[i].val=NULL;
		exts[i].header=header;
		if (exts[i].enclen) {
			exts[i].encval=(unsigned char*)ltpc_alloc(exts[i].enclen);
			if (!exts[i].encval) LTPERR(LTP_MEMORY);
			memcpy(exts[i].encval,bv,exts[i].enclen);
			bv+=exts[i].enclen;
			len-=exts[i].enclen;
			// see if we can decode the extension
			rv=ltpe_decode(&exts[i]);
			if (rv && rv!=LTP_UNKNOWN_EXT) LTPERR(rv);
		}
	}
	*blen=len;
	return(LTP_OK);
}

/*!
 * @brief figure out how many of each extension to encode
 * @param seg is the segment to encode
 * @return the byte to include in the encoding or -1 for error
 */
static int ltpc_encode_ext_counters(ltpc_seg *seg)
{
	int headers=0;
	int trailers=0;
	int byte=0;
	for (int i=0;i!=seg->nexts;i++) {
		if (seg->exts[i].header) headers++;
		else trailers++;
	}
	if (headers>15 || trailers>15) {
			byte=-1;
	} else {
			byte=headers<<4|trailers;
	}
	return(byte);
}

/*!
 * @brief decode out header numbers
 * @param byte is from the buffer
 * @param headers is (a ptr to) the number of headers
 * @param trailers is (a ptr to) the number of trailers
 * @param seg is the segment (so we can allocate space for extensions)
 */
static int ltpc_decode_ext_counters(unsigned char byte, int *headers, int *trailers, ltpc_seg *seg)
{
	if (!headers || !trailers || !seg) LTPERR(LTP_BAD_PARAM);
	*headers=byte/16;
	*trailers=byte%16;
	seg->nexts=*headers+*trailers;
	seg->exts=(ltpc_ext*)ltpc_alloc(sizeof(ltpc_ext)*seg->nexts);
	if (!seg->exts) LTPERR(LTP_MEMORY);
	return(LTP_OK);
}

/*!
 * @brief encode a data segment
 * @param seg is the abstract DS
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode_ds(
			ltpc_seg		*seg,
			int 			*obuflen,
			unsigned char 	*obuf,
			bool			fornet)
{
	ltpc_ds			*ds=&seg->val.ds;
	bool			eob=ds->eob;
	bool 			red=ds->red;
	bool 			eorp=ds->eorp;
	bool			checkpoint=ds->checkpoint;
	bool			retransmit=ds->retransmit;
	int				bytes_length=ds->bytes.len;
	unsigned char	 *bytes=ds->bytes.val; 
	int 			peerlen=seg->peer.len;
	unsigned char	*peer=seg->peer.val;
	int 			snolen=seg->sno.len;
	int				csilen=ds->csi.len;
	int				csnlen=ds->csn.len;
	int				rsnlen=ds->rsn.len;
	if (fornet) {
		if (!bytes || !peer || !obuflen || !obuf) LTPERR(LTP_BAD_PARAM);
	} else {
		if (!obuflen || !obuf) LTPERR(LTP_BAD_PARAM);
	}
	int olen=0;
	unsigned char *obufp=obuf;
	int avail=*obuflen;
	int sdnvlen;
	// handle red,eob,eorp,checkpoint combinations (sigh!)
	if (red) { 
		if (!checkpoint && !eorp && !eob) { 
			*obuf=0x00;
		} else if (checkpoint && !eorp && !eob) {
			*obuf=0x01;
		} else if (checkpoint && eorp && !eob) {
			*obuf=0x02;
		} else if (checkpoint && eorp && eob) {
			*obuf=0x03;
		} else {
			LTPERR(LTP_BAD_VALUE);
		}
	} else { // green
		if (!eob) {
			*obuf=4;
		} else {
			*obuf=7;
		} 
	}
	*obuf&=LTP_VERSION_MASK;

	obufp++; olen++; BUFLIMIT(olen,avail);
	int rv;


	// session originator (SDNV-8) - the ltpaddr in_addr
	sdnvlen=avail-olen; // remaining buffer space
	rv=sdnv8_enc(peerlen,(unsigned char*)peer,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// session number (SDNV-16) - an input
	sdnvlen=avail-olen;
	rv=sdnv16_enc(snolen,seg->sno.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	int extcount=ltpc_encode_ext_counters(seg);
	if (extcount==-1) LTPERR(LTP_BAD_VALUE);
	*obufp++=extcount; olen++; BUFLIMIT(olen,avail);
	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,true);
	if (rv) LTPERR(rv);

	// client service ID (SDNV-8) - the ltpaddr port number
	sdnvlen=avail-olen;
	rv=sdnv8_enc(csilen,ds->csi.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
	
	// offset (SDNV-16)
	sdnvlen=avail-olen;
	rv=sdnv16_enc_int(ds->offset,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// length (SDNV-16)
	sdnvlen=avail-olen;
	rv=sdnv16_enc_int(bytes_length,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// for a checkpoint (have to input that or make a new function)
	if ( (eob && red) || eorp || retransmit || checkpoint) { 
		sdnvlen=avail-olen;
		rv=sdnv8_enc(csnlen,ds->csn.val,&sdnvlen,obufp);
		if (rv) LTPERR(rv);
		obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
		if (!ds->rsn.len) { // 	report serial number (SDNV-8) - an optional input
			*obufp++=0x00; olen++; BUFLIMIT(olen,avail);
		} else {
			sdnvlen=avail-olen;
			rv=sdnv8_enc(rsnlen,ds->rsn.val,&sdnvlen,obufp);
			if (rv) LTPERR(rv);
			obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
		}
	}

	if (fornet) {
		// ... then the actual segment data
		sdnvlen=avail-olen; // remaining buffer space
		if (bytes_length>=sdnvlen) {
			LTPERR(LTP_BAD_PARAM);
		}
		memcpy(obufp,bytes,bytes_length);
		obufp+=bytes_length; olen+=bytes_length; BUFLIMIT(olen,avail);
	} else {
		// write out a zero
		*obufp++=0x00; olen++; BUFLIMIT(olen,avail);
	}

	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,false);
	if (rv) LTPERR(rv);
	
	*obuflen=olen;
	return(LTP_OK);
}


/*!
 * @brief encode a report segment
 * @param seg is an abstract report segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode_rs( ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet)
{
	ltpc_rs	*rs=&seg->val.rs;
	int olen=0;
	unsigned char *obufp=obuf;
	int avail=*obuflen;
	int sdnvlen;
	int rv;

	// header
	*obufp=LTPC_RS_HEAD; 
	obufp++; olen++; BUFLIMIT(olen,avail);

	// session originator (SDNV-8) - the ltpaddr in_addr
	sdnvlen=avail-olen; // remaining buffer space
	rv=sdnv8_enc(seg->peer.len,(unsigned char*)seg->peer.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// session number (SDNV-16) - an input
	sdnvlen=avail-olen;
	rv=sdnv16_enc(seg->sno.len,seg->sno.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	int extcount=ltpc_encode_ext_counters(seg);
	if (extcount==-1) LTPERR(LTP_BAD_VALUE);
	*obufp++=extcount; olen++; BUFLIMIT(olen,avail);
	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,true);
	if (rv) LTPERR(rv);

	// rsn
	sdnvlen=avail-olen;
	rv=sdnv8_enc(rs->rsn.len,rs->rsn.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// csn
	sdnvlen=avail-olen;
	rv=sdnv8_enc(rs->csn.len,rs->csn.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// ub
	sdnvlen=avail-olen;
	rv=sdnv16_enc_int(rs->ub,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// lb
	sdnvlen=avail-olen;
	rv=sdnv16_enc_int(rs->lb,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// cc - SDNV-8
	sdnvlen=avail-olen;
	rv=sdnv8_enc_int(rs->cc,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	for (int i=0;i!=rs->cc;i++) {
			// offset[i]
			sdnvlen=avail-olen;
			rv=sdnv16_enc_int(rs->offsets[i],&sdnvlen,obufp);
			if (rv) LTPERR(rv);
			obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

			// lengths[i]
			sdnvlen=avail-olen;
			rv=sdnv16_enc_int(rs->lengths[i],&sdnvlen,obufp);
			if (rv) LTPERR(rv);
			obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
	}

	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,false);
	if (rv) LTPERR(rv);

	*obuflen=olen;
	return(LTP_OK);
}

/*!
 * @brief encode a report ack segment
 * @param seg is an abstract report ack segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode_ra( ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet)
{
	ltpc_ra	*ra=&seg->val.ra;
	int olen=0;
	unsigned char *obufp=obuf;
	int avail=*obuflen;
	int sdnvlen;
	int rv;

	// header
	*obufp=LTPC_RA_HEAD; 
	obufp++; olen++; BUFLIMIT(olen,avail);

	// session originator (SDNV-8) - the ltpaddr in_addr
	sdnvlen=avail-olen; // remaining buffer space
	rv=sdnv8_enc(seg->peer.len,(unsigned char*)seg->peer.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	// session number (SDNV-16) - an input
	sdnvlen=avail-olen;
	rv=sdnv16_enc(seg->sno.len,seg->sno.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	int extcount=ltpc_encode_ext_counters(seg);
	if (extcount==-1) LTPERR(LTP_BAD_VALUE);
	*obufp++=extcount; olen++; BUFLIMIT(olen,avail);
	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,true);
	if (rv) LTPERR(rv);

	// rsn (SDNV-8) - an input
	sdnvlen=avail-olen;
	rv=sdnv8_enc(ra->rsn.len,ra->rsn.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,false);
	if (rv) LTPERR(rv);

	*obuflen=olen;
	return(LTP_OK);
}

/*!
 * @brief encode a cancel segment
 * @param seg is an abstract cancel segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode_co( ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet)
{
	ltpc_co	*c=&seg->val.co;
	int olen=0;
	unsigned char *obufp=obuf;
	int avail=*obuflen;
	int sdnvlen;
	int rv;
	// header
	*obufp=LTPC_CO_HEAD; 
	obufp++; olen++; BUFLIMIT(olen,avail);
	// session originator (SDNV-8) - the ltpaddr in_addr
	sdnvlen=avail-olen; // remaining buffer space
	rv=sdnv8_enc(seg->peer.len,(unsigned char*)seg->peer.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
	// session number (SDNV-16) - an input
	sdnvlen=avail-olen;
	rv=sdnv16_enc(seg->sno.len,seg->sno.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);

	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
	int extcount=ltpc_encode_ext_counters(seg);
	if (extcount==-1) LTPERR(LTP_BAD_VALUE);
	*obufp++=extcount; olen++; BUFLIMIT(olen,avail);
	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,true);
	if (rv) LTPERR(rv);

	// reasoncode - SDNV-8
	sdnvlen=avail-olen;
	rv=sdnv8_enc_int(c->reasoncode,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,false);
	if (rv) LTPERR(rv);
	
	*obuflen=olen;
	return(LTP_OK);
}

/*!
 * @brief encode a cancel ack segment
 * @param seg is an abstract cancel ack segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode_cao( ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet)
{
	ltpc_cao	*ca=&seg->val.cao;
	int olen=0;
	unsigned char *obufp=obuf;
	int avail=*obuflen;
	int sdnvlen;
	int rv;
	// header
	*obufp=LTPC_CAO_HEAD; 
	obufp++; olen++; BUFLIMIT(olen,avail);
	// session originator (SDNV-8) - the ltpaddr in_addr
	sdnvlen=avail-olen; // remaining buffer space
	rv=sdnv8_enc(seg->peer.len,(unsigned char*)seg->peer.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
	// session number (SDNV-16) - an input
	sdnvlen=avail-olen;
	rv=sdnv16_enc(seg->sno.len,seg->sno.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	int extcount=ltpc_encode_ext_counters(seg);
	if (extcount==-1) LTPERR(LTP_BAD_VALUE);
	*obufp++=extcount; olen++; BUFLIMIT(olen,avail);
	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,true);
	if (rv) LTPERR(rv);

	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,false);
	if (rv) LTPERR(rv);

	*obuflen=olen;
	return(LTP_OK);
}

/*!
 * @brief encode a cancel segment
 * @param seg is an abstract cancel segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode_cr( ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet)
{
	ltpc_cr	*c=&seg->val.cr;
	int olen=0;
	unsigned char *obufp=obuf;
	int avail=*obuflen;
	int sdnvlen;
	int rv;
	// header
	*obufp=LTPC_CR_HEAD; 
	obufp++; olen++; BUFLIMIT(olen,avail);
	// session originator (SDNV-8) - the ltpaddr in_addr
	sdnvlen=avail-olen; // remaining buffer space
	rv=sdnv8_enc(seg->peer.len,(unsigned char*)seg->peer.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
	// session number (SDNV-16) - an input
	sdnvlen=avail-olen;
	rv=sdnv16_enc(seg->sno.len,seg->sno.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	int extcount=ltpc_encode_ext_counters(seg);
	if (extcount==-1) LTPERR(LTP_BAD_VALUE);
	*obufp++=extcount; olen++; BUFLIMIT(olen,avail);
	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,true);
	if (rv) LTPERR(rv);

	// reasoncode - SDNV-8
	sdnvlen=avail-olen;
	rv=sdnv8_enc_int(c->reasoncode,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,false);
	if (rv) LTPERR(rv);
	
	*obuflen=olen;
	return(LTP_OK);
}

/*!
 * @brief encode a cancel ack segment
 * @param seg is an abstract cancel ack segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode_car( ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet)
{
	ltpc_car	*ca=&seg->val.car;
	int olen=0;
	unsigned char *obufp=obuf;
	int avail=*obuflen;
	int sdnvlen;
	int rv;
	// header
	*obufp=LTPC_CAR_HEAD; 
	obufp++; olen++; BUFLIMIT(olen,avail);
	// session originator (SDNV-8) - the ltpaddr in_addr
	sdnvlen=avail-olen; // remaining buffer space
	rv=sdnv8_enc(seg->peer.len,(unsigned char*)seg->peer.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);
	// session number (SDNV-16) - an input
	sdnvlen=avail-olen;
	rv=sdnv16_enc(seg->sno.len,seg->sno.val,&sdnvlen,obufp);
	if (rv) LTPERR(rv);
	obufp+=sdnvlen; olen+=sdnvlen; BUFLIMIT(olen,avail);

	int extcount=ltpc_encode_ext_counters(seg);
	if (extcount==-1) LTPERR(LTP_BAD_VALUE);
	*obufp++=extcount; olen++; BUFLIMIT(olen,avail);
	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,true);
	if (rv) LTPERR(rv);

	rv=ltpc_encode_ext(seg,&obufp,&olen,&avail,fornet,false);
	if (rv) LTPERR(rv);

	*obuflen=olen;
	return(LTP_OK);
}

/*!
 * @brief encode any segment at all
 * @param seg is an abstract data segment structure
 * @param obuflen on input is the size of the buffer to contain the encoded bytes and the actual encoded size on output
 * @param obuf is the buffer to contain the encoded bytes
 * @param fornet determines whether encoding is for network or file system (when buffer isn't needed)
 * @return zero or error
 */
int ltpc_encode(ltpc_seg *seg, int *obuflen, unsigned char *obuf, bool fornet)
{
	int rv;

	if (seg->version!=LTP_VERSION) LTPERR(LTP_BAD_VALUE);
	switch (seg->type) {
		case LTPC_SEG_DS:
			rv=ltpc_encode_ds(seg, obuflen, obuf, fornet); if (rv) LTPERR(rv); break;
		case LTPC_SEG_RS:
			rv=ltpc_encode_rs(seg, obuflen, obuf, fornet); if (rv) LTPERR(rv); break;
		case LTPC_SEG_RA:
			rv=ltpc_encode_ra(seg, obuflen, obuf, fornet); if (rv) LTPERR(rv); break;
		case LTPC_SEG_CO:
			rv=ltpc_encode_co(seg, obuflen, obuf, fornet); if (rv) LTPERR(rv); break;
		case LTPC_SEG_CAO:
			rv=ltpc_encode_cao(seg, obuflen, obuf, fornet); if (rv) LTPERR(rv); break;
		case LTPC_SEG_CR:
			rv=ltpc_encode_cr(seg, obuflen, obuf, fornet); if (rv) LTPERR(rv); break;
		case LTPC_SEG_CAR:
			rv=ltpc_encode_car(seg, obuflen, obuf, fornet); if (rv) LTPERR(rv); break;
		default:
			LTPERR(LTP_BAD_VALUE);
	}
	return(LTP_OK);
}

/*!
 * @section g4 decoding functions
 */


/*!
 * @brief decode a DS
 * @param buf is what you might think!
 * @param seg is the output data segment
 * @param seglen is the length of the segment within the buffer (which may be too long or short!)
 * @return zero or error
 */
int ltpc_decode_ds(ltpc_octets *buf, ltpc_seg *seg, int *seglen)
{
	ltpc_ds *ds=&seg->val.ds;
	bool eob,checkpoint,retransmit;
	unsigned char *bv=buf->val;
	int len=buf->len;

	switch(bv[0]) {
			case 0:	 ds->red=true; ds->checkpoint=false; ds->eorp=false; ds->eob=false; break;
			case 1:	 ds->red=true; ds->checkpoint=true; ds->eorp=false; ds->eob=false; break;
			case 2:	 ds->red=true; ds->checkpoint=true; ds->eorp=true; ds->eob=false; break;
			case 3:	 ds->red=true; ds->checkpoint=true; ds->eorp=true; ds->eob=true; break;
			case 4:	 ds->red=false; ds->checkpoint=false; ds->eorp=false; ds->eob=false; break;
			case 7:	 ds->red=false; ds->checkpoint=false; ds->eorp=false; ds->eob=true; break;
			default:	 // shouldn't happen - overall decoder checked already
					LTPERR(LTP_BAD_VALUE);
	}

	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);

	// get session originator (peer) SDNV-8
	int rv=sdnv8_dec(&len,bv,&seg->peer);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// get session number (sno) SDNV-16
	rv=sdnv16_dec(&len,bv,&seg->sno);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// extension handling
	int headers=0;
	int trailers=0;
	rv=ltpc_decode_ext_counters(*bv,&headers,&trailers,seg);
	if (rv) LTPERR(LTP_BAD_VALUE);
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	rv=ltpc_decode_ext(bv,&len,seg->exts,headers,true);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// get client service id (csi) SNNV-8
	rv=sdnv8_dec(&len,bv,&ds->csi);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	// get offset  SNDV-16
	rv=sdnv16_dec_int(&len,bv,&ds->offset);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	// get length SNDV-16
	ltpc_octets foo;
	rv=sdnv16_dec(&len,bv,&foo);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	if (ds->checkpoint) {
			// get csn  SDNV-8
			rv=sdnv8_dec(&len,bv,&ds->csn);
			if (rv) LTPERR(rv);
			bv=buf->val+(buf->len-len);
			// rsn (maybe zero) SDNV-8
			rv=sdnv8_dec(&len,bv,&ds->rsn);
			if (rv) LTPERR(rv);
			bv=buf->val+(buf->len-len);
			// retransmit flag is set if rsn was non-zero
			if (!ds->rsn.len || !ds->rsn.val ||  (ds->rsn.len==1 && ds->rsn.val[0]==0x00)) ds->retransmit=false;
			else ds->retransmit=true;
	}

	// get data
	ds->bytes.len=0;
	if (foo.len==0 || foo.len>sizeof(int)) LTPERR(LTPC_SEG_CRAP);
	for (int jj=0;jj!=foo.len-1;jj++) {
		ds->bytes.len+=(foo.val[jj]<<8);
	}
	ds->bytes.len+=foo.val[foo.len-1];
	ltpc_free(foo.val);
	// now consume bytes_length from buffer
	ds->bytes.val=ltpc_alloc(ds->bytes.len);
	memcpy(ds->bytes.val,bv,ds->bytes.len);
	bv+=ds->bytes.len;

	len=buf->len-(bv-buf->val);
	rv=ltpc_decode_ext(bv,&len,&seg->exts[headers],trailers,false);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	*seglen=(bv-buf->val);
	
	return(LTP_OK);
}

/*!
 * @brief decode a RS
 * @param buf is what you might think!
 * @param seg is the output report segment
 * @param seglen is the length of the segment within the buffer (which may be too long or short!)
 * @return zero or error
 */
int ltpc_decode_rs(ltpc_octets *buf, ltpc_seg *seg, int *seglen)
{
	ltpc_rs *rs=&seg->val.rs;
	unsigned char *bv=buf->val;
	int len=buf->len;

	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);

	// get session originator (peer) SDNV-8
	int rv=sdnv8_dec(&len,bv,&seg->peer);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// get session number (sno) SDNV-16
	rv=sdnv16_dec(&len,bv,&seg->sno);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// extension handling
	int headers=0;
	int trailers=0;
	rv=ltpc_decode_ext_counters(*bv,&headers,&trailers,seg);
	if (rv) LTPERR(LTP_BAD_VALUE);
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	rv=ltpc_decode_ext(bv,&len,seg->exts,headers,true);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// ltpc_octets		rsn;
	rv=sdnv8_dec(&len,bv,&rs->rsn);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// ltpc_octets		csn;
	rv=sdnv8_dec(&len,bv,&rs->csn);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// int		ub; 16!!!
	rv=sdnv16_dec_int(&len,bv,&rs->ub);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// int		lb;
	rv=sdnv16_dec_int(&len,bv,&rs->lb);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// int				cc;
	rv=sdnv8_dec_int(&len,bv,&rs->cc);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	rs->offsets=(int*)ltpc_alloc(sizeof(int)*rs->cc);
	if (!rs->offsets) LTPERR(LTP_MEMORY);
	rs->lengths=(int*)ltpc_alloc(sizeof(int)*rs->cc);
	if (!rs->lengths) LTPERR(LTP_MEMORY);

	for (int i=0;i!=rs->cc;i++) {
		// ltpc_octets		*offsets;
		rv=sdnv16_dec_int(&len,bv,&rs->offsets[i]);
		if (rv) LTPERR(rv);
		bv=buf->val+(buf->len-len);

		// ltpc_octets		*lengths;
		rv=sdnv16_dec_int(&len,bv,&rs->lengths[i]);
		if (rv) LTPERR(rv);
		bv=buf->val+(buf->len-len);
	}

	len=buf->len-(bv-buf->val);
	rv=ltpc_decode_ext(bv,&len,&seg->exts[headers],trailers,false);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	*seglen=(bv-buf->val);
	return(LTP_OK);
}

/*!
 * @brief decode a RA
 * @param buf is what you might think!
 * @param seg is the output report ack segment
 * @param seglen is the length of the segment within the buffer (which may be too long or short!)
 * @return zero or error
 */
int ltpc_decode_ra(ltpc_octets *buf, ltpc_seg *seg, int *seglen)
{
	ltpc_ra *ra=&seg->val.ra;
	unsigned char *bv=buf->val;
	int len=buf->len;

	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);

	// get session originator (peer) SDNV-8
	int rv=sdnv8_dec(&len,bv,&seg->peer);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// get session number (sno) SDNV-16
	rv=sdnv16_dec(&len,bv,&seg->sno);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// extension handling
	int headers=0;
	int trailers=0;
	rv=ltpc_decode_ext_counters(*bv,&headers,&trailers,seg);
	if (rv) LTPERR(LTP_BAD_VALUE);
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	rv=ltpc_decode_ext(bv,&len,seg->exts,headers,true);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// ltpc_octets		rsn;
	rv=sdnv8_dec(&len,bv,&ra->rsn);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	len=buf->len-(bv-buf->val);
	rv=ltpc_decode_ext(bv,&len,&seg->exts[headers],trailers,false);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	*seglen=(bv-buf->val);
	return(LTP_OK);
}

/*!
 * @brief decode a C
 * @param buf is what you might think!
 * @param seg is the output cancel segment
 * @param seglen is the length of the segment within the buffer (which may be too long or short!)
 * @return zero or error
 */
int ltpc_decode_co(ltpc_octets *buf, ltpc_seg *seg, int *seglen)
{
	ltpc_co *c=&seg->val.co;
	unsigned char *bv=buf->val;
	int len=buf->len;
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	// get session originator (peer) SDNV-8
	int rv=sdnv8_dec(&len,bv,&seg->peer);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	// get session number (sno) SDNV-16
	rv=sdnv16_dec(&len,bv,&seg->sno);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// extension handling
	int headers=0;
	int trailers=0;
	rv=ltpc_decode_ext_counters(*bv,&headers,&trailers,seg);
	if (rv) LTPERR(LTP_BAD_VALUE);
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	rv=ltpc_decode_ext(bv,&len,seg->exts,headers,true);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// int				reasoncode;
	ltpc_octets foo;
	rv=sdnv8_dec(&len,bv,&foo);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	c->reasoncode=0;
	if (foo.len==0 || foo.len>sizeof(int)) LTPERR(LTPC_SEG_CRAP);
	for (int jj=0;jj!=foo.len-1;jj++) {
		c->reasoncode+=(foo.val[jj]<<8);
	}
	c->reasoncode+=foo.val[foo.len-1];

	len=buf->len-(bv-buf->val);
	rv=ltpc_decode_ext(bv,&len,&seg->exts[headers],trailers,false);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	*seglen=(bv-buf->val);
	return(LTP_OK);
}

/*!
 * @brief decode a CA
 * @param buf is what you might think!
 * @param seg is the output cancel ack segment
 * @param seglen is the length of the segment within the buffer (which may be too long or short!)
 * @return zero or error
 */
int ltpc_decode_cao(ltpc_octets *buf, ltpc_seg *seg, int *seglen)
{
	ltpc_cao *ca=&seg->val.cao;
	unsigned char *bv=buf->val;
	int len=buf->len;
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	// get session originator (peer) SDNV-8
	int rv=sdnv8_dec(&len,bv,&seg->peer);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	// get session number (sno) SDNV-16
	rv=sdnv16_dec(&len,bv,&seg->sno);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// extension handling
	int headers=0;
	int trailers=0;
	rv=ltpc_decode_ext_counters(*bv,&headers,&trailers,seg);
	if (rv) LTPERR(LTP_BAD_VALUE);
	len--;bv++; if ((headers||trailers) & !len) LTPERR(LTPC_SEG_TRUNCATED);
	if (headers) {
		rv=ltpc_decode_ext(bv,&len,seg->exts,headers,true);
		if (rv) LTPERR(rv);
		bv=buf->val+(buf->len-len);
	}
	if (trailers) {
		len=buf->len-(bv-buf->val);
		rv=ltpc_decode_ext(bv,&len,&seg->exts[headers],trailers,false);
		if (rv) LTPERR(rv);
		bv=buf->val+(buf->len-len);
	}

	*seglen=(bv-buf->val);
	return(LTP_OK);
}

/*!
 * @brief decode a C
 * @param buf is what you might think!
 * @param seg is the output cancel segment
 * @param seglen is the length of the segment within the buffer (which may be too long or short!)
 * @return zero or error
 */
int ltpc_decode_cr(ltpc_octets *buf, ltpc_seg *seg, int *seglen)
{
	ltpc_cr *c=&seg->val.cr;
	unsigned char *bv=buf->val;
	int len=buf->len;
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	// get session originator (peer) SDNV-8
	int rv=sdnv8_dec(&len,bv,&seg->peer);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	// get session number (sno) SDNV-16
	rv=sdnv16_dec(&len,bv,&seg->sno);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// extension handling
	int headers=0;
	int trailers=0;
	rv=ltpc_decode_ext_counters(*bv,&headers,&trailers,seg);
	if (rv) LTPERR(LTP_BAD_VALUE);
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	rv=ltpc_decode_ext(bv,&len,seg->exts,headers,true);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// int				reasoncode;
	ltpc_octets foo;
	rv=sdnv8_dec(&len,bv,&foo);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	c->reasoncode=0;
	if (foo.len==0 || foo.len>sizeof(int)) LTPERR(LTPC_SEG_CRAP);
	for (int jj=0;jj!=foo.len-1;jj++) {
		c->reasoncode+=(foo.val[jj]<<8);
	}
	c->reasoncode+=foo.val[foo.len-1];

	len=buf->len-(bv-buf->val);
	rv=ltpc_decode_ext(bv,&len,&seg->exts[headers],trailers,false);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	*seglen=(bv-buf->val);
	return(LTP_OK);
}

/*!
 * @brief decode a CA
 * @param buf is what you might think!
 * @param seg is the output cancel ack segment
 * @param seglen is the length of the segment within the buffer (which may be too long or short!)
 * @return zero or error
 */
int ltpc_decode_car(ltpc_octets *buf, ltpc_seg *seg, int *seglen)
{
	ltpc_car *ca=&seg->val.car;
	unsigned char *bv=buf->val;
	int len=buf->len;
	len--;bv++; if (!len) LTPERR(LTPC_SEG_TRUNCATED);
	// get session originator (peer) SDNV-8
	int rv=sdnv8_dec(&len,bv,&seg->peer);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);
	// get session number (sno) SDNV-16
	rv=sdnv16_dec(&len,bv,&seg->sno);
	if (rv) LTPERR(rv);
	bv=buf->val+(buf->len-len);

	// extension handling
	int headers=0;
	int trailers=0;
	rv=ltpc_decode_ext_counters(*bv,&headers,&trailers,seg);
	if (rv) LTPERR(LTP_BAD_VALUE);
	len--;bv++; if ((headers||trailers) & !len) LTPERR(LTPC_SEG_TRUNCATED);
	if (headers) {
		rv=ltpc_decode_ext(bv,&len,seg->exts,headers,true);
		if (rv) LTPERR(rv);
		bv=buf->val+(buf->len-len);
	}
	if (trailers) {
		len=buf->len-(bv-buf->val);
		rv=ltpc_decode_ext(bv,&len,&seg->exts[headers],trailers,false);
		if (rv) LTPERR(rv);
		bv=buf->val+(buf->len-len);
	}

	*seglen=(bv-buf->val);
	return(LTP_OK);
}

/*!
 * @brief first level decoding of a buffer - see what type of segment it might be
 * @param len is the buffer length on input, and the remaining buffer length on output (zero if no more)
 * @param val is (a ptr to) the buffer on input, and points to the byte after the end of the 1st complete segment on output (NULL if no more)
 * @param seg is a (caller allocated) structure for the decoded stuff
 * @return is the type of segment e.g. LTPC_SEG_DS etc.
 */
int ltpc_decode(int *len, unsigned char **val, ltpc_seg *seg)
{
	if (!seg || !len ||  !val) LTPERR(LTP_BAD_PARAM);
	if (*len < LTPC_SEG_MIN_LEN) LTPERR(LTPC_SEG_TRUNCATED);
	ltpc_octets buf;
	buf.len=*len;
	buf.val=*val;
	int rtype;
	unsigned char head=buf.val[0];

	seg->version=LTP_VERSION_EXTRACTOR(head);
	if (seg->version!=LTP_VERSION) {
		return(LTPC_SEG_CRAP);
	}
	head &= LTP_VERSION_MASK;

	// only a silly set of values are good - this'll change over time!
	switch (head) {
			// a bunch of DS cases
			case 0: // fall through
			case 1: // fall through
			case 2: // fall through
			case 3: // fall through
			case 4: // fall through
			case 7: // fall through
					rtype=LTPC_SEG_DS; break;
			case LTPC_RS_HEAD: rtype=LTPC_SEG_RS; break;
			case LTPC_RA_HEAD: rtype=LTPC_SEG_RA; break;
			case LTPC_CO_HEAD: rtype=LTPC_SEG_CO; break;
			case LTPC_CAO_HEAD: rtype=LTPC_SEG_CAO; break;
			case LTPC_CR_HEAD: rtype=LTPC_SEG_CR; break;
			case LTPC_CAR_HEAD: rtype=LTPC_SEG_CAR; break;
			default:
					rtype=LTPC_SEG_CRAP;
					break;
	}
	int inner_rv;
	int inner_len;
	switch (rtype) {
		case LTPC_SEG_DS:
				inner_rv=ltpc_decode_ds(&buf,seg,&inner_len);
				break;
		case LTPC_SEG_RS:
				inner_rv=ltpc_decode_rs(&buf,seg,&inner_len);
				break;
		case LTPC_SEG_RA:
				inner_rv=ltpc_decode_ra(&buf,seg,&inner_len);
				break;
		case LTPC_SEG_CO:
				inner_rv=ltpc_decode_co(&buf,seg,&inner_len);
				break;
		case LTPC_SEG_CAO:
				inner_rv=ltpc_decode_cao(&buf,seg,&inner_len);
				break;
		case LTPC_SEG_CR:
				inner_rv=ltpc_decode_cr(&buf,seg,&inner_len);
				break;
		case LTPC_SEG_CAR:
				inner_rv=ltpc_decode_car(&buf,seg,&inner_len);
				break;
		default:
				return(LTPC_SEG_CRAP);
	}
	if (inner_rv) LTPERR(inner_rv);
	seg->type=rtype;
	*len-=inner_len;
	if (*len>0) 
		*val+=inner_len;
	else 
		*val=NULL;
	return(rtype); // pessimism !!
}

/*!
 * @brief utilituy free fnc in case we use different memory models or whatever.
 * @param foo is the thing to be freed - should have been alloc'd with ltpc_alloc()
 */
void ltpc_free(unsigned char *foo)
{
	if (foo) free(foo);
	return;
}

/*!
 * @brief utilituy malloc fnc in case we use different memory models or whatever.
 * @param len is the amount of space to allocate (won't be zero'd!)
 * @return is the pointer to the allocated space
 */
unsigned char* ltpc_alloc(size_t len)
{
	unsigned char *foo=(unsigned char*)calloc(1,len);
	return(foo);
}

/*!
 * @brief utilituy realloc fnc in case we use different memory models or whatever.
 * @param len is the amount of space to allocate (won't be zero'd!)
 * @return is the pointer to the allocated space
 */
unsigned char* ltpc_realloc(unsigned char *ptr,size_t len)
{
	unsigned char *foo=(unsigned char*)realloc(ptr,len);
	return(foo);
}

#define LTPC_FREE_COMMON(foo) \
	ltpc_free(foo->peer.val); \
	ltpc_free(foo->sno.val); \
	ltpc_free_seg_exts(foo);

/*!
 * @brief free up all the extensions in an ltpc_seg
 * @param seg is the segment to free up stuff from
 */
void ltpc_free_seg_exts(ltpc_seg *seg)
{
	for (int i=0;i!=seg->nexts;i++) {
		ltpe_free_ext(&seg->exts[i]); 
		ltpc_free((unsigned char*)seg->exts[i].val); 
		ltpc_free(seg->exts[i].encval); 
	}
	ltpc_free((unsigned char*) seg->exts);
	seg->nexts=0;
	return;
}

/*!
 * @brief free up everything inside an ltpc_seg
 * @param seg is the segment to free up.
 */
void ltpc_free_seg(ltpc_seg *seg)
{
	if (!seg) VOIDLTPERR(LTP_BAD_PARAM);
	switch (seg->type) {
		case LTPC_SEG_DS:
				LTPC_FREE_COMMON(seg)
				ltpc_free(seg->val.ds.csn.val);
				ltpc_free(seg->val.ds.rsn.val);
				ltpc_free(seg->val.ds.csi.val);
				ltpc_free(seg->val.ds.bytes.val);
				break;
		case LTPC_SEG_RS:
				LTPC_FREE_COMMON(seg)
				ltpc_free(seg->val.rs.rsn.val);
				ltpc_free(seg->val.rs.csn.val);
				ltpc_free((unsigned char*)seg->val.rs.offsets);
				ltpc_free((unsigned char*)seg->val.rs.lengths);
				break;
		case LTPC_SEG_RA:
				LTPC_FREE_COMMON(seg)
				ltpc_free(seg->val.rs.rsn.val);
				break;
		case LTPC_SEG_CO:
				LTPC_FREE_COMMON(seg)
				break;
		case LTPC_SEG_CAO:
				LTPC_FREE_COMMON(seg)
				break;
		case LTPC_SEG_CR:
				LTPC_FREE_COMMON(seg)
				break;
		case LTPC_SEG_CAR:
				LTPC_FREE_COMMON(seg)
				break;
		default:
				return;
	}
	return;
}

/*!
 * @brief map from ltpc_octets to an int if possible
 * @param oct is the (ptr to an) ltpc_octets str
 * @param val is the (ptr to the) int output
 * @return zero or error
 */
int ltpc_oct2int(ltpc_octets *oct, int *val)
{
	if (!oct || !val) LTPERR(LTP_BAD_PARAM);
	if (oct->len>sizeof(int)) LTPERR(LTPC_SEG_CRAP);
	*val=0;
	if (oct->len==0) return(LTP_OK); 
	for (int i=0;i!=oct->len-1;i++) {
		*val += oct->val[i];
		*val = *val<<8;
	}
	*val += oct->val[oct->len-1];
	return(LTP_OK);
}

/*!
 * @brief make a copy of an ltpc_octet
 * @param to is the destinatin
 * @param from is the source
 * @return zero or error
 */
int ltpc_oct_cpy(ltpc_octets *to, ltpc_octets *from)
{
	if (!from->len && from->val) LTPERR(LTP_BAD_PARAM);
	if (from->len && !from->val) LTPERR(LTP_BAD_PARAM);
	to->len=from->len;
	if (!to->len) {
		to->val=NULL;
	} else {
		to->val=ltpc_alloc(to->len);
		if (!to->val) LTPERR(LTP_MEMORY);
		memcpy(to->val,from->val,to->len);
	}
	return(LTP_OK);
}

/*!
 * @fn int sdnv_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf)
 * @brief Encode up an SDNV-8 from an input buffer
 * @param ilen is the length of the input
 * @param ibuf is the input buffer
 * @param olen is the size of the uutput buffer on input and the size of the SDNV on output
 * @param obuf is the caller allocated buffer into which the SDNV-8 should be written
 * @return zero or error as usual 
 *
 * Encode up an SDNV. See the LTP specification for details. 
 */
int sdnv_enc(int ilen, unsigned char *ibuf, int *olen, unsigned char *obuf)
{
	return(sdnv_enc_aux(ilen,ibuf,olen,obuf,8));
}

/*!
 * @brief decode an sdnv-8
 * @param blen is the length of the input buffer
 * @param buf is the input buffer
 * @param out is the output len/buffer
 * @return zero or error
 */
int sdnv_dec(int *blen, unsigned char *buf, int *olen, unsigned char *obuf)
{
	ltpc_octets out;
	out.len=*olen;
	int rv=sdnv_dec_aux(blen,buf,&out,8);
	if (rv) LTPERR(rv);
	memcpy(obuf,out.val,out.len);
	ltpc_free(out.val);
	*olen=out.len;
	return(LTP_OK);
}

/*!
 * @brief utility fnc that encodes an unsigned int an network byte order using as few bytes as possible
 * @param val is the to-be-encoded value
 * @param blen the buffer length and used buffer on output
 * @param buf is the buffer for the encoding
 * @return zero or error
 */
int ltpc_small_htonl(unsigned int val, int *blen, unsigned char *buf)
{
	if (!blen || !buf || *blen<4) LTPERR(LTP_BAD_PARAM);
	int lval=val;
	if (!(lval&0xff000000)) { lval=lval<<8; *blen=3;}
	if (!(lval&0xff000000)) { lval=lval<<8; *blen=2;}
	if (!(lval&0xff000000)) { lval=lval<<8; *blen=1;}
	switch (*blen) { 
		case 4: buf[0]=val>>24;buf[1]=val>>16;buf[2]=val>>8;buf[0]=val;break;
		case 3: buf[0]=val>>16;buf[1]=val>>8;buf[2]=val;break;
		case 2: buf[0]=val>>8;buf[1]=val;break;
		case 1: buf[0]=val;break;
			break;
		default: LTPERR(LTP_BAD_VALUE);
	}
	return(LTP_OK);
}

/*!
 * @brief binary to ascii hex converter
 * @param buf if the binary buffer
 * @param len is the buffer length
 * @param str is the caller allocated string buffer (which is sufficiently long)
 * @param slen is the length of the string buffer
 * @return zero or error
 */
int bin2AH(unsigned char *buf, int len, char *str,int slen)
{
	if (!len && slen) { // empty stuff
		snprintf(str,slen,"(null)");
		return(LTP_OK); 
	}
	if (!buf || !len || !str) return(LTP_BAD_PARAM);
	if ((2*len+3)>slen) return(LTP_BAD_VALUE);
	int foo;
	str[0]='0';
	str[1]='x';
 	for (int i=0;i!=len;i++) {
		foo=(buf[i]>>4);
		str[2*i+2]=HEXDIGIT(foo);
		foo=buf[i]&0x0f;
		str[2*i+3]=HEXDIGIT(foo);
	}
	str[2*len+2]='\0';
	return(LTP_OK);
}

/*!
 * @brief pretty print an encoded message into a buffer
 * @param msg is the encoded message
 * @param msglen is the encoded message length
 * @param buffer is the (called allocated) buffer 
 * @param len is the length of the output
 * @return zero or error
 *
 * In order to help increase the liklihood of finding bugs this code is 
 * genetically different from the encoder/decoder code (well, to the extent 
 * that one person can do that:-). The idea is that I'm less likely to make the
 * same mistake twice in the same place. We'll see.
 *
 * Well, it seems to have worked! Turned out I was putting in the csn/rsn
 * in the wrong place in checkpoint DS's! 
 */
int ltpc_pp(unsigned char *msg, int msglen, char *buffer, int *len)
{
	// type 
	if (!msglen || !msg || !buffer || !len || *len<=1) LTPERR(LTP_BAD_PARAM);
	unsigned char buf1[MAXPATH]; // this'll be used a few times I bet
	unsigned char buf2[MAXPATH]; // another
	unsigned char buf3[MAXPATH]; // another
	memset(buf1,0,MAXPATH); memset(buf2,0,MAXPATH); memset(buf2,0,MAXPATH); // try silence valgrind
	int sl=0;
	char *bp=buffer;
	int lenleft=*len;
	snprintf(bp,lenleft,"PP: Pretty printed LTP message (at %s).\nPP: -----\n",ltp_now_str()); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
	if (lenleft<=0) LTPERR(LTP_MEMORY);
	snprintf(bp,lenleft,"PP: message type: "); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
	if (lenleft<=0) LTPERR(LTP_MEMORY);
	int msgtype=0;
	int mlen=msglen;
	bool checkpoint=false; // changes decoding of DS body
	switch (msg[0]) {
		// checkpoints 1st
		case 1: // fall through
		case 2: // fall through
		case 3: // fall through
					checkpoint=true;
		case 0: // fall through
		case 4: // fall through
		case 7: // fall through
			msgtype=LTPC_SEG_DS;
			snprintf(bp,lenleft,"Data segment\t\t\t\t\t\t\t[%d]\n",msg[0]); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp); break;
			if (lenleft<=0) LTPERR(LTP_MEMORY);
		case LTPC_RS_HEAD: 
			msgtype=LTPC_SEG_RS;
			snprintf(bp,lenleft,"Report segment\n"); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp); break;
			if (lenleft<=0) LTPERR(LTP_MEMORY);
		case LTPC_RA_HEAD: 
			msgtype=LTPC_SEG_RA;
			snprintf(bp,lenleft,"Report acknowledgement\n"); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp); break;
			if (lenleft<=0) LTPERR(LTP_MEMORY);
		case LTPC_CO_HEAD: 
			msgtype=LTPC_SEG_CO;
			snprintf(bp,lenleft,"Cancel from originator\n"); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp); break;
			if (lenleft<=0) LTPERR(LTP_MEMORY);
		case LTPC_CAO_HEAD: 
			msgtype=LTPC_SEG_CAO;
			snprintf(bp,lenleft,"Cancel acknowledgement for originator\n"); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp); break;
			if (lenleft<=0) LTPERR(LTP_MEMORY);
		case LTPC_CR_HEAD: 
			msgtype=LTPC_SEG_CR;
			snprintf(bp,lenleft,"Cancel from receiver\n"); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp); break;
			if (lenleft<=0) LTPERR(LTP_MEMORY);
		case LTPC_CAR_HEAD: 
			msgtype=LTPC_SEG_CAR;
			snprintf(bp,lenleft,"Cancel acknowledgement for receiver\n"); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp); break;
			if (lenleft<=0) LTPERR(LTP_MEMORY);
		default:
			snprintf(bp,lenleft,"dunno!\n"); *len=strlen(buffer); bp+=strlen(bp); return(LTP_BAD_VALUE);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
	}

	// get session originator (peer) SDNV-8
	snprintf(bp,lenleft,"PP: peer (session originator): "); lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
	if (lenleft<=0) LTPERR(LTP_MEMORY);
	ltpc_octets peer;
	peer.len=MAXPATH;
	peer.val=buf1;
	unsigned char *mp=msg+1;
	mlen--;
	sl=mlen;
	int rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
	if (rv) LTPERR(rv);
	sl-=mlen; // this is now the #bytes consumed
	rv=bin2AH(mp,sl,(char*)buf3,MAXPATH);
	if (rv) LTPERR(rv);
	unsigned long foo;
	switch (peer.len) {
		case 4: // bet that its an IP address
			snprintf(bp,lenleft,"PP: IP: %s\t\t[%s]\n",inet_ntoa(*(in_addr*)peer.val),buf3); break;
		case 6: // bet that its an IP + port
			{
				unsigned short bar=(((unsigned short)peer.val[4])<<8)+(unsigned short)(peer.val[5]);
				snprintf(bp,lenleft,"PP: IP: %s:%d\t\t[%s]\n",inet_ntoa(*(in_addr*)peer.val),bar,buf3); break;
				break;
			}
		default: // just damn octets
			rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
			if (rv) LTPERR(rv);
			snprintf(bp,lenleft,"PP: %d Octets %s\t\t\t[%s]\n",peer.len,buf2,buf3);
	}
	mp=msg+(msglen-mlen);
	lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
	if (lenleft<=0) LTPERR(LTP_MEMORY);

	// sno - session number - an sdnv
	peer.len=MAXPATH;
	sl=mlen;
	rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
	if (rv) LTPERR(rv);
	sl-=mlen; // this is now the #bytes consumed
	rv=bin2AH(mp,sl,(char*)buf3,MAXPATH);
	if (rv) LTPERR(rv);
	rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
	if (rv) LTPERR(rv);
	snprintf(bp,lenleft,"PP: sno (session number): %s\t\t\t\t\t[%s]\n",buf2,buf3);
	mp=msg+(msglen-mlen);
	lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
	if (lenleft<=0) LTPERR(LTP_MEMORY);

	// extensions count
	int trailers=*mp&0x0f;
	int headers=*mp>>4;
	snprintf(bp,lenleft,"PP: extensions: %d headers, %d trailers\t\t\t\t\t[0x%02x]\n",headers,trailers,*mp);
	mp++;mlen--;
	lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
	if (lenleft<=0) LTPERR(LTP_MEMORY);
	
	// header extensions
	ltpc_ext	*hexts=NULL;
	hexts=(ltpc_ext*)ltpc_alloc(sizeof(ltpc_ext)*max(headers,trailers));
	if (!hexts) LTPERR(LTP_MEMORY);
	rv=ltpc_decode_ext(mp,&mlen,hexts,headers,true);
	if (rv) LTPERR(rv);
	mp=msg+(msglen-mlen);
	for (int i=0;i!=headers;i++) {
		// print
		buf1[0]='\0';
		rv=ltpe_print(&hexts[i],MAXPATH,(char*)buf1);
		if (rv) LTPERR(rv);
		snprintf(bp,lenleft,"PP: \text %d: %s\n",i,buf1);
		lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
		if (lenleft<=0) LTPERR(LTP_MEMORY);
		// free
		ltpe_free_ext(&hexts[i]);
	}
	
	// now for each thing, its body
	int dlen=0;
	int ub=0;
	int lb=0;
	int claims=0;
	int offset=0;
	int length=0;

	switch (msgtype) {
		case LTPC_SEG_DS:
			// csi
			peer.len=MAXPATH;
			sl=mlen;
			rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
			if (rv) LTPERR(rv);
			rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
			if (rv) LTPERR(rv);
			sl-=mlen; // this is now the #bytes consumed
			rv=bin2AH(mp,sl,(char*)buf3,MAXPATH);
			if (rv) LTPERR(rv);
			snprintf(bp,lenleft,"PP: csi (client service ID): %s\t\t\t\t\t\t[%s]\n",buf2,buf3);
			mp=msg+(msglen-mlen);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			// offset
			sl=mlen;
			rv=sdnv8_dec_int(&mlen,mp,&offset);
			if (rv) LTPERR(rv);
			sl-=mlen; // this is now the #bytes consumed
			rv=bin2AH(mp,sl,(char*)buf3,MAXPATH);
			if (rv) LTPERR(rv);
			snprintf(bp,lenleft,"PP: offset:  %05d (0x%04x)\t\t\t\t\t\t\t\t[%s]\n",offset,offset,buf3);
			mp=msg+(msglen-mlen);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			// length - better be an int or else!
			sl=mlen;
			rv=sdnv8_dec_int(&mlen,mp,&dlen);
			if (rv) LTPERR(rv);
			sl-=mlen; // this is now the #bytes consumed
			rv=bin2AH(mp,sl,(char*)buf3,MAXPATH);
			if (rv) LTPERR(rv);
			snprintf(bp,lenleft,"PP: data length: %05d (0x%04x)\t\t\t\t\t\t\t[%s]\n",dlen,dlen,buf3);
			mp=msg+(msglen-mlen);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			if (checkpoint) {
				// csn
				peer.len=MAXPATH;
				sl=mlen;
				rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
				if (rv) LTPERR(rv);
				sl-=mlen; // this is now the #bytes consumed
				rv=bin2AH(mp,sl,(char*)buf3,MAXPATH);
				if (rv) LTPERR(rv);
				rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
				if (rv) LTPERR(rv);
				snprintf(bp,lenleft,"PP: csn (checkpoint serial number): %s\t\t\t\t[%s]\n",buf2,buf3);
				mp=msg+(msglen-mlen);
				lenleft=*len-strlen(buffer)-1; 
				bp+=strlen(bp);
				if (lenleft<=0) LTPERR(LTP_MEMORY);
				// rsn
				peer.len=MAXPATH;
				sl=mlen;
				rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
				if (rv) LTPERR(rv);
				sl-=mlen; // this is now the #bytes consumed
				rv=bin2AH(mp,sl,(char*)buf3,MAXPATH);
				rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
				if (rv) LTPERR(rv);
				snprintf(bp,lenleft,"PP: rsn (report serial number): %s\t\t\t\t\t[%s]\n",buf2,buf3);
				mp=msg+(msglen-mlen);
				lenleft=*len-strlen(buffer)-1; 
				bp+=strlen(bp);
				if (lenleft<=0) LTPERR(LTP_MEMORY);
			}
			// data - just send out the 1st 4 bytes...
			rv=bin2AH(mp,4,(char*)buf2,MAXPATH);
			if (rv) LTPERR(rv);
			mlen-=dlen;
			mp=msg+(msglen-mlen);
			snprintf(bp,lenleft,"PP: data (1st 4 bytes or less): %s...\n",buf2);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			break;
		case LTPC_SEG_RS:
			// rsn
			peer.len=MAXPATH;
			rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
			if (rv) LTPERR(rv);
			mp=msg+(msglen-mlen);
			rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
			if (rv) LTPERR(rv);
			snprintf(bp,lenleft,"PP: rsn (report serial number): %s\n",buf2);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			// csn
			peer.len=MAXPATH;
			rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
			if (rv) LTPERR(rv);
			mp=msg+(msglen-mlen);
			rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
			if (rv) LTPERR(rv);
			snprintf(bp,lenleft,"PP: csn (checkpoint serial number): %s\n",buf2);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			// ub
			rv=sdnv8_dec_int(&mlen,mp,&ub);
			if (rv) LTPERR(rv);
			mp=msg+(msglen-mlen);
			snprintf(bp,lenleft,"PP: ub (upper bound): %d (0x%x)\n",ub,ub);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			// lb
			rv=sdnv8_dec_int(&mlen,mp,&lb);
			if (rv) LTPERR(rv);
			mp=msg+(msglen-mlen);
			snprintf(bp,lenleft,"PP: lb (lower bound): %d (0x%x)\n",lb,lb);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			// claim count
			rv=sdnv8_dec_int(&mlen,mp,&claims);
			if (rv) LTPERR(rv);
			mp=msg+(msglen-mlen);
			snprintf(bp,lenleft,"PP: claims (claim count): %d\n",claims);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			for (int i=0;i!=claims;i++) {
				// DODGY - lenleft calc is wrong
				// offset
				rv=sdnv8_dec_int(&mlen,mp,&offset);
				if (rv) LTPERR(rv);
				mp=msg+(msglen-mlen);
				snprintf(bp,lenleft,"PP: \toffset: %d (0x%x)\n",offset,offset);
				lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
				if (lenleft<=0) LTPERR(LTP_MEMORY);
				// length
				rv=sdnv8_dec_int(&mlen,mp,&length);
				if (rv) LTPERR(rv);
				mp=msg+(msglen-mlen);
				snprintf(bp,lenleft,"PP: \tlength: %d (0x%x)\n",length,length);
				lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
				if (lenleft<=0) LTPERR(LTP_MEMORY);
			}
			break;
		case LTPC_SEG_RA:
			// rsn
			peer.len=MAXPATH;
			rv=sdnv_dec(&mlen,mp,&peer.len,peer.val);
			if (rv) LTPERR(rv);
			mp=msg+(msglen-mlen);
			rv=bin2AH(peer.val,peer.len,(char*)buf2,MAXPATH);
			if (rv) LTPERR(rv);
			snprintf(bp,lenleft,"PP: rsn (report serial number): %s\n",buf2);
			lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
			if (lenleft<=0) LTPERR(LTP_MEMORY);
			break;
		case LTPC_SEG_CO:
		case LTPC_SEG_CR:
		case LTPC_SEG_CAO:
		case LTPC_SEG_CAR:
		default:
			;
	}

	// trailer extensions
	rv=ltpc_decode_ext(mp,&mlen,hexts,trailers,false);
	if (rv) LTPERR(rv);
	mp=msg+(msglen-mlen);
	for (int i=0;i!=trailers;i++) {
		// print
		buf1[0]='\0';
		rv=ltpe_print(&hexts[i],MAXPATH,(char*)buf1);
		if (rv) LTPERR(rv);
		snprintf(bp,lenleft,"PP: \text %d: %s\n",i,buf1);
		lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
		if (lenleft<=0) LTPERR(LTP_MEMORY);
		// free
		ltpe_free_ext(&hexts[i]);
	}

	ltpc_free((unsigned char*)hexts);
	
	snprintf(bp,lenleft,"PP:\nPP: -----\n"); 
	lenleft=*len-strlen(buffer)-1; bp+=strlen(bp);
	if (lenleft<=0) LTPERR(LTP_MEMORY);
	*len=strlen(buffer);
	return(LTP_OK);
}

/*! 
 * @brief octets 2 int
 * @param octets is an ltpc_octets ptr
 * @return is an integer (-1 is used for dodgy stuff)
 */
int ltpc_oct2int_r(ltpc_octets *in)
{
	int iout;
	int rv=ltpc_oct2int(in,&iout);
	if (rv) {
		iout=-1;
	}
	return(iout);
}
