/*!
 * @file ltp_crypt.cc
 * @brief Codec for specific (actually, all-known) LTP extensions
 * @version $Revision: 1.21 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This API handles key management and use of crypto libraries
 * (in this case openssl by default). The intent is that it
 * be replaceable if someone wants to use a different (typically
 * more complex) key management scheme and/or crypto library.
 *
 * There is a DUMMYCRYPTO option where no real stuff happens
 * which is handy for debugging the protocol handler.
 *
 * For non DUMMYCRYPTO cases I use openssl. There's a useful
 * HOTWO for that at: http://www.madboa.com/geek/openssl/
 *
 * The key management scheme used here is as follows:
 *
 * For symmetric key cases, we map 1:1 between LTP peers and
 * stored symmetric keys.
 *
 * The ltpd crypto configuration file should contain a line 
 * for these like:
 *
 * 0,s,fred,fred.key
 *
 * For asymmetric cases, we store a public key for each peer
 * and an optional private key.
 * The format used is an openssl self-signed cert file that
 * can be generated as follows:
 *
 * openssl req -x509 -nodes -days 365 \
 *    -newkey rsa:1024 -keyout mykey.pem -out mycert.pem
 * 
 * The corresponding ltpd crypto configuration file should 
 * contain a line like:
 * 
 * 1,s,down-ltp,mycert.pem,mykey.pem
 *
 * I've started from openssl version 0.9.8c - others and YMMV
 *
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <ctype.h> // for toupper()
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <openssl/hmac.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/objects.h>
#include <openssl/x509.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>

#include "ltp_base.h"
#include "ltp.h" // for ltp_auth_so definition
#include "ltp_crypt.h" // for ltp_auth_so definition
#include "ltp_ext.h" // for ciphersuite definitions
#include "ltp_names.h" 


/// for test purposes we can use dummy stuff - it allows us to debug all the extension handling
#undef DUMMYCRYPTO

/// the length of a dummy crypto check
#define DUMMYLEN 10

/// the value of a dummy crypto check
unsigned char dummybuf[DUMMYLEN]={0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a};

#ifdef DUMMYCRYPTO
/// give the (alert;-) programmer a hint!
#warning "Dummy crypto only!"
#endif

/// an exact KeyID is in string format
#define LTP_CRYPT_STRING_FORMAT 's'

/// an exact KeyID is in ascii-hex format
#define LTP_CRYPT_AH_FORMAT 'a'

/// a KeyID format that matches any value
#define LTP_CRYPT_WILDCARD  '*'

/// the separator characgter for crypto config files
#define LTP_CRYPT_SEP ",\n"

/// This is the hardcoded key for the NULL ciphersweet from the LTP extensions I-D/RFC
static unsigned char NULL_cs_key[20]={ 
	0xc3,0x7b,0x7e,0x64,
	0x92,0x58,0x43,0x40,
	0xbe,0xd1,0x22,0x07 ,
	0x80,0x89,0x41,0x15,
	0x50,0x68,0xf7,0x38 };

/*!
 * @struct the parameterisation for a crypto thing
 * @param cipherstuite is the ciphersuit for this record
 * @param keyid_fmt specifies whether the KeyID is in string or ascii-hex format
 * @param keyid contains the KeyID
 * @param cpar1 contains the ciphesuite-specific parameter one (usually a symmetric or public key filename)
 * @param cpar2 contains the ciphesuite-specific parameter two (usually a private key filename)
 * @param do_key contains a len/val representation of the signing/MACing key
 * @param check_key contains a len/val representation of the verification (public) key (if needed)
 */
typedef struct ltp_crypt_param_defn {
	int ciphersuite;
	char keyid_fmt;
	char *keyid;
	char *cpar1;
	char *cpar2;
	ltp_octets do_key;
	ltp_octets check_key;
} ltp_crypt_param;

/// the default map only supports the NULL ciphersuite
static ltp_crypt_param defmap[]={LTPE_CIPHERSUITE_NULL,LTP_CRYPT_WILDCARD,NULL,NULL,NULL,{0,NULL},{0,NULL}};

/// the current map
static ltp_crypt_param *map=NULL;

/// whether a map-reading exercise is needed
static bool ltp_crypt_map_reading=false;

/// the number of mappings
static int ltp_crypt_mappings=0;

/// the crypto filename
static char ltp_crypt_fname[MAXPATH]=LTP_CRYPT_FNAME_DEF;

// now for functions!! :-)

/*!
 * @brief read a MACing key from a file
 * @param fname is the file name
 * @param octs is a ptr to an ltp_octets place to put the value
 * @return zero or error
 */
static int read_mac_key_from_file(char *fname, ltp_octets *octs)
{
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Entering %s\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	FILE *fp=fopen(fname,"rb");
	if (!fp) LTPERR(LTP_FILE);
	unsigned char *key=ltpc_alloc(20); // fixed len
	if (!key) LTPERR(LTP_MEMORY);
	int rv=fread(key,1,20,fp);
	if (rv!=20) LTPERR(LTP_FILE);
	fclose(fp);
	octs->val=key;
	octs->len=16;
	return(LTP_OK);
}

/*!
 * @brief read a public key from a file
 * @param fname is the file name
 * @param octs is a ptr to an ltp_octets place to put the value
 * @return zero or error
 */
static int read_public_key_from_file(char *fname, ltp_octets *octs)
{
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Entering %s\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	FILE *fp=fopen(fname,"rb");
	if (!fp) LTPERR(LTP_FILE);
	// read key into suitable structure
	X509 *x509 = PEM_read_X509(fp, NULL, NULL, NULL);
	fclose (fp);
	if (x509 == NULL) {
		snprintf(ltpstr,LTPSTR_LEN,"PEM_read_X509 failed\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(LTP_BAD_PARAM);
	}
	EVP_PKEY *pkey;
	pkey=X509_get_pubkey(x509);
	if (pkey == NULL) {
		snprintf(ltpstr,LTPSTR_LEN,"X509_get_pubkey failed\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(LTP_BAD_PARAM);
	}
	octs->val=(unsigned char*)ltpc_alloc(sizeof(*pkey));
	if (!octs->val) LTPERR(LTP_MEMORY);
	octs->len=sizeof(*pkey);
	memcpy(octs->val,pkey,octs->len);
	return(LTP_OK);
}

/*!
 * @brief read a private key from a file
 * @param fname is the file name
 * @param octs is a ptr to an ltp_octets place to put the value
 * @return zero or error
 */
static int read_private_key_from_file(char *fname, ltp_octets *octs)
{
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Entering %s\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	FILE *fp=fopen(fname,"r");
	if (!fp) LTPERR(LTP_FILE);
	// read key into suitable structure
	EVP_PKEY *pkey;
	pkey = PEM_read_PrivateKey(fp, NULL, NULL, NULL);
	fclose(fp);
	if (!pkey) {
		snprintf(ltpstr,LTPSTR_LEN,"PEM_read_PrivateKey failed\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(LTP_BAD_PARAM);
	}
	octs->val=(unsigned char*)ltpc_alloc(sizeof(*pkey));
	if (!octs->val) LTPERR(LTP_MEMORY);
	octs->len=sizeof(*pkey);
	memcpy(octs->val,pkey,octs->len);
	return(LTP_OK);
}

/*!
 * @brief specify and read the crypto configuration file
 * @param fname is the file name 
 * @return zero or error
 */
int ltp_crypt_setup(char *fname)
{
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Entering %s\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	// just copy the thing and set a flag to re-read @ next interesting time
	if (!fname || fname[0]=='\0') { // ok to call with a NULL
		snprintf(ltp_crypt_fname,MAXPATH,"%s",fname);
	}
	ltp_crypt_map_reading=true;
	return(LTP_OK);
}

/* 
 * @brief map from a key if format character into something useful
 * @param fmt is the putative format instruction
 * @return is a good format or -1
 */
char ltp_crypt_mapkeyidfmt(char fmt)
{
	switch (toupper(fmt)) {
		case 'S':
		case '0':
			return(LTP_CRYPT_STRING_FORMAT);
		case 'A':
		case '1':
			return(LTP_CRYPT_AH_FORMAT);
		case '*':
		case '2':
			return(LTP_CRYPT_WILDCARD);
		default:
			return(-1);
	}
	return (-1);
}

/*!
 * @brief make is easy to specify a ciphersuite
 * @param str is the optarg string from a -A
 * @return is the ciphersuite or -1 for error
 */
int ltp_crypt_mapciphersuite(char *str)
{
	switch (toupper(str[0])) {
		case 'M':
		case '0':
			return LTP_CS_MAC;
		case 'S':
		case '1':
			return LTP_CS_SIG;
		case 'N':
			if (strlen(str)<4) return -1;
			if (toupper(str[1])=='U') return LTP_CS_NULL;
			if (toupper(str[1])=='O') return LTP_CS_NONE;
			return -1;
		case '2':
			// could be either 254 or 255
			if (strlen(str)<3) return -1; // too short a string
			if (str[1]!='5') return -1; // has to be 25x
			if (str[2]=='5') return LTP_CS_NULL;
			if (str[2]=='4') return LTP_CS_NONE;
		default:
			return -1; // fall through!
	}
	return -1; // can't happen
}

/*!
 * @brief initialise defualt mapping if necessary
 */
static void init_mappings(void)
{
	if (map) return;
	map=(ltp_crypt_param*)ltpc_alloc(sizeof(ltp_crypt_param));
	if (!map) VOIDLTPERR(LTP_FILE);
	memcpy(map,defmap,sizeof(ltp_crypt_param));
	ltp_crypt_mappings=1;
	ERR_load_crypto_strings();
	return;
}

/*!
 * @brief nice string for logging
 * @param cs is the ciphersuite
 * @return a nice string
 */
static const char *nice_cs(int cs)
{
	const char *cs_strs[]={
		"MACing",
		"Signature",
		"No ciphersuite at all!",
		"Null ciphersuite",
		"Weird - not a ciphersuite I know about"};
	char *rs;
	switch (cs) {
		case LTP_CS_MAC: return cs_strs[0];
		case LTP_CS_SIG: return cs_strs[1];
		case LTP_CS_NONE: return cs_strs[2];
		case LTP_CS_NULL: return cs_strs[3];
		default: return cs_strs[4];
	}
	return cs_strs[4]; // can't happen
}

/*!
 * @brief update the mapping from file
 * @return zero or error
 *
 * This fnc. uses the same ciphersuite shorthand as ltpd's commandline
 * See mapciphersuite for details.
 */
static int update_mappings(void)
{
	int rv=0;
	FILE *fp=fopen(ltp_crypt_fname,"r");
	if (!fp) LTPERR(LTP_FILE);
	// parse file, line by line
	char line[MAXPATH];
	char *fgrv=NULL;
	char *lp;
	int mapind=0;

	while ((fgrv=fgets(line,MAXPATH,fp))!=NULL) {
		if (strlen(line) > 1023) return(LTP_BAD_VALUE); /// line too long
		lp=line;
		while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
		if (!strlen(lp)) continue; 
		if (*lp=='#') continue;
		if (*lp=='\n') continue;
		int ciphersuite=ltp_crypt_mapciphersuite(lp);
		if (ciphersuite==-1) {
			snprintf(ltpstr,LTPSTR_LEN,"Bad channel value (%c,only 1-4 allowed) in microphone file (%s)\n",
						*lp,ltp_crypt_fname);
			ltp_log(LTP_LOG_MAX,ltpstr);
			fclose(fp); LTPERR(LTP_BAD_VALUE);
		}
		// make space
		mapind++;
		map=(ltp_crypt_param*)ltpc_realloc((unsigned char*)map,mapind*sizeof(ltp_crypt_param));
		if (!map) { fclose(fp);LTPERR(LTP_MEMORY); }
		ltp_crypt_param *mape=&map[mapind-1]; // the new one
		// try parse line now, format should be: int "," float "," float <<who cares>>
		// ciphersuite - I know already
		char *sr;
		char *crv=strtok_r(line,LTP_CRYPT_SEP,&sr);
		if (!crv) { fclose(fp);LTPERR(LTP_FILE); }
		mape->ciphersuite=ciphersuite;
		// keyid_fmt
		crv=strtok_r(NULL,LTP_CRYPT_SEP,&sr);
		mape->keyid_fmt=ltp_crypt_mapkeyidfmt(crv[0]);
		if (mape->keyid_fmt==LTP_CRYPT_WILDCARD) continue;
		// KeyID
		crv=strtok_r(NULL,LTP_CRYPT_SEP,&sr);
		if (!crv) { fclose(fp);LTPERR(LTP_FILE); }
		char *cpstr=ltp_name_trans(crv);
		if (cpstr) {
			mape->keyid=strdup(cpstr);
		} else {
			mape->keyid=strdup(crv);
		} 
		// cpar1 - optional
		crv=strtok_r(NULL,LTP_CRYPT_SEP,&sr);
		if (!crv) continue;
		mape->cpar1=strdup(crv);
		// cpar2 - optional
		crv=strtok_r(NULL,LTP_CRYPT_SEP,&sr);
		if (!crv) continue;
		mape->cpar2=strdup(crv);
		// ignore rest of line
	}
	fclose(fp);

	for (int i=0;i!=mapind;i++) {
		// try read in key files as required
		if (map[i].ciphersuite==LTP_CS_MAC) {
			if (map[i].cpar1==NULL) {
				snprintf(ltpstr,LTPSTR_LEN,"Macing ciphersuite (ind=%d) needs a key file\n",i);
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
			rv=read_mac_key_from_file(map[i].cpar1,&map[i].do_key);
			if (rv) LTPERR(rv);
		}
		if (map[i].ciphersuite==LTP_CS_SIG) {
			if (map[i].cpar1==NULL) {
				snprintf(ltpstr,LTPSTR_LEN,"Signature ciphersuite (ind=%d) needs a public key file\n",i);
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
			rv=read_public_key_from_file(map[i].cpar1,&map[i].check_key);
			if (rv) LTPERR(rv);
			if (map[i].cpar2!=NULL) { // private key
				rv=read_private_key_from_file(map[i].cpar2,&map[i].do_key);
				if (rv) LTPERR(rv);
			}
		}
	}
	ltp_crypt_mappings=mapind;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Updated crypto mappings:\n");
	ltp_log(LTP_LOG_MAX,ltpstr);
	for (int i=0;i!=ltp_crypt_mappings;i++) {
		snprintf(ltpstr,LTPSTR_LEN,"\tmapping[%d]: c=%s,f=%c,k=%s,c1=%s,c2=%s\n",
				i,nice_cs(map[i].ciphersuite),map[i].keyid_fmt,map[i].keyid,map[i].cpar1,map[i].cpar2);	
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
#endif
	ltp_crypt_map_reading=false;
	return(LTP_OK);
}

#define AHVAL(xxXx) ( xxXx >= '0' && xxXx <= '9' ? xxXx-'0' : toupper(xxXx)-'A')

/*!
 * @brief check for a key id match - map version vs. authinfo version
 * @param mfmt is the map format
 * @param mval is the map value
 * @param len is the length of the authinfo value
 * @param val is the authinfo value
 * @return true if a match
 *
 * For this comparison we currenly only support an exact match. Wildcards
 * are only in use for IP address ranges and for entire ciphersuites.
 */
static bool keyid_match(char mfmt, char *mval, int len, unsigned char *val)
{
	switch (mfmt) {
		case LTP_CRYPT_WILDCARD: return (true);
		case LTP_CRYPT_STRING_FORMAT: // just memcmp, easy enough
			if (strlen(mval)==len && !memcmp(mval,val,len)) return true;
			else return false;
		case LTP_CRYPT_AH_FORMAT: // un-hexify and then memcmp
			{
				if (strlen(mval)/2!=len) return(false); // has to be 2x for a match
				for (int i=0;i!=len;i++) {
					if (val[i]>>4 != AHVAL(mval[2*i])) return false;
					if (val[i]&0xf != AHVAL(mval[2*i+1])) return false;
				}
				return(true); // bingo
			}
		default: return false;
	}
	return(false); // can't get here
}

/*!
 * @brief check if an address matches an address range
 * @param rlen is the length of the address range specifier
 * @param rval is the address range specifier
 * @param vlen is the value length
 * @param val is the value
 * @return true if the value is specified by the range specifier
 */
bool match_char_ip(int rlen, char *in_rval, int vlen, void *val)
{
	// now check allowinf for address ranges
	// to recap, now we've got:
	//  	lmlem is length of lmval
	// 		lmval is a string, possible like "10.8.7/24" or "10/8"
	//		len is the length of val
	// 		val is probably an IP address in network byte order
	// first - an exact match is ok
	if (rlen>100 || !in_rval || vlen> 100 || !val) {
		snprintf(ltpstr,LTPSTR_LEN,"Mad input in %s - rlen=%d,rval=%p,vlen=%d,val=%p\n",
				__FUNCTION__,rlen,in_rval,vlen,val);
		ltp_log(LTP_LOG_MAX,ltpstr);
		return false;
	}
	if (vlen==rlen && !memcmp(in_rval,val,rlen)) {  return (true); }
	if (vlen!=sizeof(struct in_addr)) {  return (false); }
	// ok - if the lmval is an IP range then we make up a network byte order version
	struct in_addr va;
	memcpy(&va.s_addr,val,vlen);
	char *cval=strdup(in_rval); // colon finder
	char *sr1=NULL; 	
	char *sr2=NULL; 
	char *sr3=NULL;
	char *crv=strtok_r(cval,":",&sr1);
	char *sval;
	if (crv && strlen(crv)<strlen(in_rval)) { // ":foo"
		char *prv=crv+strlen(crv)+1;
		int port=atoi(prv);
		if (port<0 || port >65535) {
			free(cval); return(false);
		}
		sval=strdup(crv);
	} else {
		sval=strdup(in_rval);
	}
	free(cval);
	char *srv=strtok_r(sval,"/",&sr2);
	struct in_addr	ia;
	bool ans=false;
	if (srv && strlen(srv)<strlen(sval)) {
		char *prv=srv+strlen(srv)+1;
		int prec=atoi(prv);
		if (prec<=0 || prec>=32) {
			ans=false;
		} else {
			unsigned int b0=0,b1=0,b2=0,b3=0;
			ia.s_addr=0x00000000;
			crv=strtok_r(sval,"./",&sr3); if (crv) { b0=atoi(crv); ia.s_addr += (b0&0xff); }
			crv=strtok_r(NULL,"./",&sr3); if (crv) { b1=atoi(crv); ia.s_addr += (b1&0xff)<<8; }
			crv=strtok_r(NULL,"./",&sr3); if (crv) { b2=atoi(crv); ia.s_addr += (b2&0xff)<<16; }
			crv=strtok_r(NULL,"./",&sr3); if (crv) { b3=atoi(crv); ia.s_addr += (b3&0xff)<<24;}
			if (!((ia.s_addr ^ va.s_addr) & (0xffffffff>>prec))) {
				ans=true;
			}
		}
	} else {
		int rv=inet_aton(sval,&ia);
		if (rv) {
			if (ia.s_addr==va.s_addr) ans=true;
			else ans=false;
		}
	}
	free(sval);
	return(ans);
}

/*!
 * @brief check for a peer match
 * @param mfmt is the map format
 * @param mval is the map value
 * @param len is the length of the peer identity
 * @param val is the peer identity
 * @return true if a match
 *
 * For this comparison we support address ranges.
 */
static bool peer_match(char mfmt, char *mval, int len, unsigned char *val)
{
	char *lmval;
	int lmlen;
	switch (mfmt) {
		case LTP_CRYPT_WILDCARD: return (true); // won't happen though!
		case LTP_CRYPT_STRING_FORMAT: // just memcmp, easy enough
				lmlen=strlen(mval); lmval=strdup(mval);break;
		case LTP_CRYPT_AH_FORMAT: // un-hexify and then memcmp
				lmlen=strlen(mval)/2;
				lmval=(char*)ltpc_alloc(strlen(mval)/2+1);
				if (!lmval) {
					snprintf(ltpstr,LTPSTR_LEN,"Memory error at %s:%d\n",__FILE__,__LINE__);
					ltp_log(LTP_LOG_MAX,ltpstr);
					return (false);
				}
				for (int i=0;i!=len;i++) {
					lmval[i]= AHVAL(mval[2*i]) << 4;
					lmval[i] += AHVAL(mval[2*i+1]);
				}
				break;
		default: return false;
	}

	bool ans=match_char_ip(lmlen,lmval,len,val);

	ltpc_free((unsigned char*)lmval);
	return(ans); 
}

/*!
 * @brief check which mapping to use for given inputs
 * @param authinfo is the socket related data
 * @param peer is the LTP peer
 * @return is a ptr to the relevant ltp_crypt_param or NULL
 */
static ltp_crypt_param *check_mapping(ltp_auth_so *authinfo, ltpc_octets *peer)
{
	if (!authinfo || !peer) NULLLTPERR(LTP_BAD_INPUT);
	bool done=false;
	ltp_crypt_param *res=NULL;
	for (int i=0;!done && i!=ltp_crypt_mappings;i++) {
		if (map[i].ciphersuite==authinfo->ciphersuite) {
			if (keyid_match(map[i].keyid_fmt,map[i].keyid,authinfo->keyidlen,authinfo->keyid)) { 
				done=true; res=&map[i]; 
			}
			if (!done && peer_match(map[i].keyid_fmt,map[i].keyid,peer->len,peer->val)) {
				done=true; res=&map[i]; 
			}
		}
	}
	return res;
}

/*!
 * @brief do common crypto stuff
 * @param authinfo is the socket stuff
 * @param peer is the guy
 * @return zero or error
 */
static ltp_crypt_param *crypt_common(ltp_auth_so *authinfo, ltpc_octets *peer)
{
	init_mappings();
	if (ltp_crypt_map_reading) update_mappings();
	ltp_crypt_param *themap=check_mapping(authinfo,peer);
	if (!themap && authinfo->ciphersuite==LTP_CS_NULL) { // get out of jail!
		themap=defmap;
	}
	return(themap);
}


/*!
 * @brief do the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param olen is a ptr to the output buffer length (length used on output)
 * @param obuf is the caller allocated output buffer
 * @return zero or error
 */
int ltp_crypt_do_real(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int *olen, unsigned char **obuf)
{
	int rv;
	ltp_crypt_param *themap=crypt_common(authinfo,peer);
	if (!themap) LTPERR(LTP_BAD_VALUE);

	switch (authinfo->ciphersuite)  {
		case LTP_CS_NONE:
			LTPERR(LTP_BAD_VALUE);
		case LTP_CS_MAC:
			{
			// just to test linking against libcrypto
			if (!themap->do_key.val) LTPERR(LTP_BAD_VALUE);
			unsigned char *digest_str=HMAC(EVP_sha1(), themap->do_key.val, themap->do_key.len, buf, blen, *obuf, (unsigned int*) olen);
			if (!digest_str || *olen!=20) {
				snprintf(ltpstr,LTPSTR_LEN,"Error MACing!\n");
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
			}
			break;
		case LTP_CS_SIG:
			{
			EVP_MD_CTX md_ctx;
			unsigned int sig_len;
			unsigned char sig_buf[1024];
  			EVP_SignInit   (&md_ctx, EVP_sha256()); // seems to require openssl0.9.8
  			EVP_SignUpdate (&md_ctx, buf, blen);
			EVP_PKEY *pkey=(EVP_PKEY*)themap->do_key.val;
			if (!pkey) {
				snprintf(ltpstr,LTPSTR_LEN,"No private key!\n");
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
  			sig_len = sizeof(sig_buf);
  			rv=EVP_SignFinal (&md_ctx, sig_buf, &sig_len, pkey);
  			if (rv != 1) {
				snprintf(ltpstr,LTPSTR_LEN,"Error from EVP_SignFinal (%d)\n",rv);
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
			if (*olen<sig_len)  LTPERR(LTP_MEMORY);
			*olen=sig_len;
			memcpy(*obuf,sig_buf,sig_len);
			}
			break;
		case LTP_CS_NULL:
			{
			// just to test linking against libcrypto
			unsigned char *digest_str=HMAC(EVP_sha1(), NULL_cs_key, 20, buf, blen, *obuf, (unsigned int*) olen);
			if (!digest_str || *olen!=20) {
				snprintf(ltpstr,LTPSTR_LEN,"Error MACing!\n");
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
			}
			break;
		default:
			LTPERR(LTP_BAD_VALUE);
	}

	return(LTP_OK);
}

/*!
 * @brief check the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param alen is a ptr to the authenticator buffer length (length used on output)
 * @param abuf is the authenticator buffer
 * @param result is the check result (true = ok, false = bad)
 * @return zero or error
 */
int ltp_crypt_check_real(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int alen, unsigned char *abuf, bool *result)
{
	int rv;
	if (!result) LTPERR(LTP_BAD_PARAM);
	*result=false; // pessimist
	ltp_crypt_param *themap=crypt_common(authinfo,peer);
	if (!themap) LTPERR(LTP_BAD_VALUE);

	// now check sig

	switch (authinfo->ciphersuite)  {
		case LTP_CS_MAC:
			{
			// just to test linking against libcrypto
			if (!themap->do_key.val) LTPERR(LTP_BAD_VALUE);
			unsigned char rval[200]; // 200 is OTT but no harm
			unsigned int rlen=200;
			unsigned char *digest_str=HMAC(EVP_sha1(), themap->do_key.val, themap->do_key.len, buf, blen, rval,&rlen);
			if (!digest_str || rlen!=20) {
				snprintf(ltpstr,LTPSTR_LEN,"Error MACing!\n");
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
			if (rlen!=alen) {
				*result=false; // bummer
			} else if (memcmp(rval,abuf,rlen)) {
				*result=false; // bummer
			} else {
				*result=true; // horay!
			}
			}
			break;
		case LTP_CS_SIG:
			{
			EVP_MD_CTX md_ctx;
  			EVP_VerifyInit   (&md_ctx, EVP_sha256()); // seems to require openssl0.9.8
  			EVP_VerifyUpdate (&md_ctx, buf, blen);
			EVP_PKEY *pkey=(EVP_PKEY*)themap->do_key.val;
			if (!pkey) {
				snprintf(ltpstr,LTPSTR_LEN,"No public key!\n");
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
  			rv=EVP_VerifyFinal (&md_ctx, abuf, alen, pkey);
  			if (rv != 1) {
				snprintf(ltpstr,LTPSTR_LEN,"Error from EVP_VerifyFinal (%d)\n",rv);
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Signature check succeeded\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			*result=true; // its a good sig!!!
			}
			break;
		case LTP_CS_NULL:
			{
			unsigned char rval[200]; // 200 is OTT but no harm
			unsigned int rlen=200;
			unsigned char *digest_str=HMAC(EVP_sha1(), NULL_cs_key, 20, buf, blen, rval,&rlen);
			if (!digest_str || rlen!=20) {
				snprintf(ltpstr,LTPSTR_LEN,"Error MACing!\n");
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
			}
			if (rlen!=alen) {
				*result=false; // bummer
			} else if (memcmp(rval,abuf,rlen)) {
				*result=false; // bummer
			} else {
				*result=true; // horay!
			}
			}
			break;
		case LTP_CS_NONE:
			LTPERR(LTP_BAD_VALUE);
		default:
			LTPERR(LTP_BAD_VALUE);
	}

	return(LTP_OK);
}

/*!
 * @brief do the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param olen is a ptr to the output buffer length (length used on output)
 * @param obuf is the caller allocated output buffer
 * @return zero or error
 */
int ltp_crypt_do_dummy(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int *olen, unsigned char **obuf)
{
	// just to test linking against libcrypto
	unsigned char key[16];
	int key_len=16;
	unsigned char data[64];
	int data_len=64;
	unsigned char *foo=HMAC(EVP_md5(), key, key_len, data, data_len, NULL,NULL);
	*olen=DUMMYLEN;
	*obuf=dummybuf;
	return(LTP_OK);
}

/*!
 * @brief check the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param alen is a ptr to the authenticator buffer length (length used on output)
 * @param abuf is the authenticator buffer
 * @param result is the check result (true = ok, false = bad)
 * @return zero or error
 */
int ltp_crypt_check_dummy(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int alen, unsigned char *abuf, bool *result)
{
	if (alen==DUMMYLEN && !memcmp(abuf,dummybuf,alen)) *result=true;
	else *result=false;
	return(LTP_OK);
}

/*!
 * @brief do the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param olen is a ptr to the output buffer length (length used on output)
 * @param obuf is the caller allocated output buffer
 * @return zero or error
 */
int ltp_crypt_do(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int *olen, unsigned char **obuf)
{
#ifdef DUMMYCRYPTO
	return ltp_crypt_do_dummy(authinfo, peer, blen, buf, olen, obuf);
#else
	return ltp_crypt_do_real(authinfo, peer, blen, buf, olen, obuf);
#endif
}

/*!
 * @brief check the crypto calculations for applying an authenticator
 * @param authinfo specifies the ciphersuite and key id
 * @param blen is the length of the input buffer (the encoded segment)
 * @param buf is the input buffer
 * @param alen is a ptr to the authenticator buffer length (length used on output)
 * @param abuf is the authenticator buffer
 * @param result is the check result (true = ok, false = bad)
 * @return zero or error
 */
int ltp_crypt_check(ltp_auth_so *authinfo, ltpc_octets *peer, int blen, unsigned char *buf, int alen, unsigned char *abuf, bool *result)
{
#ifdef DUMMYCRYPTO
	return ltp_crypt_check_dummy(authinfo, peer, blen, buf, alen, abuf, result);
#else
	return ltp_crypt_check_real(authinfo, peer, blen, buf, alen, abuf, result);
#endif
}
