/*!
 * @file ltp_cues.cc
 * @brief This is implementation of the API for handling cues
 * @version $Revision: 1.14 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 * 
 * The plan here is to have a set of SPIs that we can use, e.g. the
 * perfume perfume_ltp_sched stuff, or perhaps something else. The
 * minimal SPI implementation (still TBD) could be twofold:
 * - first an always-on one where everyone is available so long as
 *   the local IP interface is "up"
 * - sendt_cues use ltp-t schedules and all that.
 * 
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <sys/socket.h>
#include <sys/types.h>
#include "ltp_base.h"
#include "ltp_int.h"
#include "ltp.h"
#ifdef THREADS
#include <pthread.h>
#endif
#ifdef SENDT_CUES
#include <sendt_cues.h>
#endif

const char *ADDRANY="PETERFARRELL"; ///< this is an address which is a wildcard (used in whos_listening etc)

static ltpaddr			localme; ///< locally set version of my IP address - set via ltp_cue_whoiam
static bool				localmeset=false; ///< indicates caller set localme
#ifdef THREADS
static pthread_mutex_t	meguard=PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP; 	///< to stop nastiness
#endif

/*!
 * @fn int ltp_cue_whos_listening(ltp_time_t now, int *nlisteners, ltpaddr *listeners)
 * @brief tells a node which other nodes are listening now (so we can Tx to them if there're packets)
 * @param now is the time I need to know about
 * @param nlisteners on input is the size of the listeners array, and on output is the number of real entries in the array
 * @param listeners is an array (caller allocated) of the LTP addresses of the currently listening nodes
 * @return zero or error
 *
 */
int ltp_cue_whos_listening(ltp_time_t now, int *nlisteners, ltpaddr *listeners)
{
	if (!nlisteners || *nlisteners<2) LTPERR(LTP_BAD_PARAM);
#ifdef SENDT_CUES
	int	alist=*nlisteners;
	char *list[alist];
	char **listp=list;
	int rv;
	rv=sendt_whos_listening(now,&alist,listp);
	if (rv && rv!=LTP_FILE) {
		LTPERR(rv);
	} else if (rv==LTP_FILE) {
		// no schedule file, so switch to "always on" mode
		listeners[0]=ltpany;
		*nlisteners=1;
	} else {
		// copy over from list to ltpaddr structures
		for (int i=0;i!=alist;i++) {
			if (list[i]==ADDRANY) { 
				listeners[i]=ltpany; // swap wildcard formats
			} else {
				rv=str2ltpaddr(list[i],&listeners[i]);
				if (rv) LTPERR(rv);
			}
		} 
		*nlisteners=alist;
		return(LTP_OK);
	}
#endif
	// anyone!! Just for now TODO: replace with ping & other schemes
	listeners[0]=ltpany;
	*nlisteners=1;
	return(LTP_OK);
}

/*!
 * @fn int ltp_cue_whoami(ltpaddr *me)
 * @brief returns my LTP address
 * @param me is the caller allocated ltpaddr structure in which to put the address
 * @return zero or error
 *
 * Note that my address can change as a result of rescheduling - so I better check it occasionally.
 * I'd say that per-reboot should be ok.
 *
 */
int ltp_cue_whoami(ltpaddr *me)
{
	if (!me) LTPERR(LTP_BAD_PARAM);
#ifdef SENDT_CUES
	char *cme=sendt_whoami();
	if (cme) {
		int rv=str2ltpaddr(cme,me);
		if (!rv) return(LTP_OK);
	}
#endif
	if (localmeset) {
#ifdef THREADS
		pthread_mutex_lock(&meguard);
#endif
	 	*me=localme; // struct copy
#ifdef THREADS
		pthread_mutex_unlock(&meguard);
#endif
	} else {
		// hard problem - just use 127.0.0.1
		ltpaddr foo={AF_LTP,UDP_DEF_PORT,0x0100007f,"\0\0\0\0\0\0\0"};  
		*me=foo;
	}
	return(LTP_OK);
}

/*!
 * @brief tells a node which other nodes ought to be talking now (so we can decrement timers)
 * @param now is the time I need to know about
 * @param ntalkers on input is the size of the listeners array, and on output is the number of real entries in the array
 * @param talkers is an array (caller allocated) of the LTP addresses of the currently chatting nodes
 * @return zero or error
 */
int ltp_cue_whos_talking(ltp_time_t now, int *ntalkers, ltpaddr *talkers)
{
	if (!ntalkers || *ntalkers<2) LTPERR(LTP_BAD_PARAM);
#ifdef SENDT_CUES
	int	alist=*ntalkers;
	char *list[alist];
	char **listp=list;
	int rv;
	rv=sendt_whos_talking(now,&alist,listp);
	if (rv && rv!=LTP_FILE) {
		LTPERR(rv);
	} else if (rv==LTP_FILE) {
		// no schedule file, so switch to "always on" mode
		talkers[0]=ltpany;
		*ntalkers=1;
	} else {
		// copy over from list to ltpaddr structures
		for (int i=0;i!=alist;i++) {
			if (list[i]==ADDRANY) {
				talkers[i]=ltpany; // swap wildcard formats
			} else {
				rv=str2ltpaddr(list[i],&talkers[i]);
				if (rv) LTPERR(rv);
			}
		}
		*ntalkers=alist;
		return(LTP_OK);
	}
#endif
	// anyone!! Just for now TODO: replace with ping & other schemes
	talkers[0]=ltpany;
	*ntalkers=1;
	return(LTP_OK);
}


/*!
 * @brief set the response from whoami to a pre-cooked one
 * @param ltpaddr *me
 * @return zero or error
 */
int ltp_cue_set_whoiam(ltpaddr *me)
{
	// this is needed since sometimes it can be hard for a library to know which IP
	// address to send out in the peer field of an LTP segment (that's used by the
	// peer to contact me back later)
	// so if a caller sets this, then it'll get the same thing back
	// this is actually normall called from an ltp setsockopt with LTP_SO_ME
#ifdef THREADS
	pthread_mutex_lock(&meguard);
#endif
	localme=*me; // struct copy 
	localmeset=true;

#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"ltp_cue_set_whoiam set to %s\n",ltpaddr2str(me));
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
#ifdef THREADS
	pthread_mutex_unlock(&meguard);
#endif
	return (LTP_OK);
}

/*!
 * @brief figure out an LTT for a node to which I'm about to transmit
 * @param peer is the node's address
 * @param now is the current time
 * @param ltt is (a ptr to) the LTT at that time
 * @return zero or error
 */
int ltp_ltt(ltpaddr *peer, ltp_time_t now, ltp_time_t *ltt)
{
	// this should be done via the SPI but no time now
	// 2010-09-02 add support for a fixed LTT for testing.
	// this will override all other LTT settings, e.g. ephemeris stuff
#define FIXEDLTTFILE "fixed.ltt"
	bool overridden=false;
	ltp_time_t localltt;
	FILE *fp=fopen(FIXEDLTTFILE,"r");
	if (fp) {
		ltp_time_t	theltt;
		char line[MAXPATH];
		char *fgrv;
		char *lp;
		bool done=false;
		bool started=false;
		int lineno=0;
		int prelineno=0;
		int postlineno=0;
		while ((fgrv=fgets(line,MAXPATH,fp))!=NULL && !overridden ) {
			lineno++;
			if (strlen(line) > (MAXPATH-1)) { fclose(fp); LTPERR(LTP_BAD_VALUE); } /// line too long
			lp=line;
			while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
			while (ltp_iswhitespace(*lp)) lp++;
			if (!strlen(lp)) continue; 
			if (*lp=='#') continue;
			if (*lp=='\n') continue;
			char *sr;
			char *crv=strtok_r(line," \n",&sr);
			if (!crv) continue;
			localltt=atof(crv);
			overridden=true;
		}
		fclose(fp);
		if (overridden) {
			*ltt=localltt;
			return(LTP_OK);
		}
		
	}
	if (!overridden) {
		return(sendt_ltt(peer,now,ltt));
	} 
}
