/*!
 * @file ltp_ext.cc
 * @brief Codec for specific (actually, all-known) LTP extensions
 * @version $Revision: 1.35 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This API "knows" abstract definitions for the various
 * extension types in LTP/LTP-T. For each we will have an encode
 * and decode function, e.g. ltpc_encode_cookie() and 
 * ltpc_decode_cookie(). We also have a table of extensions
 * with function pointers etc.
 *
 * The memory model is that the caller allocates all
 * storage - all required inputs are passed in 
 * separately.
 *
 * There are no new types defined for this api, we
 * just deal with raw 'C' types as far as posible.
 * See ltp_ext.h for definitions.
 *
 * Both of the above are aimed at avoiding me making
 * unwarranted assumptions about how the codec will
 * evolve over time. Not sure if it'll help that much
 * but worth a try.
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


#include <ctype.h> // for isprint()

#include "ltp_base.h"
#include "ltp_codec.h"
#include "ltp_ext.h"

/*!
 * @section g1 redefinitions of stuff...
 * potentially redefine this here since we might want to separate out this code
 * from the rest of the build
 */
#ifndef LTPERR
#define LTPERR(xxFoo) { return(xxFoo);}
#endif
#ifndef LTP_OK
#define LTP_OK 0
#endif
#ifndef LTP_BAD_PARAM
#define LTP_BAD_PARAM 1
#endif
#ifndef LTP_MEMORY
#define LTP_MEMORY 2
#endif
#ifndef LTP_BAD_VALUE
#define LTP_BAD_VALUE 7
#endif

/*!
 * @section g2 extensions codec
 */

/// Since callers allocated memory we have to check sizing often, this is just a bit of sugar for that
#define BUFLIMIT(xX,yY) if (xX>=yY) LTPERR(LTP_MEMORY);


/*!
 * @struct ltpe_fncs
 * @param type is the extension type tag
 * @param zerolengthallowed is a boolean specifying whether or not the type can have a zero length value
 * @param e_enc is the extension encoding fnc.
 * @param e_dec is the extension decoding fnc.
 * @param e_print is the extension printing fnc. (none implemented as yet)
 * @param e_free is the extension free fnc (if required)
 */
typedef struct ltpe_fncs_defn {
	int		type;
	bool	zerolengthallowed;
	int		(*e_enc)(ltpc_ext *ext);
	int		(*e_dec)(ltpc_ext *ext);
	int		(*e_print)(ltpc_ext *ext, int strlen, char *strbuf);
	void	(*e_free)(ltpc_ext *ext);
} ltpe_fncs;

/// the ltpe_fncs variable - a jump table for extension handlers
ltpe_fncs e_fncs[]= {
	{ LTPE_AUTH, true, ltpe_enc_auth, ltpe_dec_auth, ltpe_print_auth, ltpe_free_auth_ext }, // zero length allowed for plaintext construction
	{ LTPE_COOKIE, false, ltpe_enc_cookie, ltpe_dec_cookie, ltpe_print_cookie, ltpe_free_octets_ext },
	{ LTPTE_SOURCE, false, ltpe_enc_source, ltpe_dec_source, ltpe_print_source, ltpe_free_addr_ext },
	{ LTPTE_DEST, false, ltpe_enc_dest, ltpe_dec_dest, ltpe_print_dest, ltpe_free_addr_ext },
	{ LTPTE_EBS, false, ltpe_enc_ebs, ltpe_dec_ebs, ltpe_print_ebs, NULL },
	{ LTPTE_PORT, false, ltpe_enc_port, ltpe_dec_port, ltpe_print_port, ltpe_free_octets_ext },
	{ LTPTE_HOP_COUNT, false, ltpe_enc_hop_count, ltpe_dec_hop_count, ltpe_print_hop_count, NULL },
	{ LTPTE_VIA, true, ltpe_enc_via, ltpe_dec_via, ltpe_print_via, ltpe_free_octets_ext },
	{ LTPTE_AUTH, false, ltpe_enc_e2eauth, ltpe_dec_e2eauth, ltpe_print_e2eauth, ltpe_free_e2eauth },
	// { LTPTE_CONFID, false, ltpe_enc_t_confid, ltpe_dec_t_confid, ltpe_print_t_confid, ltpe_free_confid_ext },
	{ LTPTE_ECN, false, ltpe_enc_ecn, ltpe_dec_ecn, ltpe_print_ecn, ltpe_free_ecn_ext },
};

#define HANDLED_EXTS (sizeof(e_fncs)/sizeof(ltpe_fncs))

/*!
 * @brief populate the extension codec tables and other init
 * @return zero or error
 */
int ltpe_init(void)
{
	return LTP_OK;
}

/*!
 * @brief general encoder (calls type specific fncs.)
 * @param ext is an extension structure
 * @return zero or error
 */
int ltpe_encode(ltpc_ext *ext)
{
	int rv;
	// check a few things
	if (!ext) LTPERR(LTP_BAD_PARAM);

	// do we handle this extension?
	int extind=-1;
	for (int i=0;i!=HANDLED_EXTS && extind==-1 ;i++) {
		if (ext->type==e_fncs[i].type) {
			extind=i;
		}
	}	
	if (extind==-1) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: error unhandled extension type (%d)\n",__FUNCTION__,ext->type);
		ltp_log(LTP_LOG_ERR,ltpstr);
		LTPERR(LTP_UNKNOWN_EXT);
	}

	// already encoded? if so, we're done
	if (ext->enclen && ext->encval) return(LTP_OK);

	if (e_fncs[extind].zerolengthallowed && ext->len==0) { // ok
		return(LTP_OK);
	} else if (!e_fncs[extind].zerolengthallowed && ext->len==0) { // error
		snprintf(ltpstr,LTPSTR_LEN,"%s: zero length not allowed (ind=%d) for extsion type (%d)\n",
			__FUNCTION__,extind,ext->type);
		ltp_log(LTP_LOG_ERR,ltpstr);
		LTPERR(LTP_BAD_VALUE);
	}

	if (!ext->val || !ext->len) { 
		LTPERR(LTP_BAD_PARAM);
	}
	// silly input - catch it anyway
	if ((ext->enclen && !ext->encval) || (!ext->enclen && ext->encval)) {
		LTPERR(LTP_BAD_PARAM);
	}

	if (!e_fncs[extind].e_enc) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: weird error no encoding function (ind=%d) for extsion type (%d)\n",
			__FUNCTION__,extind,ext->type);
		ltp_log(LTP_LOG_ERR,ltpstr);
		LTPERR(LTP_BAD_STATE);
	}
	rv=e_fncs[extind].e_enc(ext);
	if (rv) LTPERR(rv);
	return(LTP_OK);
}

/*!
 * @brief general decoder (calls type specific fncs.)
 * @param ext is an extension structure
 * @return zero or error
 */
int ltpe_decode(ltpc_ext *ext)
{
	int rv;
	// check a few things
	if (!ext) LTPERR(LTP_BAD_PARAM);

	// do we handle this extension?
	int extind=-1;
	for (int i=0;i!=HANDLED_EXTS && extind==-1 ;i++) {
		if (ext->type==e_fncs[i].type) {
			extind=i;
		}
	}	
	if (extind==-1) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: error unhandled extension type (%d)\n",__FUNCTION__,ext->type);
		ltp_log(LTP_LOG_ERR,ltpstr);
		LTPERR(LTP_UNKNOWN_EXT);
	}

	if (e_fncs[extind].zerolengthallowed && ext->enclen==0) { // ok
		return(LTP_OK);
	} else if (!e_fncs[extind].zerolengthallowed && ext->enclen==0) { // error
		LTPERR(LTP_BAD_VALUE);
	}

	if (!ext->encval || !ext->enclen) LTPERR(LTP_BAD_PARAM);
	// already decoded? if so, we're done
	if (ext->len && ext->val) return(LTP_OK);
	// silly input - catch it anyway
	if ((ext->len && !ext->val) || (!ext->len && ext->val)) LTPERR(LTP_BAD_PARAM);

	if (!e_fncs[extind].e_dec) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: weird error no decoding function (ind=%d) for extsion type (%d)\n",__FUNCTION__,extind,ext->type);
		ltp_log(LTP_LOG_ERR,ltpstr);
		LTPERR(LTP_BAD_STATE);
	}
	rv=e_fncs[extind].e_dec(ext);
	if (rv) LTPERR(rv);
	return(LTP_OK);
}

/*!
 * @brief call a print fnc if it exists
 * @param ext is the guy to print
 * @param slen is the length of str 
 * @param str is the caller-allocated print buffer
 * @return zero or error
 */
int ltpe_print(ltpc_ext *ext, int slen, char *str)
{
	if (!ext || !str || !slen) return(LTP_BAD_PARAM);

	// do we handle this extension?
	int extind=-1;
	for (int i=0;i!=HANDLED_EXTS && extind==-1 ;i++) {
		if (ext->type==e_fncs[i].type) {
			extind=i;
		}
	}	
	if (extind==-1) {
		snprintf(ltpstr,LTPSTR_LEN,"%s:%d error unhandled extension type (%d)\n",__FUNCTION__,__LINE__,ext->type);
		ltp_log(LTP_LOG_ERR,ltpstr);
		snprintf(str,slen,"Unhandled extension type %d\n",ext->type);
		return(LTP_OK);
	}
	if (!e_fncs[extind].e_print) {
		snprintf(ltpstr,LTPSTR_LEN,"%s:%d error no print function for a (%d)\n",__FUNCTION__,__LINE__,ext->type);
		ltp_log(LTP_LOG_ERR,ltpstr);
		snprintf(str,slen,"No print function for ext type %d\n",ext->type);
		return(LTP_OK);
	}
	return(e_fncs[extind].e_print(ext,slen,str));
}


/*! 
 * @brief decode a auth extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_auth(ltpc_ext *ext)
{
	int len=ext->enclen;
	unsigned char *val=ext->encval;
	if (ext->header) {
		ltpe_auth_header	*hdr=(ltpe_auth_header*)ltpc_alloc(sizeof(ltpe_auth_header));
		if (!hdr) LTPERR(LTP_MEMORY);
		hdr->ciphersuite=val[0];
		if (len>=2) {
			hdr->keyid.len=len-1;
			hdr->keyid.val=(unsigned char*)ltpc_alloc(len-1);
			if (!hdr->keyid.val) LTPERR(LTP_MEMORY);
			memcpy(hdr->keyid.val,val+1,len-1);
		} else {
			hdr->keyid.len=0;
			hdr->keyid.val=NULL;
		}
		ext->val=(void*)hdr;
		ext->len=sizeof(ltpe_auth_header);
	} else {
		ltpe_auth_trailer *tlr=(ltpe_auth_trailer*)ltpc_alloc(sizeof(ltpe_auth_trailer));
		if (!tlr) LTPERR(LTP_MEMORY);
		tlr->authval.len=len;
		tlr->authval.val=(unsigned char*)ltpc_alloc(len);
		if (!tlr->authval.val) LTPERR(LTP_MEMORY);
		memcpy(tlr->authval.val,val,len);
		ext->val=(void*)tlr;
		ext->len=sizeof(ltpe_auth_trailer);
	}
	return(LTP_OK);
}

/*!
 * @brief encode a auth extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_auth(ltpc_ext *ext)
{
	if (ext->header) {
		if (ext->len!=sizeof(ltpe_auth_header)) LTPERR(LTP_BAD_VALUE);
		ltpe_auth_header *hdr=(ltpe_auth_header*)ext->val;
		int len=1+hdr->keyid.len;
		ext->enclen=len;
		ext->encval=ltpc_alloc(len);
		if (!ext->encval) LTPERR(LTP_MEMORY);
		ext->encval[0]=hdr->ciphersuite;
		if (hdr->keyid.len) memcpy(ext->encval+1,hdr->keyid.val,len-1);
	} else {
		if (ext->len!=sizeof(ltpe_auth_trailer)) LTPERR(LTP_BAD_VALUE);
		ltpe_auth_trailer *tlr=(ltpe_auth_trailer*)ext->val;
		if (!tlr->authval.len) LTPERR(LTP_BAD_VALUE);
		ext->enclen=tlr->authval.len;
		ext->encval=ltpc_alloc(tlr->authval.len);
		if (!ext->encval) LTPERR(LTP_MEMORY);
		memcpy(ext->encval,tlr->authval.val,ext->enclen);
	}
	return(LTP_OK);
}

/*!
 * @brief print a auth extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_auth(ltpc_ext *ext, int strlen, char *strbuf)
{
	int rv;
	char vals[MAXPATH];
	vals[0]='\0';
	if (ext->header) {
		if (!ext->val) {
			rv=ltpe_dec_auth(ext);
			if (rv) {
				snprintf(strbuf,strlen,"LTP Auth header - error can't decode header!");
				return(LTP_OK);
			} 
		}
		ltpe_auth_header	*aval=(ltpe_auth_header*)ext->val;
		if (!aval) {
			snprintf(strbuf,strlen,"LTP Auth header - NULL decoded header!");
			return(LTP_OK);
		}
		char kval[MAXPATH];
		kval[0]='\0';
		int rv=bin2AH(aval->keyid.val,aval->keyid.len,kval,MAXPATH);
		if (rv) LTPERR(rv);
		char cval[MAXPATH];
		switch (aval->ciphersuite) {
			case LTP_CS_MAC:snprintf(cval,MAXPATH,"OriginAuth (MAC)");break;
			case LTP_CS_SIG: snprintf(cval,MAXPATH,"Signature");break;
			case LTP_CS_NULL:snprintf(cval,MAXPATH,"NULL (hardcoded-MAC)");break;
			default:snprintf(cval,MAXPATH,"Unknown ciphersuite: %d",aval->ciphersuite);
		}
		snprintf(strbuf,strlen,"LTP Auth header, ciphersuite=%s, keyId=%s",cval,kval);
		return(LTP_OK);
	} else {
		bool truncated;
		int plen=4;
		if (ext->enclen>plen) {
			truncated=true;
		} else {
			truncated=false;
		}
		rv=bin2AH(ext->encval,plen,vals,MAXPATH);
		if (rv) LTPERR(rv);
		snprintf(strbuf,strlen,"LTP Auth val:  (%d bits) %s%s",ext->enclen*8,vals,(truncated?"...":""));
		return(LTP_OK);
	}
	return(LTP_OK);
}

/*! 
 * @brief decode a cookie extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_cookie(ltpc_ext *ext)
{
	if (!ext->enclen) LTPERR(LTP_BAD_VALUE);
	ltpe_cookie *ck=(ltpe_cookie*)ltpc_alloc(sizeof(ltpe_cookie));
	if (!ck) LTPERR(LTP_MEMORY);
	ck->cval.val=ltpc_alloc(ext->enclen);
	ck->cval.len=ext->enclen;
	memcpy(ck->cval.val,ext->encval,ext->enclen);
	return(LTP_OK);
}

/*!
 * @brief encode a cookie extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_cookie(ltpc_ext *ext)
{
	ltpe_cookie *ck=(ltpe_cookie*)ext->val;
	ext->enclen=ck->cval.len;
	ext->encval=ltpc_alloc(ck->cval.len);
	if (!ext->encval) LTPERR(LTP_MEMORY);
	memcpy(ext->encval,ck->cval.val,ext->enclen);
	return(LTP_OK);
}

/*!
 * @brief print a cookie extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_cookie(ltpc_ext *ext, int strlen, char *strbuf)
{
	char vals[MAXPATH];
	vals[0]='\0';
	int rv=bin2AH(ext->encval,ext->enclen,vals,MAXPATH);
	if (rv) LTPERR(rv);
	snprintf(strbuf,strlen,"Cookie (%d bits): %s",ext->enclen*8,vals);
	return(LTP_OK);
}

/*! 
 * @brief decode a source extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_source(ltpc_ext *ext)
{
	return ltpe_dec_dest(ext);
}

/*!
 * @brief encode a source extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_source(ltpc_ext *ext)
{
	return ltpe_enc_dest(ext); // same struct, same encoding
}

/*!
 * @brief print a source extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_source(ltpc_ext *ext, int strlen, char *strbuf)
{
	if (!ext) LTPERR(LTP_BAD_PARAM);
	ltpe_src *src=(ltpe_src*)ext->val;
	if (!src) LTPERR(LTP_BAD_VALUE);
	snprintf(strbuf,strlen,"source: %s",ltpaddr2str(&src->src));
	return(LTP_OK); 
}

/*! 
 * @brief decode a dest extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_dest(ltpc_ext *ext)
{
	int rv;
	if (!ext->enclen) LTPERR(LTP_BAD_VALUE);
	ltpe_dest *dest=(ltpe_dest*)ltpc_alloc(sizeof(ltpe_dest));
	if (!dest) LTPERR(LTP_MEMORY);
	// we treat len==4,6 as being an IP address, otherwise as a octet string
	if (ext->enclen==6 || ext->enclen==4) {
		memset(&dest->dest.sock,0,sizeof(sockaddr_in));
		ltpc_octets o_dest;
		o_dest.len=ext->enclen;
		o_dest.val=ext->encval;
		rv=octets2ltpaddr(&o_dest,&dest->dest);
		if (rv) LTPERR(rv);
		ext->val=(void*)dest;
		ext->len=sizeof(*dest);
	} else {
		dest->dest.octets.len=ext->enclen; // duplicates stuff here but not elsewhere
		dest->dest.octets.val=ltpc_alloc(dest->dest.octets.len);
		if (!dest->dest.octets.val) LTPERR(LTP_MEMORY);
		memcpy(dest->dest.octets.val,ext->encval,dest->dest.octets.len);
		dest->dest.octets.magic=LTP_ADDR_OCT_MAGIC; // magic
		ext->val=(void*)dest;
		ext->len=sizeof(*dest);
	}
	return(LTP_OK);
}

/*!
 * @brief encode a dest extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_dest(ltpc_ext *ext)
{
	int rv;
	ltpe_dest *dest=(ltpe_dest*)ext->val;
	if (dest->dest.octets.magic==LTP_ADDR_OCT_MAGIC) { // its octets
		ext->enclen=dest->dest.octets.len;
		ext->encval=ltpc_alloc(ext->enclen);
		if (!ext->encval) LTPERR(LTP_MEMORY);
		memcpy(ext->encval,dest->dest.octets.val,ext->enclen);
	} else { // its IP
		ltpc_octets o_dest;
		rv=ltpaddr2octets(&dest->dest,&o_dest);
		if (rv) LTPERR(rv);
		ext->enclen=o_dest.len;
		ext->encval=o_dest.val;
	}
	return(LTP_OK);
}

/*!
 * @brief print a dest extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_dest(ltpc_ext *ext, int strlen, char *strbuf)
{
	if (!ext) LTPERR(LTP_BAD_PARAM);
	ltpe_src *src=(ltpe_src*)ext->val;
	if (!src) LTPERR(LTP_BAD_VALUE);
	snprintf(strbuf,strlen,"dest: %s",ltpaddr2str(&src->src));
	return(LTP_OK); 
}

/*! 
 * @brief decode a ebs extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_ebs(ltpc_ext *ext)
{
	if (!ext->enclen || ext->enclen > 4) return(LTP_BAD_VALUE);
	ltpe_ebs	*ebs=(ltpe_ebs*)ltpc_alloc(sizeof(ltpe_ebs));
	if (!ebs) LTPERR(LTP_MEMORY);
	unsigned int foo;
	switch (ext->enclen) {
		case 1: foo = ext->encval[0]; break;
		case 2: foo = (ext->encval[0] << 8) + ext->encval[1]; break;
		case 3: foo = (ext->encval[0] << 16) + (ext->encval[1] << 8) + ext->encval[2]; break;
		case 4: foo = (ext->encval[0] << 24) + (ext->encval[1] << 16) + (ext->encval[2] << 8) + ext->encval[3]; break;
		default:
			LTPERR(LTP_BAD_VALUE);
	}
	ebs->ebs=foo;
	ext->val=(void*)ebs;
	ext->len=sizeof(ltpe_ebs);
	return(LTP_OK);
}

/*!
 * @brief encode a ebs extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_ebs(ltpc_ext *ext)
{
	int flen=4;
	ext->encval=ltpc_alloc(flen);
	if (!ext->encval) LTPERR(LTP_MEMORY);
	ltpe_ebs *ebs=(ltpe_ebs*)ext->val;
	unsigned int foo=ebs->ebs;
	int rv=ltpc_small_htonl(foo,&flen,ext->encval);
	if (rv) LTPERR(rv);
	ext->enclen=flen;
	return(LTP_OK);
}

/*!
 * @brief print a ebs extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_ebs(ltpc_ext *ext, int strlen, char *strbuf)
{
	if (!ext) LTPERR(LTP_BAD_PARAM);
	ltpe_ebs *ebs=(ltpe_ebs*)ext->val;
	if (!ebs) LTPERR(LTP_BAD_VALUE);
	snprintf(strbuf,strlen,"est. block size: %d",ebs->ebs);
	return(LTP_OK); 
}

/*! 
 * @brief decode a port extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_port(ltpc_ext *ext)
{
	if (!ext->enclen) LTPERR(LTP_BAD_VALUE);
	ltpe_port *port=(ltpe_port*)ltpc_alloc(sizeof(ltpe_port));
	if (!port) LTPERR(LTP_MEMORY);
	port->port.val=ltpc_alloc(ext->enclen);
	port->port.len=ext->enclen;
	memcpy(port->port.val,ext->encval,ext->enclen);
	ext->val=(void*)port;
	return(LTP_OK);
}

/*!
 * @brief encode a port extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_port(ltpc_ext *ext)
{
	ltpe_port *port=(ltpe_port*)ext->val;
	ext->enclen=port->port.len;
	ext->encval=ltpc_alloc(port->port.len);
	if (!ext->encval) LTPERR(LTP_MEMORY);
	memcpy(ext->encval,port->port.val,ext->enclen);
	return(LTP_OK);
}

/*!
 * @brief print a port extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_port(ltpc_ext *ext, int strlen, char *strbuf)
{
	if (!ext) LTPERR(LTP_BAD_PARAM);
	ltpe_port *port=(ltpe_port*)ext->val;
	if (!port) {
		int rv=ltpe_dec_port(ext);
		if (rv) LTPERR(rv);
	}
	if (!port) LTPERR(LTP_BAD_VALUE);
	char cb[MAXPATH];
	int rv=bin2AH(port->port.val,port->port.len,cb,MAXPATH);
	if (rv) LTPERR(rv);
	snprintf(strbuf,strlen,"port: %s",cb);
	return(LTP_OK); 
}

/*! 
 * @brief decode a hop_count extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_hop_count(ltpc_ext *ext)
{
	if (!ext->enclen || ext->enclen > 1) return(LTP_BAD_VALUE);
	ltpe_hop_count	*hop_count=(ltpe_hop_count*)ltpc_alloc(sizeof(ltpe_hop_count));
	if (!hop_count) LTPERR(LTP_MEMORY);
	unsigned int foo;
	switch (ext->enclen) {
		case 1: foo = ext->encval[0]; break;
		case 2: foo = (ext->encval[1] << 8) + ext->encval[0]; break;
		case 3: foo = (ext->encval[2] << 16) + (ext->encval[1] << 8) + ext->encval[0]; break;
		case 4: foo = (ext->encval[3] << 24) + (ext->encval[2] << 16) + (ext->encval[1] << 8) + ext->encval[0]; break;
		default:
			LTPERR(LTP_BAD_VALUE);
	}
	hop_count->hop_count=foo;
	if (hop_count->hop_count > LTPT_MAX_HOPS) LTPERR(LTP_BAD_VALUE);
	ext->val=(void*)hop_count;
	ext->len=sizeof(ltpe_hop_count);
	return(LTP_OK);
}

/*!
 * @brief encode a hop_count extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_hop_count(ltpc_ext *ext)
{
	// lazily we always use 1 byte for the encoding
	ext->encval=ltpc_alloc(1);
	if (!ext->encval) LTPERR(LTP_MEMORY);
	ltpe_hop_count *hop_count=(ltpe_hop_count*)ext->val;
	if (hop_count->hop_count >= LTPT_MAX_HOPS) LTPERR(LTP_BAD_VALUE);
	ext->encval[0]=(hop_count->hop_count&0xff);
	ext->enclen=1;
	return(LTP_OK);
}

/*!
 * @brief print a hop_count extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_hop_count(ltpc_ext *ext, int strlen, char *strbuf)
{
	if (!ext) LTPERR(LTP_BAD_PARAM);
	ltpe_hop_count *hop_count=(ltpe_hop_count*)ext->val;
	if (!hop_count) LTPERR(LTP_BAD_VALUE);
	snprintf(strbuf,strlen,"hop count: %d",hop_count->hop_count);
	return(LTP_OK); 
}

/*! 
 * @brief decode a via extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_via(ltpc_ext *ext)
{
	if (!ext->enclen) LTPERR(LTP_BAD_VALUE);
	ltpe_via *via=(ltpe_via*)ltpc_alloc(sizeof(ltpe_via));
	if (!via) LTPERR(LTP_MEMORY);
	via->via.val=ltpc_alloc(ext->enclen);
	via->via.len=ext->enclen;
	memcpy(via->via.val,ext->encval,ext->enclen);
	ext->val=(void*)via;
	ext->len=sizeof(via);
	return(LTP_OK);
}

/*!
 * @brief encode a via extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_via(ltpc_ext *ext)
{
	ltpe_via *via=(ltpe_via*)ext->val;
	ext->enclen=via->via.len;
	if (via->via.len) {
		ext->encval=ltpc_alloc(via->via.len);
		if (!ext->encval) LTPERR(LTP_MEMORY);
		memcpy(ext->encval,via->via.val,ext->enclen);
	}
	return(LTP_OK);
}

/*!
 * @brief print a via extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_via(ltpc_ext *ext, int strlen, char *strbuf)
{
	snprintf(strbuf,strlen,"via: (%d) %s",ext->enclen,(char*)ext->encval);
	return(LTP_OK);
}

/*! 
 * @brief decode a auth extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_t_auth(ltpc_ext *ext)
{
	// same structure, same code
	int rv=ltpe_dec_auth(ext);
	if (rv) LTPERR(rv);
	return(LTP_OK);
}

/*!
 * @brief encode a auth extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_t_auth(ltpc_ext *ext)
{
	// same structure, same code
	int rv=ltpe_enc_auth(ext);
	if (rv) LTPERR(rv);
	return(LTP_OK);
}

/*!
 * @brief print a auth extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_t_auth(ltpc_ext *ext, int strlen, char *strbuf){return(LTP_OK);}

/*! 
 * @brief decode a ecn extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_ecn(ltpc_ext *ext)
{
	ltpe_ecn *ecn=(ltpe_ecn*)ltpc_alloc(sizeof(ltpe_ecn));
	int remaining=ext->enclen;
	int index=0; 
	int totalsize=0;
	unsigned char *bp=ext->encval;
	while (remaining>0) {
		// peer-len
		int peerlen = *bp++ << 8;
		peerlen += *bp++;
		totalsize += peerlen + 4 + 4 + 4; // whatever
		// make space - should have an ltpc_ltpc_realloc fnc!! TODO
		ecn->peers=(ltpc_octets*)ltpc_realloc((unsigned char*)ecn->peers,(index+1)*sizeof(ltpc_octets));
		if (!ecn->peers) LTPERR(LTP_MEMORY);
		ecn->peers[index].val=ltpc_alloc(peerlen);
		if (!ecn->peers[index].val) LTPERR(LTP_MEMORY);
		ecn->peers[index].len=peerlen;
		// peer
		memcpy(ecn->peers[index].val,bp,peerlen);
		bp+=peerlen;
		// units
		ecn->units=(int*)ltpc_realloc((unsigned char*)ecn->units,(index+1)*sizeof(int));
		if (!ecn->units) LTPERR(LTP_MEMORY);
		char cv=*bp++; // sign extension (hopefully)
		ecn->units[index]=cv;
		// duration 
		ecn->durations=(unsigned int*)ltpc_realloc((unsigned char*)ecn->durations,(index+1)*sizeof(unsigned int));
		if (!ecn->durations) LTPERR(LTP_MEMORY);
		ecn->durations[index] = (*bp++ << 16) + (*bp++ << 8) + *bp++;
		// update 
		index++;
	}
	ecn->necn=index;
	ext->val=(void*)ecn;
	ext->len=totalsize;
	return(LTP_OK);
}

/*!
 * @brief encode a ecn extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_ecn(ltpc_ext *ext)
{
	// this is inefficient but good enough
	ltpe_ecn *ecn=(ltpe_ecn*)ext->val;
	// figure out overall length
	ext->enclen=0;
	for (int i=0;i!=ecn->necn;i++) {
		if (ecn->peers[i].len > LTPT_MAX_PEER)  LTPERR(LTP_BAD_VALUE);
		if (ecn->units[i] < LTPT_MIN_UNITS || ecn->units[i] > LTPT_MAX_UNITS) LTPERR(LTP_BAD_VALUE);
		if (ecn->durations[i] > LTPT_MAX_DURATION ) LTPERR(LTP_BAD_VALUE);
		ext->enclen += 4+2+ecn->peers[i].len; // 4 time, 2 len, len value
	}
	ext->encval=ltpc_alloc(ext->enclen);
	if (!ext->encval) LTPERR(LTP_MEMORY);
	unsigned char *bp=ext->encval; // buffer pointer
	for (int i=0;i!=ecn->necn;i++) {
		// peer len
		*bp++ = ( ecn->peers[i].len >> 8 ) & 0xff;
		*bp++ = ecn->peers[i].len & 0xff;
		// peer val
		memcpy(bp,ecn->peers[i].val, ecn->peers[i].len);
		bp += ecn->peers[i].len;
		// time units
		*bp++ = ecn->units[i] ;
		// time val
		*bp++ = (ecn->durations[i] >> 16 ) & 0xff;
		*bp++ = (ecn->durations[i] >> 8 ) & 0xff;
		*bp++ = ecn->durations[i] & 0xff;
	}
	return(LTP_OK);
}

/*!
 * @brief print a ecn extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_ecn(ltpc_ext *ext, int strlen, char *strbuf){return(LTP_OK);}

/*!
 * @brief free an extension that only contais an ltpc_octets field (cookie,port,via)
 * @param ext is the extension structure
 */
void ltpe_free_octets_ext(ltpc_ext *ext)
{
	ltpe_cookie *ck=(ltpe_cookie*)ext->val;
	if (ck && ck->cval.val) { 
		ltpc_free(ck->cval.val);
		ck->cval.val=NULL; // so ltpc_free doesn't try do it twice
	}
	return;
}

/*!
 * @brief free an extension that only contais an ltpaddr field (src,dest)
 * @param ext is the extension structure
 */
void ltpe_free_addr_ext(ltpc_ext *ext)
{
	ltpe_src *src=(ltpe_src*)ext->val;
	if (!src) return;
	if (src->src.octets.magic==LTP_ADDR_OCT_MAGIC) { // its octets
		ltpc_free(src->src.octets.val);
		src->src.octets.val=NULL;
	}
	return;
}

/*!
 * @brief free an auth extension 
 * @param ext is the extension structure
 */
void ltpe_free_auth_ext(ltpc_ext *ext)
{
	if (ext->header) {
		ltpe_auth_header *hdr=(ltpe_auth_header*)ext->val;
		if (!hdr) return;
		if (hdr->keyid.val) ltpc_free(hdr->keyid.val);
		hdr->keyid.val=NULL;
	} else {
		ltpe_auth_trailer *tlr=(ltpe_auth_trailer*)ext->val;
		if (!tlr) return;
		ltpc_free(tlr->authval.val);
		tlr->authval.val=NULL;
	}
	return;
}

/*!
 * @brief free an ecn extension 
 * @param ext is the extension structure
 */
void ltpe_free_ecn_ext(ltpc_ext *ext)
{
	ltpe_ecn *ecn=(ltpe_ecn*)ext->val;
	if (!ecn) return;
	ltpc_free((unsigned char*)ecn->peers);
	ecn->peers=NULL;
	ltpc_free((unsigned char*)ecn->units);
	ecn->units=NULL;
	ltpc_free((unsigned char*)ecn->durations);
	ecn->durations=NULL;
	ecn->necn=0;
	return;
}

/*!
 * @brief generic free fnc that calls others via jump table
 * @param ext is the extension to free the internals for
 */
void ltpe_free_ext(ltpc_ext *ext)
{
	int extind=-1;
	for (int i=0;i!=HANDLED_EXTS && extind==-1 ;i++) {
		if (ext->type==e_fncs[i].type) {
			extind=i;
		}
	}
	if (extind==-1) {
		// dunno what to do but just leak the memory
		snprintf(ltpstr,LTPSTR_LEN,"Error: unhandled extension type (%d) - not freeing at %s:%d\n",
				ext->type,__FILE__,__LINE__);
		ltp_log(LTP_LOG_ERR,ltpstr);
		return;
	}
	if (e_fncs[extind].e_free) {
		(e_fncs[extind].e_free)(ext);
	}
	return;
}

/*! 
 * @brief decode a auth extension
 * @param ext is the extension
 * @return is zero or error
 */
int ltpe_dec_e2eauth(ltpc_ext *ext)
{
	int rv;
	int len=ext->enclen;
	unsigned char *val=ext->encval;
	ltpe_e2eauth	*hdr=(ltpe_e2eauth*)ltpc_alloc(sizeof(ltpe_e2eauth));
	if (!hdr) LTPERR(LTP_MEMORY);
	hdr->ciphersuite=val[0];
	int kidlen=0;
	int enckidlen=1;
	if (val[1]) { // there's a keyid
		// decode the sdnv-encoded keyid len
		int blen=len-1;
		rv=sdnv_dec_int(&blen,val+1,&kidlen);
		enckidlen=len-blen-1; // # bytes comsumed
		if (rv) LTPERR(rv);
		hdr->keyid.len=kidlen;
		hdr->keyid.val=(unsigned char*)ltpc_alloc(kidlen+1);
		if (!hdr->keyid.val) LTPERR(LTP_MEMORY);
		memcpy(hdr->keyid.val,val+1+enckidlen,kidlen);
	} else {
		hdr->keyid.len=0;
		hdr->keyid.val=NULL;
		enckidlen=1;
		kidlen=0;
	}
	hdr->authval.len=len-(1+enckidlen+kidlen);
	hdr->authval.val=(unsigned char*)ltpc_alloc(hdr->authval.len);
	if (!hdr->authval.val) LTPERR(LTP_MEMORY);
	memcpy(hdr->authval.val,val+(1+enckidlen+kidlen),hdr->authval.len);
	ext->val=(void*)hdr;
	ext->len=sizeof(ltpe_e2eauth);
	return(LTP_OK);
}

/*!
 * @brief encode a auth extension
 * @param ext is the extension
 * @return zero or error
 */
int ltpe_enc_e2eauth(ltpc_ext *ext)
{
	if (ext->len!=sizeof(ltpe_e2eauth)) LTPERR(LTP_BAD_VALUE);
	ltpe_e2eauth *hdr=(ltpe_e2eauth*)ext->val;
	if (!hdr->authval.len) LTPERR(LTP_BAD_VALUE);
	int kidlen;
	int rv;
	int enckidlen=MAXPATH;
	unsigned char enckid[MAXPATH];
	if (hdr->keyid.len) {
		kidlen=hdr->keyid.len;
	int rv=sdnv_enc_int(kidlen,&enckidlen,enckid);
	} else {
		kidlen=0;
		enckidlen=1;
		enckid[0]=0x00;
	}
	int len= 1 + enckidlen + kidlen + hdr->authval.len;
	ext->enclen=len;
	ext->encval=ltpc_alloc(len);
	if (!ext->encval) LTPERR(LTP_MEMORY);
	ext->encval[0]=hdr->ciphersuite;
	memcpy(ext->encval+1,enckid,enckidlen);
	if (hdr->keyid.len) {
		memcpy(ext->encval+1+enckidlen,hdr->keyid.val,kidlen);
	} 
	memcpy(ext->encval + 1 + enckidlen + kidlen, hdr->authval.val,hdr->authval.len);
	return(LTP_OK);
}

/*!
 * @brief print a auth extension
 * @param ext is the extension
 * @param strlen is the size of the caller-allocated string buffer
 * @prarm strbug is the caller-allocated string buffer
 * @return zero or error
 */
int ltpe_print_e2eauth(ltpc_ext *ext, int strlen, char *strbuf)
{
	int rv;
	if (!ext->val) {
		rv=ltpe_dec_e2eauth(ext);
		if (rv) {
			snprintf(strbuf,strlen,"E2E Auth header - error can't decode header!");
				return(LTP_OK);
		} 
	}
	ltpe_e2eauth	*aval=(ltpe_e2eauth*)ext->val;
	if (!aval) {
		snprintf(strbuf,strlen,"E2E Auth header - NULL decoded header!");
		return(LTP_OK);
	}
	char kval[MAXPATH];
	kval[0]='\0';
	if (aval->keyid.len) {
		// see if the keyid is all printable, if so try stick that out
		bool allprint=true;
		for (int i=0;allprint && i!=aval->keyid.len;i++) {
			if (!isprint(aval->keyid.val[i])) {
				allprint=false;
			}
		}
		if (!allprint) {
			rv=bin2AH(aval->keyid.val,aval->keyid.len,kval,MAXPATH);
			if (rv) LTPERR(rv);
		} else {
			if (aval->keyid.len<MAXPATH) {
				memcpy(kval, (char*)aval->keyid.val, aval->keyid.len);
				kval[aval->keyid.len]='\0';
			} else {
				snprintf(kval,MAXPATH,"%s",(char*)aval->keyid.val);
				kval[MAXPATH-1]='\0';
			}
		}
	} else {
		snprintf(kval,MAXPATH,"(none)");
	}
	char cval[MAXPATH];
	switch (aval->ciphersuite) {
		case LTP_CS_MAC:snprintf(cval,MAXPATH,"OriginAuth (MAC)");break;
		case LTP_CS_SIG: snprintf(cval,MAXPATH,"Signature");break;
		case LTP_CS_NULL:snprintf(cval,MAXPATH,"NULL (hardcoded-MAC)");break;
		default:snprintf(cval,MAXPATH,"Unknown ciphersuite: %d",aval->ciphersuite);
	}
	bool truncated;
	int plen=4;
	if (aval->authval.len>plen) {
		truncated=true;
	} else {
		truncated=false;
	}
	char vals[MAXPATH];
	vals[0]='\0';
	rv=bin2AH(aval->authval.val,plen,vals,MAXPATH);
	if (rv) LTPERR(rv);
	snprintf(strbuf,strlen,"e2e auth, ciphersuite=%s, keyId=%s\nPP:\t\t\tval: (%d bits) %s%s",
				cval,kval,aval->authval.len*8,vals,(truncated?"...":""));
	return(LTP_OK);
}

/*!
 * @brief free an e2eauth extension 
 * @param ext is the extension structure
 */
void ltpe_free_e2eauth(ltpc_ext *ext)
{
	ltpe_e2eauth *aval=(ltpe_e2eauth*)ext->val;
	if (!aval) return;
	if (aval->keyid.len) { ltpc_free(aval->keyid.val); aval->keyid.len=0;}
	if (aval->authval.len) { ltpc_free(aval->authval.val); aval->authval.len=0;}
	return;
}
