/*!
 * @file ltp_file.cc
 * @brief This is the API for synching LTP state to/from disc
 * @version $Revision: 1.10 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ltp_base.h"
#include "ltp.h"
#include "ltp.h"
#include "ltp_codec.h"
#include "ltp_int.h"
#include "ltp_file.h"

/// this is the basic file name we use for current state
static char statefile[MAXPATH];

/// since those aren't alloc'd we need to know if they're initialised
static bool filesdone=false;

/// how the upper layer signals there's stuff to save
extern bool ltp_state_changed;

/// the main state variable
extern ltp_state ltp_state_var;

#define LTP_ST_WRITING 0 ///< An action for walking the state tree when syncing to disk
#define LTP_ST_READING 1 ///< An action for walking the state tree when syncing from disk
#define LTP_ST_FREEING 2 ///< An action for walking the state tree when freeing an alloc'd state tree
#define LTP_ST_COLLECT 2 ///< An action for walking the state tree when garbage collecting

/*!
 * @brief make a backup (or more than one) of the previous statefile, if any
 * @return zero or eror
 */
static int backupfile(void)
{
	if (!filesdone) return(LTP_OK); // nothing to do 
	struct stat sf_stat;
	int rv;
	if (stat(statefile,&sf_stat)) return(LTP_OK); // still nothing to do
	char backupfile[MAXPATH];
#ifdef LTP_MULTILAYER_BACKUP
	snprintf(backupfile,MAXPATH,"%s.XXXXXX",statefile);
	int fd=mkstemp(backupfile);
	if (fd==-1) LTPERR(LTP_FILE);
	close(fd);
#else
	snprintf(backupfile,MAXPATH,"%s.bup",statefile);
#endif
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"LTP_FILE: backup state file is: %s\n",backupfile);
	ltp_log(LTP_LOG_ADV,ltpstr);
#endif
	rv=rename(statefile,backupfile);
	if (rv) LTPERR(rv);
	return(LTP_OK);
}

static void free_ltpt_vals(ltpt_vals *vals)
{
	if (vals->via.len) free(vals->via.val);
	if (vals->e2e_a_kid.len) free(vals->e2e_a_kid.val);
	if (vals->e2e_a_val.len) free(vals->e2e_a_val.val);
	if (vals->necn) {
		free(vals->peers);
		free(vals->units);
		free(vals->durations);
	}
	memset(vals,0,sizeof(ltpt_vals));
	return;
}

static int write_ltpt_vals(ltpt_vals *vals,FILE *fp)
{
	int rv;
	// if (!vals) LTPERR(LTP_BAD_VALUE);
	// fwrite(vals,sizeof(ltpt_vals),1,fp);
	// if (rv!=1) LTPERR(rv);
	if (vals->via.val) {
		rv=fwrite(vals->via.val,1,vals->via.len,fp);
		if (rv!=vals->via.len) LTPERR(rv);
	}
	if (vals->e2e_a_kid.len) {
		rv=fwrite(vals->e2e_a_kid.val,1,vals->e2e_a_kid.len,fp);
		if (rv!=vals->e2e_a_kid.len) LTPERR(rv);
	}
	if (vals->e2e_a_val.len) {
		rv=fwrite(vals->e2e_a_val.val,1,vals->e2e_a_val.len,fp);
		if (rv!=vals->e2e_a_val.len) LTPERR(rv);
	}
	if (vals->necn) {
		rv=fwrite(vals->peers,sizeof(ltpaddr),vals->necn,fp);
		if (rv!=vals->necn) LTPERR(rv);
		rv=fwrite(vals->units,sizeof(int),vals->necn,fp);
		if (rv!=vals->necn) LTPERR(rv);
		rv=fwrite(vals->durations,sizeof(unsigned int),vals->necn,fp);
		if (rv!=vals->necn) LTPERR(rv);
	}
	return(LTP_OK);
}

static int read_ltpt_vals(ltpt_vals *vals,FILE *fp)
{
	int rv;
	if (vals->via.len) {
		vals->via.val=(unsigned char*)calloc(1,vals->via.len);
		if (!vals->via.val) LTPERR(LTP_MEMORY);
		rv=fread(vals->via.val,1,vals->via.len,fp);
		if (rv!=vals->via.len) LTPERR(rv);
	}
	if (vals->e2e_a_kid.len) {
		vals->e2e_a_kid.val=(unsigned char*)calloc(1,vals->e2e_a_kid.len);
		if (!vals->e2e_a_kid.val) LTPERR(LTP_MEMORY);
		rv=fread(vals->e2e_a_kid.val,1,vals->e2e_a_kid.len,fp);
		if (rv!=vals->e2e_a_kid.len) LTPERR(rv);
	}
	if (vals->e2e_a_val.len) {
		vals->e2e_a_val.val=(unsigned char*)calloc(1,vals->e2e_a_val.len);
		if (!vals->e2e_a_val.val) LTPERR(LTP_MEMORY);
		rv=fread(vals->e2e_a_val.val,1,vals->e2e_a_val.len,fp);
		if (rv!=vals->e2e_a_val.len) LTPERR(rv);
	}
	if (vals->necn) {
		vals->peers=(ltpaddr*)calloc(sizeof(ltpaddr),vals->necn);
		if (!vals->peers) LTPERR(LTP_MEMORY);
		rv=fread(vals->peers,sizeof(ltpaddr),vals->necn,fp);
		if (rv!=vals->necn) LTPERR(rv);
		vals->units=(int*)calloc(sizeof(int),vals->necn);
		if (!vals->units) LTPERR(LTP_MEMORY);
		rv=fread(vals->units,sizeof(int),vals->necn,fp);
		if (rv!=vals->necn) LTPERR(rv);
		vals->durations=(unsigned int*)calloc(sizeof(unsigned int),vals->necn);
		if (!vals->durations) LTPERR(LTP_MEMORY);
		rv=fread(vals->durations,sizeof(unsigned int),vals->necn,fp);
		if (rv!=vals->necn) LTPERR(rv);
	}
	return(LTP_OK);
}

/*!
 * @brief walk down the state tree and read or write stuff
 * @param tree is a pointer to the tree
 * @param action is READING, WRITING or FREEING
 * @return zero or error
 * @todo not sure when replay cache entry finishedwith field will ever get set - have to write some new code 
 * @todo not sure when txq segment finishedwith field will ever get set - have to write some new code 
 */
static int walkstatetree(ltp_state *tree, int action)
{
	if (action!=LTP_ST_READING && action!=LTP_ST_WRITING && action!=LTP_ST_FREEING && action!=LTP_ST_COLLECT) LTPERR(LTP_BAD_PARAM);
	if (!tree) LTPERR(LTP_BAD_PARAM);
	int rv;
	FILE *sfile=NULL;
	if (action==LTP_ST_READING) {
		sfile=fopen(statefile,"rb");
		if (!sfile) {
			snprintf(ltpstr,LTPSTR_LEN,"Can't open statefile %s\n",statefile);
			ltp_log(LTP_LOG_MAX,ltpstr);
			LTPERR(LTP_FILE);
		}
	}
	if (action==LTP_ST_WRITING) {
		rv=backupfile();
		if (rv) LTPERR(rv);
		sfile=fopen(statefile,"wb");
		if (!sfile) LTPERR(LTP_FILE);
	}

#define GCARRAY(GCAtype,GCAstart,GCAcount) { \
		GCAtype *write=GCAstart; \
		GCAtype *read=write; \
		int     count=0; \
		for (int i=0;i!=(GCAcount);i++) { \
			if (read->finishedwith) { \
				read++; \
				count++; \
			} else if (write!=read) { \
				*write++=*read++; \
			} else { \
				read++; \
				write++; \
			}\
		} \
		GCAcount -= count; \
	}

	if (action==LTP_ST_WRITING || action==LTP_ST_COLLECT) {
		// get rid of unwanted sockets etc.
		// missing code here to set finishedwith in replay cache 
		// based on some timer?
		GCARRAY(ltp_socket_t,tree->sockets,tree->nsockets);
		GCARRAY(ltp_block,tree->lbq.blocks,tree->lbq.nblocks);
		// if a segment is not (apparently) finishedwith but its parent is, then we can
		// force finishedwith for the segment - this'll happen with protocol artefacts
		// like RA etc.
		for (int i=0;i!=tree->txq.nsegments;i++) {
			if (!tree->txq.segments[i].finishedwith) {
				rv=ltp_fixparent(tree,(void**)&tree->txq.segments[i].parent,tree->txq.segments[i].parentid,LTP_PARENT_BLOCK);
				if (rv) LTPERR(rv);
				if (tree->txq.segments[i].parent->finishedwith) {
					tree->txq.segments[i].finishedwith=true; // aha!!
				}
			}
		}
		GCARRAY(ltp_segment,tree->txq.segments,tree->txq.nsegments);
		GCARRAY(ltp_cache_detail,tree->replays.details,tree->replays.nentries);
		GCARRAY(ltp_timer_detail,tree->timers.details,tree->timers.nentries);
	}

	if (action==LTP_ST_FREEING) {
		// each socket
		for (int i=0;i!=tree->nsockets;i++) {
			for (int j=0;j!=tree->sockets[i].nblocks;j++) {
				if (tree->sockets[i].blocks[j].inboundcookie.val) 
					free(tree->sockets[i].blocks[j].inboundcookie.val);
				if (tree->sockets[i].blocks[j].outboundcookie.val) 
					free(tree->sockets[i].blocks[j].outboundcookie.val);
				if (tree->sockets[i].blocks[j].decoded.val) 
					free(tree->sockets[i].blocks[j].decoded.val);
				if (tree->sockets[i].blocks[j].rx_scopes.nscopes) { 
					free(tree->sockets[i].blocks[j].rx_scopes.lbs);
					free(tree->sockets[i].blocks[j].rx_scopes.ubs);
				}
				if (tree->sockets[i].blocks[j].tx_scopes.nscopes) { 
					free(tree->sockets[i].blocks[j].tx_scopes.lbs);
					free(tree->sockets[i].blocks[j].tx_scopes.ubs);
				}
				if (tree->sockets[i].blocks[j].fwd_scopes.nscopes) { 
					free(tree->sockets[i].blocks[j].fwd_scopes.lbs);
					free(tree->sockets[i].blocks[j].fwd_scopes.ubs);
				}
				free_ltpt_vals(&tree->sockets[i].blocks[j].t_vals);
			}
			if (tree->sockets[i].blocks)
				free(tree->sockets[i].blocks);
		}
		if (tree->sockets)
			free(tree->sockets);
		// lpq loopback queue entries
		for (int i=0;i!=tree->lbq.nblocks;i++) {
			if (tree->lbq.blocks[i].inboundcookie.val) 
				free(tree->lbq.blocks[i].inboundcookie.val);
			if (tree->lbq.blocks[i].outboundcookie.val) 
				free(tree->lbq.blocks[i].outboundcookie.val);
			if (tree->lbq.blocks[i].decoded.val) 
				free(tree->lbq.blocks[i].decoded.val);
			if (tree->lbq.blocks[i].rx_scopes.nscopes) { 
				free(tree->lbq.blocks[i].rx_scopes.lbs);
				free(tree->lbq.blocks[i].rx_scopes.ubs);
			}
			if (tree->lbq.blocks[i].tx_scopes.nscopes) { 
				free(tree->lbq.blocks[i].tx_scopes.lbs);
				free(tree->lbq.blocks[i].tx_scopes.ubs);
			}
			if (tree->lbq.blocks[i].fwd_scopes.nscopes) { 
				free(tree->lbq.blocks[i].fwd_scopes.lbs);
				free(tree->lbq.blocks[i].fwd_scopes.ubs);
			}
			free_ltpt_vals(&tree->lbq.blocks[i].t_vals);
		}
		if (tree->lbq.blocks)
			free(tree->lbq.blocks);
		// txq the transmit queue entries
		// write out the encoding of each segment too
		for (int i=0;i!=tree->txq.nsegments;i++) {
			ltpc_free_seg(&tree->txq.segments[i].seg);
		}
		if (tree->txq.segments)
			free(tree->txq.segments);
		// replay entries
		for (int i=0;i!=tree->replays.nentries;i++) {
			// write hash val
			free(tree->replays.details[i].hash.val);
			// responses (if any)
			if (tree->replays.details[i].response.val)
				free(tree->replays.details[i].response.val);
		}
		if (tree->replays.details)
			free(tree->replays.details);
		// timer entries
		if (tree->timers.details)
			free(tree->timers.details);
	}

#define SFILE_LTPERR(foo) { fclose(sfile);LTPERR(foo); }

	if (action==LTP_ST_WRITING) {
		// top level
		rv=fwrite(tree,sizeof(ltp_state),1,sfile);
		if (rv!=1) SFILE_LTPERR(LTP_FILE);
		// each socket
		rv=fwrite(tree->sockets,sizeof(ltp_socket_t),tree->nsockets,sfile);
		if (rv!=tree->nsockets) SFILE_LTPERR(LTP_FILE);
		for (int i=0;i!=tree->nsockets;i++) {
			rv=fwrite(tree->sockets[i].blocks,sizeof(ltp_block),tree->sockets[i].nblocks,sfile);
			if (rv!=tree->sockets[i].nblocks) SFILE_LTPERR(LTP_FILE);
			for (int j=0;j!=tree->sockets[i].nblocks;j++) {
				if (tree->sockets[i].blocks[j].decoded.val) {
					rv=fwrite(tree->sockets[i].blocks[j].decoded.val,
							sizeof(unsigned char),
							tree->sockets[i].blocks[j].decoded.len,
							sfile);
					if (rv!= tree->sockets[i].blocks[j].decoded.len)
							SFILE_LTPERR(LTP_FILE);
				}
				if (tree->sockets[i].blocks[j].inboundcookie.val) {
					rv=fwrite(tree->sockets[i].blocks[j].inboundcookie.val,
							sizeof(unsigned char),
							tree->sockets[i].blocks[j].inboundcookie.len,
							sfile);
					if (rv!= tree->sockets[i].blocks[j].inboundcookie.len)
							SFILE_LTPERR(LTP_FILE);
				}
				if (tree->sockets[i].blocks[j].outboundcookie.val) {
					rv=fwrite(tree->sockets[i].blocks[j].outboundcookie.val,
							sizeof(unsigned char),
							tree->sockets[i].blocks[j].outboundcookie.len,
							sfile);
					if (rv!= tree->sockets[i].blocks[j].outboundcookie.len)
							SFILE_LTPERR(LTP_FILE);
				}
				if (tree->sockets[i].blocks[j].rx_scopes.nscopes) { 
					rv=fwrite(tree->sockets[i].blocks[j].rx_scopes.lbs,
							sizeof(int),
							tree->sockets[i].blocks[j].rx_scopes.nscopes,
							sfile);
					if (rv!=tree->sockets[i].blocks[j].rx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
					rv=fwrite(tree->sockets[i].blocks[j].rx_scopes.ubs,
							sizeof(int),
							tree->sockets[i].blocks[j].rx_scopes.nscopes,
							sfile);
					if (rv!=tree->sockets[i].blocks[j].rx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
				}
				if (tree->sockets[i].blocks[j].tx_scopes.nscopes) { 
					rv=fwrite(tree->sockets[i].blocks[j].tx_scopes.lbs,
							sizeof(int),
							tree->sockets[i].blocks[j].tx_scopes.nscopes,
							sfile);
					if (rv!=tree->sockets[i].blocks[j].tx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
					rv=fwrite(tree->sockets[i].blocks[j].tx_scopes.ubs,
							sizeof(int),
							tree->sockets[i].blocks[j].tx_scopes.nscopes,
							sfile);
					if (rv!=tree->sockets[i].blocks[j].tx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
				}

				if (tree->sockets[i].blocks[j].fwd_scopes.nscopes) { 
					rv=fwrite(tree->sockets[i].blocks[j].fwd_scopes.lbs,
							sizeof(int),
							tree->sockets[i].blocks[j].fwd_scopes.nscopes,
							sfile);
					if (rv!=tree->sockets[i].blocks[j].fwd_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
					rv=fwrite(tree->sockets[i].blocks[j].fwd_scopes.ubs,
							sizeof(int),
							tree->sockets[i].blocks[j].fwd_scopes.nscopes,
							sfile);
					if (rv!=tree->sockets[i].blocks[j].fwd_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
				}
				rv=write_ltpt_vals(&tree->sockets[i].blocks[j].t_vals,sfile);
				if (rv) SFILE_LTPERR(LTP_FILE);
			}
		}
		// lpq loopback queue entries
		rv=fwrite(tree->lbq.blocks,sizeof(ltp_block),tree->lbq.nblocks,sfile);
		if (rv!=tree->lbq.nblocks) SFILE_LTPERR(LTP_FILE);
		for (int i=0;i!=tree->lbq.nblocks;i++) {
			if (tree->lbq.blocks[i].decoded.val) {
				rv=fwrite(tree->lbq.blocks[i].decoded.val,
						tree->lbq.blocks[i].decoded.len,
						sizeof(unsigned char),
						sfile);
				if (rv!=sizeof(unsigned char)) SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].inboundcookie.val) {
				rv=fwrite(tree->lbq.blocks[i].inboundcookie.val,
						tree->lbq.blocks[i].inboundcookie.len,
						sizeof(unsigned char),
						sfile);
				if (rv!=sizeof(unsigned char)) SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].outboundcookie.val) {
				rv=fwrite(tree->lbq.blocks[i].outboundcookie.val,
						tree->lbq.blocks[i].outboundcookie.len,
						sizeof(unsigned char),
						sfile);
				if (rv!=sizeof(unsigned char)) SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].rx_scopes.nscopes) { 
				rv=fwrite(tree->lbq.blocks[i].rx_scopes.lbs,
						sizeof(int),
						tree->lbq.blocks[i].rx_scopes.nscopes,
						sfile);
				if (rv!=tree->lbq.blocks[i].rx_scopes.nscopes)
						SFILE_LTPERR(LTP_FILE);
				rv=fwrite(tree->lbq.blocks[i].rx_scopes.ubs,
						sizeof(int),
						tree->lbq.blocks[i].rx_scopes.nscopes,
						sfile);
				if (rv!=tree->lbq.blocks[i].rx_scopes.nscopes)
						SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].tx_scopes.nscopes) { 
				rv=fwrite(tree->lbq.blocks[i].tx_scopes.lbs,
						sizeof(int),
						tree->lbq.blocks[i].tx_scopes.nscopes,
						sfile);
				if (rv!=tree->lbq.blocks[i].tx_scopes.nscopes)
						SFILE_LTPERR(LTP_FILE);
				rv=fwrite(tree->lbq.blocks[i].tx_scopes.ubs,
						sizeof(int),
						tree->lbq.blocks[i].tx_scopes.nscopes,
						sfile);
				if (rv!=tree->lbq.blocks[i].tx_scopes.nscopes)
						SFILE_LTPERR(LTP_FILE);
			}

			if (tree->lbq.blocks[i].fwd_scopes.nscopes) { 
				rv=fwrite(tree->lbq.blocks[i].fwd_scopes.lbs,
						sizeof(int),
						tree->lbq.blocks[i].fwd_scopes.nscopes,
						sfile);
				if (rv!=tree->lbq.blocks[i].fwd_scopes.nscopes)
						SFILE_LTPERR(LTP_FILE);
				rv=fwrite(tree->lbq.blocks[i].fwd_scopes.ubs,
						sizeof(int),
						tree->lbq.blocks[i].fwd_scopes.nscopes,
						sfile);
				if (rv!=tree->lbq.blocks[i].fwd_scopes.nscopes)
						SFILE_LTPERR(LTP_FILE);
			}

			rv=write_ltpt_vals(&tree->lbq.blocks[i].t_vals,sfile);
			if (rv) SFILE_LTPERR(LTP_FILE);
		}
		// txq the transmit queue entries
		rv=fwrite(tree->txq.segments,sizeof(ltp_segment),tree->txq.nsegments,sfile);
		if (rv!=tree->txq.nsegments) SFILE_LTPERR(LTP_FILE);
		// write out the encoding of each segment too
		for (int i=0;i!=tree->txq.nsegments;i++) {
			if (tree->txq.segments[i].seg.type!=LTPC_SEG_ENCODED) {
				int blen=LTPC_MAXFSEG;
				unsigned char buf[LTPC_MAXFSEG];
				rv=ltpc_encode(&tree->txq.segments[i].seg,&blen,buf,false);
				if (rv) SFILE_LTPERR(rv);
				rv=fwrite(&blen,sizeof(blen),1,sfile);
				if (rv!=1) SFILE_LTPERR(rv);
				rv=fwrite(buf,sizeof(unsigned char),blen,sfile);
				if (rv!=blen) SFILE_LTPERR(rv);
			} else {
				rv=fwrite(&tree->txq.segments[i].encoded.len,sizeof(tree->txq.segments[i].encoded.len),1,sfile);
				if (rv!=1) SFILE_LTPERR(rv);
				rv=fwrite(tree->txq.segments[i].encoded.val,
							tree->txq.segments[i].encoded.len,
							sizeof(unsigned char),
							sfile);
				if (rv!=1) SFILE_LTPERR(rv);
			}
		}
		// replay entries
		rv=fwrite(tree->replays.details,sizeof(ltp_cache_detail),tree->replays.nentries,sfile);
		if (rv!=tree->replays.nentries) SFILE_LTPERR(LTP_FILE);
		for (int i=0;i!=tree->replays.nentries;i++) {
			// write hash val
			if (tree->replays.details[i].hash.len) {
				rv=fwrite(tree->replays.details[i].hash.val,
					tree->replays.details[i].hash.len,
					sizeof(unsigned char),
					sfile);
				if (rv!=sizeof(unsigned char)) SFILE_LTPERR(rv);
			}
			// responses (if any)
			if (tree->replays.details[i].response.len) {
				rv=fwrite(tree->replays.details[i].response.val,
					sizeof(unsigned char),
					tree->replays.details[i].response.len,
					sfile);
				if (rv!= tree->replays.details[i].response.len) 
					SFILE_LTPERR(rv);
			}
		}
		// timer entries
		rv=fwrite(tree->timers.details,sizeof(ltp_timer_detail),tree->timers.nentries,sfile);
		if (rv!=tree->timers.nentries) SFILE_LTPERR(LTP_FILE);
	}

	if (action==LTP_ST_READING) {
		// top level
		rv=fread(tree,sizeof(ltp_state),1,sfile);
		if (rv!=1) SFILE_LTPERR(LTP_FILE);
		// each socket
		if (tree->nsockets) {
			tree->sockets=(ltp_socket_t*)calloc(sizeof(ltp_socket_t),tree->nsockets);
			if (!tree->sockets) SFILE_LTPERR(LTP_MEMORY);
			tree->alloced=tree->nsockets;
			rv=fread(tree->sockets,sizeof(ltp_socket_t),tree->nsockets,sfile);
			if (rv!=tree->nsockets) SFILE_LTPERR(LTP_FILE);
		} else {
			tree->sockets=NULL;
			tree->alloced=0;
		}
		for (int i=0;i!=tree->nsockets;i++) {
			if (tree->sockets[i].nblocks) {
				tree->sockets[i].blocks=(ltp_block*)calloc(sizeof(ltp_block),tree->sockets[i].nblocks);
				if (!tree->sockets[i].blocks) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->sockets[i].blocks,sizeof(ltp_block),tree->sockets[i].nblocks,sfile);
				if (rv!=tree->sockets[i].nblocks) SFILE_LTPERR(LTP_FILE);
				for (int j=0;j!=tree->sockets[i].nblocks;j++) {
					if (tree->sockets[i].blocks[j].decoded.len) {
						tree->sockets[i].blocks[j].decoded.val=
							(unsigned char*)calloc(sizeof(unsigned char),
									tree->sockets[i].blocks[j].decoded.len);
						if (!tree->sockets[i].blocks[j].decoded.val) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].decoded.val,
							sizeof(unsigned char),
							tree->sockets[i].blocks[j].decoded.len,
							sfile);
						if (rv!= tree->sockets[i].blocks[j].decoded.len) 
								SFILE_LTPERR(LTP_FILE);
					}
					if (tree->sockets[i].blocks[j].inboundcookie.len) {
						tree->sockets[i].blocks[j].inboundcookie.val=
							(unsigned char*)calloc(sizeof(unsigned char),
									tree->sockets[i].blocks[j].inboundcookie.len);
						if (!tree->sockets[i].blocks[j].inboundcookie.val) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].inboundcookie.val,
							sizeof(unsigned char),
							tree->sockets[i].blocks[j].inboundcookie.len,
							sfile);
						if (rv!= tree->sockets[i].blocks[j].inboundcookie.len) 
								SFILE_LTPERR(LTP_FILE);
					}
					if (tree->sockets[i].blocks[j].outboundcookie.len) {
						tree->sockets[i].blocks[j].outboundcookie.val=
							(unsigned char*)calloc(sizeof(unsigned char),
									tree->sockets[i].blocks[j].outboundcookie.len);
						if (!tree->sockets[i].blocks[j].outboundcookie.val) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].outboundcookie.val,
							sizeof(unsigned char),
							tree->sockets[i].blocks[j].outboundcookie.len,
							sfile);
						if (rv!= tree->sockets[i].blocks[j].outboundcookie.len) 
								SFILE_LTPERR(LTP_FILE);
					}
					if (tree->sockets[i].blocks[j].rx_scopes.nscopes) { 
						tree->sockets[i].blocks[j].rx_scopes.lbs=
							(int*)calloc(sizeof(int),
									tree->sockets[i].blocks[j].rx_scopes.nscopes);
						if (!tree->sockets[i].blocks[j].rx_scopes.lbs) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].rx_scopes.lbs,
							sizeof(int),
							tree->sockets[i].blocks[j].rx_scopes.nscopes,
							sfile);
						if (rv!=tree->sockets[i].blocks[j].rx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
						tree->sockets[i].blocks[j].rx_scopes.ubs=
							(int*)calloc(sizeof(int),
									tree->sockets[i].blocks[j].rx_scopes.nscopes);
						if (!tree->sockets[i].blocks[j].rx_scopes.ubs) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].rx_scopes.ubs,
							sizeof(int),
							tree->sockets[i].blocks[j].rx_scopes.nscopes,
							sfile);
						if (rv!=tree->sockets[i].blocks[j].rx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
					}
					if (tree->sockets[i].blocks[j].tx_scopes.nscopes) { 
						tree->sockets[i].blocks[j].tx_scopes.lbs=
							(int*)calloc(sizeof(int),
									tree->sockets[i].blocks[j].tx_scopes.nscopes);
						if (!tree->sockets[i].blocks[j].tx_scopes.lbs) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].tx_scopes.lbs,
							sizeof(int),
							tree->sockets[i].blocks[j].tx_scopes.nscopes,
							sfile);
						if (rv!=tree->sockets[i].blocks[j].tx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
						tree->sockets[i].blocks[j].tx_scopes.ubs=
							(int*)calloc(sizeof(int),
									tree->sockets[i].blocks[j].tx_scopes.nscopes);
						if (!tree->sockets[i].blocks[j].tx_scopes.ubs) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].tx_scopes.ubs,
							sizeof(int),
							tree->sockets[i].blocks[j].tx_scopes.nscopes,
							sfile);
						if (rv!=tree->sockets[i].blocks[j].tx_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
					}
					if (tree->sockets[i].blocks[j].fwd_scopes.nscopes) { 
						tree->sockets[i].blocks[j].fwd_scopes.lbs=
							(int*)calloc(sizeof(int),
									tree->sockets[i].blocks[j].fwd_scopes.nscopes);
						if (!tree->sockets[i].blocks[j].fwd_scopes.lbs) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].fwd_scopes.lbs,
							sizeof(int),
							tree->sockets[i].blocks[j].fwd_scopes.nscopes,
							sfile);
						if (rv!=tree->sockets[i].blocks[j].fwd_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
						tree->sockets[i].blocks[j].fwd_scopes.ubs=
							(int*)calloc(sizeof(int),
									tree->sockets[i].blocks[j].fwd_scopes.nscopes);
						if (!tree->sockets[i].blocks[j].fwd_scopes.ubs) SFILE_LTPERR(LTP_MEMORY);
						rv=fread(tree->sockets[i].blocks[j].fwd_scopes.ubs,
							sizeof(int),
							tree->sockets[i].blocks[j].fwd_scopes.nscopes,
							sfile);
						if (rv!=tree->sockets[i].blocks[j].fwd_scopes.nscopes)
							SFILE_LTPERR(LTP_FILE);
					}
					rv=read_ltpt_vals(&tree->sockets[i].blocks[j].t_vals,sfile);
					if (rv) SFILE_LTPERR(LTP_FILE);
				}
			}
		}
		// lpq loopback queue entries
		if (tree->lbq.nblocks) {
			tree->lbq.blocks=(ltp_block*)calloc(sizeof(ltp_block),tree->lbq.nblocks);
			if (!tree->lbq.blocks) SFILE_LTPERR(LTP_MEMORY);
			tree->lbq.alloced=tree->lbq.nblocks;
			rv=fread(tree->lbq.blocks,sizeof(ltp_block),tree->lbq.nblocks,sfile);
			if (rv!=tree->lbq.nblocks) SFILE_LTPERR(LTP_FILE);
		} else {
			tree->lbq.blocks=NULL;
			tree->lbq.alloced=0;
		}
		for (int i=0;i!=tree->lbq.nblocks;i++) {
			if (tree->lbq.blocks[i].decoded.len) {
				tree->lbq.blocks[i].decoded.val=(unsigned char*)
					calloc(sizeof(unsigned char),
							tree->lbq.blocks[i].decoded.len);
				if (!tree->lbq.blocks[i].decoded.val) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].decoded.val,sizeof(unsigned char),tree->lbq.blocks[i].decoded.len,sfile);
				if (rv!=tree->lbq.blocks[i].decoded.len) SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].inboundcookie.len) {
				tree->lbq.blocks[i].inboundcookie.val=(unsigned char*)
					calloc(sizeof(unsigned char),
							tree->lbq.blocks[i].inboundcookie.len);
				if (!tree->lbq.blocks[i].inboundcookie.val) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].inboundcookie.val,sizeof(unsigned char),tree->lbq.blocks[i].inboundcookie.len,sfile);
				if (rv!=tree->lbq.blocks[i].inboundcookie.len) SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].outboundcookie.len) {
				tree->lbq.blocks[i].outboundcookie.val=(unsigned char*)
					calloc(sizeof(unsigned char),
							tree->lbq.blocks[i].outboundcookie.len);
				if (!tree->lbq.blocks[i].outboundcookie.val) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].outboundcookie.val,sizeof(unsigned char),tree->lbq.blocks[i].outboundcookie.len,sfile);
				if (rv!=tree->lbq.blocks[i].outboundcookie.len) SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].rx_scopes.nscopes) { 
				tree->lbq.blocks[i].rx_scopes.lbs=
					(int*)calloc(sizeof(int),
							tree->lbq.blocks[i].rx_scopes.nscopes);
				if (!tree->lbq.blocks[i].rx_scopes.lbs) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].rx_scopes.lbs,
					sizeof(int),
					tree->lbq.blocks[i].rx_scopes.nscopes,
					sfile);
				if (rv!=tree->lbq.blocks[i].rx_scopes.nscopes)
					SFILE_LTPERR(LTP_FILE);
				tree->lbq.blocks[i].rx_scopes.ubs=
					(int*)calloc(sizeof(int),
							tree->lbq.blocks[i].rx_scopes.nscopes);
				if (!tree->lbq.blocks[i].rx_scopes.ubs) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].rx_scopes.ubs,
					sizeof(int),
					tree->lbq.blocks[i].rx_scopes.nscopes,
					sfile);
				if (rv!=tree->lbq.blocks[i].rx_scopes.nscopes)
					SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].tx_scopes.nscopes) { 
				tree->lbq.blocks[i].tx_scopes.lbs=
					(int*)calloc(sizeof(int),
							tree->lbq.blocks[i].tx_scopes.nscopes);
				if (!tree->lbq.blocks[i].tx_scopes.lbs) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].tx_scopes.lbs,
					sizeof(int),
					tree->lbq.blocks[i].tx_scopes.nscopes,
					sfile);
				if (rv!=tree->lbq.blocks[i].tx_scopes.nscopes)
					SFILE_LTPERR(LTP_FILE);
				tree->lbq.blocks[i].tx_scopes.ubs=
					(int*)calloc(sizeof(int),
							tree->lbq.blocks[i].tx_scopes.nscopes);
				if (!tree->lbq.blocks[i].tx_scopes.ubs) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].tx_scopes.ubs,
					sizeof(int),
					tree->lbq.blocks[i].tx_scopes.nscopes,
					sfile);
				if (rv!=tree->lbq.blocks[i].tx_scopes.nscopes)
					SFILE_LTPERR(LTP_FILE);
			}
			if (tree->lbq.blocks[i].fwd_scopes.nscopes) { 
				tree->lbq.blocks[i].fwd_scopes.lbs=
					(int*)calloc(sizeof(int),
							tree->lbq.blocks[i].fwd_scopes.nscopes);
				if (!tree->lbq.blocks[i].fwd_scopes.lbs) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].fwd_scopes.lbs,
					sizeof(int),
					tree->lbq.blocks[i].fwd_scopes.nscopes,
					sfile);
				if (rv!=tree->lbq.blocks[i].fwd_scopes.nscopes)
					SFILE_LTPERR(LTP_FILE);
				tree->lbq.blocks[i].fwd_scopes.ubs=
					(int*)calloc(sizeof(int),
							tree->lbq.blocks[i].fwd_scopes.nscopes);
				if (!tree->lbq.blocks[i].fwd_scopes.ubs) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->lbq.blocks[i].fwd_scopes.ubs,
					sizeof(int),
					tree->lbq.blocks[i].fwd_scopes.nscopes,
					sfile);
				if (rv!=tree->lbq.blocks[i].fwd_scopes.nscopes)
					SFILE_LTPERR(LTP_FILE);
			}
			rv=read_ltpt_vals(&tree->lbq.blocks[i].t_vals,sfile);
			if (rv) SFILE_LTPERR(LTP_FILE);
		}
		// txq the transmit queue entries
		if (tree->txq.nsegments) {
			tree->txq.segments=(ltp_segment*)calloc(sizeof(ltp_segment),tree->txq.nsegments);
			if (!tree->txq.nsegments) SFILE_LTPERR(LTP_MEMORY);
			tree->txq.alloced=tree->txq.nsegments;
			rv=fread(tree->txq.segments,sizeof(ltp_segment),tree->txq.nsegments,sfile);
			if (rv!=tree->txq.nsegments) SFILE_LTPERR(LTP_FILE);
		} else {
			tree->txq.segments=NULL;
			tree->txq.alloced=0;
		}
		// now read in and decode the segments
		for (int i=0;i!=tree->txq.nsegments;i++) {
			int blen=0;
			rv=fread(&blen,sizeof(int),1,sfile);
			if (rv!=1) LTPERR(rv);
			unsigned char *buf;
			buf=(unsigned char*)malloc(blen);
			if (!buf) SFILE_LTPERR(LTP_MEMORY);
			rv=fread(buf,sizeof(unsigned char),blen,sfile);
			if (rv!=blen) SFILE_LTPERR(rv);
			rv=ltpc_decode(&blen,&buf,&tree->txq.segments[i].seg);
			if (LTPC_ERROR(rv)) SFILE_LTPERR(rv);
			free(buf);
		}
		// replay entries
		if (tree->replays.nentries) {
			tree->replays.details=(ltp_cache_detail*)calloc(sizeof(ltp_cache_detail),tree->replays.nentries);
			if (!tree->replays.details) SFILE_LTPERR(LTP_MEMORY);
			tree->replays.alloced=tree->replays.nentries;
			rv=fread(tree->replays.details,sizeof(ltp_cache_detail),tree->replays.nentries,sfile);
			if (rv!=tree->replays.nentries) SFILE_LTPERR(LTP_FILE);
		} else {
			tree->replays.details=NULL;
			tree->replays.alloced=0;
		}
		for (int i=0;i!=tree->replays.nentries;i++) {
			// read hash val
			if (tree->replays.details[i].hash.len) {
				tree->replays.details[i].hash.val=(unsigned char*)
					malloc(tree->replays.details[i].hash.len);
				if (!tree->replays.details[i].hash.val) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->replays.details[i].hash.val,
					tree->replays.details[i].hash.len,
					sizeof(unsigned char),
					sfile);
				if (rv!=sizeof(unsigned char)) 
					SFILE_LTPERR(rv);
			}
			// responses (if any)
			if (tree->replays.details[i].response.len) {
				tree->replays.details[i].response.val=(unsigned char*)
					malloc(tree->replays.details[i].response.len);
				if (!tree->replays.details[i].response.val) SFILE_LTPERR(LTP_MEMORY);
				rv=fread(tree->replays.details[i].response.val,
					sizeof(unsigned char),
					tree->replays.details[i].response.len,
					sfile);
				if (rv!= tree->replays.details[i].response.len)
					SFILE_LTPERR(rv);
			} else {
				tree->replays.details[i].response.val=NULL;
			}
		}
		// timer entries
		if (tree->timers.nentries) {
			tree->timers.details=(ltp_timer_detail*)calloc(sizeof(ltp_timer_detail),tree->timers.nentries);
			if (!tree->timers.details) SFILE_LTPERR(LTP_MEMORY);
			tree->timers.alloced=tree->timers.nentries;
			rv=fread(tree->timers.details,sizeof(ltp_timer_detail),tree->timers.nentries,sfile);
			if (rv!=tree->timers.nentries) SFILE_LTPERR(LTP_FILE);
		} else {
			tree->timers.details=NULL;
			tree->timers.alloced=0;
		}
		// I need a bunch of fixparent calls, preceeded by setting ->parent ptrs
		// to NULL so's they defo don't point to a copy of the parentid str!!!
		// (such bad copies might exist transiently during a synch cycle)
		// I couldn't do this 'till now when all ltp_instance's are present and
		// correct.
		for (int i=0;i!=tree->nsockets;i++) {
			for (int j=0;j!=tree->sockets[i].nblocks;j++) {
				tree->sockets[i].blocks[j].parent=NULL;
				rv=ltp_fixparent(tree,(void**)&tree->sockets[i].blocks[j].parent,tree->sockets[i].blocks[j].parentid,LTP_PARENT_SOCKET);
				if (rv) SFILE_LTPERR(rv);
			}
		}
		for (int i=0;i!=tree->lbq.nblocks;i++) {
			tree->lbq.blocks[i].parent=NULL;
			rv=ltp_fixparent(tree,(void**)&tree->lbq.blocks[i].parent,tree->lbq.blocks[i].parentid,LTP_PARENT_SOCKET);
			if (rv) SFILE_LTPERR(rv);
		}
		for (int i=0;i!=tree->txq.nsegments;i++) {
			tree->txq.segments[i].parent=NULL;
			rv=ltp_fixparent(tree,(void**)&tree->txq.segments[i].parent,tree->txq.segments[i].parentid,LTP_PARENT_BLOCK);
			if (rv) SFILE_LTPERR(rv);
		}
		for (int i=0;i!=tree->replays.nentries;i++) {
			// fixup the r_entry pointer of the corresponding ltp_segment
			bool foundseg=false;
			for (int j=0;j!=tree->txq.nsegments;j++) {
				if (tree->txq.segments[j].r_id==tree->replays.details[i].id) {
					foundseg=true;
					tree->txq.segments[j].r_entry=i;
				}
			}
			if (!foundseg) {
				snprintf(ltpstr,LTPSTR_LEN,"%s:%d:Can't find block (id=%d) for cache entry %d (segment id=%d)\n",
							__FILE__,__LINE__,tree->replays.details[i].parentid,i,tree->replays.details[i].id);
				ltp_log(LTP_LOG_ERR,ltpstr);
			}
		}
		for (int i=0;i!=tree->timers.nentries;i++) {
			tree->timers.details[i].segment=NULL;
			if (tree->timers.details[i].segmentid) {
				bool foundseg=false;
				rv=ltp_fixparent(tree,(void**)&tree->timers.details[i].segment,tree->timers.details[i].segmentid,LTP_PARENT_SEGMENT);
				if (rv) {
					if (tree->timers.details[i].type==LTP_TIMER_INACTIVITY || 
						tree->timers.details[i].type==LTP_TIMER_COOKIE ) {
						tree->timers.details[i].segmentid=0;
						foundseg=true;
					} else {
						SFILE_LTPERR(rv);
					}
				}
				// fixup the r_entry pointer of the corresponding ltp_segment
				for (int j=0;!foundseg && j!=tree->txq.nsegments;j++) {
					if (tree->txq.segments[j].id==tree->timers.details[i].segmentid) {
						foundseg=true;
						tree->txq.segments[j].t_entry=i;
						tree->txq.segments[j].t_id=tree->timers.details[i].id;
					}
				}
				if (!foundseg) {
					snprintf(ltpstr,LTPSTR_LEN,"%s:%d:Can't find segment (id=%d) for timer %d (id=%d)\n",
								__FILE__,__LINE__,tree->timers.details[i].segmentid,i,tree->timers.details[i].id);
					ltp_log(LTP_LOG_ERR,ltpstr);
				}
			}
			tree->timers.details[i].block=NULL;
			if (tree->timers.details[i].blockid) {
				rv=ltp_fixparent(tree,(void**)&tree->timers.details[i].block,tree->timers.details[i].blockid,LTP_PARENT_SEGMENT);
				if (rv) {
					if (tree->timers.details[i].type==LTP_TIMER_INACTIVITY || 
						tree->timers.details[i].type==LTP_TIMER_COOKIE ) {
						tree->timers.details[i].blockid=0;
					} else {
						SFILE_LTPERR(rv);
					}
				}
			}
			tree->timers.details[i].socket=NULL;
			if (tree->timers.details[i].socketid) {
				rv=ltp_fixparent(tree,(void**)&tree->timers.details[i].socket,tree->timers.details[i].socketid,LTP_PARENT_SOCKET);
				if (rv) 
					SFILE_LTPERR(rv);
			}
		}
	}
	if (sfile) fclose(sfile);
	return(LTP_OK);
}

/*!
 * @brief write the ltp_state_var to disk
 * @return zero or error
 */
int ltp_sync_2disk(void)
{
	int rv;
	if (!ltp_state_var.inited) return(LTP_OK);
	if (ltp_state_changed) { // something to do
		rv=walkstatetree(&ltp_state_var,LTP_ST_WRITING);
		if (rv) LTPERR(rv);
#undef UNITTEST
#ifdef UNITTEST
		// just for an initial test
		ltp_state nstate;
		rv=walkstatetree(&nstate,LTP_ST_READING);
		if (rv) LTPERR(rv);
		rv=walkstatetree(&ltp_state_var,LTP_ST_FREEING);
		if (rv) LTPERR(rv);
		ltp_state_var=nstate; // struct copy!
#endif
		ltp_state_changed=false;
	}
	return(LTP_OK);
}

/*!
 * @brief read the disk to ltp_state_var
 * @return zero or error
 */
int ltp_sync_fromdisk(void)
{
	int rv;
	ltp_state tmp;
	rv=walkstatetree(&tmp,LTP_ST_READING);
	if (rv) LTPERR(rv);
	if (ltp_state_var.inited) {
		rv=walkstatetree(&ltp_state_var,LTP_ST_FREEING);
		if (rv) LTPERR(rv);
	}
	tmp.inited=true;
	ltp_state_var=tmp; // shallow copy - assuming no earlier alloc's are leaked here
	return(LTP_OK);
}

/*!
 * @brief setup disk settings
 * @param dir is the directory to use
 * @return zero or error
 */
int ltp_sync_setup(char *dir)
{
#ifdef SENDT_FILE
	int rv=ltp_filetype_setup(dir,LTP_FILETYPE_LTP,statefile);
	if (rv) LTPERR(rv);
#else
	snprintf(statefile,MAXPATH,LTP_FNAME);
#endif
	filesdone=true;
	return(LTP_OK);
}

/*!
 * @brief force disk settings (mainly for testing)
 * @param file is the exact file to use
 * @return zero or error
 */
int ltp_sync_setup_file(char *file)
{
	if (!file) LTPERR(LTP_BAD_INPUT);
	snprintf(statefile,MAXPATH,"%s",file);
	filesdone=true;
	return(LTP_OK);
}



