/*!
 * @file ltp_l2.cc
 * @brief This is the external interface for our link layer protocol
 * @version $Revision: 1.28 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This is an early cut at my link layer API implementation.
 * The LTP implementation codes down to this, and from here we can
 * go down to a) an OMNET++ simulation, b) LTP/IP on cygwin and eventually
 * c) 802.11 on the SeNDT node. We'll see what happens.
 *
 * So far we're just doing the UDP/IP version. (or should I do a raw
 * socket? Naw.)
 *
 * You can initialise this into Rx,Tx or Both modes.
 *
 * For Tx's there's nothing much to do since we just send the
 * packet as requested as a UDP payload. The recipient had 
 * better be something that the sockets api can handle. If so,
 * (and firewalls permitting) it ought to get to its destination.
 *
 * Rx'ing is more difficult. When in Rx or Both modes, we start
 * a listener thread, which polls for arriving data and then
 * whenever its gotten a big enough chunk or waited a little
 * bit (maybe 1s) since it started to get stuff, then it makes
 * a call to the relevant callback function.
 *
 * Comms errors generally result in a call to the callback 
 * function.
 *
 * Inputs to this are:
 * 	port - to use is hardcoded at 3011 (TrustedWeb:-)
 *	callbacks - provided from caller, but there's a default set in here too
 *	data block size - 4k - is how much data to buffer before hitting callback button
 *	data delivery timeout - 1s - how long to wait after < a block arrived before hitting callback button
 *
 *	This file also contains default callback handlers which can be used by 
 *	higher layers. If they do use these then they will have to poll the
 *	buffers where Rx'd data is stored (and may miss such data!). I'll put
 *	the required extern statements into a new header file ltp_l2_cb.h
 *
 * There's a bunch of error generating code which is conditionally compiled
 * depending on LTP_L2_ERRORS (set that in configure.ac or config.h or,
 * I suppose, in here). The main function is l2e_do_bad_things(), but 
 * doxygen seems to miss out on it somehow.
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>

#include "ltp_base.h"
#include "ltp.h"
#include "ltp_l2.h"
#include "ltp_udp.h"
#include "ltp_cues.h"
#include "ltp_int.h"

#ifdef THREADS
#include "pthread.h"
#endif

/// the "port" to use below
static int rx_port=UDP_DEF_PORT;

// Before we start...some utility stuff

/// If a ltp_l2_id.pname field contains this string, then we'll
/// loop it back in our layer 2 code.
#define LTP_L2_LOOPBACK_PNAME ((const char*)"loopback")
/// This is the Loopback identity - use it as a destination and it'll pop back to your data handler
const ltp_l2_id l2lb={false,true,(char*)LTP_L2_LOOPBACK_PNAME,0,{0,0,0,0}};

/// If a ltp_l2_id.pname field contains this string, then we'll
/// treat it as a wildcard (like 255.255.255/8 etc).
#define LTP_L2_ANY_PNAME ((const char*)"any")
/// This is the Loopback identity - use it as a destination and it'll pop back to your data handler
const ltp_l2_id l2any={false,true,(char*)LTP_L2_ANY_PNAME,0,{0,0,0,0}};

#ifdef LTP_L2_ERRORS
/// a Trick to try get doxygen to include docs about the error generator code
#undef NOT_LTP_L2_ERRORS
#else
#define NOT_LTP_L2_ERRORS
#endif

#ifndef NOT_LTP_L2_ERRORS
// error generators - numbers and such

/// LTP_L2_ERRORS: we toss an N sided coin to decide whether to do each of the "bad things" 
/// - the coin has this many sides
#define L2E_BT_COIN_SIDES 8

/// LTP_L2_ERRORS: Header corruption - size of header
#define L2E_HEADER_LEN	50

/// LTP_L2_ERRORS: Header corruption - max length of corrupt segment
#define L2E_HEADER_CORRUPT_LEN 15

/// LTP_L2_ERRORS: Truncation - maximum percentage
#define L2E_TRUNC_MAX_PERCENT	50

/// LTP_L2_ERRORS: we toss a new N sided coin to decide whether to re-order
/// - the coin has this many sides
#define L2E_REORDER_COIN_SIDES 10

/// LTP_L2_ERRORS: Reordering - max. packets to wait
#define L2E_REORDER_MAX_WAITS	10


/*!
 * @brief check for an environment var that turns off errors
 * @return is whether to do error stuff or not
 *
 * This function checks for an environment variable called "LTP_L2E_OFF"
 * If there is no such variable in the env then layer 2 error generation
 * happens. If there is such a variable, (i.e. getenv() returns non NULL),
 * then no error generation code is called. The getenv() call is only made
 * once at initialisation time, actually, during the first call to 
 * ltp_l2_send()
 */
static bool l2errors_on(void)
{
	static bool initialised=false;
	static bool errors_on=true;
	if (!initialised) {
		errors_on=(getenv("LTP_L2E_OFF")==NULL);
		if (errors_on) {
			ltp_log(LTP_LOG_MAX,"L2E: Layer 2 errors are turned ON (LTP_L2E_OFF env var is not set)\n");
		} else {
			ltp_log(LTP_LOG_MAX,"L2E: Layer 2 errors are turned OFF (LTP_L2E_OFF env var is set)\n");
		}
		initialised=true;
	}
	return(errors_on);
}

/*!
 * @brief get a random number
 * @param n is the size of the uniform space
 * @return is an uniform random integer less than n
 */
static int l2e_ue(int n)
{
	// might delay a coredump this way:-)
	if (!n) return(1);
	// ignoring mod n truncation effects for now
	// some silliness here that hopefully an optimiser will fix 
	// needed it for debugging
	int pre_ret=(int)ltp_rand();
	// double foo = ((double)n*(double)pre_ret)/(RAND_MAX+1);
	int ret=(pre_ret>>16)%n;
#ifdef LTP_VERBOSE_XX
	snprintf(ltpstr,LTPSTR_LEN,"L2E: l2e_ue returning 0x%x (0x%x)\n",ret,pre_ret);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	return ret;
}

/*!
 * @brief toss a coin
 * @return is true 1/n of the time
 */
static bool l2e_toss_bt_coin(void) {
	int yoke;
	bool foo;
	static char *sides=NULL;
	static bool sides_print=false;
	if (!sides_print) {
		sides=getenv("LTP_L2E_COIN_SIDES");
		if (sides) {
			snprintf(ltpstr,LTPSTR_LEN,"LTP_L2E_COIN_SIDES is set to %s\n",sides);
		} else {
			snprintf(ltpstr,LTPSTR_LEN,"LTP_L2E_COIN_SIDES is not set (using %d)\n",L2E_BT_COIN_SIDES);
		}
		ltp_log(LTP_LOG_MAX,ltpstr);
		sides_print=true;
	}
	if (sides) {
		yoke=atoi(sides);
		yoke=l2e_ue(yoke);
	} else {
		yoke=l2e_ue(L2E_BT_COIN_SIDES);
	}
	foo=(yoke==0);
	return (foo);
}

/*!
 * @brief corrupt the early bytes
 * @param buf is the buffer to be corrupted (near the beginning)
 * @param len is the length of the buffer
 * @return zero or error (none possible:-)
 */
static int l2e_corrupt_header(unsigned char *buf, unsigned int len)
{
	unsigned int clen=l2e_ue(L2E_HEADER_CORRUPT_LEN); ///< length of corrupt segment
	unsigned int cpos=l2e_ue(L2E_HEADER_LEN); ///< position of corrupt segment
	if (clen>=len) clen=len; // respect bounds!!
	if (cpos+clen>=len) cpos=len-clen;
	for (int i=0;i!=clen;i++) {
		buf[cpos+i] |= l2e_ue(256) ;
	}
	snprintf(ltpstr,LTPSTR_LEN,"L2E: corrupted header bytes %d to %d\n",cpos,cpos+clen);
	ltp_log(LTP_LOG_CRITICAL,ltpstr);
	return(LTP_OK);
}

/*!
 * @brief truncate a packet
 * @param len is (a ptr to) the packet length (truncated length on output)
 * @return zero or error (not poss:-)
 */
static int l2e_truncate(unsigned int *len)
{
	unsigned int olen=*len;
	unsigned int tmax=(olen*L2E_TRUNC_MAX_PERCENT)/100;
	unsigned int tlen=l2e_ue(tmax);
	*len-=tlen;
	snprintf(ltpstr,LTPSTR_LEN,"L2E: truncated packet - truncated to %d bytes (was: %d, removed: %d, max: %d)\n",*len,olen,tlen,tmax);
	ltp_log(LTP_LOG_CRITICAL,ltpstr);
	return(LTP_OK);
}

/*!
 * @brief delete a packet - set length to zero! (easy)
 * @param len is (a ptr to) the packet length (0 length on output)
 * @return zero or error (not poss:-)
 */
static int l2e_delete_packet(unsigned int *len)
{
	*len=0;
	snprintf(ltpstr,LTPSTR_LEN,"L2E: deleted packet - removed all bytes!\n");
	ltp_log(LTP_LOG_CRITICAL,ltpstr);
	return(LTP_OK);
}

/*!
 * @brief check whether (and for how long) to hang on to this packet
 * @return is how long to hang on to packet (0 if send now)
 */
static int l2e_reorder(void)
{
	bool doit=!(l2e_ue(L2E_REORDER_COIN_SIDES));
	if (doit) {
		int times=l2e_ue(L2E_REORDER_MAX_WAITS);
		while (!times) times=l2e_ue(L2E_REORDER_MAX_WAITS); // don't want a zero wait - it confuses me:-)
		snprintf(ltpstr,LTPSTR_LEN,"L2E: reordering packet - going to wait until another %d packets are sent!\n",times);
		ltp_log(LTP_LOG_CRITICAL,ltpstr);
		return(times);
	}
	return(0);
}

/*!
 * @brief do some bad things to a buffer
 * @param buf is the buffer to be corrupted (near the beginning)
 * @param len is the length of the buffer
 * @param recip is the recipient (in case we delay the packet)
 * @return zero or error
 */
static int l2e_do_bad_things(unsigned char *buf, unsigned int *len, ltp_l2_id *recip)
{
	static unsigned char *r_buf=NULL;
	static unsigned int r_len=0;
	static int r_left=0;
	static ltp_l2_id r_recip;
	bool messed=false;
	// toss a coin for each possible bad thing
	int rv=0;
	if (l2e_toss_bt_coin()) { messed=true;rv=l2e_delete_packet(len);}
	if (rv) LTPERR(rv);
	if (!*len) return(LTP_OK); // no point going further if deleted already!!
#undef ONLY_DELETE_DS
#ifdef ONLY_DELETE_DS
	return(LTP_OK);
#endif
	if (l2e_toss_bt_coin()) { messed=true;rv=l2e_corrupt_header(buf,*len);}
	if (rv) LTPERR(rv);
	if (l2e_toss_bt_coin()) { messed=true;rv=l2e_truncate(len);}
	if (messed) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"L2E: Messed up buffer is:\n");
			ltp_log_buffer(ltpstr,(int)*len,buf);
#endif
	}
	if (rv) LTPERR(rv);
	if (r_buf==NULL) { // we only re-order one thing at a time for now
		r_left=l2e_reorder();
		if (r_left) { // do re-order
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"L2E: Keeping this buffer for later!!! (re-ordering; will wait %d goes)\n",r_left);
			ltp_log_buffer(ltpstr,(int)*len,buf);
#endif
			r_buf=(unsigned char*)malloc(*len);
			if (!r_buf) LTPERR(LTP_MEMORY); 
			memcpy(r_buf,buf,*len);
			r_len=*len;
			rv=ltp_copy_l2_id(recip,&r_recip);
			if (rv) LTPERR(rv);
			*len=0;
		} 
	} else { // see if that buffer's waited long enough
		r_left--;
		if (!r_left) {
			ltp_l2_buffer tosend;
			tosend.val=r_buf;
			tosend.len=r_len;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"L2E: Sending this buffer out of order!!! (re-ordering)\n");
			ltp_log_buffer(ltpstr,r_len,r_buf);
#endif
			rv=ltp_l2_send(&tosend,&r_recip); // note - bad things can happen again!!
			if (rv) LTPERR(rv);
			free(r_buf);
			r_buf=NULL;
			r_len=0;
			rv=ltp_free_l2_id(&r_recip);
		}
	}
	return(LTP_OK);
}

#endif // (NOT_)LTP_L2_ERRORS

/*! 
 * @brief turns on the 802.11 (or whatever!) via ifconfig - assume we're on /dev/eth0
 * @return zero or error
 */
static int l2on_via_ifconfig(void)
{
#ifdef LTP_TARGET_ARM_LINUX
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Turning on the radio (via ifconfig!)\n");
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	int rv=system("ifconfig eth0 up");
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"system(\"ifconfig eth0 up\") returned: %d\n",rv);
	ltp_log(LTP_LOG_ERR,ltpstr);
#endif
	return(LTP_OK);
#else
	return(LTP_OK);
#endif
}

/*! 
 * @brief turns off the 802.11 (or whatever!) via ifconfig - assume we're on /dev/eth0
 * @return zero or error
 */
static int l2off_via_ifconfig(void)
{
#ifdef LTP_TARGET_ARM_LINUX
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Turning off the radio (via ifconfig!)\n");
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	int rv=system("ifconfig eth0 down");
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"system(\"ifconfig eth0 down\") returned: %d\n",rv);
	ltp_log(LTP_LOG_ERR,ltpstr);
#endif
	return(LTP_OK);
#else
	return(LTP_OK);
#endif
}

/*!
 * @fn int ltp_copy_l2_id(ltp_l2_id *from, ltp_l2_id *to) 
 * @brief deep copy a ltp_l2_id
 * @param from,to guess?
 * @return zero or error
 */
int ltp_copy_l2_id(ltp_l2_id *from, ltp_l2_id *to) 
{
	if (!to || !from) LTPERR(LTP_BAD_PARAM);
	if (from->pname) {
		to->pname=strdup(from->pname);
		if (!to->pname) LTPERR(LTP_MEMORY);
	} else {
		to->pname=NULL;
	}
	to->addr=from->addr;
	to->addr_len=from->addr_len;
	to->heap=true;
	to->l2alloc=true;
	return(LTP_OK);
}

/*!
 * @fn int ltp_free_l2_id(ltp_l2_id *it)
 * @brief yada yada, doesn't free top level thing
 * @return zero or error
 */
int ltp_free_l2_id(ltp_l2_id *it) 
{
	if (!it) return(LTP_OK);
	if (!it->heap) return(LTP_OK);
	if (!it->l2alloc) return(LTP_OK);
	if (it->pname) free(it->pname);
	it->addr_len=0;
	memset(it,0,sizeof(*it));
	return (LTP_OK);
}

// First we deal with default callbacks

/// This is the default Rx queue
static ltp_l2_cb_queue	rxqueue;
ltp_l2_cb_queue	*rxqueue_p=NULL;

/// And the default Tx queue
static ltp_l2_cb_queue txqueue;
ltp_l2_cb_queue	*txqueue_p=NULL;

/// This is our best guess at a default MTU, in bytes.
/// The 1472 figure is just a number off the net, 
#define DEFAULT_MTU 1472

/// This is the reported MTU from the next layer down
static int repMTU=DEFAULT_MTU;

/*!
 * @fn int cb_def_err(int errno, char *str)
 * @brief default error handling callback
 * @param errno is the error that happened
 * @param str is the (static) error string
 * @return zero or error
 *
 * For now, we just log the error and turn off the layer2,
 * which may well be turned back on in a bit according to
 * schedule. In future we may be more sensitive to the 
 * type of error that's happened.
 *
 */
int cb_def_err(int errno, char *str) {
	snprintf(ltpstr,LTPSTR_LEN,"Layer2 error(%d) - %s\n",errno,str);
	ltp_log(LTP_LOG_ERR,ltpstr);
	int rv=ltp_l2_off(LTP_L2_OFF_BOTH);
	if (rv) LTPERR(LTP_FILE);
	return(LTP_OK);
}

/*!
 * @fn int cb_def_data(ltp_l2_buffer *buf, ltp_l2_id *peer)
 * @brief Rx data default callback (sticks rx'd data in a circular queue)
 * @param buf is the data just been rx'd.
 * @param peer is the sender's information
 * @return zero or error as usual
 */
int cb_def_data(ltp_l2_buffer *buf, ltp_l2_id *peer) 
{
	// Plan:
	// zap anything in the current position and replace it with a
	// copy of what's in this buffer.
	// and then note the time.
	int pos=rxqueue.position;
	int newpos=(pos+1)%rxqueue.size;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"cb_def_data: buf len(%d); pos is %d (-> %d); last mod %s\n",
		buf->len,
		pos,
		newpos,
		(rxqueue.times[pos]?ltp_time_str(rxqueue.times[pos]):"never"));
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
	if (rxqueue.bufs[pos].dirty==true) {
		// log that we're missing this one
		snprintf(ltpstr,LTPSTR_LEN,"cb_def_data: L2 buffer not decoded (dirty) at pos %d\n",pos);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
#endif
	GRABQUTEX;
	if (rxqueue.times[pos]) {
		// free up the val in there for a copy - but only if we alloc'd it!
		if (rxqueue.bufs[pos].heap && rxqueue.bufs[pos].l2alloc) {
			free(rxqueue.bufs[pos].val);
			rxqueue.bufs[pos].val=NULL;
		}
		// free peer as needed
		ltp_free_l2_id(&rxqueue.peers[pos]);
	}
	// now copy buf struct
	rxqueue.bufs[pos].dirty=true;
	rxqueue.bufs[pos].heap=true;
	rxqueue.bufs[pos].l2alloc=true;
	if (rxqueue.bufs[pos].val) {
		free(rxqueue.bufs[pos].val);
		rxqueue.bufs[pos].val=NULL;
	}
	rxqueue.bufs[pos].val=(unsigned char*)calloc(buf->len,sizeof(unsigned char));
	if (!rxqueue.bufs[pos].val) LTPERR(LTP_MEMORY);
	memcpy(rxqueue.bufs[pos].val,buf->val,buf->len);
	rxqueue.bufs[pos].len=buf->len;
	rxqueue.times[pos]=ltp_now();
	int rv=ltp_copy_l2_id(peer, &rxqueue.peers[pos]);
	if (rv) { DROPQUTEX; LTPERR(rv); }
	rxqueue.peers[pos].addr.sock.sin_port=ntohs(rxqueue.peers[pos].addr.sock.sin_port);
	rxqueue.position=newpos;
	DROPQUTEX;
	return(LTP_OK);
}

/*!
 * @fn int cb_def_cts(void)
 * @brief Data for Tx can now be sent (and if there's stuff in the queue then one will be sent)
 * @return zero or error
 */
int cb_def_cts(void) 
{
	// if there's something to send then do that!
	int pos=txqueue.position;
	int newpos=(pos+1)%txqueue.size;
	if (!txqueue.times[pos]) return(LTP_OK); // nothting to do
	// something to send
	int rv=ltp_l2_send(&txqueue.bufs[pos],&txqueue.peers[pos]);
	if (rv) LTPERR(rv); // leaving that to be re-sent infinitely often! (should be timed/counted!)
	/// @todo add a timer/failed-tx-counter to cb_def_cts when transmit fails so that we don't re-tx forever
	// now clean up and go home
	txqueue.times[pos]=0;
	if (txqueue.bufs[pos].heap && txqueue.bufs[pos].l2alloc) 
			free(txqueue.bufs[pos].val);
	memset(&txqueue.bufs[pos],0,sizeof(ltp_l2_buffer));
	// free peer as needed
	rv=ltp_free_l2_id(&txqueue.peers[pos]);
	if (rv) LTPERR(rv);
	txqueue.position=newpos;
	return(LTP_OK);
}

/// This is the default callback handler structure.
static ltp_l2_callbacks	defcbs = { cb_def_cts,cb_def_data,cb_def_err };

/// And a pointer to that
ltp_l2_callbacks	*defcbs_p = &defcbs;

/// Here's where we keep the callback pointers locally.
static ltp_l2_callbacks thecbs; 

/// Here's our current state variable. Right now, we don't enforce the
/// state machine, that is, we just set the state var to the right thing
/// on entry to the various functions, but we don't check that the 
/// previous state was correct. Might do that in future if necessary.
static int curr_state=LTP_L2S_RESET;

#ifdef THREADS
/// Here's the listener thread pointer
static pthread_t	listener=(pthread_t)0;

/// Here's the port the listener's on
static int 	listener_port=UDP_DEF_PORT;

/// A mutex to use in the listener thread so other code doesn't have to 
/// be re-entrant.
pthread_mutex_t l2_listener_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
/// And a special one to prevent mucking with tz & rx queues
pthread_mutex_t qutex = PTHREAD_MUTEX_INITIALIZER;
/// Take the l2 listener mutex - scheme is that the listener thread
/// waits to grab this when it need to do stuff (as opposed to hanging
/// around in a busy-wait). 
/// The implementation of most of the other l2 api functions also wait 
/// to grab the mutex on entry (and so could block if the listener thread's
/// v. busy). That should be ok for now. 

#undef MUTEXDEBUG
#ifdef MUTEXDEBUG
#define GRABMUTEX { \
	fprintf(stderr,"Grabbing listener (in %s @ %d, tid=%p)...",__FUNCTION__,__LINE__,pthread_self()); \
	pthread_mutex_lock(&l2_listener_mutex); \
	fprintf(stderr,"..got listener\n"); }
/// Give back the l2 listener mutex
#define DROPMUTEX { \
	fprintf(stderr,"Dropping listener (in %s @ %d, tid=%p)...",__FUNCTION__,__LINE__,pthread_self()); \
	pthread_mutex_unlock(&l2_listener_mutex); \
	fprintf(stderr,"...dropped listener\n"); }
#else
#define GRABMUTEX pthread_mutex_lock(&l2_listener_mutex);
#define DROPMUTEX pthread_mutex_unlock(&l2_listener_mutex);
#endif
/// Check if I'd have gotten the mutex (not sure what code wants this)
#define MUTEXAVAILABLE !pthread_mutex_lock(&l2_listener_mutex)
#else
#define GRABMUTEX
#define DROPMUTEX
#define MUTEXAVAILABLE
#endif

/// this is the buffer structure we use for arriving packets (byte
/// array therin is malloc'd when listener thread created).
static ltp_l2_buffer	thebuf;


/*!
 * @fn listener_fnc(void *foo)
 * @brief our listener thread mainline
 */
void *listener_fnc(void* foo)
{
	const char *somethingsarrivedstr="Some data's arrived";
	const char *loopbacksomethingsarrivedstr="Some data's arrived via loopback";

	/// we use this guy to determine if some loopback activity has happened
	/// with the default receive callback handler (figure out loopback yourself
	/// if you write your own handler)
	int startpos=rxqueue.position;

	// start rng stream for this thread
	int rv=ltp_rand_init(0xff0000);
	if (rv) pthread_exit(NULL);

	/// this is the structure for the peer (sender) of arrived bytes
	/// (internal arrays are malloc'd on the fly)
	ltp_l2_id		thepeer;

#ifdef LTP_EXTRA_EXTRA_VERBOSE
	int loopcount=0;
	bool nonloopback=true; // just to stop confusing logging
#endif

	bool nothingthere=true;
	while (nothingthere) {
		// BUG: TODO: FIXME:
		// something very fishy here. When this next bit is compiled in then I get
		// a livelock on suse but not cygwin
#ifdef LTP_EXTRA_EXTRA_VERBOSE_BOLLIX
			GRABMUTEX
			snprintf(ltpstr,LTPSTR_LEN,"In listener thread loop (%d)\n",loopcount++);
			ltp_log(LTP_LOG_MAX,ltpstr);
			ltp_log_flush();
			DROPMUTEX
#endif
		// this branch is only for loopback mode
		if (rxqueue.position!=startpos) {
			GRABMUTEX
			// something's arrived!!
#ifdef LTP_EXTRA_EXTRA_VERBOSE
			if (!nonloopback) ltp_log(LTP_LOG_ADVISORY,loopbacksomethingsarrivedstr);
			nonloopback=false;
#endif
			// nothing else to do!!
			DROPMUTEX
#ifndef THREADS
			nothingthere=false;
#endif
		}
		// now we'll check for UDP packet arrival
		int arr=l2_s_udp_arrival();
		if (arr != -1 && arr) {
			GRABMUTEX
#ifdef LTP_EXTRA_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"%d return from from l2_s_udp_arrival\n",arr);
			ltp_log(LTP_LOG_ADVISORY,ltpstr);
			ltp_log(LTP_LOG_ADVISORY,somethingsarrivedstr);
			ltp_log_flush();
			nonloopback=true;
#endif
			l2_s_udp_get_arrived_buf(&thebuf,&thepeer);
			thecbs.data(&thebuf,&thepeer);
#ifndef THREADS
			nothingthere=false;
#endif
			DROPMUTEX
		}
#ifdef LTP_VERBOSE
		if (arr==0) {
			GRABMUTEX
			snprintf(ltpstr,LTPSTR_LEN,"Zero return from from l2_s_udp_arrival\n");
			ltp_log(LTP_LOG_ADVISORY,ltpstr);
			DROPMUTEX
		}
#endif
#ifdef LTP_VERBOSE
		if (arr==-1) {
			GRABMUTEX
			snprintf(ltpstr,LTPSTR_LEN,"Error (%d) from l2_s_udp_arrival (%s)\n",errno, strerror(errno));
			ltp_log(LTP_LOG_ERR,ltpstr);
			int rxfd,txfd;
			int rrv=l2_udp_status(&rxfd,&txfd);
			if (!rrv) {
				snprintf(ltpstr,LTPSTR_LEN,"\tRx socket handle: %d, Tx socket handle: %d\n",rxfd,txfd);
				ltp_log(LTP_LOG_ERR,ltpstr);
			}
			DROPMUTEX
		}
#endif

#undef WAITINLOOP
#ifdef WAITINLOOP
		// now we wait for a short while
		struct timeval ltv;
		ltv.tv_sec=0;
		ltv.tv_usec=100000; // 1/10th sec
		select(0,NULL,NULL,NULL,&ltv);
#endif
#ifdef L2WATCHDOG
		// strobe watchdog
		rv=ltp_watchdog_strobe();
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"Error from ltp_watchdog_strobe: %d\n",rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			// continue on regardless! (or try to...)
		}
#endif
		
	}

	return NULL;
}

// a message
static const char		*listener_message="listener thread";

/*! 
 * @fn int ltp_l2_stop_listening(void)
 * @brief turn off a listener (if there was one)
 * @return zero or error
 *
 * This has to be done so that multiple on/off cycles are possible within 
 * a single init/shutdown cycle.
 */
static int ltp_l2_stop_listening(void)
{
	GRABMUTEX;
#ifdef THREADS
	if (!listener) {
		snprintf(ltpstr,LTPSTR_LEN,"Can't turn off - no listener!\n");
		ltp_log(LTP_LOG_ADVISORY,ltpstr);
	} else {
		int rv=pthread_cancel(listener);
		if (!rv) {
			snprintf(ltpstr,LTPSTR_LEN,"Cancelled listener thread (%p)\n",(void*)listener);
			ltp_log(LTP_LOG_ADVISORY,ltpstr);
		} else {
			ltp_log(LTP_LOG_ERR,(char*)"Failed to cancel listener thread\n");
		}
		listener=(pthread_t)0;
	}
#endif
	l2_s_udp_close();
	DROPMUTEX;
	return(LTP_OK);
}

/*! 
 * @fn ltp_l2_spawn_listener(void)
 * @brief spawn off a listener thread for rx
 * @return zero or error
 */
static int ltp_l2_spawn_listener(void)
{
	int rv;
#ifdef THREADS
	if (listener && listener_port==rx_port) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Won't start listener thread again - tid(%p) already listening! (I assume its ok)\n",(void*)listener);
		ltp_log(LTP_LOG_ADVISORY,ltpstr);
#endif
		return(LTP_OK);
	}
	if (listener_port!=rx_port) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Re-starting listener thread -  old was tid(%p)\n",(void*)listener);
		ltp_log(LTP_LOG_ADVISORY,ltpstr);
#endif
		// kill that one and start anew
		rv=ltp_l2_stop_listening();
		if (rv) LTPERR(rv);
	}
	snprintf(ltpstr,LTPSTR_LEN,"About to start listener thread on port %d\n",rx_port);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
	ltp_log_flush();

    sockaddr_in *mcp=NULL;
#ifdef LTP_MULTICAST
    ltpaddr foo;
    int wrv=ltp_cue_whoami(&foo);
    if (!ltp_is_multi(&foo)) {
#ifdef LTP_MC_DEBUG
		snprintf(ltpstr,LTPSTR_LEN,"MC: ltp_is_multi says yes %s is MC",ltpaddr2str(&foo));
	    ltp_log(LTP_LOG_MAX,ltpstr);
#endif
        mcp=&foo.sock;
    }
#endif

	// fprintf(stderr,"Grabbing listener (in %s)...",__FUNCTION__); 
	// pthread_mutex_lock(&l2_listener_mutex); 
	// fprintf(stderr,"..got it\n"); 
	listener_port=rx_port;
	rv=l2_s_udp_init(rx_port,mcp);
	if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"Failed to init UDP listener on %d (rv=%d)\n",rx_port,rv);
		ltp_log(LTP_LOG_ERR,ltpstr);
		// maybe this is ok - if I'm about to open a new listener on a different port, try that
		// LTPERR(rv);
	}
#ifdef LTP_EXTRA_EXTRA_VERBOSE
	int rxfd,txfd;
	int rrv=l2_udp_status(&rxfd,&txfd);
	if (!rrv) {
		snprintf(ltpstr,LTPSTR_LEN,"\tRx socket handle: %d, Tx socket handle: %d\n",rxfd,txfd);
		ltp_log(LTP_LOG_ERR,ltpstr);
	}
#endif
	pthread_t tmpThread;
	rv=pthread_create(&tmpThread,NULL,listener_fnc,(void*)listener_message);
	if (rv) LTPERR(rv);
	listener=tmpThread;
#ifdef LTP_TARGET_ARM_LINUX
	sleep(1);
	snprintf(ltpstr,LTPSTR_LEN,"Started listener thread - tid(%p) then slept a second\n",(void*)listener);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
	ltp_log_flush();
#else
	snprintf(ltpstr,LTPSTR_LEN,"Started listener thread - tid(%p)\n",(void*)listener);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
#endif
	// now we wait for a short while
	struct timeval ltv;
	ltv.tv_sec=0;
	ltv.tv_usec=100000; // 1/10th sec
	select(0,NULL,NULL,NULL,&ltv);
#else
	// fprintf(stderr,"Grabbing listener (in %s)...",__FUNCTION__); 
	// pthread_mutex_lock(&l2_listener_mutex); 
	// fprintf(stderr,"..got it\n"); 
	listener_port=rx_port;
	int rv=l2_s_udp_init(rx_port,mcp);
	if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"Failed to init UDP listener on %d (rv=%d)\n",rx_port,rv);
		ltp_log(LTP_LOG_ERR,ltpstr);
		DROPMUTEX
		return NULL;
	}
#ifdef LTP_EXTRA_EXTRA_VERBOSE
	int rxfd,txfd;
	int rrv=l2_udp_status(&rxfd,&txfd);
	if (!rrv) {
		snprintf(ltpstr,LTPSTR_LEN,"\tRx socket handle: %d, Tx socket handle: %d\n",rxfd,txfd);
		ltp_log(LTP_LOG_ERR,ltpstr);
	}
#endif
	listener_fnc(NULL);
#endif
	return(LTP_OK);
}


/*!
 * @fn ltp_l2_init(ltp_l2_callbacks *cbs, int port)
 * @param cbs is a pointer to a callback structure (has ptrs to c-b functions)
 * @param tport is the "port" to send from (actually in the case of e.g. UDP/IP, virtually otherwise)
 * @param rport is the "port" to listen with (actually in the case of e.g. UDP/IP, virtually otherwise)
 * @brief boot-time initialise the layer 2 stuff 
 * @return zero or error
 *
 * This function initialises a l2 implementation so that it'll be ready to 
 * be turned on using ltp_l2_on()
 *
 * We also allocate space for the default callback handlers queues.
 *
 * Call this once per reboot
 */
int ltp_l2_init(ltp_l2_callbacks *cbs, int rport) 
{
	static bool beenhere=false;
	if (!cbs) LTPERR(LTP_BAD_PARAM);
	GRABMUTEX;
	thecbs=*cbs;
	if (!beenhere) { // allocate space
		GRABQUTEX;
		rxqueue.bufs=(ltp_l2_buffer*)calloc(LTP_L2_CB_DEFAULT_QUEUE_SIZE,sizeof(ltp_l2_buffer));
		if (!rxqueue.bufs) { DROPQUTEX; DROPMUTEX; LTPERR(LTP_MEMORY); }
		rxqueue.peers=(ltp_l2_id*)calloc(LTP_L2_CB_DEFAULT_QUEUE_SIZE,sizeof(ltp_l2_id));
		if (!rxqueue.peers) { DROPQUTEX; DROPMUTEX; LTPERR(LTP_MEMORY); }
		rxqueue.times=(ltp_time_t*)calloc(LTP_L2_CB_DEFAULT_QUEUE_SIZE,sizeof(ltp_time_t));
		if (!rxqueue.times) { DROPQUTEX; DROPMUTEX; LTPERR(LTP_MEMORY); }
		rxqueue.size=LTP_L2_CB_DEFAULT_QUEUE_SIZE;
		rxqueue.position=0;
		rxqueue_p=&rxqueue;
		txqueue.bufs=(ltp_l2_buffer*)calloc(LTP_L2_CB_DEFAULT_QUEUE_SIZE,sizeof(ltp_l2_buffer));
		if (!txqueue.bufs) { DROPQUTEX; DROPMUTEX; LTPERR(LTP_MEMORY); }
		txqueue.peers=(ltp_l2_id*)calloc(LTP_L2_CB_DEFAULT_QUEUE_SIZE,sizeof(ltp_l2_id));
		if (!txqueue.peers) { DROPQUTEX; DROPMUTEX; LTPERR(LTP_MEMORY); }
		txqueue.times=(ltp_time_t*)calloc(LTP_L2_CB_DEFAULT_QUEUE_SIZE,sizeof(ltp_time_t));
		txqueue.size=LTP_L2_CB_DEFAULT_QUEUE_SIZE;
		txqueue.position=0;
		txqueue_p=&txqueue;
		if (!txqueue.times) { DROPQUTEX; DROPMUTEX; LTPERR(LTP_MEMORY); }
		beenhere=true;
		DROPQUTEX;
	} else {
		// maybe want to turn off l2 and then back on again
	}
	curr_state=LTP_L2S_INITED;
	if (rport!=UDP_DEF_PORT) rx_port=rport; 
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @fn ltp_l2_on(unsigned int *mtu, int state2go2)
 * @brief runtime initialise the layer 2 stuff and return the MTU for this link
 * @param mtu is the (caller alloc'd) place where the MTU will be returned
 * @param state2go2 is the state we want to go to, which can be LTP_L2S_ON[_TX,_RX, or _BOTH]
 * @return zero or error
 *
 * This function "turns on" a layer2 implementation, so that its ready to
 * bind/accept-connections/whatever according to the desired state (state2go2)
 *
 * MTU means the max-transmission-unit and is interpreted here as the biggest
 * payload that can be sent or rx'd by this l2.
 *
 * The state to go to input tells us whether we want to be in receive only,
 * transmit only or full duplex operation.
 *
 * If we are to receive then we will (in cygwin at least) start a thread 
 * (until we're turned off anyway) which will listen for UDP inputs on
 * our port and will then make a call to the relevant callback.
 *
 * This can be called when we're already on (e.g. to switch from
 * "Tx-only" to "both". 
 * The allowed state transitions here are therefore:
 * INIT->[Rx,Tx,Both] ; 
 * Rx->Both ; 
 * Tx->Both ; 
 * Both->[Tx,Rx] ; 
 * *->OFF/SLEEP are handled elsewhere.
 */
int ltp_l2_on(unsigned int *mtu, int state2go2)
{
	GRABMUTEX;
	int rv;
	// there's some kind of race condition in loopback modes where
	// we end up here but the current and state2go2 are the same
	// that caused an error so I won't bother:-)
	if (state2go2!=curr_state)  // then do the switch 
	switch(curr_state) {
		case LTP_L2S_RESET:
				DROPMUTEX; LTPERR(LTP_BAD_STATE);
		case LTP_L2S_INITED:
			switch (state2go2) {
					case LTP_L2S_ON_TX:
						rv=l2on_via_ifconfig();
						if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
						rv=l2_c_udp_init();
						if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Will transmit from port %d\n",l2_c_sending_port());
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						curr_state=state2go2;
						break;
					case LTP_L2S_ON_RX:
						rv=l2on_via_ifconfig();
						if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
						rv=ltp_l2_spawn_listener();
						if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
						curr_state=state2go2;
						break;
					case LTP_L2S_ON_BOTH:
						rv=l2on_via_ifconfig();
						if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
						rv=l2_c_udp_init();
						if (rv) { 
							snprintf(ltpstr,LTPSTR_LEN,"Error (%d) from l2_c_udp_init (%s)\n",errno,strerror(errno));
							ltp_log(LTP_LOG_ERR,ltpstr);
							DROPMUTEX; 
							LTPERR(rv); } // remaining in previous state!
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Will transmit from port %d\n",l2_c_sending_port());
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						rv=ltp_l2_spawn_listener();
						if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
						curr_state=state2go2;
						break;
					default:
						{ DROPMUTEX; LTPERR(LTP_BAD_PARAM); }
			}
			break;
		case LTP_L2S_ON_TX:
			if (state2go2!=LTP_L2S_ON_BOTH && state2go2!=LTP_L2S_ON_RX) { DROPMUTEX; LTPERR(LTP_BAD_PARAM); }
			rv=l2on_via_ifconfig();
			if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
			// turn on listener 
			rv=ltp_l2_spawn_listener();
			if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
			curr_state=LTP_L2S_ON_BOTH;
			break;
		case LTP_L2S_ON_RX:
			if (state2go2!=LTP_L2S_ON_BOTH && state2go2!=LTP_L2S_ON_TX) { DROPMUTEX; LTPERR(LTP_BAD_PARAM); }
			rv=l2on_via_ifconfig();
			if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
			// init c_udp
			rv=l2_c_udp_init();
			if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Will transmit from port %d\n",l2_c_sending_port());
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			curr_state=LTP_L2S_ON_BOTH;
			break;
		case LTP_L2S_ON_BOTH:
			if (state2go2!=LTP_L2S_ON_RX && state2go2!=LTP_L2S_ON_TX) { DROPMUTEX; LTPERR(LTP_BAD_PARAM); }
			if (state2go2==LTP_L2S_ON_TX) {
				// turn off listener & close s_udp
				rv=ltp_l2_stop_listening();
				if (rv) { DROPMUTEX; LTPERR(rv); } // remaining in previous state!
				curr_state=state2go2;
			}
			if (state2go2==LTP_L2S_ON_RX) {
				// close c_udp
				l2_c_udp_close();
				curr_state=state2go2;
			}
			break;
		default:
				{ DROPMUTEX; LTPERR(LTP_BAD_STATE); }
	}
	int tmp=l2_get_mtu(NULL);
	if (tmp) { 
		repMTU=tmp;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"MTU from l2_get_mtu is %d\n",repMTU);
		ltp_log(LTP_LOG_ADV,ltpstr);
#endif
	} else {
		snprintf(ltpstr,LTPSTR_LEN,"Got no MTU from l2_get_mtu, using default\n",repMTU);
		ltp_log(LTP_LOG_ADV,ltpstr);
	}
	*mtu=repMTU;
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @fn ltp_l2_off(int what)
 * @param what one of the LTP_L2_OFF_* values
 * @brief turn off the layer 2 stuff (can be turned back on with ltp_l2_on later on)
 * @return zero or error
 */
int ltp_l2_off(int what)
{
	GRABMUTEX;
	// we're ignoring returns from these since a) they always return LTP_OK in any
	// case, and b) I don't quite know how to handle such errors - better to assume
	// that closing stuff is easy (including closing my eyes:-)
	switch (what) {

		case LTP_L2_OFF_BOTH:
			ltp_l2_stop_listening();
			l2off_via_ifconfig();
			l2_c_udp_close();
			curr_state=LTP_L2S_INITED;
			break;

		case LTP_L2_OFF_RX:
			ltp_l2_stop_listening();
			if (curr_state==LTP_L2S_ON_RX) {
				l2off_via_ifconfig();
				curr_state=LTP_L2S_INITED;
			} else 
				curr_state=LTP_L2S_ON_TX;
			break;

		case LTP_L2_OFF_TX:
			l2_c_udp_close();
			if (curr_state==LTP_L2S_ON_TX) {
				l2off_via_ifconfig();
				curr_state=LTP_L2S_INITED;
			} else 
				curr_state=LTP_L2S_ON_RX;
			break;

		default:
			curr_state=LTP_L2S_INITED;
			DROPMUTEX;
			return(LTP_BAD_PARAM);
	} 
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @fn void free_queue(ltp_l2_cb_queue *foo)
 * @brief free up what we can from queues
 *
 * Some of the arrays pointed to may have to be free'd elsewhere
 * (so hopefully there're still pointer to 'em!).
 */
void free_queue(ltp_l2_cb_queue *foo)
{
	for (int i=0;i!=foo->size;i++) {
		if (foo->bufs[i].heap && foo->bufs[i].l2alloc)
				free(foo->bufs[i].val);
		ltp_free_l2_id(&foo->peers[i]);
	}
	free(foo->bufs);
	free(foo->times);
	free(foo->peers);
	return;
}

/*!
 * @fn int ltp_l2_shutdown(void)
 * @brief shutdown the layer 2 stuff, before we go to sleep or shutdown CPU (CPU reset may be needed prior to next ltp_l2_init()
 * @return zero or error
 */
int ltp_l2_shutdown(void)
{
	GRABMUTEX;
	// clean up where we can
	if (thebuf.val && thebuf.heap && thebuf.l2alloc) free(thebuf.val);
	// free default queues
	GRABQUTEX;
	free_queue(&rxqueue);
	free_queue(&txqueue);
	DROPQUTEX;
	
	curr_state=LTP_L2S_RESET;
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @fn int ltp_l2_get_state(void)
 * @brief get the state of l2
 * @return is the state of l2 (one of LTP_L2S_* above)
 */
int ltp_l2_get_state(void)
{
	// GRABMUTEX;
	// DROPMUTEX;
	return(curr_state);
}

/*!
 * @fn ltp_l2_sleep(int mode)
 * @brief set the sleep mode using one of the defined LTP_L2S_SLEEP* modes
 * @param mode is the ltp sleep state we want to get to (e.g. LTP_L2S_SLEEP1)
 * @return zero or error (e.g. if mode is not a sleep state)
 */
int ltp_l2_sleep(int mode)
{
	GRABMUTEX;
	switch (mode) {
			case LTP_L2S_SLEEP1:
			case LTP_L2S_SLEEP2:
			case LTP_L2S_SLEEP3:
			case LTP_L2S_SLEEP4:
			case LTP_L2S_SLEEP5:
					curr_state=mode;
					break;
			default:
				DROPMUTEX;
				LTPERR(LTP_BAD_PARAM);
	}
	DROPMUTEX;
	return(LTP_OK);
}


/*!
 * @fn ltp_l2_send( ltp_l2_buffer	*buf, ltp_l2_id		*recip)
 * @brief send a block to an l2 peer
 * @param buf is the buffer to sent
 * @param recip is the peer to whom it should be sent
 * @return zero or error
 */
int ltp_l2_send(
		ltp_l2_buffer	*buf,
		ltp_l2_id		*recip)
{
		// sanity check
		if (!buf || !recip) LTPERR(LTP_BAD_PARAM);
		// state ok?
		if (curr_state != LTP_L2S_ON_TX && 
			curr_state != LTP_L2S_ON_BOTH ) {
			snprintf(ltpstr,LTPSTR_LEN,"In wrong state (%d) to send!\n",curr_state);
			ltp_log(LTP_LOG_ERR,ltpstr);
			return(LTP_ERR); ///< @todo could have a better error code here
		}
		// Check if its a loopback (at this layer) and if not, 
		// check addressing and send it on if I can (i.e. if its
		// got a resolvable IP address in there).
		GRABMUTEX;
		bool loopback=false;
		int rv;
		if (recip==&l2lb) { // loopback via well known id
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Sending %d bytes via my l2 loopback (well-known id case)\n",buf->len);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			loopback=true;
			rv=thecbs.data(buf,recip);
			if (rv) {
				DROPMUTEX; LTPERR(rv);
			}
		}
		if (!loopback && recip->pname) {
			int l1=strlen(LTP_L2_LOOPBACK_PNAME);
			int l2=strlen(recip->pname);
			if (l1 && l2 && l1==l2 && !strncmp(recip->pname,LTP_L2_LOOPBACK_PNAME,l1)) {
				// here's a cheat of sorts - we use the callback fnc to 
				// put the buffer into the rxqueue
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Sending %d bytes via my l2 loopback (pname case)\n",buf->len);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				loopback=true;
				rv=thecbs.data(buf,recip);
				if (rv) {
					DROPMUTEX; LTPERR(rv);
				}
			}
		} 
		// ok, not a loopback so UDP then
		if (!loopback) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Sending %d bytes via l2 to %s:%d\n",
							buf->len,inet_ntoa(recip->addr.sock.sin_addr),recip->addr.sock.sin_port);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
#ifndef NOT_LTP_L2_ERRORS
		if (l2errors_on()) {
			ltp_l2_buffer bufcp;
			bufcp.len=buf->len;
			bufcp.val=(unsigned char*)malloc(bufcp.len);
			if (!bufcp.val) LTPERR(LTP_MEMORY);
			memcpy(bufcp.val,buf->val,buf->len);
			rv=l2e_do_bad_things(bufcp.val,&bufcp.len,recip);
			if (rv) { DROPMUTEX; LTPERR(rv); }
		 	if (bufcp.len) {
				rv=l2_c_udp_send(&bufcp,recip);
			} else {
				ltp_log(LTP_LOG_MAX,"L2E: Didn't send 'em cause there were none!");
			}
			free(bufcp.val);
		} else {
			rv=l2_c_udp_send(buf,recip);
		}
#else
			rv=l2_c_udp_send(buf,recip);
#endif
#ifdef LTP_VERBOSE
			if (rv==-1) {
				snprintf(ltpstr,LTPSTR_LEN,"Error (%d) from l2_c_udp_send (%s)\n",errno,strerror(errno));
				ltp_log(LTP_LOG_ERR,ltpstr);
			}
#endif
			if (rv) {
				DROPMUTEX; LTPERR(rv);
			}
		}
		DROPMUTEX;
		return(LTP_OK);
}

