/*!
 * @file ltp_log.cc
 * @brief This is the API for logging/error handling
 * @version $Revision: 1.25 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/utsname.h> // for uname
#include <sys/unistd.h> // for sleep
#include <sys/time.h>  // for gettimeofday
#include <errno.h> // for errno
#include <ctype.h> // for toupper()

#include <fcntl.h> // for read

#include <syslog.h>

#include <time.h> // for strptime and others

#include "ltp_base.h"
#include "ltp.h"

#ifdef LTP_TARGET_LINUX
#include <pwd.h>
#endif

#ifdef THREADS
#include <pthread.h>
#endif

#ifdef THREADS
static pthread_mutex_t time_str_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
static pthread_mutex_t logf_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#endif
#ifdef THREADS
#undef MUTEXDEBUG
#ifdef MUTEXDEBUG
#define GRABMUTEX { \
	fprintf(stderr,"Grabbing logf_mutex (@%s:%d, tid=%p)...",__FUNCTION__,__LINE__,pthread_self()); \
	pthread_mutex_lock(&logf_mutex); \
	fprintf(stderr,"...%p:%s:%d got logf_mutex\n",pthread_self(),__FUNCTION__,__LINE__); }
#define DROPMUTEX { \
	fprintf(stderr,"Dropping logf_mutex (@%s:%d)...",__FUNCTION__,__LINE__); \
	pthread_mutex_unlock(&logf_mutex); \
	fprintf(stderr,"...%p:%s:%d dropped logf_mutex\n",pthread_self(),__FUNCTION__,__LINE__); }
#else
#define GRABMUTEX pthread_mutex_lock(&logf_mutex);
#define DROPMUTEX pthread_mutex_unlock(&logf_mutex);
#endif
#else
#define GRABMUTEX
#define DROPMUTEX
#endif

#define BACKUPLOGFILE "/tmp/ltp.log"

/// This is the array of error strings
static const char *error_strings[]={
	"Success!", ///< LTP_OK
	"Some old error - no further information right now", ///< LTPERR
	"Some parameter you gave me was wrong", ///< LTP_BAD_PARAM
	"Some old memory error (e.g. malloc maybe)", ///< LTP_MEMORY
	"A problem doing some file stuff", ///< LTP_FILE
	"Some recursion went haywire", ///<LTP_BAD_RECURS
	"Some internal API is in a crazy/unexpected state!", ///<LTP_BAD_STATE
	"Encountered a crazy/unexpected value!", ///<LTP_BAD_VAL
	"Not really an error at all!!!"
};

/// Special string if input is crap
static const char *crappy="Your input error number was garbage so I've no idea what else to say!";

static FILE *ls_logf=NULL; ///< log file handle
static char *ls_log_fname=NULL; ///< log file name
static int ls_log_level=LTP_DEFAULT_LOG_LEVEL; ///< logging level
#ifndef LTP_NOTIMESTAMP
static ltp_time_t	ls_lastlog; ///< ltp time of last log event
static ltp_time_t	ls_startlog; ///< ltp time of first log event
#ifndef LTP_TIMESTAMP_FREQ
#define LTP_TIMESTAMP_FREQ 60 ///< how often (seconds) to include a timestamp
#endif
#endif

static char *ls_cfg_dir=NULL;

/*!
 * @brief close logfile etc. 
 * @return zero or error
 */
int ltp_log_close(void)
{
	int rv;
	GRABMUTEX;
	if (ls_logf) rv=fclose(ls_logf);
	ls_logf=NULL;
	DROPMUTEX;
	return(rv);
}

/*!
 * @brief re-open logfile etc. (after close, but same params otherwise)
 * @return zero or error
 */
int ltp_log_reopen(void)
{
	GRABMUTEX;
#ifdef LTP_LOGSIZE
	ls_logf=fopen(ls_log_fname,"a+");
#else
	ls_logf=fopen(ls_log_fname,"a");
#endif
	if (!ls_logf) {
		fprintf(stderr,"%s: Can't reopen %s [%s:%d]\n",__FUNCTION__,ls_log_fname,__FILE__,__LINE__);
		DROPMUTEX;
		return(LTP_FILE);
	}
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @brief rollover the logfile if we're limiting its size
 * @return zero or error
 *
 * We consider three parts of the logfile, the start the
 * latest and the last-but-one. Each part is allowed to
 * grow to 1/3rd the overall allowed size.
 */
static int ltp_log_rollover(void)
{
#ifndef LTP_LOGSIZE
	return(LTP_OK);
#else
	GRABMUTEX;
	static int donefirst=0;
	struct stat logstat;
	int chunklen=LTP_LOGSIZE/3;
	int rv;
	// is is time to rollover?
	rv=fstat(fileno(ls_logf),&logstat);
	if (rv) {
		fprintf(stderr,"%s: Can't stat %s (ls_logf=%p, rv=%d) [%s:%d]\n",__FUNCTION__,ls_log_fname,ls_logf,rv,__FILE__,__LINE__);
		DROPMUTEX;
		return(LTP_FILE);
	}
	if (!donefirst && logstat.st_size>=LTP_LOGSIZE) {
		// do stuff
		char *chunkname=NULL; ///< log file chunk name 
		FILE *chunkfp;
		int chunknamelen=0;
		int frv=0;
		unsigned char *chunk=NULL;

		chunknamelen=strlen(ls_log_fname)+10;
		chunkname=(char*)malloc(chunknamelen+1);
		if (!chunkname) {
			fprintf(stderr,"%s: memory error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			DROPMUTEX;
			return(LTP_MEMORY);
		}
		chunk=(unsigned char *)malloc(chunklen);
		if (!chunk) {
			fprintf(stderr,"%s: memory error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); 
			DROPMUTEX;
			return(LTP_MEMORY);
		}
		frv=fseek(ls_logf,0,SEEK_SET); // rewind
		if (frv) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			DROPMUTEX;
			return(LTP_FILE);
		}

		// read/write 1st 3rd
		snprintf(chunkname,chunknamelen,"%s.first",ls_log_fname);
		chunkfp=fopen(chunkname,"w");
		if (!chunkfp) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			DROPMUTEX;
			return(LTP_FILE);
		}
		frv=fread(chunk,1,chunklen,ls_logf);
		if (frv!=chunklen) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			fclose(chunkfp);
			DROPMUTEX;
			return(LTP_FILE);
		}
		frv=fwrite(chunk,1,chunklen,chunkfp);
		if (frv!=chunklen) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			fclose(chunkfp);
			DROPMUTEX;
			return(LTP_FILE);
		}
		fclose(chunkfp);

		// read/write last 3rd - the last-but-one file, we loose the middle 3rd at this stage
		frv=fseek(ls_logf,chunklen,SEEK_CUR);
		if (frv) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			DROPMUTEX;
			return(LTP_FILE);
		}
		snprintf(chunkname,chunknamelen,"%s.lbo",ls_log_fname);
		chunkfp=fopen(chunkname,"w");
		if (!chunkfp) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			DROPMUTEX;
			return(LTP_FILE);
		}
		frv=fread(chunk,1,chunklen,ls_logf);
		if (frv!=chunklen) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			fclose(chunkfp);
			DROPMUTEX;
			return(LTP_FILE);
		}
		frv=fwrite(chunk,1,chunklen,chunkfp);
		if (frv!=chunklen) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			fclose(chunkfp);
			DROPMUTEX;
			return(LTP_FILE);
		}
		fclose(chunkfp);
		// rewind the logfile now
		frv=fseek(ls_logf,0,SEEK_SET); // rewind
		if (frv) {
			fprintf(stderr,"%s: file error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			free(chunkname); free(chunk);
			DROPMUTEX;
			return(LTP_FILE);
		}
		donefirst=1;
		DROPMUTEX;
		return(LTP_OK);

	} else if (donefirst && logstat.st_size>=chunklen) {

		// rename then reopen logfile
		char *chunkname=NULL; ///< log file chunk name 
		int chunknamelen=0;
		int frv=0;

		// read/write last 3rd - the last-but-one file, we loose everything else
		chunknamelen=strlen(ls_log_fname)+10;
		chunkname=(char*)malloc(chunknamelen+1);
		if (!chunkname) {
			fprintf(stderr,"%s: memory error [%s:%d]\n",__FUNCTION__,__FILE__,__LINE__);
			DROPMUTEX;
			return(LTP_MEMORY);
		}
		snprintf(chunkname,chunknamelen,"%s.lbo",ls_log_fname);
		fclose(ls_logf);
		frv=rename(ls_log_fname,chunkname);
		if (frv) {
			fprintf(stderr,"%s: file rename error (%s->%s) [%s:%d]\n",__FUNCTION__,ls_log_fname,chunkname,__FILE__,__LINE__);
			free(chunkname);
			DROPMUTEX;
			return(LTP_MEMORY);
		}
		free(chunkname);
		frv=ltp_log_reopen();
		DROPMUTEX;
		return(frv);
	}
	DROPMUTEX;
	return(LTP_OK);
#endif
}

#ifdef DBGSCRPPLEASE
#define DBGSCR printf("At %s:%d\r\n",__FILE__,__LINE__)
#else
#define  DBGSCR
#endif

/*!
 * @fn int ltp_log_init(char *logfile, int level)
 * @brief initialise (or switch) logging to the named file at the given level
 * @param logfile pathname (platform dependent) of the logfile
 * @param level the threshold level (only this level and lower are actually logged)
 * @return zero for success LTP_OK or error value
 */
int ltp_log_init(char *logfile, int level) 
{	
	GRABMUTEX;
	char lltpstr[LTPSTR_LEN];
	int rv;
	if (ls_log_fname) {
		DBGSCR;
		snprintf(lltpstr,LTPSTR_LEN,"Setting new log file to %s, level to %d",logfile,level);
		DBGSCR;
		ltp_log(ls_log_level,lltpstr);
		DBGSCR;
		free(ls_log_fname);
		DBGSCR;
	}
	if (ls_logf!=NULL) {
		rv=fclose(ls_logf);
		DROPMUTEX;
		if (rv) LTPERR(rv);
	}
	ls_log_fname=strdup(logfile);
	DBGSCR;
#ifdef LTP_LOGSIZE
	ls_logf=fopen(ls_log_fname,"a+");
	DBGSCR;
#else
	ls_logf=fopen(ls_log_fname,"a");
	DBGSCR;
#endif
	if (!ls_logf) { // bummer: how do I tell you about this then?
	DBGSCR;
		// First I'll try again in a mo... :-)
		sleep(10);
#ifdef LTP_LOGSIZE
		ls_logf=fopen(ls_log_fname,"a+");
#else
		ls_logf=fopen(ls_log_fname,"a");
#endif
		if (ls_logf) { // phew - recovered! might tell someone...
		DBGSCR;
			snprintf(lltpstr,LTPSTR_LEN,"Had to sleep 10 before I could open this logfile.\n");
		DBGSCR;
			ltp_log(LTP_LOG_ERR,lltpstr);
		DBGSCR;
		} else { // aargh
			fprintf(stderr,"ltp: can't open logfile %s, trying %s\n",ls_log_fname,BACKUPLOGFILE);
			free(ls_log_fname);
			ls_log_fname=strdup(BACKUPLOGFILE);
#ifdef LTP_LOGSIZE
			ls_logf=fopen(ls_log_fname,"a+");
#else
			ls_logf=fopen(ls_log_fname,"a");
#endif
			if (!ls_logf) { // BIG bummer: how do I tell you about this then?
				fprintf(stderr,"ltp: can't open backup logfile either! exiting\n");
				DROPMUTEX;
				exit(-1);
			}
		}
	}
	ls_log_level=level;
#ifndef LTP_NOTIMESTAMP
	{
		// host info
		struct utsname un;
		char hostname[LTPSTR_LEN];
		char domainname[LTPSTR_LEN];
		ls_lastlog=ltp_now();
		ls_startlog=ls_lastlog;
		DBGSCR;
		snprintf(lltpstr,LTPSTR_LEN,"Starting logging at perftime: %g\n",ls_lastlog);
		DBGSCR;
		ltp_log(LTP_LOG_ERR,lltpstr);
		DBGSCR;
		rv=uname(&un);
		DBGSCR;
		if (rv) {
			snprintf(lltpstr,LTPSTR_LEN,"uname failed (%d)\n",rv);
		} else {
			snprintf(lltpstr,LTPSTR_LEN,"uname: %s %s %s %s %s,\n",
						un.sysname, un.nodename, un.release, un.version, un.machine);
		}
		DROPMUTEX;
		ltp_log(ls_log_level,lltpstr);
		GRABMUTEX;
		// host and domain
		rv=gethostname(hostname,LTPSTR_LEN);
		if (rv) {
			snprintf(lltpstr,LTPSTR_LEN,"gethostname failed (%d)\n",rv);
		} else {
			snprintf(lltpstr,LTPSTR_LEN,"gethostname: %s\n",hostname);
		}
		DROPMUTEX;
		ltp_log(ls_log_level,lltpstr);
		GRABMUTEX;
		rv=getdomainname(domainname,LTPSTR_LEN);
		if (rv) {
			snprintf(lltpstr,LTPSTR_LEN,"getdomainname failed (%d)\n",rv);
		} else {
			snprintf(lltpstr,LTPSTR_LEN,"getdomainname: %s\n",domainname);
		}
		DROPMUTEX;
		ltp_log(ls_log_level,lltpstr);
		GRABMUTEX;
		snprintf(lltpstr,LTPSTR_LEN,"ltp.h version: %s\n",LTP_BUILD_VERSION);
		DROPMUTEX;
		ltp_log(ls_log_level,lltpstr);
		GRABMUTEX;
		snprintf(lltpstr,LTPSTR_LEN,"LTPlib build: %s %s\n",LTP_BUILD_DATE,LTP_BUILD_TIME);
		DROPMUTEX;
		ltp_log(ls_log_level,lltpstr);
		GRABMUTEX;
	}
#endif
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @fn int ltp_log(int level, char *logstr)
 * @brief initialise logging to the named file
 * @param level is the loglevel of this event
 * @param logstr is the event string
 * @return zero for success LTP_OK or error value
 */
int ltp_log(int level, char *logstr) 
{
	GRABMUTEX;
	int rv=0;
	if (ls_logf==NULL) {
		DROPMUTEX;
		rv=ltp_log_init((char*)LTP_DEFAULT_LOG_FILE_NAME, LTP_DEFAULT_LOG_LEVEL);
		GRABMUTEX;
		if (rv) fprintf(stderr,"Error initialising ltp log (%s,%d,%d) at %d in %s\n",
					(char*)LTP_DEFAULT_LOG_FILE_NAME,LTP_DEFAULT_LOG_LEVEL,rv,__LINE__,__FILE__);
	}
	if (level<=ls_log_level && level >= 0) {
		int len=strlen(logstr)+100;
		char *foo=(char*) malloc(len);
		if (!foo) fprintf(stderr,"Error logging for ltp: (%d) at %d in %s\n",
						LTP_MEMORY,__LINE__,__FILE__);
#ifndef LTP_NOTIMESTAMP
		{
			ltp_time_t	now=ltp_now();
			ltp_time_t	elapsed=now-ls_startlog;
			char lltpstr[LTPSTR_LEN];
#define CONSTANT_TIMESTAMPS
#ifdef CONSTANT_TIMESTAMPS
			snprintf(lltpstr,LTPSTR_LEN,"%s,%g: ",ltp_now_str(),elapsed);
			fprintf(ls_logf,"%s",lltpstr);
#else
			if ((now-ls_lastlog)>LTP_TIMESTAMP_FREQ) {
				snprintf(lltpstr,LTPSTR_LEN,"mark: %s elapsed %g\n",ltp_now_str(),elapsed);
				ls_lastlog=now;
				fprintf(ls_logf,"%s",lltpstr);
			}
#endif
		}
#else
		fprintf(ls_logf,"%d:",level);
#endif
		// put in a CR iff there isn't one
		if (logstr[strlen(logstr)-1]=='\n') {
			snprintf(foo,len,"%s",logstr);
		} else {
			snprintf(foo,len,"%s\n",logstr);
		}
		fprintf(ls_logf,"%s",foo);
		free(foo);
	} 
	DROPMUTEX;
	if (ltp_log_rollover()) {
		return(LTP_FILE);
	}
#define LTP_ALWAYS_FLUSH_LOG
#ifdef LTP_ALWAYS_FLUSH_LOG
	ltp_log_flush();
#endif
	return(LTP_OK);
}

/*!
 * @brief How many seconds have elapsed since we started (logging)
 * @return is the elapsed time
 */
int ltp_elapsed()
{
	ltp_time_t	now=ltp_now();
	ltp_time_t	elapsed=now-ls_startlog;
	return((int)floor(elapsed));
	
}

/*!
 * @fn int ltp_log_flush(void)
 * @brief flush events to the current logfile
 * @return zero for success LTP_OK or error value
 */
int ltp_log_flush(void) 
{
	GRABMUTEX;
	if (ls_logf!=NULL) fflush(ls_logf); 
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @fn ltp_log_timestamp(const char *msg) 
 * @brief Write a timestamped message to the logfile (at LTP_LOG_CRITICAL level logging)
 * @param msg is the message to include with the time
 * @return is zero or otherwise for errors
 */
int ltp_log_timestamp(const char *msg) 
{
	GRABMUTEX;
	if (ls_logf==NULL) return(LTP_OK);
#define THISMSGLEN 255
	char thismsg[THISMSGLEN];
	time_t localt=time(NULL);
	if (msg) snprintf(thismsg,THISMSGLEN,"%s at: %s",msg,asctime(gmtime(&localt)));
	else snprintf(thismsg,THISMSGLEN,"Time: %s",asctime(gmtime(&localt)));
	DROPMUTEX;
	ltp_log(LTP_LOG_CRITICAL,thismsg);
	return(LTP_OK);
}


/*!
 * @fn ltp_logfile_fp(void)
 * @brief get a handle on the logfile (file) pointer
 * @return is the "FILE *" for the logfile pointer (or NULL if none)
 *
 * Danger! Threads may mess up if they do this
 */
FILE *ltp_logfile_fp(void)
{
	GRABMUTEX;
	int rv;
	if (ls_logf==NULL) {
		DROPMUTEX;
		rv=ltp_log_init((char*)LTP_DEFAULT_LOG_FILE_NAME, LTP_DEFAULT_LOG_LEVEL);
		GRABMUTEX;
		if (rv) fprintf(stderr,"Error initialising ltp log (%s,%d,%d) at %d in %s\n",
					(char*)LTP_DEFAULT_LOG_FILE_NAME,LTP_DEFAULT_LOG_LEVEL,rv,__LINE__,__FILE__);
	}
	DROPMUTEX;
	return (ls_logf);
}

/// This is the Julian date (as a double) for Jan1 1970 which is when time_t starts
#define JULIAN_ADDITION_TO_TIME_T 2440587.5

/*! 
 * @fn ltp_now(void)
 * @brief return the current time in appropriate format
 * @return is the current time as a ltp_time_t
 */
ltp_time_t ltp_now(void) {
	// sanity test - see if now > s/w build
	// if its not then something's up.
	// in which case, if we know when the schedule
	// starts, we'll assume that the program started
	// executing at the start of the schedule
	// This makes systems with no RTC work to some
	// extent, though not predictable, so its not really
	// a way to recover from a broken RTC.
	static int pt_inited=0;
	static int badclock=0;
	ltp_time_t	increment=0;
	ltp_time_t	decrement=0;
	ltp_time_t	build_time; // some function of =__DATE__;
	time_t			nowt=time(NULL);
	static time_t	adjust=0;
	if (!pt_inited) { // check how time and build time etc compare
		struct tm	bt; // build time
		char   tmfoo[100];
		time_t	buildt;
		memset(&bt,0,sizeof( struct tm));
		pt_inited=1; // better do this before any error since otherwise we'll recurse!
		snprintf(tmfoo,100,"%s %s ",LTP_BUILD_DATE,LTP_BUILD_TIME);
		if (!strptime(tmfoo,"%b %d %Y %T",&bt)) {
			snprintf(ltpstr,LTPSTR_LEN,"Can't parse time string: %s\n",tmfoo);
			ltp_log(LTP_LOG_MAX,ltpstr);
			badclock=1;
		} else {
			buildt=mktime(&bt);
			snprintf(ltpstr,LTPSTR_LEN,"Current time_t: %lu, biuld time_t: %lu\n",(unsigned long)nowt,(unsigned long)buildt);
			ltp_log(LTP_LOG_MAX,ltpstr);
			if (nowt<=buildt) { // weird - runinng before code was built! => bad clock
				badclock=1;
			}
		}
		if (badclock) {
			adjust=buildt-nowt;
		}
	}
#ifdef LTP_TARGET_CYGWIN
	return((ltp_time_t)nowt+adjust+JULIAN_ADDITION_TO_TIME_T);
#endif
#ifdef LTP_TARGET_LINUX
	return((ltp_time_t)nowt+adjust+JULIAN_ADDITION_TO_TIME_T);
#endif
#ifdef LTP_TARGET_ARM_LINUX
	return((ltp_time_t)nowt+adjust+JULIAN_ADDITION_TO_TIME_T);
#endif
#ifdef LTP_TARGET_OMNETPP
	return((ltp_time_t)simTime());
#endif
}

static char tbuf[MAXPATH]; /// static PRINTABLE 	time buffer
static char pbuf[MAXPATH]; /// static PRINTABLE 	time buffer

/*!
 * @brief print the time now as a string incl. milliseconds
 * @return ptr to static char * with the string 
 */
char *ltp_now_str(void)
{
	static char thestr[MAXPATH];
	struct timeval tv;
	struct tm* ptm;
	char time_string[40];
	long milliseconds;
	
	pthread_mutex_lock(&time_str_mutex); 
	/* Obtain the time of day, and convert it to a tm struct. */
	gettimeofday (&tv, NULL);
	ptm = localtime (&tv.tv_sec);
	/* Format the date and time, down to a single second. */
	strftime (time_string, sizeof (time_string), "%Y-%m-%d %H:%M:%S", ptm);
	/* Compute milliseconds from microseconds. */
	milliseconds = tv.tv_usec / 1000;
	/* Print the formatted time, in seconds, followed by a decimal point
	and the milliseconds. */
	snprintf (thestr,MAXPATH,"%s.%03ld", time_string, milliseconds);
	pthread_mutex_unlock(&time_str_mutex); 
	return(thestr);
}

/*!
 * @brief a nicer print form of ltp_time_t
 * @param thetime is a prefume_time_t value
 * @param buf is (the caller alloc'd) string in which to print the time (also returned)
 * @return is a static string
 */
char *ltp_time_str_aux(ltp_time_t thetime, char *buf)
{
#ifndef LTP_TARGET_OMNETPP
	// ok so only one process will write at a time, but reads could get corrupted, that's ok
	// for this since its only used for debug and LTP-T via's
	pthread_mutex_lock(&time_str_mutex); 
	time_t	thetime_t=(time_t)(thetime-JULIAN_ADDITION_TO_TIME_T);
	snprintf(buf,MAXPATH,"%s",asctime(gmtime(&thetime_t)));
	if (buf[strlen(buf)-1]=='\n') buf[strlen(buf)-1]='\0';
	ltp_time_t	partial=thetime-(thetime_t+JULIAN_ADDITION_TO_TIME_T);
	if (partial) {
		snprintf(buf+strlen(buf),MAXPATH-strlen(buf),"(+ %g s)",partial);
	}
	pthread_mutex_unlock(&time_str_mutex);
	return(buf);
#else
#error "do omnetpp string stuff"
#endif
}

/*!
 * @brief a nicer print form of ltp_time_t
 * @param thetime is a prefume_time_t value
 * @return is a static string
 *
 * Don't call this from within the LTP (pinger) or L2 (listener) threads. Just in case;-)
 */
char *ltp_time_str(ltp_time_t thetime)
{
	return(ltp_time_str_aux(thetime,tbuf));
}

/*!
 * @brief a nicer print form of ltp_time_t
 * @param thetime is a prefume_time_t value
 * @return is a static string
 *
 * Total hack - the ltp_time_str function to call from within the LTP (pinger) thread. 
 * You'd still better not call this from the L2 (listener) thread!
 */
char *ltp_time_str_pinger(ltp_time_t thetime)
{
	return(ltp_time_str_aux(thetime,pbuf));
}

/*!
 * @brief an xml dateTime compliant print form of ltp_time_t
 * @param thetime is a prefume_time_t value
 * @return is a static string
 */
char *ltp_xml_time_str(ltp_time_t thetime)
{
#ifndef LTP_TARGET_OMNETPP
	time_t	thetime_t=(time_t)(thetime-JULIAN_ADDITION_TO_TIME_T);
	ltp_time_t	partial=thetime-(thetime_t+JULIAN_ADDITION_TO_TIME_T);
	struct tm gt,*gtime;
	struct tm *foo=gmtime_r(&thetime_t,&gt);
	gtime=&gt;
	if (partial) {
		snprintf(tbuf,MAXPATH,"%04d-%02d-%02dT%02d:%02d:%02d.%g\n",
					gtime->tm_year+1900,gtime->tm_mon+1,gtime->tm_mday,
					gtime->tm_hour,gtime->tm_min,gtime->tm_sec,
					partial);
	} else {
		snprintf(tbuf,MAXPATH,"%04d-%02d-%02dT%02d:%02d:%02d\n",
					gtime->tm_year+1900,gtime->tm_mon+1,gtime->tm_mday,
					gtime->tm_hour,gtime->tm_min,gtime->tm_sec);
	}
	if (tbuf[strlen(tbuf)-1]=='\n') tbuf[strlen(tbuf)-1]='\0';
	return(tbuf);
#else
#error "do omnetpp string stuff"
#endif
}

/*!
 * @brief Dump an ascii hex version of a buffer into the log file
 * @param logstr is a helpful string
 * @param len is the length of the buffer
 * @param buf is the buffer
 *
 * This is always done at the LTP_LOG_MAX level
 */
int ltp_log_buffer(char *logstr, int len, unsigned char *buf)
{
	GRABMUTEX
	if (ls_log_level==LTP_LOG_MAX) {
		// put in a CR iff there isn't one
		char *foo1=NULL,*foo=NULL,*foop=NULL;
		int i;
		foo1=(char*)malloc(strlen(logstr)+10);
		if (!foo1) fprintf(stderr,"Error logging for ltp: (%d) at %d in %s\n",
						LTP_MEMORY,__LINE__,__FILE__);
		if (logstr[strlen(logstr)-1]=='\n') {
			snprintf(foo1,strlen(logstr)+10,"LB:%s",logstr);
		} else {
			snprintf(foo1,strlen(logstr)+10,"LB:%s\n",logstr);
		}
		fprintf(ls_logf,"%s",foo1);
		free(foo1);

		foo=(char*) malloc(78*(len/32+1)+100); // old: foo=(char*) malloc(3*len+7*len/72+1);
		if (!foo) fprintf(stderr,"Error logging for ltp: (%d) at %d in %s\n",
						LTP_MEMORY,__LINE__,__FILE__);
		int bar=0;
		foop=foo;
		for (i=0;i!=len;i++) {
			if (!(i%32)) { *foop++='\n'; *foop++='L'; *foop++='B'; *foop++=':'; *foop++='\t'; }
			if (!(i%4)) *foop++=' ';
			bar=buf[i]>>4;
			*foop++=HEXDIGIT(bar);
			bar=buf[i]&0x0f;
			*foop++=HEXDIGIT(bar);
		}
		*foop='\0';

		fprintf(ls_logf,"%s\n",foo);
		free(foo);
	} 
	DROPMUTEX
	return(LTP_OK);
}

/*!
 * @brief initialise the random stream for this thread
 * @param idthing is something unique to the stream (e.g. thread fnc addr)
 * @return zero or error
 *
 * Just so that each thread gets a chance
 */
int ltp_rand_init(int idthing)
{
	static int seed=0x19651117; // sf's birthday:-)
	seed ^= idthing;
#ifdef LTP_TARGET_ARM_LINUX
#define RNGFILE "/proc/sys/kernel/random/uuid"
#define DORNGFILE
#endif
#ifdef LTP_TARGET_LINUX
#ifdef PARANOID
#define RNGFILE "/dev/random"
#else
#define RNGFILE "/dev/urandom"
#endif
#define DORNGFILE
#endif
#ifdef LTP_TARGET_CYGWIN
#define RNGFILE "/dev/random"
#define DORNGFILE
#endif
#ifdef LTP_RAND_SEED
	int rngfile=open(RNGFILE,O_RDONLY);
	if (!rngfile) {
		snprintf(ltpstr,LTPSTR_LEN,"Rand: can't open rng file (%s)\n",RNGFILE);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
	int readval;
	int rv=read(rngfile,&readval,sizeof(readval));
	close(rngfile);
	if (rv!=sizeof(readval)) {
		snprintf(ltpstr,LTPSTR_LEN,"Rand: error (rv=%d,errno=%d) reading rng file (%s)\n",
						rv,errno,RNGFILE);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
	seed ^= readval;
	snprintf(ltpstr,LTPSTR_LEN,"Rand: read value (0x%x) from file (%s), seed now: 0x%x\n",readval,RNGFILE,seed);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	snprintf(ltpstr,LTPSTR_LEN,"Rand: Seed for stream %d is 0x%x\n",idthing,seed);
	ltp_log(LTP_LOG_MAX,ltpstr);
	srand(seed);
	return(LTP_OK);
}

/*!
 * @brief return a random int
 * @return is something random-ish
 */
int ltp_rand(void)
{
	static int initialised=0;
	int ret;
	if (!initialised) {
		ltp_rand_init(0xff);
		initialised=1;
	}
	while (!(ret=random()));
#ifdef RNGLOG
	snprintf(ltpstr,LTPSTR_LEN,"Rand: ret is %d\n",ret);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	return(ret);
}

/*!
 * @brief return a random octet string of the length required
 * @param len is the length in bytes of the required random value
 * @param val is a caller-allocated buffer to hold the random bytes
 * @return zero or error
 */
int ltp_rand_octets(int len, unsigned char *val)
{
	// plan - use ltp_rand a number of times and copy those to the buffer
	int 	iterations=len/sizeof(int);
	int 	lastfew=len%4;
	int 	foo;
	int 	*ival=(int*)val;

	for (int i=0;i!=iterations;i++) {
		foo=ltp_rand();
		ival[i]=foo;
	}
	if (lastfew) { // last few bytes
		foo=ltp_rand();
		switch (lastfew) {
			case 3: val[len-3]=(unsigned char)((foo>>8) & 0xff); 
			case 2: val[len-2]=(unsigned char)((foo>>16) & 0xff);
			case 1: val[len-1]=(unsigned char)((foo>>24) & 0xff);
				break;
			default:
				return(LTP_ERR);
		}
	}
	return(LTP_OK);
}


/// this is where we keep the base directory for our file store
static char basedir[MAXPATH];

/*!
 * @brief setup disk settings
 * @param dir is the directory to use
 * @param filetype is a LTP_FILETYPE_* setting e.g. LTP_FILETYPE_LTP or LTP_FILETYPE_SCHEDULE
 * @param ofile is the name of the file to setup (MAXPATH should be alloc'd already!)
 * @return zero or error
 * @todo we're touching a temp file here to make sure we can write to the file system, but I need to check
 * if there are any related attacks - don't think so, since they're all race conditions and I really
 * only create and immediately delete the file and if someone gets there first and deletes it or
 * otherwise prevents me from deleting it, then I'm ok (but think about mounted file systems and
 * possible DoS if say the remove() call blocked.
 */
int ltp_filetype_setup(char *dir, int filetype, char *ofile)
{
	struct stat dirstat;
	char testdir[MAXPATH];
	char *ldir=dir;
	if (!ldir) ldir=ls_cfg_dir;
	if (ldir && strlen(ldir)>(MAXPATH-PATHSAFETY)) LTPERR(LTP_BAD_PARAM);
	// check the dir exists and setup subsidiary file names and pointers etc.
#ifdef LTP_TARGET_CYGWIN
	switch (filetype) { // setup testdir depending on system (ignore filetype for now)
			case LTP_FILETYPE_LTP:
				if (!ldir) snprintf(testdir,MAXPATH,LTP_CFG_DEFAULT_DIR);
				else snprintf(testdir,MAXPATH,ldir);
				break;
			default:
				return(LTP_BAD_PARAM);
	}
#else
#ifdef LTP_TARGET_LINUX
	switch (filetype) {
			case LTP_FILETYPE_LTP:
				if (!ldir) {
					struct passwd *pent;
					uid_t	realuid=getuid();
					pent=getpwuid(realuid);
					if (!pent) LTPERR(LTP_LOG_SYS);
					snprintf(testdir,MAXPATH,"%s/%s",pent->pw_dir,LTP_CFG_DEFAULT_DIR);
				} else {
					snprintf(testdir,MAXPATH,"%s",ldir);
				}
				break;
			default:
				return(LTP_BAD_PARAM);
	}
#else
#ifdef LTP_TARGET_ARM_LINUX
	switch (filetype) {
			case LTP_FILETYPE_LTP:
				if (!ldir) snprintf(testdir,MAXPATH,LTP_CFG_DEFAULT_DIR);
				else snprintf(testdir,MAXPATH,ldir);
				break;
			default:
				return(LTP_BAD_PARAM);
	}
#else
#error "figure out where to put files on your system!"
#endif
#endif
#endif
	if (!stat(testdir,&dirstat)) { // possibly ok if zero return
		if (S_ISDIR(dirstat.st_mode)) { // looking good again - try touch a file there
			char tmpfile[MAXPATH];
			int fd;
			if (testdir[strlen(testdir)-1]=='/') {
				snprintf(tmpfile,MAXPATH,"%sfooXXXXXX",testdir);
			} else {
				snprintf(tmpfile,MAXPATH,"%s/fooXXXXXX",testdir);
			}
			fd=mkstemp(tmpfile);
			if (fd!=-1) { // wee-hee
				int rv;
				close(fd);
				rv=remove(tmpfile);
				if (!rv) {
					char lfname[MAXPATH];
					if (filetype==LTP_FILETYPE_LTP) snprintf(lfname,MAXPATH,LTP_FNAME);
					snprintf(basedir,MAXPATH,"%s",testdir);
					if (testdir[strlen(testdir)-1]=='/') {
						snprintf(ofile,MAXPATH,"%s%s",testdir,lfname);
					} else {
						snprintf(ofile,MAXPATH,"%s/%s",testdir,lfname);
					}
#ifdef LTP_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"LTP file (type:%d) is %s\n",filetype,ofile);
					ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				} else LTPERR(LTP_FILE);
			} else LTPERR(LTP_FILE);
		} else LTPERR(LTP_FILE);
	} else LTPERR(LTP_FILE);
	return(LTP_OK);
}

/*!
 * @brief map from YYYYMMDD etc. string to ltp_time_t
 * @param str is the above type string
 * @param timep is the pointer to the ltp_time_t for the result
 * @return zero or error
 */
int str2ltp_time_t(char *str, ltp_time_t	*timep)
{
	const char *fmtstrs[]= {
						"%a %b %d %H:%M:%S %Y",
						"%Y-%m-%dT%H:%M:%S",
						"%Y%m%d%H:%M:%S",
						"%Y%m%d%H%M%S",
						"%Y%m%d%H%M",
						"%Y%m%d%H",
						"%Y%m%d",
						NULL,
						};
	struct tm	then;
	memset(&then,0,sizeof(tm));
	time_t		then_t;
	int			done=0;
	char		**fs=(char**)fmtstrs;
	if (!timep) LTPERR(LTP_BAD_PARAM);
	while (!done && *fs) {
		if (strptime(str,*fs,&then)) {
			done=1;
		} else {
			fs++;
		}
	}
	if (!done) LTPERR(LTP_BAD_PARAM);
	then_t=mktime(&then);
	*timep=then_t+JULIAN_ADDITION_TO_TIME_T;
	return(LTP_OK);
}


/*!
 * @brief decode a double from "my" network byte order for doubles 
 * @param bp is the buffer pointer (guaranteed big enough)
 * @param val is a ptr to where to put the double
 * @return zero or error (not that there are any!)
 */
int my_double_dec(unsigned char* bp, double *val)
{
	long long bar=0;
#ifdef LTP_TARGET_ARM_LINUX
	// arm-linux seems to store the 32-bit words of the "long long" in a different order to
	// everyone else
	bar+=((long long)bp[4]) << 56;
	bar+=((long long)bp[5]) << 48;
	bar+=((long long)bp[6]) << 40;
	bar+=((long long)bp[7]) << 32;
	bar+=((long long)bp[0]) << 24;
	bar+=((long long)bp[1]) << 16;
	bar+=((long long)bp[2]) << 8;
	bar+=((long long)bp[3]) ;
#else
	bar+=((long long)bp[0]) << 56;
	bar+=((long long)bp[1]) << 48;
	bar+=((long long)bp[2]) << 40;
	bar+=((long long)bp[3]) << 32;
	bar+=((long long)bp[4]) << 24;
	bar+=((long long)bp[5]) << 16;
	bar+=((long long)bp[6]) << 8;
	bar+=((long long)bp[7]) ;
#endif
	double *foo=(double*)&bar;
	*val=*foo;
	return(LTP_OK);
}

/*!
 * @brief encode a double to "my" network byte order for doubles 
 * @param val is the double
 * @param bp is the buffer pointer (guaranteed big enough)
 * @return zero or error (not that there are any!)
 */
int my_double_enc(unsigned char *bp, double val)
{
	long long *foo;
	foo=(long long*)&val;
	long long bar=*foo;
#ifdef LTP_TARGET_ARM_LINUX
	// arm-linux seems to store the 32-bit words of the "long long" in a different order to
	// everyone else
	bp[4]=(bar&0xff00000000000000LL)>>56;
	bp[5]=(bar&0x00ff000000000000LL)>>48;
	bp[6]=(bar&0x0000ff0000000000LL)>>40;
	bp[7]=(bar&0x000000ff00000000LL)>>32;
	bp[0]=(bar&0x00000000ff000000LL)>>24;
	bp[1]=(bar&0x0000000000ff0000LL)>>16;
	bp[2]=(bar&0x000000000000ff00LL)>>8;
	bp[3]=(bar&0x00000000000000ffLL);
#else
	bp[0]=(bar&0xff00000000000000LL)>>56;
	bp[1]=(bar&0x00ff000000000000LL)>>48;
	bp[2]=(bar&0x0000ff0000000000LL)>>40;
	bp[3]=(bar&0x000000ff00000000LL)>>32;
	bp[4]=(bar&0x00000000ff000000LL)>>24;
	bp[5]=(bar&0x0000000000ff0000LL)>>16;
	bp[6]=(bar&0x000000000000ff00LL)>>8;
	bp[7]=(bar&0x00000000000000ffLL);
#endif
	return(LTP_OK);
}

/*!
 * @brief set the configuration directory (where schedules etc. are to be found)
 * @param dir is the directory to use
 * @return zero or error
 */
int ltp_set_cfg_dir(char *dir)
{
	struct stat dirstat;
	if (!dir) LTPERR(LTP_BAD_PARAM);
	// a basic check that the directory exists - could do more here
	if (stat(dir,&dirstat) || !S_ISDIR(dirstat.st_mode)) { 
		// oops something wrong, don't set & exit with an error
		snprintf(ltpstr,LTPSTR_LEN,"Can't stat configuration directory %s\n",dir);
		ltp_log(LTP_LOG_ERR,ltpstr);
		LTPERR(LTP_FILE);
	}
	// no more error checking - that'll be done when this is used
	if (ls_cfg_dir) free(ls_cfg_dir); 
	ls_cfg_dir=NULL;
	ls_cfg_dir=strdup(dir);
	return(LTP_OK);
}

/*!
 * @brief case-insensitive str comp
 * @param a first string
 * @param b 2nd string
 * @return non-zero if strings are considered the same; zero otherwise
 */
int ltp_same(char *a, char *b)
{
	if (strlen(a)!=strlen(b)) return(0);
	for (int i=0;i!=strlen(a);i++) {
		if (toupper(a[i])!=toupper(b[i])) return(0);
	}
	return(1);
}

/*!
 * @brief guess?
 * @param foo bar?
 * @return stuff
 */
int ltp_iswhitespace(char foo) 
{
	switch (foo) {
		case ' ':
		case '\t':
		case '\n':
		case '\r':
		case '\f':
			return(true);
		default:
			return(false);
	}
}

/*!
 * @brief compare an address to the ltpaddr wildcard
 * @param guy is the comparee
 * @param len is the length of the guy
 * @return true if the same
 */
bool isany(ltpaddr *guy,int len) 
{
	if (guy==&ltpany) return(true); 
	if (memcmp((void*)guy,(void*)&ltpany,len)) return(false);
	else return(true);
}

#ifdef LTP_MULTICAST
/*!
 * @fn int ltp_is_multi(addr)
 * @brief utility fnc to determine if target addr is MC or not
 * @return zero MC, error (1) if not
 */
int ltp_is_multi(ltpaddr *addr)
{
    unsigned long a32=htonl(addr->sock.sin_addr.s_addr);
    unsigned int topbyte=(a32>>24) & 0xff;
    if (topbyte >= 224 && topbyte <= 239) {
#ifdef LTP_MC_DEBUG
		snprintf(ltpstr,LTPSTR_LEN,"MC: ltp_is_multi says yes %lx is MC",a32);
	    ltp_log(LTP_LOG_MAX,ltpstr);
#endif
        return(LTP_OK);
    }
#ifdef LTP_MC_DEBUG
		snprintf(ltpstr,LTPSTR_LEN,"MC: ltp_is_multi says no %lx is not MC",a32);
	    ltp_log(LTP_LOG_MAX,ltpstr);
#endif
    return(LTP_ERR);
}
#endif

