/*!
 * @file ltp_names.cc
 * @brief Simple name -> host:port mapping
 * @version $Revision: 1.6 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ltp_base.h"
#include "ltp.h"

/*
 * @struct map from a name to another
 * @field from is the name being looked up
 * @field to is the result
 */
typedef struct name_map_defn {
	char *from;
	char *to;
} name_map;

/// the name map arrays
static name_map *themap=NULL;

/// the name map file
static char *map_fname=NULL;

/// the number of mappings
static int nmaps=0;

/// the default name map file
#define DEF_MAP_FNAME "ltp.names"

/// the separator here
#define LTP_NAME_SEP ",\n"

/*!
 * @brief read the map file
 * @return zero or error
 */
static int ltp_name_map_init(void)
{
	if (!map_fname) {
		map_fname=strdup(DEF_MAP_FNAME);
	}
	FILE *fp=fopen(map_fname,"r");
	if (!fp) {
		return(LTP_OK);
	}
	char line[MAXPATH];
	char *lp;
	char *fgrv;
	while ((fgrv=fgets(line,MAXPATH,fp))!=NULL) {
		if (strlen(line) > (MAXPATH-1)) { fclose(fp); LTPERR(LTP_BAD_VALUE); } /// line too long
		lp=line;
		// while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
		while (ltp_iswhitespace(*lp)) lp++;
		if (!strlen(lp)) continue; 
		if (*lp=='#') continue;
		if (*lp=='\n') continue;
		// read mapping
		char *sr;
		char *crv=strtok_r(line,LTP_NAME_SEP,&sr);
		if (!crv) continue;
		char *from=crv;
		crv=strtok_r(NULL,LTP_NAME_SEP,&sr);
		if (!crv) continue;
		char *to=crv;
		nmaps++;
		name_map *nm=(name_map*)realloc(themap,nmaps*sizeof(name_map));
		if (!nm) LTPERR(LTP_MEMORY);
		themap=nm;
		themap[nmaps-1].from=strdup(from);
		themap[nmaps-1].to=strdup(to);
	}
	fclose(fp);
	return(LTP_OK);
}


/*!
 * @brief translate a name if necessary
 * @param name is the input name
 * @return translated name (or NULL if no translation)
 */
char *ltp_name_trans(char *name)
{
	int rv;
	static bool inited=false;
	if (!inited) {
		rv=ltp_name_map_init();
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"ltp_name_map_init failed (%d)\n",rv);
			return(NULL);
		}
		inited=true;
	}
	// check mapping
	for (int i=0;i!=nmaps;i++) {
		if (ltp_same(name,themap[i].from)) { return(themap[i].to);}
	}
	return(NULL);
}

/*!
 * @brief set the name map file name
 * @param fname is the file name
 * @return zero or error
 */
int ltp_name_map_file(char *fname)
{
	if (map_fname) free(map_fname);
	map_fname=strdup(fname);
	return(LTP_OK);
}


