/*!
 * @file ltp_t.cc
 * @brief The LTP-T protocol's stuff
 * @version $Revision: 1.65 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <strings.h>

#include "ltp_base.h"
#include "ltp.h"
#include "ltp_int.h"
#include "ltp_crypt.h"
#include "ltp_ext.h"
#include "ltp_cues.h"
#include "ltp_codec.h"
#include "ltp_t.h"
#include "ltp_names.h"

/// separators for LTP-T config file
#define LTPT_CFG_SEP "=\n#"

/// separators for LTP-T route file
#define LTPT_ROUTE_SEP " 	,\n#"

/// default route file name
#define LTPT_DEF_ROUTE_FNAME "ltpt.routes"

#define LTPT_NUM_TAGS 11

#define LTPT_TAG_ON "ON"
#define LTPT_TAG_RELAY_SOCKETS "RELAY_SOCKETS"
#define LTPT_TAG_INCLUDE_EST "INCLUDE_ESTIMATE"
#define LTPT_TAG_PORT "PORT"
#define LTPT_TAG_HOPCOUNT "HOPCOUNT"
#define LTPT_TAG_INCLUDE_VIA "INCLUDE_VIA"
#define LTPT_TAG_E2E_CIPHER "E2E_CIPHER"
#define LTPT_TAG_E2E_KEYID "E2E_KEYID"
#define LTPT_TAG_E2E_MAP_FILE "E2E_MAP_FILE"
#define LTPT_TAG_NEED_E2E_INT "NEED_E2E_INT"
#define LTPT_TAG_ECN "ECN"

const char *ltpt_tags[]={
	LTPT_TAG_ON ,
	LTPT_TAG_RELAY_SOCKETS,
	LTPT_TAG_INCLUDE_EST ,
	LTPT_TAG_PORT ,
	LTPT_TAG_HOPCOUNT ,
	LTPT_TAG_INCLUDE_VIA ,
	LTPT_TAG_E2E_CIPHER ,
	LTPT_TAG_E2E_KEYID ,
	LTPT_TAG_E2E_MAP_FILE,
	LTPT_TAG_NEED_E2E_INT,
	LTPT_TAG_ECN 
};

/// default LTP-T cfg
ltp_t_cfg_so  ltp_t_def_cfg={
	false,
	LTP_T_DEF_RELAY_SOCKETS,
	false,
	LTP_T_NO_PORT,
	LTP_T_NO_HOPCOUNT,
	false,
	{ LTP_CS_NONE, 0, "" },
	false,
	false,
	LTP_T_DEF_MAP_FILE};

/// hack hack hack
extern ltp_state ltp_state_var;

#ifdef THREADS
#include <pthread.h>
static pthread_mutex_t tinit_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#undef MUTEXDEBUG
#ifdef MUTEXDEBUG
#define GRABMUTEX { \
	fprintf(stderr,"Grabbing tinit (in %s @ %p, tid=%p)...",__FUNCTION__,tinit_mutex,pthread_self()); \
	pthread_mutex_lock(&tinit_mutex); \
	fprintf(stderr,"..got tinit\n"); }
/// Give back the l2 listener mutex
#define DROPMUTEX { \
	fprintf(stderr,"Dropping tinit (in %s @ %p)...",__FUNCTION__,tinit_mutex); \
	pthread_mutex_unlock(&tinit_mutex); \
	fprintf(stderr,"...dropped tinit\n"); }
#else
#define GRABMUTEX pthread_mutex_lock(&tinit_mutex); 
#define DROPMUTEX pthread_mutex_unlock(&tinit_mutex);
#endif
#else
#define GRABMUTEX
#define DROPMUTEX
#endif

/// length of a route string
#define ROUTESTRLEN MAXPATH

/// place where we try print route info
static char routestr[ROUTESTRLEN];

/// whether or not crypto API mappings are initialised
static bool crypt_inited=false;

/*
 * @struct our route table type
 * @param route specifies the network and gateway
 */
typedef struct route_table_defn  {
	ltp_t_route_so			route;
	// later - add a ltp_t_cfg_so to go with that!
} route_table;

static int rt_size=0;
static route_table *rtab=NULL;

/*!
 * @brief (re-)initialise using a cfg
 * @param cfg is a ptr to an LTP-T cfg
 * @param firsttime is whether this is the first call to this or not
 * @return zero or error
 */
static int ltpt_init_aux(ltp_t_cfg_so *cfg, bool firsttime)
{
	if (!cfg) LTPERR(LTP_BAD_INPUT);
	int sock;
	int rv;
	// make relay sockets mainly
	if (!cfg->on) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"%s doing nothing - supplied cfg says LTP_T is off\n",__FUNCTION__);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		return(LTP_OK);
	}
	if (!cfg->relay_sockets) LTPERR(LTP_BAD_PARAM); // gotta have some listeners
	GRABMUTEX
	int relays_now=(ltp_state_var.t_cfg.on?ltp_state_var.t_cfg.relay_sockets:0);
	int new_relays=cfg->relay_sockets-relays_now;
	for (int i=relays_now;i<cfg->relay_sockets;i++) {
		// make new sockets
		sock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
		if (!sock) {
			DROPMUTEX
			LTPERR(LTP_SOCKET);
		}
		ltp_socket_t *sockp=&ltp_state_var.sockets[ltp_state_var.nsockets-1];
		sockp->opts.t_cfg=*cfg; // struct copy
// this is an emergency hack:-)
#define TWODAYS 60*60*24*2
		sockp->opts.inactivitylimit=TWODAYS;
		sockp->ltptrelay=true;
		/// TODO question - how to handle cookies and auth on these sockets?
		/// maybe ... dunno
	}
	ltp_state_var.t_cfg=*cfg; // struct copy
	if (new_relays>0) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Made %d %srelay sockets (had %d)\n",
			ltp_state_var.t_cfg.relay_sockets-relays_now,(firsttime?"":"more "),relays_now);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else {
		snprintf(ltpstr,LTPSTR_LEN,"%s doing nothing already have %d relays, (was asked for %d)\n",
			__FUNCTION__,relays_now,cfg->relay_sockets);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
	if (!cfg->on) { DROPMUTEX; return(LTP_OK);}
	// maybe TODO: flush existing routes if there are new ones??
	// we currently just add 'em at the end which means we match 'em first
	int nroutes=0;
	ltp_t_route_so *routes=NULL;
	rv=ltpt_load_routes((char*)LTPT_DEF_ROUTE_FNAME,&nroutes,&routes);
	if (rv) { // not fatal!
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Failed to load LTP_T routes\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Loaded %d LTP_T routes\n",nroutes);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		for (int i=0;i!=nroutes;i++) {
			rv=ltpt_set_route(&routes[i]);
			if (rv) { DROPMUTEX; LTPERR(rv); }// now its fatal
		}
		ltpc_free((unsigned char*)routes);
	}
	DROPMUTEX
	return (LTP_OK);
}

/*!
 * @brief re-initialise using a cfg
 * @param cfg is a ptr to an LTP-T cfg
 * @return zero or error
 */
int ltpt_reinit(ltp_t_cfg_so *cfg)
{
	return ltpt_init_aux(cfg,false);
}

/*!
 * @brief initialise using a cfg
 * @param cfg is a ptr to an LTP-T cfg
 * @return zero or error
 */
int ltpt_init(ltp_t_cfg_so *cfg)
{
	static bool done=false;
	if (done) return(LTP_OK);
	int rv=ltpt_init_aux(cfg,true);
	if (rv) LTPERR(rv);

	return(LTP_OK);
}

/*!
 * @brief initialise a cfg to default
 * @param cfg is a ptr to an LTP-T cfg
 * @return zero or error
 */
int ltpt_def_cfg(ltp_t_cfg_so *cfg)
{
	if (!cfg) LTPERR(LTP_BAD_PARAM);
	*cfg=ltp_t_def_cfg;
	return (LTP_OK);
}

/// how we detect a true/false value from cfg
#define TF(x) (x=='T'||x=='t')

/// pretty printing is nice
#define TFP(x) (x?"true":"false")

/*!
 * @brief load an LTP-T configuration
 * @param fname is the filename to use
 * @param cfg is a ptr to the caller allocated configuration structure
 * @return zero or error
 */
int ltpt_load_cfg(char *fname, ltp_t_cfg_so *cfg)
{
	if (!cfg) LTPERR(LTP_BAD_INPUT);
	int rv=0;
	rv=ltpt_def_cfg(cfg);
	if (rv) LTPERR(rv);
	FILE *fp=fopen(fname,"r");
	if (!fp) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Can't read LTP-T configuration: %s\n",fname);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		LTPERR(LTP_FILE);
	}
	// usual parsing 
	// parse file, line by line
	char line[MAXPATH];
	char *fgrv=NULL;
	char *lp;
	while ((fgrv=fgets(line,MAXPATH,fp))!=NULL) {
		if (strlen(line) > 1023) return(LTP_BAD_VALUE); /// line too long
		lp=line;
		while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
		if (!strlen(lp)) continue; 
		if (*lp=='#') continue;
		if (*lp=='\n') continue;
		// parse - get tag
		char *sr;
		char *crv=strtok_r(line,LTPT_CFG_SEP,&sr);
		if (!crv) { fclose(fp);LTPERR(LTP_FILE); }
		bool tagged=false;
		int tagind=-1;
		for (int j=0;!tagged && j!=LTPT_NUM_TAGS; j++) { 
			if (!strcasecmp(crv,ltpt_tags[j])) {
				crv=strtok_r(NULL,LTPT_CFG_SEP,&sr);
				if (!crv) continue;
				tagged=true;
				tagind=j;
			}
		}
		switch (tagind) {
			case 0: cfg->on=TF(crv[0]); break;
			case 1: cfg->relay_sockets=atoi(crv);break;
			case 2: cfg->include_est=TF(crv[0]); break;
			case 3: cfg->port=atoi(crv); break;
			case 4: cfg->hopcount=atoi(crv); break;
			case 5: cfg->include_via=TF(crv[0]); break;
			case 6: cfg->e2e_int.ciphersuite=ltp_crypt_mapciphersuite(crv);break;
			case 7: strncpy((char*)cfg->e2e_int.keyid,crv,min(MAXPATH-1,strlen(crv)));cfg->e2e_int.keyidlen=strlen(crv);break;
			case 8: strncpy((char*)cfg->map_file,crv,min(MAXPATH-1,strlen(crv)));break;
			case 9: cfg->need_e2e_int=TF(crv[0]);
			case 10: cfg->do_ecn=TF(crv[0]); break;
			default:
				snprintf(ltpstr,LTPSTR_LEN,"Oops - funny tag in LTP-T config. (tagind=%d,crv=%s)\n",tagind,crv);
				ltp_log(LTP_LOG_MAX,ltpstr);
				LTPERR(LTP_BAD_VALUE);
		}
	}
	fclose(fp);
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"LTP-T config:\n");
	ltp_log(LTP_LOG_MAX,ltpstr);
	snprintf(ltpstr,LTPSTR_LEN,"\ton=%s\n\trelays=%d\n\tinclude_est=%s\n\tport=%d\n\thpopcount=%d\n\tinclude_via=%s\n",
				TFP(cfg->on),cfg->relay_sockets,TFP(cfg->include_est),cfg->port,cfg->hopcount,TFP(cfg->include_via));
	ltp_log(LTP_LOG_MAX,ltpstr);
	snprintf(ltpstr,LTPSTR_LEN,"\te2e_int=(%d,%s)\n\te2e_map_file=%s\n\tneed_e2e_int=%s\n\tdo_ecn=%s\n",
				cfg->e2e_int.ciphersuite, cfg->e2e_int.keyid,cfg->map_file,TFP(cfg->need_e2e_int),TFP(cfg->do_ecn));
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	// sanity check values
	if (cfg->port<LTP_T_NO_PORT || cfg->port>LTP_T_MAX_PORT) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: out of range (%d,%d) port requested (%d) omitting port entirely\n",
					__FUNCTION__,LTP_T_NO_PORT, LTP_T_MAX_PORT, cfg->port);
		ltp_log(LTP_LOG_MAX,ltpstr);
		cfg->port=LTP_T_NO_PORT;
	}
	if (cfg->hopcount<LTP_T_NO_HOPCOUNT || cfg->hopcount>LTP_T_MAX_HOPCOUNT) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: out of range (%d,%d) hopcount requested (%d) omitting hopcount entirely\n",
					__FUNCTION__,LTP_T_NO_HOPCOUNT, LTP_T_MAX_HOPCOUNT, cfg->hopcount);
		ltp_log(LTP_LOG_MAX,ltpstr);
		cfg->hopcount=LTP_T_NO_HOPCOUNT;
	}
	if (cfg->relay_sockets<=0 || cfg->relay_sockets>LTP_T_MAX_RELAY_SOCKETS) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: out of range (%d,%d) relay_sockets requested (%d) setting default (%d)\n",
					__FUNCTION__,0,LTP_T_MAX_RELAY_SOCKETS, cfg->relay_sockets, LTP_T_DEF_RELAY_SOCKETS);
		ltp_log(LTP_LOG_MAX,ltpstr);
		cfg->relay_sockets=LTP_T_DEF_RELAY_SOCKETS;
	}
	return(LTP_OK);
}


/*!
 * @brief estimate the length of stuff extension encodings take up
 * @param block is the block concerned
 * @param sock is the socket
 * @return the estimeate
 */
int ltpt_extensions_estimate(ltp_block *block, ltp_socket_t *sock)
{
	if (!block || !sock) LTPERR(LTP_BAD_PARAM);
	int accum=0;
	int rv;
	ltp_t_cfg_so *cfg=&sock->opts.t_cfg;
	if (!cfg->on && !block->t_vals.flags) return(0);

	// source
	if (block->t_vals.flags & LTPT_SET_SOURCE) {
		accum+=2; // tag, len
		accum+=sizeof(block->t_vals.source);
	} else {
		ltpaddr my_addr;
		rv=ltp_cue_whoami(&my_addr);
		if (rv) {
			// oops - not much I can do now (no error return) but I'll crap out in a minute anyway so just log it
			snprintf(ltpstr,LTPSTR_LEN,"%s: ltp_cue_whoami failed (rv=%d)\n",
					__FUNCTION__,rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			return(0);
		}
		if (my_addr.octets.magic==LTP_ADDR_OCT_MAGIC) {
			accum+= 2+my_addr.octets.len;
		} else {
			accum+=6;
		}
	}

	if (block->t_vals.flags & LTPT_SET_DEST) {
		accum+=2; // tag, len
		accum+=sizeof(block->t_vals.dest);
	} else {
		int peerlen=0;
		ltpaddr *peer=NULL;
		if (sock->outbound) {
			peerlen=sock->obpeerlen;
			peer=&sock->obpeer;
		} else { 
			peerlen=sock->ibpeerlen;
			peer=&sock->ibpeer;
		}
		if (peer->octets.magic==LTP_ADDR_OCT_MAGIC) {
			accum+= 2+peer->octets.len;
		} else {
			accum+=6;
		}
	}

	if (sock->opts.t_cfg.include_est || (block->t_vals.flags & LTPT_SET_EBS) ) {
		accum += 6; // maybe overestimate but easy enough
	}
	if (sock->opts.t_cfg.port!=LTP_T_NO_PORT || (block->t_vals.flags & LTPT_SET_PORT)) {
		accum += 6; // maybe overestimate but easy enough
	}
	if (sock->opts.t_cfg.hopcount!=LTP_T_NO_HOPCOUNT || (block->t_vals.flags & LTPT_SET_HOP_COUNT)) {
		accum += 3;  // exact
	}
	if (sock->opts.t_cfg.include_via || (block->t_vals.flags & LTPT_SET_VIA)) {
		accum += 2; // usual
		accum += block->t_vals.via.len;
		accum += 59; // the "departure" string we'll add in a bit
	} else if (sock->opts.t_cfg.include_via) {
		accum += 2; // revisit this when via is filling up - but start out empty
	}
	if ((sock->opts.t_cfg.e2e_int.ciphersuite!=LTP_CS_NONE) || (block->t_vals.flags & LTPT_SET_E2E_AUTH)) {
		if (block->t_vals.flags & LTPT_SET_E2E_AUTH) {
			accum+=2+block->t_vals.e2e_a_kid.len+block->t_vals.e2e_a_val.len+6;
		} else {
			accum+=2+sock->opts.t_cfg.e2e_int.keyidlen;
			if (sock->opts.t_cfg.e2e_int.ciphersuite==LTP_CS_SIG) {
				accum+=6+128;
			} else {
				accum+=2+20;
			}
		}
	}
	return(accum);
}

/*!
 * @brief add basic LTP-T related extensions
 * @param seg is the ltpc_seg
 * @param block is the block
 * @param sock is the socket
 * @return zero or error
 *
 * This could be more efficient if I remembered the calculated values for earlier segments
 * of the block. But I don't care about that for now. And maybe this way will get more
 * flexibility later (though I doubt it). Certainly the code's easier this way.
 */
static int add_new_ltpt_exts(ltpc_seg *seg, ltp_block *block, ltp_socket_t *sock)
{
	int nexts;
	int rv;
	// add source ext
	{
		ltpaddr my_addr;
		rv=ltp_cue_whoami(&my_addr);
		if (rv) LTPERR(rv);
		ltpe_src *source=(ltpe_src*)ltpc_alloc(sizeof(ltpe_src));
		if (!source) LTPERR(LTP_MEMORY);
		// has to be a deep copy - stuff can get moved
		source->src=my_addr; // struct (actually union) copy
		if (my_addr.octets.magic==LTP_ADDR_OCT_MAGIC) {
			// deep copy octets value
			unsigned char *vcp=ltpc_alloc(my_addr.octets.len);
			if (!vcp) LTPERR(LTP_MEMORY);
			memcpy(vcp,my_addr.octets.val,my_addr.octets.len);
			source->src.octets.val=vcp;
		}
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_SOURCE;
		seg->exts[nexts-1].val=(void*)source;
		seg->exts[nexts-1].len=sizeof(*source);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}
	// add dest ext
	if (!(block->t_vals.flags & LTPT_SET_DEST)) {
		ltpe_dest *dst=(ltpe_dest*)ltpc_alloc(sizeof(ltpe_dest));
		if (!dst) LTPERR(LTP_MEMORY);
		// has to be a deep copy - stuff can get moved
		dst->dest=sock->obpeer; // struct (actually union) copy
		if (sock->obpeer.octets.magic==LTP_ADDR_OCT_MAGIC) {
			// deep copy octets value
			unsigned char *vcp=ltpc_alloc(sock->obpeer.octets.len);
			if (!vcp) LTPERR(LTP_MEMORY);
			memcpy(vcp,sock->obpeer.octets.val,sock->obpeer.octets.len);
			dst->dest.octets.val=vcp;
		}
		block->t_vals.dest=sock->obpeer;
		block->t_vals.flags != LTPT_SET_DEST;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_DEST;
		seg->exts[nexts-1].val=(void*)dst;
		seg->exts[nexts-1].len=sizeof(*dst);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	} else {
		ltpe_dest *dst=(ltpe_dest*)ltpc_alloc(sizeof(ltpe_dest));
		if (!dst) LTPERR(LTP_MEMORY);
		// has to be a deep copy - stuff can get moved
		dst->dest=block->t_vals.dest; // struct (actually union) copy
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_DEST;
		seg->exts[nexts-1].val=(void*)dst;
		seg->exts[nexts-1].len=sizeof(*dst);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}

	// maybe est block sie
	if (sock->opts.t_cfg.include_est) {
		ltpe_ebs *ebs=(ltpe_ebs*)ltpc_alloc(sizeof(ltpe_ebs));
		if (!ebs) LTPERR(LTP_MEMORY);
		ebs->ebs=block->decoded.len;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_EBS;
		seg->exts[nexts-1].val=(void*)ebs;
		seg->exts[nexts-1].len=sizeof(*ebs);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}
	// maybe port
	if (sock->opts.t_cfg.port!=LTP_T_NO_PORT) {
		ltpe_port *port=(ltpe_port*)ltpc_alloc(sizeof(ltpe_port));
		if (!port) LTPERR(LTP_MEMORY);
		port->port.len=sizeof(int);
		port->port.val=ltpc_alloc(port->port.len);
		if (!port->port.val) LTPERR(LTP_MEMORY);
		rv=ltpc_small_htonl(sock->opts.t_cfg.port,&port->port.len,port->port.val);
		if (rv) LTPERR(rv);
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_PORT;
		seg->exts[nexts-1].val=(void*)port;
		seg->exts[nexts-1].len=sizeof(*port);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}
	// maybe hop count
	if (sock->opts.t_cfg.hopcount!=LTP_T_NO_HOPCOUNT) {
		ltpe_hop_count *hop_count=(ltpe_hop_count*)ltpc_alloc(sizeof(ltpe_hop_count));
		if (!hop_count) LTPERR(LTP_MEMORY);
		hop_count->hop_count=sock->opts.t_cfg.hopcount;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_HOP_COUNT;
		seg->exts[nexts-1].val=(void*)hop_count;
		seg->exts[nexts-1].len=sizeof(*hop_count);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}
	// maybe via - initially empty
	if (sock->opts.t_cfg.include_via) {
		ltpe_via *via=(ltpe_via*)ltpc_alloc(sizeof(ltpe_via));
		if (!via) LTPERR(LTP_MEMORY);
		via->via.len=0; via->via.val=NULL; 
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_VIA;
		seg->exts[nexts-1].val=(void*)via;
		seg->exts[nexts-1].len=sizeof(*via);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}
	// this guy better be last
	if (sock->opts.t_cfg.e2e_int.ciphersuite!=LTP_CS_NONE) {

		if (block->t_vals.flags &  LTPT_SET_E2E_AUTH) {

			ltpe_e2eauth *auth=(ltpe_e2eauth*)ltpc_alloc(sizeof(ltpe_e2eauth));
			if (!auth) LTPERR(LTP_MEMORY);
			auth->ciphersuite=block->t_vals.e2e_a_cs;
			auth->keyid.len=block->t_vals.e2e_a_kid.len;
			auth->keyid.val=ltpc_alloc(block->t_vals.e2e_a_kid.len);
			if (!auth->keyid.val) LTPERR(LTP_MEMORY);
			memcpy(auth->keyid.val,
					block->t_vals.e2e_a_kid.val,
					block->t_vals.e2e_a_kid.len);
			auth->authval.len=block->t_vals.e2e_a_val.len;
			auth->authval.val=ltpc_alloc(auth->authval.len);
			if (!auth->authval.val) LTPERR(LTP_MEMORY);
			memcpy( auth->authval.val,
					block->t_vals.e2e_a_val.val,
					block->t_vals.e2e_a_val.len);
			nexts=add_ext(seg);
			seg->exts[nexts-1].type=LTPTE_AUTH;
			seg->exts[nexts-1].val=(void*)auth;
			seg->exts[nexts-1].len=sizeof(*auth);
			seg->exts[nexts-1].header=true;
			seg->exts[nexts-1].enclen=0;
			seg->exts[nexts-1].encval=NULL;

		} else { // new sig/mac needed
			// load a map file etc 
			if (!crypt_inited) {
				rv=ltp_crypt_setup(sock->opts.t_cfg.map_file);
				if (rv) LTPERR(rv);
				crypt_inited=true;
			}
			// better add an e2e-auth ext
	#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Adding an e2e-auth extension\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
	#endif
			// c14n 
			ltp_octets	c14n_in; 	// c14n'd buffer that's input to signing/macing
			int inclexts=1; 		// figure it out in a mo
			ltp_octets	src_in; 	// encoded source
			ltp_octets	dest_in; 	// encoded dest
			// these are optional, if not include len=0;
			ltp_octets	ebs_in;		// ebs
			ltp_octets	port_in;	// port
			ebs_in.len=0;
			port_in.len=0;
			ltpe_dest *dst=NULL;
			for (int i=0;i!=seg->nexts;i++) {
				if (seg->exts[i].type==LTPTE_SOURCE) {
					rv=ltpe_encode(&seg->exts[i]);
					if (rv) LTPERR(rv);
					src_in.len=seg->exts[i].enclen;
					src_in.val=seg->exts[i].encval;
					inclexts++;
				}
				if (seg->exts[i].type==LTPTE_DEST) {
					rv=ltpe_encode(&seg->exts[i]);
					if (rv) LTPERR(rv);
					dest_in.len=seg->exts[i].enclen;
					dest_in.val=seg->exts[i].encval;
					dst=(ltpe_dest*)seg->exts[i].val;
					inclexts++;
				}
				if (seg->exts[i].type==LTPTE_EBS) {
					rv=ltpe_encode(&seg->exts[i]);
					if (rv) LTPERR(rv);
					ebs_in.len=seg->exts[i].enclen;
					ebs_in.val=seg->exts[i].encval;
					inclexts++;
				}
				if (seg->exts[i].type==LTPTE_PORT) {
					rv=ltpe_encode(&seg->exts[i]);
					if (rv) LTPERR(rv);
					port_in.len=seg->exts[i].enclen;
					port_in.val=seg->exts[i].encval;
					inclexts++;
				}
			}
			// add those all up
			if (!src_in.len || !dest_in.len) LTPERR(LTP_BAD_VALUE);
			ltp_octets encblocklen; // sdnv encoded block length
#define EBV_SIZE	10
			unsigned char ebv[EBV_SIZE]; 
			encblocklen.len=EBV_SIZE;
			encblocklen.val=ebv;
			rv=sdnv_enc_int(block->decoded.len,(int*)&encblocklen.len,ebv);
			if (rv) LTPERR(rv);
			c14n_in.len= 2 + sock->opts.t_cfg.e2e_int.keyidlen + 
							src_in.len + dest_in.len + ebs_in.len + port_in.len + 
							encblocklen.len + block->decoded.len;
			// ok this is inefficient if the block is big
			// to avoid the inefficiency, would have to allocated empty space before the block when I copy it
			// below ltp_sendto, but that'd be difficult, so maybe later
			c14n_in.val=ltpc_alloc(c14n_in.len);
			if (!c14n_in.val) LTPERR(LTP_MEMORY);
			c14n_in.val[0]=(unsigned char)inclexts; 
			c14n_in.val[1]=sock->opts.t_cfg.e2e_int.ciphersuite; 
			int offset=2; 
			if (sock->opts.t_cfg.e2e_int.keyidlen) {
				memcpy(c14n_in.val+offset,
						sock->opts.t_cfg.e2e_int.keyid,
						sock->opts.t_cfg.e2e_int.keyidlen);
				offset+=sock->opts.t_cfg.e2e_int.keyidlen;
			}
			memcpy(c14n_in.val+offset,src_in.val,src_in.len); offset += src_in.len;
			memcpy(c14n_in.val+offset,dest_in.val,dest_in.len); offset += dest_in.len;
			if (ebs_in.len) memcpy(c14n_in.val+offset,ebs_in.val,ebs_in.len); offset += ebs_in.len;
			if (port_in.len) memcpy(c14n_in.val+offset,port_in.val,port_in.len); offset += port_in.len;
			memcpy(c14n_in.val+offset,encblocklen.val,encblocklen.len); offset += encblocklen.len;
			memcpy(c14n_in.val+offset,block->decoded.val,block->decoded.len); offset += block->decoded.len;
			// sign/mac
			unsigned char *authbuf=ltpc_alloc(LTP_AUTHBUFSIZE);
			if (!authbuf) LTPERR(LTP_MEMORY);
			int authlen=LTP_AUTHBUFSIZE;
			ltpc_octets o_dest;
			rv=ltpaddr2octets(&dst->dest,&o_dest);
			if (rv) LTPERR(rv);
			o_dest.len=4; // nix any port number for now!!
#ifdef LTP_EXTRA_EXTRA_VERBOSE
			// print input to checking
			snprintf(ltpstr,LTPSTR_LEN,"c14n'd block %d to be signed (%d bytes)\n",block->id,c14n_in.len);
			ltp_log_buffer(ltpstr,c14n_in.len,c14n_in.val);
#endif
			rv=ltp_crypt_do(&sock->opts.t_cfg.e2e_int,&o_dest,c14n_in.len,c14n_in.val,&authlen,&authbuf);
			if (rv) LTPERR(rv);
			ltpc_free(o_dest.val);
			ltpc_free(c14n_in.val);
			// make up extension value
			ltpe_e2eauth *auth=(ltpe_e2eauth*)ltpc_alloc(sizeof(ltpe_e2eauth));
			if (!auth) LTPERR(LTP_MEMORY);
			auth->ciphersuite=sock->opts.t_cfg.e2e_int.ciphersuite;
			auth->keyid.len=sock->opts.t_cfg.e2e_int.keyidlen;
			auth->keyid.val=NULL;
			if (auth->keyid.len) {
				auth->keyid.val=ltpc_alloc(auth->keyid.len);
				if (!auth->keyid.val) LTPERR(LTP_MEMORY);
				memcpy(auth->keyid.val,sock->opts.t_cfg.e2e_int.keyid,auth->keyid.len);
			}
			auth->authval.val=authbuf;
			auth->authval.len=authlen;

			nexts=add_ext(seg);
			seg->exts[nexts-1].type=LTPTE_AUTH;
			seg->exts[nexts-1].val=(void*)auth;
			seg->exts[nexts-1].len=sizeof(*auth);
			seg->exts[nexts-1].header=true;
			seg->exts[nexts-1].enclen=0;
			seg->exts[nexts-1].encval=NULL;
			// remember all this now
			block->t_vals.flags |= LTPT_SET_E2E_AUTH;
			block->t_vals.e2e_a_cs=sock->opts.t_cfg.e2e_int.ciphersuite;
			block->t_vals.e2e_a_kid.len=sock->opts.t_cfg.e2e_int.keyidlen; // shallow copy ok? nope!
			block->t_vals.e2e_a_kid.val=ltpc_alloc(block->t_vals.e2e_a_kid.len);
			if (! block->t_vals.e2e_a_kid.val) LTPERR(LTP_MEMORY);
			memcpy(block->t_vals.e2e_a_kid.val,
					sock->opts.t_cfg.e2e_int.keyid,
					block->t_vals.e2e_a_kid.len);
			block->t_vals.e2e_a_val.len=auth->authval.len; // shallow copy ok? nope!
			block->t_vals.e2e_a_val.val=ltpc_alloc(auth->authval.len);
			if (!block->t_vals.e2e_a_val.val) LTPERR(LTP_MEMORY);
			memcpy(block->t_vals.e2e_a_val.val,
					auth->authval.val,
					block->t_vals.e2e_a_val.len);

		}
	}
#if 0
// I'll do these later on
LTPTE_CONFID
LTPTE_ECN
#endif
	return(LTP_OK);
}

/*!
 * @brief add basic LTP-T related extensions
 * @param seg is the ltpc_seg
 * @param block is the block
 * @param sock is the socket
 * @return zero or error
 */
int ltpt_add_departure_exts(ltpc_seg *seg, ltp_block *block, ltp_socket_t *sock)
{
	// base our work on the current value of block->t_vals;
	if (!seg || !block || !sock) LTPERR(LTP_BAD_PARAM);
	int nexts=0;
	int rv;
	if (seg->type==LTPC_SEG_DS && block->t_vals.flags & LTPT_SET_VIA) {
		// this was a bug, should do nothing now
		// add deprature info & off it goes - leave blank value alone
		int newlen=block->t_vals.via.len;
		char added[MAXPATH];
		ltpaddr my_addr;
		int rv=ltp_cue_whoami(&my_addr);
		if (rv) LTPERR(rv);
		snprintf(added,MAXPATH,"\nVIA: %s:%d at %s (departure)",
			inet_ntoa(my_addr.sock.sin_addr),my_addr.sock.sin_port,ltp_now_str());
		newlen+=strlen(added);
		ltp_octets newvia;
		newvia.len=newlen;
		newvia.val=ltpc_alloc(newlen);
		if (!newvia.val) LTPERR(LTP_MEMORY);
		memcpy(newvia.val,block->t_vals.via.val,block->t_vals.via.len);
		memcpy(newvia.val+block->t_vals.via.len-1,added,strlen(added));
		newvia.val[newvia.len-1]='\0';
		ltpe_via *via=(ltpe_via*)ltpc_alloc(sizeof(ltpe_via));
		if (!via) LTPERR(LTP_MEMORY);
		via->via.len=newvia.len;
		via->via.val=newvia.val;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_VIA;
		seg->exts[nexts-1].val=(void*)via;
		seg->exts[nexts-1].len=sizeof(*via);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"LTP-T: via processing outbound seg: via was: %s, and I'm adding %s\n",
								block->t_vals.via.val,added);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	}
	return(LTP_OK);
}

/*!
 * @brief add basic LTP-T related extensions
 * @param seg is the ltpc_seg
 * @param block is the block
 * @param sock is the socket
 * @return zero or error
 */
static int add_fwd_ltpt_exts(ltpc_seg *seg, ltp_block *block, ltp_socket_t *sock)
{
	// base our work on the current value of block->t_vals;
	if (!seg || !block || !sock) LTPERR(LTP_BAD_PARAM);
	int nexts=0;
	int rv;

	if (block->t_vals.flags & LTPT_SET_SOURCE) {
		ltpe_src *src=(ltpe_src*)ltpc_alloc(sizeof(ltpe_src));
		if (!src) LTPERR(LTP_MEMORY);
		src->src=block->t_vals.source;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_SOURCE;
		seg->exts[nexts-1].val=(void*)src;
		seg->exts[nexts-1].len=sizeof(*src);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}

	if (block->t_vals.flags & LTPT_SET_DEST) {
		ltpe_dest *dest=(ltpe_dest*)ltpc_alloc(sizeof(ltpe_dest));
		if (!dest) LTPERR(LTP_MEMORY);
		dest->dest=block->t_vals.dest;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_DEST;
		seg->exts[nexts-1].val=(void*)dest;
		seg->exts[nexts-1].len=sizeof(*dest);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}

	if (block->t_vals.flags & LTPT_SET_PORT) {
		ltpe_port *port=(ltpe_port*)ltpc_alloc(sizeof(ltpe_port));
		if (!port) LTPERR(LTP_MEMORY);
		port->port.len=sizeof(int);
		port->port.val=ltpc_alloc(port->port.len);
		if (!port->port.val) LTPERR(LTP_MEMORY);
		rv=ltpc_small_htonl( block->t_vals.port, &port->port.len, port->port.val);
		if (rv) LTPERR(rv);
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_PORT;
		seg->exts[nexts-1].val=(void*)port;
		seg->exts[nexts-1].len=sizeof(*port);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}

	if (block->t_vals.flags & LTPT_SET_EBS) {
		ltpe_ebs *ebs=(ltpe_ebs*)ltpc_alloc(sizeof(ltpe_ebs));
		if (!ebs) LTPERR(LTP_MEMORY);
		ebs->ebs=block->t_vals.ebs;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_EBS;
		seg->exts[nexts-1].val=(void*)ebs;
		seg->exts[nexts-1].len=sizeof(*ebs);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}

	if (block->t_vals.flags & LTPT_SET_HOP_COUNT) {
		ltpe_hop_count *hop_count=(ltpe_hop_count*)ltpc_alloc(sizeof(ltpe_hop_count));
		if (!hop_count) LTPERR(LTP_MEMORY);
		hop_count->hop_count=block->t_vals.hop_count-1; // don't decrement the block value cause its for >1 seg
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_HOP_COUNT;
		seg->exts[nexts-1].val=(void*)hop_count;
		seg->exts[nexts-1].len=sizeof(*hop_count);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
	}

	if (block->t_vals.flags & LTPT_SET_VIA) {
#ifdef DEPONARRIVAL
		// this was a bug, should do nothing now
		// add deprature info & off it goes - leave blank value alone
		int newlen=block->t_vals.via.len;
		char added[MAXPATH];
		ltpaddr my_addr;
		int rv=ltp_cue_whoami(&my_addr);
		if (rv) LTPERR(rv);
		snprintf(added,MAXPATH,"\nVIA: %s:%d at %s (departure)",
			inet_ntoa(my_addr.sock.sin_addr),my_addr.sock.sin_port,ltp_now_str());
		newlen+=strlen(added);
		ltp_octets newvia;
		newvia.len=newlen;
		newvia.val=ltpc_alloc(newlen);
		if (!newvia.val) LTPERR(LTP_MEMORY);
		memcpy(newvia.val,block->t_vals.via.val,block->t_vals.via.len);
		memcpy(newvia.val+block->t_vals.via.len-1,added,strlen(added));
		newvia.val[newvia.len-1]='\0';
		ltpe_via *via=(ltpe_via*)ltpc_alloc(sizeof(ltpe_via));
		if (!via) LTPERR(LTP_MEMORY);
		via->via.len=newvia.len;
		via->via.val=newvia.val;
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_VIA;
		seg->exts[nexts-1].val=(void*)via;
		seg->exts[nexts-1].len=sizeof(*via);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"LTP-T: via processing outbound seg: via was: %s, and I'm adding %s\n",
								block->t_vals.via.val,added);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
#endif
	}

	if (block->t_vals.flags & LTPT_SET_E2E_AUTH) {
		// just fwd it
		ltpe_e2eauth *auth=(ltpe_e2eauth*)ltpc_alloc(sizeof(ltpe_e2eauth));
		if (!auth) LTPERR(LTP_MEMORY);
		auth->ciphersuite=block->t_vals.e2e_a_cs;
		auth->keyid.len=block->t_vals.e2e_a_kid.len;
		auth->keyid.val=ltpc_alloc(block->t_vals.e2e_a_kid.len);
		if (!auth->keyid.val) LTPERR(LTP_MEMORY);
		memcpy(auth->keyid.val,
				block->t_vals.e2e_a_kid.val,
				block->t_vals.e2e_a_kid.len);
		auth->authval.len=block->t_vals.e2e_a_val.len;
		auth->authval.val=ltpc_alloc(auth->authval.len);
		if (!auth->authval.val) LTPERR(LTP_MEMORY);
		memcpy( auth->authval.val,
				block->t_vals.e2e_a_val.val,
				block->t_vals.e2e_a_val.len);
		nexts=add_ext(seg);
		seg->exts[nexts-1].type=LTPTE_AUTH;
		seg->exts[nexts-1].val=(void*)auth;
		seg->exts[nexts-1].len=sizeof(*auth);
		seg->exts[nexts-1].header=true;
		seg->exts[nexts-1].enclen=0;
		seg->exts[nexts-1].encval=NULL;
#ifdef LTP_EXTRA_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"LTP-T: e2e auth added to outbound seg\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	}

	return(LTP_OK);
}

/*!
 * @brief add basic LTP-T related extensions
 * @param seg is the ltpc_seg
 * @param block is the block
 * @param sock is the socket
 * @return zero or error
 */
int add_ltpt_exts(ltpc_seg *seg, ltp_block *block, ltp_socket_t *sock)
{
	if (!sock) LTPERR(LTP_BAD_PARAM);
	if (!sock->inbound && sock->outbound) return add_new_ltpt_exts(seg,block,sock); 
	if (sock->inbound && sock->outbound) return add_fwd_ltpt_exts(seg,block,sock);
	return(LTP_BAD_VALUE);
}

/*!
 * @brief figure out whether this segment is an LTP-T one or not
 * @param seg is the segment
 * @param bool true if the segment has any LTP-T extensions
 */
bool is_ltpt_seg(ltpc_seg *seg)
{
	if (!seg) LTPERR(LTP_BAD_INPUT);
	if (!seg->nexts) {
		return (false);
	}
	if (!seg->exts) { // nexts says there are some - they better be there
		return (false);
	}
	for (int i=0;i!=seg->nexts;i++) {
		if (seg->exts[i].type==LTPTE_SOURCE) { return(true); }
		if (seg->exts[i].type==LTPTE_DEST) { return(true); }
		if (seg->exts[i].type==LTPTE_EBS) { return(true); }
		if (seg->exts[i].type==LTPTE_PORT) { return(true); }
		if (seg->exts[i].type==LTPTE_HOP_COUNT) { return(true); }
		if (seg->exts[i].type==LTPTE_VIA) { return(true); }
#if 0
// I'll do these later on
LTPTE_AUTH	 	
LTPTE_CONFID
LTPTE_ECN
#endif
	}
	return(false);
}

/*!
 * @brief figure out whether this LTP-T segment is from me 
 * @param seg is the segment
 * @return true if the segment is for me
 */
bool is_ltpt_fromme(ltpc_seg *seg)
{
	bool fromme=false;
	ltpaddr my_addr;
	int rv=ltp_cue_whoami(&my_addr);
	if (rv) LTPERR(rv);
	int	port=-1;
	ltpe_src srccp,*src=NULL;
	int dlen=0;
	for (int i=0;!src && port==-1 && i!=seg->nexts;i++) {
		if (seg->exts[i].type==LTPTE_SOURCE) { // ignore 2nd and subsequent DEST extensions
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Got an LTPTE_SOURCE\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			src=(ltpe_src*)seg->exts[i].val;
			dlen=seg->exts[i].len;
			if (!src) LTPERR(LTP_BAD_VALUE);
		}
		if (seg->exts[i].type==LTPTE_PORT) { // ignore 2nd and subsequent PORT extensions
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Got an LTPTE_PORT\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			ltpe_port *portp=(ltpe_port*)seg->exts[i].val;
			if (!portp) LTPERR(LTP_MEMORY);
			if (!portp->port.len) port=0;
			else if (portp->port.len==1) port=portp->port.val[0];
			else if (portp->port.len==2) port=(portp->port.val[0]<<8)+portp->port.val[1];
			else LTPERR(LTP_BAD_VALUE);
		}
	}
	if (port==-1) port=UDP_DEF_PORT;
	if (src) {
		srccp=*src;
		if (srccp.src.octets.magic!=LTP_ADDR_OCT_MAGIC) {
			if (!srccp.src.sock.sin_port) {
				srccp.src.sock.sin_port=port;
			}
			srccp.src.sock.sin_family=AF_LTP;
		}
		if (!ltpaddr_cmp(&srccp.src,&my_addr,dlen)) {
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"It's from me\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			fromme=true;
			return fromme;
		} else {
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"It's NOT from me\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			fromme=false;
			return fromme;
		}
	}
	return false; // cannot happen!
}

/*!
 * @brief figure out whether this LTP-T segment is for me (LTP peer= LTP-T dest)
 * @param seg is the segment
 * @return true if the segment is for me
 */
bool is_ltpt_forme(ltpc_seg *seg)
{
	bool forme=false;
	ltpaddr my_addr;
	int rv=ltp_cue_whoami(&my_addr);
	if (rv) LTPERR(rv);
	int	port=-1;
	ltpe_dest dstcp,*dst=NULL;
	int dlen=0;
	for (int i=0;!dst && port==-1 && i!=seg->nexts;i++) {
		if (seg->exts[i].type==LTPTE_DEST) { // ignore 2nd and subsequent DEST extensions
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Got an LTPTE_DEST\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			dst=(ltpe_dest*)seg->exts[i].val;
			dlen=seg->exts[i].len;
			if (!dst) LTPERR(LTP_BAD_VALUE);
		}
		if (seg->exts[i].type==LTPTE_PORT) { // ignore 2nd and subsequent PORT extensions
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Got an LTPTE_PORT\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			ltpe_port *portp=(ltpe_port*)seg->exts[i].val;
			if (!portp) LTPERR(LTP_MEMORY);
			if (!portp->port.len) port=0;
			else if (portp->port.len==1) port=portp->port.val[0];
			else if (portp->port.len==2) port=(portp->port.val[0]<<8)+portp->port.val[1];
			else LTPERR(LTP_BAD_VALUE);
		}
	}
	if (port==-1) port=UDP_DEF_PORT;
	if (dst) {
		dstcp=*dst;
		if (dstcp.dest.octets.magic!=LTP_ADDR_OCT_MAGIC) {
			if (!dstcp.dest.sock.sin_port) {
				dstcp.dest.sock.sin_port=port;
			}
			dstcp.dest.sock.sin_family=AF_LTP;
		}
		if (!ltpaddr_cmp(&dstcp.dest,&my_addr,dlen)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"It's for me!\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			forme=true;
			return forme;
		} else {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"It's NOT for me - will try forward\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			forme=false;
			return forme;
		}
	}
	return false; // cannot happen!
}

/*!
 * @brief do the LTP-T processing for this just received segment
 * @param segstr is the segment structure
 * @param theblock is the parent block (it can be NULL)
 * @param thesock is the parent socket (can it be NULL - not sure?)
 * @param badseg is an output - true if this segments cookies are awry in some way  
 * @param inlen is the length of the encoded segment
 * @return zero or error
 */
int ltpt_inbound_processing(ltpc_seg *seg, ltp_block *theblock, ltp_socket_t *sock, bool *badseg, int inlen)
{
	// first question - is there any LTP-T stuff there?
	// redundant (caller checks) but do it again anyway:-)
	bool ltpt_seg=is_ltpt_seg(seg);
	if (!ltpt_seg) return(LTP_OK);
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Got an LTP-T segment!\n");
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif

	for (int i=0;i!=seg->nexts;i++) {
		// ignore 2nd and subsequent values ??
		// source 
		if (seg->exts[i].type==LTPTE_SOURCE) {
			ltpe_src *source=(ltpe_src*)seg->exts[i].val;
			if (source) {
				theblock->t_vals.flags|=LTPT_SET_SOURCE;
				theblock->t_vals.source=source->src;
			}
		}
		// DEST 
		if (seg->exts[i].type==LTPTE_DEST) { 
			ltpe_dest *dst=(ltpe_dest*)seg->exts[i].val;
			if (dst) {
				theblock->t_vals.flags|=LTPT_SET_DEST;
				theblock->t_vals.dest=dst->dest;
			}
		}
		// ebs - nowt - only need when ECN
		if (seg->exts[i].type==LTPTE_EBS) {
			ltpe_ebs *ebs=(ltpe_ebs*)seg->exts[i].val;
			if (ebs) {
				theblock->t_vals.flags|=LTPT_SET_EBS;
				theblock->t_vals.ebs=ebs->ebs;
			}
		}
		// port 
		if (seg->exts[i].type==LTPTE_PORT) {
			ltpe_port *port=(ltpe_port*)seg->exts[i].val;
			if (port) {
				theblock->t_vals.flags|=LTPT_SET_PORT;
				unsigned int foo=0;
				if (port->port.len==1) foo=port->port.val[0];
				else if (port->port.len==2) foo=(port->port.val[0]<<8)+port->port.val[1];
				else {
					snprintf(ltpstr,LTPSTR_LEN,"Got an odd port length (%d) using value of zero now\n",port->port.len);
					ltp_log(LTP_LOG_MAX,ltpstr);
					foo=0;
				}
				theblock->t_vals.port=foo;
			}
		}
		// hop_count - maybe ditch
		if (seg->exts[i].type==LTPTE_HOP_COUNT) {
			ltpe_hop_count *hc=(ltpe_hop_count*)seg->exts[i].val;
			if (hc) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Got an hop_count of %d\n",hc->hop_count);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				// don't decrement here but just bounce if necessary
				if (!hc->hop_count) {
					*badseg=true;
#ifdef LTP_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"\tHop count says ditch that seg\n");
					ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				} else {
					theblock->t_vals.flags|=LTPT_SET_HOP_COUNT;
					theblock->t_vals.hop_count=hc->hop_count;
				}
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"\tHop count has null value - ditch that seg\n");
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				*badseg=true;
			}
		}
		// via - not nowt
		if (seg->exts[i].type==LTPTE_VIA) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Got a VIA\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			// maybe have to decode ?
			ltpe_via *via=(ltpe_via*)seg->exts[i].val;
			int newlen=0;
			if (via) { // note via can be NULL if I'm the 1st router
				newlen=via->via.len;
			} else { 
				newlen=1; // for the "NUL" character, even if we won't send it 
			}
			int viacplen=newlen-1; // how many bytes of old thing to copy
			char added[MAXPATH];
			ltpaddr my_addr;
			int rv=ltp_cue_whoami(&my_addr);
			if (rv) LTPERR(rv);
			snprintf(added,MAXPATH,"\nVIA: %s:%d at %s (arrival)",
				inet_ntoa(my_addr.sock.sin_addr),my_addr.sock.sin_port,ltp_now_str());
			newlen+=strlen(added);
			ltp_octets newvia;
			newvia.len=newlen;
			newvia.val=ltpc_alloc(newlen);
			if (!newvia.val) LTPERR(LTP_MEMORY);
			if (viacplen) {
				memcpy(newvia.val,via->via.val,via->via.len);
			}
			memcpy(newvia.val+viacplen,added,strlen(added));
			newvia.val[newvia.len-1]='\0';
			if (theblock->t_vals.via.val) ltpc_free(theblock->t_vals.via.val);
			theblock->t_vals.via=newvia; // struct cp
			theblock->t_vals.flags|=LTPT_SET_VIA;
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"LTP-T: via processing inbound seg: via was: %s, and I'm adding %s\n",
								(viacplen?(char*)via->via.val:"null"),added);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		}
		// if we've not seen an e2e auth before...
		if (seg->exts[i].type==LTPTE_AUTH && !(theblock->t_vals.flags&LTPT_SET_E2E_AUTH)) {
			// copy to block is all really
			ltpe_e2eauth *auth=(ltpe_e2eauth*)seg->exts[i].val;
			if (!auth) LTPERR(LTP_MEMORY);
			theblock->t_vals.e2e_a_cs=auth->ciphersuite;
			theblock->t_vals.e2e_a_kid.len=auth->keyid.len;
			if (auth->keyid.len)  {
				theblock->t_vals.e2e_a_kid.val=ltpc_alloc(theblock->t_vals.e2e_a_kid.len);
				if (!theblock->t_vals.e2e_a_kid.val) LTPERR(LTP_MEMORY);
				memcpy(
					theblock->t_vals.e2e_a_kid.val,
					auth->keyid.val,
					theblock->t_vals.e2e_a_kid.len);
			} else theblock->t_vals.e2e_a_kid.val=NULL;
			theblock->t_vals.e2e_a_val.len=auth->authval.len;
			theblock->t_vals.e2e_a_val.val=ltpc_alloc(auth->authval.len);
			if (!theblock->t_vals.e2e_a_val.val) LTPERR(LTP_MEMORY);
			memcpy( 
					theblock->t_vals.e2e_a_val.val,
					auth->authval.val,
					theblock->t_vals.e2e_a_val.len);
			theblock->t_vals.flags|=LTPT_SET_E2E_AUTH;
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"LTP-T: e2e_auth seen on inbound seg");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		}
		// ecn - TODO LATER
		// confid - nowt and LATER
	}
	sock->outbound=true;  // magic!! tell the guy to send the packets too!
	sock->finishedwith=false; // might have to mark it more later
	theblock->t_vals.forwarded=false; // there's work to do

	return(LTP_OK);
}


/*!
 * @brief final LTP-T processing before delivery
 * @param theblock is the parent block (it can be NULL)
 * @param thesock is the parent socket (can it be NULL - not sure?)
 * @param status returns whether things look good or not
 * @param status_str explains why not, if not
 * @param sstr_len is the size of the status_str buffer
 * @return zero or error
 */
int ltpt_final_processing(ltp_block *block, ltp_socket_t *sock, bool *status, char *status_str, int sstr_len)
{
	*status=false; // pessimist!
	int rv;
	int inclexts=1;
	bool ebs_present=false;
	bool port_present=false;
	bool src_present=false;
	bool dest_present=false;
	ltpaddr *src_in=NULL;
	ltpaddr *dest_in=NULL;
	int *ebs_in=NULL;
	int *port_in=NULL;

	snprintf(status_str,sstr_len,"%s","Dunno yet!");

	if (block->t_vals.flags & LTPT_SET_SOURCE) {
		src_present=true;
		src_in=&block->t_vals.source;
		inclexts++;
	}
	if (block->t_vals.flags & LTPT_SET_DEST) {
		dest_present=true;
		dest_in=&block->t_vals.dest;
		inclexts++;
	}
	if (block->t_vals.flags & LTPT_SET_EBS) {
		ebs_present=true;
		ebs_in=&block->t_vals.ebs;
		inclexts++;
	}
	if (block->t_vals.flags & LTPT_SET_PORT) {
		port_present=true;
		port_in=&block->t_vals.port;
		inclexts++;
	}

	if (!src_present || !dest_present) {
		snprintf(status_str,sstr_len,"%s","No source or dest present.");
		LTPERR(LTP_BAD_VALUE);
	}

	if (block->t_vals.flags & LTPT_SET_E2E_AUTH) {
		// check that sig
		// load a map file etc 
		if (!crypt_inited) {
			rv=ltp_crypt_setup(sock->opts.t_cfg.map_file);
			if (rv) {
				snprintf(status_str,sstr_len,"%s","ltp_crypt_setup failed");
				LTPERR(rv);
			}
			crypt_inited=true;
		}
		ltp_auth_so ai_so;
		ai_so.ciphersuite=block->t_vals.e2e_a_cs;
		ai_so.keyidlen=block->t_vals.e2e_a_kid.len; 
		memcpy(ai_so.keyid,block->t_vals.e2e_a_kid.val,ai_so.keyidlen);
		// c14n the inputs - PITA but has to be done
		// ciphersuite
		// keyid or zero byte
		int keyidlen=0;
		if (ai_so.keyidlen) {
			keyidlen=ai_so.keyidlen;
		} else  {
			keyidlen=1;
		}
		// sdnv-encoded block len
		int sdnv_blen_len=MAXPATH;
		unsigned char sdnv_blen[MAXPATH];
		rv=sdnv_enc_int(block->decoded.len,(int*)&sdnv_blen_len,sdnv_blen);
		if (rv) {
			snprintf(status_str,sstr_len,"%s","sdnv_enc_int 1 failed");
			LTPERR(rv);
		}
		// various extensions
		ltpc_octets src_buf;
		ltpc_octets dest_buf;
		ltpc_octets ebs_buf;
		ltpc_octets port_buf;
		rv=ltpaddr2octets(src_in,&src_buf);
		if (rv) {
			snprintf(status_str,sstr_len,"%s","ltpaddr2octets(src) failed");
			LTPERR(rv);
		}
		rv=ltpaddr2octets(dest_in,&dest_buf);
		if (rv) {
			snprintf(status_str,sstr_len,"%s","ltpaddr2octets(dest) failed");
			LTPERR(rv);
		}
		if (ebs_present) {
			// re-encode it, think this is non-leaking, but I may be wrong!
			ltpc_ext tmpext;
			ltpe_ebs lebs;
			lebs.ebs=*ebs_in;
			tmpext.type=LTPTE_EBS;
			tmpext.val=(void*)&lebs;
			tmpext.len=sizeof(lebs);
			tmpext.header=true; tmpext.enclen=0;tmpext.encval=NULL;
			rv=ltpe_encode(&tmpext);
			ebs_buf.len=tmpext.enclen;
			ebs_buf.val=tmpext.encval;
		} else {
			ebs_buf.len=0;
			ebs_buf.val=NULL;
		}
		if (port_present) {
			// re-encode it, think this is non-leaking, but I may be wrong!
			ltpc_ext tmpext;
			ltpe_port lport;
			lport.port.len=sizeof(int);
			lport.port.val=ltpc_alloc(lport.port.len);
			if (!lport.port.val) {
				snprintf(status_str,sstr_len,"%s","alloc failed 1");
				LTPERR(LTP_MEMORY);
			}
			rv=ltpc_small_htonl(*port_in,&lport.port.len,lport.port.val);
			if (!lport.port.val) {
				snprintf(status_str,sstr_len,"%s","ltpc_small_htonl failed 1");
				LTPERR(LTP_MEMORY);
			}
			tmpext.type=LTPTE_PORT;
			tmpext.header=true; tmpext.enclen=0;tmpext.encval=NULL;
			tmpext.val=(void*)&lport;
			tmpext.len=sizeof(lport);
			rv=ltpe_encode(&tmpext);
			port_buf.len=tmpext.enclen;
			port_buf.val=tmpext.encval;
			ltpc_free(lport.port.val);
		} else {
			port_buf.len=0;
			port_buf.val=NULL;
		}
		// bytes of block
		int auth_input_len= 2 + keyidlen +
		 	src_buf.len + dest_buf.len + ebs_buf.len + port_buf.len
				+ sdnv_blen_len + block->decoded.len;
		unsigned char *auth_input=ltpc_alloc(auth_input_len);
		if (!auth_input) {
			snprintf(status_str,sstr_len,"%s","alloc failed 2");
			LTPERR(LTP_MEMORY);
		}
		auth_input[0]=inclexts;
		auth_input[1]=ai_so.ciphersuite;
		int offset=2;
		if (ai_so.keyidlen) {
			memcpy(auth_input+offset,ai_so.keyid,ai_so.keyidlen);
			offset+=ai_so.keyidlen;
		} else {
			auth_input[2]=0x00;
			offset=3;
		}
		// src
		memcpy(auth_input+offset,src_buf.val,src_buf.len);
		offset += src_buf.len;
		ltpc_free(src_buf.val);
		// dest
		memcpy(auth_input+offset,dest_buf.val,dest_buf.len);
		offset += dest_buf.len;
		ltpc_free(dest_buf.val);
		// ebs ??
		if (ebs_present) {
			memcpy(auth_input+offset,ebs_buf.val,ebs_buf.len);
			offset += ebs_buf.len;
			ltpc_free(ebs_buf.val);
		}
		// port ??
		if (port_present) {
			memcpy(auth_input+offset,port_buf.val,port_buf.len);
			offset += port_buf.len;
			ltpc_free(port_buf.val);
		}
		memcpy(auth_input+offset,sdnv_blen,sdnv_blen_len);
		offset+=sdnv_blen_len;
		memcpy(auth_input+offset,block->decoded.val,block->decoded.len);
		// now do crypto stuff
		unsigned char *auth_val=block->t_vals.e2e_a_val.val;
		int auth_val_len=block->t_vals.e2e_a_val.len;
		bool auth_res=false;
		ltpc_octets o_src;
		rv=ltpaddr2octets(&sock->ibpeer,&o_src);
		if (rv) {
			snprintf(status_str,sstr_len,"%s","ltpaddr2octets failed 2");
			LTPERR(rv);
		}
		o_src.len=4; // just IP address, no port for now
#ifdef LTP_EXTRA_EXTRA_VERBOSE
		// print input to checking
		snprintf(ltpstr,LTPSTR_LEN,"c14n'd block %d to be verified (%d bytes)\n",block->id,auth_input_len);
		ltp_log_buffer(ltpstr,auth_input_len,auth_input);
#endif
		rv=ltp_crypt_check(&ai_so,&o_src,auth_input_len,auth_input,auth_val_len,auth_val,&auth_res);
		ltpc_free(o_src.val);
		ltpc_free(auth_input);
		if (rv) {
			snprintf(status_str,sstr_len,"%s","ltpc_free failed");
			LTPERR(rv); 
		}
		if (auth_res) { // yippee!!
			*status=true;
		} else { // bummer
			*status=false;
		}
	}
	// guess we made it!
	*status=true;
	return(LTP_OK);
}

/*!
 * @brief load LTP-T routes
 * @param fname is the filename to use
 * @param nroutes is a ptr to an integer on output containing the number of routes
 * @param routes is a ptr to an array of ltp_t_route_so's allocated inside this function
 * @return zero or error
 */
int ltpt_load_routes(char *fname, int *nroutes,ltp_t_route_so **routes)
{
	if (!nroutes || !routes || !fname) LTPERR(LTP_BAD_INPUT);
	int rv=0;
	FILE *fp=fopen(fname,"r");
	if (!fp) LTPERR(LTP_FILE);
	// usual parsing 
	// parse file, line by line
	char line[MAXPATH];
	char *fgrv=NULL;
	char *lp;
	int sofar=0;
	ltp_t_route_so *rsofar=NULL;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"LTP-T routes:\n");
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	while ((fgrv=fgets(line,MAXPATH,fp))!=NULL) {
		if (strlen(line) > 1023) return(LTP_BAD_VALUE); /// line too long
		lp=line;
		while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
		if (!strlen(lp)) continue; 
		if (*lp=='#') continue;
		if (*lp=='\n') continue;
		// parse - get tag
		char *sr1;
		char *nws=strtok_r(line,LTPT_ROUTE_SEP,&sr1);
		char *gws=strtok_r(NULL,LTPT_ROUTE_SEP,&sr1);
		if (!nws) { fclose(fp);LTPERR(LTP_FILE); }
		nws=strdup(nws);
		if (!gws) { free(nws); fclose(fp);LTPERR(LTP_FILE); }
		char *cp=ltp_name_trans(nws); if (cp) { free(nws); nws=strdup(cp); } 
		gws=strdup(gws);
		cp=ltp_name_trans(gws); if (cp) { free(gws); gws=strdup(cp); } 
		sofar++;
		rsofar=(ltp_t_route_so*)ltpc_realloc((unsigned char*)rsofar,sofar*sizeof(ltp_t_route_so));
		if (!rsofar) { fclose(fp); LTPERR(LTP_MEMORY); }
		snprintf(rsofar[sofar-1].network,LTP_T_ROUTE_SLEN,"%s",nws);
		snprintf(rsofar[sofar-1].gateway,LTP_T_ROUTE_SLEN,"%s",gws);
		free(nws); free(gws);
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"\tnw: %s, gw: %s\n",
			rsofar[sofar-1].network, rsofar[sofar-1].gateway);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	}
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"That's a total of %d routes (zero being acceptable!)\n",sofar);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	fclose(fp);
	*nroutes=sofar;
	*routes=rsofar;
	return(LTP_OK);
}

/*!
 * @brief figure out the LTP-T gateway to use for an LTP-T address
 * @param dest is the destination
 * @param gw is the gateway to use (called allocated)
 * @param nlisteners is the number of current contacts
 * @param listeners are the current contacts
 * @param matched is true (on return) if some route matches, otherwise false
 * @return zero or error
 */
int ltpt_gateway_listeners(ltpaddr *dest, ltpaddr *gw, int nlisteners, ltpaddr *listeners, bool *matched)
{
	// maybe change this later to prefer the actual destination over an indirect gateway
	if (!matched || !dest || !gw) LTPERR(LTP_BAD_PARAM);
	bool foundone=false;
	if (!nlisteners) {
		// nothing to do!
		*matched=false;
		*gw=*dest;

#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"ltpt_gateway_listeners: no listeners at all!\n"); 
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		return(LTP_OK);
	}
	ltpaddr foob=*dest;
	ltpaddr my_addr;
	int rv=ltp_cue_whoami(&my_addr);
	if (rv) LTPERR(rv);
	// first if one of the listeners is the destination, go with that
	// (later some metrics might determine otherwise, but not now)
	for (int i=0;i!=nlisteners;i++) {
		if (!isany(&listeners[i],sizeof(listeners[i])) &&
			!ltpaddr_cmp(&listeners[i],dest,sizeof(*dest))) { // if the dest is a listener then pick it!
			*gw=*dest;
			*matched=true;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"ltpt_gateway_listeners: matched listener (%d) with dest.\n",i);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			return(LTP_OK);
		}
	}
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"ltpt_gateway_listeners: no listener (out of %d) match dest.\n",nlisteners);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	// somewhere in the code there's another of these before match_char_ip is called, so 
	// now I need one here too!
	// foob.sock.sin_addr.s_addr=htonl(dest->sock.sin_addr.s_addr);
	// used to be - now I don't want it:-(
	foob.sock.sin_addr.s_addr=dest->sock.sin_addr.s_addr;
	int found_ind_r=-1;
	int found_ind_l=-1;
	for (int i=rt_size-1; found_ind_l==-1 && i>=0;i--) { // routes are added to the end so start there - most recent first match
		if (match_char_ip(strlen(rtab[i].route.network),rtab[i].route.network,sizeof(in_addr),(void*)&foob.sock.sin_addr)) {
			ltpaddr gw_addr;
			int rv=str2ltpaddr(rtab[i].route.gateway,&gw_addr);
			if (rv) LTPERR(rv);
			if (ltpaddr_cmp(&my_addr,&gw_addr,sizeof(gw_addr))) { // if I'm not the gateway in question, then keep looking
				// is the gw a listener?
				for (int j=0;j!=nlisteners;j++) {
					if (!ltpaddr_cmp(&listeners[j],&gw_addr,sizeof(gw_addr))) { // if the g/w is a listener then pick it!
						found_ind_l=j;
						found_ind_r=i;
						foundone=true;
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"ltpt_gateway_listeners: matched gw (%d) with listener (%d)\n",i,j);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					}
				}
			}
		}
	}
	if (found_ind_l==-1) {
		// just shallow copy dest to gw - i.e. try go direct
		*gw=*dest;
		if (!rt_size) foundone=true; // if no routes exist then we should send direct (what else is possible?)
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"ltpt_gateway_listeners: no listener (out of %d) matched a gw.\n",nlisteners);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else {
		// the gw cannot be a wildcard dude, need a concrete addr
		// see which is best (hopefully one is)
		if (isany(&listeners[found_ind_l],sizeof(ltpaddr))) {
			gw->sock.sin_family=AF_LTP;
			char *stin=strdup(rtab[found_ind_r].route.gateway);
			int rv=str2ltpaddr(stin,gw);
			if (rv) LTPERR(rv);
			if (isany(gw,sizeof(ltpaddr))) { /// crap - both were
				snprintf(ltpstr,LTPSTR_LEN,"Gateway code picked dumb gateway at %s:%d\n\tListener: %s, Route: %s\n",
						__FILE__,__LINE__,ltpaddr2str(&listeners[found_ind_l]),stin);
				ltp_log(LTP_LOG_MAX,ltpstr);
				return(LTP_BAD_VALUE);
			} 
			free(stin);
		} else {
			memset(gw,0,sizeof(ltpaddr));
			memcpy(gw,&listeners[found_ind_l],sizeof(ltpaddr));
		}

	}
	*matched=foundone;
	return(LTP_OK);
}

/*!
 * @brief figure out the LTP-T gateway to use for an LTP-T address
 * @param dest is the destination
 * @param gw is the gateway to use (called allocated)
 * @return zero or error
 */
int ltpt_gateway(ltpaddr *dest, ltpaddr *gw)
{
	if (!dest || !gw) LTPERR(LTP_BAD_PARAM);
	ltpaddr foob=*dest;
	ltpaddr my_addr;
	int rv=ltp_cue_whoami(&my_addr);
	if (rv) LTPERR(rv);
	// somewhere in the code there's another of these before match_char_ip is called, so 
	// now I need one here too!
	// foob.sock.sin_addr.s_addr=htonl(dest->sock.sin_addr.s_addr);
	// used to be - now I don't want it:-(
	foob.sock.sin_addr.s_addr=dest->sock.sin_addr.s_addr;
	int found_ind=-1;
	for (int i=rt_size-1;found_ind==-1 && i>=0;i--) { // routes are added to the end so start there - most recent first match
		if (match_char_ip(strlen(rtab[i].route.network),rtab[i].route.network,sizeof(in_addr),(void*)&foob.sock.sin_addr)) {
			ltpaddr gw_addr;
			int rv=str2ltpaddr(rtab[i].route.gateway,&gw_addr);
			if (rv) LTPERR(rv);
			if (ltpaddr_cmp(&my_addr,&gw_addr,sizeof(gw_addr))) { // if I'm not the gateway in question, then we've a route
				found_ind=i;
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"ltpt_gateway: gw %d matched.\n",i);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			}
		}
	}
	if (found_ind==-1) {
		// just shallow copy dest to gw - i.e. try go direct
		*gw=*dest;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"ltpt_gateway: no gw matched.\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else {
		memset(gw,0,sizeof(ltpaddr));
		gw->sock.sin_family=AF_LTP;
		char *stin=strdup(rtab[found_ind].route.gateway);
		int rv=str2ltpaddr(stin,gw);
		if (rv) LTPERR(rv);
		free(stin);
	}
	return(LTP_OK);
}


/*!
 * @brief add a route to the local set
 * @param route is a ptr to an ltp_t_route_so
 * @return zero or error
 */
int ltpt_set_route(ltp_t_route_so *route)
{
	if (!route) LTPERR(LTP_BAD_PARAM);
	rt_size ++;
	rtab=(route_table*)ltpc_realloc((unsigned char*)rtab,rt_size*sizeof(route_table));
	if (!rtab) LTPERR(LTP_MEMORY);
	// rtab[rt_size-1].route=*route;
	snprintf(rtab[rt_size-1].route.network,LTP_T_ROUTE_SLEN,"%s",route->network);
	snprintf(rtab[rt_size-1].route.gateway,LTP_T_ROUTE_SLEN,"%s",route->gateway);
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"%d-th route added (nw=%s,gw=%s)\n",rt_size,route->network,route->gateway);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	return(LTP_OK);
}

/*!
 * @brief return a printable version of the local routes
 * @return a ptr to a static char * with route info
 */
char *ltpt_print_routes(void)
{
	if (!rt_size) {
		snprintf(routestr,ROUTESTRLEN,"No info.");
		return(routestr);
	}
	char *cp=routestr;
	int cpl=ROUTESTRLEN;
	for (int i=0;cpl>0 && i!=rt_size;i++) {
		snprintf(cp,cpl,"nw: %s, gw: %s\n\t",rtab[i].route.network,rtab[i].route.gateway);
		cp=routestr+strlen(routestr);
		cpl=ROUTESTRLEN-strlen(routestr);
	}
	return(routestr);
}
