/*!
 * @file ltp_l2_udp.cc
 * @brief These are functions that our l2 needs to do UDP
 * @version $Revision: 1.12 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * Main idea is just to keep the socket stuff together here.
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <unistd.h>
#include <net/if.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>

#include "ltp_base.h"
#include "ltp_int.h"
#include "ltp.h"
#include "ltp_l2.h"


// I love global vars:-)

/// Transmit socket descriptor
static int	t_sockfd=0;
/// Receive socket descriptor
static int	r_sockfd=0;
/// Receive file descriptor set (for select)
static fd_set rfds;
/// Port on which we listen for things
static int 	r_port=UDP_DEF_PORT;

/// Although this is called <<blah>>_udp.cc and the fncs are
/// all called l2_[c|s]_udp_*, I'm gonna try it with SOCK_RAW
/// instead of SOCK_DGRAM and see if that works better (a few
/// less headers! Tried it. Didn't work. Back to UDP!
#define REALLY_UDP
#ifdef REALLY_UDP
#define MYSOCKTYPE SOCK_DGRAM
#else
#define MYSOCKTYPE SOCK_RAW
#endif

/*!
 * @fn int l2_s_udp_init(int port)
 * @brief initialise a listener on some port
 * @param port is the place to listen:-)
 * @param mcp is NULL or (should be) a multicast group address
 * @return zero or error
 */
int l2_s_udp_init(int port, sockaddr_in *mcp) 
{
	r_port=port;
	if ((r_sockfd=socket(AF_INET,MYSOCKTYPE,0))==-1) {
		return(-1);
	}
	struct sockaddr_in Target;
	bzero ((unsigned char*)&Target,sizeof(Target));
	Target.sin_family=AF_INET;
	Target.sin_addr.s_addr=htonl(INADDR_ANY);
	if (port) r_port=port;
	Target.sin_port=htons(r_port);

#ifdef LTP_MULTICAST
    if (mcp) {
        // should be a MC group, higher layers checked, but let's be anal:-)
        ltpaddr foo;
        foo.sock=*mcp;
        if (!ltp_is_multi(&foo)) {
            // set to listen to MC group
            struct ip_mreq bar;
            bar.imr_multiaddr=mcp->sin_addr;
            in_addr local;
            local.s_addr=htonl(INADDR_ANY);
            bar.imr_interface=local;
            int sorv=setsockopt(r_sockfd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&bar, sizeof(bar));
            if (sorv<0) return(sorv);
        }
        
    }
#endif

	if (bind(r_sockfd,(struct sockaddr*)&Target,sizeof(Target))==-1) {
		return(-2);
	}
	FD_ZERO(&rfds);
	FD_SET(r_sockfd,&rfds);
	int bufsz=1500*200; // 100 is 12ms worth and should be what we want!!!
	socklen_t	szlen=sizeof(bufsz);
	int rv=setsockopt(r_sockfd,SOL_SOCKET,SO_RCVBUF,(void*)&bufsz,szlen);
	if (rv) return(rv);
	return(LTP_OK);
}

/// for now I can't get the select code to work on cygwin or suse (sigh...)
#undef SELECT_WORKS

#ifndef SELECT_WORKS

/// Length of local buffer to hang onto rx'd packet (while select isn't working!).
/// If this is less than the MTU, we're quite wasteful and may even loose data,
/// so I should fix that later (move lbuf to the heap, alloc'ing at least an
/// MTU). However, since we're effectively hardcoding MTU for now, there's no
/// point just yet.
#define LBUFSIZE 4096

/// Actual length of data in local buffer to hang onto rx'd packet (while select isn't working!)
static int lbuflen=0;

/// Local buffer to hang onto rx'd packet (while select isn't working!)
static unsigned char lbuf[LBUFSIZE];

/// Local structure to hang on't rx'd packet source (while slect isn't working)
static struct sockaddr_in lSource;

#endif

/*!
 * @fn int l2_s_udp_arrival(void)
 * @brief check if some bytes have arrived there
 * @return true if something's waiting
 *
 * We're going to wait forever in select until something arrives. That
 * ought be ok since we should be inside a listener thread here.
 */
int l2_s_udp_arrival() 
{
#ifdef SELECT_WORKS
	struct timeval tv;
	tv.tv_sec=0;
	tv.tv_usec=50000; // 1/20-th of a s
	int rv=select(1,&rfds,NULL,NULL,&tv);
	// int rv=select(1,&rfds,NULL,NULL,NULL);
	return(rv);
#else
	bzero((unsigned char*)&lSource,sizeof(lSource));
	socklen_t ssize=sizeof(lSource);
	lbuflen=recvfrom(r_sockfd,(void*)lbuf,LBUFSIZE,0,(struct sockaddr*)&lSource, &ssize);
	return(lbuflen);
#endif
}

/*!
 * @fn int l2_s_udp_get_arrived_buf(ltp_l2_buffer *buf, ltp_l2_id *peer) 
 * @brief get the bytes that've just arrived
 * @param buf is the buffer into which they ought be put
 * @param peer is the guy who sent 'em to us (as far as we can tell)
 * @return zero or error
 *
 * Note that this function allocates nothing!!! You better copy the outputs
 * as soon as you can or they'll be overwritten. Luckily, that's just what
 * we do do in ltp_l2.cc's default callback handler!
 */
int l2_s_udp_get_arrived_buf(ltp_l2_buffer *buf, ltp_l2_id *peer) 
{
#ifdef SELECT_WORKS
	struct sockaddr_in Source;
	bzero((unsigned char*)&Source,sizeof(Source));

	socklen_t ssize=sizeof(Source);
	int rxd=recvfrom(r_sockfd,(void*)buf->val,buf->len,0,(struct sockaddr*)&Source, &ssize);
	/// @todo map Source to peer
#else
	buf->val=lbuf;
	buf->len=lbuflen;
	buf->heap=false; // don't want this free'd
	buf->l2alloc=true; // may ws well say that again :-)
	peer->heap=false;
	peer->l2alloc=false;
	peer->addr.sock=lSource;
	peer->addr_len=sizeof(lSource);
	peer->pname=NULL;
#endif
	return(LTP_OK);
}

/*!
 * @fn int l2_s_udp_close(void)
 * @brief finish up the listener
 * @return zero or error
 */
int l2_s_udp_close(void) 
{
	close(r_sockfd);
	r_sockfd=0;
	return(LTP_OK);
}


/*!
 * @fn int l2_c_udp_init(void) 
 * @brief initialise a UDP client to be able to send in a moment
 * @param port is the place to send to
 * @return zero or error 
 */
int l2_c_udp_init(void) 
{

// TXINONE means to setup, send and close the tx udp socket 
// each time we send. Otherwise, some combination with 
// scheduling means we're leaving UDP sockets all over the place
#define TXINONE
#ifndef TXINONE

	if ((t_sockfd=socket(AF_INET,MYSOCKTYPE,0))==-1) {
		return(-1);
	}
	struct sockaddr_in Source;
	bzero ((unsigned char*)&Source,sizeof(Source));
	Source.sin_family=AF_INET;
	Source.sin_addr.s_addr=htonl(INADDR_ANY);
	Source.sin_port=0;
	if (bind(t_sockfd,(struct sockaddr*)&Source,sizeof(Source))==-1) {
		return(-1);
	}
	int bufsz=1500*200; // 100 is 12ms worth and should be what we want!!!
	socklen_t	szlen=sizeof(bufsz);
	int rv=setsockopt(t_sockfd,SOL_SOCKET,SO_SNDBUF,&bufsz,szlen);
	if (rv) return(rv);

#endif

	return(LTP_OK);
}

/*!
 * @fn int l2_c_udp_send(ltp_l2_buffer *buf, ltp_l2_id *peer) 
 * @brief send some bytes to that guy
 * @param buf are the bytes
 * @param peer is the indended victim of those bytes
 * @return zero or error
 */
int l2_c_udp_send(ltp_l2_buffer *buf, ltp_l2_id *peer) 
{

#ifdef TXINONE
	if ((t_sockfd=socket(AF_INET,MYSOCKTYPE,0))==-1) {
		return(-1);
	}
	struct sockaddr_in Source;
	bzero ((unsigned char*)&Source,sizeof(Source));
	Source.sin_family=AF_INET;
	Source.sin_addr.s_addr=htonl(INADDR_ANY);
	Source.sin_port=0;
	if (bind(t_sockfd,(struct sockaddr*)&Source,sizeof(Source))==-1) {
		return(-1);
	}
	int bufsz=1500*200; // 100 is 12ms worth and should be what we want!!!
	socklen_t	szlen=sizeof(bufsz);
	int rv=setsockopt(t_sockfd,SOL_SOCKET,SO_SNDBUF,&bufsz,szlen);
	if (rv) return(rv);
#endif

	struct sockaddr_in Target;
	bzero((unsigned char*)&Target,sizeof(Target));
	Target.sin_family=AF_INET;
	Target.sin_port=htons(peer->addr.sock.sin_port);
	Target.sin_addr.s_addr=peer->addr.sock.sin_addr.s_addr;

#ifdef LTP_MULTICAST
// stephen@tolerantnetworks.com: if the destination adddress is a 
// multicast group, then set the TTL to 2 (for a one-router scope). This
// is currently hardcoded based on the address matching 224/4 (which
// means multicast). 
// to do this we just set a socket option
    if (!ltp_is_multi(&peer->addr)) {
        // set TTL to 2
        unsigned char ttl=2;
        int srv=setsockopt(t_sockfd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
#ifdef LTP_MC_DEBUG
        if (srv) {
		    snprintf(ltpstr,LTPSTR_LEN,"MC: error (%d) setting ttl to %d",srv,ttl);
        } else {
		    snprintf(ltpstr,LTPSTR_LEN,"MC: set ttl to %d",ttl);
        }
	    ltp_log(LTP_LOG_MAX,ltpstr);
#endif
    }
#endif

	int sent=sendto(t_sockfd,(void*)buf->val,buf->len,0,(struct sockaddr*)&Target, sizeof(Target));
	if (sent!=buf->len) {
		return (-1);
	}
#ifdef TXINONE
	close(t_sockfd);
	t_sockfd=0;
#endif
	return(LTP_OK);
}

/*!
 * @fn int l2_c_udp_close(void)
 * @brief finish up the client (nothing to do really)
 * @return zero or error
 */
int l2_c_udp_close(void)
{
#ifndef TXINONE
	close(t_sockfd);
	t_sockfd=0;
#endif
	return(LTP_OK);
}

/*!
 * @brief let the client caller know the current socket's port
 * @return the port
 */
int l2_c_sending_port(void)
{
	if (t_sockfd==0) return(-2);
	struct sockaddr_in addr;
	int alen=sizeof(addr);
	int rv=getsockname(t_sockfd,(struct sockaddr*)&addr,(socklen_t*)&alen);
	if (rv) return(rv);
	return addr.sin_port;
}

/*!
 * @fn int l2_get_mtu(name)
 * @param name is the interface name
 * @brief return our best guess at the MTU
 * @return is the guessed MTU or zero if we've no good guess!
 */
int l2_get_mtu(char *name)
{
	int sock2try=0; ///< the socket to try get an MTU option from (one of r_sockfd or t_sockfd)
	if (t_sockfd) sock2try=t_sockfd;
	if (!sock2try && r_sockfd) sock2try=r_sockfd;
	if (!sock2try) return(0);
	int foo=1500;
#undef USINGSOCKOPT
#define USINGIOCTL

#ifdef USINGIOCTL
#define WAYTOOBIG (1<<24)
	foo=WAYTOOBIG;
	struct ifreq ifr;
	if (name && name[0]!='\0') {
		strncpy(ifr.ifr_name, name, IFNAMSIZ);
		ioctl(sock2try, SIOCGIFMTU, &ifr);
		foo=ifr.ifr_mtu; 
	} else {
		memset(&ifr,0,sizeof(ifr));
		// see if we can do better than crapology
		ioctl(sock2try, SIOCGIFFLAGS, &ifr);
		if (ifr.ifr_flags & IFF_LOOPBACK) {
			const char *defloname="lo";
			strncpy(ifr.ifr_name, defloname, IFNAMSIZ);
			ioctl(sock2try, SIOCGIFMTU, &ifr);
			foo=ifr.ifr_mtu;
		} else {
			const char *defname="eth0";
			strncpy(ifr.ifr_name, defname, IFNAMSIZ);
			ioctl(sock2try, SIOCGIFMTU, &ifr);
			foo=ifr.ifr_mtu; 
		}
	}
#endif

#ifdef USINGSOCKOPT
	int sockmtu;
	// didn't get the right #include to get IP_MTU - gave up after 10 mins:-)
	int rv=getsockopt(sock2try,SOL_SOCKET,IP_MTU,(void*)&sockmtu,(socklen_t)sizeof(sockmtu));
	if (rv==-1) {
		foo=1500;
	} else {
		foo=sockmtu;
	}
	 
#endif

	foo -= 28; // take off UDP overhead at this point
	
	return(foo);
}


/*!
 * @brief return raw socket handles for debugging purposes
 * @param rx is a ptr that will contain the rx socket handle
 * @param tx is a ptr that will contain the rx socket handle
 * @return zero or error
 */
int l2_udp_status(int *rx, int *tx)
{
	if (!rx || !tx) LTPERR(LTP_BAD_PARAM);
	*rx=r_sockfd;
	*tx=t_sockfd;
	return(LTP_OK);
}
