/*!
 * @file ltpd.cc
 * @brief This is the LTP ltp_daemon
 * @version $Revision: 1.67 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This is my LTP/LTP-T client, sever and router.
 * See the usage() function for details.
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <getopt.h>
#include <unistd.h> // for sleep()
#include <errno.h> 
#include <pthread.h> 
#include <netdb.h> 
#include <sys/stat.h> 
#include <ctype.h> // for toupper()

#include "ltp_base.h"
#include "ltp.h"
#include "ltp_crypt.h"
#include "ltp_t.h"
#include "ltp_cues.h"

// local parameters 

/// whether to wait for a client to hit a character before exiting or not
static bool dowait=true;

/// the default time that a client will sleep if not waiting on a user keyboard hit
#define LTPD_DEF_SLEEPTIME 10

/// how long for a non-waiting client to sleep before exiting
static int  sleeptime=LTPD_DEF_SLEEPTIME;

/// whether to produce verbose output or not
static bool verbose=false;

/// macro to implement verbosity - careful with syntax here
#define MAYBEPRINT(__XxX__) if (verbose) printf("%s",__XxX__)

/// whether to run as a ltp_daemon or not
static bool ltp_daemon=false;

/// logfile name
char log_fname[MAXPATH];

/// default logfile name
#define LTPD_DEF_LOG "ltpd.log"

/// config file name
char ccfg_fname[MAXPATH];

/// default crypto cfg file name
#define LTPD_DEF_CCFG "ltpd.crypt"

/// default LTP-T config file name
#define LTPD_DEF_TCFG "ltpt.cfg"

/// the LTP-T configuration file name
char ltpt_cfg_fname[MAXPATH];

/// whether an LTP-T configuration file name was supplied on the command line
bool ltpt_cfg_given=false;

/// default LTP-T routing file name
#define LTPD_DEF_RCFG "ltpt.routes"

/// the LTP-T routes file name
char ltpt_routes_fname[MAXPATH];

/// whether an LTP-T configuration file name was supplied on the command line
bool ltpt_routes_given=false;

/// the input (to the program) file name
char inputfname[MAXPATH];

/// an input file name was given
static bool inputfnamegiven=false;

/// the default input file name
#define LTPD_DEF_INPUT "ltpd.in"

/// the output (from the program) file name
char outputfname[MAXPATH];

/// the default output file name
#define LTPD_DEF_OUTPUT "ltpd.out"

/// the typical size of an RX buffer
#define LTPD_DEF_RXBUF 65536

/// the size of buffer to allocate for data rx'd from network
static int rxbufsize=LTPD_DEF_RXBUF;

/// the various modes
#define LTPD_MODE_CLIENT 0 	///< act as an LTP client
#define LTPD_MODE_SERVER 1	///< act as an LTP responder/server
#define LTPD_MODE_ROUTER 2	///< act as an LTP-T router and as a responder/server
#define LTPD_MODE_BOTH	 3	///< act as both a client and server (for loopback testing)

/// the default mode
#define LTPD_DEF_MODE LTPD_MODE_CLIENT

/// the operational mode 
static int mode=LTPD_DEF_MODE;

/// the layer2 mtu
static int l2mtu=1472;

/// whether the layer2 mtu was set from command line
static bool l2mtu_set=false;

/// whether we're in file server mode or not
static bool fileserver=false;

/// if in fileserver mode and a client, then ask for this file
static char requestedfile[MAXPATH];

/// if a requestedfile was given
static bool requestedfilegiven=false;

/// the default filename to request as a client
#define LTPD_DEF_RF "ltpd.fetch"

// protocol parameters

/// whether or not cookies are turned on
static int cookies=0;

/// whether or not sockets block 
static bool blocking=false;

/// the default cookie grace period
#define LTPD_DEF_CGL 5

/// the cookie grace period
static int cookie_grace=LTPD_DEF_CGL;

/// whether a cooke grace was provided
static bool cookie_grace_given=false;

/// biggest port number allowed - not sure where this is properly so here's fine
#define MAXPORT 65535

/// the destination
static ltpaddr dest;

/// the source
static ltpaddr source;

/// whether a source is supplied
static bool source_given=false;

/// the listener
static ltpaddr listener;

/// the redlen setting - defaults to allred
static int redlen=LTP_ALLRED; 

/// the flags for sendto - not sure what to do with this yet
static int flags=0;

/// the authentication stuff
static ltp_auth_so	auth={ LTP_CS_NONE, 0, ""};

/// whether a ciphersuite was specified from the command line
static bool auth_ciphersuite_given=false;

/// whether a key id was specified from the command line
static bool auth_keyid_given=false;

/// whether to sync to/from disk 
static bool disksync=false;

/// whether to use very short numbers suitable for interoperatign with ION (space code:-)
static bool ion_mode=false;

/*!
 * @brief make it easy to specify a mode of operation
 * @param str is the optarg string
 * @return the mode or -1 for an error
 */
static int mapmode(char *str)
{
	// anything that starts
	if (!str) return(-1);
	switch (toupper(str[0])) {
		case 'C':
		case '0':
			return(LTPD_MODE_CLIENT);
		case 'S':
		case '1':
			return(LTPD_MODE_SERVER);
		case 'R':
		case '2':
			return(LTPD_MODE_ROUTER);
		case 'B':
		case '3':
			return(LTPD_MODE_BOTH);
		default:
			return(-1);
	}
	return (-1);  // can't happen
}

/*!
 * @brief map a string into a socket address 
 * @param str is the string (can be DNS or dot.sep. IP)
 * @param addr is the resuting address
 * @return zero or error
 */
static int mapaddr(char *str, ltpaddr *addr)
{
	// note - error handling a but tricky here - can't use ltp_log since we're called from
	// within the getopt() loop, so have to use stderr if necessary
	if (!str) {
		fprintf(stderr,"Null string ptr given to mapaddr - bad idea - exiting\n");
		exit(-1);
	}
	if (!addr) {
		fprintf(stderr,"Null addr ptr given to mapaddr - bad idea - exiting\n");
		exit(-1);
	}
	if (strlen(str)<=1) { 
		fprintf(stderr,"String (%s) too short for mapaddr - bad idea - exiting\n",str);
		exit(-1);
	}
	hostent	*hent;
	// was a port specified?
	char *cpstr=strdup(str);
	char *pstr=NULL;
	for (int i=strlen(cpstr)-1;i!=0;i--) {
		if (cpstr[i]==':') { 
			pstr=cpstr+i+1;
			cpstr[i]='\0';
			break;
		}
	}
	int port;
	if (pstr) {
		port=atoi(pstr);
	} else {
		port=UDP_DEF_PORT;
	}
	if (port<=0 || port>MAXPORT) {
		fprintf(stderr,"Requested port (%d) out of range (min=1,max=%d), host=%s\n",port,MAXPORT,cpstr);
		exit(-1);
	}
	hent=gethostbyname(cpstr);
	free(cpstr);
	if (hent==NULL) {
		fprintf(stderr,"gethostnyname(%s) returned NULL - exiting\n",str);
		exit(-1);
	}
	if (hent->h_name) {
		bzero(addr,sizeof(sockaddr_in));
		addr->sock.sin_family=AF_LTP;
		addr->sock.sin_port=port;
		memcpy(&addr->sock.sin_addr.s_addr,hent->h_addr_list[0],hent->h_length);
		if (verbose) {
			printf("Mapped %s to %s:%d\n",str,inet_ntoa(addr->sock.sin_addr),addr->sock.sin_port);
		}
		return(0);
	} else {
		fprintf(stderr,"gethostnyname(%s) returned an empty record - exiting\n",str);
		exit(-1);
	}
	return(-1); // can't happen
}

/*!
 * @brief guess what usage() does?
 */
static void usage(void)
{
	printf("ltpd is the LTP and LTP-T ltp_daemon.\n");
	printf("ltpd [-2 l2mtu] [-b buffersize] [-c crypto-cfg] [-d] [-g [fname]] [-i input] [-l log] [-m mode] [-o output] [-r routes-cfg] [-s] [-t LTP-T-cfg] [-v] [-w sleeptime]\n");
	printf("\t\t[-A ciphersuite] [-C cookielen] [-D dest[:port]] [-G cookie-grace-period] [-L listener[:port]] [-K keyid] [-S source[:port]] [-R redlen] [-I]\n");
	printf("Local parameters:\n");
	
	printf("\t[-2 l2mtu] sets the layer 2 maximum transmission unit (packet size) via a sockopt() call (min=%d,max=%d)\n",
					LTP_MINMTU,LTP_MAXMTU);
	printf("\t[-b buffersize] specifies the size of receive buffer for the application to allocate (def=%d)\n",rxbufsize);
	printf("\t[-c crypto-cfg] sets the crypto-configuration file name to crypto-cfg [%s]\n",LTPD_DEF_CCFG);
	printf("\t[-d] instructs ltpd to sync state information to disk and re-load state on start-up\n");
	printf("\t[-g [file]] instructs ltpd to act as a file server or to ask the server for a named file. [def: ltpd.fetch]\n");
	printf("\t[-i input] sets the input file name [%s]\n",LTPD_DEF_INPUT);
	printf("\t[-l log] sets the log file name [%s]\n",LTPD_DEF_LOG);
	printf("\t[-m mode] the operational mode - Client, Server or Router [Client]\n");
	printf("\t[-o output] sets the output file name(s) [%s]\n",LTPD_DEF_OUTPUT);
	printf("\t[-r routes-cfg] specifies where LTP-T routing configuration file [%s]\n",LTPD_DEF_RCFG);
	printf("\t[-s] instructs ltpd to spawn a process to run as a daemon - the \n\t\tinitial process exits leaving one behind [false]\n");
	printf("\t[-t [LTP-T-cfg]] use LTP-T & optionally specify configuration file to use (default: %s)\n",LTPD_DEF_TCFG);
	printf("\t[-v] requests verbose output (to stdio) [false]\n");
	printf("\t[-w [sleeptime]] when operating as a client, instructs ltpd to exit without waiting for the user to hit a character [%d]\n",
					LTPD_DEF_SLEEPTIME);
	printf("\t\t  - a value of \"b\" makes the sending socket block until the session is complete, i.e. \"-w b\"\n");
	printf("Protocol parameters:\n");
	printf("\t[-A ciphersuite] specifies which LTP-Auth ciphersuite to use [\"MAC\"/0, \"Signature\"/1, \"NULL\",255 ] default is none\n");
	printf("\t[-C len] turn on cookies of len bytes (cookies are off by default, same as len=0)\n");
	printf("\t[-D dest[:port]] sets the destination IP address or DNS name [%s]\n",ltpaddr2str(&dest));
	printf("\t[-G cookie-grace-period] specifies the number of seconds before cookies must be seen in responses [%d]\n",LTPD_DEF_CGL);
	printf("\t[-I] forces the stack to use very short numbers, suitable for interoperating with the ION LTP code [false]\n");
	printf("\t[-K keyid] specifies the key id to use if some LTP Auth stuff is wanted.\n");
	printf("\t[-L listener[:port]] sets the host IP address and more commonly, port on which to listen as a server\n");
	printf("\t[-R redlen] sets the red/green length (default:allred=%d;allgreen=%d,max=%d)\n",
					LTP_ALLRED,LTP_ALLGREEN,LTP_MAXRED);
	printf("\t[-S source[:port]] sets the source IP address or DNS name\n");
	printf("Notes:\n");
	printf("- The \"Router\" mode of operation is only meaningful for LTP-T of course.\n");
	printf("- Options using uppercase (e.g. D,G,S) relate to protocol parameters, \n  others are local\n");
	printf("- Protocol parameters all default to 127.0.0.1:1113\n");
	printf("- See the ltp extension document for ciphersuite descriptions but note that NULL is not \"off\" its a test mode\n");
	printf("- If you set a source (-S) while in server mode (-m S) then all responses will be forced to go to that address:port\n");
	exit(-1);
}

/*!
 * @brief sanity checking for parameters
 */
static void checkparams(void)
{
	if (mode<LTPD_MODE_CLIENT || mode >LTPD_MODE_BOTH) {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Bad mode value (%d) requested - using default (%d)\n",mode,LTPD_DEF_MODE);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		mode=LTPD_DEF_MODE;
	}
	if (cookie_grace_given && !cookies) {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: It seems silly to set a cookie grace periond with no cookies!\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	}
	if (auth.ciphersuite==LTP_CS_NONE || auth.ciphersuite==LTP_CS_NULL && auth_keyid_given) {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: It seems silly to set a keyid (%s) for the off or NULL ciphersuite (%d)!\n",
				auth.keyid, auth.ciphersuite);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	}
	return;
}

#define  LTPD_GETSTR_FRAG1 "<ltpdget file=\""
#define  LTPD_GETSTR_FRAG2 "\" size=\""
#define  LTPD_GETSTR_FRAG3 "\" />"
#define  LDPD_GETSTR_MIN  (2+strlen(LTPD_GETSTR_FRAG1)+strlen(LTPD_GETSTR_FRAG2)+strlen(LTPD_GETSTR_FRAG3))

/*!
 * @brief make up a fileserver request
 * @param inbuf is the request buffer
 * @param insize is the size of the request buffer
 * @param requestedFlie is the name of the requested file
 * @param retsize is the max buffer size expected in return
 * @return is the size of the formatted request or -1 for error
 */
static int make_request(char *inbuf,int insize,char *requestedfile,int retsize)
{
	snprintf(inbuf,insize,"%s%s%s%d%s\n", LTPD_GETSTR_FRAG1, requestedfile,
		LTPD_GETSTR_FRAG2, retsize, LTPD_GETSTR_FRAG3);
	return(strlen(inbuf));
}

/*!
 * @brief parse an incoming block to see if its a file server request
 * @param req is the request
 * @param reqlen is the request length
 * @param fname is the file requested
 * @param fnamelen is the file requested
 * @param delay is the length of time to wait before sending (seconds) NOT YET IMPLEMENTED
 * @param to is the entity to which the file should be sent NOT YET IMPLEMENTED (CAREFUL - DoS!!!)
 * @return zero or error
 */
static int parse_request(unsigned char *req, int reqlen, char *fname, int fnamelen, time_t *delay, ltpaddr *to)
{
#ifndef LTP_DODGY_FILESERVER
	// use this to make the dodginess visible in configure.ac 
	return(LTP_BAD_PARAM);
#endif
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: You have compiled ltpd as a DODGY FILE SERVER take a look at %s:%d\n", __FILE__,__LINE__);
	ltp_log(LTP_LOG_MAX,ltpstr);
	MAYBEPRINT(ltpstr);
	char *reqcp=(char*)req;
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: Request is %s\n",reqcp);
	ltp_log(LTP_LOG_MAX,ltpstr);
	MAYBEPRINT(ltpstr);
	// gonna be a tad anal about this! for good reasons
	if (strlen(reqcp) < LDPD_GETSTR_MIN) LTPERR(LTP_BAD_PARAM);
	if (strncmp(reqcp,LTPD_GETSTR_FRAG1,strlen(LTPD_GETSTR_FRAG1))) LTPERR(LTP_BAD_PARAM);
	reqcp+=strlen(LTPD_GETSTR_FRAG1);
	// next up to end quote should be file name
	char *cp=reqcp;
	// don't like "/" inside filenames - no messing with ../../../etc/passwd or similar
	// TODO this isn't properly handling i18n perhaps, but LATER
	while (*cp && *cp!='\"') {
		if (*cp=='/') LTPERR(LTP_BAD_VALUE);
		cp++;
	}
	char keeper=*cp;
	*cp='\0';
	snprintf(fname,fnamelen,"%s",reqcp);
	reqcp=cp;
	*cp=keeper;
	if (strncmp(reqcp,LTPD_GETSTR_FRAG2,strlen(LTPD_GETSTR_FRAG2))) LTPERR(LTP_BAD_PARAM);
	reqcp+=strlen(LTPD_GETSTR_FRAG2);
	int bufsize=atoi(reqcp);
	if (bufsize<=0) LTPERR(LTP_BAD_VALUE);
	// if (strncmp(reqcp,LTPD_GETSTR_FRAG3,strlen(LTPD_GETSTR_FRAG3))) LTPERR(LTP_BAD_PARAM);
	return(LTP_OK);
}

/*!
 * @brief read, parse and action config file settings via sockopts
 * @param sock is the socket to use
 * @return zero or error
 */
static int ltpd_sockopts(int sock)
{
	int rv;
	// open, read, parse cfg file
	// then call relevant sockopts on socket
	if (false && mode==LTPD_MODE_CLIENT && redlen==LTP_ALLGREEN) { // no need to listen
		int foo=0;
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_CLIENTONLY,&foo,sizeof(foo));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_CLIENTONLY: rv=%d\n", rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (redlen!=LTP_ALLRED) { // needn't bother since this is the default but it makes debugging easier... a bit
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_RED,&redlen,sizeof(redlen));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_RED: rv=%d,redlen=%d\n",rv,redlen);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (cookies) {
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_COOKIE,&cookies,sizeof(cookies));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_COOKIE: rv=%d,val=%d\n",rv,cookies);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (cookies && cookie_grace!=LTPD_DEF_CGL) {
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_COOKIE_GRACE,&cookie_grace,sizeof(cookie_grace));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_COOKIE_GRACE: rv=%d,val=%d\n",rv,cookie_grace);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (auth_ciphersuite_given || auth_keyid_given) {
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_AUTH,&auth,sizeof(auth));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_AUTH: rv=%d,val=(%d,%s)\n",
				rv,auth.ciphersuite,auth.keyid);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (l2mtu_set) {
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_L2MTU,&l2mtu,sizeof(l2mtu));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_L2MTU: rv=%d,val=%d\n",
				rv,l2mtu);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (ltpt_cfg_given) {
		ltp_t_cfg_so ltpt_cfg;
		memset(&ltpt_cfg,0,sizeof(ltp_t_cfg_so));
		// read the file
		rv=ltpt_load_cfg(ltpt_cfg_fname,&ltpt_cfg);
		if (rv) LTPERR(rv);
		// set the sockopt()
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_T_CFG,&ltpt_cfg,sizeof(ltp_t_cfg_so));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_T_CFG: rv=%d\n", rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (ltpt_routes_given) {
		// read each route from file and call setsocktop() for each
		int 			nroutes=0;
		ltp_t_route_so *routes;
		rv=ltpt_load_routes(ltpt_routes_fname,&nroutes,&routes);
		if (rv) LTPERR(rv);
		// LTP_SO_T_ROUTE
		for (int i=0;i!=nroutes;i++) {
			rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_T_ROUTE,&routes[i],sizeof(ltp_t_route_so));
			if (rv) {
				snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_T_ROUTE: rv=%d,route=%d,(%s,%s)\n", 
						rv,i,routes[i].network,routes[i].gateway);
				ltp_log(LTP_LOG_MAX,ltpstr);
				MAYBEPRINT(ltpstr);
			}
		}
	}
	if (blocking) {
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_BLOCK,&blocking,sizeof(blocking));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_BLOCK: rv=%d,val=%d\n",rv,blocking);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (mode==LTPD_MODE_SERVER && source_given) {
		// force the source:port to which LTP responses are to be sent
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_FORCE,&source,sizeof(source));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_FORCE: rv=%d\n",rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	if (ion_mode) {
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_ION,&ion_mode,sizeof(bool));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_ION: rv=%d,val=%d\n",rv,ion_mode);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}
	return(LTP_OK);
}

/*!
 * @brief do client stuff
 * @return zero or error
 */
static int ltpd_client(void)
{
	int rv;
	int sock;
	// only used if fileserver is set
	int retsock;
	int retsize=rxbufsize; // figure out what to do there
	unsigned char *retbuf=NULL;

	snprintf(ltpstr,LTPSTR_LEN,"LTPD: Starting %s\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
	
	unsigned char *inbuf=NULL;
	int insize=0;

	if (!fileserver) {
		// want to send something
		FILE *fp=fopen(inputfname,"r");
		if (!fp) { 
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: Can't open input file %s\n",inputfname);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
			LTPERR(LTP_FILE);
		}
		struct stat instat;
		rv=fstat(fileno(fp),&instat);
		if (rv) { fclose(fp); LTPERR(LTP_FILE);}
		insize=instat.st_size;
		if (!insize) { fclose(fp); LTPERR(LTP_FILE);}
		inbuf=(unsigned char*)calloc(insize,1);
		if (!inbuf)  { fclose(fp); LTPERR(LTP_MEMORY);}
		rv=fread(inbuf,1,insize,fp);
		if (rv!=insize) { fclose(fp); free(inbuf); LTPERR(LTP_FILE);}
		fclose(fp);
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Read %d bytes from %s OK\n",insize,inputfname);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	} else {
		// make up request
		insize=strlen(requestedfile)+100;
		inbuf=(unsigned char*)calloc(insize,1);
		if (!inbuf) LTPERR(LTP_MEMORY);
		rv=make_request((char*)inbuf,insize,requestedfile,retsize);
		if (rv==-1) {
			LTPERR(LTP_BAD_PARAM);
		}
		insize=rv;
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Request to server is for file %s (request size=%d)\n",requestedfile,insize);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		// better have a listener for the answer I guess?
	}

	// setup crypto maps if necessary
	if (auth_ciphersuite_given) {
		rv=ltp_crypt_setup(ccfg_fname);
		if (rv) LTPERR(rv);
	}

	sock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
	if (!sock) LTPERR(LTP_SOCKET);

	// use configuration stuff
	rv=ltpd_sockopts(sock);
	if (rv) LTPERR(rv);

	if (blocking && redlen!=LTP_ALLGREEN) {
		// wait for the last RA to be sent before terminating
		int foo=1;
		rv=ltp_setsockopt(sock,SOL_SOCKET,LTP_SO_LINGER,&foo,sizeof(foo));
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_LINGER: rv=%d\n", rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}

	// setup source addr
	rv=ltp_bind(sock,(ltpaddr*)&source,sizeof(source));
	if (rv) { free(inbuf); ltp_close(sock); LTPERR(LTP_SOCKET);}

	// need to prepare a socket if I expect stuff back - it may arrive quickly!
	if (fileserver) {
		retsock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
		if (!retsock) LTPERR(LTP_SOCKET);
		// use configuration stuff
		rv=ltpd_sockopts(retsock);
		if (rv) LTPERR(rv);
		// setup listener addr
		rv=ltp_bind(retsock,&listener,sizeof(listener));
		if (rv) { free(retbuf); ltp_close(retsock); LTPERR(LTP_SOCKET);}
	}

	// shoot it off (in DTN fashion of course!)
	char *now_str=ltp_now_str(); 
	char *dest_str=strdup(ltpaddr2str(&dest));
	if (!fileserver) {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: sending %d bytes from file %s at %s to %s\n",
				insize,inputfname,now_str,dest_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	} else {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: sending %d bytes asking for file %s at %s to %s\n",
				insize,inputfname,now_str,dest_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	}
	rv=ltp_sendto(sock,inbuf,insize,flags,(ltpaddr*)&dest,sizeof(dest));
	now_str=ltp_now_str();
	if (!fileserver) {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: finished sending %d bytes from file %s at %s to %s\n",
				insize,inputfname,now_str,dest_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	} else {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: finished sending %d bytes asking for file %s at %s to %s\n",
				insize,inputfname,now_str,dest_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	}
	if (rv!=insize) { 
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: ltp_sendto failed! (ret=%d)\n",rv);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		free(dest_str); free(inbuf); ltp_close(sock); LTPERR(LTP_SOCKET);
	} 
	free(dest_str);

	if (fileserver) {
		retbuf=(unsigned char*)calloc(retsize,1);
		if (!retbuf) { LTPERR(LTP_MEMORY);}
		int bytesrxd=0;
		ltpaddr from;
		int fromlen=sizeof(from);
		bool done=false;
		now_str=ltp_now_str();
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: calling recvfrom %s\n", now_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		while ((bytesrxd=ltp_recvfrom(retsock,retbuf,retsize,flags,(ltpaddr*)&from,(ltpaddr_len*)&fromlen))==0) {
			if (!disksync) {
#undef LONGSLEEP
#ifdef LONGSLEEP
				sleep(1);
#endif
#define SHORTSLEEP
#ifdef SHORTSLEEP
				struct timespec	ts,ts1;
				memset(&ts,0,sizeof(ts));
				memset(&ts1,0,sizeof(ts));
				// 20ms seems to be the quickest I can get on Ubuntu regardless of what I ask for
				ts.tv_nsec=1000*1000*20;
				rv=nanosleep(&ts,&ts1);
#endif
			}
		}
		now_str=ltp_now_str();
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Got %d bytes back from %s at %s\n",bytesrxd,ltpaddr2str(&from),now_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		ltp_close(retsock);
	}
	
	if (verbose && ltpt_routes_given) {
		// just print 'em out at the end
		char routestr[4096];
		int olen=4096;
		rv=ltp_getsockopt(sock,SOL_SOCKET,LTP_SO_T_ROUTE,routestr,(socklen_t*)&olen);
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: getsockopt() failed for LTP_SO_T_ROUTE - (%d)\n",rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		} else {
			ltp_log(LTP_LOG_MAX,routestr);
			MAYBEPRINT(routestr);
		}
	}

	// hang about a bit to give things a chance
	if (dowait && !ltp_daemon) {
		fprintf(stderr,"Hit return to exit\n");
		char ch=getchar();
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: User asked to finish\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
	} else {
		if (!ltp_daemon && !blocking) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: Not waiting for a keyboard hit - will exit in %d\n",sleeptime);
			sleep(sleeptime);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
	}


	if (disksync) {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Synching to disk\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		rv=ltp_save();
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: Problem synching to disk (rv=%d)\n",rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
			return(rv);
		}
	}

	// and we're done (at this level)
	if (!ltp_daemon) {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Closing socket\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		ltp_close(sock);
		sleep(1);
	} else {
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Looping...forever...\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		while (true) sleep(100);// wait forever
	}
	free(inbuf);

	return(LTP_OK);
}

/*!
 * @brief printable version of a key id
 * @param kid are the key id octets
 * @param len is the length
 * @return is a ptr to a static array that's a printable version of the input
 */
static char *octets2str(unsigned char *kid, int len)
{
	// TODO make a real octets2str later for binary crap
	static char str[LTP_CS_MAX_KID_LEN];
	int thislen=min(LTP_CS_MAX_KID_LEN,len);
	memcpy(str,kid,thislen);
	str[thislen]='\0';
	// something anyway
	for (int i=0;i!=thislen;i++) {
		if (!isprint(str[i])) str[i]='%';
	}
	return (str);
}

/*!
 * @brief write a state-of-play struct to a file 
 * @param sop is a ptr to the state-of-plat struct
 * @param fname is the name of the file for the block - we'll use <<fname>>.sop
 * @return zero or error
 */
static int sop2file(ltp_sop *sop, char *fname)
{
	char fname1[MAXPATH];
	snprintf(fname1,MAXPATH,"%s.sop",fname);
	FILE *fp=fopen(fname1,"a+");
	if (strlen(fname)>MAXPATH-5) { // crap - but overwrite a bit anyway
		fname1[MAXPATH-5]='.';
		fname1[MAXPATH-4]='s';
		fname1[MAXPATH-3]='o';
		fname1[MAXPATH-2]='p';
		fname1[MAXPATH-1]='\0';
	}
	if (!fp) LTPERR(LTP_FILE);
	fprintf(fp,"----- SOP at %s -----\n",ltp_now_str());
	fprintf(fp,"Blocks %d ",sop->nblocks);
	if (sop->nblocks!=1) fprintf(fp,"\n");
	int lastb=min(MAXBINFOS,sop->nblocks);
	for (int i=0;i!=lastb;i++) {
		ltp_binfo *bi=&sop->infos[i];
		char bid_str[MAXPATH];
		int rv=bin2AH(sop->infos[i].bid,sop->infos[i].bidlen,bid_str,MAXPATH);
		if (rv) fprintf(fp,"Block ID: Cannot read!\n");
		else fprintf(fp,"Block ID: %s, Complete %s, Redlen: %d, RA-done: %s, Scopes: %d ",
				bid_str,(bi->complete?"true":"false"),bi->redlen,(bi->radone?"true":"false"),bi->nscopes);
		int lastscope=min(bi->nscopes,MAXSCOPES);
		if (lastscope>4) fprintf(fp,"\n");
		for (int j=0;j!=lastscope;j++) {
			fprintf(fp,"(%d,%d)",bi->lbs[j],bi->ubs[j]);
			if (j==(lastscope-1)) fprintf(fp,"\n");
			else if (((j+1)%4)==0) fprintf(fp,"\n");
			else fprintf(fp,",");
		}
	}
	fclose(fp);
	return(LTP_OK);
}


/*!
 * @brief do LTP server/responder/listener/router stuff
 * @return zero or error
 */
static int ltpd_server(void)
{
	int rv;
	int recviters=0;
	int sock;
	bool done=false;
	int bytesrxd;
	int fromlen;
	ltpaddr from;
	int insize=rxbufsize; // figure out what to do there
	unsigned char *inbuf;
	memset(&from,0,sizeof(from));
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: Starting %s\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
	inbuf=(unsigned char*)calloc(insize,1);
	if (!inbuf) { LTPERR(LTP_MEMORY);}

	// setup crypto maps if necessary
	if (auth_ciphersuite_given) {
		rv=ltp_crypt_setup(ccfg_fname);
		if (rv) LTPERR(rv);
	}
#undef ONESOCK
#ifdef ONESOCK
	sock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
	if (!sock) LTPERR(LTP_SOCKET);
	// use configuration stuff
	rv=ltpd_sockopts(sock);
	if (rv) LTPERR(rv);
	// setup listener addr
	rv=ltp_bind(sock,&listener,sizeof(listener));
	if (rv) { free(inbuf); ltp_close(sock); LTPERR(LTP_SOCKET);}
#endif
	// file naming stuff
	char fname[MAXPATH];
	int iteration=0;
	char *now_str;
	while (!done) {
		// shoot it off (in DTN fashion of course!)
#ifndef ONESOCK
		sock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
		if (!sock) LTPERR(LTP_SOCKET);
		// use configuration stuff
		rv=ltpd_sockopts(sock);
		if (rv) LTPERR(rv);
		// setup listener addr
		rv=ltp_bind(sock,&listener,sizeof(listener));
		if (rv) { free(inbuf); ltp_close(sock); LTPERR(LTP_SOCKET);}
#endif
		now_str=ltp_now_str();
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: calling recvfrom %s\n", now_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		while ((bytesrxd=ltp_recvfrom(sock,inbuf,insize,flags,(ltpaddr*)&from,(ltpaddr_len*)&fromlen))==0) {
			if (!disksync) {
#undef LONGSLEEP
#ifdef LONGSLEEP
				sleep(1);
#endif
#define SHORTSLEEP
#ifdef SHORTSLEEP
				struct timespec	ts,ts1;
				memset(&ts,0,sizeof(ts));
				memset(&ts1,0,sizeof(ts));
				// 20ms seems to be the quickest I can get on Ubuntu regardless of what I ask for
				ts.tv_nsec=1000*1000*20;
				rv=nanosleep(&ts,&ts1);
#endif
			} else {
				rv=ltp_save();
				if (rv) done=true;
			}
		}
		int soperrcount=0;
		now_str=ltp_now_str();
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Got %d bytes back (%d-th time) from %s at %s\n",bytesrxd,recviters++,ltpaddr2str(&from),now_str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		MAYBEPRINT(ltpstr);
		ltp_sop sop;
		socklen_t soplen=sizeof(ltp_sop);
		rv=ltp_getsockopt(sock,SOL_SOCKET,LTP_SO_SOP,&sop,(socklen_t*)&soplen);
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: getsockopt(LTP_SO_SOP) returned error %d\n",rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		} else {
			// if its got red, then wait until RA done
			if (sop.infos[0].redlen!=0) {
				while (soperrcount<10 && !sop.infos[0].radone) { // HACK HACK enough for now
					soplen=sizeof(ltp_sop);
					rv=ltp_getsockopt(sock,SOL_SOCKET,LTP_SO_SOP,&sop,(socklen_t*)&soplen);
					if (rv) {
						soperrcount++;
						snprintf(ltpstr,LTPSTR_LEN,"LTPD: getsockopt(LTP_SO_SOP) returned error %d\n",rv);
						ltp_log(LTP_LOG_MAX,ltpstr);
						MAYBEPRINT(ltpstr);
					}
					sleep(1);
				}
			}
			if (soperrcount<10) {
				snprintf(fname,MAXPATH,"%s-%s.%d",outputfname,inet_ntoa(from.sock.sin_addr),from.sock.sin_port);
				rv=sop2file(&sop,fname);
				if (rv) {
					snprintf(ltpstr,LTPSTR_LEN,"LTPD: sop2file failed - %d\n",rv);
					ltp_log(LTP_LOG_MAX,ltpstr);
					MAYBEPRINT(ltpstr);
				}  
			} else {
				snprintf(ltpstr,LTPSTR_LEN,"LTPD: too many getsockopt errors - %d\n",soperrcount);
				ltp_log(LTP_LOG_MAX,ltpstr);
				MAYBEPRINT(ltpstr);
			}
		}
		// if (bytesrxd==-1) { free(inbuf); ltp_close(sock); LTPERR(LTP_SOCKET);}
		bool fileserved=false;
		if (bytesrxd>0 && fileserver) {
			// parse request and serve file back (maybe);
			// later request might have abs/rel time and destination - for now just send back to same guy
			time_t	delay=0;
			ltpaddr	to=from;
			rv=parse_request(inbuf,insize,fname,MAXPATH,&delay,&to);
			if (!rv) {
				FILE *fp=fopen(fname,"r");
				if (fp) { 
					// read the file and send it back down to the requestor
					struct stat outstat;
					rv=fstat(fileno(fp),&outstat);
					if (rv) { fclose(fp); LTPERR(LTP_FILE);}
					int outsize=outstat.st_size;
					if (!outsize) { fclose(fp); LTPERR(LTP_FILE);}
					unsigned char *outbuf=(unsigned char*)calloc(outsize,1);
					if (!outbuf)  { fclose(fp); LTPERR(LTP_MEMORY);}
					rv=fread(outbuf,1,outsize,fp);
					if (rv!=outsize) { fclose(fp); free(outbuf); LTPERR(LTP_FILE);}
					fclose(fp);
					snprintf(ltpstr,LTPSTR_LEN,"LTPD: Read %d bytes from %s OK\n",outsize,fname);
					ltp_log(LTP_LOG_MAX,ltpstr);
					MAYBEPRINT(ltpstr);
					// eventually...
					int outsock;
					outsock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
					if (!outsock) LTPERR(LTP_SOCKET);
					// use configuration stuff
					rv=ltpd_sockopts(outsock);
					if (rv) LTPERR(rv);
					int foo=1;
					rv=ltp_setsockopt(outsock,SOL_SOCKET,LTP_SO_LINGER,&foo,sizeof(foo));
					if (rv) {
						snprintf(ltpstr,LTPSTR_LEN,"LTPD: setsockopt() failed for LTP_SO_LINGER: rv=%d\n", rv);
						ltp_log(LTP_LOG_MAX,ltpstr);
						MAYBEPRINT(ltpstr);
					}
					// setup source addr
					rv=ltp_bind(outsock,(ltpaddr*)&listener,sizeof(listener));
					if (rv) { free(inbuf); ltp_close(outsock); LTPERR(LTP_SOCKET);}
					rv=ltp_sendto(outsock,outbuf,outsize,flags,(ltpaddr*)&to,sizeof(to));
					if (rv!=outsize) { free(outbuf); ltp_close(sock); LTPERR(LTP_SOCKET);}
					fileserved=true;
					free(outbuf);
					ltp_close(outsock);
				} else {
					// log but continue
					snprintf(ltpstr,LTPSTR_LEN,"LTPD: Fileserver can't open requested file %s\n",fname);
					ltp_log(LTP_LOG_MAX,ltpstr);
					MAYBEPRINT(ltpstr);
				}
			}
		} 
#ifndef ONESOCK
		ltp_close(sock);
		sock=0;
		if (disksync) {
			rv=ltp_save();
			if (rv) done=true;
			rv=ltp_restore();
			if (rv) done=true;
		}
#endif
		if (!fileserved && bytesrxd>0) { // save whatever the guy sent me
			// write file out
			snprintf(fname,MAXPATH,"%s-%s.%d",outputfname,inet_ntoa(from.sock.sin_addr),from.sock.sin_port);
			FILE *fp=fopen(fname,"a+"); // for append!! might save the day if >1 buffer from same src. addr & port
			if (!fp) { free(inbuf); ltp_close(sock); LTPERR(LTP_FILE);}
			rv=fwrite(inbuf,1,bytesrxd,fp);
			if (rv!=bytesrxd) { 
				free(inbuf);fclose(fp); ltp_close(sock); LTPERR(LTP_FILE);
			}
			fclose(fp);
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: Wrote %d bytes to %s\n",bytesrxd,fname);
			ltp_log(LTP_LOG_MAX,ltpstr);
			MAYBEPRINT(ltpstr);
		}
		// for loopback mode we can now exit
		if (mode==LTPD_MODE_BOTH) done=true;
	}
#ifdef ONESOCK
	ltp_close(sock);
#endif
	return(LTP_OK);
}

/*!
 * @brief do LTP server/responder/listener/router stuff
 * @return zero or error
 */
static int ltpd_router(void)
{
	int rv;
	int sock;
	bool done=false;
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: Starting %s\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
	sock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
	if (!sock) LTPERR(LTP_SOCKET);
	// setup listener addr
	rv=ltp_bind(sock,&listener,sizeof(listener));
	if (rv) { ltp_close(sock); LTPERR(LTP_SOCKET);}
	// use configuration stuff
	rv=ltpd_sockopts(sock);
	if (rv) LTPERR(rv);
	ltp_close(sock);
	while (!done) {
		// need to sleep a bit?
		// TODO: think later
		if (!disksync) {
			sleep(1);
		} else {
			rv=ltp_save();
			if (rv) done=true;
		}
	}
	return(LTP_OK);
}

/*!
 * @brief server wrapper for thread proto
 */
void *ltpd_server_thread_fnc (void* foo)
{
	int rv=ltpd_server();
	return NULL;
}

// a thread :-)
static pthread_t	bothserver=(pthread_t)0;
// a message :-)
static const char	*bothserver_message="ltpd_server_thread";

/*!
 * @fn int main(int argc, char *argv[])
 * @brief see usage
 * @param argc you should know (uses getopt)
 * @param argv you should know (uses getopt)
 * @return something or other
 */
main(int argc, char *argv[])
{
	int rv;

    extern char *optarg;/* used by getopt */
    extern int optind, opterr, optopt;
	int c;
	int errflag=0;

	snprintf(ccfg_fname,MAXPATH,LTPD_DEF_CCFG);
	snprintf(log_fname,MAXPATH,LTPD_DEF_LOG);
	snprintf(inputfname,MAXPATH,LTPD_DEF_INPUT);
	snprintf(outputfname,MAXPATH,LTPD_DEF_OUTPUT);
	snprintf(ltpt_cfg_fname,MAXPATH,LTPD_DEF_TCFG);
	snprintf(ltpt_routes_fname,MAXPATH,LTPD_DEF_RCFG);
	snprintf(requestedfile,MAXPATH,LTPD_DEF_RF);

	// dummy is 127.0.0.1:1113 - good for loopback
	ltpaddr dummy={AF_LTP,UDP_DEF_PORT,0x0100007f,"\0\0\0\0\0\0\0"};
	memcpy(&dest,&dummy,sizeof(sockaddr_in));
	memcpy(&source,&dummy,sizeof(sockaddr_in));
	memcpy(&listener,&dummy,sizeof(sockaddr_in));

    while ((c = getopt(argc, argv, "2:A:C:D:G:K:IL:R:S:b:c:dg::hi:l:m:o:r:st::vw::?")) != EOF) {
    	switch(c) {
			case '2':
				l2mtu=atoi(optarg);
				l2mtu_set=true;
				break;
			case 'A':
				auth.ciphersuite=ltp_crypt_mapciphersuite(optarg);
				auth_ciphersuite_given=true;
				break;
			case 'C':
				cookies=atoi(optarg);
				break;
			case 'D':
				str2ltpaddr(optarg,&dest);
				break;
			case 'G':
				cookie_grace=atoi(optarg);
				cookie_grace_given=true;
				break;
			case 'K':
				auth.keyidlen=strlen(optarg);
				memcpy(auth.keyid,optarg,min(auth.keyidlen+1,MAXPATH));
				auth_keyid_given=true;
				break;
			case 'I':
				ion_mode=true;
				break;
			case 'L':
				str2ltpaddr(optarg,&listener);
				break;
			case 'R':
				redlen=atoi(optarg);
				break;
			case 'S':
				str2ltpaddr(optarg,&source);
				source_given=true;
				break;
			case 'b':
				rxbufsize=atoi(optarg);
				if (rxbufsize < LTPD_DEF_RXBUF) {	
					// assume megabytes
					rxbufsize *= (1024*1024);
				}
				break;
			case 'c':
				snprintf(ccfg_fname,MAXPATH,"%s",optarg);
				break;
			case 'd':
				disksync=true;
				break;
			case 'g':
				fileserver=true;
				if (optarg) {
					snprintf(requestedfile,MAXPATH,"%s",optarg);
					requestedfilegiven=true;
				}
				break;
			case 'i':
				inputfnamegiven=true;
				snprintf(inputfname,MAXPATH,"%s",optarg);
				break;
			case 'l':
				snprintf(log_fname,MAXPATH,"%s",optarg);
				break;
			case 'm':
				mode=mapmode(optarg);
				break;
			case 'o':
				snprintf(outputfname,MAXPATH,"%s",optarg);
				break;
			case 'r':
				snprintf(ltpt_routes_fname,MAXPATH,"%s",optarg);
				ltpt_routes_given=true;
				break;
			case 's':
				ltp_daemon=true;
				break;
			case 't':
				if (optarg) {
					snprintf(ltpt_cfg_fname,MAXPATH,"%s",optarg);
				}
				ltpt_cfg_given=true;
				break;
			case 'v':
				verbose=true;
				break;
			case 'w':
				dowait=false;
				if (optarg) {
					if (*optarg=='b' || *optarg=='B') {
						sleeptime=LTPD_DEF_SLEEPTIME;
						blocking=true;
					} else {
						sleeptime=atoi(optarg);
						blocking=false;
					}
				}
				break;
			case '?':
			case 'h':
			default:
				usage();
		}
	}

	MAYBEPRINT("ltpd running in verbose mode\n");

	ltp_log_init(log_fname,LTP_LOG_MAX);
	ltp_log_timestamp("Starting ltpd");

	checkparams();

	snprintf(ltpstr,LTPSTR_LEN,"LTPD: Local parameters are:\nLTPD:\tcrypto-cfg=%s\nLTPD:\tdisksync=%s\nLTPD:\tdaemon=%s\n",
				ccfg_fname,(disksync?"true":"false"),(ltp_daemon?"true":"false"));
	ltp_log(LTP_LOG_MAX,ltpstr);
	MAYBEPRINT(ltpstr);
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: \tinput=%s\nLTPD:\tlog=%s\nLTPD:\tmode=%d\nLTPD:\toutput=%s\nLTPD:\tverbose=%s\n",
				inputfname,log_fname,mode,outputfname,(verbose?"true":"false"));
	ltp_log(LTP_LOG_MAX,ltpstr);
	MAYBEPRINT(ltpstr);
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: \tl2mtu=%d\nLTPD:\tLTP-T cfg=%s\nLTPD:\tdowait=%s,sleeptime=%d\nLTPD:\trxbufsize=%d\n",
					l2mtu,ltpt_cfg_fname,(dowait?"true":"false"),sleeptime,rxbufsize);
	ltp_log(LTP_LOG_MAX,ltpstr);
	MAYBEPRINT(ltpstr);
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: \tblocking=%s\n",
					(blocking?"true":"false"));
	ltp_log(LTP_LOG_MAX,ltpstr);
	MAYBEPRINT(ltpstr);
	char *dstr=strdup(inet_ntoa(dest.sock.sin_addr));
	char *sstr=strdup(inet_ntoa(source.sock.sin_addr));
	char *lstr=strdup(inet_ntoa(listener.sock.sin_addr));
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: Protocol parameters are:\nLTPD:\tdest=%s:%d\nLTPD:\tsource=%s:%d\nLTPD:\tlisten=%s:%d\nLTPD:\tredlen=%d\nLTPD:\tcookies=%d\nLTPD:\tcookie_grace=%d\nLTPD:\tCipher=%d,KeyID=%s\n",
				dstr,dest.sock.sin_port,
				sstr,source.sock.sin_port,
				lstr,listener.sock.sin_port,
				redlen,cookies,cookie_grace,
				auth.ciphersuite,auth.keyid);
	free(dstr);
	free(sstr);
	ltp_log(LTP_LOG_MAX,ltpstr);
	MAYBEPRINT(ltpstr);

	// new file server mode checks
	if (fileserver) {
		// client - doesn't make sense to have both -i and -g
		if (mode==LTPD_MODE_CLIENT) {
			if (inputfnamegiven) {
				snprintf(ltpstr,LTPSTR_LEN,"LTPD: Silly to both send and receive a file - only one of -i or -g for clients!");
				MAYBEPRINT(ltpstr);
				usage();
			}
			if (!requestedfilegiven) {
				snprintf(ltpstr,LTPSTR_LEN,"LTPD: A client has to specify a file for -g\n");
				MAYBEPRINT(ltpstr);
				usage();
			}
		} else if (mode==LTPD_MODE_SERVER) {
			// anything goes I guess
		} else {
			// silly - exit out 
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: File server mode only for client or server, not both or routers\n");
			MAYBEPRINT(ltpstr);
			usage();
		}
	}

	char *now_str=ltp_now_str();
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: Starting run at: %s\n",now_str);
	MAYBEPRINT(ltpstr);

	// check setup, initialise stuff etc.
	if (ltp_daemon) { // fork off a process and let it do the rest of the work
		MAYBEPRINT("...and as a ltp_daemon\n");
		verbose=false; // turn off verbosity now, no point in it anyway
		char cwd[MAXPATH];
		char *crv=getcwd(cwd,MAXPATH);
		if (crv==NULL) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: getcwd failed - exiting all (errno=%d)\n",errno);
			ltp_log(LTP_LOG_MAX,ltpstr);
			ltp_log_close();
			fprintf(stderr,"%s",ltpstr);
			return(errno);
		}
		rv=daemon(0,0); // this guy auto-exits the parent
		if (rv==-1) { // oops
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: daemon failed - exiting all (errno=%d)\n",errno);
			ltp_log(LTP_LOG_MAX,ltpstr);
			ltp_log_close();
			fprintf(stderr,"%s",ltpstr);
			return(errno);
		}
		snprintf(ltpstr,LTPSTR_LEN,"LTPD: Daemon child chdir to %s\n",cwd);
		MAYBEPRINT(ltpstr);
		rv=chdir(cwd);
		if (rv==-1) {
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: chdir failed - exiting all (errno=%d)\n",errno);
			ltp_log(LTP_LOG_MAX,ltpstr);
			ltp_log_close();
			fprintf(stderr,"%s",ltpstr);
			return(errno);
		}
	} 

	// bit of useless info
	snprintf(ltpstr,LTPSTR_LEN,"LTPD: ltpd running in process: %d\n",getpid());
	MAYBEPRINT(ltpstr);
	ltp_log(LTP_LOG_MAX,ltpstr);

	if (disksync) {
		rv=ltp_restore();
		if (rv) LTPERR(rv);
	}

	// first we say what we're gonna try do ('cause it helps me figure it out:-)
	switch (mode) {
		case LTPD_MODE_BOTH:
			{
				rv=ltp_cue_set_whoiam(&source);
				if (rv) LTPERR(rv);
				pthread_t tmpThread;
				pthread_create(&tmpThread,NULL,ltpd_server_thread_fnc,(void*)bothserver_message);
				bothserver=tmpThread;
#ifdef LTP_TARGET_ARM_LINUX
				sleep(1);
				snprintf(ltpstr,LTPSTR_LEN,"LTPD: Started ltpd_server thread - tid(%p) then slept a second\n",bothserver);
				ltp_log(LTP_LOG_ADVISORY,ltpstr);
				ltp_log_flush();
#else
				snprintf(ltpstr,LTPSTR_LEN,"LTPD: Started ltpd_server thread - tid(%p)\n",(void*)bothserver);
				ltp_log(LTP_LOG_ADVISORY,ltpstr);
#endif
				sleep(2);
				rv=ltpd_client();
			}
			break;
		case LTPD_MODE_CLIENT:
			rv=ltp_cue_set_whoiam(&source);
			if (rv) LTPERR(rv);
			// try open input file, then send it, once done exit
			rv=ltpd_client();
			break;
		case LTPD_MODE_SERVER:
			// start listener, when file/session arrives, drop to output
			rv=ltp_cue_set_whoiam(&listener);
			if (rv) LTPERR(rv);
			rv=ltpd_server();
			break;
		case LTPD_MODE_ROUTER:
			// same as server, but also turn on LTP-T, (mode is visible in ltpd_server()
			rv=ltp_cue_set_whoiam(&listener);
			if (rv) LTPERR(rv);
			rv=ltpd_router();
			break;
		default:
			snprintf(ltpstr,LTPSTR_LEN,"LTPD: Unexpected mode (%d) - exiting\n",mode);
			ltp_log(LTP_LOG_MAX,ltpstr);
			ltp_log_close();
			MAYBEPRINT(ltpstr);
			exit(-1);
	}

	// if we get here we're either an ltp_daemon's child process or else we're running synchronously
	ltp_log_timestamp("Finishing ltpd");
	ltp_log_close();

	return(0);
}
