/*!
 * @file sendt_cues.cc
 * @brief This is the external interface for our event handling API
 * @version $Revision: 1.66 $ $Author: stephen $
 * @version Copyright (C) 2004 Trinity College Dublin
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <ctype.h> // for isalpha()
#include <time.h> // for time()
#include "ltp_base.h"
#include "ltp.h"
#include "ltp_int.h" 			
#include "ltp_names.h" 			
#include "ltp_cues.h"

#undef GOTHEREIN
#ifdef GOTHEREIN
#define GOTHERE { fprintf(stderr,"At %d in %s\n",__LINE__,__FUNCTION__); }
#else 
#define GOTHERE
#endif

#ifdef LTP_VERBOSE
static time_t lastprinted=0;
#define CONTOUT if (1) { \
	time_t now=time(0); \
	if (now>(lastprinted+5)) { \
		char lltpstr[LTPSTR_LEN];\
		snprintf(lltpstr,LTPSTR_LEN,"Continue at %s from %s:%d \n",ltp_now_str(),__FILE__,__LINE__); \
		ltp_log(LTP_LOG_MAX,lltpstr);\
		lastprinted=now; \
	} \
	continue; }
#else
#define CONTOUT continue
#endif

/// the default schedule file name
static char def_fname[]="ltp.sched";

/// the local var with the schedule file name
static char *fname=def_fname;

/// entry types
#define COMMS_ENTRY 0

/// pretty print
#define TFP(x) (x?"true":"false")

// these arrays will hold the LTT values for the times we'll need in a bit
int 		arrdim=0; // size of alloc'd space - done in chunks of 100
int			arrents=0; // used space
ltp_time_t	*relarr=NULL; // the relative times
ltp_time_t	*lttarr=NULL; // the corresponding ltt's
static ltp_time_t	epoch=0; // the start of an epoch

typedef struct lttinfo_defn {
	int		arrdim;
	int 	arrents;
	ltp_time_t	*relarr;
	ltp_time_t	*lttarr;
	ltp_time_t	epoch;
} lttinfo_t;

int nlttinfos=0;
static lttinfo_t	*arrarr=NULL;

#define SET_LTT(sc) \
	arrdim=arrarr[sc->vis->ltt_inf_ind].arrdim; \
	arrents=arrarr[sc->vis->ltt_inf_ind].arrents; \
	relarr=arrarr[sc->vis->ltt_inf_ind].relarr; \
	lttarr=arrarr[sc->vis->ltt_inf_ind].lttarr; \
	epoch=arrarr[sc->vis->ltt_inf_ind].epoch;

/*!
 * @struct a schedule entry
 * @field entry_type is the type, COMMS (others defined elsewher)
 * @field reltime is the relative time of the entry
 * @field peers is the addr-spec (e.g. 127.0/16, 127.0.0.1:4321)
 * @field tx specifies whether this is a tx or rx entry
 * @field on specifies whether this is an on or off entry
 * @field ntags specifies how many more tag=value paires there are 
 * @field is an array of tag strings
 * @field is an array of value strings
 */
typedef struct sc_entry_defn {
	int			entry_type;
	float		reltime;
	char 		*peers;
	bool		tx;
	bool		on;
	int			ntags;
	char		**tags;
	char		**vals;
} sc_entry;

#define SENDT_CUE_ABS 0
#define SENDT_CUE_REL 1
#define SENDT_CUE_EPH 2

/*
 * @struct a visibility entry
 * @field reltime is the relative time
 * @field from is the looker
 * @field bitrates is an array of bitrates (0=invisible) settings for that time (up to the next)
 */
typedef struct vis_ent_defn {
	ltp_time_t	reltime;
	int			from;
	int			*bitrates; 
} vis_ent;

/*
 * @struct visibility information
 * @field absstart says whether start is absolute or relative
 * @field start determines the start time
 * @field ltt_fname is the ltt file name (EPH)
 * @field nto says how many "to" entries there are 
 * @field tos is an array of "to" guys
 * @field nfrom says how many "from" entries there are 
 * @field froms is an array of "from" guys - I only use the local "me"/"whoami" value
 * @field nentries is the number of entries
 * @field entries are the entries
 */
typedef struct vis_info_defn {
	bool		absstart;
	ltp_time_t	start;
	char		*ltt_fname;
	int			ltt_inf_ind;
	int			nto;
	char		**tos;
	int 		nfrom;
	char		**froms;
	int 		nentries;
	vis_ent		*entries;
} vis_info;



// an "always-on" vis_info to use if we can't load files or stuff

static const char v_all[30]="255.255.255.255";
static char *vv_all=(char*)v_all;
static int v_rates[2]={1,1};
static int *vv_rates=v_rates;
static vis_ent v_ent={0,1,vv_rates};
static vis_info vis_all_always={
	true, 					// abstart
	0, 						// start - will be replaced with ltp_now
	(char*)"No file at all",		// ltt_fname
	-1,
	1,
	&vv_all,
	1,
	&vv_all,
	1,
	&v_ent};

/*!
 * @struct schedule structure
 * @field type is ABS,REL or EPH
 * @field start is the start time (REL or ABS)
 * @field vis_fname is the visibility file name (EPH)
 * @field dutycycle is the dutycycle of the schedule
 * @field lastentry is the last event time (relative to dutycycle)
 * @field nentries is the number of entries
 * @field entries is the array of entries
 */
typedef struct sc_sched_defn {
	int			type;
	ltp_time_t	start;
	char		*vis_fname;
	vis_info	*vis;
	ltp_time_t	dutycycle;
	ltp_time_t	lastentry;
	int 		nentries;
	sc_entry	*entries;
} sc_sched;

/// the number of schedules
static int nschedules=1;

/// the max # of schedules allowed
#define LTP_MAXSCHED 100

/// the default schedule entry
static sc_entry def_entry = { 
	0, // type = comms
	0, // rel time = 0
	NULL, // peers
	true, // tx
	true, // on/off
	0, // ntags
	NULL, // tags
	NULL  // vals
};

/// the default schedule - always on
static sc_sched def_sched = {
	0, // start
	0,
	NULL,
	NULL,
	0,
	0,
	1, // entries
	&def_entry
};

/// the schedule variable
static sc_sched *schedules=&def_sched;

/// the char* version of whoami
static char *mestr=NULL;

/// the visibility information for this process
static vis_info *vises=NULL;

/// the number of those loaded
static int nvises=0;

/// the number of visibility files loaded already
static int nvis_fnames=0;

/// the names of the loaded visibility files
static char **vis_fnames=NULL;

/// an artefact for strrok_r
static char *st_sched=NULL;

/// another artefact for strrok_r
static char *st_vis=NULL;

/*!
 * @brief read in an xml formatter schedule file
 * @param fname is the file name
 * @return zero or error
 */
static int sendt_try_xml(char *fname)
{
	// implement LATER, for now returning error is right
	return(LTP_FILE);
}

#define VERSION_STR "VERSION"
#define NS_STR		"NSCHEDULES"
#define START_STR	"START"
#define COMMS_STR	"COMMS"
#define END_STR		"END"
#define DUTY_STR	"DUTY"
#define ABS_STR		"ABS"
#define REL_STR		"REL"
#define TX_STR		"TX"
#define RX_STR		"RX"
#define ON_STR		"ON"
#define OFF_STR		"OFF"
#define ME_STR		"ME"
#define EPH_STR		"EPH"
#define FROM_STR	"FROM"
#define LTT_STR		"LTT"
#define TO_STR		"TO"
#define VIS_STR		"VIS"
#define DUR_STR		"DUR"
#define ADDED_STR	"ADDED"
#define NEV_STR		"NEV"

#define LTP_COMMA_SEP ",\n"
#define LTP_EQ_SEP "=\n"


static void kill_sched(void) 
{
	if (!schedules) return;
	if (schedules!=&def_sched) {
		for (int i=0;i!=nschedules;i++) free(schedules[i].entries);
		free(schedules);
		schedules=NULL;
	}
	return;
}

typedef struct durstr_defn {
	const char *str;
	int units;
} durstr_t;

const durstr_t	durstr[]={
	{  "seconds", 1 },
	{  "second", 1 },
	{  "sec", 1 },
	{  "secs", 1 },
	{  "s", 1 },
	{  "minutes", 60 },
	{  "minute", 60 },
	{  "min", 60 },
	{  "m", 60 },
	{  "hours", 3600 },
	{  "hour", 3600 },
	{  "hr", 3600 },
	{  "h", 3600 },
	{  "days", 3600*24 },
	{  "day", 3600*24 },
	{  "dy", 3600*24 },
	{  "d", 3600*24 },
	{  "week", 3600*24*7 },
	{  "weeks", 3600*24*7 },
	{  "wk", 3600*24*7 },
	{  "w", 3600*24*7 },
};

static int parse_time_str(char *str, ltp_time_t *dur)
{
	// str format is nnnn.mmmmSTR
	// nnnn.mmm are floating point units
	// STR can be any of the units in durstr above
	char *dup=strdup(str);
	char *cp=dup;
	while (isdigit(*cp) || *cp=='.') cp++;
	int multiplier=1;
	if (*cp) {
		char *units=strdup(cp);
		*cp='\0';
		int nstrs=sizeof(durstr)/sizeof(durstr_t);
		for(int i=0;i!=nstrs;i++) {
			if (ltp_same(units,(char*)durstr[i].str)) {
				multiplier=durstr[i].units;
				break;
			}
		}
	} 
	float num=atof(dup);
	*dur=num*multiplier;
	return(LTP_OK);
}

static ltp_time_t	getrel(sc_sched *sc, ltp_time_t abs)
{
	ltp_time_t reltime=-1;
	if (sc->type==SENDT_CUE_ABS) {
		if (sc->start>abs) {
			ltp_time_t tt=(floor)((sc->start-abs)/sc->dutycycle) +1 ;
			ltp_time_t tt1=sc->start - tt*sc->dutycycle;
			reltime=abs-tt1;
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"%s: abs (%.0f) preceeds start (%.0f) reltime is: %.0f, rstart is %.0f\n",
					__FUNCTION__,abs,sc->start,reltime,tt1);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			return(reltime);
		}
		ltp_time_t tmp=(abs-sc->start)/sc->dutycycle;
		reltime=abs-sc->start-(floor(tmp)*sc->dutycycle);
		return(reltime);
	} 
	if (sc->type==SENDT_CUE_REL) {
		// need to include sc->start in this
		char *now_str=ltp_time_str(abs);
		// this may be non-robust ;-)
		now_str[11]='0';
		now_str[12]='0';
		now_str[14]='0';
		now_str[15]='0';
		now_str[17]='0';
		now_str[18]='0';
		ltp_time_t midnight;
		int rv=str2ltp_time_t(now_str,&midnight);
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"%s: error (%d) from str2ltp_time_t\n",__FUNCTION__,rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
		}
		ltp_time_t tmp=(abs-midnight)/sc->dutycycle;
		reltime=abs-midnight-(floor(tmp)*sc->dutycycle);
		return(reltime);
	} 
	if (sc->type==SENDT_CUE_EPH) {
		vis_info *vi=sc->vis;
		if (vi->absstart) {
			ltp_time_t tmp=(abs-vi->start)/sc->dutycycle;
			reltime=abs-vi->start-(floor(tmp)*sc->dutycycle);
			return(reltime);
		} else {
			snprintf(ltpstr,LTPSTR_LEN,"%s: weird value in EPH schedule\n",__FUNCTION__);
			ltp_log(LTP_LOG_MAX,ltpstr);
			return(reltime);
		}
	}
	snprintf(ltpstr,LTPSTR_LEN,"%s: weird schedule type (%d)\n",__FUNCTION__,sc->type);
	ltp_log(LTP_LOG_MAX,ltpstr);
	return(reltime);
}

#define MINUTE 60
#define HOUR (60*MINUTE)
#define DAY (24*HOUR)
#define WEEK (7*DAY)

static char *nicediff(ltp_time_t	diff)
{
	static char retstr[MAXPATH];
	if (diff<MINUTE) { snprintf(retstr,MAXPATH,"%.3fs",diff); return(retstr); }
	if (diff<HOUR) { snprintf(retstr,MAXPATH,"%.3fm",diff/MINUTE); return(retstr); }
	if (diff<DAY) { snprintf(retstr,MAXPATH,"%.3fh",diff/HOUR); return(retstr); }
	if (diff<WEEK) { snprintf(retstr,MAXPATH,"%.3fd",diff/DAY); return(retstr); }
	snprintf(retstr,MAXPATH,"%.3fw",diff/WEEK); return(retstr); 
}


static ltp_time_t getltt(sc_sched *sc,ltp_time_t abs)
{
	SET_LTT(sc);
	ltp_time_t	visrel=abs-epoch;
	if (arrents<=0) return(0);
	int entry=arrents-1;
	while (entry && relarr[entry] && relarr[entry]>visrel) {
		entry=(int)floor((visrel-relarr[0])/(relarr[entry]-relarr[0])*1000)*entry/1000;
	}
	for (int i=0;i!=arrents;i++) {
		if (relarr[entry]>=visrel) {
			if (entry==0) return(lttarr[0]);
			else return((lttarr[entry]+lttarr[entry-1])/2);
		}
		entry++;
		if (entry==arrents) entry=0;
	}
	return(0);
}

/*!
 * @brief print schedules to buffer
 * @param buf is a caller allocated buffer for the output
 * @param blen is the size of buf
 */
void print_scheds(char *buf, int blen)
{
	if (blen <=0 || !buf) return;
	buf[0]='\0';
	int lenleft=blen;
	ltp_time_t	now=ltp_now();
	ltp_time_t	reltime=-1;
	for (int i=0;i!=nschedules;i++) {
		int nextind=-1;
		reltime=getrel(&schedules[i],now);
		if (schedules[i].type==SENDT_CUE_ABS) {
			snprintf(buf+strlen(buf),lenleft,"PS: NE: ABS Schedule %d start: %s duty-cycle: %.0f\n",i+1,ltp_time_str(schedules[i].start),schedules[i].dutycycle);
			lenleft=blen-strlen(buf);
		} else if (schedules[i].type==SENDT_CUE_REL) {
			snprintf(buf+strlen(buf),lenleft,"PS: NE: REL Schedule %d start: %.0f duty-cycle: %.0f\n",i+1,schedules[i].start,schedules[i].dutycycle);
			lenleft=blen-strlen(buf);
		} else if (schedules[i].type==SENDT_CUE_EPH) {
			snprintf(buf+strlen(buf),lenleft,"PS: NE: EPH Schedule %d vis file: %s duty-cycle: %.0f\n",i+1,schedules[i].vis_fname,schedules[i].dutycycle);
			lenleft=blen-strlen(buf);
		}
		snprintf(buf+strlen(buf),lenleft,"PS: \t %d entries (last: %.0f)\n",schedules[i].nentries,schedules[i].lastentry);
		lenleft=blen-strlen(buf);
		sc_entry *se=schedules[i].entries;
		int ents2p=min(schedules[i].nentries,20);
		bool truncated=(ents2p!=schedules[i].nentries);
		for (int j=0;j!=ents2p;j++) {
			char *nt= ltp_name_trans(se->peers);
			if (!nt) nt=se->peers;
			snprintf(buf+strlen(buf),lenleft,"PS: \treltime=%.0f,%s,%s,peers=%s\n", 
						se->reltime,
						(se->tx?"tx":"rx"),
						(se->on?"on":"off"),
						nt);
			lenleft=blen-strlen(buf);
			for (int k=0;k!=se->ntags;k++) {
				snprintf(buf+strlen(buf),lenleft,"PS: \t\ttag: %s, value: %s\n",se->tags[k],se->vals[k]);
				lenleft=blen-strlen(buf);
			}
			if (nextind==-1 && se->reltime > reltime) {
				nextind=j;
			}
			se++;
		}
		if (truncated) {
			snprintf(buf+strlen(buf),lenleft,"PS: \t...truncating: only showing %d of %d...\n",ents2p+1,schedules[i].nentries);
			lenleft=blen-strlen(buf);
			// go find next ind - might be after truncation
			se=&schedules[i].entries[schedules[i].nentries-1];
			char *nt= ltp_name_trans(se->peers);
			if (!nt) nt=se->peers;
			snprintf(buf+strlen(buf),lenleft,"PS: \treltime=%.0f,%s,%s,peers=%s (last one)\n", 
						se->reltime,
						(se->tx?"tx":"rx"),
						(se->on?"on":"off"),
						nt);
			lenleft=blen-strlen(buf);
			for (int k=0;k!=se->ntags;k++) {
				snprintf(buf+strlen(buf),lenleft,"PS: \t\ttag: %s, value: %s\n",se->tags[k],se->vals[k]);
				lenleft=blen-strlen(buf);
			}
			nextind=-1;
			for (int j=0;j!=schedules[i].nentries;j++) {
				if (nextind==-1 && schedules[i].entries[j].reltime > reltime) {
					nextind=j;
					break;
				}
			}
		}
		if (nextind==-1) {
			nextind=0;
		}
		if (schedules[i].nentries) {
			ltp_time_t tmp=schedules[i].entries[nextind].reltime-reltime;
			if (tmp<0) tmp+=schedules[i].dutycycle;
			if (schedules[i].vis) {
				SET_LTT((&schedules[i]));
			}
			snprintf(buf+strlen(buf),lenleft,"PS: NE: Now is:\nPS: NE:\t%s,sched-rel: %.0f, epoch-rel: %.0f\n", 
					ltp_time_str(now),
					reltime,
					schedules[i].start-epoch+reltime);
			lenleft=blen-strlen(buf);
			// figure out what TX/RX state I'm now in
			bool txon=true;
			bool rxon=true;
			int txind=-1;
			int rxind=-1;
			bool txdone=false;
			bool rxdone=false;
			int findind=nextind;
			if (findind==0) findind=schedules[i].nentries-1;
			else findind=nextind-1;
			while (findind!=nextind && !(txdone && rxdone)) {
				if (!txdone && schedules[i].entries[findind].tx) {
					txdone=true;
					txon=schedules[i].entries[findind].on;
					txind=findind;
				} else if (!rxdone) {
					rxdone=true;
					rxon=schedules[i].entries[findind].on;
					rxind=findind;
				}
				findind--;
				if (findind<0) findind=schedules[i].nentries-1;
			}
			snprintf(buf+strlen(buf),lenleft,"PS: NE: Current status: TX: %s (ind=%d), RX: %s, (ind=%d)\n",
					(txon?"on":"off"),txind,(rxon?"on":"off"),rxind);
			lenleft=blen-strlen(buf);
			snprintf(buf+strlen(buf),lenleft,"PS: NE: Next event is:\nPS: NE:\t%s (index %d,sched-rel=%.0f)\n",
					ltp_time_str(now+tmp),
					nextind,
					schedules[i].entries[nextind].reltime);
			lenleft=blen-strlen(buf);
			snprintf(buf+strlen(buf),lenleft,"PS: NE: \tsched-rel: %s\n",
					ltp_time_str(schedules[i].start+schedules[i].entries[nextind].reltime));
			lenleft=blen-strlen(buf);
			if (schedules[i].type==SENDT_CUE_EPH) {
				snprintf(buf+strlen(buf),lenleft,"PS: NE: \tDiff=%s,LTT=%f\n",
					nicediff(tmp),
					getltt(&schedules[i],schedules[i].start+schedules[i].entries[nextind].reltime));
				lenleft=blen-strlen(buf);
			} else {
				snprintf(buf+strlen(buf),lenleft,"PS: NE: \tDiff=%s\n",nicediff(tmp));
				lenleft=blen-strlen(buf);
			}
		} else {
			snprintf(buf+strlen(buf),lenleft,"PS: NE: Now is relative (%.0f), but there are NO events at alll\n",reltime);
			lenleft=blen-strlen(buf);
		}
	}
	// visibility stuff
	if (nvis_fnames) {
		if (nvis_fnames!=nvises) {
			snprintf(buf+strlen(buf),lenleft,"PS: Visibility oddness: %d files, %d data structures??",nvis_fnames,nvises);
			lenleft=blen-strlen(buf);
		}
		if (!nvises) {
			snprintf(buf+strlen(buf),lenleft,"PS: \nNo vis_infos!\n");
			lenleft=blen-strlen(buf);
		} else {
			snprintf(buf+strlen(buf),lenleft,"PS: Visibility (%d files):\n",nvises);
			lenleft=blen-strlen(buf);
			for (int i=0;i!=nvises;i++) {
				char *fn=vis_fnames[i];
				char *lttfn=vises[i].ltt_fname;
				snprintf(buf+strlen(buf),lenleft,"PS: Vis file: %s: (LTT: %s,start: %s)\n",fn,lttfn,ltp_time_str(vises[i].start));
				lenleft=blen-strlen(buf);
				snprintf(buf+strlen(buf),lenleft,"PS: \tfroms (%d): ",vises[i].nfrom);
				lenleft=blen-strlen(buf);
				for (int j=0;j!=vises[i].nfrom;j++) {
					char *nt=ltp_name_trans(vises[i].froms[j]);
					if (!nt) nt=vises[i].froms[j];
					snprintf(buf+strlen(buf),lenleft,"%s ",nt);
					lenleft=blen-strlen(buf);
				}
				snprintf(buf+strlen(buf),lenleft,"\nPS: \ttos (%d): ",vises[i].nto);
				lenleft=blen-strlen(buf);
				for (int j=0;j!=vises[i].nto;j++) {
					char *nt=ltp_name_trans(vises[i].tos[j]);
					if (!nt) nt=vises[i].tos[j];
					snprintf(buf+strlen(buf),lenleft,"%s ",nt);
					lenleft=blen-strlen(buf);
				}
				snprintf(buf+strlen(buf),lenleft,"\nPS: \tentries (%d):",vises[i].nentries);
				lenleft=blen-strlen(buf);
				int ents2print=min(vises[i].nentries,10);
				for (int j=0;j!=ents2print;j++) {
					vis_ent *ve=&vises[i].entries[j];
					char *nt=ltp_name_trans(vises[i].froms[ve->from]);
					if (!nt) nt=vises[i].froms[ve->from];
					snprintf(buf+strlen(buf),lenleft,"\nPS: \t\t%.0f,%s:", ve->reltime,nt);
					lenleft=blen-strlen(buf);
					for (int k=0;k!=vises[i].nto;k++) {
						snprintf(buf+strlen(buf),lenleft,"%s ", TFP(ve->bitrates[k]));
						lenleft=blen-strlen(buf);
					}
				}
				if (ents2print<vises[i].nentries) {
					snprintf(buf+strlen(buf),lenleft,"...(%d) omitted\n",vises[i].nentries-ents2print);
					lenleft=blen-strlen(buf);
				} else {
					snprintf(buf+strlen(buf),lenleft,"\n");
					lenleft=blen-strlen(buf);
				}
			}
		}
	}
	ltp_log_flush();
	return;
}


char *addrspec2dotsep(char *as)
{
	char dotsep[21]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};;
	int port=UDP_DEF_PORT;
	int slashnum=0;
	
	// map from a/8 to a.255.255.255
	// map from a.b/16 to a.b.255.255
	// map from a.b.c/24 to a.b.c.255
	// if there's a port, leave it be
	char *cp=strdup(as); // rest is TBD
	char *slash=NULL;
	char *colon=NULL;
	for (char *ch=cp;*ch;ch++) {
		if (*ch==':') colon=ch;
		if (*ch=='/') slash=ch;
	}
	if (colon) {
		port=atoi(colon+1);
		*colon='\0';
	}
	if (slash) {
		slashnum=atoi(slash+1);
		*slash='\0';
	}
	strncpy(dotsep,cp,21);
	if (slashnum) {
		switch (slashnum) {
			case 8: strncat(dotsep,".255.255.255",21); break;
			case 16: strncat(dotsep,".255.255",21); break;
			case 24: strncat(dotsep,".255",21); break;
			default:
				snprintf(ltpstr,LTPSTR_LEN,"Unsupported addr-spec %s - has to be /8/16 or /24\n",as);
				ltp_log(LTP_LOG_MAX,ltpstr);
				free(cp);	
				return(NULL);
		}
	}
	if (port!=UDP_DEF_PORT) {
		char tmp[20];
		snprintf(tmp,20,":%d",port);
		strncat(dotsep,tmp,21);
	}
	free(cp);
	return(strdup(dotsep));
}

// silly little thing to parse "N,thing1,...thingn"
static int parse_tofrom(char *crv, int *n, char ***arr, char *compstr)
{
	int rv=0;
	if (!crv || !n || !arr) LTPERR(LTP_BAD_PARAM);
	if (strlen(crv)!=strlen(compstr) || strncmp(crv,compstr,strlen(crv))) return(LTP_BAD_VALUE);
	crv=strtok_r(NULL,LTP_COMMA_SEP,&st_vis);
	if (!crv) return(LTP_BAD_VALUE);
	int count=atoi(crv);
	if (count<=0) return(LTP_BAD_VALUE);
	*n=count;
	char **list=(char**)calloc(count,sizeof(char*));
	if (!list) LTPERR(LTP_MEMORY);
	for (int i=0;i!=count;i++) {
		crv=strtok_r(NULL, LTP_COMMA_SEP,&st_vis);
		if (!crv) { free(list); return(LTP_BAD_VALUE);}
		list[i]=strdup(crv);
	}
	*arr=list;
	return(LTP_OK);
}

/*
 * @brief read in a visibilty file name
 * @param fname is the file name
 * @return zero or error
 *
 * Note that this could be called >1 time, since each separate schedule could use
 * a file (same file could even be used >1 time)
 */
static int load_vis(char *fname)
{
	int rv;
	// consider this read now - if we barf in a minute we always will
	if (!fname) LTPERR(LTP_BAD_PARAM);
	for (int i=0;i!=nvis_fnames;i++) {
		if (strlen(fname)==strlen(vis_fnames[i]) && !strncmp(fname,vis_fnames[i],strlen(fname))) {
			snprintf(ltpstr,LTPSTR_LEN,"Already loaded visibility information from fname=%s\n",fname);
			ltp_log(LTP_LOG_MAX,ltpstr);
			return(LTP_OK);
		}
	} 
	if (!nvis_fnames) {
		vis_fnames=(char**)calloc(1,sizeof(char*));
		if (!vis_fnames) LTPERR(LTP_MEMORY);
		vis_fnames[0]=strdup(fname);
		nvis_fnames=1;
	} else {
		vis_fnames=(char**)realloc(vis_fnames,(nvis_fnames+1)*sizeof(char*));
		if (!vis_fnames) LTPERR(LTP_MEMORY);
		vis_fnames[nvis_fnames]=strdup(fname);
		nvis_fnames++;
	}
	FILE *fp=fopen(fname,"r");
	if (!fp) {
		snprintf(ltpstr,LTPSTR_LEN,"Can't open visibility file fname=%s\n",fname);
		ltp_log(LTP_LOG_MAX,ltpstr);
		return(LTP_FILE);
	} else {
		snprintf(ltpstr,LTPSTR_LEN,"Loading visibility infomation from fname=%s\n",fname);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
	// usual story - read lines
	char line[MAXPATH];
	char *fgrv;
	char *lp;
	char *ltt_file=NULL;
	bool fromseen=false;
	bool toseen=false;
	bool startseen=false;
	bool lttseen=false;
	int froms=0;
	int ntos=0;
	char **ltos=NULL;
	char **lfroms=NULL;
	ltp_time_t	reltime;
	while ((fgrv=fgets(line,MAXPATH,fp))!=NULL) {
		if (strlen(line) > (MAXPATH-1)) { fclose(fp); LTPERR(LTP_BAD_VALUE); } /// line too long
		lp=line;
		// while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
		while (ltp_iswhitespace(*lp)) lp++;
		if (!strlen(lp)) continue; 
		if (*lp=='#') continue;
		if (*lp=='\n') continue;
		// expect FROM, TO and START first
		if (!lttseen || !fromseen || !toseen || !startseen) {
			ltp_time_t rectime; // time for this record
			bool absrec=false;
			char *crv=strtok_r(line,LTP_COMMA_SEP,&st_vis);
			if (!crv) CONTOUT;
			char *crv1=NULL;
			char *crv2=NULL;
			switch (toupper(crv[0])){
				case 'L': // LTT
					if (strlen(crv)!=strlen(LTT_STR) || strncmp(crv,LTT_STR,strlen(crv))) CONTOUT;
					ltt_file=strdup(crv+strlen(LTT_STR)+1);
					if (ltt_file[strlen(ltt_file)-1]=='\n') ltt_file[strlen(ltt_file)-1]='\0';
					lttseen=true;
					break;
				case 'F': // froms
					rv=parse_tofrom(crv,&froms,&lfroms,(char*)FROM_STR);
					if (rv) CONTOUT;
					fromseen=true;
					break;
				case 'T': // tos
					rv=parse_tofrom(crv,&ntos,&ltos,(char*)TO_STR);
					if (rv) CONTOUT;
					toseen=true;
					break;
				case 'S': // strart
					if (strlen(crv)!=strlen(START_STR) || strncmp(crv,START_STR,strlen(crv))) CONTOUT;
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_vis);
					if (!crv) CONTOUT;
					crv1=strdup(crv);
					crv2=strtok_r(crv1,LTP_EQ_SEP,&st_vis);
					if (!crv2) { free(crv1); CONTOUT; }
					switch(toupper(crv1[0])) {
						case 'A':
							if (strlen(crv2)!=strlen(ABS_STR) || strncmp(crv2,ABS_STR,strlen(crv2))) CONTOUT;
							crv2=strtok_r(NULL,LTP_EQ_SEP,&st_vis);
							if (!crv2) { free(crv1); CONTOUT; }
							rv=str2ltp_time_t(crv2,&rectime);
							if (rv) CONTOUT;
							absrec=true;
							break;
						case 'R':
							if (strlen(crv2)!=strlen(REL_STR) || strncmp(crv2,REL_STR,strlen(crv2))) CONTOUT;
							crv2=strtok_r(NULL,LTP_EQ_SEP,&st_vis);
							if (!crv2) { free(crv1); CONTOUT; }
							rv=parse_time_str(crv2,&rectime);
							if (rv) CONTOUT;
							absrec=false;
							break;
						default:
							CONTOUT;
					}
					free(crv1);
					startseen=true;
					break;
				default: CONTOUT;
			}
			if (lttseen && fromseen && toseen && startseen) { // ok to make a new vis_info now
				// setup this guy
				if (!vises) {
					vises=(vis_info*)calloc(1,sizeof(vis_info));
					if (!vises) LTPERR(LTP_MEMORY);
					nvises=1;
				} else {
					vises=(vis_info*)realloc(vises,(nvises+1)*sizeof(vis_info));
					if (!vises) LTPERR(LTP_MEMORY);
					nvises++;
				}
				vises[nvises-1].absstart=absrec;
				vises[nvises-1].start=rectime;
				vises[nvises-1].ltt_fname=ltt_file;
				vises[nvises-1].nto=ntos;
				vises[nvises-1].tos=ltos;
				vises[nvises-1].nfrom=froms;
				vises[nvises-1].froms=lfroms;
				vises[nvises-1].nentries=0;
				vises[nvises-1].entries=NULL;
			}
		} else {
			// a line of info: reltime,from-ind,vis-to1,vis-to2,...
			char *crv=strtok_r(line,LTP_COMMA_SEP,&st_vis);
			if (!crv) CONTOUT;
			if (strlen(crv)!=strlen(VIS_STR) || strncmp(crv,VIS_STR,strlen(crv))) CONTOUT;
			crv=strtok_r(NULL,LTP_COMMA_SEP,&st_vis);
			if (!crv) CONTOUT;
			rv=parse_time_str(crv,&reltime);
			if (rv) CONTOUT;
			if (reltime<0) CONTOUT;
			crv=strtok_r(NULL,LTP_COMMA_SEP,&st_vis);
			if (!crv) CONTOUT;
			int fromind=atoi(crv);
			if (fromind<=0 || fromind>froms) CONTOUT;
			int *lvises=(int*)calloc(ntos,sizeof(int));
			if (!lvises) LTPERR(LTP_MEMORY);
			for (int j=0;j!=ntos;j++) {
				crv=strtok_r(NULL,LTP_COMMA_SEP,&st_vis);
				if (!crv) 
					CONTOUT;
				lvises[j]=atoi(crv);
				if (lvises[j]<0) CONTOUT;
			}
			// add entry
			vis_ent *ve=vises[nvises-1].entries;
			int vn=vises[nvises-1].nentries;
			if (!ve) {
				ve=(vis_ent*)calloc(1,sizeof(vis_ent));
				if (!ve) LTPERR(LTP_MEMORY);
				vn=1;
			} else {
				ve=(vis_ent*)realloc(ve,(vn+1)*sizeof(vis_ent));
				if (!ve) LTPERR(LTP_MEMORY);
				vn++;
			}
			ve[vn-1].reltime=reltime;
			ve[vn-1].bitrates=lvises;
			ve[vn-1].from=fromind-1; // go back to zero-based index
			vises[nvises-1].entries=ve;
			vises[nvises-1].nentries=vn;
		}
	}
	
	if (vises && lttseen && fromseen && toseen && startseen) { // ok - got all headers and at least one entry
		return(LTP_OK);
	} else {
		LTPERR(LTP_BAD_VALUE);
	}
}

static ltp_time_t getdur(sc_entry *se)
{
	ltp_time_t	ret=0;
	for (int i=0;i!=se->ntags;i++) {
		if (strlen(se->tags[i])==strlen(DUR_STR) && !strncmp(se->tags[i],DUR_STR,strlen(DUR_STR))) {
			ret=atof(se->vals[i]);
		}
	}
	return(ret);
}

static int getnev(sc_entry *se)
{
	int	ret=0;
	for (int i=0;i!=se->ntags;i++) {
		if (strlen(se->tags[i])==strlen(NEV_STR) && !strncmp(se->tags[i],NEV_STR,strlen(NEV_STR))) {
			ret=atoi(se->vals[i]);
		}
	}
	return(ret);
}

static int getinds(char *from, char *to, sc_sched *sc, int *toindp, int *fromindp)
{
	if (!from || !to || !sc || !toindp || !fromindp) LTPERR(LTP_BAD_PARAM);
	if (!sc->vis) {
		*toindp=-1;
		*fromindp=-1;
		return(LTP_OK);
	}
	int toind=-1;
	ltpaddr to_in;
	int rv=str2ltpaddr(to,&to_in);
	if (rv) LTPERR(rv);
	ltpaddr from_in;
	rv=str2ltpaddr(from,&from_in);
	if (rv) LTPERR(rv);
	ltpaddr tmp;
	int fromind=-1;
	*toindp=toind;
	*fromindp=fromind;
	bool tointolist=false;
#ifdef LTP_EXTRA_EXTRA_VERBOSE_AND_BROKEN
	///TODO: figure out if this is needed & if so fix it
	snprintf(ltpstr,LTPSTR_LEN,"%s: from: %s, to: %s, sc-ind: %d, relwant: %.0f, dur: %.0f\n",
					__FUNCTION__, from, to, sind,relwant,dur);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	// which to are we? better be one of 'em
	for (int j=0;j!=sc->vis->nto;j++) {
		rv=str2ltpaddr(sc->vis->tos[j],&tmp);
		if (rv) LTPERR(rv);
		if (!ltpaddr_cmp(&to_in,&tmp,sizeof(to_in))) {
			toind=j;
			tointolist=true;
			break;
		}
	}
	if (toind==-1) {
		for (int j=0;j!=sc->vis->nfrom;j++) {
			rv=str2ltpaddr(sc->vis->froms[j],&tmp);
			if (rv) LTPERR(rv);
			if (!ltpaddr_cmp(&to_in,&tmp,sizeof(to_in))) {
				toind=j;
				break;
			}
		}
	}
	if (toind==-1) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: can't find peer %s in list\n",__FUNCTION__,to);
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(LTP_BAD_VALUE);
	}
	if (tointolist) {
		// then from better be in from list
		for (int j=0;j!=sc->vis->nfrom;j++) {
			rv=str2ltpaddr(sc->vis->froms[j],&tmp);
			if (rv) LTPERR(rv);
			if (!ltpaddr_cmp(&from_in,&tmp,sizeof(from_in))) {
				fromind=j;
				break;
			}
		}
		if (fromind==-1) {
			snprintf(ltpstr,LTPSTR_LEN,"%s: can't find peer %s in list\n",__FUNCTION__,from);
			ltp_log(LTP_LOG_MAX,ltpstr);
			LTPERR(LTP_BAD_VALUE);
		}
	} else {
		// to is in from list, so from better be in to list
		for (int j=0;j!=sc->vis->nto;j++) {
			rv=str2ltpaddr(sc->vis->tos[j],&tmp);
			if (rv) LTPERR(rv);
			if (!ltpaddr_cmp(&from_in,&tmp,sizeof(from_in))) {
				fromind=j;
				break;
			}
		}
		if (fromind==-1) {
			snprintf(ltpstr,LTPSTR_LEN,"%s: can't find peer %s in list\n",__FUNCTION__,from);
			ltp_log(LTP_LOG_MAX,ltpstr);
			LTPERR(LTP_BAD_VALUE);
		}
	}
	if (toind==-1 || fromind==-1) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: odd place to get - error though\n",__FUNCTION__);
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(LTP_BAD_VALUE);
	}
	if (!tointolist) { // have to reverse toind and fromind if I start out backwards
		int xxx=toind;
		toind=fromind;
		fromind=xxx;
	}
	*toindp=toind;
	*fromindp=fromind;
	return(LTP_OK);
}

/*
 * @brief search through the vis infos to find the next poss time for whatever
 * @param from is the entity who's scheulde we're processing
 * @param to is the peer
 * @param sc is the schedule
 * @param relwant is the relative time we want to start tx/rx-ing
 * @param dur is the duration we want 
 * @param nextstart is the actual next time we can start
 * @param nextend is the end of that period (may be shorter than dur due to invisibilty)
 * @return zero or error
 */
static int getnextvis(char *from, char *to, sc_sched *sc, 
	ltp_time_t relwant, ltp_time_t dur, ltp_time_t *nextstart, ltp_time_t *nextend, int sind)
{

	int toind, fromind;
	int rv=getinds(from,to,sc,&toind,&fromind);
	if (rv) LTPERR(rv);

	bool vbefore_done=false;
	bool vafter_done=false;
	vis_ent *von_before=NULL; // nearest "on" entry, can start before or after wanted
	vis_ent *voff_before=NULL; // next off, after the above
	vis_ent *von_after=NULL; // nearest "on" entry, can start before or after wanted
	vis_ent *voff_after=NULL; // nearest "on" entry, can start before or after wanted
	ltp_time_t	von_before_duration=0;
	ltp_time_t	von_after_duration=0;

	int entry=sc->vis->nentries-1;
	ltp_time_t	rel0=sc->vis->entries[0].reltime;
	while (sc->vis->entries[entry].reltime > relwant) {
		entry=(int)floor((relwant-rel0)/(sc->vis->entries[entry].reltime-rel0)*1000)*entry/1000;
	}
	// go back far enough that we can find a von_before
	while (entry && sc->vis->entries[entry].from!=fromind && !sc->vis->entries[entry].bitrates[toind]) {
		entry--;
	}
	for (int i=0;!vbefore_done && !vafter_done && i<sc->vis->nentries;i++) {
		if (sc->vis->entries[entry].from != fromind) {
			entry++; if (entry==sc->vis->nentries) entry=0;
			continue;
		}
		if (!vbefore_done && sc->vis->entries[entry].bitrates[toind] && sc->vis->entries[entry].reltime <= relwant) {
			von_before=&sc->vis->entries[entry];
		}
		if (!vbefore_done && !sc->vis->entries[entry].bitrates[toind]) {
			voff_before=&sc->vis->entries[entry];
		}
		if (sc->vis->entries[entry].bitrates[toind] && sc->vis->entries[entry].reltime > relwant) {
			vbefore_done=true;
			von_after=&sc->vis->entries[entry];
		}
		if (!sc->vis->entries[entry].bitrates[toind] && sc->vis->entries[entry].reltime > relwant) {
			voff_after=&sc->vis->entries[entry];
			vafter_done=true;
		}
		entry++; if (entry==sc->vis->nentries) entry=0;
	}
	// which to are we? better be one of 'em
	if (von_before && voff_before) {
		von_before_duration=voff_before->reltime-von_before->reltime;
	} else if (von_before) { // must only go of at the end
		von_before_duration=sc->lastentry-von_before->reltime;
	} else {
		von_before_duration=0;
	}
	if (von_after && voff_after) {
		von_after_duration=voff_after->reltime-von_after->reltime;
	} else if (von_after) {
		von_after_duration=sc->lastentry-von_after->reltime;
	} else {
		von_after_duration=0;
	}
	if (von_before && von_before->reltime+von_before_duration > relwant+dur) {
		*nextstart=relwant;
		*nextend=*nextstart+dur;
		return(LTP_OK);
	} 
	if (von_after && von_after_duration > dur) {
		*nextstart=von_after->reltime;
		*nextend=*nextstart+dur;
		return(LTP_OK);
	} 
	if (von_after) { // can't get what I want, maybe less is worth it?
		*nextstart=von_after->reltime;
		*nextend=*nextstart+von_after_duration;
	}	
	return(LTP_BAD_VALUE);
}

static int load_ltts(sc_sched *sc)
{
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"%s called\n",__FUNCTION__);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	// chew up the first lines of the ltt file
	ltp_time_t start=sc->vis->start; 
	int 		rv;
#define Y2K "2000-01-01T00:00:00"
	if (!arrarr) {
		arrarr=(lttinfo_t*)calloc(1,sizeof(lttinfo_t));
		if (!arrarr) LTPERR(LTP_MEMORY);
		nlttinfos=1;
		sc->vis->ltt_inf_ind=0;
	} else {
		lttinfo_t *tmp=(lttinfo_t*)realloc(arrarr,(nlttinfos+1)*sizeof(lttinfo_t));
		if (!tmp) LTPERR(LTP_MEMORY);
		arrarr=tmp;
		sc->vis->ltt_inf_ind=nlttinfos;
		memset(&(arrarr[nlttinfos]),0,sizeof(lttinfo_t));
		nlttinfos++;
	}
	SET_LTT(sc);
	rv=str2ltp_time_t((char*)Y2K,&epoch);
	if (rv) LTPERR(rv);
	// want relative time in file to start from
	ltp_time_t	relstart=start-epoch; 
	if (relstart<=0) LTPERR(LTP_BAD_VALUE);
	ltp_time_t	relend=relstart+sc->dutycycle;

	// now transform into a set of on/off durations 
	// need an LTT info file
	FILE *fp=fopen(sc->vis->ltt_fname,"r");
	if (!fp) {
		snprintf(ltpstr,LTPSTR_LEN,"Can't open file %s, exiting %s\n",
					sc->vis->ltt_fname,__FUNCTION__);
		ltp_log(LTP_LOG_MAX,ltpstr);
		return(LTP_OK); // leave all alone!
	}
	char line[MAXPATH];
	char *fgrv;
	char *lp;
	bool done=false;
	bool started=false;
	int lineno=0;
	int prelineno=0;
	int postlineno=0;
	while ((fgrv=fgets(line,MAXPATH,fp))!=NULL && !done ) {
		lineno++;
		if (strlen(line) > (MAXPATH-1)) { fclose(fp); LTPERR(LTP_BAD_VALUE); } /// line too long
		lp=line;
		while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
		while (ltp_iswhitespace(*lp)) lp++;
		if (!strlen(lp)) continue; 
		if (*lp=='#') continue;
		if (*lp=='\n') continue;
		if (!started) {
			prelineno++;
			// <tag> reltime ltt
			char *sr;
			char *crv=strtok_r(line," \n",&sr);
			if (!crv) continue;
			crv=strtok_r(NULL," \n",&sr);
			if (!crv) continue;
			ltp_time_t relt=atof(crv);
			if (relt<=0) continue;
			if (relt>=relstart) started=true;
		} 
		if (started) {
			postlineno++;
			// memory mgmt
			if (arrents==arrdim) { // more space needed
				arrdim+=100;
				relarr=(ltp_time_t*)realloc(relarr,arrdim*sizeof(ltp_time_t));
				if (!relarr) LTPERR(LTP_MEMORY);
				lttarr=(ltp_time_t*)realloc(lttarr,arrdim*sizeof(ltp_time_t));
				if (!lttarr) LTPERR(LTP_MEMORY);
			}
			// read line <tag> reltime ltt
			char *sr1;
			char *crv=strtok_r(line," \n",&sr1);
			if (!crv) continue;
			crv=strtok_r(NULL," \n",&sr1);
			if (!crv) continue;
			ltp_time_t relt=atof(crv);
			if (relt<=0) continue;
			if (relt>=relend) done=true;
			crv=strtok_r(NULL," \n",&sr1);
			if (!crv) continue;
			ltp_time_t ltt=atof(crv);
			if (ltt<=0) continue;
			// assign
			relarr[arrents]=relt;
			lttarr[arrents]=ltt;
			arrents++;
		}
	}
	// put back into arrarr
	arrarr[sc->vis->ltt_inf_ind].arrdim=arrdim;
	arrarr[sc->vis->ltt_inf_ind].arrents=arrents;
	arrarr[sc->vis->ltt_inf_ind].relarr=relarr;
	arrarr[sc->vis->ltt_inf_ind].lttarr=lttarr;
	arrarr[sc->vis->ltt_inf_ind].epoch=epoch;
#ifdef LTP_VERBOSE
	{
		char *epochstr=strdup(ltp_time_str(epoch));
		char *startstr=strdup(ltp_time_str(start));
		char *endstr=strdup(ltp_time_str(start+sc->dutycycle));
		if (arrents) {
			snprintf(ltpstr,LTPSTR_LEN,"%s:read %d lines of %s to get to start, then %d more\n\tepoch=%s\n\tstart=%s (LTT: %f)\n\tend=  %s (LTT: %f)",
						__FUNCTION__,prelineno,sc->vis->ltt_fname,postlineno,epochstr,startstr,lttarr[0],endstr,lttarr[arrents-1]);
		} else {
			snprintf(ltpstr,LTPSTR_LEN,"%s:read %d lines of %s to get to start, then %d more\n\tepoch=%s\n\tstart=%s (LTT: NONE!)\n\tend=  %s (LTT: NONE!)",
						__FUNCTION__,prelineno,sc->vis->ltt_fname,postlineno,epochstr,startstr,endstr);
		}
		ltp_log(LTP_LOG_MAX,ltpstr);
		free(epochstr);
		free(startstr);
	}
#endif
	fclose(fp);
	return(LTP_OK);
}


// this just hide memory allocs
static int newev(int *alloced, int *nents, sc_entry **ents,sc_entry *newent)
{
	static int incr=100;
	if (!alloced || !nents || !ents || !newent) LTPERR(LTP_BAD_PARAM);
	if (!*nents) {
		sc_entry *lents=(sc_entry*)realloc(*ents,(*nents+incr)*sizeof(sc_entry));
		if (!lents) LTPERR(LTP_MEMORY);
		*alloced=*nents+incr;
		*ents=lents;
	}
	if (*nents >= *alloced) {
		sc_entry *lents=(sc_entry*)realloc(*ents,(*nents+incr)*sizeof(sc_entry));
		if (!lents) LTPERR(LTP_MEMORY);
		*alloced=*nents+incr;
		*ents=lents;
	}
	(*ents)[*nents]=*newent;
	*nents=*nents+1;
	return(LTP_OK);
}

#define EPH_RX_GRACE 2

/*!
 * @brief extend the set of events for an EPH schedule given the LTT info
 * @param sc is the schedule
 * @return zero or error
 */
static int eph_extend(sc_sched *sc,int sind)
{
	// check inputs
	if (!sc) LTPERR(LTP_BAD_PARAM);
	if (!sc->vis) LTPERR(LTP_BAD_PARAM);
	int rv=load_ltts(sc);

	// now create new events for each scheduled event 
	// scheduled events specify a start, duration and how many 2xLTT repeats to use
	// for follow-up tx/rx events
	// we only schedule tx/rx events for times when vises says we should
	// so things may be postponed when compared to the input schedule
	// we try to accumulate the requested duration even if it takes >1 set of 
	// events 
	// if a durataion has to be spread, we still generate corresponding opposite events
	// for the end of each contact, but the NEV events only start from the end of the
	// fully used duration.
	// lastly we sort the extended set of events, then remove any duplicates (or 
	// totally overlapped events)

	int ne=sc->nentries;
	// new ones to replace the old
	sc_entry *nse=NULL;
	int nne=0;
	int nnalloced=0;
	for (int i=0;i<ne;i++) {
		// process each
		sc_entry *se=&sc->entries[i];
		sc_entry  newent;
		ltp_time_t	duration=getdur(se);
		int nevlim=getnev(se);
		if (!nevlim) nevlim=1; // need original in any case
		ltp_time_t	nextstart=0;
		ltp_time_t	nextend=se->reltime;
		for (int j=0;j<nevlim;j++) {
			
			rv=getnextvis(mestr,se->peers,sc,nextend,duration,&nextstart,&nextend,sind);
			if (rv) break; // presumably no visibilty?

			if (nextend-nextstart==duration) { 
				// add an "on"
				newent.entry_type=se->entry_type;
				newent.reltime=floor(nextstart);
				newent.peers=se->peers;
				newent.tx=se->tx;
				if (!newent.tx) newent.reltime-=EPH_RX_GRACE; // grab a second's more rx on-time
				newent.on=true;
				newent.ntags=se->ntags;
				newent.tags=se->tags;
				newent.vals=se->vals;
				rv=newev(&nnalloced,&nne,&nse,&newent);
				if (rv) LTPERR(rv);
				// and an "off"
				newent.entry_type=se->entry_type;
				newent.reltime=ceil(nextend);
				newent.peers=se->peers;
				newent.tx=se->tx;
				newent.on=false;
				newent.ntags=se->ntags;
				newent.tags=se->tags;
				newent.vals=se->vals;
				rv=newev(&nnalloced,&nne,&nse,&newent);
			} else {
				// gotta spread the duration over >1 on/off cycle
				ltp_time_t	nstart=0;
				ltp_time_t	nend=0;
				ltp_time_t	lastone=nextend;
				ltp_time_t	durleft=duration-(nextend-nextstart);
				int toomanygoes=100;
				while (!rv && durleft>0 && toomanygoes--) {
					rv=getnextvis(mestr,se->peers,sc,lastone,durleft,&nstart,&nend,sind);
					if (rv) break;
					newent.entry_type=se->entry_type;
					newent.reltime=floor(nstart);
					newent.peers=se->peers;
					newent.tx=se->tx;
					if (!newent.tx) newent.reltime-=EPH_RX_GRACE; // grab a second's more rx on-time
					newent.on=true;
					newent.ntags=se->ntags;
					newent.tags=se->tags;
					newent.vals=se->vals;
					rv=newev(&nnalloced,&nne,&nse,&newent);
					if (rv) LTPERR(rv);
					// and an "off"
					newent.entry_type=se->entry_type;
					newent.reltime=ceil(nextend);
					newent.peers=se->peers;
					newent.tx=se->tx;
					newent.on=false;
					newent.ntags=se->ntags;
					newent.tags=se->tags;
					newent.vals=se->vals;
					rv=newev(&nnalloced,&nne,&nse,&newent);
					lastone=nend;
					durleft=durleft-(nend-nstart);
				}
				if (!toomanygoes) {
					snprintf(ltpstr,LTPSTR_LEN,"%s:%d toomanygoes!\n",__FUNCTION__,__LINE__);
					ltp_log(LTP_LOG_MAX,ltpstr);
				}
				nextstart=nstart;
				nextend=nend;

			}
			// so now we've allocated "duration" worth of "on", we want an amount of opposite time
			// a couple of LTT's later
			// we're gonna go for duration/2's worth of time (for no particular reason!)
			// we first  see what's the ltt
			ltp_time_t ltt=getltt(sc,sc->start+nextstart);
#ifdef LTP_VERBOSE
			// just once to allow us verify that routers and apps are in-whack
			static bool ltt_logged=false;
			if (!ltt_logged) {
				snprintf(ltpstr,LTPSTR_LEN,"At %s LTT is %f\n",
						ltp_time_str(sc->start+nextstart),ltt);
				ltp_log(LTP_LOG_MAX,ltpstr);
				ltt_logged=true;
			}
#endif
			// then figure when next the other guy can start talking back (after the signal's
			// arrived with him)
			rv=getnextvis(mestr,se->peers,sc,nextstart+ltt,duration/2,&nextstart,&nextend,sind);
			if (rv) break; // presumably no visibilty?
			if ((nextend-nextstart)==(duration/2)) {
				// then I want to be "on" again another ltt later
				newent.entry_type=se->entry_type;
				newent.reltime=floor(nextstart+ltt);
				newent.peers=se->peers;
				newent.tx=!se->tx;
				if (!newent.tx) newent.reltime-=EPH_RX_GRACE; // grab a second's more rx on-time
				newent.on=true;
				newent.ntags=se->ntags;
				newent.tags=se->tags;
				newent.vals=se->vals;
				rv=newev(&nnalloced,&nne,&nse,&newent);
				if (rv) LTPERR(rv);
				// and off again at nextend+ltt
				newent.entry_type=se->entry_type;
				newent.reltime=ceil(nextend+ltt);
				newent.peers=se->peers;
				newent.tx=!se->tx;
				newent.on=false;
				newent.ntags=se->ntags;
				newent.tags=se->tags;
				newent.vals=se->vals;
				rv=newev(&nnalloced,&nne,&nse,&newent);
			} else {
				// crap - have to loop to fill duration/2's worth of time
				// gotta spread the duration over >1 on/off cycle
				ltp_time_t	nstart=0;
				ltp_time_t	nend=0;
				ltp_time_t	lastone=nextend;
				ltp_time_t	durleft=(duration/2)-(nextend-nextstart);
				int toomanygoes=100;
				while (!rv && durleft>0 && toomanygoes--) {
					rv=getnextvis(mestr,se->peers,sc,lastone,durleft,&nstart,&nend,sind);
					if (rv) break;
					newent.entry_type=se->entry_type;
					newent.reltime=floor(nstart);
					newent.peers=se->peers;
					newent.tx=!se->tx;
					if (!newent.tx) newent.reltime-=EPH_RX_GRACE; // grab a second's more rx on-time
					newent.on=true;
					newent.ntags=se->ntags;
					newent.tags=se->tags;
					newent.vals=se->vals;
					rv=newev(&nnalloced,&nne,&nse,&newent);
					if (rv) LTPERR(rv);
					// and an "off"
					newent.entry_type=se->entry_type;
					newent.reltime=ceil(nextend);
					newent.peers=se->peers;
					newent.tx=!se->tx;
					newent.on=false;
					newent.ntags=se->ntags;
					newent.tags=se->tags;
					newent.vals=se->vals;
					rv=newev(&nnalloced,&nne,&nse,&newent);
					lastone=nend;
					durleft=durleft-(nend-nstart);
				}
				if (!toomanygoes) {
					snprintf(ltpstr,LTPSTR_LEN,"%s:%d toomanygoes!\n",__FUNCTION__,__LINE__);
					ltp_log(LTP_LOG_MAX,ltpstr);
				}
				nextstart=nstart;
				nextend=nend;
			GOTHERE
			}
			// ok now we've added events to get us 1 duration's worth of stuff 
			// next is to go again starting at nextend
			// that happens at the top of the loop
			GOTHERE
		}
			GOTHERE
	}
			GOTHERE
	// now ditch old entries
	free(sc->entries);
			GOTHERE
	// and adopt new
	sc->nentries=nne;
	sc->entries=nse;
	return(LTP_OK);
}

#define ONLY_RELTIME 100

/*!
 * @brief check if two entries duplicate one another
 * @param a is one entry
 * @param b is the other
 * @return zero for equal, non-zero otherwise
 */
static int entry_cmp(sc_entry *a, sc_entry *b)
{
	if (a->tx != b->tx) return(2);
	if (a->on != b->on) return(3);
	if (!a->peers || !b->peers) return(4);
	if (strlen(a->peers)!=strlen(b->peers)) return(5);
	if (!ltp_same(a->peers,b->peers)) return(6);
	if (a->reltime != b->reltime) return(ONLY_RELTIME);
	return(0);
}

static int clean_dups(sc_sched *sc, int sind)
{
	if (!sc->nentries) {
		snprintf(ltpstr,LTPSTR_LEN,"Schedule %d has zero entries!!\n",sind);
		ltp_log(LTP_LOG_MAX,ltpstr);
		return(LTP_OK);
	}
	// last pass for duplicates
	sc_entry *se=sc->entries;
	int ns=sc->nentries;
	int dups=0;
	int contigs=0;
	int writehead=0;
	for (int readhead=1;readhead<ns;readhead++) {
		while (readhead<ns && !entry_cmp(&se[writehead],&se[readhead])) {
			dups++;
			readhead++;
		}  
		// check for contiguous ones
		if (entry_cmp(&se[writehead],&se[readhead])==ONLY_RELTIME) {
			// only differ in reltime, so its by definition a contig!
			contigs++;
		} else {
			se[++writehead]=se[readhead];
		}
	}
	sc->nentries-=(dups+contigs);
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Removed %d duplicates and %d contigs from schedule %d (used to be %d)\n",dups,contigs,sind,ns);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	return(LTP_OK);
}

/*!
 * @brief read in a schedule file
 * @param fname is the file name
 * @return zero or error
 */
static int sendt_read_file(char *fname)
{
	int rv;
	char *lfname;
	int nstartsseen=0;
	int nendsseen=0;

	if (!fname) lfname=def_fname;
	else lfname=fname;
	FILE *fp=fopen(lfname,"r");
	if (!fp) return(LTP_FILE);
	// now guess if the file's text or xml
	// 1st non blank == "<" says try xml (TODO: LATER), otherwise proceed as text
	char ch;
	while (ltp_iswhitespace(ch=fgetc(fp)));
	if (ch=='<') {
		snprintf(ltpstr,LTPSTR_LEN,"Trying to parse %s as an xml file\n",lfname);
		ltp_log(LTP_LOG_MAX,ltpstr);
		fclose(fp);
		rv=sendt_try_xml(lfname);
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"Failed to parse %s as an xml file\n",lfname);
			ltp_log(LTP_LOG_MAX,ltpstr);
			// back where we started
		} else {
			return(LTP_OK);
		}
	} 
	fclose(fp);
	fp=fopen(lfname,"r");
	if (!fp) LTPERR(LTP_FILE);
	snprintf(ltpstr,LTPSTR_LEN,"Trying to parse %s as a text file\n",lfname);
	ltp_log(LTP_LOG_MAX,ltpstr);
	char line[MAXPATH];
	bool versionseen=false;
	bool nschedulesseen=false;
	bool startsseen=false;
	bool meseen=false;
	bool allendsseen=false;
	char *fgrv;
	char *lp;
	while ((fgrv=fgets(line,MAXPATH,fp))!=NULL) {
		if (strlen(line) > (MAXPATH-1)) { fclose(fp); LTPERR(LTP_BAD_VALUE); } /// line too long
		lp=line;
		// while (*lp && *lp!='\n' && (*lp==' '|| *lp=='\t')) lp++; // chew whitespace
		while (ltp_iswhitespace(*lp)) lp++;
		if (!strlen(lp)) continue; 
		if (*lp=='#') continue;
		if (*lp=='\n') continue;
		if (!versionseen || !nschedulesseen || !meseen) { // find these two before schedule info
			char *crv=strtok_r(line,LTP_EQ_SEP,&st_sched);
			if (!crv) CONTOUT;
			if (ltp_same(crv,(char*)VERSION_STR)) { // version
				versionseen=true;
				// nothing more to do for now, since this is version 1.0 code:-)
			} else if (ltp_same(crv,(char*)NS_STR)) { // nschedules
				crv=strtok_r(NULL,LTP_EQ_SEP,&st_sched);
				if (!crv) CONTOUT;
				int new_nschedules=atoi(crv);
				if (new_nschedules>LTP_MAXSCHED) {
					snprintf(ltpstr,LTPSTR_LEN,"File %s calls for %d schedules, max is %d - exiting\n",
								fname,new_nschedules,LTP_MAXSCHED);
					ltp_log(LTP_LOG_MAX,ltpstr);
				}
				kill_sched();
				schedules=(sc_sched*)calloc(new_nschedules,sizeof(sc_sched));
				if (!schedules) LTPERR(LTP_MEMORY);
				nschedules=new_nschedules;
				nschedulesseen=true;
			} else if (ltp_same(crv,(char*)ME_STR)) {
				crv=strtok_r(NULL,LTP_EQ_SEP,&st_sched);
				if (!crv) CONTOUT;
				mestr=strdup(crv);
				meseen=true;
			}
		} else {
			char *crv=strtok_r(line,LTP_COMMA_SEP,&st_sched);
			if (!crv) { CONTOUT; }
			if (!startsseen) { // look for schedule start info - has to preceed entries
				if (!ltp_same(crv,(char*)START_STR)) CONTOUT;
				crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
				if (!crv) CONTOUT;
				int schedno=atoi(crv);
				if (schedno<=0 || schedno>nschedules) CONTOUT;
				// abs/rel time to start
				crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
				if (!crv) CONTOUT;
				ltp_time_t reltime=-1;
				ltp_time_t	abstime=0;
				int	stype=-1;
				char *this_vis=NULL;
				switch(crv[0]) {
					case 'R': // relative
						if (strlen(crv)<=strlen(REL_STR) || strncmp(crv,REL_STR,strlen(REL_STR))) CONTOUT;
						rv=parse_time_str(crv+strlen(REL_STR)+1,&reltime);
						if (rv) CONTOUT;
						if (reltime<=0) CONTOUT;
						stype=SENDT_CUE_REL;
						break;
					case 'A': // absolute
						if (strlen(crv)<=strlen(ABS_STR) || strncmp(crv,ABS_STR,strlen(ABS_STR))) CONTOUT;
						rv=str2ltp_time_t(crv+4,&abstime);
						if (rv) CONTOUT;
						stype=SENDT_CUE_ABS;
						break;
					case 'E':
						if (strlen(crv)<=strlen(EPH_STR) || strncmp(crv,EPH_STR,strlen(EPH_STR))) CONTOUT;
						this_vis=crv+strlen(EPH_STR)+1;
						rv=load_vis(this_vis);
						if (rv && rv!=LTP_FILE) LTPERR(rv);
						stype=SENDT_CUE_EPH;
						break;
					default:
						CONTOUT;
				}
				// Duration
				crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
				if (!crv) 
					CONTOUT;
				if (strlen(crv)<=5) CONTOUT;
				if (crv[0]!=DUTY_STR[0]) CONTOUT;
				if (crv[1]!=DUTY_STR[1]) CONTOUT;
				if (crv[2]!=DUTY_STR[2]) CONTOUT;
				if (crv[3]!=DUTY_STR[3]) CONTOUT;
				if (crv[4]!='=') CONTOUT;
				// parse dutycycle string
				ltp_time_t	dutycycle;
				rv=parse_time_str(crv+5,&dutycycle);
				// all done create the schedule 
				schedules[schedno-1].type=stype;
				if (stype==SENDT_CUE_ABS) {
					schedules[schedno-1].start=abstime;
				} else if (stype==SENDT_CUE_REL) { // other two the same
					schedules[schedno-1].start=(ltp_time_t)reltime;
				} else if (stype==SENDT_CUE_EPH) {
					schedules[schedno-1].vis_fname=strdup(this_vis);
					// copy up the start time from the vis info, rel/abs info's there too, but not here sorry
					if (vises) {
						schedules[schedno-1].start=vises[nvises-1].start;
						schedules[schedno-1].vis=&vises[nvises-1];
					} else {
						snprintf(ltpstr,LTPSTR_LEN,"Can't access visibility information - assuming all all visible\n");
						ltp_log(LTP_LOG_MAX,ltpstr);
						vis_all_always.start=ltp_now();
						schedules[schedno-1].start=vis_all_always.start;
						schedules[schedno-1].vis=&vis_all_always;
					}
				}
				schedules[schedno-1].dutycycle=dutycycle;
				if (++nstartsseen==nschedules) startsseen=true;
			} else if (!allendsseen) { // look for entries and/or ENDs
				if (ltp_same(crv,(char*)COMMS_STR)) {
					// schedno (int)
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					if (!crv) CONTOUT;
					int schedno=atoi(crv);
					if (schedno<=0 || schedno>nschedules) CONTOUT;
					ltp_time_t reltime;  
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					if (!crv) CONTOUT;
					rv=parse_time_str(crv,&reltime);
					if (rv) CONTOUT;
					if (reltime<0 || reltime>schedules[schedno-1].dutycycle) CONTOUT;
					// host[:port] or netmask
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					if (!crv) CONTOUT;
					char *peers=addrspec2dotsep(crv); // we're leaking this much for now (in error cases)!! TODO: fix LATER
					if (!peers) CONTOUT;
					// TX/RX
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					if (!crv) CONTOUT;
					bool ltx=false;
					if (ltp_same(crv,(char*)TX_STR)) { ltx=true;}
					else if (ltp_same(crv,(char*)RX_STR)) { ltx=false;}
					else CONTOUT;
					// ON/OFF
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					if (!crv) CONTOUT;
					bool lon=false;
					if (ltp_same(crv,(char*)ON_STR)) { lon=true;}
					else if (ltp_same(crv,(char*)OFF_STR)) { lon=false;}
					else CONTOUT;
					// ntags
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					int ntags=0;
					char **ltags=NULL;
					char **lvals=NULL;
					if (crv) {
						// Other tags unsorted RATE,IF,MTU
						ntags=atoi(crv);
						if (ntags<=0) CONTOUT;
						ltags=(char**)calloc(ntags,sizeof(char*));
						if (!ltags) LTPERR(LTP_MEMORY);
						lvals=(char**)calloc(ntags,sizeof(char*));
						if (!lvals) { free(ltags); LTPERR(LTP_MEMORY); }
						bool cleanup=false;
						for (int i=0;!cleanup && i!=ntags;i++) {
							crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
							if (!crv) cleanup=true;
							else {
								ltags[i]=strdup(crv);
								char *cp=ltags[i];
								while (*cp && *cp!='=') cp++;
								if (*cp=='=') {
									lvals[i]=strdup(cp+1);
									*cp='\0';
								} else cleanup=true;
							}
						}
						if (cleanup) {
							for (int i=0;i!=ntags;i++) {
								if (ltags[i]) free(ltags[i]);
								if (lvals[i]) free(lvals[i]);
							}
							free(ltags);
							free(lvals);
							CONTOUT;
						}
					}
					// create new entry
					if (!schedules[schedno-1].nentries) {
						schedules[schedno-1].entries=(sc_entry*)calloc(1,sizeof(sc_entry));
						if (!schedules[schedno-1].entries) LTPERR(LTP_MEMORY);
						schedules[schedno-1].nentries=1;
					} else {
						sc_entry *se=(sc_entry*)realloc(schedules[schedno-1].entries,(schedules[schedno-1].nentries+1)*sizeof(sc_entry));
						if (!se) LTPERR(LTP_MEMORY);
						schedules[schedno-1].entries=se;
						schedules[schedno-1].nentries++;
					}
					sc_entry *se=&schedules[schedno-1].entries[schedules[schedno-1].nentries-1];
					se->entry_type=COMMS_ENTRY;
					se->reltime=reltime;
					se->peers=peers;
					se->tx=ltx;
					se->on=lon;
					se->ntags=ntags;
					se->tags=ltags;
					se->vals=lvals;
				} else { // look for END's or just chew up material that I don't care about
					if (!ltp_same(crv,(char*)END_STR)) CONTOUT;
					// schedno
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					if (!crv) CONTOUT;
					int schedno=atoi(crv);
					if (schedno<=0 || schedno>nschedules) CONTOUT;
					// reltime (float)
					crv=strtok_r(NULL,LTP_COMMA_SEP,&st_sched);
					if (!crv) CONTOUT;
					ltp_time_t lastentry;
					rv=parse_time_str(crv,&lastentry);
					if (rv) CONTOUT;
					if (lastentry<0 || lastentry>schedules[schedno-1].dutycycle) CONTOUT;
					schedules[schedno-1].lastentry=lastentry;
					if (++nendsseen==nschedules) allendsseen=true;
				}
			}
		}
	}
	// fix up visibilty info agsin - can be moved by realloc (line 784)
	// and earlier schedule entries will point to old spot - not good
	// this is not a good fix, but I've no time this week
	for (int i=0;i!=nschedules;i++) {
		if (schedules[i].type==SENDT_CUE_EPH) {
			for (int vi=0;vi!=nvises;vi++) {
				if (strlen(vis_fnames[vi])==strlen(schedules[i].vis_fname) &&
					!strcmp(vis_fnames[vi],schedules[i].vis_fname)) {
					schedules[i].vis=&vises[vi];
				}
			}
		}
	}
	// sort schedules
	for (int i=0;i!=nschedules;i++) {
		// sort each schedule
		// no pressing need for efficiency here yet - bubble sort
		if (schedules[i].type==SENDT_CUE_EPH) {
			// go through generating subsequent events for each configured one
			rv=eph_extend(&schedules[i],i);
			if (rv) LTPERR(rv);
		}
		sc_entry tmp;
		sc_entry *se=schedules[i].entries;
		int ns=schedules[i].nentries;
		for (int j=0;j!=ns;j++) {
			for (int k=j;k!=ns;k++) {
				if (se[j].reltime > se[k].reltime) {
					tmp=se[j];
					se[j]=se[k];
					se[k]=tmp;
				}
			}
		}

		rv=clean_dups(&schedules[i],i);
		if (rv) LTPERR(rv);

	}
#ifdef LTP_VERBOSE
	//  print schedules to log
	char pbuf[20*MAXPATH];
	print_scheds(pbuf,20*MAXPATH);
	ltp_log(LTP_LOG_MAX,pbuf);
	ltp_log_flush();
#endif
	return(LTP_OK);
}

/*!
 * @brief re-initialise shcedule loading
 * @param fname schedule file name
 * @return zero or error
 *
 * when fname is NULL, then the current setup is re-initialised - files
 * re-read, structures re-inited, etc.
 *
 * this is meant to be called from main, before doing anything else
 * but is called internally once if necessary
 */
int sendt_init(char *fname)
{
	static bool sendt_inited=false;
	if (sendt_inited) return(LTP_OK);
	int rv=sendt_read_file(fname);
	if (rv) {
		if (rv==LTP_FILE) {
			return(rv);
		} else {
			sendt_inited=true; // no recovering from that error whatever it was
			LTPERR(rv);
		}
	}
	sendt_inited=true;
	return(LTP_OK);
}

/*!
 * @brief tells a node which other nodes are listening now (so we can Tx to them if there're packets)
 * @param now is the time I need to know about
 * @param nlisteners on input is the size of the listeners array, and on output is the number of real entries in the array
 * @param listeners is an array (caller allocated) of the LTP addresses of the currently listening nodes
 * @param tx is a bool - true implies I want to know about tx, false about rx
 * @return zero or error
 */
int sendt_who(ltp_time_t now, int *cnt, char **listp, bool tx)
{
	int rv;
	ltp_time_t	reltime=0;
	ltp_time_t	tmp=0;
	int lcnt=0;
	for (int i=0;i!=nschedules;i++) {
		if (!schedules[i].nentries) {
			continue;
		}
		reltime=getrel(&schedules[i],now);
		// if (reltime > schedules[i].lastentry) CONTOUT;
		// ok so we know we've a reltime for this
		// so search for the listeners/talkers around then
		float lon=-1;
		sc_entry laston;
		// try guess where we are assuming events are more or less uniformly distributed
		int entry=schedules[i].nentries-1;
		sc_entry *se=&schedules[i].entries[entry];
		while (entry && se->reltime>reltime) {
			entry=(int)floor(reltime/se->reltime*1000)*entry/1000;
			se=&schedules[i].entries[entry];
		}
		// go forward to 1st time beyond reltime, then backwards
		// to see if we're on or off at reltime
		while (entry != schedules[i].nentries-1 && se->reltime<reltime) {
			entry++;
			se=&schedules[i].entries[entry];
		}
		// now se->reltime >= reltime
		// so we step back until we hit a match for our guy - this could wrap around to
		// the previous cycle though, so be careful
		bool found=false;
		while (!found && entry>=0) {
			if (se->reltime<=reltime && se->tx==tx) {
				found=true;
			} else {
				entry--;
				if (entry>=0) {
					se=&schedules[i].entries[entry];
				}
			}
		}
		if (found) {
			if (se->on) {
				lcnt++;
				if (lcnt<*cnt) {
					listp[lcnt-1]=se->peers;
				} else {
					LTPERR(LTP_BAD_PARAM);
				}
			} // otherwise its off and we're done
		} else {
			// start from the end and search back until we get
			entry=schedules[i].nentries-1;
			se=&schedules[i].entries[entry];
			while (!found && entry>=0 && se->reltime > reltime) {
				if (se->tx==tx) {
					found=true;
				} else {
					entry--;
					if (entry>=0) {
						se=&schedules[i].entries[entry];
					}
				}
			}
			if (found) {
				if (se->on) {
					lcnt++;
					if (lcnt<*cnt) {
						listp[lcnt-1]=se->peers;
					} else {
						LTPERR(LTP_BAD_PARAM);
					}
				} // otherwise its off and we're done
			}
		}

#ifdef BROKENCRAP		
		// this is nonsense - should be going backwards!
		// go back to 1st entry for that time
		while (entry && schedules[i].entries[entry-1].reltime==se->reltime) {
			entry--;
			se=&schedules[i].entries[entry];
		}
		for (int j=0;se->reltime<=reltime && entry!=schedules[i].nentries && j!=schedules[i].nentries;j++) {
			// have we found it or gone beyond it?
			// found if reltime is between the last "on" and this "off"
			if (lon==-1 && se->tx==tx && se->on && reltime >= se->reltime )  {
				lon=se->reltime;
				laston=*se;
			}
			if (lon!=-1 && se->tx==tx && !se->on && reltime < se->reltime && lon <= reltime) {
				// add that peer
				lcnt++;
				if (lcnt<*cnt) {
					listp[lcnt-1]=se->peers;
				} else {
					LTPERR(LTP_BAD_PARAM);
				}
				lon=-1;
			}
			entry++;
			if (entry!=schedules[i].nentries) {
				se=&schedules[i].entries[entry];
			}
		}
		// can happen that lon is still on
		if (lon!=-1) {
			lcnt++;
			if (lcnt<*cnt) {
				listp[lcnt-1]=laston.peers;
			} else {
				LTPERR(LTP_BAD_PARAM);
			}
		}
#endif
	}

	*cnt=lcnt;
	return(LTP_OK);
}

/*!
 * @brief tells a node which other nodes are listening now (so we can Tx to them if there're packets)
 * @param now is the time I need to know about
 * @param nlisteners on input is the size of the listeners array, and on output is the number of real entries in the array
 * @param listeners is an array (caller allocated) of the LTP addresses of the currently listening nodes
 * @return zero or error
 */
int sendt_whos_listening(ltp_time_t now, int *cnt, char **listp)
{
	int rv=sendt_init(NULL);
	if (rv==LTP_FILE) {
		return(rv);
	} else if (rv) {
		LTPERR(rv);
	}
	rv=sendt_who(now,cnt,listp,false);
#ifdef LTP_VERBOSE
	static int lastlcnt=0;
	static bool ever=false;
	if (!ever || lastlcnt!=*cnt) {
		ever=true;
		snprintf(ltpstr,LTPSTR_LEN,"%s lists %d at %s\n",
				__FUNCTION__,*cnt,ltp_time_str(now));
		ltp_log(LTP_LOG_MAX,ltpstr);
		lastlcnt=*cnt;
		for (int i=0;i!=*cnt;i++) {
			snprintf(ltpstr,LTPSTR_LEN,"\tlistener %d: %s\n", i,listp[i]);
			ltp_log(LTP_LOG_MAX,ltpstr);
		}
		ltp_log_flush();
	}
#endif
	return(rv);
}

/*!
 * @brief returns my LTP address
 * @param me is the caller allocated string in which to put the address
 * @return zero or error
 *
 * Note that my address can change as a result of rescheduling - so I better check it occasionally.
 * I'd say that per-reboot should be ok.
 */
char *sendt_whoami(void)
{
	int rv=sendt_init(NULL);
	if (rv==LTP_FILE) {
		return(NULL);
	} else if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"%s error from sendt_init (%d)\n",__FUNCTION__,rv);
		ltp_log(LTP_LOG_MAX,ltpstr);
		return(NULL);
	}
#ifdef LTP_EXTRA_EXTRA_VERBOSE
	if (mestr) {
		snprintf(ltpstr,LTPSTR_LEN,"%s says I'm %s\n",__FUNCTION__,mestr);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
#endif
	return(mestr);
}

/*!
 * @brief tells a node which other nodes ought to be talking now (so we can decrement timers)
 * @param now is the time I need to know about
 * @param ntalkers on input is the size of the listeners array, and on output is the number of real entries in the array
 * @param talkers is an array (caller allocated) of the LTP addresses of the currently chatting nodes
 * @return zero or error
 */
int sendt_whos_talking(ltp_time_t now, int *cnt,char **listp)
{
	int rv=sendt_init(NULL);
	if (rv==LTP_FILE) {
		return(rv);
	} else if (rv) {
		LTPERR(rv);
	}
	rv=sendt_who(now,cnt,listp,true);
#ifdef LTP_VERBOSE
	static int lastlcnt=0;
	static bool ever=false;
	if (!ever || lastlcnt!=*cnt) {
		ever=true;
		snprintf(ltpstr,LTPSTR_LEN,"%s lists %d at %s\n", __FUNCTION__,*cnt,ltp_time_str(now));
		ltp_log(LTP_LOG_MAX,ltpstr);
		for (int i=0;i!=*cnt;i++) {
			snprintf(ltpstr,LTPSTR_LEN,"\ttalker %d: %s\n", i,listp[i]);
			ltp_log(LTP_LOG_MAX,ltpstr);
		}
		ltp_log_flush();
		lastlcnt=*cnt;
	}
#endif
	return(rv);
}

/*!
 * @brief figure out an LTT for a node to which I'm about to transmit
 * @param peer is the node's address
 * @param now is the current time
 * @param ltt is (a ptr to) the LTT at that time
 * @return zero or error
 */
int sendt_ltt(ltpaddr *peer, ltp_time_t now, ltp_time_t *ltt)
{
	if (!peer || !ltt) LTPERR(LTP_BAD_PARAM);
	ltp_time_t ans=(ltp_time_t)100/(ltp_time_t)1000; // 100ms
	bool done=false;
	int rv;
	char *peerstr=strdup(ltpaddr2str(peer));
	ltpaddr mea;
	rv=ltp_cue_whoami(&mea);
	char *lme=strdup(ltpaddr2str(&mea));
	for (int i=0;!done && i!=nschedules;i++) {
		int toind, fromind;
		rv=getinds(peerstr,lme,&schedules[i],&toind,&fromind);
		if (rv && rv!=LTP_BAD_VALUE) LTPERR(rv);
		if (toind!=-1 && fromind!= -1) {
			// gotcha, try find LTT now
			ltp_time_t reltime=getrel(&schedules[i],now);
			ans=getltt(&schedules[i],schedules[i].start+reltime);
			done=true;
		}
	}
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"sendt_ltt: %sltt from %s to %s at %s is %g\n",
				(done?"":"fake "),peerstr,lme,ltp_time_str(now),ans);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	free(lme);
	free(peerstr);
	*ltt=ans;
	return(LTP_OK);
}
