/*!
 * @file test_comms.cc
 * @brief Tests the protocol handlers for LTP and L2 
 * @version $Revision: 1.5 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <getopt.h>
#include <unistd.h> // for sleep()
#include "ltp_base.h"
#include "ltp.h"
#include "ltp_l2.h"


static void usage()
{
	printf("Tests the protocol handlers for LTP and L2 (currently UDP)\n");
	printf("test_comms [-l layer] [ -c | -s | -b ] [ -1 ] [-m size] [-r repeats]\n");
	printf("where:\n");
	printf("\t [-l layer] determines the layer/level to test (can be UDP or LTP) [LTP]\n");
	printf("\t [-c | -s | -b ] indicates if this instance is to run as a client, server or both [b]\n");
	printf("\t [-1 ] indicates that the 127.0.0.1 loopback address should be used instead of the layer2 loopback [off]\n");
	printf("\t [-2 ] indicates that the layer 2 loopback address should be used [on for layer 2 tests]\n");
	printf("\t [-3 ] indicates that the layer 3 loopback address should be used [on for layer 3 tests]\n");
	printf("\t [-d addr ] indicates that the addr (in decimal dot.sep notation) should be used as the destination for transmission.\n");
	printf("\t [-m size ] indicates that the test LTP block (ignored for UDP testing) should have size bytes [4567]\n");
	printf("\t [-r repeats] how many additional messages to send [0]\n");
	exit(-1);
}


/// length of a string describing the parameters of this run
#define DESCRIPTION_LEN 1000
/// a string describing the parameters of this run
static char	description[DESCRIPTION_LEN];

/// the layer/level to test (can be 2 or 3) [2]
static int level2test=3;

/// The mode in which to run tests. 0=>client;1=>server;2=>both [2]
static int mode=2;

/// Whether to loopback using 127.0.0.1
static bool l1loop=false;

/// Whether to loopback in our layer2 or using 127.0.0.1
static bool l2loop=true;

/// Whether to loopback in layer 3 or not
static bool l3loop=true;

/// The destination address for transmissions
static char *dest;

/// Whether a destination address was provided via [-d addr]
static bool destprovided=false;

/// Default message size want this >1 MTU (i.e. >1492)
#define DEF_MSGLEN 4567 

/// Actual message length
static int msglen=DEF_MSGLEN;

/// Whether a user defined message length was supplied
static bool msglengiven=false;

/// Whether re-send a 2nd, and/or subsequent messages as well
static int msgrepeats=0;

/// Just helpful strings (indexed by mode var)
static char* modestrs[]={
		"a client",
		"a server",
		"both a client and server"
};

#define TAKEAMO(mos) \
	{ printf("Taking a 'mo...");  \
	sleep(1); \
	for (int hgthgt=0;hgthgt!=mos;hgthgt++)  { printf("%d...",hgthgt); sleep(1); fflush(stdout); } \
	printf("\n"); fflush(stdout); }

static int testlevel2(void) 
{
	int rv;
	// initialise
	rv=ltp_l2_init(defcbs_p,0);
	if (rv) LTPERR(rv);
	// turn on
	unsigned int mtu;
	switch (mode) {
		case 0: rv=ltp_l2_on(&mtu,LTP_L2S_ON_TX); break;
		case 1: rv=ltp_l2_on(&mtu,LTP_L2S_ON_RX); break;
		case 2: rv=ltp_l2_on(&mtu,LTP_L2S_ON_BOTH); TAKEAMO(3) break;
		default: printf("Bad mode! (%d)",mode);usage();exit(-1);
	}
	if (rv) LTPERR(rv);
	// send
	if (mode==0 || mode ==2 ) {
		ltp_l2_buffer	buf;
		char *strval="This is the buffer to send.";
		buf.heap=true;
		buf.l2alloc=false;
		buf.len=strlen(strval)+1;
		buf.val=(unsigned char*)strval;
		ltp_l2_id		me;
		char *me_addr_str="127.0.0.1"; // doesn't matter though
		if (l2loop) {
			rv=ltp_l2_send(&buf,&l2lb);
			for (int jj=0;jj!=msgrepeats;jj++) {
				if (rv) break;
				buf.val[0]='1'+(jj%10);
				buf.val[2]=' ';
				rv=ltp_l2_send(&buf,&l2lb);
			}
		} else if (destprovided) {
			me.pname=dest;
			bzero(&me.addr,sizeof(ltpaddr));
			me.addr.sock.sin_family=AF_LTP;
			me.addr.sock.sin_port=UDP_DEF_PORT;
			me.addr.sock.sin_addr.s_addr=inet_addr(dest);
			me.addr_len=sizeof(me.addr);
			rv=ltp_l2_send(&buf,&me);
			for (int jj=0;jj!=msgrepeats;jj++) {
				if (rv) break;
				buf.val[0]='1'+(jj%10);
				buf.val[2]=' ';
				rv=ltp_l2_send(&buf,&me);
			}
		} else { // 127.0.0.1 loopback
			me.pname=me_addr_str;
			bzero(&me.addr,sizeof(ltpaddr));
			me.addr.sock.sin_family=AF_LTP;
			me.addr.sock.sin_port=UDP_DEF_PORT;
			me.addr.sock.sin_addr.s_addr=inet_addr(me_addr_str);
			me.addr_len=sizeof(me.addr);
			rv=ltp_l2_send(&buf,&me);
			for (int jj=0;jj!=msgrepeats;jj++) {
				if (rv) break;
				buf.val[0]='1'+(jj%10);
				buf.val[2]=' ';
				rv=ltp_l2_send(&buf,&me);
			}
		}
		if (rv) LTPERR(rv);
		printf("Sent 28 bytes.\n");
	}
	if (mode==1 || mode==2) {
		// wait 'till 1st packet rx'd
		while (!rxqueue_p->times[0])
			sleep(1); // busy, sleepy wait!
		printf("Got %d bytes\n",rxqueue_p->bufs[0].len);
		if (msgrepeats) {
			int jj=msgrepeats;
			int oldpos=rxqueue_p->position;
			while (jj--) {
				while (rxqueue_p->position!=oldpos) { printf(".");sleep(1);}
				printf("Got %d bytes\n",rxqueue_p->bufs[0].len);
			}
		}
	}
	// turn off
	rv=ltp_l2_off(LTP_L2_OFF_BOTH);
	if (rv) LTPERR(rv);
	// closedown
	rv=ltp_l2_shutdown();
	if (rv) LTPERR(rv);
	return (LTP_OK);
}

/*!
 * @fn int test_ltp(int nsteps,int *steps)
 * @brief do an LTP test
 * @param nsteps is the number of test steps to do here
 * @param steps is the list of which test steps to perform
 * @return zero or error as usual
 *
 * This function acts as a fairly typical client and/or
 * server sockets code tester. The wrinkle is that each
 * step has a number and we will only do the numbers 
 * that are requested. That way we can test use of LTP
 * across multiple runs of the program. E.g. client step
 * 1 is to create a socket and bind. Step 2 is to send 
 * a packet. Step 3 to send another. Step 4 to close
 * the socket. We want to be able to do that in 4
 * separate runs of the program. (Note: it doesn't work
 * that way actually, but you should get the idea.)
 */
static int test_ltp(int nsteps, int *steps)
{
	/* 
	 * So how to implement that ? We'll write the code inline
	 * then make all vars global and break the code into a
	 * separate fnc for each step, and then make an array of
	 * function pointers to which we call.
	 */
	int c_sock=0; 
	int rv;
	ltpaddr my_addr={AF_LTP,UDP_DEF_PORT,0x0100007f,"\0\0\0\0\0\0\0"};
	ltpaddr_len my_addr_len=sizeof(my_addr);
	if (destprovided) {
		my_addr.sock.sin_addr.s_addr=inet_addr((const char*)dest);
	}
	int s_sock; ///< server LTP socket
	char *msgbuf;
	msgbuf=(char*)malloc(msglen+1);
	if (!msgbuf) { 
		printf("Memory error %s:%d - exiting\n",__FILE__,__LINE__);
		exit(-1);
	}
	char *msg=msgbuf;
	unsigned char *rmsg;
	rmsg=(unsigned char*)malloc(2*msglen+100);
	if (!rmsg) { 
		printf("Memory error %s:%d - exiting\n",__FILE__,__LINE__);
		exit(-1);
	}
	int rmsglen=msglen*2;
	ltpaddr from;
	ltpaddr_len fromlen;

#define PSTEP \
	snprintf(ltpstr,LTPSTR_LEN,"test_ltp step %d\n",steps[i]); \
	ltp_log(LTP_LOG_MAX,ltpstr);

	for (int i=0;i!=nsteps;i++) {
		switch(steps[i]) {

				case 0:
					PSTEP;
					c_sock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
					if (!c_sock) LTPERR(LTP_ERROR);
					break;

				case 1:
					PSTEP;
					rv=ltp_bind(c_sock,&my_addr,my_addr_len);
					if (rv) LTPERR(LTP_ERROR);
					break;

				case 2:
					PSTEP;
					s_sock=ltp_socket(AF_LTP,SOCK_LTP_SESSION,0);
					if (!s_sock) LTPERR(LTP_ERROR);
					break;

				case 3:
					PSTEP;
					rv=ltp_bind(s_sock,&ltpany,LTPANYLEN);
					if (rv) LTPERR(LTP_ERROR);
					break;

				case 4:
					PSTEP;
					for (int ii=0;ii!=msglen;ii++) msgbuf[ii]='A'+(ii%26);
					msgbuf[msglen]='\0'; // EOS - just so strlen works ok in the debugger
					for (int ii=0;ii!=msglen;ii++) msgbuf[ii]='A'+(ii%26);
					msgbuf[msglen]='\0'; // EOS - just so strlen works ok in the debugger
					if (l3loop) {
						rv=ltp_sendto(c_sock,msg,msglen,0,&ltplb,LTPLBLEN);
					} else {
						rv=ltp_sendto(c_sock,msg,msglen,0,&my_addr,my_addr_len);
					}
					if (rv==-1) LTPERR(LTP_ERROR);
					snprintf(ltpstr,LTPSTR_LEN,"Sent %d bytes\n",msglen+1);
					ltp_log(LTP_LOG_MAX,ltpstr);
					if (msgrepeats) {
						for (int jj=0;jj!=msgrepeats;jj++) {
							msg[0]='1'+(jj%10);
							msg[1]=' ';
							if (l3loop) {
								rv=ltp_sendto(c_sock,msg,msglen,0,&ltplb,LTPLBLEN);
							} else {
								rv=ltp_sendto(c_sock,msg,msglen,0,&my_addr,my_addr_len);
							}
						}
					}
					break;

				case 5:
					{
						PSTEP;
						int s5lim=5;
						// TAKEAMO(3);
						for (int s5it=0;s5it!=s5lim;s5it++) {
							rv=ltp_ping();
							TAKEAMO(1)
							if (rv) LTPERR(LTP_ERROR);
							ltp_log(LTP_LOG_MAX,"Pinged LTP");
						}
					}
					break;

				case 6:
					PSTEP;
					rv=ltp_recvfrom(s_sock,rmsg,msglen*2,0,&from,&fromlen);
					if (rv==-1) LTPERR(LTP_ERROR);
					rmsglen=rv;
					snprintf(ltpstr,LTPSTR_LEN,"Received %d bytes\n",rmsglen);
					printf("ltp_recvfrom(): %s\n",ltpstr);
					ltp_log(LTP_LOG_MAX,ltpstr);
					break;

				case 7:
					PSTEP;
					ltp_close(c_sock);
					c_sock=0;
					break;

				case 8:
					PSTEP;
					ltp_close(s_sock);
					s_sock=0;
					break;

				case 9:
					PSTEP;
					rv=ltp_save();
					if (rv) LTPERR(LTP_ERROR);
					break;

				case 10: // hang about until recvfrom produces something
					{
					PSTEP;
					rv=0; int count=0;
					printf("%d recvfrom calls done\n",count);
					sleep(10); 
					while (rv==0) { 
						rv=ltp_recvfrom(s_sock,rmsg,msglen*2,0,&from,&fromlen); 
						count++;
						if (!(count%10)) printf("%d recvfrom calls done\n",count);
						sleep(1); 
					}
					rmsglen=rv;
					snprintf(ltpstr,LTPSTR_LEN,"Received %d bytes\n",rmsglen);
					printf("ltp_recvfrom(): %s\n",ltpstr);
					ltp_log(LTP_LOG_MAX,ltpstr);
					if (rv<0) LTPERR(LTP_ERROR);
					rv=ltp_recvfrom(s_sock,rmsg,msglen*2,0,&from,&fromlen); 
					if (rv<0) LTPERR(LTP_ERROR);
					}
					break;

				case 11: // hang about until a keybd hit
					PSTEP;
					{
					rv=0; 
					int ch=0;
					printf("Hit 'q' to stop.\n");
					while (!rv && ch!='q' && ch!='Q') { rv=ltp_ping(); sleep(1);  ch=getchar(); }
					if (rv==-1) LTPERR(LTP_ERROR);
					rmsglen=rv;
					}
					break;

				default:
					LTPERR(LTP_BAD_PARAM);
					break;
		}

	}

	return(LTP_OK);
}

/*!
 * @fn int main(int argc, char *argv[])
 * @brief see usage
 * @param argc you should know (uses getopt)
 * @param argv you should know (uses getopt)
 * @return something or other
 */
int main(int argc, char *argv[])
{
	ltp_log_init("test_comms.log",LTP_LOG_MAX);
	ltp_log_timestamp("Starting test_comms");

    extern char *optarg;/* used by getopt */
    extern int optind, opterr, optopt;
	int c;
	int errflag=0;

    while ((c = getopt(argc, argv, "123bcd:hl:m:r:s?")) != EOF) {
    	switch(c) {
			case '1':
					l1loop=true;
					l2loop=false;
					l3loop=false;
					break;
			case '2':
					l1loop=false;
					l2loop=true;
					l3loop=false;
					break;
			case '3':
					l1loop=false;
					l2loop=false;
					l3loop=true;
					break;
			case 'b':
					mode=2; break;
			case 'd':
					l2loop=false;
					l3loop=false;
					destprovided=true;
					dest=strdup(optarg);
					break;
			case 'c':
					mode=0; break;
			case 'l':
					if (!strncasecmp("LTP",optarg,3)) level2test=3;
					if (!strncasecmp("UDP",optarg,3)) level2test=2;
					break;
			case 'm':
					msglengiven=true;
					msglen=atoi(optarg);
					if (msglen<2) msglen=2;
					break;
			case 'r':
					msgrepeats=atoi(optarg);
					if (msgrepeats<0) msgrepeats=0;
					break;
			case 's':
					mode=1; break;
			case 'h':
			case '?':
			default:
				errflag++;
				usage();
				break;
    	}
    }
	snprintf(description,DESCRIPTION_LEN,"Testing Layer %d as %s",level2test, modestrs[mode]);
	ltp_log(LTP_LOG_MAX,description);
	if (level2test==2) testlevel2();
	if (level2test==3) {
		int allsteps[11]={0,1,2,3,4,10,7,8,9,11,5};
		switch (mode) {
			case 2:  // both
					test_ltp(10,allsteps); break;
			case 1:  // server
					allsteps[0]=2;
					allsteps[1]=3;
					allsteps[2]=5;
					allsteps[3]=10;
					allsteps[4]=11;
					test_ltp(5,allsteps); break;
			case 0:  // client
					allsteps[0]=0;
					allsteps[1]=1;
					allsteps[2]=4;
					allsteps[3]=5;
					allsteps[4]=7;
					allsteps[5]=11;
					test_ltp(6,allsteps); break;
			default:
					printf("Silly mode!\n");
					usage();
		} 
	}
	ltp_log_timestamp("Finished test_comms");
	return 0;
}

