/*!
 * @file test_sdnv.cc
 * @brief Tests the sdnv codec
 * @version $Revision: 1.9 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <getopt.h>
#include <sys/stat.h>
#include <unistd.h> // for sleep()
#include "ltp_base.h"
#include "ltp.h"
#include "ltp_l2.h"
#include "ltp_codec.h"

static void usage()
{
	printf("Tests the sdnv codec\n");
	printf("test_sdnv [-b] [-1|-2|-3|-4|-m|-z] [-i ifile] [-o ofile]\n");
	printf("where:\n");
	printf("\t [-b] means go backwards - treat innput as encoded file [off]\n");
	printf("\t [-i ifile] means take the input (binary) bytes from ifile [no default]\n");
	printf("\t [-o ofile] means place the output (binary) bytes into ofile [no default]\n");
	printf("\t [-1|-2|-3|-4] means use sample #1, #2 etc from the LTP spec (sec 2(17)). [no default]\n");
	printf("\t [-m] means use a 100 byte sample.\n");
	printf("\t [-z] means use a 100 byte sample of all zeros (except the MSB).\n");
	exit(-1);
}

/*!
 * @fn int main(int argc, char *argv[])
 * @brief see usage
 * @param argc you should know (uses getopt)
 * @param argv you should know (uses getopt)
 * @return something or other
 */
int main(int argc, char *argv[])
{
	int result=1;
	ltp_log_init("test_sdnv.log",LTP_LOG_MAX);
	ltp_log_timestamp("Starting test_sdnv");

    extern char *optarg;/* used by getopt */
    extern int optind, opterr, optopt;
	int c;
	int errflag=0;

	char *ifile=NULL;
	char *ofile=NULL;
	bool ifilegiven=false;
	bool ofilegiven=false;
	bool decoder=false;
	int	 sample2use=1;
    while ((c = getopt(argc, argv, "1234bi:mo:zh?")) != EOF) {
    	switch(c) {
			case '1': sample2use=1; break;
			case '2': sample2use=2; break;
			case '3': sample2use=3; break;
			case '4': sample2use=4; break;
			case 'm': sample2use=5; break;
			case 'z': sample2use=6; break;
			case 'b': decoder=true; break;
			case 'i':
				ifilegiven=true;
				ifile=strdup(optarg);
				break;
			case 'o':
				ofilegiven=true;
				ofile=strdup(optarg);
				break;
			case 'h':
			case '?':
			default:
				errflag++;
				usage();
				break;
    	}
    }

	if (decoder && !ifilegiven) {
		fprintf(stderr,"-b requires an input file (-i)\n\n");
		usage();
	}

	unsigned char defibuf1[]={ 0x0A, 0xBC } ;
	unsigned char defibuf2[]={ 0x12, 0x34 } ;
	unsigned char defibuf3[]={ 0x42, 0x34 } ;
	unsigned char defibuf5[100] = {
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
	};
	unsigned char defibuf4[]={ 0x7F } ;
	unsigned char defobuf[6]; // too big but whatever
	int ilen=2;
	int olen=6;
	int rv;

	unsigned char *obuf=NULL;
	unsigned char *ibuf=defibuf1;
	switch (sample2use) {
		case 2: ibuf=defibuf2; ilen=2; olen=3; break;
		case 3: ibuf=defibuf3; ilen=2; olen=3; break;
		case 4: ibuf=defibuf4; ilen=1; olen=3; break;
		case 5: ibuf=defibuf5; ilen=100; olen=200; break;
		case 6: 	
			ilen=100; 	
			ibuf=(unsigned char*)malloc(ilen); 	
			memset(ibuf,0x00,ilen); 	
			ibuf[0]=0x80;	
			olen=500; 	
			break;
		case 1:
		default:
			ibuf=defibuf1;
	}

	if (ifilegiven) {
		FILE *fp=fopen(ifile,"rb");
		struct stat lstat;
		int rv=fstat(fileno(fp),&lstat);
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"Cannot open file %s\n",ifile);
			ltp_log(LTP_LOG_MAX,ltpstr);
			fprintf(stderr,ltpstr);
			if (ofilegiven) {
				FILE *opf=fopen(ofile,"w");
				fprintf(opf,ltpstr);
			}
			exit(-1);
		}
		unsigned int fsz=lstat.st_size;
		unsigned char *fbuf=(unsigned char*) malloc(fsz);
		if (!fbuf) {
			snprintf(ltpstr,LTPSTR_LEN,"memory error allocating %d bytes %s\n",fsz,ifile);
			ltp_log(LTP_LOG_MAX,ltpstr);
			fprintf(stderr,ltpstr);
			if (ofilegiven) {
				FILE *opf=fopen(ofile,"w");
				fprintf(opf,ltpstr);
			}
			exit(-1);
		}
		rv=fread(fbuf,1,fsz,fp);
		if (rv!=fsz) {
			snprintf(ltpstr,LTPSTR_LEN,"Only read %d bytes from file %s (expecting %d)\n",rv,ifile,fsz);
			ltp_log(LTP_LOG_MAX,ltpstr);
			fprintf(stderr,ltpstr);
			if (ofilegiven) {
				FILE *opf=fopen(ofile,"w");
				fprintf(opf,ltpstr);
			}
			exit(-1);
		}
		fclose(fp);
		ibuf=fbuf;
		ilen=fsz;
		olen=(ilen/7)*8+100;
		snprintf(ltpstr,LTPSTR_LEN,"Input is %d bytes from file %s\n",rv,ifile,fsz);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}

	int keepilen=ilen;

	if (!decoder) {
		// standard enc/dec/compare test
		snprintf(ltpstr,LTPSTR_LEN,"Standard test.");
		ltp_log(LTP_LOG_MAX,ltpstr);

		obuf=(unsigned char*)malloc(olen);
		memset(obuf,0xAF,olen);

		snprintf(ltpstr,LTPSTR_LEN,"ibuf");
		ltp_log_buffer(ltpstr,ilen,ibuf);

		rv=sdnv_enc(ilen,ibuf,&olen,obuf);
		if (rv) LTPERR(rv);

		snprintf(ltpstr,LTPSTR_LEN,"obuf");
		ltp_log_buffer(ltpstr,olen,obuf);

	} else {
		// dec/enc/compare test
		snprintf(ltpstr,LTPSTR_LEN,"Decoding test.");
		ltp_log(LTP_LOG_MAX,ltpstr);

		obuf=(unsigned char*)malloc(olen);
		memset(obuf,0xAF,olen);

		snprintf(ltpstr,LTPSTR_LEN,"ibuf");
		ltp_log_buffer(ltpstr,ilen,ibuf);

		rv=sdnv_dec(&ilen,ibuf,&olen,obuf);
		if (rv) LTPERR(rv);

		snprintf(ltpstr,LTPSTR_LEN,"obuf");
		ltp_log_buffer(ltpstr,olen,obuf);
	}

	if (ofilegiven) {
		FILE *of=fopen(ofile,"wb");
		if (!of) {
			fprintf(stderr,"Can't write output file: %s\n",ofile);
			LTPERR(LTPC_C_OTHER);
		}
		int wrote=fwrite(obuf,1,olen,of);
		if (wrote!=olen) {
			fprintf(stderr,"Error writing output file: %s (%d of %d written)\n",ofile,wrote,olen);
			LTPERR(LTPC_C_OTHER);
		}
		fclose(of);
	}

	int rolen=2*olen;// too big but whatever
	unsigned char *robuf; 
	robuf=(unsigned char*)malloc(rolen);

	if (!decoder) {
		rv=sdnv_dec(&olen,obuf,&rolen,robuf);
		if (rv) LTPERR(rv);
	} else {
		rv=sdnv_enc(olen,obuf,&rolen,robuf);
		if (rv) LTPERR(rv);
	}

	snprintf(ltpstr,LTPSTR_LEN,"robuf");
	ltp_log_buffer(ltpstr,rolen,robuf);

	if (keepilen!=rolen) LTPERR(LTPC_C_OTHER);

	for (int i=0;i!=rolen;i++) {
		if (ibuf[i]!=robuf[i]) LTPERR(LTPC_C_OTHER);
	}

	result=0;

	fprintf(stderr,"SDNV test worked fine.\n");

	ltp_log_timestamp("Finished test_sdnv");
	return result;
}

