/*!
 * @file test_sdnv_cgi.cc
 * @brief Tests the sdnv codec
 * @version $Revision: 1.9 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


#include <getopt.h>
#include <sys/stat.h>
#include <unistd.h> // for sleep()
#include <ctype.h> // for toupper()
#include "ltp_base.h"
#include "ltp.h"
#include "ltp_l2.h"
#include "ltp_codec.h"

#define LOGFILE "test_sdnv_cgi.log"
#define MAXFILE 1000000

#ifdef __cplusplus
extern "C" {
#endif
#include "cgic.h"
#ifdef __cplusplus
}
#endif

static void usage()
{
	printf("Tests the sdnv codec as a CGI program\n");
	printf("Input is complicated CGI stuff - look to https://down.dsg.cs.tcd.ie/ltplib/web/\n");
	printf("if you've got an account\n");
	exit(-1);
}

static int readcgifile(unsigned char **buf, int *blen);

#undef HVMACRO
#ifdef HVMACRO
#define HV(XX) ( ( (XX) >= '0' && (XX) < '9' ) ? (XX)-'0' : (XX)-'A'+10 )
#else
static unsigned char HV(unsigned char x) {
	if (x >= '0' && x <= '9') return (x-'0');
	else if (x >= 'A' && x <='F') return (x-'A'+10);
	else if (x >= 'a' && x <='f') return (x-'a'+10);
	else return(0);
}
#endif

/*!
 * @fn int main(int argc, char *argv[])
 * @brief see usage
 * @param argc you should know (uses getopt)
 * @param argv you should know (uses getopt)
 * @return something or other
 */
int cgiMain()
{
	int rv=0;
	ltp_log_init(LOGFILE,LTP_LOG_MAX);
	ltp_log_timestamp("Starting test_sdnv");

	cgiHeaderContentType("text/html");
    /* Top of the page */
    fprintf(cgiOut, "<HTML><HEAD>\n");
    fprintf(cgiOut, "<TITLE>SDNV test</TITLE></HEAD>\n");
    fprintf(cgiOut, "<BODY><H1>SDNV test</H1>\n");

	bool dofile=true;
	char dofilestr[1024];
	if (cgiFormString("dofile",dofilestr,1024) != cgiFormSuccess) {
		fprintf(cgiOut,"<p>Not doing a file.</p>");
		dofile=false;
	}

	bool already_encoded=false;
	if (cgiFormCheckboxSingle("decode") == cgiFormSuccess) {
		already_encoded=true;
	}

	bool view_log=false;
	if (cgiFormCheckboxSingle("verbose") == cgiFormSuccess) {
		view_log=true;
	}

	bool hexoutput=false;
	if (cgiFormCheckboxSingle("hexoutput") == cgiFormSuccess) {
		hexoutput=true;
	}

	unsigned char *ibuf;
	int ilen;
	if (dofile) {
		rv=readcgifile(&ibuf,&ilen);
		if (rv) { // error
    		fprintf(cgiOut, "<p>File error - crapola</p></BODY></HTML>\n");
			return rv;
		}
	} else {
		
    	fprintf(cgiOut, "<p>Hexval handling</p>\n");
		char lbuf[1024];
		if (cgiFormString("hexval",lbuf,1024)!=cgiFormSuccess) {
    		fprintf(cgiOut, "<p>Hexval error - crapola</p></BODY></HTML>\n");
			return rv;
		}
		// decode to binary!
		int llen=strlen(lbuf);
		ibuf=(unsigned char*)malloc(ilen);
		ilen=0;
		bool topnibble=true;
		for (int i=0;i!=llen;i++) {
			switch (toupper(lbuf[i])) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				case 'A':
				case 'B':
				case 'C':
				case 'D':
				case 'E':
				case 'F':
					if (topnibble) {
						ibuf[ilen]=HV(toupper(lbuf[i]))<<4;
						topnibble=false;
						break;
					} else {
						ibuf[ilen++] += HV(toupper(lbuf[i]));
						topnibble=true;
						break;
					}
				default:
					; // nothing
			}
		}

		fprintf(cgiOut,"<p>Input length %d</p>",ilen);
		// output ibuf
		char *foo=(char*) malloc(3*ilen+4*ilen/72+1);
		if (!foo)  {
			fprintf(cgiOut,"<p>Memory error (%d)</p>",__LINE__);
    		fprintf(cgiOut, "</BODY></HTML>\n");
			return(rv);
		}
		int bar;
		char *foop=foo;
		for (int i=0;i!=ilen;i++) {
			if (!(i%36)) { *foop++='\n'; *foop++='\t'; }
			if (!(i%4)) *foop++=' ';
			bar=ibuf[i]>>4;
			*foop++=HEXDIGIT(bar);
			bar=ibuf[i]&0x0f;
			*foop++=HEXDIGIT(bar);
		}
		*foop='\0';
		fprintf(cgiOut,"<pre>");
		fprintf(cgiOut,foo);
		fprintf(cgiOut,"</pre>");
	}

	// fprintf(cgiOut,"<p>Input length %d</p>",ilen);

	unsigned char *obuf;
	int olen=2*ilen; // safety in numbers!

	if (!already_encoded) {
		// standard enc/dec/compare test
		snprintf(ltpstr,LTPSTR_LEN,"Standard test.");
		ltp_log(LTP_LOG_MAX,ltpstr);

		obuf=(unsigned char*)malloc(olen);
		memset(obuf,0xAF,olen);

		snprintf(ltpstr,LTPSTR_LEN,"ibuf");
		ltp_log_buffer(ltpstr,ilen,ibuf);

		rv=sdnv_enc(ilen,ibuf,&olen,obuf);
		if (rv) LTPERR(rv);

		snprintf(ltpstr,LTPSTR_LEN,"obuf");
		ltp_log_buffer(ltpstr,olen,obuf);

	} else {
		// dec/enc/compare test
		snprintf(ltpstr,LTPSTR_LEN,"Decoding test.");
		ltp_log(LTP_LOG_MAX,ltpstr);

		obuf=(unsigned char*)malloc(olen);
		memset(obuf,0xAF,olen);

		snprintf(ltpstr,LTPSTR_LEN,"ibuf");
		ltp_log_buffer(ltpstr,ilen,ibuf);

		rv=sdnv_dec(&ilen,ibuf,&olen,obuf);
		if (rv) LTPERR(rv);

		snprintf(ltpstr,LTPSTR_LEN,"obuf");
		ltp_log_buffer(ltpstr,olen,obuf);
	}

	fprintf(cgiOut,"<h2>Output</h2>");
	if (already_encoded) {
		if (!hexoutput) fprintf(cgiOut, "<p>Decoded output</p>\n");
		else fprintf(cgiOut, "<p>Decoded output (hex)</p>\n");
	} else {
		if (!hexoutput) fprintf(cgiOut, "<p>Encoded output</p>\n");
		else fprintf(cgiOut, "<p>Encoded output (hex)</p>\n");
	}
	fprintf(cgiOut,"<p>Output length: %d</p>",olen);

	if (!hexoutput) {
		fwrite(obuf,1,olen,cgiOut);
	} else {

		char *foo=(char*) malloc(3*olen+4*olen/72+1);
		if (!foo)  {
			fprintf(cgiOut,"<p>Memory error (%d)</p>",__LINE__);
    		fprintf(cgiOut, "</BODY></HTML>\n");
			return(rv);
		}
		int bar;
		char *foop=foo;
		for (int i=0;i!=olen;i++) {
			if (!(i%36)) { *foop++='\n'; *foop++='\t'; }
			if (!(i%4)) *foop++=' ';
			bar=obuf[i]>>4;
			*foop++=HEXDIGIT(bar);
			bar=obuf[i]&0x0f;
			*foop++=HEXDIGIT(bar);
		}
		*foop='\0';
		fprintf(cgiOut,"<pre>");
		fprintf(cgiOut,foo);
		fprintf(cgiOut,"</pre>");

	}

	ltp_log_timestamp("Finished test_sdnv");
	ltp_log_flush();

	if (view_log) {
		// just dump it out
		int logsize;
		fprintf(cgiOut, "<h2>Log file</h2>\n");

		// takes a mo?
		sleep(1);

		FILE *lfp=fopen(LOGFILE,"r");
		if (!lfp) {
			fprintf(cgiOut, "Can't alloc memory");
    		fprintf(cgiOut, "</BODY></HTML>\n");
			return(rv);
		}

		struct stat logstat;
		rv=fstat(fileno(lfp),&logstat);
		if (rv) {
			fprintf(cgiOut, "Can't stat file");
    		fprintf(cgiOut, "</BODY></HTML>\n");
			fclose(lfp);
			return(rv);
		}

		logsize=logstat.st_size;

		unsigned char *lbuf=(unsigned char*)malloc(logsize+1);
		if (!lbuf) {
			fprintf(cgiOut, "Can't alloc memory");
    		fprintf(cgiOut, "</BODY></HTML>\n");
			fclose(lfp);
			return(rv);
		}
		
		rv=fread(lbuf,1,logsize,lfp);
		if (rv!=logsize) {
			fprintf(cgiOut, "Can't read file");
    		fprintf(cgiOut, "</BODY></HTML>\n");
			fclose(lfp);
			free(lbuf);
			return(rv);
		}
		fclose(lfp);

		fprintf(cgiOut, "<p>Logfile is %d bytes</p>",logsize);
		fprintf(cgiOut, "<pre>\n");
		fwrite(lbuf,1,logsize,cgiOut);
		fprintf(cgiOut, "</pre>\n");
		free(lbuf);

		fprintf(cgiOut, "</pre>\n");
	}
    /* Finish up the page */
    fprintf(cgiOut, "</BODY></HTML>\n");

	return 0;
}

static int readcgifile(unsigned char **buf, int *blen)
{
	int rv=1;
	cgiFilePtr file;
	char name[1024];
	char contentType[1024];
	char* buffer;
	int size;
	int got;
	if (cgiFormFileName("filename", name, sizeof(name)) != cgiFormSuccess) {
		// printf("<p>No file was uploaded.<p>\n");
		return(rv);
	} 
	// fprintf(cgiOut, "The filename submitted was: ");
	// cgiHtmlEscape(name);
	// fprintf(cgiOut, "<p>\n");
	cgiFormFileSize("filename", &size);
	if (size>MAXFILE) {
		return(rv);
	}
	// fprintf(cgiOut, "The file size was: %d bytes<p>\n", size);
	cgiFormFileContentType("filename", contentType, sizeof(contentType));
	// fprintf(cgiOut, "The alleged content type of the file was: ");
	// cgiHtmlEscape(contentType);
	// fprintf(cgiOut, "<p>\n");
	// fprintf(cgiOut, "Of course, this is only the claim the browser made when uploading the file. Much like the filename, it cannot be trusted.<p>\n");
	// fprintf(cgiOut, "The file's contents are shown here:<p>\n");
	buffer=(char*)malloc(size);
	if (!buffer) return(rv);
	if (cgiFormFileOpen("filename", &file) != cgiFormSuccess) {
		// fprintf(cgiOut, "Could not open the file.<p>\n");
		free(buffer);
		return(rv);
	}
	// fprintf(cgiOut, "<pre>\n");
	char *bufp=buffer;
	int remaining=size;
	while (cgiFormFileRead(file, bufp, remaining, &got) == cgiFormSuccess) {
		bufp+=got;
		remaining-=got;
	}
	if (remaining) {
		cgiFormFileClose(file);
		free(buffer);
		return(rv);
	}
	cgiFormFileClose(file);
	*buf=(unsigned char*)buffer;
	*blen=size;
	return(0);
}

