/*!
 * @file ltp.cc
 * @brief This is the API for our "transport" protocol
 * @version $Revision: 1.167 $ $Author: stephen $
 * @version Copyright (C) 2006 Trinity College Dublin
 *
 * This is my LTP implementation. The interface is like sockets
 * but all calls are non-blocking (the layer below will run
 * a blocking poll/recv in a thread or else will have an
 * interrupt handler for packet arrival), unless overridden
 * via a setsockopt() call.
 *
 * The model for using this is the following. Our application
 * has data for transmission (say a sensor reading and some
 * diagnostics). The application packages up its data according
 * to whatever protocol it talks. The result is a bunch of 
 * messages and destinations (each message has a destination;
 * not ruling out multicast, the destination could be an MC
 * group in future).
 *
 * The application simply "sends" the data. All queueing and
 * other stuff happens below this API. Similarly for arriving 
 * data. 
 *
 * Data for forwarding is, however not handled here. So the 
 * application is really dealing directly with the messaging
 * layer API, which handles routing, and not LTP, but we ignore 
 * that here.)
 *
 * Once data has been handed to LTP then it will be delivered
 * or else an error message will be returned (as if it were 
 * a network packet).
 *
 * This means that the LTP layer has to hold onto data (and
 * state!) even through reboot cycles. Hence we need to be
 * able to "flatten" our state into the filesystem on
 * demand (and obviously unflatten out as well).
 *
 * We assume that this API is only called when the lower
 * layer comms h/w is powered.
 *
 * This LTP layer handles fragmentation of blocks into 
 * LTP segments, i.e. the API works with blocks, not segments
 * and the codec en/decodes segments not entire blocks.
 *
 * We will take a garbage collecting approach to free up 
 * blocks, segments and other things which are no longer 
 * needed.
 *
 * Timers use "punctuated" time (see below for definition).
 * So each timer stores the number of seconds (or iterations)
 * remaining, and is only decremented when the relevant
 * connection is established (or relevant re-Tx happens).
 *
 * Files:
 * 	include/ltp.h - defines the external API 
 * 	include/ltp_int.h - defines types and local statics
 * 		use within the implementation here
 *  src/ltp.cc - this file
 *  include/ltp_l2.h - the external API of the lower layer
 *  include/ltp_codec.h - hides nitty-gritty details of packet formats
 *
 * There is now an LTP layer loopback mode for blocks - a block
 * destined for this address (ltplb) via ltp_sendto() will 
 * be sent to or queued for the 1st subsequent call to 
 * ltp_recvfrom() (where there's no real data available). 
 * We do this using a special, separate queue of blocks that's 
 * part of the state. That mode should result in loopback working
 * even across reboots.
 *
 * @todo collect stats
 */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Trinity College Dublin (TCD) Licklider
 * transmission protocol (LTP) implementation.
 *
 * The Initial Developer of the Original Code is Trinity College Dublin.
 * Portions created by the Initial Developer are Copyright (C) 2006. All
 * Rights Reserved.
 *
 * Contributor(s): Stephen Farrell <stephen.farrell@cs.tcd.ie>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <unistd.h> // for sleep()
#include <ctype.h> // for isprint()
#include <netdb.h> // for hosent etc.
#include <sys/time.h>  // for gettimeofday

#include "ltp_base.h"
#include "ltp.h"
#include "ltp_l2.h"
#include "ltp_codec.h"
#include "ltp_ext.h"
#include "ltp_int.h"
#include "ltp_file.h"
#include "ltp_cues.h"
#include "ltp_crypt.h"
#include "ltp_t.h"
#include "ltp_names.h"

// LTP-T compile time options - for now mutually exclusive
/// set this if you want to only forward a block after its been fully received by the router
#undef LTPT_ONLYSENDWHENFULL
/// set this if you want to trickle out segments of to-be-forwarded blocks as and when you can
#define LTPT_TRICKLE_FORWARD

/// We want to use SHA-256 for our hash algorithm for replay detection
#define SHA256HASH
#ifdef SHA256HASH
#include "sha256.h"
#endif

#ifdef THREADS
#include <pthread.h>
#define PING_THREADS
#else
#undef PING_THREADS
#endif

#ifdef GOTHEREIN
#define GOTHERE { fprintf(stderr,"At %d in %s\n",__LINE__,__FUNCTION__); }
#else 
#define GOTHERE
#endif

/// A "special" LTP address - the LTP layer loopback - 0x0100007f is 127.0.0.1
ltpaddr ltplb={AF_LTP,UDP_DEF_PORT,0x0100007f,"\0\0\0\0\0\0\0"}; // 7 '\0' characters in quotes = 8 zero bytes!
/// Another "special" address - "any" (equivalent to broadcast IP - 255.255.255.255)
ltpaddr ltpany={AF_LTP,UDP_DEF_PORT,0xffffffff,"\0\0\0\0\0\0\0"};
/// Another "special" address - this time an "empty" one
ltpaddr ltpnull={AF_LTP,UDP_DEF_PORT,0,"\0\0\0\0\0\0\0"};
/// Finally my address, as told to me by layer 2
ltpaddr my_addr;
bool my_addr_inited=false;

/// A macro to initialise my_addr if neccessary
#define INITMYADDR if (!my_addr_inited) { int xxRv=ltp_cue_whoami(&my_addr); if (xxRv) LTPERR(-1) else my_addr_inited=true; }

/// whether to shutdown l2 on close of l3
static bool l2shutdownonclose=false;

/// the layer 2 MTU we get back from ltp_l2_on()
static unsigned int l2mtu=ltp_mib_var.segmentsize; 

/// whether application has set the l2mtu via a sockopt() call - if so, use that
static bool l2mtu_set=false;

/// the state we want for l2 (tx,rx or both) [both]
static int l2desiredstate=LTP_L2S_ON_BOTH;

/// Whether layer 2 has been initialised as yet
static int l2state=LTP_L2S_RESET;

/// Our global state variable structure
ltp_state ltp_state_var = { 
		0, 
		false, 
		0, 
		0, 
		NULL, 
		{ 1, 0, 0, NULL}, // loopback queue
		{ 2, 0, 0, NULL}, // tx queue
		{ 3, 0, 0, NULL}, // replays
		{ 4, 0, 0, NULL}, // timers
		0 };
/// We need to know if some state's changed (usually it will have I guess)
bool ltp_state_changed=false;

/// To work with ION I need to be able to force the responder address via LTP_SO_FORCE
static ltpaddr forced_addr=ltpany;
static bool forced_addr_set=false;

/// To wotk with ION I need to force some identfiers to be short
static bool ion_mode=false;

// look ahead a bit :-)
static int ltp_ping_init(void);
static int ltp_init_aux(void);
static int springclean(ltp_socket_t* socksess);

/// "ping" the LTP layer - get it to check for in/out bound stuff and 
/// handle as appropriate - called occasionally (a thread or from within
/// LTPINIT or else directly by the application).
#ifdef PING_THREADS
#define LTPPING { int rrzzz=ltp_ping_init(); if (rrzzz) { if (rrzzz>0) rrzzz *=-1; LTPERR(rrzzz); } }
#else
#define LTPPING { int rrzzz=ltp_ping(); if (rrzzz) { if (rrzzz>0) rrzzz *=-1; LTPERR(rrzzz); } }
#endif

/// Common initialisation macro - called at entry to most ltp_* functions
#define LTPINIT { int zzz=ltp_init_aux(); if (zzz) { if (zzz>0) zzz *=-1; LTPERR(zzz); } LTPPING }

/// We hit this whenever we've made a change to the state of memory that might have
/// to be saved to CF.
#define LTP_STATE_CHANGED ltp_state_changed=true;

#ifdef MUDSAT
// SF MUDSAT testing: see what happens if I just set all the segments of this

// SF MUDSAT testing: default ciphersuite stuff
// TODO: fixme
static int using_def_crypto=0;
static int def_ciphersuite=LTP_CS_SIG;
static char *def_keyid="foo";

#endif

/*!
 * @fn ltp_instance ltp_next_instance(void)
 * @return is a new (unused) instance number
 */
static ltp_instance ltp_next_instance(void);

// Some thread stuff from here on down
#ifdef PING_THREADS
/// Here's the pinger thread pointer
static pthread_t	pinger=(pthread_t)0;
static pthread_t	forwarder=(pthread_t)0;
/// A mutex to use in the pinger thread so other code doesn't have to 
/// be re-entrant. Also used to isolate LTP-T forwarder thread.
#define TRYTHIS
#ifdef TRYTHIS
pthread_mutex_t pinger_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#else
pthread_mutex_t pinger_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
/// Take the pinger mutex - scheme is that the pinger thread
/// waits to grab this when it needs to do stuff (as opposed to hanging
/// around in a busy-wait). 
/// The implementation of the external LTP api functions also wait 
/// to grab the mutex on entry (and so could block if the pinger thread's
/// v. busy). That should be ok for now. 
#undef MUTEXDEBUG
#ifdef MUTEXDEBUG
#define GRABMUTEX { \
	fprintf(stderr,"Grabbing pinger (at %s:%d)...",__FILE__,__LINE__); \
	pthread_mutex_lock(&pinger_mutex); \
	fprintf(stderr,"..got pinger\n"); }
/// Give back the pinger mutex
#define DROPMUTEX { \
	fprintf(stderr,"Dropping pinger (at %s:%d)...",__FILE__,__LINE__); \
	pthread_mutex_unlock(&pinger_mutex); \
	fprintf(stderr,"...dropped pinger\n"); }
#else
#define GRABMUTEX pthread_mutex_lock(&pinger_mutex);
/// Give back the l2 pinger mutex
#define DROPMUTEX pthread_mutex_unlock(&pinger_mutex);
/// Check if I'd have gotten the mutex (not sure what code wants this)
#define MUTEXAVAILABLE !pthread_mutex_lock(&pinger_mutex)
#endif
#else
#define GRABMUTEX
#define DROPMUTEX
#define MUTEXAVAILABLE
#endif

/// a macro for that
#define NI (ltp_next_instance())

/// The number of chunks to allocate for each call to FOOREALLOC.
#define LTP_ALLOC_SPARES 	100

/// Macro to declare a few things and get ready to send a new segment
#define NEWSEGMENT \
			int nsegments=ltp_state_var.txq.nsegments; \
			int alloced=ltp_state_var.txq.alloced; \
			ltp_segment *segments=ltp_state_var.txq.segments; \
			FOOREALLOC(ltp_segment,segments,nsegments+1,alloced); \
			ltp_state_var.txq.segments=segments; \
			ltp_state_var.txq.alloced=alloced;   \
			ltp_state_var.txq.nsegments=nsegments+1; \
			segments[nsegments].id=NI; \
			segments[nsegments].state=LTP_PS_WAITING_SEND; \
			segments[nsegments].r_entry=LTP_NOCACHE; \
			segments[nsegments].r_id=0; \
			segments[nsegments].t_entry=LTP_NOTIMER; \
			segments[nsegments].t_id=0; \
			segments[nsegments].finishedwith=false; \

/// This reallocates space of type xXxtype for xXxvar so that there's now 
/// room for xXxmany.
/// There's a tweak required here. When realloc moves things, it leaves
/// behind the ->id field values, but since that's how we find the moved
/// stuff we don't want it to happen (I did get crashes that way!) since
/// if we find one of those its not really in a structure anymore.
/// So we calloc and memcpy instead of really reallocing. To make this a
/// bit less painful we always allocate LTP_ALLOC_SPARES 
/// additional spaces when we have to do anything.
#define FOOREALLOC(xXxtype,xXxvar,xXxwanted,xXxthere) \
	{ \
		if (xXxwanted >= xXxthere ) { \
			size_t fooxXxthere=xXxwanted+LTP_ALLOC_SPARES; \
			xXxtype *xXxfoo=(xXxtype*)calloc(fooxXxthere,sizeof(xXxtype)); \
			if (xXxfoo==NULL) { DROPMUTEX; LTPERR(LTP_MEMORY); } \
			if (xXxvar) { \
				memcpy(xXxfoo,xXxvar,xXxthere*sizeof(xXxtype)); \
				free(xXxvar); \
			} \
			xXxvar=xXxfoo; \
			xXxthere=fooxXxthere; \
		} \
	}

/// This is the same as FOOREALLOC but the error case return is a void
/// and not an int (i.e. VOIDLTPERR(foo) instead of LTPERR(foo)
#define VOIDFOOREALLOC(xXxtype,xXxvar,xXxwanted,xXxthere) \
	{ \
		if (xXxwanted >= xXxthere ) { \
			size_t fooxXxthere=xXxwanted+LTP_ALLOC_SPARES; \
			xXxtype *xXxfoo=(xXxtype*)calloc(fooxXxthere,sizeof(xXxtype)); \
			if (xXxfoo==NULL) { DROPMUTEX; VOIDLTPERR(LTP_MEMORY); } \
			if (xXxvar) { \
				memcpy(xXxfoo,xXxvar,xXxthere*sizeof(xXxtype)); \
				free(xXxvar); \
			} \
			xXxvar=xXxfoo; \
			xXxthere=fooxXxthere; \
		} \
	}

/*!
 * @section sfunctions static functions from here on down.
 */

#ifdef MUDSAT
// SF MUDSAT testing: see what happens if I just set all the segments of this
// all green block to indicate they were not sent, but just once
int ltp_green_re_tx(ltp_segment *seg)
{
	int blockid=seg->parent->id;
	ltp_segment *segments=ltp_state_var.txq.segments; 
	// go through all segments for that block and set their
	// state to LTP_PS_WAITING_SEND and see what happens
	for (int i=0;i!=ltp_state_var.txq.nsegments;i++) {
		if (segments[i].parent->id==blockid) {
			// mark for re-tx
			segments[i].state=LTP_PS_WAITING_SEND;
		}
	}
	return(LTP_OK);
}
#endif

/*!
 * @brief estimate the length of stuff extension encodings take up
 * @param theblock is the block concerned
 * @param thesock is the socket
 * @return the estimeate
 */
int extensions_estimate(ltp_block *theblock, ltp_socket_t *thesock)
{
	int accumulated=0;
	// cookie
	if (thesock->opts.cookie) {
		accumulated+=2+thesock->opts.cookie; // max cookie is 20 so no SDNV complications
	}
	// inboundcookie
	if (theblock->inboundcookie.len) {
		accumulated+=2+theblock->inboundcookie.len; // max cookie is 20 so no SDNV complications
	}
	// authent
	if (thesock->opts.auth.ciphersuite==LTP_CS_MAC) {
		// header = tag(1), len (1) + optional-keyid
		// trailer = tag(1), len (1), MAC (20)
		accumulated+=24+thesock->opts.auth.keyidlen;
	}
	if (thesock->opts.auth.ciphersuite==LTP_CS_SIG) {
		// header = tag(1), len (1) + optional-keyid
		// trailer = tag(1), len (2), Sig (128)
		accumulated+=133+thesock->opts.auth.keyidlen;
	}
	if (thesock->opts.auth.ciphersuite==LTP_CS_NULL) {
		// header = tag(1), len (1) 
		// trailer = tag(1), len (1), MAC (20)
		accumulated+=24;
	}
	if (thesock->opts.t_cfg.on || theblock->t_vals.flags) {
		accumulated+=ltpt_extensions_estimate(theblock,thesock);
	}
	return(accumulated);
}

/*!
 * @brief figure out the latest MTU info
 * @param block is the block in question
 * @param sock is the socket in question
 * @return mtu is returned (or zero for error)
 */
static int latestmtu(ltp_block *block, ltp_socket_t *sock)
{
	// Note: block can be NULL and that ought be ok, just take from sock then

	// TODO: take mtu from socket option if provided
	/// the layer 2 MTU less overhead 
	static unsigned int l2mtuover;
	// recalc - could have been changed via sockopt()
	if (sock && sock->opts.mtu_set) {
		l2mtuover=sock->opts.mtu-LTPC_DS_OVERHEAD; 
		snprintf(ltpstr,LTPSTR_LEN,"Overridden MTU is %d (from %d)\n",l2mtuover,sock->opts.mtu);
		ltp_log(LTP_LOG_MAX,ltpstr);
	} else {
		l2mtuover=l2mtu-LTPC_DS_OVERHEAD; 
	}
	// adjust that based on the extensions we're using now
	l2mtuover-=extensions_estimate(block,sock);
#ifdef LTP_VERBOSE
	static bool printed=false;
	if (!printed) {
		snprintf(ltpstr,LTPSTR_LEN,"Reduced MTU is %d (from %d)\n",l2mtuover,l2mtu);
		ltp_log(LTP_LOG_MAX,ltpstr);
		printed=true;
	}
#endif
	if (l2mtuover<=0) { return(0); }
	return(l2mtuover);
}

/*
 * @brief my non-kernel limited nanosleep
 */
int ltp_microsleep(int microseconds)
{
	// has to be a busy wait
	static bool calibrated=false;
	static int multiplier=0;
	static int dummy=0;
	struct timeval tv;
	
	/* Obtain the time of day, and convert it to a tm struct. */
	if (!calibrated) {
#define ITERTODO 10
		int avg_calib;
		int totaldiff=0;
		gettimeofday (&tv, NULL);
		int seconds=tv.tv_sec;
		int micros=tv.tv_usec;
		for (int i=0;i!=ITERTODO;i++) {
			gettimeofday (&tv, NULL);
			int inner_seconds=tv.tv_sec;
			int inner_micros=tv.tv_usec;
			dummy=0;
			while ((inner_micros-tv.tv_usec)==0) {
				gettimeofday (&tv, NULL); dummy++;
			}
			multiplier=dummy;
			int diff=tv.tv_usec-inner_micros;
			if (tv.tv_sec!=inner_seconds) { diff+=(1000000*(tv.tv_sec-inner_seconds)); }
			int lc_mult=multiplier;
			multiplier=diff/lc_mult+1;
			snprintf(ltpstr,LTPSTR_LEN,"Calibrating ltp_microsleep iterations=%d,diff=%d,multiplier for 1us = %d\n",
					lc_mult,diff,multiplier);
			ltp_log(LTP_LOG_MAX,ltpstr);
			avg_calib+=multiplier;
		}
		multiplier=avg_calib/ITERTODO;
		calibrated=true;
		int bigdiff=tv.tv_usec-micros;
		if (tv.tv_sec!=seconds) { bigdiff+=(1000000*(tv.tv_sec-seconds)); }
		if (bigdiff >= microseconds) {
			snprintf(ltpstr,LTPSTR_LEN,"Calibrated ltp_microsleep multiplier for 1us = %d\n",multiplier);
			ltp_log(LTP_LOG_MAX,ltpstr);
			return(0);
		}
		snprintf(ltpstr,LTPSTR_LEN,"Calibrated ltp_microsleep multiplier for 1us = %d (waiting until %d us)\n",
				multiplier,(microseconds-bigdiff));
		ltp_log(LTP_LOG_MAX,ltpstr);
		// otherwise wait that much
		dummy=0;
		for (int i=0;i!=multiplier*(microseconds-bigdiff);i++) {
			gettimeofday (&tv, NULL); dummy++;
		}
		return(0);
	}
#ifdef LTP_VERBOSE
	gettimeofday (&tv, NULL);
	int lv_seconds=tv.tv_sec;
	int lv_micros=tv.tv_usec;
#endif
	dummy=0;
	for (int i=0;i!=multiplier*microseconds;i++) {
		gettimeofday (&tv, NULL); dummy++;
	}
#ifdef LTP_VERBOSE
	int lv_diff=tv.tv_usec-lv_micros;
	if (tv.tv_sec!=lv_seconds) { lv_diff+=(1000000*(tv.tv_sec-lv_seconds)); }
	snprintf(ltpstr,LTPSTR_LEN,"ltp_microsleep waited %d us (iterations=%d)\n",
			lv_diff,dummy);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	return(0);
}

/*
 * @brief wait a bit so as not to max out the CPU and give other threads a go
 */
static void ltp_waitabit(void)
{
#undef WAITINLOOP
#ifdef WAITINLOOP
	// now we wait for a short while
	struct timeval ltv;
	ltv.tv_sec=0;
	ltv.tv_usec=100000; // 1/10th sec
	select(0,NULL,NULL,NULL,&ltv);
#else
	struct timespec	ts,ts1;
	memset(&ts,0,sizeof(ts));
	memset(&ts1,0,sizeof(ts));
	ts.tv_nsec=1000*1000*20;  // 20ms
#undef NANOPRINT
#ifdef NANOPRINT
	char 		befstr[MAXPATH],aftstr[MAXPATH];
	snprintf(befstr,MAXPATH,ltp_now_str());
#endif
	int rv=nanosleep(&ts,&ts1);
#ifdef NANOPRINT
	snprintf(aftstr,MAXPATH,ltp_now_str());
	snprintf(ltpstr,LTPSTR_LEN,"waitabit/nanosleep at %d from %s to %s\n",__LINE__,befstr,aftstr);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
#endif
#ifdef SENDT
		// strobe watchdog
	int rv=ltp_watchdog_strobe();
	if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"Error from ltp_watchdog_strobe: %d\n",rv);
		ltp_log(LTP_LOG_MAX,ltpstr);
		// continue on regardless! (or try to...)
	}
#endif
}

/*!
 * @brief initialise and start a cookie timer 
 * @param block is (a ptr to) the block
 * @return zero or error
 */
static int init_cookie_timer(ltp_socket_t *sock, ltp_block *block)
{
	ltp_segment	*seg=NULL;
	int pval=(int)floor(sock->opts.cookie_grace); 
	int cval=1; 
	int alloced=ltp_state_var.timers.alloced;
	int nentries=ltp_state_var.timers.nentries;
	nentries++;
	FOOREALLOC(ltp_timer_detail,ltp_state_var.timers.details,nentries,alloced); 
	ltp_timer_detail *det=&ltp_state_var.timers.details[nentries-1];
	det->id=NI;
	det->type=LTP_TIMER_COOKIE;
	det->finishedwith=false;
	det->active=true;
	det->lastdecremented=(ltp_time_t)ltp_now();
	det->remaining=pval;
	det->counter=cval;
	det->block=block;
	if (det->block) det->blockid=block->id;
	else det->blockid=0;
	det->segment=seg;
	if (det->segment) det->segmentid=seg->id;
	else det->segmentid=0;
	det->socket=sock;
	if (det->socket) det->socketid=sock->id;
	else det->socketid=0;
	ltp_state_var.timers.alloced=alloced;
	ltp_state_var.timers.nentries=nentries;
	LTP_STATE_CHANGED;
	return(LTP_OK);
}

/*!
 * @brief add a new, or the first extension
 * @param seg to the guy to add to 
 * @param added is the new guy
 * @return is the new total number of segments or a negative number for errors
 */
int add_ext(ltpc_seg *seg)
{
	if (!seg) LTPERR(-LTP_BAD_VALUE);
	if (!seg->nexts) {
		seg->exts=(ltpc_ext*)ltpc_alloc(sizeof(ltpc_ext));
		if (!seg->exts) LTPERR(-LTP_MEMORY);
		seg->nexts=1;
	} else {
		seg->nexts++;
		seg->exts=(ltpc_ext*)ltpc_realloc((unsigned char*)seg->exts,sizeof(ltpc_ext)*seg->nexts);
		if (!seg->exts) LTPERR(-LTP_MEMORY);
	}
	memset(&seg->exts[seg->nexts-1],0,sizeof(ltpc_ext));
	return(seg->nexts);
}

/*! 
 * @brief check around to see what extensions to add before encoding
 * @param seg is the segment
 * @param added_len is (a ptr to) the additional encoded length we'll need
 * @return zero or error
 */
static int ltp_add_exts_before_encode(ltp_segment *segment, int *added_len)
{
	int rv=0;
	int accum_new_len=0;
	if (!segment || !segment->parent || !segment->parent->parent || !added_len) LTPERR(LTP_BAD_PARAM);
	ltp_block*block=segment->parent;

	// ok no idea why I'm adding these but it works let's see...
	rv=ltp_fixparent(&ltp_state_var,(void**)&segment->parent,segment->parentid,LTP_PARENT_BLOCK);
	if (rv) LTPERR(rv);
	rv=ltp_fixparent(&ltp_state_var,(void**)&block->parent,block->parentid,LTP_PARENT_SOCKET);
	if (rv) LTPERR(rv);

	ltp_socket_t *sock=block->parent;
	ltpc_seg	*seg=&segment->seg;

	*added_len=0;

	// if setup for cookies then generate or add mine as required.
	// implicit in this is that we get a new cookie for each block sent via this socket
	// I may prefer some other behavour later, e.g. one cookie per destination for as long
	// as I can remember 'em, or one per socket, or whatever
	if (block->outboundcookie.len==0 && sock->opts.cookie) {
		// first time - need to generate a cookie
		ltp_octets *cook=&block->outboundcookie;
		cook->len=sock->opts.cookie;
		cook->val=(unsigned char*)calloc(cook->len,sizeof(char));
		if (!cook->val) LTPERR(LTP_MEMORY);
		rv=ltp_rand_octets(cook->len,cook->val);
		if (rv) LTPERR(rv);
		// start timer - this'll be sent momentarily
		if (sock->opts.cookie_grace) {
			rv=init_cookie_timer(sock,block);
			if (rv) LTPERR(rv);
		} else { // no need for timer if grace is zero
			block->cookie_grace_expired=true;
		}
	}
 	if (block->outboundcookie.len) {
		// add a cookie extension and accumulate some additional length
		rv=add_ext(seg);
		if (rv<=0) LTPERR(rv-1); // the -1 makes sure to not return zero
		ltpc_ext *theext=&seg->exts[seg->nexts-1];
		memset(theext,0,sizeof(ltpc_ext));
		ltpe_cookie *ck=(ltpe_cookie*)ltpc_alloc(sizeof(ltpe_cookie));
		if (!ck) LTPERR(LTP_MEMORY);
		ck->cval.len=block->outboundcookie.len;
		ck->cval.val=ltpc_alloc(ck->cval.len);
		if (!ck->cval.val) LTPERR(LTP_MEMORY);
		memcpy(ck->cval.val,block->outboundcookie.val,ck->cval.len);
		// make up the generic cookie
		theext->header=true;
		theext->type=LTPE_COOKIE;
		theext->len=sizeof(*ck); // doesn't matter, not used
		theext->val=(void*)ck;
		theext->enclen=0;
		theext->encval=NULL;
		*added_len+=ck->cval.len+10; // overestimate but ok
	}
	// inbound cookie?
 	if (block->inboundcookie.len) {
		// add a cookie extension and accumulate some additional length
		rv=add_ext(seg);
		if (rv<=0) LTPERR(rv-1); // the -1 makes sure to not return zero
		ltpc_ext *theext=&seg->exts[seg->nexts-1];
		memset(theext,0,sizeof(ltpc_ext));
		ltpe_cookie *ck=(ltpe_cookie*)ltpc_alloc(sizeof(ltpe_cookie));
		if (!ck) LTPERR(LTP_MEMORY);
		ck->cval.len=block->inboundcookie.len;
		ck->cval.val=ltpc_alloc(ck->cval.len);
		if (!ck->cval.val) LTPERR(LTP_MEMORY);
		memcpy(ck->cval.val,block->inboundcookie.val,ck->cval.len);
		// make up the generic cookie
		theext->header=true;
		theext->type=LTPE_COOKIE;
		theext->len=sizeof(*ck); // doesn't matter, not used
		theext->val=(void*)ck;
		theext->enclen=0;
		theext->encval=NULL;
		*added_len+=ck->cval.len+10; // overestimate but ok
	}

#ifdef MUDSAT
	// SF: MUDSAT - just keep any existing authenticators
	// and don't recalculate 'em (nothing ought change I hope)
	// the code here really fixes a bug 
#endif
	int already_authed=0;
	for (int oe=0;oe!=seg->nexts;oe++) {
		if (seg->exts[oe].type==LTPE_AUTH) {
			already_authed=1;
			*added_len+=100; // TODO: fix this later, 100 is loads though
		}
	}

	// now we see what, if anything to do about outbound authentication
	if (!already_authed && sock->opts.auth.ciphersuite!=LTP_CS_NONE) {
		// something!
		int enclen,authlen;
		unsigned char *encbuf,*authbuf;


		// put in anything that's part of plaintext
		rv=add_ext(seg);
		if (rv<=0) LTPERR(rv-1); // the -1 makes sure to not return zero
		ltpc_ext *theext=&seg->exts[seg->nexts-1];
		ltpe_auth_header *ahd=(ltpe_auth_header*)calloc(1,sizeof(ltpe_auth_header));
		if (!ahd) LTPERR(LTP_MEMORY);
		int kblen=sock->opts.auth.keyidlen;
		unsigned char *kb=(unsigned char*)calloc(1,kblen+1); // space for a gratuitous null char too
		if (!kb) LTPERR(LTP_MEMORY);
		memcpy(kb,sock->opts.auth.keyid,kblen);
		kb[kblen]='\0'; // so a naive printf etc is ok
		switch (sock->opts.auth.ciphersuite) {
			case LTP_CS_MAC:
				ahd->ciphersuite=LTPE_CIPHERSUITE_ORIGINAUTH;
				ahd->keyid.len=kblen;
				ahd->keyid.val=kb;
				break;
			case  LTP_CS_SIG:
				ahd->ciphersuite=LTPE_CIPHERSUITE_SIGNATURE;
				ahd->keyid.len=kblen;
				ahd->keyid.val=kb;
				break;
			case LTP_CS_NULL:
				ahd->ciphersuite=LTPE_CIPHERSUITE_NULL;
				ahd->keyid.len=0;
				ahd->keyid.val=NULL;
				free(kb); // let's not leak that!
				break;
			default:
				LTPERR(LTP_BAD_VALUE);
		}
		theext->header=true;
		theext->type=LTPE_AUTH;
		theext->len=sizeof(*ahd); // doesn't matter, not used
		theext->val=(void*)ahd;
		theext->enclen=0;
		theext->encval=NULL;
		// encode up the segment into encbuf
		ltp_octets ptbuf;
		switch (seg->type) {
			case LTPC_SEG_DS: ptbuf.len=seg->val.ds.bytes.len+LTPC_DS_OVERHEAD; break;
			case LTPC_SEG_RS: ptbuf.len=LTPC_RS_OVERHEAD+seg->val.rs.cc*(2*10); break;
			default:
				ptbuf.len=LTPC_RA_OVERHEAD;
		}
		// add the trailer with no value (so right plaintext generated)
		rv=add_ext(seg);
		theext=&seg->exts[seg->nexts-1];
		theext->header=false;
		theext->type=LTPE_AUTH;
		theext->len=0;
		theext->val=NULL;
		theext->enclen=0;
		theext->encval=NULL;
		ptbuf.len+=kblen+10;
#ifdef MUDSAT
		// SF: MUDSAT testing - this is not an addition but a bugfix
		// if there are existing extensions then we need to allow
		// space for their encoding in the ptbuf
#endif
		int otherextslen=0;
		for (int oe=0;oe!=seg->nexts;oe++) {
			int a=seg->exts[oe].len;
			int b=seg->exts[oe].enclen;
			otherextslen+=max(a,b)*2;
		}
		ptbuf.len+=otherextslen;
		ptbuf.val=ltpc_alloc(ptbuf.len);
		if (!ptbuf.val) LTPERR(LTP_MEMORY);
		rv=ltpc_encode(seg, (int*)&ptbuf.len, ptbuf.val, true);
		if (rv) LTPERR(rv);
		// now do crypto
		authbuf=ltpc_alloc(LTP_AUTHBUFSIZE);
		if (!authbuf) LTPERR(LTP_MEMORY);
		authlen=LTP_AUTHBUFSIZE;
		rv=ltp_crypt_do(&sock->opts.auth,&seg->peer,ptbuf.len,ptbuf.val,&authlen,&authbuf);
		if (rv) LTPERR(rv);
		ltpc_free(ptbuf.val);
		// now put authval back into segment
		ltpe_auth_trailer *tlr=(ltpe_auth_trailer*)calloc(1,sizeof(ltpe_auth_trailer));
		if (!tlr) LTPERR(LTP_MEMORY);
		tlr->authval.len=authlen;
		tlr->authval.val=authbuf;
		theext->len=sizeof(*tlr); // doesn't matter, not used
		theext->val=(void*)tlr;

		// accumulate more space
		*added_len+=kblen+authlen+1000; // overestimate but ok 
	}
	if (sock->ltptrelay) {
		rv=ltpt_add_departure_exts(seg,block,sock);
		if (rv) LTPERR(rv);
	}
	return LTP_OK;
}

/*!
 * @brief printable version of a key id
 * @param kid are the key id octets
 * @param len is the length
 * @return is a ptr to a static array that's a printable version of the input
 */
static char *octets2str(unsigned char *kid, int len)
{
	// TODO make a real octets2str later for binary crap
	static char str[LTP_CS_MAX_KID_LEN];
	int thislen=min(LTP_CS_MAX_KID_LEN,len);
	memcpy(str,kid,thislen);
	str[thislen]='\0';
	// something anyway
	for (int i=0;i!=thislen;i++) {
		if (!isprint(str[i])) str[i]='%';
	}
	return (str);
}

/*!
 * @brief do the LTP auth processing for this segment
 * @param segstr is the segment structure
 * @param theblock is the parent block (it can be NULL)
 * @param thesock is the parent socket (can it be NULL - not sure?)
 * @param badcookie is an output - true if this segments cookies are awry in some way  
 * @param inlen is the length of the encoded segment
 */
static int auth_processing(ltpc_seg *segstr, ltp_block *theblock, ltp_socket_t *thesock, bool *badauth, int inlen) 
{
	// now what about LTP Auth extensions that I get?
	int rv;
	unsigned char *auth_input=NULL;
	int auth_input_len=0;
	unsigned char *auth_val=NULL;
	int auth_val_len=0;
	bool auth_res=false;
	*badauth=false; // optimist!
	for (int i=0;!auth_res && i!=segstr->nexts;i++) {
		if (segstr->exts[i].type==LTPE_AUTH && segstr->exts[i].header) {
			// read header
			ltp_auth_so ai_so;
			ltpe_auth_header *hdr=(ltpe_auth_header*)segstr->exts[i].val;
			if (!hdr) LTPERR(LTP_BAD_VALUE);
			ai_so.ciphersuite=hdr->ciphersuite;
			memcpy(ai_so.keyid,hdr->keyid.val,min(hdr->keyid.len,LTP_CS_MAX_KID_LEN));
			ai_so.keyidlen=hdr->keyid.len;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Extension #%d is an Auth Header (c=%d,k=%s)\n",
				i,ai_so.ciphersuite,octets2str(ai_so.keyid,ai_so.keyidlen));
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			// check stuff
			if (!auth_input) {
				// generate an encoding with no AuthVals at all
				auth_input_len=inlen;
				auth_input=ltpc_alloc(inlen);
				if (!auth_input) LTPERR(LTP_MEMORY);
				ltpc_seg segcp=*segstr; // (nearly) shallow copy
				segcp.exts=(ltpc_ext*)ltpc_alloc(sizeof(ltpc_ext)*segcp.nexts);
				if (!segcp.exts) LTPERR(LTP_MEMORY);
				memcpy(segcp.exts,segstr->exts,sizeof(ltpc_ext)*segcp.nexts);
				// now we can mess with segcp.exts[i] values 
				for (int j=0;j!=segcp.nexts;j++) { // zap authvals
					if (segcp.exts[j].type==LTPE_AUTH && !segcp.exts[j].header) {
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Extension #%d is an Auth Trailer\n",j);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						segcp.exts[j].val=NULL; segcp.exts[j].len=0;
						segcp.exts[j].encval=NULL; segcp.exts[j].enclen=0;
					}
				}
				rv=ltpc_encode(&segcp, &auth_input_len, auth_input, true);
				if (rv) LTPERR(rv);
				// tidy up time
				ltpc_free((unsigned char*)segcp.exts);

			}
			// find matching trailer - don't bother just try all?
			for (int j=i;!auth_res && j!=segstr->nexts;j++) {
				if (segstr->exts[j].type==LTPE_AUTH && !segstr->exts[j].header) {
					auth_val=(unsigned char*)segstr->exts[j].encval;
					auth_val_len=segstr->exts[j].enclen;
					rv=ltp_crypt_check(&ai_so,&segstr->peer,auth_input_len,auth_input,auth_val_len,auth_val,&auth_res);
					if (rv) LTPERR(rv); 
					if (auth_res) { // yippee!!
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Extension #%d is a GOOD Auth Trailer - Yippee!\n",j);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					} else { // bummer
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Extension #%d is a BAD Auth Trailer - Yuk!\n",j);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						*badauth=true;
					}
				}
			}
		}
	}
	if (auth_input) ltpc_free(auth_input);
	return(LTP_OK);
}

/*!
 * @brief do the cookie processing for this segment
 * @param segstr is the segment structure
 * @param theblock is the parent block (it can be NULL)
 * @param thesock is the parent socket (can it be NULL - not sure?)
 * @param badcookie is an output - true if this segments cookies are awry in some way  
 */
static int cookie_processing(ltpc_seg *segstr, ltp_block *theblock, ltp_socket_t *thesock, bool *badcookie) 
{
	if (!segstr || !badcookie) LTPERR(LTP_BAD_INPUT);
	// do I expect cookies?
	bool cookies_wanted=(thesock && (thesock->opts.cookie!=0));
	bool grace_expired=(theblock && theblock->cookie_grace_expired);
	bool cookie_present=false;
	int	mycookie_index=-1;
	int	hiscookie_index=-1;
	int unknowncookie_index=-1;
	for (int i=0;i!=segstr->nexts;i++) {
		if (segstr->exts[i].type==LTPE_COOKIE) {
			cookie_present=true;
			// see if its mine 
			if (theblock && theblock->outboundcookie.len && 
				segstr->exts[i].encval && segstr->exts[i].enclen>=thesock->opts.cookie) {
				if (!memcmp(segstr->exts[i].encval,theblock->outboundcookie.val,thesock->opts.cookie)) {
					mycookie_index=i;
					// handle cookie lengthening - even the peer is allowed make it longer
					if (segstr->exts[i].encval && segstr->exts[i].enclen>thesock->opts.cookie) {
						unsigned char *ccv=ltpc_alloc(segstr->exts[i].enclen);
						if (!ccv) { LTPERR(LTP_MEMORY); }
						memcpy(ccv,segstr->exts[i].encval,segstr->exts[i].enclen);
						free(theblock->outboundcookie.val);
						theblock->outboundcookie.val=ccv;
						theblock->outboundcookie.len=segstr->exts[i].enclen;
					}
				}
			}
			if (mycookie_index!=i && theblock && theblock->inboundcookie.len &&
				segstr->exts[i].encval && segstr->exts[i].enclen>=theblock->inboundcookie.len) {
				// check against inbound
				if (!memcmp(segstr->exts[i].encval,theblock->inboundcookie.val,theblock->inboundcookie.len)) {
					hiscookie_index=i;
					// handle cookie lengthening - this time the cookie setter is making it longer
					if (segstr->exts[i].encval && segstr->exts[i].enclen>theblock->inboundcookie.len) {
						unsigned char *ccv=ltpc_alloc(segstr->exts[i].enclen);
						if (!ccv) { LTPERR(LTP_MEMORY); }
						memcpy(ccv,segstr->exts[i].encval,segstr->exts[i].enclen);
						free(theblock->inboundcookie.val);
						theblock->inboundcookie.val=ccv;
						theblock->inboundcookie.len=segstr->exts[i].enclen;
					}
				}
			} 
			if (mycookie_index!=i && hiscookie_index!=i) unknowncookie_index=i;
		}
	}
#ifdef LTP_VERBOSE
	if (cookie_present) { // don't bother the reader if there's nothing there
		snprintf(ltpstr,LTPSTR_LEN,"Cookie status: wanted=%s (grace expired=%s),mine(ind)=%d, his(ind)=%d,unk(ind)=%d\n",
				(cookies_wanted?"true":"false"), (grace_expired?"true":"false"),
				mycookie_index,hiscookie_index,unknowncookie_index);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
#endif
	// so now enforcement
	// outbound analysis
	if ( cookies_wanted && grace_expired && cookie_present) {
		// ditch the seg
		*badcookie=false;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Got expected cookie as extension #%d\n",mycookie_index);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else if ( cookies_wanted && grace_expired && !cookie_present) {
		// ditch the seg
		*badcookie=true;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Ditched seg 'cause of cookies at %s:%d\n",__FILE__,__LINE__);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else if (cookies_wanted && grace_expired && mycookie_index==-1) {
		// ditch the seg
		*badcookie=true;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Ditched seg 'cause of cookies at %s:%d\n",__FILE__,__LINE__);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} 
	// inbound analysis
	if (theblock && theblock->inboundcookie.len && !cookie_present) {
		// oops no cookies at all
		// ditch the seg
		*badcookie=true;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Ditched seg 'cause of cookies at %s:%d\n",__FILE__,__LINE__);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else if (theblock && theblock->inboundcookie.len && cookie_present &&
		hiscookie_index==-1) {
		// oops some cookie but not his
		// ditch the seg
		*badcookie=true;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Ditched seg 'cause of cookies at %s:%d\n",__FILE__,__LINE__);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} 
	// maybe add a new inbound cookie
	if (!*badcookie && theblock && !theblock->inboundcookie.len && cookie_present && 
		unknowncookie_index != -1) {
		unsigned char *ccv=ltpc_alloc(segstr->exts[unknowncookie_index].enclen);
		if (!ccv) { LTPERR(LTP_MEMORY); }
			memcpy(ccv,segstr->exts[unknowncookie_index].encval,segstr->exts[unknowncookie_index].enclen);
			// ltpc_free(theblock->inboundcookie.val);
			theblock->inboundcookie.val=ccv;
			theblock->inboundcookie.len=segstr->exts[unknowncookie_index].enclen;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Added inbound cookie at %s:%d\n",__FILE__,__LINE__);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} 
	return(LTP_OK);
}

/*!
 * @fn fixl2ifnecessary(void)
 * @brief check state of l2 and if necessary fix it to what we want
 * @return zero or error
 */
static int fixl2ifnecessary(void) 
{
	int rv;
	int rport=UDP_DEF_PORT; // take the default below
	l2state=ltp_l2_get_state();
	if (l2state==LTP_L2S_RESET) {
#ifdef LTP_EXTRA_VERBOSE
		char 		befstr[MAXPATH],aftstr[MAXPATH];
		snprintf(befstr,MAXPATH,"%s",ltp_now_str());
#endif
		rv=ltp_l2_init(defcbs_p,rport);
		if (rv) LTPERR(rv);
		l2state=ltp_l2_get_state();
#ifdef LTP_EXTRA_VERBOSE
		snprintf(aftstr,MAXPATH,"%s",ltp_now_str());
		snprintf(ltpstr,LTPSTR_LEN,"fixl2ifnecessary/ltp_l2_init at %d from %s to %s\n",__LINE__,befstr,aftstr);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	}
/// turns on l2 if necessary, alternative is to have ltp_cue i/f do it???
#define L2ON_NOT_CUE 
#ifdef L2ON_NOT_CUE
	if (l2state==LTP_L2S_INITED) {
#ifdef LTP_EXTRA_VERBOSE
		char 		befstr[MAXPATH],aftstr[MAXPATH];
		snprintf(befstr,MAXPATH,"%s",ltp_now_str());
#endif
		rv=ltp_l2_on(&l2mtu,l2desiredstate);
		if (rv) LTPERR(rv);
		l2state=ltp_l2_get_state();
#ifdef LTP_EXTRA_VERBOSE
		snprintf(aftstr,MAXPATH,"%s",ltp_now_str());
		snprintf(ltpstr,LTPSTR_LEN,"fixl2ifnecessary/ltp_l2_on at %d from %s to %s\n",__LINE__,befstr,aftstr);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	}
#endif
	return (LTP_OK);
}

#ifdef PING_THREADS

/*!
 * @brief thread initialisiation function that repeatedly calls ltp_ping()
 */
void *pinger_fnc (void* foo)
{
	static int pinger_errors=0; // the number of consequetive errors seen from ltp_ping
	bool lasterr=false;	 // whether or not the last call returned an error
	// start rng stream for this thread
	int rv=ltp_rand_init(0xff00);
	if (rv) pthread_exit(NULL);
	while (true) {
		GRABMUTEX
		rv=ltp_ping();
		if (rv) {
			if (lasterr) {
				pinger_errors++;
			} else {
				pinger_errors=0;
				lasterr=true;
			}
			if (pinger_errors >= LTP_MAX_PING_ERRORS) {
				snprintf(ltpstr,LTPSTR_LEN,"ltp_ping returned %d - already got %d errors, so exiting!\n",rv,pinger_errors);
				ltp_log(LTP_LOG_ERR,ltpstr);
				DROPMUTEX
				pthread_exit(NULL);
			} else {
				if (lasterr) {
					snprintf(ltpstr,LTPSTR_LEN,"ltp_ping returned %d-th consequtive error: %d - sleeping a second to let things improve!\n",
							pinger_errors,rv);
				} else {
					snprintf(ltpstr,LTPSTR_LEN,"ltp_ping returned %d - sleeping a second to let things improve!\n",rv);
				}
				ltp_log(LTP_LOG_ERR,ltpstr);
				sleep(1);
			}
		} else {
			lasterr=false;
		}
		DROPMUTEX
		ltp_waitabit();
	}
	return NULL;
}

// a message
static const char		*pinger_message="pinger_thread";

// a message
static const char		*forwarder_message="forwarder_thread";

/*!
 * @brief initialise a "pinging" thread
 * @return zero or error
 */
static int ltp_ping_init(void)
{
	static bool done=false;
	if (done) return(LTP_OK);
	// try to set l2 into our desired state
	int rv=fixl2ifnecessary();
	if (rv) LTPERR(rv);
	// now start a thread for pinging
	snprintf(ltpstr,LTPSTR_LEN,"About to start pinger thread\n");
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
	ltp_log_flush();
	pthread_t tmpThread;
	rv=pthread_create(&tmpThread,NULL,pinger_fnc,(void*)pinger_message);
	if (rv) LTPERR(rv);
	pinger=tmpThread;
#ifdef LTP_TARGET_ARM_LINUX
	sleep(1);
	snprintf(ltpstr,LTPSTR_LEN,"Started pinger thread - tid(%p) then slept a second\n",pinger);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
	ltp_log_flush();
#else
	snprintf(ltpstr,LTPSTR_LEN,"Started pinger thread - tid(%p)\n",(void*)pinger);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
#endif
	done=true;
	return(LTP_OK);
}
#endif

/*!
 * @brief fix a parent pointer that's been moved if necessary
 * @param tree is the ltp_state variable containing everything
 * @param parent is the pointer that may have been moved - on output its the correct value
 * @param parentid is the ltp_instance that we really want
 * @param type is the type of list we're on about
 * @return zero or error as usual 
 *
 * This function is needed since realloc() can move pointers. This
 * could happen as follows: during ltp_sendto() we allocate 
 * a new block in a socket, and then allocate some segments whose
 * parent pointers point to that newly allocated block. Later on
 * another block is added to this socket, but causing the blocks
 * pointer to be moved in order to get more contiguous space (whether
 * this happens will be OS/compiler/runtime dependent, but we do
 * know it might happen). At that point the parent fields in the
 * segment structures are bogus, but the parentid fields are still
 * good.
 * 
 * What we therefore do is before using any parent pointer
 * we first call here to see if the parent pointer and parentid
 * ltp_instance are still in whack. If they're not then we 
 * search through our lists and fix the parent pointer or
 * else return a system error (in which case: gulp!)
 *
 * Okay I know someone'll say I ought to be using some STL or
 * other C++ list type or something, but I still don't know
 * what I've to port and hell this should work! If sbrk() helps
 * out then we mightn't even have to do very much work in here
 * very often (hmm....)
 */
int ltp_fixparent(ltp_state *tree,void **parent,ltp_instance parentid, ltp_parent type)
{
	if (!parent) LTPERR(LTP_BAD_PARAM);
	if (*parent) { // see if its ok already
		switch (type) {
			case LTP_PARENT_SOCKET:
			{
				// if the thing is in the relevant range then assume we can de-reference for checking purposes
				if (*parent >= tree->sockets && *parent <= &(tree->sockets[tree->nsockets-1])) {
					ltp_socket_t *dad=(ltp_socket_t*)*parent;
					if (dad->id==parentid) return(LTP_OK);
				}
			}
			break;
			case LTP_PARENT_BLOCK:
			{
				for (int i=0;i!=tree->nsockets;i++) {
					if (*parent >= tree->sockets[i].blocks && *parent <= &(tree->sockets[i].blocks[tree->sockets[i].nblocks-1])) {
						ltp_block *dad=(ltp_block*)*parent;
						if (dad->id==parentid) return(LTP_OK);
					}
				}
			}
			break;
			case LTP_PARENT_SEGMENT:
			{
				if (*parent >= tree->txq.segments && *parent <= &(tree->txq.segments[tree->txq.nsegments-1])) {
					ltp_segment *dad=(ltp_segment*)*parent;
					if (dad->id==parentid) return(LTP_OK);
				}
			}
			break;
			default:
				LTPERR(LTP_BAD_PARAM);
		}
	}
	// oops - got here means that dad's id ain't what we hoped. Better
	// search down all the ble&*in lists :-(
	if (type==LTP_PARENT_SOCKET) {
		for (int i=0;i!=tree->nsockets;i++) {
			if (tree->sockets[i].id==parentid) {
#ifdef LTP_EXTRA_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Fixed ptr to parent socket %d (id=%d)\n",i,parentid);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				*parent=(void*)&tree->sockets[i];
				return(LTP_OK);
			}
		}
		// if we get here then ...
		LTPERR(LTP_ERR); // bo*&ix!!!
	}
	if (type==LTP_PARENT_BLOCK) {
		for (int i=0;i!=tree->nsockets;i++) {
			for (int j=0;j!=tree->sockets[i].nblocks;j++) {
				if (tree->sockets[i].blocks[j].id==parentid) {
#ifdef LTP_EXTRA_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"Fixed ptr to parent block %d of socket %d (id=%d)\n",j,i,parentid);
					ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					*parent=(void*)&tree->sockets[i].blocks[j];
					return(LTP_OK);
				}
			}
		}
		// if we get here then ...
		LTPERR(LTP_ERR); // bo*&ix!!!
	}
	if (type==LTP_PARENT_SEGMENT) {
		// segments can be found either in blocks or else in other ltp_state_var places
		for (int i=tree->txq.nsegments-1;i>=0;i--) { // search down, later are probably more likely hits
			if (tree->txq.segments[i].id==parentid) {
#ifdef LTP_EXTRA_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Fixed ptr to parent segment %d of (id=%d)\n",i,parentid);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				*parent=(void*)&tree->txq.segments[i];
				return(LTP_OK);
			}
		}
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Can't find parent segment with id=%d\n",parentid);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		LTPERR(LTP_ERR); // bo*&ix!!!
	}
	return(LTP_OK);
}

/*!
 * @brief get rid of a timer
 * @param type is the type of timer e.g. LTP_TIMER_CP
 * @param thing helps us find the timer, either as a serial number or else a socket id
 * @return zero or error
 */
static int kill_timer(ltp_timer_type type, void *thing)
{
	int rv;
	ltpc_octets *number=(ltpc_octets*)thing;
	ltp_socket_t* sock=(ltp_socket_t*)thing;
	bool killed=false;
	ltp_timer_detail *details=ltp_state_var.timers.details;
	switch (type) {
		case LTP_TIMER_CP:
			for (int i=0;!killed && i!=ltp_state_var.timers.nentries;i++) {
				if (!details[i].finishedwith && details[i].type==LTP_TIMER_CP) {
					rv=ltp_fixparent(&ltp_state_var,(void**)&details[i].segment,details[i].segmentid,LTP_PARENT_SEGMENT);
					if (rv) LTPERR(rv);
					if (details[i].segment->seg.val.ds.csn.len==number->len
						&&
						!memcmp(details[i].segment->seg.val.ds.csn.val,number->val,number->len)) {
						details[i].finishedwith=true;
#ifdef LTP_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Killing CP timer %d of %d, for thing %d\n",
											i,ltp_state_var.timers.nentries,details[i].segmentid);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						killed=true;
					}
				}
			}
			break;

		case LTP_TIMER_RS:
			for (int i=0;!killed && i!=ltp_state_var.timers.nentries;i++) {
				if (!details[i].finishedwith && details[i].type==LTP_TIMER_RS) {
					rv=ltp_fixparent(&ltp_state_var,(void**)&details[i].segment,details[i].segmentid,LTP_PARENT_SEGMENT);
					if (rv) LTPERR(rv);
					if (details[i].segment->seg.val.rs.rsn.len==number->len
						&&
						!memcmp(details[i].segment->seg.val.rs.rsn.val,number->val,number->len)) {
						details[i].finishedwith=true;
#ifdef LTP_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Killing RS timer %d of %d, for thing %d\n",
											i,ltp_state_var.timers.nentries,details[i].segmentid);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						killed=true;
					}
				}
			}
			break;

		case LTP_TIMER_C:
			for (int i=0;!killed && i!=ltp_state_var.timers.nentries;i++) {
				if (!details[i].finishedwith && details[i].type==LTP_TIMER_C) {
					rv=ltp_fixparent(&ltp_state_var,(void**)&details[i].segment,details[i].segmentid,LTP_PARENT_SEGMENT);
					if (rv) LTPERR(rv);
					if (details[i].segment->seg.sno.len==number->len
						&&
						!memcmp(details[i].segment->seg.sno.val,number->val,number->len)) {
						details[i].finishedwith=true;
#ifdef LTP_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Killing C timer %d of %d, for thing %d\n",
											i,ltp_state_var.timers.nentries,details[i].segmentid);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						killed=true;
					}
				}
			}
			break;

		case LTP_TIMER_INACTIVITY:
			for (int i=0;!killed && i!=ltp_state_var.timers.nentries;i++) {
				if (!details[i].finishedwith && details[i].type==LTP_TIMER_INACTIVITY) {
					rv=ltp_fixparent(&ltp_state_var,(void**)&details[i].socket,details[i].socketid,LTP_PARENT_SOCKET);
					if (rv) LTPERR(rv);
					if (details[i].socket->id==sock->id) {
						killed=true;
#ifdef LTP_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Killing socket timer %d of %d, for thing %d\n",
											i,ltp_state_var.timers.nentries,details[i].socketid);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						details[i].finishedwith=true;
					}
				}
			}
			break;

		case LTP_TIMER_COOKIE:
			for (int i=0;!killed && i!=ltp_state_var.timers.nentries;i++) {
				if (!details[i].finishedwith && details[i].type==LTP_TIMER_COOKIE) {
					rv=ltp_fixparent(&ltp_state_var,(void**)&details[i].socket,details[i].socketid,LTP_PARENT_SOCKET);
					if (rv) LTPERR(rv);
					if (details[i].socket->id==sock->id) {
						killed=true;
#ifdef LTP_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Killing cookie timer %d of %d, for thing %d\n",
											i,ltp_state_var.timers.nentries,details[i].socketid);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						details[i].finishedwith=true;
					}
				}
			}
			break;

		default:
			LTPERR(LTP_BAD_PARAM);
	}
	if (killed)
		return(LTP_OK);
	else
		return(LTP_BAD_VALUE);
}

/*!
 * @brief initialise (but don't start!) an inactivity timer 
 * @param type is the type of thing that indexes the timer, one of LTP_PARENT_BLOCK, LTP_PARENT_SEGMENT, LAP_PARENT_SOCKET
 * @param parent is (a ptr to) the thing
 * @param existing is whether this is a new initialisation or a re-initialisation
 * @return zero or error
 */
static int init_inactivity_timer(int type, void *parent, bool existing) 
{
	ltp_socket_t	*sock=NULL;
	ltp_block	*block=NULL;
	ltp_segment	*seg=NULL;
	void *dad=parent;
	int rv;
	switch (type) {
	case LTP_PARENT_SEGMENT:
				seg=(ltp_segment*)dad;
				rv=ltp_fixparent(&ltp_state_var,(void**)&seg->parent,seg->parentid,LTP_PARENT_BLOCK);
				if (rv) LTPERR(rv);
				dad=seg->parent;
	case LTP_PARENT_BLOCK:
				block=(ltp_block*)dad;
				rv=ltp_fixparent(&ltp_state_var,(void**)&block->parent,block->parentid,LTP_PARENT_SOCKET);
				if (rv) LTPERR(rv);
				dad=block->parent;
	case LTP_PARENT_SOCKET:
				sock=(ltp_socket_t*)dad;
				break;
	default:
			LTPERR(LTP_BAD_PARAM);
	}
	int pval=(int)floor(sock->opts.inactivitylimit); 
	int cval=1; 

	if (!existing) {
		int alloced=ltp_state_var.timers.alloced;
		int nentries=ltp_state_var.timers.nentries;
		nentries++;
		FOOREALLOC(ltp_timer_detail,ltp_state_var.timers.details,nentries,alloced); 
		ltp_timer_detail *det=&ltp_state_var.timers.details[nentries-1];
		det->id=NI;
		det->type=LTP_TIMER_INACTIVITY;
		det->finishedwith=false;
		det->active=true;
		det->lastdecremented=(ltp_time_t)ltp_now();
		det->remaining=pval;
		det->counter=cval;
		det->block=block;
		if (det->block) det->blockid=block->id;
		else det->blockid=0;
		det->segment=seg;
		if (det->segment) det->segmentid=seg->id;
		else det->segmentid=0;
		det->socket=sock;
		if (det->socket) det->socketid=sock->id;
		else det->socketid=0;
		ltp_state_var.timers.alloced=alloced;
		ltp_state_var.timers.nentries=nentries;
	} else { // re-init
		bool found=false;
		int index;
		ltp_timer_detail *ent=NULL;
		for (index=0;!found && index!=ltp_state_var.timers.nentries;index++) {
			if (sock->id==ltp_state_var.timers.details[index].socketid) {
				found=true;
				ent=&ltp_state_var.timers.details[index];
			}
		}
		if (!found) {
			LTPERR(LTP_BAD_INPUT);
		} 
		ent->finishedwith=false;
		ent->active=true;
		ent->remaining=pval;
		ent->lastdecremented=(ltp_time_t)ltp_now();
		ent->counter=cval;
	}
	LTP_STATE_CHANGED;
	return(LTP_OK);
}

/*!
 * @brief initialise (but don't start!) a timer of the specified type for the specified segment
 * @param type is the type of timer e.g. LTP_TIMER_CP
 * @param parent is (a ptr to) the segment who's transmission starts the timer
 * @return zero or error
 */
static int init_timer(
				ltp_timer_type type,
				ltp_segment *parent) 
{
	int		cval=0;
	ltp_time_t	pval=(ltp_time_t)0;

	if (!parent) LTPERR(LTP_BAD_PARAM);

	// setup values to store and sanity check in one
	ltp_time_t now=ltp_now();
	ltp_time_t ltt;
#define THISWAY
#ifdef THISWAY
	int rv=ltp_ltt(&parent->peer,now,&ltt);
	if (rv) LTPERR(rv);
#else
	int rv=ltp_fixparent(&ltp_state_var,(void**)&parent->parent,parent->parentid,LTP_PARENT_BLOCK);
	if (rv) LTPERR(rv);
	ltp_block *theblock=parent->parent;
	rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
	if (rv) LTPERR(rv);
	rv=ltp_ltt(&theblock->parent->obpeer,now,&ltt);
	if (rv) LTPERR(rv);
#endif
#ifdef LTP_EXTRA_EXTRA_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"init_timer: type is %d parent (seg) is %d ltt is %g\n",
				type,parent->id,ltt);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	switch (type) {
		case LTP_TIMER_CP: pval=ltp_mib_var.time2report; cval=ltp_mib_var.txlimit; break;
		case LTP_TIMER_RS: pval=ltp_mib_var.time2rack; cval=ltp_mib_var.reptxlimit; break;
		case LTP_TIMER_C: pval=ltp_mib_var.time2cack; cval=ltp_mib_var.cantxlimit; break;

		// this isn't for inactivity timers
		case LTP_TIMER_INACTIVITY: 
		default:
			LTPERR(LTP_BAD_PARAM);
	}

	int alloced=ltp_state_var.timers.alloced;
	int nentries=ltp_state_var.timers.nentries;
	nentries++;
	FOOREALLOC(ltp_timer_detail,ltp_state_var.timers.details,nentries,alloced); 
	ltp_timer_detail *det=&ltp_state_var.timers.details[nentries-1];
	det->id=NI;
	det->type=type;
	det->finishedwith=false;
	det->active=false; // until the segment's actually sent
	det->everexpired=false;
	det->lastdecremented=(ltp_time_t)ltp_now();
	det->remaining=pval;
	det->counter=cval;
	det->segment=parent;
	det->segmentid=parent->id;
	det->block=NULL;
	det->blockid=0;
	det->socket=NULL;
	det->socketid=0;
	det->absstart=now+2*ltt;
	det->rtt=2*ltt;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"init_timer: absstart is %s, rtt is %g\n",
				ltp_time_str(det->absstart),det->rtt);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	// store index of timer in seg 
	parent->t_entry=nentries-1;
	ltp_state_var.timers.alloced=alloced;
	ltp_state_var.timers.nentries=nentries;
	LTP_STATE_CHANGED;
	return(LTP_OK);
}

/*!
 * @brief say if a timer has gone around already (indicate a re-tx)
 * @param entry is the timer entry
 * @return true if this has every expired before
 */
static bool old_timer(int entry)
{
	if (entry < 0 || entry >= ltp_state_var.timers.nentries) {
		return false;
	}
	ltp_timer_detail *det=&ltp_state_var.timers.details[entry];
	return det->everexpired;
}

/*!
 * @brief re-initialise a timer
 * @param entry is the timer entry
 * @return zero or error
 */
static int reinit_timer(int entry)
{
	if (entry >= ltp_state_var.timers.nentries) {
		snprintf(ltpstr,LTPSTR_LEN,"%s: bad entry (%d) for timer (only %d exist)\n",__FUNCTION__,entry,
				ltp_state_var.timers.nentries);
		ltp_log(LTP_LOG_ERR,ltpstr);
		LTPERR(LTP_BAD_PARAM);
	}
	ltp_timer_detail *det=&ltp_state_var.timers.details[entry];
	ltp_time_t	pval=(ltp_time_t)0;

	// setup values to store and sanity check in one
	switch (det->type) {
		case LTP_TIMER_CP: pval=ltp_mib_var.time2report; break;
		case LTP_TIMER_RS: pval=ltp_mib_var.time2rack; break;
		case LTP_TIMER_C: pval=ltp_mib_var.time2cack; break;
		default:
			LTPERR(LTP_BAD_PARAM);
	}
	det->lastdecremented=ltp_now();
	det->remaining=pval;
	det->active=true;
	ltp_time_t now=ltp_now();
	det->absstart=now+det->rtt;
	return(LTP_OK);
}



/*!
 * @brief hash a buffer
 * @param buf is the ptr to the buffer
 * @param len is the length of the buffer
 * @param hash is the output hash
 * @return zero or error
 */
static int do_hash(unsigned char *buf, int len, ltpc_octets *hash)
{
#ifdef SUMHASH
	int sum=0;
	for (int i=0;i!=len;i++) sum+=buf[i];
	hash->len=sizeof(int);
	hash->val=(unsigned char*)calloc(1,sizeof(int));
	if (!hash->val) LTPERR(LTP_MEMORY);
	memcpy(hash->val,&sum,hash->len);
	return(LTP_OK);
#endif

#ifdef SHA256HASH
	
   	SHA256_ctx ctx;
	int i;
	long secondCount;
	unsigned char *digest=(unsigned char*)calloc(1,32);
	if (!digest) LTPERR(LTP_MEMORY);

	SHA256_init( &ctx );
	SHA256_update( &ctx, buf, len);
	SHA256_final( &ctx );

	SHA256_digest( &ctx, digest );

	hash->val=digest;
	hash->len=32;

	return(LTP_OK);

#endif
}

/*!
 * @brief check if an rx'd buffer is a replay and record (a hash of) it if not
 * @param buf is the ptr to the buffer
 * @param len is the length of the buffer
 * @param remember is true if this buffer's hash ought to be added into the replay cache
 * @param entry is a handle for the colliding or new hash entry 
 * @param responded is set to true if there is a previous response for this entry
 * @return true if we've seen this before
 */
static bool is_replay(
				unsigned char *buf, 
				int len, 
				bool remember, 
				int *entry,
				bool *responded)
{
	bool found=false;
	int nentries=ltp_state_var.replays.nentries;
	int alloced=ltp_state_var.replays.alloced;
	ltp_cache_detail	*details=ltp_state_var.replays.details;
	ltpc_octets *hp;
	ltpc_octets hash;
	do_hash(buf,len,&hash);
	int theentry=LTP_NOCACHE;
	// loop goes backwards since replay is more like to hit recently?
	for (int i=nentries;!found && i!=0;i--) {
		hp=&details[i-1].hash;
		if (hash.len==hp->len
			&&
			!memcmp(hash.val,hp->val,hash.len)) {
				found=true; // oops!
				theentry=i;
			}
	}
	if (remember && !found) {
		nentries++;
		FOOREALLOC(ltp_cache_detail,details,nentries,alloced);
		ltp_state_var.replays.nentries=nentries;
		ltp_state_var.replays.alloced=alloced;
		ltp_state_var.replays.details=details;

		details[nentries-1].id=NI;
		details[nentries-1].hash.len=hash.len;
		details[nentries-1].hash.val=hash.val;
		details[nentries-1].response.len=0;
		details[nentries-1].response.val=NULL;
		details[nentries-1].finishedwith=false;
		theentry=nentries-1;
	}
	*entry=theentry;
	if (details && theentry!= LTP_NOCACHE && details[theentry].response.len) *responded=true;
	else *responded=false;
	if (!remember) {
		free(hash.val);
	}
	return(found);
}

/*!
 * @brief add a response matched with a replay cache entry
 * @param seg is the ltp_segment structure which we're potentially going to replay
 * @param buf is the ptr to the buffer
 * @param len is the length of the buffer
 * @return zero or error
 */
static int add_replay_response(
				ltp_segment *seg,
				unsigned char *buf, 
				int len)
{
	if (!seg) LTPERR(LTP_BAD_PARAM);
	int entry=seg->r_entry;
	ltp_segment_type type=seg->seg.type;
	ltp_instance pid=seg->parentid;
	int nentries=ltp_state_var.replays.nentries;
	int alloced=ltp_state_var.replays.alloced;
	ltp_cache_detail *details=ltp_state_var.replays.details;
	if (!len || !buf || (entry <0 && entry!=LTP_CACHE) || entry > nentries) LTPERR(LTP_BAD_PARAM);
	if (entry==LTP_CACHE) { // make more space 
		entry=nentries;
		nentries++;
		FOOREALLOC(ltp_cache_detail,details,nentries,alloced);
		ltp_state_var.replays.nentries=nentries;
		ltp_state_var.replays.alloced=alloced;
		ltp_state_var.replays.details=details;
		memset(&details[entry],0,sizeof(ltp_cache_detail));
		details[entry].id=NI;
		seg->r_entry=entry;
		seg->r_id=details[entry].id;
	}
	ltpc_octets *resp=&ltp_state_var.replays.details[entry].response;
	if (resp->val != NULL || resp->len!=0) {
#ifdef LTP_VERBOSE
		ltp_log(LTP_LOG_MAX,(char*)"Ignoring newer replay response in favour of earlier\n");
#endif
	} else {
		resp->val=ltpc_alloc(len);
		if (!resp->val) LTPERR(LTP_MEMORY);
		memcpy(resp->val,buf,len);
		resp->len=len;
		ltp_state_var.replays.details[entry].type=type;
		ltp_state_var.replays.details[entry].parentid=pid;
	}
	return(LTP_OK);
}

/*!
 * @brief send out a previously sent segment
 * @param entry is the index in the replay cache
 * @param peer is the guy to sent it to
 * @param plen is the size of peer
 * @return zero or error
 */
static int resend_from_replay(int entry, ltpaddr *peer, int plen)
{

	if (entry <0 || entry >= ltp_state_var.replays.nentries) 
			LTPERR(LTP_BAD_PARAM);
	ltpc_octets *resp=&ltp_state_var.replays.details[entry].response;
	if (resp->val == NULL || resp->len==0) LTPERR(LTP_BAD_VALUE);

	// add a new segment for tx
	NEWSEGMENT
	segments[nsegments].seg.type=LTPC_SEG_ENCODED;

	// let's try shallow copy first - might be ok
	segments[nsegments].encoded=*resp;

	segments[nsegments].r_entry=entry; // in case we need to go again
	segments[nsegments].r_id=ltp_state_var.replays.details[entry].id;

	segments[nsegments].parent=0;
	segments[nsegments].parentid=ltp_state_var.replays.details[entry].parentid;
	memcpy(&segments[nsegments].peer,peer,plen);
	segments[nsegments].peerlen=plen;
	segments[nsegments].encoded_type=ltp_state_var.replays.details[entry].type;
	switch (ltp_state_var.replays.details[entry].type) {
		case LTPC_SEG_DS:
				init_timer(LTP_TIMER_CP,&segments[nsegments]); break;
		case LTPC_SEG_RS:
				init_timer(LTP_TIMER_RS,&segments[nsegments]); break;
		case LTPC_SEG_RA:
		case LTPC_SEG_CAO:
		case LTPC_SEG_CAR:
				// nothing to do
				; break;
		default:
			LTPERR(LTP_BAD_VALUE);
	}

#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,
					"MSG: Queueing an encoded segment for LTP[id:%s,port:%d]\n",
					inet_ntoa(peer->sock.sin_addr),peer->sock.sin_port);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif

	LTP_STATE_CHANGED;
	return(LTP_OK);
}

/*!
 * @brief another wrapper for dumping scope -> log
 * @param socpe is a scope ptr
 * @param lineno is the calling code line
 * @param scopetype is the type of scope we're logging
 * @return zero or error
 */
static int lowlogscope(int nscopes, int *lbs, int *ubs, int lineno, int scopetype, int id)
{
	char *dbgstr=NULL;
	const char *dbgstrs[5]={"dunno", "rx", "tx", "fwd", "sub"};
	dbgstr=(char*)dbgstrs[scopetype];
	ltpstr[0]='\0'; 
	char *psfoo=ltpstr; 
	int pslenleft=LTPSTR_LEN; 
	snprintf(psfoo,pslenleft,"Scope %s, Line: %d, Block %d has %d scopes: ",dbgstr,lineno,id,nscopes); 
	pslenleft= LTPSTR_LEN-strlen(ltpstr); 
	psfoo=ltpstr+strlen(ltpstr); 
	if (nscopes >= 8 ) {
		snprintf(psfoo,pslenleft, "...too many to bother with, here's 4\n");
		pslenleft= LTPSTR_LEN-strlen(ltpstr); 
		psfoo=ltpstr+strlen(ltpstr); 
		int scopes2print=4;
		for (int psPS=0;pslenleft>0 && psPS!=scopes2print/2;psPS++) { 
			snprintf(psfoo,pslenleft,"(%d,%d),", lbs[psPS], ubs[psPS]); 
			pslenleft= LTPSTR_LEN-strlen(ltpstr); 
			psfoo=ltpstr+strlen(ltpstr); 
		}
		snprintf(psfoo,pslenleft,"...");
		pslenleft= LTPSTR_LEN-strlen(ltpstr); 
		psfoo=ltpstr+strlen(ltpstr); 
		for (int psPS=nscopes-scopes2print/2;pslenleft>0 && psPS!=nscopes;psPS++) { 
			snprintf(psfoo,pslenleft,"(%d,%d),", lbs[psPS], ubs[psPS]); 
			pslenleft= LTPSTR_LEN-strlen(ltpstr); 
			psfoo=ltpstr+strlen(ltpstr); 
		}
	} else {
		for (int psPS=0;pslenleft>0 && psPS!=nscopes;psPS++) { 
			snprintf(psfoo,pslenleft,"(%d,%d),", lbs[psPS], ubs[psPS]); 
			pslenleft= LTPSTR_LEN-strlen(ltpstr); 
			psfoo=ltpstr+strlen(ltpstr); 
		}
	}
	if (pslenleft>0) snprintf(psfoo,pslenleft,"\n"); 
	else { ltp_log(LTP_LOG_MAX,(char*)"ltpstr not big enough for scopes!\n"); } 
	ltp_log(LTP_LOG_MAX,ltpstr); 
	return(LTP_OK);
}

/* @brief dump a scope to the log
 * @param theblock is a guess what...
 * @param lineno says where we're called from
 * @praam scopetype says tx,rx or fwd
 * @return zero - no errors
 */
static int logscope(ltp_block *theblock, int lineno, int scopetype)
{ 
	int *lbs=NULL;
	int *ubs=NULL;
	int nscopes=0;
	if (scopetype==LTP_TX_SCOPES) {
		lbs=theblock->tx_scopes.lbs;
		ubs=theblock->tx_scopes.ubs;
		nscopes=theblock->tx_scopes.nscopes;
	} else if (scopetype==LTP_RX_SCOPES) {
		lbs=theblock->rx_scopes.lbs;
		ubs=theblock->rx_scopes.ubs;
		nscopes=theblock->rx_scopes.nscopes;
	} else if (scopetype==LTP_FWD_SCOPES) {
		lbs=theblock->fwd_scopes.lbs;
		ubs=theblock->fwd_scopes.ubs;
		nscopes=theblock->fwd_scopes.nscopes;
	} else {
		LTPERR(LTP_BAD_VALUE);
	}
	int rv=lowlogscope(nscopes, lbs, ubs, lineno, scopetype, theblock->id);
	if (rv) LTPERR(rv);
	return(LTP_OK);
} 

#ifdef LTP_VERBOSE
#define PRINTSCOPE logscope(theblock,__LINE__,scopetype);
#else
#define PRINTSCOPE 
#endif

/*!
 * @brief say if a block's scopes are now complete
 * @param theblock is the block 
 * @return true if no gaps remain to be filled
 */
static bool complete_scope(ltp_block *theblock)
{
	bool complete=true;
	int *lbs=theblock->rx_scopes.lbs;
	int *ubs=theblock->rx_scopes.ubs;
	int nscopes=theblock->rx_scopes.nscopes;
	if (lbs[0]!=0) complete=false;
	for (int i=0;complete && i!=nscopes-1;i++) {
		if (lbs[i+1]!=ubs[i]) {
			complete=false; // bugger a gap exists
		}
	}
	return(complete);
}

/*!
 * @brief add a new pair of values to a block's scopes
 * @param theblock is what you'd guess
 * @param newlb is the new lower bound value to add
 * @param newub is the new upper bound value to add
 * @param scopetype if tx,rx or fwd
 * @return zero or error
 */
static int addnewscope(ltp_block *theblock, int newlb, int newub, int scopetype)
{
	int alloced;
	int *lbs=NULL;
	int *ubs=NULL;
	int nscopes=0;

	char *dbgstr=NULL;
	const char *dbgstrs[5]={"dunno", "rx", "tx", "fwd", "sub"};
	if (scopetype==LTP_TX_SCOPES) {
		lbs=theblock->tx_scopes.lbs;
		ubs=theblock->tx_scopes.ubs;
		nscopes=theblock->tx_scopes.nscopes;
		alloced=theblock->tx_scopes.alloced;
	} else if (scopetype==LTP_RX_SCOPES) {
		lbs=theblock->rx_scopes.lbs;
		ubs=theblock->rx_scopes.ubs;
		nscopes=theblock->rx_scopes.nscopes;
		alloced=theblock->rx_scopes.alloced;
			} else if (scopetype==LTP_FWD_SCOPES) {
				lbs=theblock->fwd_scopes.lbs;
		ubs=theblock->fwd_scopes.ubs;
		nscopes=theblock->fwd_scopes.nscopes;
		alloced=theblock->fwd_scopes.alloced;
	} else {
		LTPERR(LTP_BAD_VALUE);
	}
	dbgstr=(char*)dbgstrs[scopetype];

	// actually seems to have happened in an error generation case and 
	// messed up mightily!
	if (newlb>=newub || newlb <0 || newub<=0 ) {
#ifdef LTP_EXTRA_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): dumb scopes!!\n",dbgstr,newlb,newub);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		LTPERR(LTP_BAD_PARAM);
	}

	if (nscopes==0) { // 1st time in - easy
#ifdef LTP_EXTRA_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): adding 1st scope\n",dbgstr,newlb,newub);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		nscopes=1;
		alloced=1;
		int *newlbs=(int*)calloc(nscopes,sizeof(int));
		if (!newlbs) LTPERR(LTP_MEMORY);
		int *newubs=(int*)calloc(nscopes,sizeof(int));
		if (!newubs) LTPERR(LTP_MEMORY);
		newlbs[0]=newlb;
		newubs[0]=newub;

		if (scopetype==LTP_TX_SCOPES) {
			theblock->tx_scopes.lbs=newlbs;
			theblock->tx_scopes.ubs=newubs;
			theblock->tx_scopes.nscopes=nscopes;
			theblock->tx_scopes.alloced=alloced;
		} else if (scopetype==LTP_RX_SCOPES) {
			theblock->rx_scopes.lbs=newlbs;
			theblock->rx_scopes.ubs=newubs;
			theblock->rx_scopes.nscopes=nscopes;
			theblock->rx_scopes.alloced=alloced;
		} else if (scopetype==LTP_FWD_SCOPES) {
			theblock->fwd_scopes.lbs=newlbs;
			theblock->fwd_scopes.ubs=newubs;
			theblock->fwd_scopes.nscopes=nscopes;
			theblock->fwd_scopes.alloced=alloced;
		}
		PRINTSCOPE;

		return(LTP_OK);
	}
	
	// new ones will often completely span the current set
	// so both left and right are beyond previous limits we go to one scope so
	if (newlb<=lbs[0] && newub>=ubs[nscopes-1]) {
		nscopes=1;
		lbs[0]=newlb;
		ubs[0]=newub;

		if (scopetype==LTP_TX_SCOPES) {
			theblock->tx_scopes.nscopes=nscopes;
		} else if (scopetype==LTP_RX_SCOPES) {
			theblock->rx_scopes.nscopes=nscopes;
		} else if (scopetype==LTP_FWD_SCOPES) {
			theblock->fwd_scopes.nscopes=nscopes;
		}
		PRINTSCOPE;

		return(LTP_OK);
	}

	// now take care of case where new stuff is just after the end
	// which ought to be the most common case (if layer 2 is being
	// good)
	if (newub >= ubs[nscopes-1]) {
		if (newlb <= (ubs[nscopes-1]+1)) { // adjacent or overlapping
			if (newlb>= lbs[nscopes-1]) { // just extend right
#ifdef LTP_EXTRA_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): extending last scope\n",dbgstr,newlb,newub);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				ubs[nscopes-1]=newub;
			} else { // bit of munging to do - maybe some other gaps are filled too!
				int leftplace=0;
				bool leftfound=false;
				for (int sc=0;!leftfound && sc!=nscopes;sc++) {
						if (newlb <= lbs[sc]) {
									leftplace=sc;
									leftfound=true;
						}
				}
				if (!leftfound) { // hmm...shouldn't happen
					LTPERR(LTP_BAD_VALUE);
				} else {
					// reduce the number of scopes appropriately and 
					// fix the ub for that place
#ifdef LTP_EXTRA_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): right-spanning scope\n",dbgstr,newlb,newub);
					ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					lbs[leftplace]=newlb;
					ubs[leftplace]=newub;
					nscopes=leftplace+1;

					if (scopetype==LTP_TX_SCOPES) {
						theblock->tx_scopes.nscopes=nscopes;
					} else if (scopetype==LTP_RX_SCOPES) {
						theblock->rx_scopes.nscopes=nscopes;
					} else if (scopetype==LTP_FWD_SCOPES) {
						theblock->fwd_scopes.nscopes=nscopes;
					}
				}
			}
			PRINTSCOPE;
			return(LTP_OK);
		} else { // beyond current, but preceeded by a gap
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): adding new scope\n",dbgstr,newlb,newub);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			nscopes++;
			alloced+=1;
			int *newlbs=(int*)realloc(lbs,nscopes*sizeof(int));
			if (!newlbs) LTPERR(LTP_MEMORY);
			int *newubs=(int*)realloc(ubs,nscopes*sizeof(int));
			if (!newubs) LTPERR(LTP_MEMORY);
			newlbs[nscopes-1]=newlb;
			newubs[nscopes-1]=newub;

			if (scopetype==LTP_TX_SCOPES) {
				theblock->tx_scopes.lbs=newlbs;
				theblock->tx_scopes.ubs=newubs;
				theblock->tx_scopes.nscopes=nscopes;
				theblock->tx_scopes.alloced=alloced;
			} else if (scopetype==LTP_RX_SCOPES) {
				theblock->rx_scopes.lbs=newlbs;
				theblock->rx_scopes.ubs=newubs;
				theblock->rx_scopes.nscopes=nscopes;
				theblock->rx_scopes.alloced=alloced;
			} else if (scopetype==LTP_FWD_SCOPES) {
				theblock->fwd_scopes.lbs=newlbs;
				theblock->fwd_scopes.ubs=newubs;
				theblock->fwd_scopes.nscopes=nscopes;
				theblock->fwd_scopes.alloced=alloced;
			}
			PRINTSCOPE;

			return(LTP_OK);
		}
	}

	// there is a chance that this scope is for bytes which are to the "left"
	// of those I've already gotten, so better handle that now.
	if (newlb < lbs[0] ) {
		if ( newub >= (lbs[0]-1)) { // left-adjacent or overlapping
			if ( newub <= ubs[0] ) { // simple left-extend 
#ifdef LTP_EXTRA_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): left-extending 1st scope\n",dbgstr,newlb,newub);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				lbs[0]=newlb;
				PRINTSCOPE;
				return(LTP_OK);
			} else { // left extend with some spanning
				int rightplace=0;
				bool rightfound=false;
				for (int sc=0;!rightfound && sc!=nscopes;sc++) {
						if (newub <= ubs[sc]) {
									rightplace=sc;
									rightfound=true;
						}
				}
				if (!rightfound) { // shouldn't happen
					LTPERR(LTP_BAD_PARAM);
				} else {
					// reduce the number of scopes appropriately and 
					// fix the ub for that place
#ifdef LTP_EXTRA_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): left-spanning scope\n",dbgstr,newlb,newub);
					ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					lbs[0]=newlb;
					ubs[0]=newub;
					nscopes-=rightplace;
					// copy down others to fix order
					for (int sc=rightplace;sc!=nscopes;sc++) {
						lbs[sc-rightplace+1]=lbs[sc];
						ubs[sc-rightplace+1]=ubs[sc];
					}
				}
			}
		} else { // new 0-th element, rest are copied
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): adding new 1st scope\n",dbgstr,newlb,newub);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			nscopes++;
			alloced+=1;
			int *newlbs=(int*)calloc(nscopes,sizeof(int));
			if (!newlbs) LTPERR(LTP_MEMORY);
			int *newubs=(int*)calloc(nscopes,sizeof(int));
			if (!newubs) LTPERR(LTP_MEMORY);
			newlbs[0]=newlb;
			newubs[0]=newub;
			memcpy(newlbs+1,lbs,(nscopes-1)*sizeof(int));
			memcpy(newubs+1,ubs,(nscopes-1)*sizeof(int));
			free(lbs);
			free(ubs);

			if (scopetype==LTP_TX_SCOPES) {
				theblock->tx_scopes.lbs=newlbs;
				theblock->tx_scopes.ubs=newubs;
				theblock->tx_scopes.nscopes=nscopes;
				theblock->tx_scopes.alloced=alloced;
			} else if (scopetype==LTP_RX_SCOPES) {
				theblock->rx_scopes.lbs=newlbs;
				theblock->rx_scopes.ubs=newubs;
				theblock->rx_scopes.nscopes=nscopes;
				theblock->rx_scopes.alloced=alloced;
			} else if (scopetype==LTP_FWD_SCOPES) {
				theblock->fwd_scopes.lbs=newlbs;
				theblock->fwd_scopes.ubs=newubs;
				theblock->fwd_scopes.nscopes=nscopes;
				theblock->fwd_scopes.alloced=alloced;
			}
			PRINTSCOPE;

			return(LTP_OK);
		}
	}

	// let's check that we don't already have those bytes
	for (int sc=0;sc!=nscopes;sc++) {
		if (lbs[sc]<=newlb && ubs[sc]>=newub) {
			// done!!!
			PRINTSCOPE;
			return(LTP_OK);
		}
	}

	// Ok so this is a gap filler somewhere in the middle
	// possibilities are:-
	// 		it partially fills a gap (need to allocate more memory)
	// 		it extends an existing scope to the right (adjacent or overlapping)
	// 		it extends an existing scope to the left (adjacent or overlapping)
	// 		it completely (though not necessarily exactly) fills a gap, changing two (or more!) 
	// 		    current scopes to one
	// We analyse this and determine whether the new socpe is leftadjacent (or overlapping)
	// and/or rightadjacent (referred to as LA/RA below).
	// Then we know what to do depending on the values of LA & RA:
	// 		LA	RA	Action
	// 		T	T	fills gap - eliminate scopes
	// 		T	F	extends a scope to the right (up)
	// 		F	T	extends a scope to the left (down)
	// 		F	F	create a new scope in between


	
	bool leftadjacent=false;
	bool rightadjacent=false;
	int  leftplace=0;
	int  rightplace=0;
	for (int sc=0;!rightplace && sc!=nscopes;sc++) {
		if (ubs[sc] <= newlb) leftplace=sc; // happens each time until we set rightplace
		if (lbs[sc] >= newub) { // we're in the vicinity
			rightplace=sc;
		} 
	}
	if (newlb <= (ubs[leftplace]+1)) {
		leftadjacent=true;
	} else if (newlb>=lbs[leftplace+1]) { // corner-case check 
		leftplace++;
		leftadjacent=true;
	}
	if (newub >= (lbs[rightplace]-1)) {
		rightadjacent=true;
	} else if (newub<=ubs[rightplace-1]) { // corner-case check 
		rightplace--;
		rightadjacent=true;
	}
#ifdef LTP_EXTRA_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"addnewscope %s (%d,%d): complex case (LA:%s,RA:%s,LP:%d,RP:%d)\n",
					dbgstr,newlb,newub,
					(leftadjacent?"true":"false"),
					(rightadjacent?"true":"false"),
					leftplace,
					rightplace);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	if (leftadjacent && rightadjacent) {
		int onscopes=nscopes;
		nscopes-=(rightplace-leftplace);
		ubs[leftplace]=ubs[rightplace];
		// down-shift remainder
		for (int sc=rightplace;sc!=onscopes-1;sc++) {
			lbs[sc-rightplace+leftplace+1]=lbs[sc+1];
			ubs[sc-rightplace+leftplace+1]=ubs[sc+1];
		}
		if (scopetype==LTP_TX_SCOPES) {
			theblock->tx_scopes.nscopes=nscopes;
		} else if (scopetype==LTP_RX_SCOPES) {
			theblock->rx_scopes.nscopes=nscopes;
		} else if (scopetype==LTP_FWD_SCOPES) {
			theblock->fwd_scopes.nscopes=nscopes;
		}
	} else if (leftadjacent && !rightadjacent) {
		ubs[leftplace]=newub;
	} else if (!leftadjacent && rightadjacent) {
		lbs[rightplace]=newlb;
	} else if (!leftadjacent && !rightadjacent) {
		int onscopes=nscopes;
		nscopes+= 1 - (rightplace-leftplace-1);
		alloced+=1;
		int *newlbs=(int*)calloc(nscopes,sizeof(int));
		if (!newlbs) LTPERR(LTP_MEMORY);
		int *newubs=(int*)calloc(nscopes,sizeof(int));
		if (!newubs) LTPERR(LTP_MEMORY);
		memcpy(newlbs,lbs,(leftplace+1)*sizeof(int));
		memcpy(newubs,ubs,(leftplace+1)*sizeof(int));
		newlbs[leftplace+1]=newlb;
		newubs[leftplace+1]=newub;
		memcpy(&newlbs[leftplace+2],&lbs[rightplace],(onscopes-rightplace)*sizeof(int));
		memcpy(&newubs[leftplace+2],&ubs[rightplace],(onscopes-rightplace)*sizeof(int));
		free(lbs);
		free(ubs);

		if (scopetype==LTP_TX_SCOPES) {
			theblock->tx_scopes.lbs=newlbs;
			theblock->tx_scopes.ubs=newubs;
			theblock->tx_scopes.nscopes=nscopes;
			theblock->tx_scopes.alloced=alloced;
		} else if (scopetype==LTP_RX_SCOPES) {
			theblock->rx_scopes.lbs=newlbs;
			theblock->rx_scopes.ubs=newubs;
			theblock->rx_scopes.nscopes=nscopes;
			theblock->rx_scopes.alloced=alloced;
		} else if (scopetype==LTP_FWD_SCOPES) {
			theblock->fwd_scopes.lbs=newlbs;
			theblock->fwd_scopes.ubs=newubs;
			theblock->fwd_scopes.nscopes=nscopes;
			theblock->fwd_scopes.alloced=alloced;
		}
		PRINTSCOPE;
	}
	return(LTP_OK);
}


/*!
 * @brief yet another replacement for ntohl (thanks cygwin)
 * @param in is a buffer
 * @param out is a ptr to a long
 * @return zero or error
 */
int octets2long(ltpc_octets *in, long *out)
{
	if (!in || !out) LTPERR(LTP_BAD_PARAM);
	if (in->len>4) LTPERR(LTP_BAD_VALUE);
	if (in->len<1) LTPERR(LTP_BAD_VALUE);
	long foo=0;
	switch (in->len) {
		case 4: foo += (in->val[0] << 24)+(in->val[1]<<16)+(in->val[2]<<8)+(in->val[3]); break;
		case 3: foo += (in->val[0] << 16)+(in->val[1]<<8)+(in->val[2]); break;
		case 2: foo += (in->val[0] << 8)+(in->val[1]); break;
		case 1: foo += (in->val[0]); break;
	}
	foo=ntohl(foo);
	*out=foo;
	return(LTP_OK);
}

/*!
 * @brief find the block for a given LTP session number within a socket
 * @param thesock is the socket we're checking within
 * @param theblock is the pointer which will end up at the right place all going well
 * @param sno is the session number field from an LTP message
 * @return true if found, false otherwise
 */
static int get_block_from_sock(ltp_socket_t *thesock, ltp_block **theblock, ltpc_octets *sno)
{
	int rv;
	bool done=false;
	if (!thesock || !theblock || !sno) LTPERR(LTP_BAD_PARAM);
	for (int bid=0;!done && bid!=thesock->nblocks;bid++) {
		long foo=htonl(thesock->blocks[bid].id);
		// check first if it matches a locally produced block
		if (sno->len==sizeof(int) ) {
			long bar; 
			rv=octets2long(sno,&bar);
			bar=htonl(bar);
			if (foo==bar) {
				done=true;
				*theblock=&thesock->blocks[bid];
#ifdef LTP_EXTRA_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Matched socket (%d) and local block id (index: %d, len: %d) for packet.\n",
						thesock->id,bid,(*theblock)->decoded.len);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			}
		}
		// now we'll see if the sno is the peer's choice
		if (!done && sno->len==thesock->blocks[bid].remoteidlen 
			&& 
			!memcmp(sno->val,thesock->blocks[bid].remoteid,sno->len)) {
			done=true;
			*theblock=&thesock->blocks[bid];
#ifdef LTP_EXTRA_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Matched socket (%d) and remote block id (index: %d, len: %d) for packet.\n",
						thesock->id,bid,(*theblock)->decoded.len);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		}
	}
	return(LTP_OK);
}

/*!
 * @brief find the block for a given LTP session number field
 * @param theblock is the pointer which will end up at the right place all going well
 * @param sno is the session number field from an LTP message
 * @param dir is the direction of the blockl, inbound (true) or outbound(false)
 * @return true if found, false otherwise
 */
static bool get_block_from_sno(ltp_block **theblock, ltpc_octets *sno, bool dir)
{
	if (!theblock || !sno) return(false);
	bool found=false;
	ltp_instance snoid;
	int rv=ltpc_oct2int(sno,&snoid);
	if (rv) LTPERR(false);
	
	ltp_socket_t *sock=ltp_state_var.sockets;
	for (int i=0;!found && i!=ltp_state_var.nsockets;i++,sock++) {
		ltp_block *bl=sock->blocks;
		for (int j=0;!found && j!=sock->nblocks;j++,bl++) {
			if (!dir && bl->id==snoid) {
				found=true;
				*theblock=bl;
			} else if (dir && bl->remoteidlen==sno->len && !memcmp(bl->remoteid,sno->val,sno->len)) {
				found=true;
				*theblock=bl;
			}
		}
	}
	return(found);
}


/*!
 * @brief adds a new block to loopback queue.
 * @param msg is the bytes to add
 * @param len is the number of bytes
 * @return zero or error
 */
static int ltp_add_lb(const void *msg, int len)
{
	int nblocks=ltp_state_var.lbq.nblocks;
	ltp_block *blocks=ltp_state_var.lbq.blocks;
	int alloced=ltp_state_var.lbq.alloced;
	nblocks++;
	FOOREALLOC(ltp_block,blocks,nblocks,alloced);
	ltp_state_var.lbq.blocks=blocks;
	ltp_state_var.lbq.nblocks=nblocks;
	ltp_state_var.lbq.alloced=alloced;
	blocks[nblocks-1].decoded.val=(unsigned char*)calloc(len,1);
	if (!blocks[nblocks-1].decoded.val) return(-1);
	blocks[nblocks-1].id=NI;
	blocks[nblocks-1].finishedwith=true;
	blocks[nblocks-1].outboundcookie.len=0;
	blocks[nblocks-1].outboundcookie.val=NULL;
	blocks[nblocks-1].inboundcookie.len=0;
	blocks[nblocks-1].inboundcookie.val=NULL;
	blocks[nblocks-1].cookie_grace_expired=false;
	blocks[nblocks-1].complete=true;
	blocks[nblocks-1].delivered=false;
	blocks[nblocks-1].finishedwithwhendelivered=false;
	blocks[nblocks-1].created=ltp_now();
	blocks[nblocks-1].decoded.len=len;
	memcpy(blocks[nblocks-1].decoded.val,msg,len);
	blocks[nblocks-1].parent=0; // special case - parent socket is "main" (-ish)
	blocks[nblocks-1].parentid=0; // special case - parent is ltp_state_var
	blocks[nblocks-1].remoteidlen=0;
	blocks[nblocks-1].remoteid=NULL;
	blocks[nblocks-1].tx_scopes.nscopes=0;
	blocks[nblocks-1].tx_scopes.alloced=0;
	blocks[nblocks-1].tx_scopes.lbs=NULL;
	blocks[nblocks-1].tx_scopes.ubs=NULL;
	blocks[nblocks-1].rx_scopes.nscopes=0;
	blocks[nblocks-1].rx_scopes.alloced=0;
	blocks[nblocks-1].rx_scopes.lbs=NULL;
	blocks[nblocks-1].rx_scopes.ubs=NULL;
	blocks[nblocks-1].fwd_scopes.nscopes=0;
	blocks[nblocks-1].fwd_scopes.alloced=0;
	blocks[nblocks-1].fwd_scopes.lbs=NULL;
	blocks[nblocks-1].fwd_scopes.ubs=NULL;
	blocks[nblocks-1].redtide=LTP_REDTIDE_INIT;
	blocks[nblocks-1].redcomplete=false;
	blocks[nblocks-1].eobsent=false;
#ifdef MUDSAT
#ifdef LTP_VERBOSE
    snprintf(ltpstr,LTPSTR_LEN,"MUDSAT setiting blcok re-tx ctr to 2\n");
    ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	blocks[nblocks-1].mudsat_re_tx_ctr=2;
#endif
	LTP_STATE_CHANGED
	return(LTP_OK);
}

/*!
 * @brief adds a new block based on an incoming segment
 * @param socksess is the socket to which a new block must be added
 * @param seg is the data segment we rx'd
 * @param block on output is (a ptr to) the new block
 * @return zero or error
 */
static int ltp_add_block_from_seg(ltp_socket_t *socksess, ltpc_seg *seg, ltp_block **block)
{
	ltpc_ds *ds=&seg->val.ds;
	int nblocks=socksess->nblocks;
	ltp_block *blocks=socksess->blocks;
	int alloced=socksess->alloced;
	nblocks++;
	FOOREALLOC(ltp_block,blocks,nblocks,alloced);
	socksess->blocks=blocks;
	socksess->nblocks=nblocks;
	socksess->alloced=alloced;

	int len=ds->bytes.len;
	int offset=ds->offset;


	int totallen=len+offset;

	if (ds->csi.len==2) {
		blocks[nblocks-1].csi=ds->csi.val[1]*256+ds->csi.val[0];
	} else {
		snprintf(ltpstr,LTPSTR_LEN, "Inbound segment has odd csi (making block), using port\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		blocks[nblocks-1].csi=htons(socksess->obpeer.sock.sin_port);
	}
	blocks[nblocks-1].decoded.val=(unsigned char*)calloc(1,totallen);
	if (!blocks[nblocks-1].decoded.val) return(-1);
	blocks[nblocks-1].id=NI;
	blocks[nblocks-1].finishedwith=false;
	blocks[nblocks-1].inboundcookie.len=0;
	blocks[nblocks-1].inboundcookie.val=NULL;
	blocks[nblocks-1].outboundcookie.len=0;
	blocks[nblocks-1].outboundcookie.val=NULL;
	blocks[nblocks-1].cookie_grace_expired=false;
	blocks[nblocks-1].eob_seen=ds->eob;
	blocks[nblocks-1].eorp_seen=ds->eorp;
	blocks[nblocks-1].eorp_sent=false;
	blocks[nblocks-1].created=ltp_now();
	blocks[nblocks-1].decoded.len=len;
	memcpy(blocks[nblocks-1].decoded.val+offset,ds->bytes.val,len);
	blocks[nblocks-1].parent=socksess;
	blocks[nblocks-1].parentid=socksess->id; // special case - parent is ltp_state_var
	blocks[nblocks-1].remoteidlen=seg->sno.len;
	blocks[nblocks-1].remoteid=(unsigned char*)calloc(1,seg->sno.len);
	if (!blocks[nblocks-1].remoteid) LTPERR(LTP_MEMORY);
	memcpy(blocks[nblocks-1].remoteid,seg->sno.val,seg->sno.len);
	blocks[nblocks-1].tx_scopes.nscopes=0;
	blocks[nblocks-1].tx_scopes.alloced=0;
	blocks[nblocks-1].tx_scopes.lbs=NULL;
	blocks[nblocks-1].tx_scopes.ubs=NULL;
	blocks[nblocks-1].rx_scopes.nscopes=0;
	blocks[nblocks-1].rx_scopes.alloced=0;
	blocks[nblocks-1].rx_scopes.lbs=NULL;
	blocks[nblocks-1].rx_scopes.ubs=NULL;
	blocks[nblocks-1].fwd_scopes.nscopes=0;
	blocks[nblocks-1].fwd_scopes.alloced=0;
	blocks[nblocks-1].fwd_scopes.lbs=NULL;
	blocks[nblocks-1].fwd_scopes.ubs=NULL;
	int rv=addnewscope(&blocks[nblocks-1],offset,offset+len,LTP_RX_SCOPES);
	if (rv) LTPERR(rv);
	blocks[nblocks-1].complete=(ds->eob && complete_scope(&blocks[nblocks-1]));
	blocks[nblocks-1].delivered=false;
	blocks[nblocks-1].finishedwithwhendelivered=false;
	blocks[nblocks-1].current_rsn=0;
	if (!ds->red) {
		blocks[nblocks-1].redtide=LTP_REDTIDE_INIT;
	} else {
		blocks[nblocks-1].redtide=totallen;
	}
	blocks[nblocks-1].redcomplete=false;
	blocks[nblocks-1].eobsent=false;
#ifdef MUDSAT
#ifdef LTP_VERBOSE
    snprintf(ltpstr,LTPSTR_LEN,"MUDSAT setiting blcok re-tx ctr to 2\n");
    ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	blocks[nblocks-1].mudsat_re_tx_ctr=2;
#endif
	// maybe check extensions to see what's what
	LTP_STATE_CHANGED;
	*block=&blocks[nblocks-1];
	return(LTP_OK);
}

/*!
 * @fn int ltp_socket(int domain, int type, int protocol)
 * @brief create (for originator) a new LTP socket (analagous to the socket() function)
 * @param domain has to be AF_LTP
 * @param type has to be SOCK_LTP_SESSION
 * @param protocol has to be LTP_BUNDLING (not here though) or LTP_FARRELL (all that works here!)
 * @return is the socket number for use in later calls (e.g. to ltp_sendto())
 */
int ltp_socket(int domain, int type, int protocol)
{
	LTPINIT;
	GRABMUTEX;
	/* just create the socket and put it into the state var */
	int sockfd=NI; ///< to be returned, its the next instance
	int id=NI;

	int nsockets=ltp_state_var.nsockets;
	ltp_socket_t	*sockets=ltp_state_var.sockets;
	int alloced=ltp_state_var.alloced;

	nsockets++;
	FOOREALLOC(ltp_socket_t,sockets,nsockets,alloced);
	ltp_state_var.nsockets=nsockets;
	ltp_state_var.sockets=sockets; // realloc *can* move things!
	ltp_state_var.alloced=alloced;

	sockets[nsockets-1].id=id;
	sockets[nsockets-1].inbound=false;
	sockets[nsockets-1].outbound=false;
	sockets[nsockets-1].finishedwith=false;
	sockets[nsockets-1].sockfd=sockfd;
	sockets[nsockets-1].nblocks=0;
	sockets[nsockets-1].blocks=NULL;
	sockets[nsockets-1].ibpeer=ltpnull;
	sockets[nsockets-1].ibpeerlen=0;
	sockets[nsockets-1].obpeer=ltpnull;
	sockets[nsockets-1].obpeerlen=0;
	sockets[nsockets-1].opts=LTP_DEFAULT_SOCKOPT;
	sockets[nsockets-1].ltptrelay=false;
	sockets[nsockets-1].close_called=false;

#ifdef MUDSAT
	// SF: MUDSAT - use default crypto
	if (using_def_crypto==1) {
		ltp_auth_so ov;
		ov.ciphersuite=def_ciphersuite;
		ov.keyidlen=strlen(def_keyid);
		memcpy(ov.keyid,def_keyid,ov.keyidlen);
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"MUDSAT Setting DEFAULT ciphersuite to %d, keyid to |%s| in ltp_setsockopt\n",
			ov.ciphersuite,octets2str(ov.keyid,ov.keyidlen));
		ltp_log(LTP_LOG_ADV,ltpstr);
#endif
		sockets[nsockets-1].opts.auth=ov;
	}
#endif

	int rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)&sockets[nsockets-1],false);
	if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"Cannot initialise inactivity timer for socket exiting\n");
		ltp_log(LTP_LOG_ERR,ltpstr);
		DROPMUTEX;
		return(-1);
	}

#ifdef LTP_EXTRA_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Made my %d-th new socket: %d\n",nsockets,sockets[nsockets-1].id);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif

	LTP_STATE_CHANGED;
	DROPMUTEX;
	return(sockfd);
}

/// A macro to declare "found" and "sess" and check if "sock" is a socket
/// somewhere in our state. If it is sess is the pointer you want and
/// found sez' if its there.
#define GETSOCKSESS \
	ltp_socket_t *socksess; \
	bool found=false; \
	for (int ii=0;!found && ii!=ltp_state_var.nsockets;ii++) { \
		if ( (sock==ltp_state_var.sockets[ii].sockfd  || sock == ltp_state_var.sockets[ii].id) && \
			ltp_state_var.sockets[ii].finishedwith==false) { \
			found=true; \
			socksess=&ltp_state_var.sockets[ii]; \
		} \
	} 


/*!
 * @brief compare two LTP addresses
 * @param a1 is a comparee
 * @param a2 is a comparee
 * @param a1len is the length of significant bytes of a1
 * @return zero if they're the same, non-zero otherwise
 *
 * This compares a1len bytes of a1 with a2 (allowing NULL=NULL to return zero). The funny
 * length handling is due to idiosyncracies of the underlying sockaddr structure.
 * 
 * We allow ".255" to be wildcard inside IP addresses, so 127.0.255.255 matches 127.0.0.1
 * If there're any 255's then we don't check for ports.
 *
 * We allow zero as a wildcard port
 * 
 */
int ltpaddr_cmp(ltpaddr *a1, ltpaddr *a2,int a1len )
{
	if (a1==NULL && a2==NULL) return(0);
	if (isany(a1,a1len)) return(0); // a1's a wildcard
	if (isany(a2,a1len)) return(0); // someone's a wildcard
	if (!a1 || !a2) return(1);
	if (a1->octets.magic==LTP_ADDR_OCT_MAGIC && a2->octets.magic!=LTP_ADDR_OCT_MAGIC ) return(2);
	if (a1->octets.magic!=LTP_ADDR_OCT_MAGIC && a2->octets.magic==LTP_ADDR_OCT_MAGIC ) return(3);

	// both use same structure anyway
	if (a1->octets.magic==LTP_ADDR_OCT_MAGIC) {
		if (a1->octets.len!=a2->octets.len) return(4);
		if (memcmp(a1->octets.val,a2->octets.val,a1->octets.len)) return(5);
	} else {
		if (a1->sock.sin_family!=a2->sock.sin_family) return(6);
		// allow .255 to be a wildcard inside 
		bool saw255=false;
		for (int i=0;i!=sizeof(a1->sock.sin_addr.s_addr);i++) {
			int a1ch=(a1->sock.sin_addr.s_addr>>(8*i))&0xff;
			int a2ch=(a2->sock.sin_addr.s_addr>>(8*i))&0xff;
			saw255=(a1ch==255 || a2ch==255);
			if (!saw255 && a1ch!=a2ch) {
				return(9);
			}
		}
		// if no 255's check ports if port!=0
		if (!saw255 && a1->sock.sin_port && a2->sock.sin_port &&
				a1->sock.sin_port!=a2->sock.sin_port) return(7);
	}
	return(0);
}

/*!
 * @brief string-ify an ltpaddr
 * @param addr is the address to string-fy
 * @return is the string-form (in a static buffer)
 */
char *ltpaddr2str(ltpaddr *addr)
{
	static char strbuf[100]={0x00};
	if (!addr) snprintf(strbuf,100,"<<NULL>> input");
	if (addr->octets.magic==LTP_ADDR_OCT_MAGIC) {
		snprintf(strbuf,100,"%s\n",octets2str(addr->octets.val,addr->octets.len));
	} else {
		snprintf(strbuf,100,"%s:%d",inet_ntoa(addr->sock.sin_addr),addr->sock.sin_port);
	}
	return(strbuf);
}

/*!
 * @brief map ltpaddr to an octet string 
 * @param ad is the ltp address
 * @param sp is the octet string
 * @return zero or error
 *
 * The mapping is to use an IP address in network byte order with an optional
 * port number, if the default isn't in use.
 */
int ltpaddr2octets(ltpaddr *ad, ltpc_octets *sp)
{
	if (!ad || !sp) LTPERR(LTP_BAD_PARAM);
	int pvallen;
	if (ad->octets.magic==LTP_ADDR_OCT_MAGIC) {
		sp->len=ad->octets.len;
		sp->val=ltpc_alloc(sp->len);
		if (!sp->val) LTPERR(LTP_MEMORY);
		memcpy(sp->val,ad->octets.val,sp->len);
		return (LTP_OK);
	}
	if (ad->sock.sin_port==UDP_DEF_PORT) {
		pvallen=4;
	} else {
		pvallen=6;
	}
	long foo=htonl(ad->sock.sin_addr.s_addr);
	sp->len=pvallen;
	sp->val=(unsigned char*)calloc(1,pvallen);
	if (!sp->val) LTPERR(LTP_MEMORY);
	sp->val[0]=(foo>>24) & 0xff;
	sp->val[1]=(foo>>16) & 0xff;
	sp->val[2]=(foo>>8) & 0xff;
	sp->val[3]=foo & 0xff;
	if (ad->sock.sin_port!=UDP_DEF_PORT) {
		// short bar=htons(ad->sock.sin_port);
		short bar=ad->sock.sin_port;
		sp->val[4]=(bar>>8)&0xff;
		sp->val[5]=bar & 0xff;
	}
	return(LTP_OK);
}

/*!
 * @brief map an octet string to an LTP address
 * @param sp is the octet string
 * @param ad is the LTP address
 * @return zero or error
 *
 * The inverse of ltpaddr2octets
 */
int octets2ltpaddr(ltpc_octets *sp, ltpaddr *ad)
{
	int rv;
	if (!ad || !sp || !sp->val) LTPERR(LTP_BAD_PARAM);
	if (sp->len==4) {
		ad->sock.sin_family=AF_LTP;
		ad->sock.sin_port=UDP_DEF_PORT;
		rv=octets2long(sp,(long*)&ad->sock.sin_addr.s_addr);
		return(LTP_OK);
	}
	if (sp->len==6) {
		ad->sock.sin_family=AF_LTP;
		sp->len=4; // hack hack
		rv=octets2long(sp,(long*)&ad->sock.sin_addr.s_addr);
		ad->sock.sin_port=(sp->val[4]<<8)+sp->val[5];
		sp->len=6; // un-hack un-hack
		return(LTP_OK);
	}
	ad->octets.magic=LTP_ADDR_OCT_MAGIC;
	ad->octets.len=sp->len;
	ad->octets.val=ltpc_alloc(sp->len);
	if (!ad->octets.val) LTPERR(LTP_MEMORY);
	memcpy(ad->octets.val,sp->val,sp->len);
	return (LTP_OK);
}

/*!
 * @fn int ltp_bind(int sock, ltpaddr *ad, ltpaddr_len addrlen); 
 * @brief bind socket to lower layer OS 
 * @param sock is the socket
 * @param ad is the my address
 * @param addrlen is the length of my address
 * @return is zero for success and -1 for error
 */
int ltp_bind(int sock, ltpaddr *ad, ltpaddr_len addrlen) 
{
	LTPINIT;
	GRABMUTEX;
	// all we've to do is associate the addr with the sockfd
	GETSOCKSESS;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"ltp_bind: asked to bind %s to socket %d\n",
				ltpaddr2str(ad),sock);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	if (!found) { DROPMUTEX; return(-1); }
	if (!ad) { DROPMUTEX; return(-2);}
	if (ad==&ltpany) {
		socksess->inbound=true;
	} else {
		memcpy(&socksess->obpeer,ad,addrlen);
		socksess->obpeerlen=addrlen;
		// socksess->outbound=true; maybe not!
	}
	int rv;
	// non-zero or non-default port? better kick l2
	if (ad->octets.magic!=LTP_ADDR_OCT_MAGIC && ad->sock.sin_port && ad->sock.sin_port!=UDP_DEF_PORT) {
		rv=ltp_l2_init(defcbs_p,ad->sock.sin_port);
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"Cannot re-initialise l2 for different port (%d) - rv=%d\n",
						ad->sock.sin_port,rv);
			ltp_log(LTP_LOG_ERR,ltpstr);
			DROPMUTEX;
			return(-3);
		}
	}
	rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)socksess,true);
	if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"Cannot re-initialise inactivity timer for socket - exiting\n");
		ltp_log(LTP_LOG_ERR,ltpstr);
		DROPMUTEX;
		return(-4);
	}
	LTP_STATE_CHANGED;
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @brief make a ds from a scope range
 * @param to is where to send this segment
 * @param theblock is the block in question
 * @param seglen is the length of the data to send
 * @param offset is the offset of the data
 * @return zero or error
 */
static int ds_from_block(ltpaddr *to, ltp_block *theblock,int seglen, int offset, bool checkpoint, ltpc_octets *rsn)
{
	int rv;
	if (!to || !theblock) LTPERR(LTP_BAD_PARAM);
	NEWSEGMENT
	segments[nsegments].parent=theblock;
	segments[nsegments].parentid=theblock->id;
	rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
	if (rv) LTPERR(rv);
	segments[nsegments].peer=*to;
	segments[nsegments].peerlen=sizeof(*to);
	segments[nsegments].r_entry=LTP_NOCACHE; // no need to store this guy
	segments[nsegments].r_id=0;
	segments[nsegments].t_entry=LTP_NOTIMER; // no need to store this guy
	segments[nsegments].t_id=0;
	ltpc_seg *seg=&segments[nsegments].seg;
	seg->type=LTPC_SEG_DS;
	// the "peer" in the segment is me!! it'll be used on arrival, not transmit!
	INITMYADDR;
	// Careful this - might not work!!!
	if (ion_mode) {
		short engid=htons(my_addr.sock.sin_port);
		seg->peer.len=2;
		seg->peer.val=(unsigned char*)calloc(1,seg->peer.len);
		if (!seg->peer.val) LTPERR(LTP_MEMORY);
		memcpy(seg->peer.val,&engid,seg->peer.len);
	} else {
		// int pvallen=sizeof(my_addr.sock.sin_addr.s_addr);
		rv=ltpaddr2octets(&my_addr,&seg->peer);
		if (rv) LTPERR(rv);
	}
	if (ion_mode) {
		short csid=1;
		seg->val.ds.csi.len=1;
		seg->val.ds.csi.val=(unsigned char*)calloc(1,sizeof(csid));
		if (!seg->val.ds.csi.val) LTPERR(LTP_MEMORY);
		seg->val.ds.csi.val[0]=0x01;
	} else {
		short csid=theblock->csi;
		seg->val.ds.csi.len=sizeof(csid);
		seg->val.ds.csi.val=(unsigned char*)calloc(1,sizeof(csid));
		if (!seg->val.ds.csi.val) LTPERR(LTP_MEMORY);
		memcpy(seg->val.ds.csi.val,&csid,sizeof(csid));
	}
	int sno=theblock->id;
	seg->sno.val=(unsigned char*)calloc(1,sizeof(sno));
	if (!seg->sno.val) LTPERR(LTP_MEMORY);
	memcpy(seg->sno.val,&sno,sizeof(sno));
	seg->sno.len=sizeof(sno);
	seg->val.ds.offset=offset; 
	seg->val.ds.bytes.val=NULL; // fill in later on
	seg->val.ds.bytes.len=seglen; 
	seg->val.ds.eob=false; 
	// well maybe it is eob, let's see...
	if (theblock->eob_seen && ((offset+seglen)==theblock->decoded.len)) seg->val.ds.eob=true;
	seg->val.ds.checkpoint=false; 
	seg->val.ds.retransmit=false; 
	memset(&seg->val.ds.csn,0,sizeof(ltp_octets)); // initially no checkpoint serial number
	memset(&seg->val.ds.rsn,0,sizeof(ltp_octets)); // initially no checkpoint serial number
	// are these bytes red or not?
	int redlen=theblock->parent->opts.redlen;
	// if we're forwarding the block then LTP_ALLRED isn't appropeiate - follow the leader
	if (redlen!=LTP_ALLRED && theblock->redtide > redlen) {
		// the socket is set for something (e.g ALLRED/ALLGREEN, but the inbound
		// block has a different setting (only happens in the case of LTP-T fowarding I think)
		redlen=theblock->redtide;
	}
	if (redlen!=LTP_ALLGREEN) {
		if (redlen==LTP_ALLRED || (seg->val.ds.offset <= redlen)) {
			seg->val.ds.red=true;
			if (seg->val.ds.eob) {
				seg->val.ds.checkpoint=true;
				seg->val.ds.eorp=true;
			}
		}
		// is this the last red segment?
		// bug - only do this once!
		// still buggy redcomplete isn't necessarily set in time (TODO: think!)
		// probably add an eorp_sent to the block str
		if (redlen!=LTP_ALLRED && !theblock->eorp_sent && theblock->eorp_seen && seg->val.ds.offset+seg->val.ds.bytes.len > redlen) {
			seg->val.ds.eorp=true;
			seg->val.ds.checkpoint=true;
			theblock->eorp_sent=true;
		}
	}
	if (seg->val.ds.checkpoint || checkpoint) { // last red DS has to be CP
		seg->val.ds.checkpoint=true;
		ltpc_ds *ds=&seg->val.ds;
		ds->retransmit=true;
		ds->rsn.len=rsn->len;
		ds->rsn.val=ltpc_alloc(rsn->len);
		if (!ds->rsn.val) LTPERR(LTP_MEMORY);
		memcpy(ds->rsn.val,rsn->val,ds->rsn.len);
		short csni=htons(NI&0x7fff)&0x7fff;
		ds->csn.len=2;
		ds->csn.val=(unsigned char*)calloc(1,ds->csn.len);
		if (!ds->csn.val) LTPERR(LTP_MEMORY);
		ds->csn.val[0]=csni>>8;
		ds->csn.val[1]=csni&0xff;
		if (ds->csn.val[0]==0x00) {
			ds->csn.len=1;
			ds->csn.val[0]=ds->csn.val[1];
		}
		rv=init_timer(LTP_TIMER_CP,&segments[nsegments]);
		if (rv) LTPERR(rv);
	}
	if (theblock->t_vals.flags) {
		rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
		if (rv) LTPERR(rv);
		add_ltpt_exts(seg,theblock,theblock->parent);
	}
	return(LTP_OK);
}

/*!
 * @fn int ltp_sendto_aux(int sock, const void *msg, size_t len, int flags, ltpaddr *to, ltpaddr_len tolen)
 * @brief sendto equivalent for LTP
 * @param sock is a destination LTP "socket" 
 * @param msg are the bytes to send
 * @param len is the number of bytes
 * @param flags contains flags for the LTP socket
 * @param to is the destination address
 * @param tolen is the length of the destination address
 * @return is the number of bytes sent or -1 for an error
 *
 * Same as external ltp_sendto other than not copying the buffer
 */
static int ltp_sendto_aux(int sock, const void *msg, size_t len, int flags, ltpaddr *to, ltpaddr_len tolen)
{
	// let's just do the minimum so's we don't loose those bytes
	GRABMUTEX;
	GETSOCKSESS;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Entering ltp_sendto_aux with %zu bytes to send to %s on socket %d\n",
					len,ltpaddr2str(to),sock);
	ltp_log(LTP_LOG_MAX,ltpstr);
	ltp_log_flush();
#endif
	if (!found) { 
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Socket (%d) not found\n",sock);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		DROPMUTEX; 
		return(-1); 
	}
	if (!socksess->outbound) { 
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Socket (%d,id:%d) being made ready for outbound\n",sock,socksess->id);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		socksess->outbound=true;
	}
	// bogus optimisation deleted
	INITMYADDR;
#ifdef LTP_VERBOSE
	if (socksess->ibpeerlen) {
		// sheesh - maybe force it to be me/now?
		char *s1=strdup(ltpaddr2str(&socksess->ibpeer));
		char *s2=strdup(ltpaddr2str(&my_addr));
		snprintf(ltpstr,LTPSTR_LEN,"Overwriting ibpeer for socket (%d) %s -> %s\n",
					socksess->id,s1,s2);
		ltp_log(LTP_LOG_MAX,ltpstr);
		free(s1); free(s2);
	}
#endif
	socksess->ibpeer=my_addr;
	socksess->ibpeerlen=sizeof(my_addr);

	int rv;

	// calling sendto means this is an outbound case!
	if (to==&ltplb) { // loopback case
#ifdef LTP_VERBOSE
		ltp_log(LTP_LOG_MAX,(char*)"Using LTP loopback\n");
#endif
		rv=ltp_add_lb(msg,len);
		DROPMUTEX;
		if (rv) LTPERR(-1);
		return(len);
	} else {
		// record to as peer for later possible C segments etc.
		memcpy(&socksess->obpeer,to,tolen);
		socksess->obpeerlen=tolen;
	}

	ltp_block *blocks=socksess->blocks;
	int	nblocks=socksess->nblocks;
	int alloced=socksess->alloced;

	nblocks++;
	FOOREALLOC(ltp_block,blocks,nblocks,alloced);
	socksess->blocks=blocks;
	socksess->alloced=alloced;

	socksess->nblocks=nblocks; // waited 'till after calloc, just in case!

	blocks[nblocks-1].id=NI;
	blocks[nblocks-1].finishedwith=false;
	blocks[nblocks-1].outboundcookie.len=0;
	blocks[nblocks-1].outboundcookie.val=NULL;
	blocks[nblocks-1].inboundcookie.len=0;
	blocks[nblocks-1].inboundcookie.val=NULL;
	blocks[nblocks-1].cookie_grace_expired=false;
	blocks[nblocks-1].cookie_grace_expired=false;
	blocks[nblocks-1].complete=false;
	blocks[nblocks-1].delivered=false;
	blocks[nblocks-1].finishedwithwhendelivered=false;
	blocks[nblocks-1].eob_seen=false;
	blocks[nblocks-1].eorp_seen=false;
	blocks[nblocks-1].eorp_sent=false;
	blocks[nblocks-1].created=ltp_now();
	blocks[nblocks-1].decoded.len=len;
	blocks[nblocks-1].decoded.val=(unsigned char*)msg;
	blocks[nblocks-1].parent=socksess; 
	blocks[nblocks-1].parentid=socksess->id; 
	blocks[nblocks-1].remoteidlen=0;
	blocks[nblocks-1].remoteid=NULL;
	blocks[nblocks-1].tx_scopes.nscopes=0;
	blocks[nblocks-1].tx_scopes.alloced=0;
	blocks[nblocks-1].tx_scopes.lbs=NULL;
	blocks[nblocks-1].tx_scopes.ubs=NULL;
	blocks[nblocks-1].rx_scopes.nscopes=0;
	blocks[nblocks-1].rx_scopes.alloced=0;
	blocks[nblocks-1].rx_scopes.lbs=NULL;
	blocks[nblocks-1].rx_scopes.ubs=NULL;
	blocks[nblocks-1].fwd_scopes.nscopes=0;
	blocks[nblocks-1].fwd_scopes.alloced=0;
	blocks[nblocks-1].fwd_scopes.lbs=NULL;
	blocks[nblocks-1].fwd_scopes.ubs=NULL;
#ifdef MUDSAT
#ifdef LTP_VERBOSE
    snprintf(ltpstr,LTPSTR_LEN,"MUDSAT setiting blcok re-tx ctr to 2\n");
    ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	blocks[nblocks-1].mudsat_re_tx_ctr=2;
#endif
	if (ion_mode) {
		blocks[nblocks-1].csi=1;
	} else {
		blocks[nblocks-1].csi=htons(socksess->obpeer.sock.sin_port);
	}
	blocks[nblocks-1].current_rsn=0;
	blocks[nblocks-1].t_vals.flags=0;

	// for later use when we're in a blocking call state
	ltp_block *theblock=&blocks[nblocks-1];

	int lc_mtu=latestmtu(&blocks[nblocks-1],socksess);

	if (lc_mtu<=0) LTPERR(LTP_BAD_VALUE);

	// now make up segments for transmission
	int newsegs=len/lc_mtu+(len%lc_mtu?1:0);

	int nsegments=ltp_state_var.txq.nsegments;
	alloced=ltp_state_var.txq.alloced;
	ltp_segment *segments=ltp_state_var.txq.segments;

	FOOREALLOC(ltp_segment,segments,nsegments+newsegs,alloced);
	ltp_state_var.txq.segments=segments; // realloc might have moved him!
	ltp_state_var.txq.alloced=alloced;

	// figure out how many segments are red/green
	int redlen=socksess->opts.redlen;
	int redsegs;
	if (redlen==LTP_ALLRED) {
		redsegs=newsegs;
	} else if (redlen==LTP_ALLGREEN) {
		redsegs=0;
	} else {
		redsegs=redlen/lc_mtu+((len%lc_mtu)?1:0);
	}

	if (socksess->opts.t_cfg.on) {
		int nlisteners=LTP_LISTENERS_PER_GO;
		ltpaddr	listeners[LTP_LISTENERS_PER_GO];
		bool matched=false;
		ltpaddr nexthop;
		while (!matched) {
			nlisteners=LTP_LISTENERS_PER_GO;
			ltp_time_t now=ltp_now();
			int ltp_cue_rv=ltp_cue_whos_listening(now,&nlisteners,listeners);
			if (ltp_cue_rv) LTPERR(ltp_cue_rv);
			if (nlisteners) {
				rv=ltpt_gateway_listeners(to,&nexthop,nlisteners,listeners,&matched);
				if (rv) { 
					snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltpt_gateway_listeners (%d)\n",rv);
					ltp_log(LTP_LOG_MAX,ltpstr);
					ltp_log_flush();
					LTPERR(rv);
				}
			}
			if (!matched) {
				DROPMUTEX;
				ltp_waitabit();
				GRABMUTEX;
			} else {
				int nexthoplen=sizeof(nexthop);
				for (int k=0;k!=nlisteners;k++) { // ignore that unless the nexthop is now listening (we'll come back to it I hope)
					if (!ltpaddr_cmp(&nexthop,&listeners[k],nexthoplen)) { // ok we may have something to send
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Forwarding block (size:%d,id:%d) to %s\n",
							blocks[nblocks-1].decoded.len,blocks[nblocks-1].id,ltpaddr2str(&nexthop));
						ltp_log(LTP_LOG_MAX,ltpstr);
						ltp_log_flush();
#endif
						memcpy(&socksess->obpeer,&nexthop,sizeof(nexthop));
						socksess->obpeerlen=sizeof(nexthop);
						// block->t_vals.gw=nexthop; - copy below should be better
						blocks[nblocks-1].csi=htons(socksess->obpeer.sock.sin_port);
						// don't forget the intended "to"
						blocks[nblocks-1].t_vals.dest=*to;
						blocks[nblocks-1].t_vals.flags|=LTPT_SET_DEST;
					} 
				}
			} 
		}

#ifdef NOGWLISTENERS
	// old, silly method
		rv=ltpt_gateway(to,&gw);
		if (rv) { DROPMUTEX; LTPERR(-1); }
		char *s1=strdup(ltpaddr2str(to));
		char *s2=strdup(ltpaddr2str(&gw));
		if (ltpaddr_cmp(&gw,to,tolen)) { // they differ
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Changing target peer from %s to %s\n", s1,s2);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			memcpy(&socksess->obpeer,&gw,sizeof(gw));
			socksess->obpeerlen=sizeof(gw);
			blocks[nblocks-1].csi=htons(socksess->obpeer.sock.sin_port);
			// don't forget the intended "to"
			blocks[nblocks-1].t_vals.dest=*to;
			blocks[nblocks-1].t_vals.flags|=LTPT_SET_DEST;
		}
		free(s1);free(s2);
#endif
	}

	bool cpset=false;
	for (int i=0;i!=newsegs;i++) {
		segments[nsegments+i].id=NI;
		segments[nsegments+i].state=LTP_PS_WAITING_SEND;
		segments[nsegments+i].parent=&blocks[nblocks-1]; // if socksess->blocks moves due to a later realloc() we're screwed!! (except for parentid)
		segments[nsegments+i].parentid=blocks[nblocks-1].id; // this is a saver!!
		memcpy(&segments[nsegments+i].peer,&socksess->obpeer,socksess->obpeerlen);
		segments[nsegments+i].peerlen=socksess->obpeerlen;
		segments[nsegments+i].r_entry=LTP_NOCACHE; // no need to store this guy
		segments[nsegments+i].r_id=0; // no need to store this guy
		segments[nsegments+i].t_entry=LTP_NOTIMER; // no need to store this guy
		segments[nsegments+i].t_id=0; // no need to store this guy

		ltpc_seg	*seg=&segments[nsegments+i].seg;

		seg->type=LTPC_SEG_DS; // data segment
		seg->nexts=0;
		seg->exts=NULL;

		// do LTP-T stuff here?
		if (socksess->opts.t_cfg.on) {
			add_ltpt_exts(seg,&blocks[nblocks-1],socksess);
		}

		// we split peer in ds.peer and ds.sno/ds.csi at this point!

		// the "peer" in the segment is me!! it'll be used on arrival, not transmit!
		INITMYADDR;
		if (ion_mode) {
			short engid=htons(my_addr.sock.sin_port);
			seg->peer.len=2;
			seg->peer.val=(unsigned char*)calloc(1,seg->peer.len);
			if (!seg->peer.val) LTPERR(LTP_MEMORY);
			memcpy(seg->peer.val,&engid,seg->peer.len);
		} else {
			rv=ltpaddr2octets(&my_addr,&seg->peer);
			if (rv) LTPERR(rv);
		}

		if (ion_mode) {
			short csid=1;
			seg->val.ds.csi.len=1;
			seg->val.ds.csi.val=(unsigned char*)calloc(1,sizeof(csid));
			if (!seg->val.ds.csi.val) LTPERR(LTP_MEMORY);
			seg->val.ds.csi.val[0]=0x01;
		} else {
			short csid=blocks[nblocks-1].csi;
			seg->val.ds.csi.len=sizeof(csid);
			seg->val.ds.csi.val=(unsigned char*)calloc(1,sizeof(csid));
			if (!seg->val.ds.csi.val) LTPERR(LTP_MEMORY);
			memcpy(seg->val.ds.csi.val,&csid,sizeof(csid));
		}
		int sno=blocks[nblocks-1].id;
		seg->sno.val=(unsigned char*)calloc(1,sizeof(sno));
		if (!seg->sno.val) LTPERR(LTP_MEMORY);
		memcpy(seg->sno.val,&sno,sizeof(sno));
		seg->sno.len=sizeof(sno);
		seg->val.ds.offset=i*lc_mtu;
		seg->val.ds.bytes.val=NULL; // fill in later on
		seg->val.ds.bytes.len= ( 
								(i==(newsegs-1) && len%lc_mtu)  // if we're the last segment and there's a remainder
								? 
								len-(i*lc_mtu)  // then the length is that remainder
								: 
								lc_mtu); // otherwise its just the nice even segment size
		seg->val.ds.eob=false; 
		seg->val.ds.checkpoint=false; 
		seg->val.ds.retransmit=false; 
		memset(&seg->val.ds.csn,0,sizeof(ltp_octets)); // initially no checkpoint serial number
		memset(&seg->val.ds.rsn,0,sizeof(ltp_octets)); // initially no report serial number

		// setup red & maybe eorp
		if (redsegs && i<redsegs) {
			seg->val.ds.red=true;
		}
		if (redsegs && (i==(redsegs-1))) { 
			seg->val.ds.eorp=true;
			seg->val.ds.checkpoint=true;
			// set a 15-bit CSN for the last red segment in the block 
			short csni=htons(NI&0x7fff)&0x7ff;
			ltpc_ds	*ds=&seg->val.ds;
			ds->csn.len=2;
			ds->csn.val=(unsigned char*)calloc(1,ds->csn.len);
			if (!ds->csn.val) { DROPMUTEX; LTPERR(LTP_MEMORY); }
			ds->csn.val[0]=csni>>8;
			ds->csn.val[1]=csni&0xff;
			if (ds->csn.val[0]==0x00) {
				// shorten
				ds->csn.len=1;
				ds->csn.val[0]=ds->csn.val[1];
			}
			rv=init_timer(LTP_TIMER_CP,&segments[nsegments+i]);
			if (rv) { DROPMUTEX; LTPERR(rv); }
			cpset=true;
		}

	}
	// set a 15-bit CSN for the last segment in the block (it'll often be needed and ignored if not)
	// bit buggy if csn is one byte, so a test to force that so I can test it...
	// looks ok, back to random 15 bit values
#undef SHORTCSNTEST
#ifdef SHORTCSNTEST
	short csni=0x1f;
#else
	short csni=htons(NI&0x7fff) & 0x7ff;
#endif
	ltpc_ds	*ds=&segments[nsegments+newsegs-1].seg.val.ds;
	ds->csn.len=2;
	ds->csn.val=(unsigned char*)calloc(1,ds->csn.len);
	if (!ds->csn.val) { DROPMUTEX; LTPERR(LTP_MEMORY); }
	ds->csn.val[0]=csni>>8;
	ds->csn.val[1]=csni&0xff;
	if (ds->csn.val[0]==0x00) {
		// shorten
		ds->csn.len=1;
		ds->csn.val[0]=ds->csn.val[1];
	}
	ds->eob=true;
	// setup red & maybe eorp
	if (redsegs && newsegs<=redsegs) {
		ds->red=true;
		ds->eorp=true;
		ds->checkpoint=true;
	}
	segments[nsegments+newsegs-1].r_entry=LTP_CACHE;

	if (ds->red && !cpset) {
		rv=init_timer(LTP_TIMER_CP,&segments[nsegments+newsegs-1]);
		if (rv) { DROPMUTEX; LTPERR(rv); }
	}

	ltp_state_var.txq.nsegments+=newsegs;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Made %d new data segments for later transmission via layer 2\n",newsegs);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif

	rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)socksess,true);
	if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"Cannot re-initialise inactivity timer for socket - exiting\n");
		ltp_log(LTP_LOG_ERR,ltpstr);
		DROPMUTEX;
		return(-1);
	}
	bool nonblocking=socksess->opts.nonblocking;
	LTP_STATE_CHANGED;
	DROPMUTEX;
	if (nonblocking) return(len);

	// enter wait loop until block is complete
	while (true) {
		// should readlly ltp_fixparent() here
		// wait a bit
		ltp_waitabit();
		// check again
		GRABMUTEX;
		GETSOCKSESS;
		if (!found) { 
			snprintf(ltpstr,LTPSTR_LEN,"Socket deleted while blocking - exiting ltp_sendto()\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
			DROPMUTEX; // it got deleted!
			return(-1);
		}
		if (theblock->complete) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Socket completed while blocking - exiting ltp_sendto() normally\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			DROPMUTEX;
			return(len);
		}
		if (socksess->finishedwith) {
			snprintf(ltpstr,LTPSTR_LEN,"Socket finishedwith while blocking - exiting ltp_sendto()\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
			DROPMUTEX;
			return(-1);
		}
		DROPMUTEX;
	};
	return(len);
}

/*!
 * @fn int ltp_sendto(int sock, const void *msg, size_t len, int flags, ltpaddr *to, ltpaddr_len tolen)
 * @brief sendto equivalent for LTP
 * @param sock is a destination LTP "socket" 
 * @param msg are the bytes to send
 * @param len is the number of bytes
 * @param flags contains flags for the LTP socket
 * @param to is the destination address
 * @param tolen is the length of the destination address
 * @return is the number of bytes sent or -1 for an error
 */
int ltp_sendto(int sock, const void *msg, size_t len, int flags, ltpaddr *to, ltpaddr_len tolen)
{
	unsigned char *msgcp=(unsigned char*)calloc(1,len);
	if (!msgcp) LTPERR(-1);
	memcpy(msgcp,(unsigned char*)msg,len);
	int rv=ltp_sendto_aux(sock,msgcp,len,flags,to,tolen);
	if (rv==-1) {
		free(msgcp);
	}
	return(rv);
}

/*!
 * @fn int ltp_block2buf(ltp_block *b,void *buf,int buflen) 
 * @brief get the bytes of a block from wherever & put 'em in a buffer if they fit
 * @param b is an ltp_block pointer (any option is ok here)
 * @param buf is a buffer (allocated already) into which we put the bytes (if they fit)
 * @param buflen is the size of the buffer
 * @return -1 for an error otherwise the actual number of bytes from the block
 */
static int ltp_block2buf(ltp_block *b,void *buf,int buflen) 
{
	int blen; // actual length of block
	if (b->decoded.val && b->decoded.len) { // easy!!!
		blen=b->decoded.len;
		if (blen > buflen) return(-blen);
		memcpy(buf,b->decoded.val,blen);
		return(blen);
	}
	/// @todo add file and file pointer handlers here when I get to that
	return(-1);
}

/*!
 * @fn ltp_recvfrom(int sock, unsigned char *msg, size_t len, int flags, ltpaddr *addr, ltpaddr_len *addr_len)
 * @brief sendto equivalent for LTP
 * @param sock is a destination LTP "socket" (like a socket)
 * @param msg are the bytes read
 * @param len is the number of bytes
 * @param flags contains flags for the LTP socket
 * @param addr contains the address (engine ID) of the LTP peer
 * @param addr_len contains the size of the above
 * @return is the number of bytes sent or a negative number  for an error (if the input buffer was too small
 *     we return the negative of the number that would have been ok and a subsequent call with that sized
 *     buffer should be fine. (Note this differs from UDP's recvfrom return, because LTP blocks can be BIG!)
 */
int ltp_recvfrom(int sock, unsigned char *msg, size_t len, int flags, ltpaddr *addr, ltpaddr_len *addr_len)
{
	LTPINIT;
	GRABMUTEX;
	GETSOCKSESS;
#ifdef LTP_EXTRa_EXTRA_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Entering ltp_recvfrom with a %d byte buffer on socket %d\n",
					len,ltpaddr2str(addr),sock);
	ltp_log(LTP_LOG_MAX,ltpstr);
	ltp_log_flush();
#endif
	if (!found) { GOTHERE; DROPMUTEX; return(-1); }
	if (!socksess->inbound) { 
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Socket (%d) being made ready for inbound\n",socksess->id);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		socksess->inbound=true;
	}

	int rv=fixl2ifnecessary();
	if (rv) { GOTHERE; DROPMUTEX; LTPERR(rv); }

	// now check if there're any inbound, complete, not finishedwith blocks.
	// if there are - deliver 'em and mark 'em finishedwith
	// I should change this to add a new bool "delivered" which is checked near here
	// that way I can deliver the block before the RS/RA cycle is complete but only
	// send it up once (unlike now)
	bool gotblock=false;
	ltp_block *theblock;
	for (int i=0;!gotblock && i!=socksess->nblocks;i++) {
		if (socksess->blocks[i].complete &&
			!socksess->blocks[i].delivered) { // bingo!!
			theblock=&socksess->blocks[i];
			gotblock=true;
		}
	}
	if (!gotblock) {
		// check if there's anything in the loopback queue.
		// and return that if there is.
		if (ltp_state_var.lbq.nblocks) {
			// aha - something was once in the loopback queue
			// let's see if any of it's still for delivery
			ltp_block *blocks=ltp_state_var.lbq.blocks;
			bool foundone=false;
			int theone=0;
			for (int i=0;!foundone && i!=ltp_state_var.lbq.nblocks;i++) {
				if (!blocks[i].finishedwith) {
					foundone=true;
					theone=i;
				}
			}
			if (!foundone) { GOTHERE; DROPMUTEX; return(0); }
			int blen=ltp_block2buf(&blocks[theone],msg,len);
			if (blen<0) { 
				snprintf(ltpstr,LTPSTR_LEN,"Error: ltp_block2buf failed (%d) - is your input to ltp_recvfrom big enough?\n",blen);
				ltp_log(LTP_LOG_MAX,ltpstr);
				DROPMUTEX; return(blen); 
			}
			*addr=ltplb; // might need a strdup here sometime?
			*addr_len=LTPLBLEN;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",blocks[theone].id,__LINE__);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			blocks[theone].finishedwith=true;
			LTP_STATE_CHANGED;
			return(blen);
		} else {
			GOTHERE; DROPMUTEX;
			return(0);
		}
	}
	if (gotblock && 
			(socksess->opts.t_cfg.need_e2e_int  // expecting some - change this to a new boolean!
			||
			theblock->t_vals.flags&LTPT_SET_E2E_AUTH)) { // got some
		char sstr[MAXPATH];
		bool status=true;
		rv=ltpt_final_processing(theblock,socksess,&status,sstr,MAXPATH);
		if (rv) {
			snprintf(ltpstr,LTPSTR_LEN,"ltpt_final_processing error on block id %d - rv=%d\n",
					theblock->id,rv);
			ltp_log(LTP_LOG_MAX,ltpstr);
			GOTHERE; DROPMUTEX;
			return(0);
		}
		if (!status) {
			snprintf(ltpstr,LTPSTR_LEN,"ltpt_final_processing doesn't like block %d because: %s\n",
					theblock->id,sstr);
			ltp_log(LTP_LOG_MAX,ltpstr);
		} 
#ifdef LTP_VERBOSE
		else {
			snprintf(ltpstr,LTPSTR_LEN,"ltpt_final_processing likes block %d - Yipee!\n",theblock->id);
			ltp_log(LTP_LOG_MAX,ltpstr);

		}
#endif
		// if status and we're asking for e2e_auth then barf, otherwise allow it out
		if (!status && socksess->opts.t_cfg.need_e2e_int) { // oops!
			snprintf(ltpstr,LTPSTR_LEN,"Ditching block since LTP-T final processing says so and I'm configured to honour that\n");
			ltp_log(LTP_LOG_MAX,ltpstr);
			GOTHERE; DROPMUTEX;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",theblock->id,__LINE__);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			theblock->finishedwith=true; // shame, but them's the rules!
			return(0);
		} 
	}

	int blen=ltp_block2buf(theblock,msg,len);
	if (blen<0) { 
		snprintf(ltpstr,LTPSTR_LEN,"Error: ltp_block2buf failed (%d) - is your input to ltp_recvfrom big enough?\n",blen);
		ltp_log(LTP_LOG_MAX,ltpstr);
		DROPMUTEX; return(blen); 
	}
	// now addr is peer
	// hang on - maybe we got it via LTP-T in which case:
	if (theblock->t_vals.flags & LTPT_SET_SOURCE) {
		*addr=theblock->t_vals.source;
		*addr_len=sizeof(theblock->t_vals.source);
	} else {
		*addr=socksess->ibpeer;
		*addr_len=socksess->ibpeerlen;
	}
	theblock->delivered=true;
	if (theblock->finishedwithwhendelivered) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",theblock->id,__LINE__);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		theblock->finishedwith=true;
	}
	rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)socksess,true);
	if (rv) {
		snprintf(ltpstr,LTPSTR_LEN,"Cannot re-initialise inactivity timer for socket - exiting\n");
		ltp_log(LTP_LOG_ERR,ltpstr);
		GOTHERE; DROPMUTEX;
		return(-1);
	}
	LTP_STATE_CHANGED;
	GOTHERE; DROPMUTEX;
	return (blen);
}

/*!
 * @brief make up a C from a socket and the given code
 * @param seg is the segment
 * @param reasoncode is the reason for killing the session
 * @param sno is the session number to cancel
 * @return zero or error
 */
static int make_c_from_sock(ltpc_seg *seg, int reasoncode, ltpc_octets *sno)
{
	// peer
	int rv;
	INITMYADDR;
	ltpc_co *c=&seg->val.co;
	seg->nexts=0;
	seg->exts=NULL;
	if (ion_mode) {
		short engid=htons(my_addr.sock.sin_port);
		seg->peer.len=2;
		seg->peer.val=(unsigned char*)calloc(1,seg->peer.len);
		if (!seg->peer.val) LTPERR(LTP_MEMORY);
		memcpy(seg->peer.val,&engid,seg->peer.len);
	} else {
		rv=ltpaddr2octets(&my_addr,&seg->peer);
		if (rv) LTPERR(rv);
	}
	// sno
	seg->sno.len=sno->len;
	seg->sno.val=ltpc_alloc(sno->len);
	if (!seg->sno.val) LTPERR(LTP_MEMORY);
	memcpy(seg->sno.val,sno->val,sno->len);
	// exp
	// reasoncode
	c->reasoncode=reasoncode;
	return(LTP_OK);
}

/*!
 * @brief shutdown a socket (auxilliary func for use when already got mutex)
 * @param sock is the socket to close
 */
static void ltp_close_aux(int sock) 
{
	GETSOCKSESS;
	if (!found) {
		snprintf(ltpstr,LTPSTR_LEN,"Attempt to close unknown socket %d\n",sock);
		ltp_log(LTP_LOG_MAX,ltpstr);
		return;
	}
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"About to close socket %d\n",socksess->id);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	int rv;
	// send out a C segment if this is an open socket with some associated block of data
	int peerlen=0;
	ltpaddr *peer=NULL;
	if (socksess->outbound) {
		peerlen=socksess->obpeerlen;
		peer=&socksess->obpeer;
	} else { 
		peerlen=socksess->ibpeerlen;
		peer=&socksess->ibpeer;
	}
	if (peerlen && !socksess->finishedwith && socksess->nblocks>0) {

		// but only if there was some block in play (sometime) for the session!
		ltpc_octets sno;
		sno.len=0;
		sno.val=NULL;
		bool snofound=false;
		int bindex=0;

		// go through all blocks and take the 1st unfinished one, 
		// if they're all otherwise finishedwith, then there's no need to send a Cancel
		for (int i=0;!snofound && i!=socksess->nblocks;i++) {
			if (!socksess->blocks[i].finishedwith) {
				bindex=i;
				snofound=true;
			}  
			if (snofound && socksess->outbound) {
				sno.len=sizeof(socksess->blocks[i].id);
				sno.val=ltpc_alloc(sno.len);
				if (!sno.val) VOIDLTPERR(LTP_MEMORY);
				int foo=htonl(socksess->blocks[i].id);
				if (sno.len==4) {
					sno.val[0]=(foo>>24) & 0xff;
					sno.val[1]=(foo>>16) & 0xff;
					sno.val[2]=(foo>>8) & 0xff;
					sno.val[3]=foo & 0xff;
				} else {
					memcpy(sno.val,&foo,sno.len);
				}
			}
			if (snofound && socksess->inbound) {
				sno.len=socksess->blocks[i].remoteidlen;
				sno.val=ltpc_alloc(sno.len);
				if (!sno.val) VOIDLTPERR(LTP_MEMORY);
				memcpy(sno.val,socksess->blocks[i].remoteid,sno.len);
			}
		}
		if (!snofound) { // nothing to send, and can springclean
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Not sending Cancel segment for socket %d since all %d blocks are finishedWith\n",
									sock,socksess->nblocks);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif

			rv=kill_timer(LTP_TIMER_INACTIVITY,socksess);
			if (rv) VOIDLTPERR(rv);
			rv=springclean(socksess);
			if (rv) VOIDLTPERR(rv);

		} else {
		
			// add a new segment for tx
			int nsegments=ltp_state_var.txq.nsegments;
			int alloced=ltp_state_var.txq.alloced;
			ltp_segment *segments=ltp_state_var.txq.segments;
		
		
			VOIDFOOREALLOC(ltp_segment,segments,nsegments+1,alloced);
			ltp_state_var.txq.segments=segments; // realloc might have moved him!
			ltp_state_var.txq.alloced=alloced;
			ltp_state_var.txq.nsegments=nsegments+1;
		
			// make up an CO or CR with the latest info
			segments[nsegments].id=NI;
			segments[nsegments].state=LTP_PS_WAITING_SEND;
			segments[nsegments].r_entry=LTP_NOCACHE;
			segments[nsegments].r_id=0;
			segments[nsegments].t_entry=LTP_NOTIMER;
			segments[nsegments].t_id=0;
			segments[nsegments].finishedwith=false;
			if (socksess->outbound) {
				segments[nsegments].seg.type=LTPC_SEG_CO;
			} else {
				segments[nsegments].seg.type=LTPC_SEG_CR;
			}
			// hack hack - depending on ltpc_co and ltpc_cr having same structure here!
			ltpc_co *c=&segments[nsegments].seg.val.co;
			rv=make_c_from_sock(&segments[nsegments].seg,LTPC_RC_CLOSE,&sno);
			if (rv) { VOIDLTPERR(rv); }
			ltpc_free(sno.val);
			segments[nsegments].parent=&socksess->blocks[bindex]; 
			segments[nsegments].parentid=socksess->blocks[bindex].id;
			memcpy(&segments[nsegments].peer,peer,sizeof(ltpaddr));
			segments[nsegments].peerlen=sizeof(ltpaddr);
			if (!socksess->outbound) {
				// fixup peer inside C segment (sigh)
				if (segments[nsegments].seg.peer.len) free (segments[nsegments].seg.peer.val);
				rv=ltpaddr2octets(peer,&segments[nsegments].seg.peer);
				if (rv) VOIDLTPERR(rv);
			}
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,
					"MSG: Queueing a %s for LTP[id:%s,port:%d] - reasoncode: %x\n",
					(socksess->outbound?"CO":"CR"),inet_ntoa(peer->sock.sin_addr),peer->sock.sin_port,c->reasoncode);
			ltp_log(LTP_LOG_MAX,ltpstr);
#endif
			rv=springclean(socksess);
			if (rv) VOIDLTPERR(rv);
		}

		LTP_STATE_CHANGED;
	} else {
		socksess->finishedwith=true;
		rv=springclean(socksess);
		if (rv) VOIDLTPERR(rv);
	}

	return;
}

/*! 
 * @brief spring clean a socket and associated stuff
 * @param socksess is the socket who's stuff I'm cleaning
 * @return zero or error
 */
static int springclean(ltp_socket_t* socksess) 
{
	// do a bit of spring cleaning
	socksess->finishedwith=true;
	// zap its blocks, segments, timers etc.
	for (int i=0;i!=socksess->nblocks;i++) {
		socksess->blocks[i].finishedwith=true;
		for (int j=0;j!=ltp_state_var.txq.nsegments;j++) {
			if (socksess->blocks[i].id ==
				ltp_state_var.txq.segments[j].parentid) {
				// zap dat segment
				ltp_instance segint=ltp_state_var.txq.segments[j].id;
				ltp_state_var.txq.segments[j].finishedwith=true;
				// and any associated replay cache entries 
				for (int k=0;k!=ltp_state_var.replays.nentries;k++) {
					if (segint==ltp_state_var.replays.details[k].parentid) {
						ltp_state_var.replays.details[k].finishedwith=true;
					}
				}
				// and any associated timers
				for (int k=0;k!=ltp_state_var.timers.nentries;k++) {
					if (segint==ltp_state_var.timers.details[k].segmentid) {
						ltp_state_var.timers.details[k].finishedwith=true;
					}
				}
			}
		}
	}
	// zap any inactivity timers for the socket 
	// there could be others hanging about as well - what about them? TODO
	for (int k=0;k!=ltp_state_var.timers.nentries;k++) {
		if (ltp_state_var.timers.details[k].type==LTP_TIMER_INACTIVITY && 
			ltp_state_var.timers.details[k].socketid==socksess->id) {
			ltp_state_var.timers.details[k].finishedwith=true;
		}
	}
	return(LTP_OK);
}

/*!
 * @fn ltp_close(int sock)
 * @brief shutdown a socket
 * @param sock is the socket to close
 * 
 * The non-obvious bit: If SO_LINGER is set for the socket and if there're
 * already outstanding data segments to be sent, then we don't close this
 * guy now. What'll happen is whenever the last RA is sent out, then we
 * will call ltp_close_aux() if SO_LINGER was set and an application called
 * this (sock->close_called==true).
 */
void ltp_close(int sock) 
{
	GRABMUTEX;
	GETSOCKSESS;
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Application called ltp_close at: %s for socket %d\n",ltp_time_str(ltp_now()),sock);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	if (!found) { DROPMUTEX; return; }
	socksess->close_called=true;
	bool data2go=false;
	// is outbound && any data segments queued already?
	if (socksess->opts.linger && socksess->outbound) {
		for (int i=0;i!=socksess->nblocks;i++) {
			for (int j=0;!data2go && j!=ltp_state_var.txq.nsegments;j++) {
				if (ltp_state_var.txq.segments[j].parentid==socksess->blocks[i].id && 
					ltp_state_var.txq.segments[j].seg.type==LTPC_SEG_DS) {
						data2go=true;
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Not closing yet linger set and data segments outstanding on an outbound session\n");
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				}
			}
		}
	}
	// try not close inbound ones until the block is complete
	// might not be enough though, when's complete set?
	if (!data2go && socksess->opts.linger && socksess->inbound) {
		for (int i=0;!data2go && i!=socksess->nblocks;i++) {
			ltp_block *theblock=&socksess->blocks[1];
			if (!theblock->complete || !theblock->finishedwith) {
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Not closing yet linger set and something outstanding on an inbound socket\n");
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				data2go=true;
			}
		}
	}
	if (!data2go) {
		ltp_close_aux(sock);
	}
	DROPMUTEX;
	return;
}

/*!
* @fn ltp_get_status(int *n, ltp_status *stats, bool *more2go)
* @brief check the status of the LTP layer
* @param n is the number of status reports 
* @param stats is an array of status reports
* @param more2go specifies whether there are more status reports available (which don't fit in the current stats)
* @return is zero for success, or non-zero for error
* On input, the parameter n specifies the size of the array stats (which is allocated by the 
* caller); on output it contains the number of status reports which have been "filled in";
* if there are more status reports that the (input) n, then more2go will be set to true.
* Pointers within report structures are allocated by the implementation and must not be
* free'd by the caller (@todo think about threads).
*/
int ltp_get_status(int *n,ltp_status *stats, bool *more2go)
{
	return(LTP_OK);
}

/*!
 * @brief do a bit of tailored reallocing
 * @param thing is an ltp_scopes ptr
 * @param newsize is the (minimum) new size we want
 * @return zero or error
 */
static int morescopespace(ltp_scopes *thing, int newsize)
{
	if (!thing) return(LTP_BAD_PARAM);
	if (newsize<=thing->alloced) return(LTP_OK);
	int ans=newsize+100; // don't realloc too often
	int *newlbs;
	if (thing->lbs) {
		newlbs=(int*)realloc(thing->lbs,ans*sizeof(int));
	} else {
		newlbs=(int*)calloc(ans,sizeof(int));
	}
	if (!newlbs) LTPERR(LTP_MEMORY);
	int *newubs;
	if (thing->ubs) {
		newubs=(int*)realloc(thing->ubs,ans*sizeof(int));
	} else {
		newubs=(int*)calloc(ans,sizeof(int));
	}
	if (!newubs) LTPERR(LTP_MEMORY);
	thing->nscopes=newsize;
	thing->alloced=ans;
	thing->lbs=newlbs;
	thing->ubs=newubs;
	return(LTP_OK);
}

/*!
 * @brief deep-copy a scope
 * @param from is one to copy
 * @param to is the place to put the copy
 * @return zero or error
 */
static int scope_cpy(ltp_scopes *from,ltp_scopes *to)
{
	int *cclbs;
	int *ccubs;
	if (!to || !from) LTPERR(LTP_BAD_PARAM);
	if (from->nscopes && (!from->lbs || !from->ubs)) LTPERR(LTP_BAD_VALUE);
	*to=*from;
	to->ubs=(int*)malloc(from->nscopes*sizeof(int));
	if (!to->ubs) LTPERR(LTP_MEMORY);
	to->lbs=(int*)malloc(from->nscopes*sizeof(int));
	if (!to->lbs) LTPERR(LTP_MEMORY);
	memcpy(to->lbs,from->lbs,from->nscopes*sizeof(int));
	memcpy(to->ubs,from->ubs,from->nscopes*sizeof(int));
	return(LTP_OK);
}

/*!
 * @brief "subtract" two scopes
 * @param s1 is first operand
 * @param s2 is the 2nd operand
 * @param res is the result, i.e. scope1-scope2 
 * @return zero or error
 */
static int scope_sub(ltp_scopes *s1, ltp_scopes *s2, ltp_scopes *res)
{
	// we're substracting ranges here
	// there are 5 possibilities...we'll denote the current s1 range by [] and the current s2 range by {}, 
	// and we derive a res range
	if (!s1 || !s2 || !res) LTPERR(LTP_BAD_PARAM);
#ifdef LTP_EXTRA_EXTRA_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"%s:%s:%d, s1.n=%d, s2.n=%d\n", __FUNCTION__,__FILE__,__LINE__,
				s1->nscopes,s2->nscopes);
	ltp_log(LTP_LOG_MAX,ltpstr);
	lowlogscope(s1->nscopes, s1->lbs, s1->ubs, __LINE__, LTP_SUB_SCOPES, 0);
	lowlogscope(s2->nscopes, s2->lbs, s2->ubs, __LINE__, LTP_SUB_SCOPES, 0);
#endif
	int rv=LTP_OK;
	res->nscopes=0;
	if (s2->nscopes==0) {
		// special case, answer is s1
		int rv=scope_cpy(s1,res);
		return(rv);
	}
	if (s1->nscopes==0 && s2->nscopes==0) {
		// special case, answer is either
		int rv=scope_cpy(s1,res);
		return(rv);
	}
	if (s1->nscopes==0 && s2->nscopes!=0) {
		// I'm not clever enough to figure out a good answer there!
		LTPERR(LTP_BAD_VALUE);
	}
	for (int i=0,j=0,k=0;i!=s1->nscopes;) {
#ifdef LTP_EXTRA_EXTRA_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"%s:%s:%d, i=%d,j=%d,k=%d\n", __FUNCTION__,__FILE__,__LINE__,i,j,k);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		if (s1->lbs[i] < s2->lbs[j] && s1->ubs[i] < s2->ubs[j]) {
			// option 1:  [{]} -> [{
			morescopespace(res,k+1);
			res->lbs[k]=s1->lbs[i];
			res->ubs[k]=s2->lbs[j];
			i++; k++;
		} else if (s1->ubs[i]<s2->lbs[j]) {
			// option 2: []{} -> []
			morescopespace(res,k+1);
			res->lbs[k]=s1->lbs[i];
			res->ubs[k]=s1->ubs[i];
			i++; k++;
		} else if (s1->lbs[i]>=s2->lbs[j] && s1->ubs[i]>s2->ubs[j]) {
			// option 3: {[}] -> }]
			morescopespace(res,k+1);
			res->lbs[k]=s2->ubs[j];
			res->ubs[k]=s1->ubs[i];
			i++; k++;
		} else if (s1->lbs[i]<s2->lbs[j] && s1->ubs[i]>s2->ubs[j]){
			// option 4: [{}] -> [{ and }] (two new ranges)
			morescopespace(res,k+2);
			res->lbs[k]=s1->lbs[i];
			res->ubs[k]=s2->lbs[j];
			k++;
			res->lbs[k]=s2->ubs[j];
			res->ubs[k]=s1->ubs[i];
			i++;
		} else if (s2->lbs[j] <= s1->lbs[i] && s2->ubs[j] >= s1->ubs[i]) {
			// option 5: {[]} -> no new res
			i++;
		} else {
			// crap!!
			snprintf(ltpstr,LTPSTR_LEN,"%s:%s:%d, oops !!! i=%d,j=%d,k=%d\n", __FUNCTION__,__FILE__,__LINE__,i,j,k);
			ltp_log(LTP_LOG_MAX,ltpstr);
			LTPERR(LTP_BAD_VALUE);
		}
		// chew up any unnecessary s2's
		if (i!=s1->nscopes) {
			while (j<(s2->nscopes-1) && s2->ubs[j]<s1->lbs[i]) { 
				j++; 
				if (j==s2->nscopes) { 
					j--;
				}
			}
		}
	}
	return(LTP_OK);
}

/*
 * @brief free scope memory
 * @param sc is the one to free up
 */
static void scope_free(ltp_scopes *sc)
{
	if (!sc) VOIDLTPERR(LTP_BAD_PARAM);
	if (sc->nscopes) {
		if (!sc->ubs) VOIDLTPERR(LTP_BAD_VALUE);
		free(sc->ubs);
		if (!sc->lbs) VOIDLTPERR(LTP_BAD_VALUE);
		free(sc->lbs);
	}
	memset(sc,0,sizeof(ltp_scopes));
	return;
}

/*!
 * @brief send on bits of a block as and when I can
 * @param to is the next hop for this block
 * @param block is the block in question
 * @param sock is its socket
 * @return zero or error
 */
static int ltpt_forward_block(ltpaddr *to, ltp_block *block, ltp_socket_t *sock)
{
	int rv=LTP_OK;
	if (!to || !block || !sock) LTPERR(LTP_BAD_PARAM);
	// for each chunk of the block that's not yet been sent, queue it for sending
	// first figure current mtu
	int rscopes=block->rx_scopes.nscopes;
	int fscopes=block->fwd_scopes.nscopes;

	ltp_scopes tbs; // this is the set of ranges that are To-Be-Sent, which is the
					// rx_scopes "minus" fwd_scopes
					// this'll end up being the "new" value of fwd_scopes
	if (!fscopes) {
		tbs.nscopes=block->rx_scopes.nscopes;
		tbs.alloced=tbs.nscopes;
		tbs.ubs=(int*)calloc(tbs.nscopes,sizeof(int));
		if (!tbs.ubs) LTPERR(LTP_MEMORY);
		memcpy(tbs.ubs,block->rx_scopes.ubs,tbs.nscopes*sizeof(int));
		tbs.lbs=(int*)calloc(tbs.nscopes,sizeof(int));
		if (!tbs.lbs) LTPERR(LTP_MEMORY);
		memcpy(tbs.lbs,block->rx_scopes.lbs,tbs.nscopes*sizeof(int));
	} else {
		tbs.nscopes=0; tbs.alloced=0; tbs.ubs=NULL; tbs.lbs=NULL;
		rv=scope_sub(&block->rx_scopes,&block->fwd_scopes,&tbs);
		if (rv) LTPERR(rv);
	}

#ifdef LTP_VERBOSE
	{
		static ltp_scopes lastprinted={0,0,NULL,NULL};
		static int samecnt=0;
		ltp_scopes diff={0,0,NULL,NULL};
		// if nothing to be sent, print nothin
		if (tbs.nscopes!=0) {
			rv=scope_sub(&tbs,&lastprinted,&diff);
			if (rv) {
				snprintf(ltpstr,LTPSTR_LEN,"%s: scope_sub error (%d)  diff %d times at %s\n",
					__FUNCTION__,rv,samecnt,ltp_now_str());
				ltp_log(LTP_LOG_MAX,ltpstr);
				logscope(block,__LINE__,LTP_RX_SCOPES);
				logscope(block,__LINE__,LTP_FWD_SCOPES);
				lowlogscope(tbs.nscopes, tbs.lbs, tbs.ubs, __LINE__, LTP_SUB_SCOPES, block->id);
				
			} else if (diff.nscopes) {
				logscope(block,__LINE__,LTP_RX_SCOPES);
				logscope(block,__LINE__,LTP_FWD_SCOPES);
				lowlogscope(tbs.nscopes, tbs.lbs, tbs.ubs, __LINE__, LTP_SUB_SCOPES, block->id);
				scope_free(&lastprinted);
				scope_cpy(&tbs,&lastprinted);
				samecnt=0;
			} else {
				samecnt++;
				if (samecnt%1000==0) {
					snprintf(ltpstr,LTPSTR_LEN,"%s: same scope diff %d times at %s\n",
							__FUNCTION__,samecnt,ltp_now_str());
					ltp_log(LTP_LOG_MAX,ltpstr);
					logscope(block,__LINE__,LTP_RX_SCOPES);
					logscope(block,__LINE__,LTP_FWD_SCOPES);
					lowlogscope(tbs.nscopes, tbs.lbs, tbs.ubs, __LINE__, LTP_SUB_SCOPES, block->id);
					lowlogscope(lastprinted.nscopes, lastprinted.lbs, lastprinted.ubs, __LINE__, LTP_SUB_SCOPES, block->id);
				}
			}
		}
	}
#endif

	// TODO: I may want to modify the ranges here to take account of filling MTUs or other factors 
	// that'd mean messing with tbs a bit more, for now we'll take it as we have it, except we
	// won't try send >1 MTU in a DS of course

	// now make up DSes for those bytes, at this point we assume that all marked bytes are
	// to be sent
	int mtu=latestmtu(block,sock);
	ltpc_octets rsn;
	rsn.len=0;
	rsn.val=NULL;
	for (int i=0;i!=tbs.nscopes;i++) {
		int lastseg=((tbs.ubs[i]-tbs.lbs[i])%mtu);
		int segsperscope=(tbs.ubs[i]-tbs.lbs[i])/mtu + (lastseg?1:0);
		for (int j=0;j!=segsperscope;j++) {
			int seglen=((j==segsperscope-1) ? lastseg : mtu);
			int offset=tbs.lbs[i]+(j*mtu);
			bool checkpoint=false; // the ds will be a cp iff its the last red seg, and in this case is automaticly figured
			rv=ds_from_block(to,block,seglen,offset,checkpoint,&rsn);
			if (rv) LTPERR(rv);
		}
	}

	// and remember the new fwd_scopes for next time
	if (block->fwd_scopes.nscopes) {
		free(block->fwd_scopes.ubs);
		free(block->fwd_scopes.lbs);
	}
	// Correct version - now that I've sent the diff, the rx is what's been fwded
	if (tbs.nscopes) {
		free(tbs.ubs);
		free(tbs.lbs);
	}
	if (block->rx_scopes.nscopes) { // deep copy
		block->fwd_scopes.nscopes=block->rx_scopes.nscopes; 
		block->fwd_scopes.ubs=(int*)calloc(block->fwd_scopes.nscopes,sizeof(int));
		if (!block->fwd_scopes.ubs) LTPERR(LTP_MEMORY);
		memcpy(block->fwd_scopes.ubs,block->rx_scopes.ubs,block->fwd_scopes.nscopes*sizeof(int));
		block->fwd_scopes.lbs=(int*)calloc(block->fwd_scopes.nscopes,sizeof(int));
		if (!block->fwd_scopes.lbs) LTPERR(LTP_MEMORY);
		memcpy(block->fwd_scopes.lbs,block->rx_scopes.lbs,block->fwd_scopes.nscopes*sizeof(int));
	} 

	// when are we done forwarding?
	if (block->eob_seen && block->fwd_scopes.nscopes==1
			&& block->fwd_scopes.lbs[0]==0
			&& block->fwd_scopes.ubs[0]==block->decoded.len) {
		block->t_vals.forwarded=true; // might need a new boolean, let's see
	}

	return(LTP_OK);
}

/*!
 * @brief figure out what LTP-T segments to send forward (and put relevant segments into the tx queue)
 * @param nlisteners is the number of guys listening
 * @param listeneres is an array of their names/addresses
 * @return zero or error
 */
static int ltpt_trickle_forward(int nlisteners, ltpaddr *listeners)
{
	if (!nlisteners) return(LTP_OK);
	int rv=LTP_OK;
	ltpaddr nexthop;
	ltpaddr_len nexthoplen=0;
#define GWLISTENERS
#ifdef GWLISTENERS

	// got anything going there?
	// if so then send what we have
	for (int sockind=0;sockind!=ltp_state_var.nsockets;sockind++) {
		ltp_socket_t	*sock=&ltp_state_var.sockets[sockind];
		if (sock->ltptrelay) {
			for (int blockind=0;blockind!=sock->nblocks;blockind++) {
				ltp_block	*block=&sock->blocks[blockind];
				bool		matched=false;
				if (block->t_vals.flags && !block->finishedwith && !block->t_vals.forwarded 
					&& block->fwd_scopes.nscopes==0) { // more criteria too no doubt
					// LTP_ALLRED isn't appropriate for forwarding so force it to zero if needed
					if (sock->opts.redlen==LTP_ALLRED) {
						snprintf(ltpstr,LTPSTR_LEN,"Forcing redlen to 0 for socket %d (was %d)\n",sock->id,sock->opts.redlen);
						ltp_log(LTP_LOG_MAX,ltpstr);
						sock->opts.redlen=0;
					}
					// who to?
					rv=ltpt_gateway_listeners(&block->t_vals.dest,&nexthop,nlisteners,listeners,&matched);
					if (rv) { 
						snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltpt_gateway_listeners (%d)\n",rv);
						ltp_log(LTP_LOG_MAX,ltpstr);
						ltp_log_flush();
						LTPERR(rv);
					}
					if (matched) {
						nexthoplen=sizeof(nexthop);
						for (int k=0;k!=nlisteners;k++) { // ignore that unless the nexthop is now listening (we'll come back to it I hope)
							if (!ltpaddr_cmp(&nexthop,&listeners[k],nexthoplen)) { // ok we may have something to send
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Trickle forwarding block (size:%d,id:%d) to %s\n",
									block->decoded.len,block->id,ltpaddr2str(&nexthop));
								ltp_log(LTP_LOG_MAX,ltpstr);
								ltp_log_flush();
#endif
								memcpy(&sock->obpeer,&nexthop,sizeof(nexthop));
								sock->obpeerlen=sizeof(nexthop);
								// block->t_vals.gw=nexthop; - copy below should be better
								block->t_vals.gw=sock->obpeer;
								rv=ltpt_forward_block(&nexthop,block,sock);
								if (rv) {
									snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltpt_forward_block (%d)\n",rv);
									ltp_log(LTP_LOG_MAX,ltpstr);
									ltp_log_flush();
									LTPERR(rv);
								}
							} else {
#ifdef LTP_EXTRA_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Not Trickle forwarding block yet (size:%d,id:%d) for %s (at %d)\n",
							block->decoded.len,block->id,ltpaddr2str(&block->t_vals.dest),__LINE__);
						ltp_log(LTP_LOG_MAX,ltpstr);
						ltp_log_flush();
#endif
							}
						}
					} else {
#ifdef LTP_EXTRA_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Not Trickle forwarding block yet (size:%d,id:%d) for %s (at %d)\n",
							block->decoded.len,block->id,ltpaddr2str(&block->t_vals.dest),__LINE__);
						ltp_log(LTP_LOG_MAX,ltpstr);
						ltp_log_flush();
#endif
					}

				} else if (block->t_vals.flags && !block->finishedwith && !block->t_vals.forwarded
					&& block->fwd_scopes.nscopes>0) { // more criteria too no doubt
					// maybe more to forward? we've started already
					// but only queue more up if that blocks gw is a listener now
					// otherwise we may waste bandwidth with small MTUs
					nexthop=block->t_vals.gw;
					nexthoplen=sizeof(nexthop);
#ifdef LTP_EXTRA_EXTRA_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Trickle forwarding block (size:%d,id:%d) for %s (at %d)\n",
						block->decoded.len,block->id,ltpaddr2str(&block->t_vals.dest),__LINE__);
					ltp_log(LTP_LOG_MAX,ltpstr);
					ltp_log_flush();
#endif
					for (int k=0;k!=nlisteners;k++) {
						if (!ltpaddr_cmp(&nexthop,&listeners[k],nexthoplen)) { // ok we may have something to send
#ifdef LTP_EXTRA_EXTRA_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Trickle forwarding block (size:%d,id:%d) for %s (at %d)\n",
						block->decoded.len,block->id,ltpaddr2str(&block->t_vals.dest),__LINE__);
					ltp_log(LTP_LOG_MAX,ltpstr);
					ltp_log_flush();
#endif
							memcpy(&sock->obpeer,&nexthop,sizeof(nexthop));
							sock->obpeerlen=sizeof(nexthop);
							rv=ltpt_forward_block(&nexthop,block,sock);
							if (rv) {
								snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltpt_forward_block (%d)\n",rv);
								ltp_log(LTP_LOG_MAX,ltpstr);
								ltp_log_flush();
								LTPERR(rv);
							}
						}
					}
				}
			}
		}
	}

#else
	// old way

	if (!nlisteners) return(LTP_OK);
	for (int l=0;l!=nlisteners;l++) {
		// got anything going there?
		// if so then send what we have
		ltpaddr *listener=&listeners[l];
		for (int sockind=0;sockind!=ltp_state_var.nsockets;sockind++) {
			ltp_socket_t	*sock=&ltp_state_var.sockets[sockind];
			if (sock->ltptrelay) {
				for (int blockind=0;blockind!=sock->nblocks;blockind++) {
					ltp_block	*block=&sock->blocks[blockind];
					if (block->t_vals.flags && !block->finishedwith && !block->t_vals.forwarded 
						&& block->fwd_scopes.nscopes==0) { // more criteria too no doubt
						// who to?
						rv=ltpt_gateway(&block->t_vals.dest,&nexthop);
						if (rv) { 
							snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltpt_gateway (%d)\n",rv);
							ltp_log(LTP_LOG_MAX,ltpstr);
							ltp_log_flush();
							LTPERR(rv);
						}
						nexthoplen=sizeof(nexthop);
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Trickle forwarding block (size:%d,id:%d) to %s\n",
								block->decoded.len,block->id,ltpaddr2str(&nexthop));
						ltp_log(LTP_LOG_MAX,ltpstr);
						ltp_log_flush();
#endif
						memcpy(&sock->obpeer,&nexthop,sizeof(nexthop));
						sock->obpeerlen=sizeof(nexthop);
						block->t_vals.gw=nexthop;
						if (!ltpaddr_cmp(&nexthop,listener,nexthoplen)) { // ok we may have something to send
							rv=ltpt_forward_block(&nexthop,block,sock);
							if (rv) {
								snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltpt_forward_block (%d)\n",rv);
								ltp_log(LTP_LOG_MAX,ltpstr);
								ltp_log_flush();
								LTPERR(rv);
							}
						}
					} else {
						// maybe more to forward? we've started already
						nexthop=block->t_vals.gw;
						memcpy(&sock->obpeer,&nexthop,sizeof(nexthop));
						sock->obpeerlen=sizeof(nexthop);
						if (!ltpaddr_cmp(&nexthop,listener,nexthoplen)) { // ok we may have something to send
							rv=ltpt_forward_block(&nexthop,block,sock);
							if (rv) {
								snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltpt_forward_block (%d)\n",rv);
								ltp_log(LTP_LOG_MAX,ltpstr);
								ltp_log_flush();
								LTPERR(rv);
							}
						}
					}
				}
			}
		}
	}

#endif
	return(rv);
}

/*
 * @brief run around forwarding (new) segment from (partly) received blocks
 */
void *forwarder_fnc (void* foo)
{
	int rv;
	snprintf(ltpstr,LTPSTR_LEN,"LTP-T: In forwarder thread - tid(%p)\n",(void*)forwarder);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
#ifdef LTP_EXTRA_EXTRA_VERBOSE
	int loopcount=0;
#endif
	while (true) {
		GRABMUTEX
		// do stuff
		// now go through relay sockets to see who has what for forwarding and handle it!
		for (int i=0;i!=ltp_state_var.nsockets;i++) {
			ltp_socket_t	*sock=&ltp_state_var.sockets[i];
			if (sock->ltptrelay && sock->outbound && !sock->finishedwith) { // worth a look
				for (int j=0;j!=sock->nblocks;j++) {
					ltp_block	*block=&sock->blocks[j];

					if (!block->finishedwith && block->complete) {
						ltpaddr to;
						ltpaddr_len tolen=0;
						ltpaddr dest;
						// who to?
						rv=ltpt_gateway(&block->t_vals.dest,&to);
						if (rv) { DROPMUTEX; pthread_exit(NULL); }
						tolen=sizeof(to);
#ifdef LTP_EXTRA_EXTRA_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Forwarding complete block (size:%d,id:%d) to %s\n",
							block->decoded.len,block->id,ltpaddr2str(&to));
						ltp_log(LTP_LOG_MAX,ltpstr);
						ltp_log_flush();
#endif
						// fix up socket data
						DROPMUTEX;
						int rv=ltp_sendto_aux(sock->sockfd, block->decoded.val, block->decoded.len, 0, &to, tolen);
						if (rv==-1) {
							snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Error from ltp_sendto_aux (%d)\n",rv);
							ltp_log(LTP_LOG_MAX,ltpstr);
							ltp_log_flush();
							pthread_exit(NULL);
						}
						GRABMUTEX;
						// sent off and another copy was made can crap this one now
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",block->id,__LINE__);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						block->finishedwith=true;
					}
				}
			}
		}
		
		// don't do stuff
		DROPMUTEX
		// ltp_waitabit();
	}
	pthread_exit(NULL);
}


/*
 * @brief kick off a thread for forwarding (new) segment from (partly) received blocks
 * @return zero or error
 */
static int forwarder_thread_start(void)
{
#ifdef LTPT_ONLYSENDWHENFULL
	int rv;
	snprintf(ltpstr,LTPSTR_LEN,"LTP-T: About to start forwarder thread\n");
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
	ltp_log_flush();
	rv=pthread_create(&forwarder,NULL,forwarder_fnc,forwarder_message);
	if (rv) LTPERR(rv);
#ifdef LTP_TARGET_ARM_LINUX
	sleep(1);
	snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Started forwarder thread - tid(%p) then slept a second\n",forwarder);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
	ltp_log_flush();
#else
	snprintf(ltpstr,LTPSTR_LEN,"LTP-T: Started forwarder thread - tid(%p)\n",forwarder);
	ltp_log(LTP_LOG_ADVISORY,ltpstr);
#endif
#endif
	return(LTP_OK);
}

/*!
 * @fn ltp_init_aux(void)
 * @brief initialisation function - called upon 1st call to a ltp_*
 * @return is zero for success, or else an error
 *
 * This is the one to call after you've grabbed the MUTEX
 */
static int ltp_init_aux(void)
{
	if (ltp_state_var.inited) {
		return(LTP_OK);
	}
	int rv=ltp_sync_setup(NULL);
	if (rv) { 
		LTPERR(rv);
	}
	rv=fixl2ifnecessary();
	if (rv) {
		LTPERR(rv);
	}
	ltp_state_var.inited=true;
	// start ltp-t relays if necessary
	rv=ltpt_def_cfg(&ltp_state_var.t_cfg);
	if (rv) LTPERR(rv);
	rv=ltp_sync_setup(NULL);
	if (rv) LTPERR(rv);
	rv=fixl2ifnecessary();
	if (rv) LTPERR(rv);
	ltp_state_var.inited=true;
	// start ltp-t relays if necessary
	rv=ltpt_def_cfg(&ltp_state_var.t_cfg);
	if (rv) LTPERR(rv);
	rv=ltpt_init(&ltp_state_var.t_cfg);
	if (rv) LTPERR(rv);
	if (ltp_state_var.t_cfg.on) {
		// maybe kick off a thread for forwarding (new) segment from (partly) received blocks
		rv=forwarder_thread_start();
		if (rv) LTPERR(rv);
	}
#undef BITOFTESTCODE
#ifdef BITOFTESTCODE
	// I just want to see what happens here.
	ltp_scopes s1;
	ltp_scopes s2;
	ltp_scopes res;
	s1.nscopes=1; int s1lbs[1]={0}; int s1ubs[1]={262144}; s1.lbs=s1lbs; s1.ubs=s1ubs;
	s2.nscopes=1; int s2lbs[1]={0}; int s2ubs[1]={209560}; s2.lbs=s2lbs; s2.ubs=s2ubs;
	rv=scope_sub(&s1, &s2, &res);
	lowlogscope(s1.nscopes, s1.lbs, s1.ubs, __LINE__, LTP_SUB_SCOPES, 0);
	lowlogscope(s2.nscopes, s2.lbs, s2.ubs, __LINE__, LTP_SUB_SCOPES, 0);
	lowlogscope(res.nscopes, res.lbs, res.ubs, __LINE__, LTP_SUB_SCOPES, 0);
#endif

#ifdef MUDSAT
	// SF: MUDSAT
	// if there's a crypto cfg in the working directory then
	// use that and set up a ciphersuite
	// TODO: Parse it out to check
	FILE *fp=fopen(LTP_CRYPT_FNAME_DEF,"r");
	if (fp) { 
		// just openable is enough for now
		fclose(fp);
		rv=ltp_crypt_setup(LTP_CRYPT_FNAME_DEF);
		if (rv) LTPERR(rv);
		using_def_crypto=1;
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Will use default crypto (sig,keyid=foo,file=%s)\n",LTP_CRYPT_FNAME_DEF);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	} else {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Not using default crypto (no sign of file=%s)\n",LTP_CRYPT_FNAME_DEF);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	}
#endif

	return(LTP_OK);
}

/*!
 * @fn ltp_init(void)
 * @brief initialisation function - called upon 1st call to a ltp_*
 * @return is zero for success, or else an error
 */
int ltp_init(void)
{
	GRABMUTEX;
	int rv=ltp_init_aux();
	DROPMUTEX;
	return(rv);
}

/*!
 * @fn ltp_instance ltp_next_instance(void)
 * @return is a new (unused) instance number
 */
static ltp_instance ltp_next_instance(void)
{
	/// ok we're not going to care about the unlikely possibility of a collision!!!
	ltp_instance ret=0;
	while (!ret) ret=(ltp_instance) ltp_rand(); // no zeros please
	// feckin ION doesn't like all 32 bits used here in session numbers
	// easiest fix is to zero the top bit here
	ret&=0x7fffffff;
	return ret;
}

/*!
 * @fn int ltp_save(void)
 * @brief tells LTP to save its state etc to non-volatile storage
 * @return zero or error
 */
int ltp_save(void)
{
	LTPINIT;
	GRABMUTEX;
	int rv=ltp_sync_2disk();
	DROPMUTEX;
	if (rv) LTPERR(rv);
	return(rv);
}

/*!
 * @fn int ltp_restore
 * @brief tells LTP to restore its state etc from non-volatile storage
 * @return zero or error
 */
int ltp_restore(void)
{
	LTPINIT;
	GRABMUTEX;
	int rv=ltp_sync_fromdisk();
	if (rv==LTP_FILE) {
		snprintf(ltpstr,LTPSTR_LEN,"ltp_restore: can't read current.ltp file - continuing anyway\n");
		ltp_log(LTP_LOG_ERR,ltpstr);
	}
	else { 
		if (rv) { 
			snprintf(ltpstr,LTPSTR_LEN,"ltp_restore: ltp_sync_fromdisk failed (%d)\n",rv);
			ltp_log(LTP_LOG_ERR,ltpstr);
			DROPMUTEX; LTPERR(rv); 
		}
	}
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"ltp_restore: success\n");
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @brief make a CA from a C
 * @param ca_seg is the CA
 * @param c_seg is the C
 */
static int make_ca_from_c(ltpc_seg *ca_seg, ltpc_seg *c_seg)
{
	// header stuff etc.

	INITMYADDR
	ltpc_cao *ca=&ca_seg->val.cao;
	ltpc_co *c=&c_seg->val.co;
	ca_seg->nexts=0;
	ca_seg->exts=NULL;
	int rv=1;
	if (ion_mode) {
		short engid=htons(my_addr.sock.sin_port);
		ca_seg->peer.len=2;
		ca_seg->peer.val=(unsigned char*)calloc(1,ca_seg->peer.len);
		if (!ca_seg->peer.val) LTPERR(LTP_MEMORY);
		memcpy(ca_seg->peer.val,&engid,ca_seg->peer.len);
	} else {
		rv=ltpaddr2octets(&my_addr,&ca_seg->peer);
		if (rv) LTPERR(rv);
	}
	rv=ltpc_oct_cpy(&ca_seg->sno,&c_seg->sno);
	if (rv) LTPERR(rv);
	return(LTP_OK);
}

/*!
 * @brief make an RA from an RS
 * @param rs_seg is the RS
 * @param ra_seg is the RA
 */
static int make_ra_from_rs(ltpc_seg *rs_seg, ltpc_seg *ra_seg)
{
	// header stuff etc.

	INITMYADDR
	ltpc_ra *ra=&ra_seg->val.ra;
	ltpc_rs *rs=&rs_seg->val.rs;
	int pvallen=sizeof(my_addr.sock.sin_addr.s_addr);
	ra_seg->nexts=0;
	ra_seg->exts=NULL;
	// int rv=ltpaddr2octets(&my_addr,&ra_seg->peer);
	int rv=ltpc_oct_cpy(&ra_seg->peer,&rs_seg->peer);
	if (rv) LTPERR(rv);
	rv=ltpc_oct_cpy(&ra_seg->sno,&rs_seg->sno);
	if (rv) LTPERR(rv);
	rv=ltpc_oct_cpy(&ra->rsn,&rs->rsn);
	if (rv) LTPERR(rv);
	return(LTP_OK);
}

/*!
 * @brief make up an RS based on a block
 * @param theblock is the block 
 * @param ds_seg is the most recently rx'd data segment
 * @param rses is the number of rs'es created
 * @return zero or error
 */
static int make_rses_from_block(ltp_block *theblock, ltpc_seg *ds_seg, int *rses)
{
	ltpc_ds *ds=&ds_seg->val.ds;
	if (!ds->checkpoint) LTPERR(LTP_BAD_VALUE);

	// header stuff etc.
	NEWSEGMENT
	ltp_segment *firstseg=&segments[nsegments];
	segments[nsegments].seg.type=LTPC_SEG_RS;
	ltpc_rs *rs=&segments[nsegments].seg.val.rs;
	ltpc_seg *rs_seg=&segments[nsegments].seg;

	rs_seg->nexts=0;
	rs_seg->exts=NULL;
	int rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
	if (rv) LTPERR(rv);
	rv=ltpaddr2octets(&theblock->parent->ibpeer,&rs_seg->peer);
	if (rv) LTPERR(rv);
	rv=ltpc_oct_cpy(&rs_seg->sno,&ds_seg->sno);
	if (rv) LTPERR(rv);
	rv=ltpc_oct_cpy(&rs->csn,&ds->csn);
	if (rv) LTPERR(rv);
	// add a newly created rsn
#undef RND_RSN
#ifdef RND_RSN
	short foo=htons(NI&0x7fff);
	rs->rsn.len=2;
	rs->rsn.val=ltpc_alloc(2);
	rs->rsn.val[0]=foo>>8;
	rs->rsn.val[1]=foo%256;
	if (rs->rsn.val[0]==0x00) {
		// shorten
		rs->rsn.len=1;
		rs->rsn.val[0]=ds->csn.val[1];
	}
#else
	if (theblock->current_rsn==0) { // 1st time
		short foo=(NI&0x7fff);
		theblock->current_rsn=foo;
		foo=htons(foo);
		rs->rsn.len=2;
		rs->rsn.val=ltpc_alloc(2);
		rs->rsn.val[0]=foo>>8;
		rs->rsn.val[1]=foo&0xff;
		if (rs->rsn.val[0]==0x00) {
			// shorten
			rs->rsn.len=1;
			rs->rsn.val[0]=ds->csn.val[1];
		}
	} else {
		theblock->current_rsn++;
		if (theblock->current_rsn>=0x7fff) {
			snprintf(ltpstr,LTPSTR_LEN,"Wrapping the rsn for block %d\n",theblock->id);
			ltp_log(LTP_LOG_MAX,ltpstr);
			theblock->current_rsn=0;
		}
		short foo=htons(theblock->current_rsn);
		rs->rsn.len=2;
		rs->rsn.val=ltpc_alloc(2);
		rs->rsn.val[0]=foo>>8;
		rs->rsn.val[1]=foo&0xff;
		if (rs->rsn.val[0]==0x00) {
			// shorten
			rs->rsn.len=1;
			rs->rsn.val[0]=ds->csn.val[1];
		}
	}
#endif
	int nscopes=theblock->rx_scopes.nscopes;
	rs->lb=theblock->rx_scopes.lbs[0];
	rs->ub=theblock->rx_scopes.ubs[nscopes-1];
	int last_lb=rs->lb;
	int next_ub=rs->ub;
	int *lbs=theblock->rx_scopes.lbs;
	int *ubs=theblock->rx_scopes.ubs;
	int cccount=0;
	rs->offsets=NULL;
	rs->lengths=NULL;
	for (int sc=0;sc!=nscopes-1;sc++) {
		if (lbs[sc+1]!=ubs[sc]) { // there's a gap, so make a report
			// add an offset & length
			cccount++;
			rs->offsets=(int*)realloc(rs->offsets,cccount*sizeof(int));
			rs->lengths=(int*)realloc(rs->lengths,cccount*sizeof(int));
			rs->offsets[cccount-1]=last_lb-rs->lb;
			rs->lengths[cccount-1]=ubs[sc]-last_lb;
			last_lb=lbs[sc+1];
		}
	}
	// check for last one
	if (cccount && lbs[nscopes-1]!=ubs[nscopes-1]) {
		// add a new claim then
		cccount++;
		rs->offsets=(int*)realloc(rs->offsets,cccount*sizeof(int));
		rs->lengths=(int*)realloc(rs->lengths,cccount*sizeof(int));
		rs->offsets[cccount-1]=last_lb-rs->lb;
		rs->lengths[cccount-1]=ubs[nscopes-1]-last_lb;
	}
	if (cccount==0) {
		cccount++;
		rs->offsets=(int*)realloc(rs->offsets,cccount*sizeof(int));
		rs->lengths=(int*)realloc(rs->lengths,cccount*sizeof(int));
		rs->offsets[0]=0;
		rs->lengths[0]=rs->ub-rs->lb;
		rs->cc=1;
	} 
	rs->cc=cccount;

	segments[nsegments].parent=theblock; 
	segments[nsegments].parentid=theblock->id;
	ltpaddr to=theblock->parent->ibpeer; // we answer to the inbound peer!
	to.sock.sin_port=theblock->parent->ibpeer.sock.sin_port;
	to.sock.sin_family=AF_LTP;
	segments[nsegments].peer=to;
	segments[nsegments].peerlen=sizeof(to);

	*rses=1;

	// now we need to figure out if one RS is enough
	// if we need more than one, we'll split up the "main" one above
	int mtu=latestmtu(theblock,theblock->parent);

	int rsoverhead= // adding 1 byte for each sdnv encoding field
			7 + // peer
			5 + // sno
			// extensions are taken care of in latestmtu
			5 + // rsn
			5 + // csn
			5 + // ub
			5 + // lb
			5 ; // claim count
	int bytes4claims=mtu-rsoverhead;
	int bytesperclaim=10; // 2 x 32 bit offsets in sdnv
	int claimspersegment=bytes4claims/bytesperclaim;

	if (rs->cc <= claimspersegment) {
		// prepare a timer for the one and only
		rv=init_timer(LTP_TIMER_RS,&segments[nsegments]);
		if (rv) LTPERR(rv);
		return(LTP_OK);
	}

	// multi-segment cases
	// now split the current rs structure into different ones, each with claimspersegment;

	int segmentsneeded=rs->cc/claimspersegment + (rs->cc%claimspersegment?1:0);
#ifdef LTP_VERBOSE
	snprintf(ltpstr,LTPSTR_LEN,"Multi-segment Report: mtu=%d,rsoh=%d,4claims=%d,perclaim=%d,claimspersegment=%d, segmentsneeded=%d, total claims=%d\n",
			mtu,rsoverhead,bytes4claims,bytesperclaim,claimspersegment,segmentsneeded,rs->cc);
	ltp_log(LTP_LOG_MAX,ltpstr);
#endif

	int overallclaimnumber=0;

	for (int i=0;i!=segmentsneeded;i++) {
		nsegments++;
		FOOREALLOC(ltp_segment,segments,nsegments+1,alloced); 
		ltp_state_var.txq.segments=segments; 
		ltp_state_var.txq.alloced=alloced;   
		ltp_state_var.txq.nsegments=nsegments; 
		segments[nsegments].id=NI; 
		segments[nsegments].state=LTP_PS_WAITING_SEND; 
		segments[nsegments].r_entry=LTP_NOCACHE; 
		segments[nsegments].r_id=0; 
		segments[nsegments].t_entry=LTP_NOTIMER; 
		segments[nsegments].t_id=0; 
		segments[nsegments].finishedwith=false; 
		segments[nsegments].seg.type=LTPC_SEG_RS;
		ltpc_rs *rs_i=&segments[nsegments].seg.val.rs;
		ltpc_seg *rs_seg_i=&segments[nsegments].seg;

		// copy headers
		rv=ltpc_oct_cpy(&rs_seg_i->peer,&rs_seg->peer);
		if (rv) { LTPERR(LTP_MEMORY); }
		rv=ltpc_oct_cpy(&rs_seg_i->sno,&rs_seg->sno);
		if (rv) { LTPERR(LTP_MEMORY); }
		rv=ltpc_oct_cpy(&rs_i->csn,&ds->csn);
		if (rv) LTPERR(rv);
		// first RS copies rsn from "master", thereafter we increment
#ifdef RND_RSN
		#error "fix this crapology"
#else
		if (i==0) { // copy
			rv=ltpc_oct_cpy(&rs_i->rsn,&rs->rsn);
			if (rv) { LTPERR(LTP_MEMORY); }
		} else { // increment
			theblock->current_rsn++;
			if (theblock->current_rsn>=0x7fff) {
				snprintf(ltpstr,LTPSTR_LEN,"Wrapping the rsn for block %d\n",theblock->id);
				ltp_log(LTP_LOG_MAX,ltpstr);
				theblock->current_rsn=0;
			}
			short foo=htons(theblock->current_rsn);
			rs_i->rsn.len=2;
			rs_i->rsn.val=ltpc_alloc(2);
			rs_i->rsn.val[0]=foo>>8;
			rs_i->rsn.val[1]=foo&0xff;
			if (rs_i->rsn.val[0]==0x00) {
				// shorten
				rs_i->rsn.len=1;
				rs_i->rsn.val[0]=ds->csn.val[1];
			}
		} 
#endif
		// not optimising lb,ub (or claims) for now
		// now we are - was making a horlics of re-tx cause of that!
		// each RS was resulting in re-tx of all then-outstanding bytes
		// ouch
		rs_i->lb=rs->offsets[i*claimspersegment];
		// update our bounds
		if (i!=segmentsneeded) {
			rs_i->ub=rs->offsets[(i+1)*claimspersegment]+rs->lengths[(i+1)*claimspersegment];
		} else { // overall one
			rs_i->ub=rs->ub;
		}

		rs_i->cc=claimspersegment;
		if (i==segmentsneeded-1) { // last one
			rs_i->cc=rs->cc % claimspersegment;
		}
		rs_i->offsets=(int*)ltpc_alloc(rs_i->cc*sizeof(int));
		if (!rs_i->offsets) LTPERR(LTP_MEMORY);
		rs_i->lengths=(int*)ltpc_alloc(rs_i->cc*sizeof(int));
		if (!rs_i->lengths) LTPERR(LTP_MEMORY);

		for (int j=0;j!=rs_i->cc;j++) {
			// copy claims
			rs_i->offsets[j]=rs->offsets[i*claimspersegment+j]-rs_i->lb;;
			rs_i->lengths[j]=rs->lengths[i*claimspersegment+j];
		}
		
		segments[nsegments].parent=theblock; 
		segments[nsegments].parentid=theblock->id;
		ltpaddr to=theblock->parent->ibpeer; // we answer to the inbound peer!
		to.sock.sin_port=theblock->parent->ibpeer.sock.sin_port;
		to.sock.sin_family=AF_LTP;
		segments[nsegments].peer=to;
		segments[nsegments].peerlen=sizeof(to);
		// prepare a timer for the last one
		rv=init_timer(LTP_TIMER_RS,&segments[nsegments]);
		if (rv) LTPERR(rv);

	}
	// fix last one
	*rses=segmentsneeded;
						
	// free up original humungous one
	firstseg->state=LTP_PS_RS_ACKED; // no need for that segment to be sent anymore
	firstseg->finishedwith=true; // kill it again!
	ltpc_free_seg(rs_seg);
	// firstseg will be cleaned up next garbage collection time

	return(LTP_OK);
}

/*!
 * @brief figure out when we've tx'd enough in this go
 * @param segs is the number of segments just queued
 * @param bytes is the number of bytes enqueued
 * @return true if enough queued already, otherwise false
 * @todo put in proper MIB-based calculations
 */
static bool ltp_sent_enough_for_now(int segs, int bytes)
{
	
#ifdef FIFTY
	// for now I assume 802.11b at 11mbs
	// we'll be full-duplex generally
	// want to send 50ms worth and no more in a go (dunno why, just because!)
	double bytespergo_d=(11*1204*1024)/8; // 1s worth of bytes
	bytespergo_d /= 20;
	int bytespergo=(int)floor(bytespergo_d);
	
	if (bytes >= bytespergo) return(true);
#endif
#define SEGBASED
#ifdef SEGBASED
	// segs based TIMING TUNING
	// pinger_fnc waits 12ms between calls to ping which gives enough time
	// for 100 1500 byte segments to be sent out @ 100 Mbps (Ethernet) or 120 us/MTU
	if (segs >= 100) return(true);
#endif
	return false;
}

/*!
 * @brief is this an inbound or outbound segment
 * @param type is the segment type
 * @return true for inbound segments
 */
static bool inbound_seg(ltp_segment_type type)
{
	switch(type) {
		case LTPC_SEG_DS:
		case LTPC_SEG_RA:
		case LTPC_SEG_CO:
		case LTPC_SEG_CAR:
			return true;
		case LTPC_SEG_RS:
		case LTPC_SEG_CR:
		case LTPC_SEG_CAO:
			return false;
	}
	snprintf(ltpstr,LTPSTR_LEN,"Weird - unknown segment type! (%d) - pretending its inbound\n",type);
	ltp_log(LTP_LOG_MAX,ltpstr);
	return(true);
}

/*!
 * @fn int ltp_ping(void)
 * @brief give the LTP layer a chance to do some work
 * @return zero or error
 *
 * This function allows the LTP layer a chance to see if it can tx or rx anything.
 * ok - depending on what state l2's in we check for outgoing and/or incoming data.
 * we don't explicitly mess with the l2 state (other than the above) since higher
 * layer code (scheduler/router) should do that.
 *
 */
int ltp_ping(void)
{
	// try to set l2 into our desired state
	int rv=fixl2ifnecessary();
	if (rv) LTPERR(rv);
	ltp_time_t	now=ltp_now();

	// initialise the read position first so we're ready to read incoming data
	static int oldpos=0;
	static bool rxqueuepositioninited=false;
	if (!rxqueuepositioninited) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Initialising rxqueueposition to %d (statically was %d)\n",
			rxqueue_p->position,oldpos);
		ltp_log(LTP_LOG_MAX,ltpstr);
#endif
		GRABQUTEX;
		oldpos=rxqueue_p->position;
		DROPQUTEX;
		rxqueuepositioninited=true;
	}

	if (l2state==LTP_L2S_ON_TX || l2state==LTP_L2S_ON_BOTH) { // transmitting
		now=ltp_now();
		// go find all data segments for sending and try send 'em
		// how to do that:
		// when a block is not finishedwith but has no segments then
		// we need to make segments
		// when a block is not finishedwith and has segments in the
		// LTP_PS_WAITING_SEND state (other states?) then its on the
		// list
		// when we know what's on the list we make a separate list
		// of those for possible later transmission using l2 calls
		int bytes_sent=0; /// remember how many we're queuring so we can stop when we ought
		int segs_sent=0; /// remember how may segments we're queueing so we can stop when we ought
		ltp_l2_buffer	l2buf; ///< just stuff for l2 api
		l2buf.heap=false;
		l2buf.l2alloc=false;
		int nlisteners=LTP_LISTENERS_PER_GO;
		ltpaddr	listeners[LTP_LISTENERS_PER_GO];
		int ltp_cue_rv=ltp_cue_whos_listening(now,&nlisteners,listeners);
		if (ltp_cue_rv) LTPERR(ltp_cue_rv);
		if (nlisteners) {
		
#undef WAITHALF
#ifdef WAITHALF
			{
				struct timespec	ts,ts1;
				memset(&ts,0,sizeof(ts));
				memset(&ts1,0,sizeof(ts));
				ts.tv_nsec=1000*1000*20;  // 20 ms (please)
				char 		befstr[MAXPATH],aftstr[MAXPATH];
				snprintf(befstr,MAXPATH,ltp_now_str());
				int rv=nanosleep(&ts,&ts1);
				snprintf(aftstr,MAXPATH,ltp_now_str());
				snprintf(ltpstr,LTPSTR_LEN,"waithalf/nanosleep at %d from %s to %s\n",__LINE__,befstr,aftstr);
				ltp_log(LTP_LOG_MAX,ltpstr);
			}
#endif

#ifdef LTPT_TRICKLE_FORWARD
			// first see what's what with ltp_t segments
			// this has to be done now since otherwise I might miss out on some data
			// that arrived when this guy wasn't listening
			if (ltp_state_var.t_cfg.on) {
				rv=ltpt_trickle_forward(nlisteners,listeners);
				if (rv) LTPERR(rv);
			}
#endif

			// see if there's anything for one of 'em
			int nsegments=ltp_state_var.txq.nsegments;
			ltp_segment *segments=ltp_state_var.txq.segments;
			bool done=false;
			while (!done) {
				// pick out next segment to transmit in the following order:
				// according to the listener 
				// with control segments 1st
				// and then data segments
				// NEW: If sock has SO_LINGER set (.linger) then don't send a cancel before data
				int firsti=0;
				bool firstdone=false; // set if we've seen a data seg to send to that listener
				bool ctrlseg=false; // set if we've seen a ctrol seg to send to that listerer
				// int i=0; // not really needed, but whatever
				done=true; // assume there're none left to send for a moment
				for (int j=0;!ctrlseg && j!=nlisteners;j++) {
					for (int i=0;!ctrlseg && i!=nsegments;i++) {
						if (segments[i].state==LTP_PS_WAITING_SEND // ready to be sent
							&&
							!ltpaddr_cmp(&segments[i].peer,
									&listeners[j],
									segments[i].peerlen)) {  // the other guy's listening
	
							done=false; // ahh - we're not really done!
							int stype=segments[i].seg.type;
							if (stype==LTPC_SEG_ENCODED) stype=segments[i].encoded_type;

							if (LTPC_CTRL_SEG(stype)) {
#ifdef LTP_VERBOSE
								if (firstdone) {
									snprintf(ltpstr,LTPSTR_LEN,"Going to send control segment index %d instead of data segment index %d\n",
											i,firsti);
								} else {
									snprintf(ltpstr,LTPSTR_LEN,"Going to send control segment index %d \n",i);
								}
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								ctrlseg=true;
								firsti=i;
							} else {
								if (!firstdone) {
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN,"First data segment to be sent is segment index %d \n",i);
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									firstdone=true;
									firsti=i;
								}
							}
						} 
					}
				}

				if (!done) {
					int i=firsti;

					// figure out how much space to alloc for the encoded segment.
					switch (segments[i].seg.type) {
								case LTPC_SEG_ENCODED:
									{
										l2buf.len=segments[i].encoded.len;
									}
									break;

								case LTPC_SEG_DS:
									{   // also have to fix the buffer ptr since we only store the block buffer once
										rv=ltp_fixparent(&ltp_state_var,(void**)&segments[i].parent,segments[i].parentid,LTP_PARENT_BLOCK);
										if (rv) LTPERR(rv);
										rv=ltp_fixparent(&ltp_state_var,(void**)&segments[i].parent->parent,segments[i].parent->parentid,LTP_PARENT_SOCKET);
										if (rv) LTPERR(rv);
										if (
											!segments[i].parent  // oops - orphaned
						   				) {
											LTPERR(LTP_MEMORY);
										}
										ltp_block *block=segments[i].parent;
										ltp_socket_t *sock=block->parent;
										int offset=segments[i].seg.val.ds.offset;
										int length=segments[i].seg.val.ds.bytes.len;
										if (
											!block->decoded.val // no buffer 
											||
											block->decoded.len < offset+length // not enough data!
						   				) {
											LTPERR(LTP_MEMORY);
										}
										// overestimate
										l2buf.len=latestmtu(block,sock)+LTPC_DS_OVERHEAD;
										// pass segment info to codec
										segments[i].seg.val.ds.bytes.val=segments[i].parent->decoded.val+offset;
									}
									break;

								case LTPC_SEG_RS: l2buf.len=LTPC_RS_OVERHEAD+segments[i].seg.val.rs.cc*(2*10); break;
								case LTPC_SEG_RA: 
												  l2buf.len=LTPC_RA_OVERHEAD; 
												  break;
								case LTPC_SEG_CO: l2buf.len=LTPC_CO_OVERHEAD; break;
								case LTPC_SEG_CAO: l2buf.len=LTPC_CAO_OVERHEAD; break;
								case LTPC_SEG_CR: l2buf.len=LTPC_CR_OVERHEAD; break;
								case LTPC_SEG_CAR: l2buf.len=LTPC_CAR_OVERHEAD; break;
								default:
									LTPERR(LTP_BAD_VALUE);
					}
					// do the encode, log some stuff and send to l2

#ifdef LTP_VERBOSE
					const char *segtypestr[]={
									" Data",
									" Report",
									" Report ACK",
									" Cancel(O)",
									" Cancel ack(O)",
									" Cancel(R)",
									" Cancel ack(R)",
									"n Encoded"};
					snprintf(ltpstr,LTPSTR_LEN, "PP: About to encode a%s segment for transmission to LTP[id:%s,port:%d]",
															segtypestr[segments[i].seg.type-LTPC_SEG_DS],
															inet_ntoa(segments[i].peer.sock.sin_addr),
															segments[i].peer.sock.sin_port);
					ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					// check whether any extensions need to be added to the segment
					bool retx=old_timer(segments[i].t_entry);
					if (!retx) {
						int extensions_len=0;
						rv=ltp_add_exts_before_encode(&segments[i],&extensions_len); // maybe need more inputs later 
						if (rv) LTPERR(rv);
						l2buf.len += extensions_len;
					}
					if (segments[i].seg.type!=LTPC_SEG_ENCODED) {
						l2buf.val=(unsigned char*)calloc(l2buf.len,1);
						if (!l2buf.val) LTPERR(LTP_MEMORY);
						rv=ltpc_encode(&segments[i].seg, (int*)&l2buf.len, l2buf.val, true);
						if (rv) LTPERR(rv);
					} else {
						l2buf.val=segments[i].encoded.val;
					}
#ifdef LTP_VERBOSE
#undef LTP_LOG_DS_CONTENT
#ifdef LTP_LOG_DS_CONTENT
					snprintf(ltpstr,LTPSTR_LEN,"%s segment id=%d,ind=%d for blockid %d\n",
							segtypestr[segments[i].seg.type-LTPC_SEG_DS],segments[i].id,i,segments[i].parentid);
					ltp_log_buffer(ltpstr,l2buf.len,l2buf.val);
#endif
					// and a nice decoded version
					char *buffer;
					int blen=10*MAXPATH;
					buffer=(char*)calloc(1,blen);
					if (!buffer) LTPERR(LTP_MEMORY);
					rv=ltpc_pp(l2buf.val,l2buf.len,buffer,&blen);
					if (rv) LTPERR(rv);
					
					ltp_log(LTP_LOG_MAX,buffer);
					free(buffer);
#endif

					ltp_l2_id l2id; 
					l2id.heap=false;
					l2id.l2alloc=false;
					l2id.addr_len=0;
					l2id.addr=segments[i].peer;
					l2id.addr_len=segments[i].peerlen;
					l2id.pname=NULL;
					// pass encoded bytes to layer 2
					rv=ltp_l2_send(&l2buf,&l2id);
					if (rv) LTPERR(rv);

					rv=ltp_fixparent(&ltp_state_var,(void**)&segments[i].parent,segments[i].parentid,LTP_PARENT_BLOCK);
					if (rv) LTPERR(rv);
					rv=ltp_fixparent(&ltp_state_var,(void**)&segments[i].parent->parent,segments[i].parent->parentid,LTP_PARENT_SOCKET);
					if (rv) LTPERR(rv);
					// if (!segments[i].parent->parent->ltptrelay) {
#define SLOWLY
#ifdef SLOWLY
					// only do rate control if I'm the source
					// oops! that's wrong, if I'm a router and in schedule mode, then essentially
					// I'm re-inserting all the packets at once and so have lost whatever rate 
					// control hapened before
					{
						static double last_time=0;
						// odd doesn't seem to work this 
						struct timeval lnow;
						gettimeofday(&lnow,NULL);
						double dnow=lnow.tv_sec*1000+lnow.tv_usec/1000;
						// struct timespec ts;
						// clock_gettime(CLOCK_REALTIME,&ts);
						// double dnow=ts.tv_sec*1000+ts.tv_nsec/1000000;
						double gap=dnow-last_time;
						// sending more quickly results in losts UDP packets
						// try a shorter gap, used to be 20ms
						if (gap == 0 ) {
							struct timespec	ts,ts1;
							memset(&ts,0,sizeof(ts));
							memset(&ts1,0,sizeof(ts));
							// 20ms seems to be the quickest I can get on Ubuntu regardless of what I ask for
							int notneeded=(int)(floor(gap));
							if (notneeded>20) notneeded=0;
							if (notneeded<0) notneeded=0;
							ts.tv_nsec=1000*1000*(20-notneeded);
							char 		befstr[MAXPATH],aftstr[MAXPATH];
							snprintf(befstr,MAXPATH,"%s",ltp_now_str());
							int rv=nanosleep(&ts,&ts1);
							snprintf(aftstr,MAXPATH,"%s",ltp_now_str());
							snprintf(ltpstr,LTPSTR_LEN,"slowly/nanosleep at %d from %s to %s (gap: %g ms)\n",
								__LINE__,befstr,aftstr,gap);
							ltp_log(LTP_LOG_MAX,ltpstr);
						} else {
							snprintf(ltpstr,LTPSTR_LEN,"no-slowly/nanosleep needed at %d time %s (gap: %g ms)\n",
								__LINE__,ltp_now_str(),gap);
							ltp_log(LTP_LOG_MAX,ltpstr);
						}
						// clock_gettime(CLOCK_REALTIME,&ts);
						// last_time=ts.tv_sec*1000+ts.tv_nsec/1000000;
						gettimeofday(&lnow,NULL);
						last_time=lnow.tv_sec*1000+lnow.tv_usec/1000;
					}
#endif
#undef QUICKLY
#ifdef QUICKLY
						ltp_microsleep(120);
#endif
					// }

					// start the associated timer, if any
					if (segments[i].t_entry!=LTP_NOTIMER) {
						rv=reinit_timer(segments[i].t_entry);
						if (rv) LTPERR(rv);
					}

					// restart the socket's inactivity timer
					rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)segments[i].parent->parent,true);
					if (rv) LTPERR(rv);

					// accumulate bytes_sent counter
					bytes_sent+=l2buf.len;
					segs_sent++;

					// if necessary remember those bytes in case of request replay
					if (segments[i].r_entry!=LTP_NOCACHE) {  // its cacheable
#ifdef LTP_VERBOSE
						if (segments[i].r_entry==LTP_CACHE) {
							snprintf(ltpstr,LTPSTR_LEN,"Caching segment %d of type %d as new entry (len=%d)\n",
													i,segments[i].seg.type,l2buf.len);
						} else {
							snprintf(ltpstr,LTPSTR_LEN,"Caching segment %d of type %d as entry %d (len=%d)\n",
											i,segments[i].seg.type,segments[i].r_entry,l2buf.len);
						}
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						// using rv here was bad - don't do that! need it below
						int rrv=add_replay_response(
								&segments[i],
								l2buf.val,
								l2buf.len);
						if (rrv) LTPERR(rrv);
					}

					if (segments[i].seg.type!=LTPC_SEG_ENCODED) {
						free(l2buf.val);
					}

					// update segment state
					segments[i].state=LTP_PS_SENT;
					LTP_STATE_CHANGED;

					// if it was a DS then we can NULL out the bytes ptr which reduces the crap
					// we have to fixup after reading back from disk
					if (segments[i].seg.type==LTPC_SEG_DS) {
						segments[i].seg.val.ds.bytes.val=NULL;
					}

					// we kill an inbound session is if we've just sent a CAO
					if (segments[i].seg.type==LTPC_SEG_CAO) {
						rv=ltp_fixparent(&ltp_state_var,(void**)&segments[i].parent,segments[i].parentid,LTP_PARENT_BLOCK);
						if (rv) LTPERR(rv);
						ltp_block *par=segments[i].parent;
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",par->id,__LINE__);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						par->finishedwith=true;
					}

					// if we're sending a CAR then that's the end of an outbound session
					if (segments[i].seg.type==LTPC_SEG_CAR) {
						rv=ltp_fixparent(&ltp_state_var,(void**)&segments[i].parent,segments[i].parentid,LTP_PARENT_BLOCK);
						if (rv) LTPERR(rv);
						ltp_block *par=segments[i].parent;
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",par->id,__LINE__);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						par->finishedwith=true;
					}

					// ok, just check as to whether we've just sent the last required RA for an outbound session
					if (segments[i].seg.type==LTPC_SEG_RA 
						|| 
						(segments[i].seg.type==LTPC_SEG_ENCODED && segments[i].encoded_type==LTPC_SEG_RA)) {
						rv=ltp_fixparent(&ltp_state_var,(void**)&segments[i].parent,segments[i].parentid,LTP_PARENT_BLOCK);
						if (rv) LTPERR(rv);
						ltp_block *par=segments[i].parent;
						rv=ltp_fixparent(&ltp_state_var,(void**)&par->parent,par->parentid,LTP_PARENT_SOCKET);
						if (rv) LTPERR(rv);
						if (par->parent->opts.redlen > 0 &&
							par->tx_scopes.nscopes>=1 && 
							par->tx_scopes.lbs[0]==0 &&
							par->tx_scopes.ubs[0] >= par->parent->opts.redlen) {
#ifdef LTP_VERBOSE
							snprintf(ltpstr,LTPSTR_LEN,"Marking redcomplete for block (%d) (ubs=%d,redlen=%d)\n",
								par->id,par->tx_scopes.ubs[0],par->parent->opts.redlen);
							ltp_log(LTP_LOG_MAX,ltpstr);
#endif
							par->redcomplete=true;
						}
						// if its complete and has been delivered then that's that.
						if (par->complete) {
#ifdef LTP_VERBOSE
							snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",par->id,__LINE__);
							ltp_log(LTP_LOG_MAX,ltpstr);
#endif
							par->finishedwith=true;
						}
					}

					// if that's the last of an all green block then we're complete
					if (segments[i].parent->parent->opts.redlen==LTP_ALLGREEN 
						&& segments[i].seg.type==LTPC_SEG_DS
						&& segments[i].seg.val.ds.eob) {

#ifdef MUDSAT
							// SF MUDSAT testing: see what happens if I just set all the segments of this
							// all green block to indicate they were not sent, but just once
							if (segments[i].parent->mudsat_re_tx_ctr>0) {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"MUDSAT Re-TXing the block\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								int rtrv=ltp_green_re_tx(&segments[i]);
								if (rtrv) LTPERR(rtrv);
								segments[i].parent->mudsat_re_tx_ctr--;
							} else {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"MUDSAT NO longer Re-TXing the block\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
							    segments[i].parent->complete=true;
                            } 
#endif
							
					}
					// if that's the last regardless and the redpart is done then we're all done
					if (segments[i].seg.type==LTPC_SEG_DS
						&& segments[i].seg.val.ds.eob) {
							segments[i].parent->eobsent=true;
					}
					if (segments[i].parent->redcomplete &&
						segments[i].parent->eobsent) {
						segments[i].parent->complete=true;
					}
					if (segments[i].parent->complete) {
						bool 		ltpt_seg=is_ltpt_seg(&segments[i].seg);
						if (ltpt_seg) {
							bool		ltpt_forme=is_ltpt_forme(&segments[i].seg); // if I'm the final dest, or its not an LTP-T seg
							if (ltpt_forme) {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"Setting finishedwith for block %d (its for me)\n",segments[i].parentid);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								segments[i].parent->finishedwith=true;
							} else if (is_ltpt_fromme(&segments[i].seg)) { // if I'm the source
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"Setting finishedwith for block %d (I'm the source)\n",segments[i].parentid);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								segments[i].parent->finishedwith=true;
							} else {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"NOT Setting finishedwith for block %d (not for me)\n",segments[i].parentid);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
							}
						} else {
							if (segments[i].seg.type==LTPC_SEG_RA) {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"Setting finishedwith for block %d (plain LTP, last RA sent)\n",segments[i].parentid);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								segments[i].parent->finishedwith=true;
							}
						}
					}

					// check l2state is still good for this loop
					l2state=ltp_l2_get_state();
					if (l2state!=LTP_L2S_ON_TX && l2state!=LTP_L2S_ON_BOTH) { // finished transmitting
						done=true;
					}
				}
				if (ltp_sent_enough_for_now(segs_sent,bytes_sent)) {
#ifdef LTP_VERBOSE
					snprintf(ltpstr,LTPSTR_LEN,"ping() sent enough already!\n");
					ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					done=true;
				}
			}
		}
#ifdef LTP_VERBOSE
		if (bytes_sent) {
			snprintf(ltpstr,LTPSTR_LEN,"One call to ping() sent %d segments totalling %d bytes (elapsed time: %d)\n",
				segs_sent,bytes_sent,ltp_elapsed());
			ltp_log(LTP_LOG_MAX,ltpstr);
		}
#endif
	}

#ifdef LTP_EXTRA_VERBOSE
	STARTTIMEDTRACEBLOCK(x1,true)

	static int lastl2state=-1;
	if (l2state!=lastl2state) {
		char *statestr[]={
			"LTP_L2S_RESET",
			"LTP_L2S_INITED",
			"LTP_L2S_ON_TX",
			"LTP_L2S_ON_RX",
			"LTP_L2S_ON_BOTH",
			"LTP_L2S_SLEEP1",
			"LTP_L2S_SLEEP2",
			"LTP_L2S_SLEEP3",
			"LTP_L2S_SLEEP4",
			"LTP_L2S_SLEEP5"};
		snprintf(ltpstr,LTPSTR_LEN,"We're now (elapsed %d) in %s state (%d)\n",ltp_elapsed(),statestr[l2state],l2state);
		ltp_log(LTP_LOG_MAX,ltpstr);
		lastl2state=l2state;
	}

	ENDTIMEDTRACEBLOCK(x1)
#endif

	if (l2state==LTP_L2S_ON_RX || l2state==LTP_L2S_ON_BOTH) { // receiving
		now=ltp_now();
		// do some reading of data (hopefully segments)
		// and (depending on handling) if all segments are present, then
		// mark the block as complete
		GRABQUTEX;
		int newpos=rxqueue_p->position;
		int posdiff=newpos-oldpos;
		if (posdiff<0) posdiff+=rxqueue_p->size;
#ifdef LTP_VERBOSE
		if (newpos!=oldpos) {
				char *l2peeraddr=inet_ntoa(rxqueue_p->peers[oldpos].addr.sock.sin_addr);
				int l2peerport=rxqueue_p->peers[oldpos].addr.sock.sin_port;
				snprintf(ltpstr,LTPSTR_LEN,"Rx'd %d buffers, 1st of which is %d bytes, from LTP[id:%s,port:%d]\n",
							posdiff,rxqueue_p->bufs[oldpos].len,l2peeraddr,l2peerport);
				ltp_log(LTP_LOG_MAX,ltpstr);
		}
#endif
#ifdef LTP_EXTRA_VERBOSE
	STARTTIMEDTRACEBLOCK(x2,true)
		if (rxqueue_p->position!=oldpos) {
			snprintf(ltpstr,LTPSTR_LEN,"Rx state and pos is %d (was %d)\n",
				rxqueue_p->position,oldpos);
			ltp_log(LTP_LOG_MAX,ltpstr);
			ltp_log_flush();
		}
	ENDTIMEDTRACEBLOCK(x2)
#endif
		DROPQUTEX;

		if (newpos!=oldpos) { // stuff's arrived!! (or else we missed an entire queue-full of blocks)

			// now let's try to decode that data!
			
			// We may have gotten >1 buffer from layer 2 due to lower layer fragmentation or 
			// the fact that we weren't listening. In order to be more robust
			// against lower layer fragmentation we catenate all those buffers that are
			// from the same peer and handle them as one 

			int *alllen,*l2peers;
			alllen=(int*)calloc(1,sizeof(int)*(posdiff+1));
			if (!alllen) LTPERR(LTP_MEMORY);
			l2peers=(int*)calloc(1,sizeof(int)*(posdiff+1));
			if (!l2peers) LTPERR(LTP_MEMORY);
			int npeers=0; //< the number of actual peers

			int qsize=rxqueue_p->size;

			GRABQUTEX
			// size things up
			for (int i=oldpos;i!=newpos;i=(i+1)%qsize) { 
				int *l2p=l2peers;
				bool done=false;
				while (!done && *l2p) {
					if ((rxqueue_p->peers[i].addr_len==rxqueue_p->peers[*l2p-1].addr_len) && 
						!ltpaddr_cmp(&rxqueue_p->peers[i].addr,
									&rxqueue_p->peers[*l2p-1].addr,
									rxqueue_p->peers[i].addr_len)) {
						done=true;
					} else {
						l2p++;
					}
				}
				if (!done || !*l2p) {
					npeers++;
					*l2p=i+1;
				}
#ifdef LTP_EXTRA_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"i: %d (q: %d, o: %d n: %d), offset: %td, peer: %d (of %d)\n",
								i,qsize,oldpos,newpos,l2p-l2peers,*l2p,npeers);
				ltp_log(LTP_LOG_MAX,ltpstr);
#endif
				int bilen=rxqueue_p->bufs[i].len; 
				alllen[l2p-l2peers]+=bilen;
			}

			// now do catenations and then decodes, one l2-peer at a time
			for (int pp=0;pp!=npeers;pp++) {
				unsigned char *allbuf,*nextseg;
				if (alllen[pp]<=0) {
					// weird shit, what to do?
					int forcelen=8*MAXPATH;
					snprintf(ltpstr,LTPSTR_LEN,"Error: Oddness at %d: alllen[%d] is %d - forcing to %d\n",__LINE__,pp,alllen[pp],forcelen);
					ltp_log(LTP_LOG_MAX,ltpstr);
					allbuf=(unsigned char*)calloc(1,forcelen);
				} else {
					allbuf=(unsigned char*)calloc(1,alllen[pp]+100);
				}
				if (!allbuf) LTPERR(LTP_MEMORY);
				nextseg=allbuf;
				/// remember the edges of the l2 buffers, in case we need to recover from garbage
				unsigned char *l2endpoints[posdiff+1]; 
				unsigned char **l2ep=l2endpoints;
				for (int i=0;i!=posdiff+1;i++) { l2endpoints[i]=NULL;}

				// these guys are for handling detected replays
				int r_entry=LTP_NOCACHE;
				bool r_responded=false;

				int pid=l2peers[pp]-1;
				int pplen=rxqueue_p->peers[pid].addr_len;
				// catenate up relevant buffers for peer 'pp'
				for (int i=oldpos;i!=newpos;i=(i+1)%qsize) {
					// if its for this peer
					int pilen=rxqueue_p->peers[i].addr_len;
					ltpaddr *pia=&rxqueue_p->peers[i].addr;
					ltpaddr *ppa=&rxqueue_p->peers[pid].addr;
					if ((pilen==pplen && 
						!ltpaddr_cmp(pia,ppa,pilen))) {
						if (is_replay(
								rxqueue_p->bufs[i].val,
								rxqueue_p->bufs[i].len,
								false,
								&r_entry,
								&r_responded)) {
#ifdef LTP_VERBOSE
							snprintf(ltpstr,LTPSTR_LEN,"Buffer %d IS A REPLAY!!!\n",i);
							ltp_log(LTP_LOG_ERR,ltpstr);
#endif
							if (r_responded) {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"There is a replay response for that too! We'll re-send it.\n");
								ltp_log(LTP_LOG_ERR,ltpstr);
#endif
								int rv=resend_from_replay(
											r_entry,
											&rxqueue_p->peers[i].addr,
											rxqueue_p->peers[i].addr_len);
								if (rv) { free(allbuf); DROPQUTEX; LTPERR(rv); }
							}
						} else { 
							*l2ep++=nextseg; 
							memcpy(nextseg,rxqueue_p->bufs[i].val,rxqueue_p->bufs[i].len);
							nextseg+=rxqueue_p->bufs[i].len;
						}
						// mark that buffer as no longer dirty
						rxqueue_p->bufs[i].dirty=false;
					}
				}

				// we have paddr from the segment, but l2peeraddr from layer 2
				// TODO which to use for what?
				// we prefer the Layer2 reporting of the peer name for talking-back purposes
				ltpaddr paddr;
				// paddr=rxqueue_p->peers[l2peers[pp]-1].addr; // struct copy
				paddr=rxqueue_p->peers[pid].addr; // struct copy

				if (nextseg!=allbuf) { // something there to decode
					l2ep=l2endpoints;
					nextseg=allbuf;
					int currbuflen=alllen[pp];
					int currlen=0;
					unsigned char *currseg;
					int bufnum=0;
					ltpc_seg segstr;
					// ltpaddr	l2addr=rxqueue_p->peers[l2peers[pp]-1].addr;
					// int		l2addrlen=rxqueue_p->peers[l2peers[pp]-1].addr_len;
					while	(
								nextseg!=NULL  // haven't run out of bytes
								&&
								currbuflen>0 // something happened
						) {
						currseg=nextseg;
						currlen=currbuflen;
						memset(&segstr,0,sizeof(ltpc_seg));

#define PP_INCOMING
#ifdef PP_INCOMING
						// and a nice decoded version
						snprintf(ltpstr,LTPSTR_LEN,"Pretty Printing received buffer before decode\n");
						ltp_log(LTP_LOG_MAX,ltpstr);
						char buffer[10*MAXPATH];
						int blen=10*MAXPATH;
						memset(buffer,0,blen);
						rv=ltpc_pp(currseg,currbuflen,buffer,&blen);
						ltp_log(LTP_LOG_MAX,buffer);
#endif
						int packet_type=ltpc_decode(&currbuflen,&nextseg,&segstr);
						// if (ISLTPERR(packet_type)) { LTPERR(packet_type); }
						currlen-=currbuflen;
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"Decoded peer %d's buffer %d of (total) %d, type is %d, length is %d (bytes left: %d)\n",
								pp,bufnum,posdiff,packet_type,currlen,currbuflen);
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						bufnum++;
						if (currbuflen < 0 ) { // oops truncation!
							packet_type=LTPC_SEG_TRUNC;
						}
						// update our view of the layer 2 endpoints.
						l2ep++;
						if (LTPC_ERROR(packet_type)) {
#ifdef LTP_VERBOSE
							snprintf(ltpstr,LTPSTR_LEN,"Buffer %d isn't a real segment!!!!\n",bufnum-1);
							ltp_log(LTP_LOG_ERR,ltpstr);
							snprintf(ltpstr,LTPSTR_LEN,"Dodgy buffer is:\n");
							ltp_log_buffer(ltpstr,currlen,currseg);
#endif
							// throw it away
							nextseg=*l2ep;
							if (nextseg) currbuflen=alllen[pp]-(nextseg-allbuf);
							if (bufnum>posdiff) {
								// we're done
								currbuflen=0; // to trip up the while loop
								// we're leaking the bytes here?
							} else {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"Now looking at l2 buffer %td of %d\n",l2ep-l2endpoints,posdiff);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
							}
						} else {
	
							// might not actually be a ds, but the cast works
							// (maybe make a separate header part of the struct later on - not sure)
							// doing just that now!
							ltpc_ds *ds=&segstr.val.ds;
							// now go off put this data into the right block etc.
							// find the socket with that peer and session id
							int plen=segstr.peer.len;
							unsigned char *pval=segstr.peer.val;

							// SF: CHANGE IN Philly interop
							bool hackforscott=false;
							if (forced_addr_set) {
								// SF: CHANGE IN Philly interop
								char *s1=strdup(ltpaddr2str(&paddr));
								char *s2=strdup(ltpaddr2str(&forced_addr));
								snprintf(ltpstr,LTPSTR_LEN,"Layer 2 I'm being forced to use %s as the return (and not %s)\n",s2,s1);
								ltp_log(LTP_LOG_MAX,ltpstr);
								paddr=forced_addr; // struct copy
								free(s1); free(s2);
							} else {
								// but may as well note when that differs from the session ID
								// maybe a NAT box there, eh?
								ltpaddr naddr;
								rv=octets2ltpaddr(&segstr.peer,&naddr);
								if (rv) { free(allbuf); LTPERR(rv); }
								if (ltpaddr_cmp(&naddr,&paddr,sizeof(paddr))) {
									if (ion_mode) {
										short engid=htons(my_addr.sock.sin_port);
										if (segstr.peer.len==2 && !memcmp(segstr.peer.val,&engid,2)) {
										hackforscott=true;
										snprintf(ltpstr,LTPSTR_LEN,"Layer 2 seems to be an inbound port, probably from ION");
										ltp_log(LTP_LOG_MAX,ltpstr);
										paddr=my_addr; // struct copy
										}
									}
									if (!hackforscott) {
										char *s1=strdup(ltpaddr2str(&paddr));
										char *s2=strdup(ltpaddr2str(&naddr));
										if (naddr.octets.magic==LTP_ADDR_OCT_MAGIC) {
											snprintf(ltpstr,LTPSTR_LEN,"Layer 2 (%s) and engine ID (%s) variant peer addresses differ! - I prefer layer 2\n",s1,s2);
											ltp_log(LTP_LOG_MAX,ltpstr);
											hackforscott=true;
										} else {
											snprintf(ltpstr,LTPSTR_LEN,"Layer 2 (%s) and engine ID (%s) variant peer addresses differ! - I prefer engine ID\n",s1,s2);
											ltp_log(LTP_LOG_MAX,ltpstr);
											paddr=naddr; // struct copy
										}
										free(s1); free(s2);
									}
								}
							}

							ltp_socket_t *thesock=NULL;
							ltp_socket_t *emptylistener=NULL;
							ltp_socket_t *emptyrelay=NULL;
							ltp_block *theblock=NULL;
							bool		ditched_seg=false;
							bool 		ltpt_seg=is_ltpt_seg(&segstr);
							bool		ltpt_forme=(ltpt_seg?is_ltpt_forme(&segstr):false); // if I'm the final dest
							for (int i=0;!thesock && i!=ltp_state_var.nsockets;i++) {
								ltp_socket_t *socksess=&ltp_state_var.sockets[i];
								if (socksess->finishedwith) continue;
								// note if we've an unbound listening socket in case
								// we find nothing else
								if (!emptylistener && !socksess->ibpeerlen && !socksess->ltptrelay) {
								 	emptylistener=socksess;
								}
								if (!ltpt_forme && ltpt_seg && !emptyrelay && !socksess->ibpeerlen && socksess->ltptrelay) {
									emptyrelay=socksess;
								}
								// this may be for an existing inbound socket
								if (!thesock && inbound_seg(segstr.type) && socksess->inbound && !ltpaddr_cmp(&paddr,&socksess->ibpeer,sizeof(paddr))) {
									thesock=socksess;
									rv=get_block_from_sock(socksess, &theblock,&segstr.sno);
									if (rv) { free(allbuf); LTPERR(rv); }
								}

								// this may be for an existing outbound socket
								if (!thesock && !inbound_seg(segstr.type) && socksess->outbound && !ltpaddr_cmp(&paddr,&socksess->ibpeer,sizeof(paddr))) {
									rv=get_block_from_sock(socksess, &theblock,&segstr.sno);
									if (rv) { free(allbuf); LTPERR(rv); }
									if (theblock) { // will return as NULL if not the right socket
										thesock=socksess;
									}
								}
								// this may be for an existing ION outbound socket
								if (ion_mode && !thesock && !inbound_seg(segstr.type) && socksess->outbound && hackforscott) {
									if (!ltpaddr_cmp(&paddr,&socksess->ibpeer,sizeof(paddr))) {
										rv=get_block_from_sock(socksess, &theblock,&segstr.sno);
										if (rv) { free(allbuf); LTPERR(rv); }
										if (theblock) { // will return as NULL if not the right socket
											thesock=socksess;
										}
									} else {
										ltpaddr taddr=paddr;
										taddr.sock.sin_port=socksess->obpeer.sock.sin_port;
										taddr.sock.sin_family=socksess->obpeer.sock.sin_family;
										rv=ltpaddr_cmp(&taddr,&socksess->obpeer,sizeof(taddr));
										if (!rv) {
											rv=get_block_from_sock(socksess, &theblock,&segstr.sno);
											if (rv) { free(allbuf); LTPERR(rv); }
											if (theblock) { // will return as NULL if not the right socket
												thesock=socksess;
												paddr=taddr;
											}
										} 
									}
								}

								// new case for LTP-T sesssions - this may be dodgy but is worth a try
								if (!thesock && socksess->ltptrelay && !inbound_seg(segstr.type) ) {
									bool donehere=false;
									for (int j=0;!donehere && j!=socksess->nblocks;j++) {	
										rv=get_block_from_sock(socksess, &theblock,&segstr.sno);
										if (rv) { free(allbuf); LTPERR(rv); }
										if (theblock) { // will return as NULL if not the right socket
											thesock=socksess;
											donehere=true;
										}
									}
								}
							}
							if (!thesock) {
								if (segstr.type==LTPC_SEG_DS && (!ltpt_seg && emptylistener)  || (ltpt_seg && ltpt_forme && emptylistener)) {
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN, "Filling in emptylistener (%d for %s)\n",
										emptylistener->id,ltpaddr2str(&paddr));
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									thesock=emptylistener;
									/// @bug A possible bug. I'm only associating the layer2 peer with a socket 
									/// @bug the 1st time I get something for that socket. The alternative is
									/// @bug to do that afresh each time, but then I've to worry about hijacking
									/// @bug of sessions to a bad l2 address. Not sure what's right there really.
									thesock->ibpeer=paddr;
									thesock->ibpeerlen=sizeof(paddr);
									
									// create a new inactivity timer for this guy
									rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)thesock,true);
									if (rv) { free(allbuf); LTPERR(rv); }
								} else if (segstr.type==LTPC_SEG_DS && !ltpt_forme && ltpt_seg && emptyrelay) {
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN, "LTP-T segment found - filling in emptyrelay (%d for %s)\n",
										emptyrelay->id,ltpaddr2str(&paddr));
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									thesock=emptyrelay;
									/// @bug A possible bug. I'm only associating the layer2 peer with a socket 
									/// @bug the 1st time I get something for that socket. The alternative is
									/// @bug to do that afresh each time, but then I've to worry about hijacking
									/// @bug of sessions to a bad l2 address. Not sure what's right there really.
									thesock->ibpeer=paddr;
									thesock->ibpeerlen=sizeof(paddr);
									
									// create a new inactivity timer for this guy
									rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)thesock,true);
									if (rv) { free(allbuf); LTPERR(rv); }
								} else {
									// ditch segment and break
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN, "Can't find matching or free socket for %s\n",ltpaddr2str(&paddr));
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									ditched_seg=true;
								}
							} else {
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN, "Socket found - (%d for %s)\n",
										thesock->id,ltpaddr2str(&paddr));
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								// re-init socket's inact. timer
								rv=init_inactivity_timer(LTP_PARENT_SOCKET,(void*)thesock,true);
								if (rv) { free(allbuf); LTPERR(rv); }
							}

							if (!theblock && segstr.type!=LTPC_SEG_DS) {
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,
											"Can't find matching or free block (aka session)\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								ditched_seg=true;
							}

							// we may need to create a new block if none exists
							// that's needed for proper extension handling so do it now if we must
							// can only really happen if we get the 1st DS
							if (!ditched_seg && segstr.type==LTPC_SEG_DS && !theblock) {
								// create a new one!
								thesock->inbound=true;
								rv=ltp_add_block_from_seg(thesock,&segstr,&theblock);
								if (rv) { free(allbuf); LTPERR(rv); }
							}

							// lumps of extension handling code
							if (!ditched_seg) rv=cookie_processing(&segstr, theblock, thesock, &ditched_seg);
							if (rv) { free(allbuf); LTPERR(rv); }
							if (!ditched_seg) rv=auth_processing(&segstr, theblock, thesock, &ditched_seg, currlen);
							if (rv) { free(allbuf); LTPERR(rv); }
							if (ltpt_seg && !ditched_seg) rv=ltpt_inbound_processing(&segstr, theblock, thesock, &ditched_seg, currlen);
							if (rv) { free(allbuf); LTPERR(rv); }

							if (!ditched_seg) {
								switch (segstr.type) {
	
								case LTPC_SEG_DS:
								{
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN, "MSG: Got a DS\n");
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									thesock->inbound=true; // by definition since we just got a segment!
									if (!theblock) {
										// this code would be wrong, but I don't think I can get here any longer since 
										// this is called above already!
										// create a new one!
										rv=ltp_add_block_from_seg(thesock,&segstr,&theblock);
										if (rv) { free(allbuf); LTPERR(rv); }
									} else {
										// update the block's buffer
										int currlen=theblock->decoded.len;
										int offset=ds->offset;
										int len=ds->bytes.len;
										if (offset+len>currlen) { // need more space
											unsigned char *foo=(unsigned char*)calloc(1,offset+len);
											if (!foo) LTPERR(LTP_MEMORY);
											// copy in original block
											memcpy(foo,theblock->decoded.val,theblock->decoded.len);
											// and add new segment data
											// note: last segment in overwrites if bytes overlap! fine that though
											memcpy(foo+offset,ds->bytes.val,len);
											free(theblock->decoded.val);
											theblock->decoded.val=foo;
											theblock->decoded.len=offset+len;
										} else { // just copy
											memcpy(theblock->decoded.val+offset,ds->bytes.val,len);
										}
										rv=addnewscope(theblock,offset,offset+len,LTP_RX_SCOPES);
										if (rv) {
											snprintf(ltpstr,LTPSTR_LEN,"addnewscope returned %d - ignoring rest of this RS\n",rv);
											ltp_log(LTP_LOG_MAX,ltpstr);
											continue;
										}
										// if (rv) LTPERR(rv);

										if (ds->eob) {
											theblock->eob_seen=true;
										}
										if (theblock->eob_seen && complete_scope(theblock)) { // all red or all rx'd 
											theblock->complete=true;
										}
										if (theblock->eob_seen && theblock->redcomplete) { // partly red and got all the red bytes
											theblock->complete=true;
										}
										if (theblock->eob_seen && (theblock->redtide==0 || theblock->redtide==LTP_REDTIDE_INIT)) { // all green
#ifdef MUDSAT
											// SF: MUDSAT - with our 3 x Tx idea we can wait a bit
											// or... forever ;-) TODO: add a timer
											if (complete_scope(theblock)) {
												theblock->complete=true;
											}
#else
											theblock->complete=true;
#endif
											// so it'll be finishedwithwhendelivered if its for me
										}

									}
									// is it red?
									if (ds->red) {
										if (theblock->redtide < ds->offset+ds->bytes.len) {
											// extend red len
#ifdef LTP_VERBOSE
											snprintf(ltpstr,LTPSTR_LEN,"Extending red length of block id %d, (%d -> %d)\n",
																theblock->id,theblock->parent->opts.redlen,ds->offset+ds->bytes.len);
											ltp_log(LTP_LOG_MAX,ltpstr);
#endif
											theblock->redtide=ds->offset+ds->bytes.len;
										}
									} else { // green
										if (theblock->redtide==LTP_REDTIDE_INIT) {
											theblock->redtide=0;
										}
									}
									// handle checkpointing
									if (ds->checkpoint) {

										if (ds->eorp) {
											// if we've got all the red stuff, then we're good to go (maybe, 
											// should we wait for the RA?)
											if (theblock->rx_scopes.nscopes > 0 && 
												theblock->rx_scopes.lbs[0]==0 &&
												theblock->rx_scopes.ubs[0] >= ds->offset ) {
												theblock->redcomplete=true;
												theblock->eorp_seen=true;
											}
										}

										// add one (or maybe more) new segment for tx

										rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
										if (rv) { free(allbuf); LTPERR(rv); }
	
										int rses=0;
										rv=make_rses_from_block(theblock,&segstr,&rses);
										if (rv) { free(allbuf); LTPERR(rv); }

#ifdef LTP_VERBOSE
										ltpaddr to=theblock->parent->ibpeer;
										snprintf(ltpstr,LTPSTR_LEN,
												"MSG: Queueing %d RSes for LTP[id:%s,port:%d]\n",
												rses,inet_ntoa(to.sock.sin_addr),to.sock.sin_port);
										ltp_log(LTP_LOG_MAX,ltpstr);
#endif
	
									}

									// TODO: May be buggy! I'm still not setting finishedwith for green in a router
									if (theblock->eob_seen && 
										(theblock->redtide==0 || theblock->redtide==LTP_REDTIDE_INIT)) {
#ifdef LTP_VERBOSE
										snprintf(ltpstr,LTPSTR_LEN,"Setting finishwithwhendelivered=true for %d at %d\n",theblock->id,__LINE__);
										ltp_log(LTP_LOG_MAX,ltpstr);
#endif
										theblock->finishedwithwhendelivered=true;
									}
								}
								break;
	
								case LTPC_SEG_RS:
								{
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN, "MSG: Got an RS\n");
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									// prepare & send an RA & relevant DS's

									// find the relevant block
									ltpc_rs *rs=&segstr.val.rs;

									if (rs->ub > theblock->decoded.len) { // so long sucker!
#ifdef LTP_VERBOSE
										snprintf(ltpstr,LTPSTR_LEN,
												"An RS upper bound (%d) is greater than the block length! (%d) - ditching that dummy!\n",
													rs->ub,theblock->decoded.len);
										ltp_log(LTP_LOG_MAX,ltpstr);
#endif
										break; // ignoring RS!!!
									}

									// kill any timer for the checkpoint for that RS)
									rv=kill_timer(LTP_TIMER_CP,&segstr.val.rs.csn);
									if (rv) {
										snprintf(ltpstr,LTPSTR_LEN,"Can't kill CP timer for RS - csn=0x%x...Ignoring\n",
												ltpc_oct2int_r(&segstr.val.rs.csn));
										ltp_log(LTP_LOG_MAX,ltpstr);
									} 
#ifdef LTP_VERBOSE
										else {
										snprintf(ltpstr,LTPSTR_LEN,"Killed CP timer for RS - csn=0x%x\n",
												ltpc_oct2int_r(&segstr.val.rs.csn));
										ltp_log(LTP_LOG_MAX,ltpstr);
									}
#endif

									// update block's scopes
									for (int jj=0;jj!=rs->cc;jj++) {
										int slb=rs->lb+rs->offsets[jj];
										int sub=slb+rs->lengths[jj];
										rv=addnewscope(theblock,slb,sub,LTP_TX_SCOPES);
										if (rv) {
											snprintf(ltpstr,LTPSTR_LEN,"addnewscope returned %d - ignoring rest of this RS\n",rv);
											ltp_log(LTP_LOG_MAX,ltpstr);
											free(allbuf);
											continue;
										}
										// maybe exiting there is wrong?
									}

									// find the gaps - this is based on the scopes, not the 
									// last rx'd RS
									int ngaps=0;
									int *gapstarts=NULL;
									int startsalloced=0; // just to use FOOREALLOC
									int *gapends=NULL;
									int endsalloced=0; // just to use FOOREALLOC
									int nscopes=theblock->tx_scopes.nscopes;
									if (!nscopes) {
										ngaps=1;
										FOOREALLOC(int,gapstarts,ngaps,startsalloced);
										gapstarts[0]=0;
										FOOREALLOC(int,gapends,ngaps,endsalloced);
										gapends[0]=rs->ub;
									} else {
										// 1st gap 
										if (theblock->tx_scopes.lbs[0]!=0) {
											ngaps++;
											FOOREALLOC(int,gapstarts,ngaps,startsalloced);
											FOOREALLOC(int,gapends,ngaps,endsalloced);
											gapstarts[ngaps-1]=0;
											gapends[ngaps-1]=min(rs->ub,theblock->tx_scopes.lbs[0]);
										}
										// middle gaps
										bool pastub=(theblock->tx_scopes.lbs[0]>=rs->ub); // true if we're further than RS ub
										for (int gg=0;!pastub && gg!=nscopes-1;gg++) {
											int start=theblock->tx_scopes.lbs[gg];
											pastub=(theblock->tx_scopes.lbs[gg]>=rs->ub); // true if we're further than RS ub
											int end=theblock->tx_scopes.ubs[gg];
											if (!pastub && end<theblock->tx_scopes.lbs[gg+1]) {
												ngaps++;
												FOOREALLOC(int,gapstarts,ngaps,startsalloced);
												FOOREALLOC(int,gapends,ngaps,endsalloced);
												gapstarts[ngaps-1]=end;
												gapends[ngaps-1]=min(theblock->tx_scopes.lbs[gg+1],rs->ub);
											}
										}
										// last gap
										int lastgood=theblock->tx_scopes.ubs[nscopes-1];
										if (rs->ub > lastgood) {
											ngaps++;
											FOOREALLOC(int,gapstarts,ngaps,startsalloced);
											FOOREALLOC(int,gapends,ngaps,endsalloced);
											gapstarts[ngaps-1]=lastgood;
											gapends[ngaps-1]=rs->ub;
										}
									}
									if (!ngaps) {
										// can mark this as complete?
										if (rs->ub==theblock->decoded.len) {
											// no gaps and the RS goes as far as the end of the block, then declare victory
											// or, mark the block as complete and then we'll set finishedwith when we've
											// sent the required RA
											theblock->complete=true;
										}
									} else { // there's a gap somewhere
										// quick sanity clause - should only happen due to bugs above!
										int lastgapend=gapends[ngaps-1];
										if (lastgapend>theblock->decoded.len) {
											// bummer something wrong 
#ifdef LTP_VERBOSE
											snprintf(ltpstr,LTPSTR_LEN,
												"A gap upper bound (%d) is greater than the block length! (%d) - ditching that dummy!\n",
													lastgapend,theblock->decoded.len);
											ltp_log(LTP_LOG_MAX,ltpstr);
#endif
											break;
										}
										// make up DS'es for 'em - honouring the L2 MTU though!
										for (int gg=0;gg!=ngaps;gg++) {
											// a.k.a. "mind the gaps" :-)
											int gapsize=gapends[gg]-gapstarts[gg];
											int lc_mtu=latestmtu(theblock,theblock->parent);
											int gapsegs=gapsize/lc_mtu+(gapsize%lc_mtu?1:0);
											for (int nseg=0;nseg!=gapsegs;nseg++) {

				
												int seglen= ( (nseg==(gapsegs-1) && gapsize%lc_mtu)  ? // if we're the last segment and there's a remainder
																	gapsize-(nseg*lc_mtu)  // then the length is that remainder
																	: 
																	lc_mtu); // otherwise its just the nice even segment size
												int offset=gapstarts[gg]+nseg*lc_mtu;
												bool checkpoint=(gg==(ngaps-1) && nseg==(gapsegs-1)); 
												rv=ds_from_block(&theblock->parent->obpeer,theblock,seglen,offset,checkpoint,&rs->rsn);
												if (rv) { free(allbuf); LTPERR(rv); }
											}
										}
									}

									// add a new segment for tx
									NEWSEGMENT
									segments[nsegments].seg.type=LTPC_SEG_RA;
									ltpc_ra *ra=&segments[nsegments].seg.val.ra;
									rv=make_ra_from_rs(&segstr,&segments[nsegments].seg);
									if (rv) { free(allbuf); LTPERR(rv); }
									segments[nsegments].parent=theblock; 
									segments[nsegments].parentid=theblock->id;
									rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
									if (rv) { free(allbuf); LTPERR(rv); }
									ltpaddr to;
									to=theblock->parent->obpeer;
									memcpy(&segments[nsegments].peer,&to,sizeof(to));
									segments[nsegments].peerlen=sizeof(to);
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN,
											"MSG: Queueing an RA for LTP[id:%s,port:%d]\n",
											inet_ntoa(to.sock.sin_addr),to.sock.sin_port);
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									// in SO_LINGER cases, it may be time now to close the thing
									if (!thesock->opts.linger && theblock->complete && thesock->close_called) {
										ltp_close_aux(thesock->sockfd);
									}
							}
							break;

							case LTPC_SEG_RA:
							{
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN, "MSG: Got an RA\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								// kill any timer for the checkpoint for that RS)
								rv=kill_timer(LTP_TIMER_RS,&segstr.val.ra.rsn);
								if (rv) {
									snprintf(ltpstr,LTPSTR_LEN,"Can't kill RS timer for RA - rsn=0x%x...Ignoring\n",
												ltpc_oct2int_r(&segstr.val.ra.rsn));
									ltp_log(LTP_LOG_MAX,ltpstr);
								}
#ifdef LTP_VERBOSE
										else {
										snprintf(ltpstr,LTPSTR_LEN,"Killed RS timer for RA - rsn=0x%x\n",
												ltpc_oct2int_r(&segstr.val.ra.rsn));
										ltp_log(LTP_LOG_MAX,ltpstr);
									}
#endif
								// Maybe we're done now with an inbound block?
								if (theblock->complete && theblock->delivered) {
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN,"Setting finishedwith for block %d (at %d)\n",theblock->id,__LINE__);
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									theblock->finishedwith=true;
								}
								// regardless of status, record facts  - TODO: make this a sockopt
								if (theblock->complete && !theblock->delivered) {
#ifdef LTP_VERBOSE
									snprintf(ltpstr,LTPSTR_LEN,"Setting finishedwithwhendelivered for block %d (at %d)\n",theblock->id,__LINE__);
									ltp_log(LTP_LOG_MAX,ltpstr);
#endif
									theblock->finishedwithwhendelivered=true;
								}
								if (theblock->complete || theblock->redcomplete) {
									// buggy - in the case of partial redness I'm not complete
									// but may be radone, need to have another check that this
									// is the last RA needed (and maybe have to fix sop stuff
									// too)
									theblock->radone=true;
								}
							}
							break;

							case LTPC_SEG_CO:
							{
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN, "MSG: Got a CO\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								// add a new segment for tx
								NEWSEGMENT
								segments[nsegments].seg.type=LTPC_SEG_CAO;
								ltpc_cao *ca=&segments[nsegments].seg.val.cao;
								ltpc_co *c=&segstr.val.co;
	
								rv=make_ca_from_c(&segments[nsegments].seg,&segstr);
								if (rv) { free(allbuf); LTPERR(rv); }
	
								segments[nsegments].parent=theblock; 
								segments[nsegments].parentid=theblock->id;
								rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
								if (rv) { free(allbuf); LTPERR(rv); }
								ltpaddr to;
								to=theblock->parent->ibpeer;
								to.sock.sin_family=AF_LTP;
								memcpy(&segments[nsegments].peer,&to,sizeof(to));
								segments[nsegments].peerlen=sizeof(to);
								rv=init_timer(LTP_TIMER_C,&segments[nsegments]);
								if (rv) { free(allbuf); LTPERR(rv); }
								// remember that we want to cache the CO bytes
								segments[nsegments].r_entry=LTP_CACHE;
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,
												"MSG: Queueing a CAO for LTP[id:%s,port:%d]\n",
												inet_ntoa(to.sock.sin_addr),to.sock.sin_port);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								// stop further forwarding
								// TODO: properly handle ltp-t cases!
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"Set finishedwith=true for block %d at %d\n",theblock->id,__LINE__);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								theblock->finishedwith=true;
								break;
							}

							case LTPC_SEG_CAO:
							{
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN, "MSG: Got a CAO\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								// kill any timer for the checkpoint for that RS)
								rv=kill_timer(LTP_TIMER_C,&segstr.sno);
								if (rv) {
									snprintf(ltpstr,LTPSTR_LEN,"Can't kill C timer for CAO - sno=0x%x...Ignoring\n",
												ltpc_oct2int_r(&segstr.sno));
									ltp_log(LTP_LOG_MAX,ltpstr);
								}
#ifdef LTP_VERBOSE
										else {
										snprintf(ltpstr,LTPSTR_LEN,"Killed C timer for CAO - sno=0x%x\n",
												ltpc_oct2int_r(&segstr.sno));
										ltp_log(LTP_LOG_MAX,ltpstr);
									}
#endif
								break;
							}

							case LTPC_SEG_CR:
							{
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN, "MSG: Got a CR\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								// add a new segment for tx
								NEWSEGMENT
								segments[nsegments].seg.type=LTPC_SEG_CAR;
								// hack hack - depending on CAO/CAR being the same structure really
								ltpc_cao *ca=&segments[nsegments].seg.val.cao;
								ltpc_co *c=&segstr.val.co;
	
								rv=make_ca_from_c(&segments[nsegments].seg,&segstr);
								if (rv) { free(allbuf); LTPERR(rv); }
	
								segments[nsegments].parent=theblock; 
								segments[nsegments].parentid=theblock->id;
								rv=ltp_fixparent(&ltp_state_var,(void**)&theblock->parent,theblock->parentid,LTP_PARENT_SOCKET);
								if (rv) { free(allbuf); LTPERR(rv); }
								ltpaddr to;
								to=theblock->parent->ibpeer;
								to.sock.sin_family=AF_LTP;
								memcpy(&segments[nsegments].peer,&to,sizeof(to));
								segments[nsegments].peerlen=sizeof(to);
								/// @todo should have separate timer for different directions
								rv=init_timer(LTP_TIMER_C,&segments[nsegments]);
								if (rv) { free(allbuf); LTPERR(rv); }
								// remember that we want to cache the CR bytes
								segments[nsegments].r_entry=LTP_CACHE;
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,
												"MSG: Queueing a CAR for LTP[id:%s,port:%d]\n",
												inet_ntoa(to.sock.sin_addr),to.sock.sin_port);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
							}

							case LTPC_SEG_CAR:
							{
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN, "MSG: Got a CAR\n");
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								// kill any timer for the checkpoint for that RS)
								rv=kill_timer(LTP_TIMER_C,&segstr.sno);
								if (rv) {
									snprintf(ltpstr,LTPSTR_LEN,"Can't kill C timer for CAR - sno=0x%x...Ignoring\n",
												ltpc_oct2int_r(&segstr.sno));
									ltp_log(LTP_LOG_MAX,ltpstr);
								}
#ifdef LTP_VERBOSE
										else {
										snprintf(ltpstr,LTPSTR_LEN,"Killed C timer for CAR - sno=0x%x\n",
												ltpc_oct2int_r(&segstr.sno));
										ltp_log(LTP_LOG_MAX,ltpstr);
									}
#endif
								break;
							}

							default:
								// no idea - just ditch it
#ifdef LTP_VERBOSE
								snprintf(ltpstr,LTPSTR_LEN,"Ditching segment of unknown type (%d)\n",packet_type);
								ltp_log(LTP_LOG_MAX,ltpstr);
#endif
								;
							} // switch

							} // if (thesock)
							ltpc_free_seg(&segstr);
						} // not replay
					} // while
				} // done with allbuf (had something for me)
				free(allbuf);
			} // for each peer
			DROPQUTEX
			free(alllen);
			free(l2peers);
			oldpos=newpos; // note: we now no longer care if new buffers arrive - we've got the data
		} 
	} // in Rx mode

	if (l2state==LTP_L2S_ON_RX || l2state==LTP_L2S_ON_BOTH) { // handling timers
		now=ltp_now();
		int ntalkers=LTP_TALKERS_PER_GO;
		ltpaddr	talkers[LTP_TALKERS_PER_GO];
		rv=ltp_cue_whos_talking(now,&ntalkers,talkers);
		if (rv) LTPERR(rv);
		if (ntalkers) {
			ltp_timer_detail *details=ltp_state_var.timers.details;
			int nentries=ltp_state_var.timers.nentries;
			int active_guys[nentries];
			int *agp=active_guys;
			for (int j=0;j!=nentries;j++) active_guys[j]=-1;
			// figure out which timers should be active
			for (int i=0;i!=ntalkers;i++) {
				for (int j=0;j!=nentries;j++) {
					if (details[j].absstart<=now && details[j].active && !details[j].finishedwith && details[j].lastdecremented!=now) {
						if (details[j].type!=LTP_TIMER_INACTIVITY && details[j].type!=LTP_TIMER_COOKIE) {
							rv=ltp_fixparent(&ltp_state_var,(void**)&details[j].segment,details[j].segmentid,LTP_PARENT_SEGMENT);
							if (rv) 
								LTPERR(rv);
							ltp_segment *seg=details[j].segment;
#ifdef LTP_EXTRA_VERBOSE
	STARTTIMEDTRACEBLOCK(x3,true)
							snprintf(ltpstr,LTPSTR_LEN,"Timer %d's segment parent id is %d (ptr: %p)\n",j,seg->parentid,seg->parent);
							ltp_log(LTP_LOG_MAX,ltpstr);
	ENDTIMEDTRACEBLOCK(x3)
#endif
							rv=ltp_fixparent(&ltp_state_var,(void**)&seg->parent,seg->parentid,LTP_PARENT_BLOCK);
							if (rv) 
								LTPERR(rv);
	
							if(!ltpaddr_cmp(&seg->peer, &talkers[i], seg->peerlen)) {  // the other guy's chatting
								*agp++=j;  // include him in the list of active counters
							}
						} else { // its an inactivity timer and so is always "live" - NO!!! Use punctuated time there too!
#ifdef LTP_EXTRA_VERBOSE
	STARTTIMEDTRACEBLOCK(x3,true)
							snprintf(ltpstr,LTPSTR_LEN,"Timer %d is an inactivity timer (%g remaining)\n",
									j,details[j].remaining);
							ltp_log(LTP_LOG_MAX,ltpstr);
	ENDTIMEDTRACEBLOCK(x3)
#endif
							// if the socket's peer is a talker then decrement
							rv=ltp_fixparent(&ltp_state_var,(void**)&details[j].socket,details[j].socketid,LTP_PARENT_SOCKET);
							if (rv) 
								LTPERR(rv);
							ltp_socket_t *ia_sock=details[j].socket;
							if(ia_sock->obpeerlen && !ltpaddr_cmp(&ia_sock->obpeer, &talkers[i], ia_sock->obpeerlen)) {  // the other guy's chatting
								*agp++=j;
							} else if(ia_sock->ibpeerlen && !ltpaddr_cmp(&ia_sock->ibpeer, &talkers[i], ia_sock->ibpeerlen)) {  // the other guy's chatting
								*agp++=j;
							}
						}
					} 
				}
			}
			int agc=agp-active_guys;
			agp=active_guys;
#ifdef LTP_EXTRA_VERBOSE
	STARTTIMEDTRACEBLOCK(x4,true)
			snprintf(ltpstr,LTPSTR_LEN,"%d timers active at elapsed time %d\n",agc,ltp_elapsed());
			ltp_log(LTP_LOG_MAX,ltpstr);
	ENDTIMEDTRACEBLOCK(x4)
#endif
			// now go through these and the active timer and handle required stuff (decrementing and expiry actions)
			for (int j=0;j!=agc;j++,agp++) {
				// max to decrement is 1s
				details[*agp].remaining -= min(1,now-details[*agp].lastdecremented);
				details[*agp].lastdecremented=now;
#ifdef LTP_EXTRA_VERBOSE
	STARTTIMEDTRACEBLOCK(x5,true)
				snprintf(ltpstr,LTPSTR_LEN,"timer %d (of %d) has %g seconds remaining at elapsed time %d\n",j,agc,details[*agp].remaining,ltp_elapsed());
				ltp_log(LTP_LOG_MAX,ltpstr);
	ENDTIMEDTRACEBLOCK(x5)
#endif
				if (details[*agp].remaining <= 0 ) { // expiry!!
					details[*agp].counter--;
#ifdef LTP_VERBOSE
					char ttstr[MAXPATH];
					int thingid;
					switch (details[*agp].type) {
						case LTP_TIMER_CP: snprintf(ttstr,MAXPATH,"checkpoint"); thingid=details[*agp].segmentid; break;
						case LTP_TIMER_RS: snprintf(ttstr,MAXPATH,"report segment"); thingid=details[*agp].segmentid; break;
						case LTP_TIMER_C: snprintf(ttstr,MAXPATH,"cancel"); thingid=details[*agp].segmentid; break;
						case LTP_TIMER_INACTIVITY: snprintf(ttstr,MAXPATH,"inactivity"); thingid=details[*agp].socketid; break;
						case LTP_TIMER_COOKIE: snprintf(ttstr,MAXPATH,"cookie"); thingid=details[*agp].blockid; break;
						default: snprintf(ttstr,MAXPATH,"unknown"); thingid=0;
					}
					snprintf(ltpstr,LTPSTR_LEN,"timer %d (of %d) has expired  at elapsed time %d - timer type: %d(%s) for thing: %d, with %d goes left\n",
									j,agc,ltp_elapsed(),details[*agp].type, ttstr, thingid, details[*agp].counter);
					ltp_log(LTP_LOG_MAX,ltpstr);
#endif
					details[*agp].everexpired=true;
					if (!details[*agp].counter) {
						int sockfd=0;
						if (details[*agp].type==LTP_TIMER_COOKIE) {
							rv=ltp_fixparent(&ltp_state_var,(void**)&details[*agp].block,details[*agp].blockid,LTP_PARENT_BLOCK);
							if (rv) LTPERR(rv);
							details[*agp].block->cookie_grace_expired=true;
						} else if (details[*agp].type==LTP_TIMER_INACTIVITY) {
							sockfd=details[*agp].socketid;
						} else {
							rv=ltp_fixparent(&ltp_state_var,(void**)&details[*agp].segment, details[*agp].segmentid,LTP_PARENT_SEGMENT);
							if (rv) LTPERR(rv);
							rv=ltp_fixparent(&ltp_state_var,(void**)&details[*agp].segment->parent, details[*agp].segment->parentid,LTP_PARENT_BLOCK);
							if (rv) LTPERR(rv);
							sockfd=details[*agp].segment->parent->parentid;
						}
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"timer %d (of %d) expiry closes socket %d at elapsed time %d\n",
								j,agc,sockfd, ltp_elapsed());
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						// zap it in any case - even if nothing else works
						details[*agp].finishedwith=true;
						details[*agp].active=false;
						// close socket, cancel etc.
						ltp_close_aux(sockfd);
					} else {
						// re-tx the parent segment 
#ifdef LTP_VERBOSE
						snprintf(ltpstr,LTPSTR_LEN,"timer %d (of %d) expiry causes replay of segment id %d at elapsed time %d\n",
										j,agc,details[*agp].segment->id,ltp_elapsed());
						ltp_log(LTP_LOG_MAX,ltpstr);
#endif
						// but its not active yet (until the re-Tx actually happens)
						details[*agp].active=false;
						// we'll need to de-ref the parent either way
						rv=ltp_fixparent(&ltp_state_var,(void**)&details[*agp].segment, details[*agp].segmentid,LTP_PARENT_SEGMENT);
						if (rv) LTPERR(rv);
						// try to just reset state - not sure if it'll work though ;-)
						details[*agp].segment->state=LTP_PS_WAITING_SEND;
					}
				}
			}
			if (agc) {
				LTP_STATE_CHANGED;
			}
		}
	}
#ifdef LTP_VERBOSE
	ltp_log_flush();
#endif
	return(LTP_OK);
}

/// TODO: all this ltp_get_sop stuff should be in its own file really. Whenever I actually
/// properly handle the memory stuff then that'll make for loads of code. No time now though.

///< MAX size of a binary bliock id
#define MAXLOCALIDSIZE 4

///< a place to put lower bound scopes
static int sop_lbs[MAXBINFOS][MAXSCOPES];

///< a place to put upper bound scopes
static int sop_ubs[MAXBINFOS][MAXSCOPES];

///< a place to put block infos 
static ltp_binfo sop_binfos[MAXBINFOS];

///< a place to put local block ids
static unsigned char sop_bids[MAXBINFOS][MAXLOCALIDSIZE];

/*!
 * @brief initialise sop storage
 * @return zero or error
 */
static int ltp_sop_init(void)
{
	memset(sop_binfos,0,sizeof(sop_binfos));
	memset(sop_lbs,0,sizeof(sop_lbs));
	memset(sop_ubs,0,sizeof(sop_ubs));
	for (int i=0;i!=MAXBINFOS;i++) {
		sop_binfos[i].lbs=&sop_lbs[i][0];
		sop_binfos[i].ubs=&sop_ubs[i][0];
	}
	return(LTP_OK);
}

/*!
 * @brief get the state of play for a socket
 * @param sock is the socket
 * @param sop is a ptr to the ltp_sop str provided in the call to getsockopt()
 * @return zero or error
 */
static int ltp_get_sop(int sock, ltp_sop *sop)
{
	if (!sop) LTPERR(LTP_BAD_PARAM);
	GETSOCKSESS;
	if (!found) LTPERR(LTP_BAD_PARAM);
	ltp_sop_init(); // ZERO	everything
	sop->nblocks=socksess->nblocks;
	int lastb=min(socksess->nblocks,MAXBINFOS);
	if (socksess->nblocks>MAXBINFOS) {
		snprintf(ltpstr,LTPSTR_LEN,"Socket %d has more (%d) blocks than I can report on (%d), just doing the first %d\n",
				sock,socksess->nblocks,MAXBINFOS,MAXBINFOS);
		ltp_log(LTP_LOG_MAX,ltpstr);
	}
	for (int i=0;i!=lastb;i++) {
		ltp_block *tb=&socksess->blocks[i];
		ltp_binfo *bi=&sop_binfos[i];

		bi->inbound=socksess->inbound;
		bi->outbound=socksess->outbound;
		bi->complete=tb->complete;
		if (tb->remoteidlen) {
			bi->bidlen=tb->remoteidlen;
			bi->bid=tb->remoteid;
			bi->redlen=tb->redtide;
		} else {
			memcpy(sop_bids[i],(void*)&tb->id,sizeof(tb->id));
			bi->bidlen=sizeof(tb->id);
			bi->bid=sop_bids[i];
			bi->redlen=socksess->opts.redlen;
		}
		/// TODO: properly handle bi-directional and relay sockets, not done now;-)
		/// for now I just use whichever is populated
		ltp_scopes *scopes=NULL;
		if (tb->tx_scopes.nscopes) scopes=&tb->tx_scopes;
		if (tb->rx_scopes.nscopes) scopes=&tb->rx_scopes;
		if (tb->fwd_scopes.nscopes) scopes=&tb->fwd_scopes;
		if (scopes) {
			bi->nscopes=scopes->nscopes;
			int lastscope=min(bi->nscopes,MAXSCOPES);
			if (lastscope!=bi->nscopes) {
		 		snprintf(ltpstr,LTPSTR_LEN,"Socket %d, block %d has too many scopes (%d, doing 1st %d)\n",
					sock,tb->id,bi->nscopes,MAXSCOPES);
				ltp_log(LTP_LOG_MAX,ltpstr);
			}
			memcpy(bi->lbs,scopes->lbs,lastscope*sizeof(int));
			memcpy(bi->ubs,scopes->ubs,lastscope*sizeof(int));
		} 
#ifdef LTP_VERBOSE
		 else {
		 	snprintf(ltpstr,LTPSTR_LEN,"Socket %d, block %d has no scopes\n",sock,tb->id);
			ltp_log(LTP_LOG_MAX,ltpstr);
		}
#endif
		if (tb->radone && tb->complete) bi->radone=true;
	}
	sop->infos=sop_binfos;
	return(LTP_OK);
}


/*!
 * @brief really do the [set|get]sockopt() function
 * @param set is true for a setsockopt() call, false for a getsockopt() call
 * @param sock is the socket
 * @param level is the level to be used and MUST be SOL_SOCKET here
 * @param optname is the #define'd name of the option, e.g. LTP_SO_LINGER
 * @param optval is the value of the option
 * @param optlenp is a ptr to the length of the value (modified on output)
 * @return zero or error
 *
 * Hopefully this allows me to write checking code once only.
 */
static int sockopt_aux(bool set, int sock, int level, int optname, void *optval, socklen_t *optlenp)
{
	LTPINIT;
	GRABMUTEX;
	if (!optlenp) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Bad input (%d,%d,%p,%p) parameter to %s\n",
			level,optname,optval,optlenp, set?"ltp_setsockopt":"ltp_getsockopt");
		ltp_log(LTP_LOG_ERR,ltpstr);
#endif
		DROPMUTEX;return(-1);
	}
	int optlen=*optlenp;
	if (!optval || optlen<=0 || (set && optlen>LTP_MAX_OPTLEN) || optname<0 || optname >LTP_MAX_SO) {
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Bad input (%d,%d,%p,%d) parameter to %s\n",
			level,optname,optval,optlen, set?"ltp_setsockopt":"ltp_getsockopt");
		ltp_log(LTP_LOG_ERR,ltpstr);
#endif
		DROPMUTEX;return(-1);
	}
	GETSOCKSESS;
	if (!found) { 
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Unknown socket (%d) input to %s\n",
			sock,set?"ltp_setsockopt":"ltp_getsockopt");
		ltp_log(LTP_LOG_ERR,ltpstr);
#endif
		DROPMUTEX; return(-1); 
	}
	// do stuff
	switch (optname) {

	case LTP_SO_BLOCK:
		if (optlen<sizeof(bool)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_BLOCK\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			bool *ov=(bool*)optval;
			bool theval=!*ov; // NOTE REVERSAL OF SENSE!!!
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_BLOCK to %s for ltp_setsockopt\n",theval?"true":"false");
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			socksess->opts.nonblocking=theval; 
		} else {
			if (optlen<sizeof(socksess->opts.nonblocking)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_BLOCK\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
				bool theval=!socksess->opts.nonblocking ; // NOTE REVERSAL OF SENSE!!!
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_BLOCK as %s for ltp_getsockopt\n",theval?"true":"false");
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&theval,sizeof(theval));
				*optlenp=sizeof(theval);
			}
		}
		break;

	case LTP_SO_LINGER:
		if (optlen!=sizeof(int)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_LINGER\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			int *ov=(int*)optval;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_LINGER to %s for ltp_setsockopt\n",*ov?"true":"false");
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			socksess->opts.linger=*ov;
		} else {
			if (optlen<sizeof(socksess->opts.linger)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_LINGER\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_LINGER as %s for ltp_getsockopt\n",socksess->opts.linger?"true":"false");
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&socksess->opts.linger,sizeof(socksess->opts.linger));
				*optlenp=sizeof(socksess->opts.linger);
			}
		}
		break;

	case LTP_SO_INACTIVITY:
		if (optlen!=sizeof(int)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_ACTIVITY\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			int *ov=(int*)optval;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_ACTIVITY to %d for ltp_setsockopt\n",*ov);
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			socksess->opts.inactivitylimit=*ov;
		} else {
			if (optlen<sizeof(int)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_ACTIVITY\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
				int foo=(int)floor(socksess->opts.inactivitylimit);
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_ACTIVITY as %d for ltp_getsockopt\n",foo);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&foo,sizeof(int));
				*optlenp=sizeof(int);
			}
		}
		break;

	case LTP_SO_COOKIE:
		if (set) {
			int *ov=(int*)optval;
			if (!ov) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Disabling cookies LTP_SO_COOKIE ltp_setsockopt\n");
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				socksess->opts.cookie=0;
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_COOKIE to %d for ltp_setsockopt\n",*ov);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				socksess->opts.cookie=*ov;
				// check for too long or short. Note that sort of makes the MIN the default on the
				// basis that many callers won't set a realistic value, but will just set a nonzero
				// value
				if (*ov > LTP_MAX_COOKIE) {
					snprintf(ltpstr,LTPSTR_LEN,"Asking for overlong (%d) cookie via IOCTL - setting max (%d)\n",
										*ov,LTP_MAX_COOKIE);
					ltp_log(LTP_LOG_ERR,ltpstr);
					socksess->opts.cookie=LTP_MAX_COOKIE;
				}
				if (*ov < LTP_MIN_COOKIE) {
					snprintf(ltpstr,LTPSTR_LEN,"Asking for too short (%d) a cookie via IOCTL - setting min (%d)\n",
										*ov,LTP_MIN_COOKIE);
					ltp_log(LTP_LOG_ERR,ltpstr);
					socksess->opts.cookie=LTP_MIN_COOKIE;
				}
			}
			// TODO: add grace period handling - maybe via a new IOCTL?
		} else {
			if (optlen<sizeof(int)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_COOKIE\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
				int foo=(int)socksess->opts.cookie;
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_COOKIE as %d for ltp_getsockopt\n",foo);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&foo,sizeof(int));
				*optlenp=sizeof(int);
			}
		}
		break;

	case LTP_SO_COOKIE_GRACE:
		if (set) {
			int *ov=(int*)optval;
			if (!ov) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"LTP_SO_COOKIE_GRACE set to zero in ltp_setsockopt!\n");
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				socksess->opts.cookie_grace=0;
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_COOKIE_GRACE to %d in ltp_setsockopt\n",*ov);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				socksess->opts.cookie_grace=*ov;
				// check for too long or short. Note that sort of makes the MIN the default on the
				// basis that many callers won't set a realistic value, but will just set a nonzero
				// value
				if (*ov > LTP_MAX_COOKIE_GRACE) {
					snprintf(ltpstr,LTPSTR_LEN,"Asking for overlong (%d) cookie grace period via IOCTL - setting max (%d)\n",
										*ov,LTP_MAX_COOKIE);
					ltp_log(LTP_LOG_ERR,ltpstr);
					socksess->opts.cookie_grace=LTP_MAX_COOKIE_GRACE;
				}
				if (*ov < LTP_MIN_COOKIE_GRACE) {
					snprintf(ltpstr,LTPSTR_LEN,"Asking for too short (%d) a cookie grace period via IOCTL - setting min (%d)\n",
										*ov,LTP_MIN_COOKIE);
					ltp_log(LTP_LOG_ERR,ltpstr);
					socksess->opts.cookie_grace=LTP_MIN_COOKIE_GRACE;
				}
			}
			// TODO: add grace period handling - maybe via a new IOCTL?
		} else {
			if (optlen<sizeof(int)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_COOKIE_GRACE\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
				int foo=(int)socksess->opts.cookie_grace;
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_COOKIE_GRACE as %d for ltp_getsockopt\n",foo);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&foo,sizeof(int));
				*optlenp=sizeof(int);
			}
		}
		break;

		case LTP_SO_RED:
		if (optlen!=sizeof(int)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_RED\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			int *ov=(int*)optval;
			if (*ov < LTP_ALLRED || *ov >LTP_MAXRED) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Not setting LTP_SO_RED to %d for ltp_setsockopt - out of range (%d,%d)\n",
					*ov,LTP_ALLRED,LTP_MAXRED);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_RED to %d for ltp_setsockopt\n",*ov);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				socksess->opts.redlen=*ov;
			}
		} else {
			if (optlen<sizeof(socksess->opts.redlen)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_RED\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_RED as %d for ltp_getsockopt\n",socksess->opts.redlen);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&socksess->opts.redlen,sizeof(socksess->opts.redlen));
				*optlenp=sizeof(socksess->opts.redlen);
			}
		}
		break;

		case LTP_SO_AUTH:
		if (optlen!=sizeof(ltp_auth_so)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_AUTH\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			ltp_auth_so *ov=(ltp_auth_so*)optval;
			if (ov->ciphersuite < LTP_CS_MAC || ov->ciphersuite >LTP_CS_NULL) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad ciphersuite for LTP_SO_AUTH (%d) given to ltp_setsockopt - out of range (%d,%d)\n",
					ov->ciphersuite,LTP_CS_MAC,LTP_CS_NULL);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_AUTH ciphersuite to %d, keyid to |%s| in ltp_setsockopt\n",
					ov->ciphersuite,octets2str(ov->keyid,ov->keyidlen));
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				socksess->opts.auth=*ov;
			}
		} else {
			if (optlen<sizeof(socksess->opts.auth)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_AUTH\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_AUTH ciphersuite to %d, keyid as |%s| in ltp_setsockopt\n",
					socksess->opts.auth.ciphersuite,octets2str(socksess->opts.auth.keyid,socksess->opts.auth.keyidlen));
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&socksess->opts.auth,sizeof(socksess->opts.auth));
				*optlenp=sizeof(socksess->opts.auth);
			}
		}
		break;

	case LTP_SO_CLIENTONLY:
		if (set) {
			l2desiredstate=LTP_L2S_ON_TX;
			unsigned int lc_mtu=latestmtu(NULL,socksess);
			int rv=ltp_l2_on(&lc_mtu,l2desiredstate);
			if (rv) LTPERR(rv);
			l2state=ltp_l2_get_state();
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_CLIENTONLY - I won't be listening anyway\n");
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
		} else {
			if (optlen<sizeof(l2state)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_CLIENTONLY\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_CLIENTONLY as %d for ltp_getsockopt\n",l2state);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&l2state,sizeof(l2state));
				*optlenp=sizeof(l2state);
			}
		}
		break;

		case LTP_SO_L2MTU:
		if (optlen!=sizeof(int)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_L2MTU\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			int *ov=(int*)optval;
			if (*ov < LTP_MINMTU || *ov >LTP_MAXMTU) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Not setting LTP_SO_L2MTU to %d for ltp_setsockopt - out of range (%d,%d)\n",
					*ov,LTP_MINMTU,LTP_MAXMTU);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_L2MTU to %d for ltp_setsockopt\n",*ov);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				l2mtu=*ov;
				l2mtu_set=true;
				// TODO: set mtu into socket option
				socksess->opts.mtu=l2mtu;
				socksess->opts.mtu_set=true;
			}
		} else {
			if (optlen<sizeof(int)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_L2MTU\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_L2MTU as %d for ltp_getsockopt\n",socksess->opts.redlen);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				// TODO: read mtu from socket option
				unsigned int lc_mtu=l2mtu;
				if (socksess->opts.mtu_set) {
					lc_mtu=socksess->opts.mtu;	
				}
				memcpy(optval,&lc_mtu,sizeof(int));
				*optlenp=sizeof(int);
			}
		}
		break;

	case LTP_SO_T_CFG:
		if (optlen!=sizeof(ltp_t_cfg_so)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_T_CFG\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			ltp_t_cfg_so *ov=(ltp_t_cfg_so*)optval;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_T_CFG via ltp_setsockopt\n");
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			socksess->opts.t_cfg=*ov; // struct copy
			int rv=ltpt_reinit(ov);
			if (rv) {
				snprintf(ltpstr,LTPSTR_LEN,"ltpt_reinit failed - %d\n",rv);
				ltp_log(LTP_LOG_ADV,ltpstr);
				DROPMUTEX;
				return(-1);
			}
			if (ltp_state_var.t_cfg.on) {
				// kick off a thread for forwarding (new) segment from (partly) received blocks
				rv=forwarder_thread_start();
				if (rv) {
					snprintf(ltpstr,LTPSTR_LEN,"forwarder_thread_start failed - %d\n",rv);
					ltp_log(LTP_LOG_ADV,ltpstr);
					DROPMUTEX;
					return(-1);
				}
			}
		} else {
			if (optlen<sizeof(ltp_t_cfg_so)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_T_CFG\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_T_CFG from ltp_getsockopt\n");
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&socksess->opts.t_cfg,sizeof(socksess->opts.t_cfg));
				*optlenp=sizeof(ltp_t_cfg_so);
			}
		}
		break;

	case LTP_SO_T_ROUTE:
		if (set && optlen!=sizeof(ltp_t_route_so)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_T_ROUTE\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			ltp_t_route_so *ov=(ltp_t_route_so*)optval;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_T_ROUTE via ltp_setsockopt\n");
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			int rv=ltpt_set_route(ov);
			if (rv) {
				DROPMUTEX;
				return(-1);
			}
		} else {
			if (optlen<10) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_T_ROUTE\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_T_ROUTE from ltp_getsockopt\n");
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				char *routestr=ltpt_print_routes();
				int cancopy=min(optlen,strlen(routestr)+1);
				memcpy(optval,routestr,cancopy);
				*optlenp=cancopy;
			}
		}
		break;

	case LTP_SO_T_AUTH:
		if (set && optlen!=sizeof(int)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_T_AUTH\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			return(-1);
		}
		if (set) {
			int *ov=(int*)optval;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_T_AUTH via ltp_setsockopt\n");
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			if (ov) {
				socksess->opts.t_cfg.need_e2e_int=true;
			} else {
				socksess->opts.t_cfg.need_e2e_int=false;
			}
		} else {
			if (optlen!=sizeof(int)) {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_T_AUTH\n",
					optlen, set?"ltp_setsockopt":"ltp_getsockopt");
				ltp_log(LTP_LOG_ERR,ltpstr);
#endif
				DROPMUTEX;
				return(-1);
			} else {
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_T_AUTH from ltp_getsockopt\n");
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
				memcpy(optval,&socksess->opts.t_cfg.need_e2e_int,sizeof(int));
				*optlenp=sizeof(int);
			}
		}
		break;

	case LTP_SO_SOP:
		if (optlen<sizeof(ltp_sop)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_SOP\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			LTPERR(-1);
		}
		if (set) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_SOP is not supported\n");
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			return(-1);
		} else {
				ltp_sop *theval=(ltp_sop*)optval;
				*optlenp=sizeof(ltp_sop);
				int rv=ltp_get_sop(sock,theval);
				if (rv) { DROPMUTEX; LTPERR(rv); }
#ifdef LTP_VERBOSE
				snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_SOP with %d for block infos\n",theval->nblocks);
				ltp_log(LTP_LOG_ADV,ltpstr);
#endif
		}
		break;

	case LTP_SO_FORCE:
		if (optlen!=sizeof(ltpaddr)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_FORCE\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			LTPERR(-1);
		}
		if (set) {
			memcpy(&forced_addr,optval,sizeof(ltpaddr));
			forced_addr_set=true;
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Forced LTP_SO_FORCE to %s\n",ltpaddr2str(&forced_addr));
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
		} else {
			memcpy(optval,&forced_addr,sizeof(ltpaddr));
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_FORCE as %s (forced_addr_set=%s)\n",
				ltpaddr2str(&forced_addr),(forced_addr_set?"true":"false"));
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
		}
		break;

	case LTP_SO_ION:
		if (optlen!=sizeof(bool)) {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) size to %s, for LTP_SO_ION\n",
				optlen, set?"ltp_setsockopt":"ltp_getsockopt");
			ltp_log(LTP_LOG_ERR,ltpstr);
#endif
			DROPMUTEX;
			LTPERR(-1);
		}
		if (set) {
			ion_mode=*((bool*)optval);
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Setting LTP_SO_ION to %s\n",(ion_mode?"true":"false"));
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
		} else {
#ifdef LTP_VERBOSE
			snprintf(ltpstr,LTPSTR_LEN,"Returning LTP_SO_ION (as %s)\n",(ion_mode?"true":"false"));
			ltp_log(LTP_LOG_ADV,ltpstr);
#endif
			memcpy(optval,&ion_mode,sizeof(ion_mode));
		}
		break;

	default:
#ifdef LTP_VERBOSE
		snprintf(ltpstr,LTPSTR_LEN,"Bad optname (%d) parameter to %s\n",
			optname, set?"ltp_setsockopt":"ltp_getsockopt");
		ltp_log(LTP_LOG_ERR,ltpstr);
#endif
		DROPMUTEX;
		return(-1);
	}
	DROPMUTEX;
	return(LTP_OK);
}

/*!
 * @brief setsockopt() equivalent
 * @param s is the socket
 * @param level is the level to be used and MUST be SOL_SOCKET here
 * @param optname is the #define'd name of the option, e.g. LTP_SO_LINGER
 * @param optval is the value of the option
 * @param optlen is the length of the value
 * @return zero or error
 */
int ltp_setsockopt(int s, int level, int optname, const void *optval, socklen_t optlen)
{
	return sockopt_aux(true,s,level,optname,(void*)optval,&optlen);
}

/*!
 * @brief getsockopt() equivalent
 * @param s is the socket
 * @param level is the level to be used and MUST be SOL_SOCKET here
 * @param optname is the #define'd name of the option, e.g. LTP_SO_LINGER
 * @param optval is the value of the option
 * @param optlen is the length of the value
 * @return zero or error
 */
int ltp_getsockopt(int s, int level, int optname, void *optval, socklen_t *optlen)
{
	return sockopt_aux(false,s,level,optname,optval,optlen);
}

#ifndef MAXPORT
/// biggest port number allowed - not sure where this is properly so here's fine
#define MAXPORT 65535
#endif


/*!
 * @brief map a string into an LTP socket address 
 * @param str is the string (can be DNS or dot.sep. IP)
 * @param addr is the resuting address
 * @return zero or error
 */
int str2ltpaddr(char *str, ltpaddr *addr)
{
	// note - error handling a but tricky here - can't use ltp_log since we're called from
	// within the getopt() loop, so have to use stderr if necessary
	if (!str) {
		snprintf(ltpstr,LTPSTR_LEN,"Null string ptr given to mapaddr - bad idea - exiting\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(-1);
	}
	if (!addr) {
		snprintf(ltpstr,LTPSTR_LEN,"Null addr ptr given to mapaddr - bad idea - exiting\n");
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(-1);
	}
	if (strlen(str)<=1) { 
		snprintf(ltpstr,LTPSTR_LEN,"String (%s) too short for mapaddr - bad idea - LTPERR\n",str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(-1);
	}
	char *cpstr=ltp_name_trans(str);
	if (cpstr) {
		cpstr=strdup(cpstr);
	} else {
		cpstr=strdup(str);
	}
	hostent	*hent;
	// was a port specified?
	char *pstr=NULL;
	for (int i=strlen(cpstr)-1;i!=0;i--) {
		if (cpstr[i]==':') { 
			pstr=cpstr+i+1;
			cpstr[i]='\0';
			break;
		}
	}
	int port;
	if (pstr) {
		port=atoi(pstr);
	} else {
		port=UDP_DEF_PORT;
	}
	if (port<0 || port>MAXPORT) {
		snprintf(ltpstr,LTPSTR_LEN,"Requested port (%d) out of range (min=1,max=%d), host=%s\n",port,MAXPORT,cpstr);
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(-1);
	}
	hent=gethostbyname(cpstr);
	free(cpstr);
	if (hent==NULL) {
		snprintf(ltpstr,LTPSTR_LEN,"gethostnyname(%s) returned NULL - exiting\n",str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(-1);
	}
	if (hent->h_name) {
		bzero(addr,sizeof(sockaddr_in));
		addr->sock.sin_family=AF_LTP;
		addr->sock.sin_port=port;
		memcpy(&addr->sock.sin_addr.s_addr,hent->h_addr_list[0],hent->h_length);
		return(0);
	} else {
		snprintf(ltpstr,LTPSTR_LEN,"gethostnyname(%s) returned an empty record - exiting\n",str);
		ltp_log(LTP_LOG_MAX,ltpstr);
		LTPERR(-1);
	}
	return(-1); // can't happen
}


/*!
 * @brief set the response from whoami to a pre-cooked one
 * @param ltpaddr *me
 * @return zero or error
 *
 * This is needed when you want to get responses to the LTP
 * segments you send to a peer LTP engine, and where you 
 * don't have any relevant LTP configuration (e.g. schedules
 * that determine who you are.)
 */
int ltp_set_whoiam(ltpaddr *me) 
{
	// just pass this on to the cues handler. (Only reason 
	// for the indirection is to have tidy include files.
	return (ltp_cue_set_whoiam(me));
}


/*!
 * @brief tells a node which other nodes are listening now (so we can Tx to them if there're packets)
 * @param nlisteners on input is the size of the listeners array, and on output is the number of real entries in the array
 * @param listeners is an array (caller allocated) of the LTP addresses of the currently listening nodes
 * @return zero or error
 *
 * This can be called by an application that stores up LTP blocks until a 
 * contact is open, e.g. a bundle protocol agent like DTN2's dtnd.
 */
int ltp_whos_listening_now(int *nlisteners, ltpaddr *listeners)
{
	// just pass this on to the cues handler. (Only reason 
	// for the indirection is to have tidy include files.
	ltp_time_t now=ltp_now();
	return (ltp_cue_whos_listening(now, nlisteners, listeners));
}

